/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.actions;

import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.algorithms.Algorithm.SearchType;
import de.uni_due.inf.ti.raven.algorithms.BisimulationUpToCongruenceInvariantChecking;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * Handles invocation of the bisimulation-up-to-congruence invariant check.
 * @author Christoph Blume
 * @version $Revision: 1.19 $
 *
 */
public class BisimulationUpToCongruenceInvariantCheckingAction extends InvariantCheckingAction {
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new {@code Action} to execute the bisimulation-up-to-congruence invariant check 
	 * for the automaton {@code automaton} and the left- resp. right-hand side {@code leftHandSide} 
	 * and {@code rightHandSide}.
	 * @param type the search type used by the algorithm, either {@code forward} or {@code backward}
	 * @param computeCounterExample if {@code true}, the algorithm will compute a counter-example
	 * @param automaton the automaton used by the algorithm
	 * @param leftHandSide the left-hand side used by the algorithm
	 * @param rightHandSide the right-hand side used by the algorithm
	 */
	public BisimulationUpToCongruenceInvariantCheckingAction(SearchType type, boolean computeCounterExample, 
			Automaton automaton, CospanDecomposition leftHandSide, CospanDecomposition rightHandSide) {
		if(type==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SEARCH_TYPE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(automaton==null) {
			throw new NullPointerException(
					RavenExceptionMessages.AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(leftHandSide==null) {
			throw new NullPointerException(
					RavenExceptionMessages.LHS_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(rightHandSide==null) {
			throw new NullPointerException(
					RavenExceptionMessages.RHS_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		this.algorithmName = "Bisimulation Up To Congruence Invariant Checking";
		this.computeCounterExample = computeCounterExample;
		this.automaton = automaton;
		this.leftHandSide = leftHandSide;
		this.rightHandSide = rightHandSide;
		invariantCheckingAlgorithm = new BisimulationUpToCongruenceInvariantChecking(type, computeCounterExample);
		logger = Logger.getLogger(getClass().getName());
	}
	
	@Override
	public void cancelComputation() {
		invariantCheckingAlgorithm.cancelComputation();
	}

}
