/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.actions;

import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.algorithms.Algorithm.SearchType;
import de.uni_due.inf.ti.raven.algorithms.BisimulationUpToCongruenceLanguageEquivalence;
import de.uni_due.inf.ti.raven.algorithms.BisimulationUpToCongruenceLanguageInclusion;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * Handles invocation of the bisimulation-up-to-congruence language inclusion
 * check.
 * 
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.24 $
 * 
 */
public class BisimulationUpToCongruenceLanguageInclusionAction extends
		LanguageInclusionAction {

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new {@code Action} to execute the bisimulation-up-to-congruence
	 * language inclusion check for the two automata {@code firstAutomaton} and
	 * {@code secondAutomaton}.
	 * @param type forward or backward
	 * @param computeCounterExample if {@code true}, the algorithm will compute 
	 * a counter-example
	 * @param firstAutomaton the first automaton used by the algorithm
	 * @param secondAutomaton the second automaton used by the algorithm
	 * @see BisimulationUpToCongruenceLanguageEquivalence
	 */
	public BisimulationUpToCongruenceLanguageInclusionAction(SearchType type, 
			boolean computeCounterExample, Automaton firstAutomaton, 
			Automaton secondAutomaton) {
		if(type==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SEARCH_TYPE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(firstAutomaton==null) {
			throw new NullPointerException(
					RavenExceptionMessages.FIRST_AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(secondAutomaton==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SECOND_AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		this.algorithmName = "Bisimulation Up To Congruence Language Inclusion";
		this.type = type;
		this.firstAutomaton = firstAutomaton;
		this.secondAutomaton = secondAutomaton;
		this.computeCounterExample = computeCounterExample;

		languageInclusionAlgorithm = new BisimulationUpToCongruenceLanguageInclusion(
				type, computeCounterExample);
		logger = Logger.getLogger(getClass().getName());
	}

	/**
	 * Creates a new {@code Action} to execute the bisimulation-up-to-congruence
	 * language inclusion check for the two automata {@code firstAutomaton} and
	 * {@code secondAutomaton}.
	 * @param type forward or backward
	 * @param firstAutomaton the first automaton used by the algorithm
	 * @param secondAutomaton the second automaton used by the algorithm
	 * 
	 * @see BisimulationUpToCongruenceLanguageEquivalence
	 */
	public BisimulationUpToCongruenceLanguageInclusionAction(SearchType type,
			Automaton firstAutomaton, Automaton secondAutomaton) {
		this(type, true, firstAutomaton, secondAutomaton);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public void cancelComputation() {
		languageInclusionAlgorithm.cancelComputation();
	}

}
