/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.actions;

import java.util.EventObject;

/**
 * An event which indicates that the computation of a(n intermediate) result has been finished. 
 * The event is passed to every {@code ComputationListener} object that registered to receive such 
 * events using the {@code addComputationListener} method.
 * @author Christoph Blume
 * @version $Revision: 1.12 $
 *
 * @see ComputationAction
 * @see ComputationListener
 */
public class ComputationEvent extends EventObject {
	
	private static final long serialVersionUID = 6796467245118414055L;

	/**
	 * Enumeration of event types which can occur during the computation. 
	 * @author Christoph Blume
	 *
	 */
	public static enum EventType {
		/** Indicates that the computation has been started */
		COMPUTATION_STARTED,
		
		/** Indicates that the computation has been advanced */
		COMPUTATION_ADVANCED,
		
		/** Indicates that the computation has been finished */
		COMPUTATION_FINISHED,
		
		/** Indicates that the computation has been cancelled */
		COMPUTATION_CANCELLED;
	}	
	
	private final EventType type;
	private final String name;
	private final int maximum;
	private final int value;
	
	/**
	 * Creates a new ComputationEvent
	 * @param source The object that originated the event
	 * @param type the type of the event
	 * @param name the name of the event
	 * @param maximum the maximum value
	 * @param value the current value
	 */
	public ComputationEvent(Object source, EventType type, String name, int maximum, int value) {
		super(source);
		this.type = type;
		this.name = name;
		this.maximum = maximum;
		this.value = value;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the event type.
	 * @return the event type
	 */
	public final EventType getType() {
		return type;
	}
	
	/**
	 * Returns the name of this event.
	 * @return the name of this event
	 */
	public final String getName() {
		return name;
	}

	/**
	 * Returns the maximum value.
	 * @return the maximum value
	 */
	public final int getMaximum() {
		return maximum;
	}
	
	/**
	 * Returns the current value.
	 * @return the current value
	 */
	public final int getValue() {
		return value;
	}

}
