/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.actions;

import java.util.Arrays;
import java.util.Random;

import nl.uu.cs.treewidth.input.GraphInput.InputData;
import nl.uu.cs.treewidth.input.InputException;
import nl.uu.cs.treewidth.input.RandomGraphGenerator;
import nl.uu.cs.treewidth.ngraph.NGraph;
import de.uni_due.inf.ti.raven.algorithms.LibTWNGraphToGraphAlgorithm;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.Node;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * Creates a random graph.
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.16 $
 *
 */
public class CreateRandomGraphAction extends ComputationAction<Graph> {

	/**
	 * 
	 * @author Christoph Blume
	 * @version $Revision: 1.16 $
	 *
	 */
	private enum Type {
		/** Default Random Graph Generator */
		DEFAULT_GENERATOR,
		
		/** LibTW Random Graph Generator */
		LIB_TW_GENERATOR
	}
	
	private static final String DEFAULT_NODE_NAME = "n";
	private static final String DEFAULT_EDGE_NAME = "e";

	private final Type type;

	private Graph graph;
	private String graphName;
	private int maxEdges;
	private int numberOfNodes;
	private float density;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * 
	 * @param graphname the repository name of the generated graph
	 * @param numberOfNodes the number of nodes of the random graph
	 * @param density the probability density with which an edge is added between two nodes
	 */
	public CreateRandomGraphAction(String graphname, int numberOfNodes, float density){
		if(graphname==null) {
			throw new NullPointerException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(graphname.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		if(numberOfNodes<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.NUMBER_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(density<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.DENSITY_MUST_NOT_BE_LESS_THAN_ZERO_EXCEPTION);
		}
		if(density>1) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.DENSITY_MUST_NOT_BE_GREATER_THAN_ONE_EXCEPTION);
		}
		
		type = Type.DEFAULT_GENERATOR;
		this.graphName = graphname;
		this.numberOfNodes = numberOfNodes;
		this.density = density;
	}
	
	/**
	 * 
	 * @param graphname the repository name of the generated graph
	 * @param maxEdges the maximum number of edges
	 */
	public CreateRandomGraphAction(String graphname, int maxEdges){
		if(graphname==null) {
			throw new NullPointerException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(graphname.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		if(maxEdges<1) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_NUMBER_OF_EDGES_MUST_POSITIVE_EXCEPTION);
		}
		
		type = Type.LIB_TW_GENERATOR;
		
		graph = null;
		this.maxEdges = maxEdges;
		this.graphName = graphname;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public void execute() {
		Random rand = new Random();

		switch (type) {
		case DEFAULT_GENERATOR:
			graph = new Graph(graphName);
			
			//Create all (isolated) nodes
			for(int i=0;i<numberOfNodes;i++) {
				graph.addNode(DEFAULT_NODE_NAME + i);
			}
			
			//Create edges between the nodes (dependent on the given density)
			if(density>0) {
				boolean createClique = (density==1);
				
				for(int i=0;i<numberOfNodes;i++) {
					Node firstNode = graph.getNode(DEFAULT_NODE_NAME + i);
					
					for(int j=i+1;j<numberOfNodes;j++) {
						if(createClique || rand.nextFloat()<density) {
							Node secondNode = graph.getNode(DEFAULT_NODE_NAME + j);
							graph.addEdge(DEFAULT_EDGE_NAME + graph.getEdgeCount(), 
									Arrays.asList(firstNode, secondNode));
						}
					}
				}
			}
			break;
			
		case LIB_TW_GENERATOR:
			int edgeCount = Math.abs(rand.nextInt(maxEdges));
			double circleProb = Math.abs(rand.nextInt(101))/100.0;
			RandomGraphGenerator generator = new RandomGraphGenerator(edgeCount, circleProb);
			try {
				NGraph<InputData> nGraph = generator.get();
				graph = LibTWNGraphToGraphAlgorithm.convert(graphName, nGraph);
			} catch (InputException e) {
				e.printStackTrace();
			}
			break;

		default:
			//Do nothing, i.e. create no graph
			break;
		}		
	}

	@Override
	public Graph getResult() {
		return graph;
	}

	@Override
	public void addComputationListener(ComputationListener l) {
		// Do nothing since the graph generator does not provide any computation information	
	}

	@Override
	public void removeComputationListener(ComputationListener l) {
		// Do nothing since the graph generator does not provide any computation information
	}

	@Override
	public void cancelComputation() {
		// Do nothing (as long as the computation can not be cancelled reasonably)
	}
	
}
