/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.actions;

import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.algorithms.Algorithm.SearchType;
import de.uni_due.inf.ti.raven.algorithms.EmptinessCheckAlgorithm;
import de.uni_due.inf.ti.raven.algorithms.BackwardEmptinessCheck;
import de.uni_due.inf.ti.raven.algorithms.ForwardEmptinessCheck;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * Handles invocation of the antichain-based language inclusion check.
 * @author Christoph Blume
 * @version $Revision: 1.12 $
 *
 */
public class EmptinessCheckAction extends ComputationAction<Boolean> {

	private Automaton automaton;
	private EmptinessCheckAlgorithm emptinessCheckAlgorithm;
	private Logger logger;
	private SearchType type;
	
	private Boolean result;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new {@code Action} to execute the emptiness check for the automaton {@code automaton}.
	 * @param type the search type used by the algorithm, either {@code forward} or {@code backward}
	 * @param automaton the automaton for which the emptiness is checked
	 * 
	 * @see ForwardEmptinessCheck
	 * @see BackwardEmptinessCheck
	 */
	public EmptinessCheckAction(SearchType type, Automaton automaton) {
		if(type==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SEARCH_TYPE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(automaton==null) {
			throw new NullPointerException(
					RavenExceptionMessages.AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		this.type = type;
		this.automaton = automaton;
		
		if (type.equals(SearchType.forward)){
			emptinessCheckAlgorithm = new ForwardEmptinessCheck();
		}
		else {
			emptinessCheckAlgorithm = new BackwardEmptinessCheck();
		}
		
		logger = Logger.getLogger(getClass().getName());
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public void  execute() {
		//Notify that the computation has been started
		fireComputationStarted();
		
		logger.info("Perform emptiness check ("+type+"):");
		result = emptinessCheckAlgorithm.acceptsEmptyLanguage(automaton);

		//Notify that the computation has been finished
		fireComputationFinished();
	}
	
	@Override
	public Boolean getResult() {
		return result;
	}

	@Override
	public void addComputationListener(ComputationListener l) {
		emptinessCheckAlgorithm.addComputationListener(l);
	}

	@Override
	public void removeComputationListener(ComputationListener l) {
		emptinessCheckAlgorithm.removeComputationListener(l);
	}
	
	@Override
	public void cancelComputation() {
		emptinessCheckAlgorithm.cancelComputation();
	}

}
