/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.actions;

import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.algorithms.Algorithm.SearchType;
import de.uni_due.inf.ti.raven.algorithms.InvariantCheckingAlgorithm;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.RuleCounterExample;
import de.uni_due.inf.ti.raven.data.automata.Automaton;

/**
 * Handles invocation of the invariant check.
 * @author Christoph Blume
 * @version $Revision: 1.10 $
 *
 */
public abstract class InvariantCheckingAction extends ComputationAction<Boolean> {

	/**
	 * The name of the algorithm which is used.
	 */
	protected String algorithmName;
	
	/** The type of the algorithm which is started by this Action, either {@code Forward} or 
	 * {@code Backward}
	 * @see SearchType
	 */
	protected SearchType type;
	
	/** Indicates whether a counter-example should be computed (if the result is negative) */
	protected boolean computeCounterExample;
	
	/** The automaton used for the invariant check */
	protected Automaton automaton;

	/** The left hand-side used for the invariant check */
	protected CospanDecomposition leftHandSide;
	
	/** The right hand-side used for the invariant check */
	protected CospanDecomposition rightHandSide;
		
	/** The concrete implementation of language inclusion check to be used */
	protected InvariantCheckingAlgorithm invariantCheckingAlgorithm;
	
	/** The logger used to log additional information during the computation */
	protected Logger logger;

	private Boolean isInvariant;
	private RuleCounterExample counterExample;

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public final void execute() {
		//Notify that the computation has been started
		fireComputationStarted();
		
		logger.info("Compute antichain language inclusion ("+type+") using "+algorithmName+" algorithm");
		
		isInvariant = invariantCheckingAlgorithm.isInvariant(automaton, leftHandSide, rightHandSide);
		if(computeCounterExample && !isInvariant) {
			counterExample = invariantCheckingAlgorithm.getInvariantCounterExample();
		}

		//Notify that the computation has been finished
		fireComputationFinished();
	}
	
	@Override
	public final Boolean getResult() {
		return isInvariant;
	}
	
	/**
	 * Returns the counter example, if the language of the automaton is not 
	 * an invariant for the graph transformation rule, and {@code null} otherwise
	 * @return the counter example (if the invariant check failed) or {@code null}
	 * otherwise
	 */
	public final RuleCounterExample getCounterExample() {
		return counterExample;
	}

	@Override
	public final void addComputationListener(ComputationListener l) {
		invariantCheckingAlgorithm.addComputationListener(l);
	}

	@Override
	public final void removeComputationListener(ComputationListener l) {
		invariantCheckingAlgorithm.removeComputationListener(l);
	}

}
