/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.actions;

import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.algorithms.Algorithm.SearchType;
import de.uni_due.inf.ti.raven.algorithms.LanguageInclusionAlgorithm;
import de.uni_due.inf.ti.raven.data.CospanCounterExample;
import de.uni_due.inf.ti.raven.data.automata.Automaton;

/**
 * Handles invocation of the language inclusion check.
 * @author Christoph Blume
 * @version $Revision: 1.11 $
 *
 */
public abstract class LanguageInclusionAction extends ComputationAction<Boolean> {
	
	/**
	 * The name of the algorithm which is used.
	 */
	protected String algorithmName;

	/** 
	 * The type of the algorithm which is started by this Action, either {@code Forward} or 
	 * {@code Backward}
	 * @see SearchType
	 */
	protected SearchType type;
	
	/** Indicates whether a counter-example should be computed (if the result is negative) */
	protected boolean computeCounterExample;
	
	/** The first automaton used for the language inclusion check */
	protected Automaton firstAutomaton;
	
	/** The second automaton used for the language inclusion check */
	protected Automaton secondAutomaton;
	
	/** The concrete implementation of language inclusion check to be used */
	protected LanguageInclusionAlgorithm languageInclusionAlgorithm;
	
	/** The logger used to log additional information during the computation */
	protected Logger logger;
	
	private boolean isIncluded;
	private CospanCounterExample counterExample;

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public final void execute() {
		//Notify that the computation has been started
		fireComputationStarted();
		
		logger.info("Compute language inclusion ("+type+") using "+algorithmName+" algorithm");
		
		if(firstAutomaton.equals(secondAutomaton)) {
			isIncluded = true;
		}
		else {
			isIncluded = languageInclusionAlgorithm.isIncluded(firstAutomaton, secondAutomaton);
		}

		if(computeCounterExample && !isIncluded){
			counterExample = languageInclusionAlgorithm.getCounterExample();
		}

		//Notify that the computation has been finished
		fireComputationFinished();
	}
	
	@Override
	public final Boolean getResult() {
		return isIncluded;
	}
	
	/**
	 * Returns the counter example, if the language of the first automaton is 
	 * not contained in the language of the second automaton, and {@code null}
	 * otherwise
	 * @return the counter example (if the inclusion check failed) or {@code null}
	 * otherwise
	 */
	public final CospanCounterExample getCounterExample() {
		return counterExample;
	}

	@Override
	public final void addComputationListener(ComputationListener l) {
		languageInclusionAlgorithm.addComputationListener(l);
	}

	@Override
	public final void removeComputationListener(ComputationListener l) {
		languageInclusionAlgorithm.removeComputationListener(l);
	}

}
