/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.actions;

import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.algorithms.LibTWGreedyFillInAlgorithm;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.TreeDecomposition;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * Handles invocation of the Greedy Fill-In algorithm.
 * @author Christoph Blume
 * @version $Revision: 1.14 $
 *
 */
public class LibTWGreedyFillInAction extends ComputationAction<TreeDecomposition> {

	private LibTWGreedyFillInAlgorithm decompositionAlgorithm;
	private Graph graph;
	private Logger logger;
	private TreeDecomposition decomposition;

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new {@code Action} to start the Greedy Fill-In algorithm decomposing the graph 
	 * {@code graph}. 
	 * @param graph the graph which will be decomposed
	 */
	public LibTWGreedyFillInAction(Graph graph) {
		if(graph==null) {
			throw new NullPointerException(
					RavenExceptionMessages.GRAPH_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		this.graph = graph;
		decompositionAlgorithm = new LibTWGreedyFillInAlgorithm();
		logger = Logger.getLogger(getClass().getName());
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public void execute() {
		//Notify that the computation has been started
		fireComputationStarted();
		
		logger.info("Compute greedy fill-in graph decomposition");		
		decomposition = (TreeDecomposition) decompositionAlgorithm.decompose(graph);

		//Notify that the computation has been finished
		fireComputationFinished();
	}

	@Override
	public TreeDecomposition getResult() {
		return decomposition;
	}

	@Override
	public void addComputationListener(ComputationListener l) {
		decompositionAlgorithm.addComputationListener(l);
	}

	@Override
	public void removeComputationListener(ComputationListener l) {
		decompositionAlgorithm.removeComputationListener(l);
	}

	@Override
	public void cancelComputation() {
		decompositionAlgorithm.cancelComputation();
	}

}
