/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.actions;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;

import de.uni_due.inf.ti.raven.data.Cospan;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.data.Signature;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.io.GXLReader;
import de.uni_due.inf.ti.raven.io.OperationFileReader;
import de.uni_due.inf.ti.raven.io.SignatureFileReader;
import de.uni_due.inf.ti.raven.io.AutomatonFileReader;
import de.uni_due.inf.ti.raven.main.RavenMain;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * Handles persistence of data from the repository.
 * 
 * @author Christoph Blume, Laura Steinert, Dennis Nolte
 * @version $Revision: 1.38 $
 * 
 */
public class LoadAction implements IOAction {

	/**
	 * Enumeration of load types indicating the data which is to be loaded.
	 * @author Christoph Blume, Laura Steinert, Dennis Nolte
	 * 
	 */
	public enum LoadType {
		/** The automaton type */
		AUTOMATON,

		/** The cospan GXL type */
		COSPANGXL,

		/** The decomposition type */
		DECOMPOSITION,

		/** The graph GXL type */
		GRAPHGXL,

		/** The signature type */
		SIGNATURE
	}

	private LoadType type;
	private String name;
	private Repository repo;
	private Exception e;
	private String filename;

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * 
	 * @param type the type of data object to be loaded
	 * @param name the name under which the loaded data object can be found in the repository
	 * @param filename the filename of the file which will be loaded
	 * @throws FileNotFoundException thrown if the file specified by the filename can not be found
	 */
	public LoadAction(LoadType type, String name, String filename)
			throws FileNotFoundException {
		if (type==null) {
			throw new NullPointerException(
					RavenExceptionMessages.LOAD_TYPE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if (name==null) {
			throw new NullPointerException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if (name.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		if (filename==null) {
			throw new NullPointerException(
					RavenExceptionMessages.FILENAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if (filename.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.FILENAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		if (!new File(filename).exists()) {
			throw new FileNotFoundException(RavenExceptionMessages.FILE_NOT_FOUND_EXCEPTION);
		}

		repo = Repository.getRepository();
		if (repo.contains(name)) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.DUPLICATE_NAME_EXCEPTION);
		}

		this.type = type;
		this.name = name;
		this.filename = filename;

	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public void execute() {
		CospanDecomposition decomp;

		switch (type) {
		case AUTOMATON:

			Automaton automaton = null;
			AutomatonFileReader reader = new AutomatonFileReader();
			if (!filename.endsWith("."+RavenMain.AUTOMATON_FILE_EXTENSION)) {
				filename += "." + RavenMain.AUTOMATON_FILE_EXTENSION;
			}
			try {
				automaton = reader.readAutomaton(filename);
				Repository.getRepository().put(name, automaton);
				
				// put the signature into the repository if it is not already there
				Signature sig = automaton.getAlphabet();
				Repository repos = Repository.getRepository();
				if (!repos.contains(sig)){
					repos.put(name + "Sig", sig);
				}
			} catch (Exception e) {
				this.e = e;
				e.printStackTrace();
			}

			if (automaton != null) {
				repo.put(name, automaton);
			}

			break;

		case COSPANGXL:
			// Read cospan from the GXL-file and put it into the repository
			try {
				Cospan cospan = GXLReader.readCospan(filename);
				repo.put(name, cospan);
			} catch (IOException e) {
				this.e = e;
			}
			break;
			
		case DECOMPOSITION:
			try {
				decomp = OperationFileReader.readOperations(filename);
				repo.put(name, decomp);
			} catch (Exception e) {
				this.e = e;
			}
			break;

		case GRAPHGXL:
			// Read graph from the GXL-file and put it into the repository
			try {
				Graph graph = GXLReader.readGraph(filename);
				repo.put(name, graph);
			} catch (IOException e) {
				this.e = e;
			}
			break;

		case SIGNATURE:
			try {
				Signature signature = SignatureFileReader.read(filename);
				repo.put(name, signature);
			} catch (Exception e) {
				this.e = e;
			}
			break;

		default:
			break;
		}
	}

	@Override
	public boolean hasExceptionOccured() {
		return (e != null);
	}

	@Override
	public Exception getException() {
		return e;
	}

}
