/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.actions;

import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.algorithms.CospanDecompositionAlgorithm;
import de.uni_due.inf.ti.raven.algorithms.DefaultMembershipAlgorithm;
import de.uni_due.inf.ti.raven.algorithms.GraphDecompositionAlgorithm;
import de.uni_due.inf.ti.raven.algorithms.LibTWGreedyFillInAlgorithm;
import de.uni_due.inf.ti.raven.algorithms.MembershipAlgorithm;
import de.uni_due.inf.ti.raven.data.Cospan;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * Handles invocation of the membership check.
 * @author Christoph Blume
 * @version $Revision: 1.26 $
 *
 */
public class MembershipAction extends ComputationAction<Boolean> {

	private String automatonName;
	private String cospanName;
	private String decompositionName;
	private String graphName;
	private Automaton automaton;
	private Graph graph;
	private Cospan cospan;
	private CospanDecomposition decomposition;
	private CospanDecompositionAlgorithm cospanDecompositionAlgorithm;
	private GraphDecompositionAlgorithm graphDecompositionAlgorithm;
	private MembershipAlgorithm memberAlgorithm;
	private Logger logger;

	private boolean isMember;

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new {@code Action} to execute the membership check for the automaton
	 * {@code automaton} and the cospan decomposition {@code decomposition}, i.e. it will be 
	 * checked whether the cospan decomposition (of an input graph) is accepted by the automaton.
	 * @param automaton the automaton used by the algorithm
	 * @param decomposition the cospan decomposition to be checked
	 */
	public MembershipAction(Automaton automaton, CospanDecomposition decomposition) {
		if(automaton==null) {
			throw new NullPointerException(
					RavenExceptionMessages.AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(decomposition==null) {
			throw new NullPointerException(
					RavenExceptionMessages.DECOMPOSITION_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(!automaton.getAlphabet().contains(decomposition.getSignature())) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.DECOMPOSITION_INCOMPATIBLE_EXCEPTION);
		}

		this.automaton = automaton;
		this.decomposition = decomposition;
		graphDecompositionAlgorithm = new LibTWGreedyFillInAlgorithm();
		memberAlgorithm = new DefaultMembershipAlgorithm();
		logger = Logger.getLogger(getClass().getName());

		Repository repository = Repository.getRepository();
		automatonName = repository.getName(automaton);
		decompositionName = repository.getName(decomposition);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public void execute() {
		//Notify that the computation has been started
		fireComputationStarted();
		
		isCancelled = false;
		if(graph!=null && !isCancelled) {
			logger.info("Compute decomposition of graph '"+graphName+"'");

			decomposition = CospanDecomposition.getCospanDecomposition(
					graphDecompositionAlgorithm.decompose(graph));
			decompositionName = graphName;
		}
		else if(cospan!=null && !isCancelled) {
			logger.info("Compute decomposition of cospan '"+cospanName+"'");

			decomposition = cospanDecompositionAlgorithm.decompose(cospan);
			decompositionName = cospanName;
		}

		logger.info("Check acceptance of decomposition '"+decompositionName+"' for automaton '" +
				automatonName + "'");
		if(!isCancelled) {
			isMember = memberAlgorithm.isMemberOf(automaton, decomposition);
		}
		
		//Notify that the computation has been finished
		fireComputationFinished();
	}

	@Override
	public Boolean getResult() {
		return isMember;
	}

	@Override
	public void addComputationListener(ComputationListener l) {
		graphDecompositionAlgorithm.addComputationListener(l);
		memberAlgorithm.addComputationListener(l);
	}

	@Override
	public void removeComputationListener(ComputationListener l) {
		graphDecompositionAlgorithm.removeComputationListener(l);
		memberAlgorithm.removeComputationListener(l);
	}

	@Override
	public void cancelComputation() {
		isCancelled = true;
		graphDecompositionAlgorithm.cancelComputation();
		memberAlgorithm.cancelComputation();
	}

}
