/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.actions;

import de.uni_due.inf.ti.raven.data.Cospan;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.GraphDecomposition;
import de.uni_due.inf.ti.raven.data.Signature;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.io.AutomatonPrinter;
import de.uni_due.inf.ti.raven.io.BDDPrinter;
import de.uni_due.inf.ti.raven.io.CospanPrinter;
import de.uni_due.inf.ti.raven.io.DecompositionPrinter;
import de.uni_due.inf.ti.raven.io.GraphPrinter;
import de.uni_due.inf.ti.raven.io.SignaturePrinter;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * Handles printout of data from the repository.
 * @author Christoph Blume
 * @version $Revision: 1.20 $
 *
 */
public class PrintAction implements Action {

	private AutomatonPrinter automatonPrinter;
	private BDDPrinter bddPrinter;
	private CospanPrinter cospanPrinter;
	private DecompositionPrinter decompositionPrinter;
	private GraphPrinter graphPrinter;
	private SignaturePrinter signaturePrinter;
	private Automaton automaton;
	private RavenBDD bdd;
	private Cospan cospan;
	private CospanDecomposition cospanDecomposition;
	private Graph graph;
	private GraphDecomposition graphDecomposition;
	private Signature signature;

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates an action which prints an automaton to a given printer.
	 * @param printer the printer to be used
	 * @param automaton the automaton to be printed
	 */
	public PrintAction(AutomatonPrinter printer, Automaton automaton) {
		if(printer==null) {
			throw new NullPointerException(
					RavenExceptionMessages.PRINTER_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(automaton==null) {
			throw new NullPointerException(
					RavenExceptionMessages.AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}

		this.automatonPrinter = printer;
		this.automaton = automaton;
	}

	/**
	 * Creates an action which prints a BDD to a given printer.
	 * @param printer the printer to be used
	 * @param bdd the BDD to be printed
	 */
	public PrintAction(BDDPrinter printer, RavenBDD bdd) {
		if(printer==null) {
			throw new NullPointerException(
					RavenExceptionMessages.PRINTER_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(bdd==null) {
			throw new NullPointerException(
					RavenExceptionMessages.BDD_MUST_NOT_BE_NULL_EXCEPTION);
		}

		this.bddPrinter = printer;
		this.bdd = bdd;
	}
	
	/**
	 * Creates an action which prints a cospan to a given printer.
	 * @param printer the printer to be used
	 * @param cospan the cospan to be printed
	 */
	public PrintAction(CospanPrinter printer, Cospan cospan) {
		if(printer==null) {
			throw new NullPointerException(
					RavenExceptionMessages.PRINTER_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(cospan==null) {
			throw new NullPointerException(
					RavenExceptionMessages.COSPAN_MUST_NOT_BE_NULL_EXCEPTION);
		}

		this.cospanPrinter = printer;
		this.cospan = cospan;
	}

	/**
	 * Creates an action which prints a graph decomposition to a given printer.
	 * @param printer the printer to be used
	 * @param decomposition the graph decomposition to be printed
	 */
	public PrintAction(DecompositionPrinter printer, CospanDecomposition decomposition) {
		if(printer==null) {
			throw new NullPointerException(
					RavenExceptionMessages.PRINTER_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(decomposition==null) {
			throw new NullPointerException(
					RavenExceptionMessages.DECOMPOSITION_MUST_NOT_BE_NULL_EXCEPTION);
		}

		this.decompositionPrinter = printer;
		this.cospanDecomposition = decomposition;
	}

	/**
	 * Creates an action which prints a graph to a given printer.
	 * @param printer the printer to be used
	 * @param graph the graph to be printed
	 */
	public PrintAction(GraphPrinter printer, Graph graph) {
		if(printer==null) {
			throw new NullPointerException(
					RavenExceptionMessages.PRINTER_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(graph==null) {
			throw new NullPointerException(
					RavenExceptionMessages.GRAPH_MUST_NOT_BE_NULL_EXCEPTION);
		}

		this.graphPrinter = printer;
		this.graph = graph;
	}

	/**
	 * Creates an action which prints a graph decomposition to a given printer.
	 * @param printer the printer to be used
	 * @param decomposition the graph decomposition to be printed
	 */
	public PrintAction(DecompositionPrinter printer, GraphDecomposition decomposition) {
		if(printer==null) {
			throw new NullPointerException(
					RavenExceptionMessages.PRINTER_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(decomposition==null) {
			throw new NullPointerException(
					RavenExceptionMessages.DECOMPOSITION_MUST_NOT_BE_NULL_EXCEPTION);
		}

		this.decompositionPrinter = printer;
		this.graphDecomposition = decomposition;
	}

	/**
	 * Creates an action which prints a signature to a given printer.
	 * @param printer the printer to be used
	 * @param signature the signature to be printed
	 */
	public PrintAction(SignaturePrinter printer, Signature signature) {
		if(printer==null) {
			throw new NullPointerException(
					RavenExceptionMessages.PRINTER_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(signature==null) {
			throw new NullPointerException(RavenExceptionMessages.SIGNATURE_MUST_NOT_BE_NULL_EXCEPTION);
		}

		this.signaturePrinter = printer;
		this.signature = signature;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public void execute() {
		if(automaton!=null) {
			automatonPrinter.printAutomaton(automaton);
		}
		else if(bdd!=null) {
			bddPrinter.print(bdd);
		}
		else if(cospan!=null) {
			cospanPrinter.print(cospan);
		}
		else if(cospanDecomposition!=null) {
			decompositionPrinter.print(cospanDecomposition);
		}
		else if(graph!=null) {
			graphPrinter.print(graph);
		}
		else if(graphDecomposition!=null) {
			decompositionPrinter.print(graphDecomposition);
		}
		else if(signature!=null) {
			signaturePrinter.print(signature);
		}
	}

}
