/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.actions;

import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * Handles the computation of the reachables states of an automaton.
 * @author Christoph Blume
 * @version $Revision: 1.15 $
 *
 */
public class ReachabilityAction extends ComputationAction<RavenBDD> {

	private Automaton automaton;
	private CospanDecomposition decomposition;
	private RavenBDD reachableStates;

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates an action which computes all states for the automaton {@code automaton} which are 
	 * reachable by processing the cospan decomposition {@code decomposition}. The states, which are 
	 * represented by a BDD, are stored in the repository with the name {@code name}.
	 * @param automaton the automaton to be used
	 * @param decomposition the cospan decomposition to be used
	 */
	public ReachabilityAction(Automaton automaton, CospanDecomposition 
			decomposition) {
		if(automaton==null) {
			throw new NullPointerException(
					RavenExceptionMessages.AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(decomposition==null) {
			throw new NullPointerException(
					RavenExceptionMessages.DECOMPOSITION_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		this.automaton = automaton;
		this.decomposition = decomposition;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public void execute() {
		//Notify that the computation has been started
		fireComputationStarted();

		reachableStates = automaton.getReachableStates(decomposition);

		//Notify that the computation has been finished
		fireComputationFinished();
	}

	@Override
	public RavenBDD getResult() {
		return reachableStates;
	}

	@Override
	public void cancelComputation() {
		// Do nothing (as long as the computation can not be cancelled reasonably
	}

}
