/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.actions;

import java.io.FileNotFoundException;
import java.io.IOException;

import de.uni_due.inf.ti.raven.data.Cospan;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.Signature;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.io.AutomatonWriter;
import de.uni_due.inf.ti.raven.io.CospanDecompositionFileWriter;
import de.uni_due.inf.ti.raven.io.CospanWriter;
import de.uni_due.inf.ti.raven.io.GXLWriter;
import de.uni_due.inf.ti.raven.io.GraphWriter;
import de.uni_due.inf.ti.raven.io.PlainTextWriter;
import de.uni_due.inf.ti.raven.io.SVGWriter;
import de.uni_due.inf.ti.raven.io.SignatureFilePrinter;
import de.uni_due.inf.ti.raven.io.AutomatonFileWriter;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * Handles persistence of data from the repository.
 * @author Christoph Blume, Laura Steinert, Dennis Nolte
 * @version $Revision: 1.35 $
 *
 */
public class SaveAction implements IOAction {

	/**
	 * Enumeration of persistence types usable for automata. 
	 * @author Christoph Blume
	 *
	 */
	public enum AutomatonPersistenceType {
		/** Serialized automaton persistence type */
		SERIALIZED_FORMAT,

		/** Plain (textual) automaton persistence type */
		PLAIN_FORMAT
	}
	
	/**
	 * Enumeration of persistence types usable for graphs.
	 * @author Christoph Blume
	 *
	 */
	public enum GraphPersistenceType {
		/** GXL graph persistence type */
		GXL_FORMAT,

		/** SVG graph persistence type */
		SVG_FORMAT
	}

	private AutomatonWriter automatonWriter;
	private CospanWriter cospanWriter;
	private GraphWriter graphWriter;
	private Graph graph;
	private Cospan cospan;
	private CospanDecomposition decomposition;
	private Signature signature;
	private Automaton automaton;
	private String filename;
	private IOException e;

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new save action which saves the automaton {@code automaton} to the file named by
	 * {@code filename}.
	 * @param type the type indicates which printer is used to save the automaton
	 * @param automaton the automaton to be saved
	 * @param filename the name of the file to which the automaton will be saved
	 */
	public SaveAction(AutomatonPersistenceType type, Automaton automaton, String filename) {
		if(type==null) {
			throw new NullPointerException(
					RavenExceptionMessages.PERSISTENCE_TYPE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(automaton==null) {
			throw new NullPointerException(
					RavenExceptionMessages.AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(filename==null) {
			throw new NullPointerException(
					RavenExceptionMessages.FILENAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(filename.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.FILENAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}

		this.automaton = automaton;
		this. filename = filename;
		switch (type) {
		case SERIALIZED_FORMAT:
			automatonWriter = AutomatonFileWriter.getWriter();
			break;

		case PLAIN_FORMAT:
			automatonWriter = PlainTextWriter.getWriter();
			break;
		}
	}

	/**
	 * Creates a new save action which saves the graph {@code graph} as a gxl file named by
	 * {@code filename}.
	 * @param type the type indicates which printer is used to save the graph
	 * @param graph the graph to be saved
	 * @param filename the name of the file to which the graph will be saved (as a GXL)
	 * @throws FileNotFoundException thrown if the file given by the filename can not be found
	 */
	public SaveAction(GraphPersistenceType type, Graph graph, String filename) throws FileNotFoundException {
		if(graph==null) {
			throw new NullPointerException(RavenExceptionMessages.GRAPH_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(filename==null) {
			throw new NullPointerException(
					RavenExceptionMessages.FILENAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(filename.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.FILENAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}

		this.graph = graph;
		this.filename = filename;
		switch (type) {
		case GXL_FORMAT:
			graphWriter = GXLWriter.getWriter();
			break;

		case SVG_FORMAT:
			graphWriter = new SVGWriter();
			break;
		}
	}

	/**
	 * Creates a new save action which saves the cospan {@code cospan} as a cos file named by
	 * {@code filename}.
	 * @param cospan the cospan to be saved
	 * @param filename the name of the file to which the cospan will be saved (as a GXL)
	 * @throws FileNotFoundException thrown if the file given by the filename can not be found
	 */
	public SaveAction(Cospan cospan, String filename) throws FileNotFoundException {
		if(cospan==null) {
			throw new NullPointerException(RavenExceptionMessages.COSPAN_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(filename==null) {
			throw new NullPointerException(
					RavenExceptionMessages.FILENAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(filename.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.FILENAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		
		this.cospan = cospan;
		this.filename = filename;
		cospanWriter = GXLWriter.getWriter();
	}
	
	/**
	 * Creates a new save action which saves the signature {@code signature} to the file 
	 * named by {@code filename}.
	 * @param signature the signature to be saved
	 * @param filename the name of the file to which the signature will be saved
	 */
	public SaveAction(Signature signature, String filename) {
		if(signature==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SIGNATURE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(filename==null) {
			throw new NullPointerException(
					RavenExceptionMessages.FILENAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(filename.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.FILENAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}

		this.signature = signature;
		this.filename = filename;
	}

	/**
	 * Creates a new save action which saves the decomposition {@code decomp} to the file 
	 * named by {@code filename}.
	 * @param decomposition the decomposition to be saved
	 * @param filename the name of the file to which the signature will be saved
	 */
	public SaveAction(CospanDecomposition decomposition, String filename) {
		if(decomposition==null) {
			throw new NullPointerException(
					RavenExceptionMessages.DECOMPOSITION_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(filename==null) {
			throw new NullPointerException(
					RavenExceptionMessages.FILENAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(filename.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.FILENAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		
		this.decomposition = decomposition;
		this.filename = filename;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public void execute() {
		try {
			if (automatonWriter!=null) {
					automatonWriter.write(automaton, filename);
			}
			else if (cospanWriter!=null) {
				cospanWriter.write(cospan, filename);
			}
			else if (decomposition!=null){
				//TODO Encapsulate the CospanDecompositionWriter
				CospanDecompositionFileWriter.writeOperations(filename, decomposition);
			}
			else if (graphWriter!=null) {
				graphWriter.write(graph, filename);
			}
			else if (signature!=null){
				try {
					//TODO Encapsulate the CospanDecompositionWriter
					SignatureFilePrinter.writeSignature(filename, signature);
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		} catch (IOException e) {
			this.e = e;
		}
	}

	@Override
	public boolean hasExceptionOccured() {
		return (e != null);
	}

	@Override
	public Exception getException() {
		return e;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/



}
