/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.actions;

import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.algorithms.DefaultCospanDecompositionAlgorithm;
import de.uni_due.inf.ti.raven.data.Cospan;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.automata.AutomatonFactory;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * Handles creation of a new product automaton.
 * @author Christoph Blume
 * @version $Revision: 1.4 $
 *
 */
public class ShiftAutomatonAction extends ComputationAction<Automaton> {

	private Automaton automaton;
	private CospanDecomposition decomposition;
	private AutomatonFactory factory;
	private Logger logger;

	private Automaton shiftedAutomaton;

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates an action which constructs the product automaton for two given automata.
	 * @param automaton the first automaton used for the product automaton
	 * @param decomposition the second automaton used for the product automaton
	 */
	public ShiftAutomatonAction(Automaton automaton, CospanDecomposition decomposition) {
		if(automaton==null) {
			throw new NullPointerException(
					RavenExceptionMessages.AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(decomposition==null) {
			throw new NullPointerException(
					RavenExceptionMessages.DECOMPOSITION_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(!automaton.getAlphabet().contains(decomposition.getSignature())) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ALPHABETS_MUST_BE_EQUAL_EXCEPTION);
		}
		if(automaton.getMaximumInterface()<decomposition.getWidth()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_INTERFACE_SIZES_MUST_NOT_BE_DIFFERENT_EXCEPTION);
		}
		if(automaton.getInnerInterface()!=decomposition.getInnerInterfaceSize()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZES_MUST_NOT_BE_DIFFERENT_EXCEPTION);
		}

		this.automaton = automaton;
		this.decomposition = decomposition;
		
		//Initialize factory
		factory = AutomatonFactory.getFactory();
		//Initialize logger
		logger = Logger.getLogger(getClass().getName());
	}
	
	/**
	 * Creates an action which constructs the product automaton for two given automata.
	 * @param automaton the first automaton used for the product automaton
	 * @param cospan the second automaton used for the product automaton
	 */
	public ShiftAutomatonAction(Automaton automaton, Cospan cospan) {
		this(automaton, new DefaultCospanDecompositionAlgorithm().decompose(cospan));
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public void execute() {
		//Log entering 
		logger.entering(getClass().getName(), "execute");

		//Notify that the computation has been started
		fireComputationStarted();
		//Create new product automaton
		shiftedAutomaton = AutomatonFactory.shiftAutomaton(automaton, decomposition);
		//Notify that the computation has been finished
		fireComputationFinished();

		//Log exiting
		logger.exiting(getClass().getName(), "execute");
	}
	
	@Override
	public Automaton getResult() {
		return shiftedAutomaton;
	}

	@Override
	public void addComputationListener(ComputationListener l) {
		factory.addComputationListener(l);
	}
	
	@Override
	public void removeComputationListener(ComputationListener l) {
		factory.removeComputationListener(l);
	}
	
	@Override
	public void cancelComputation() {
		// Do nothing (as long as the computation can not be cancelled reasonably
	}

}
