/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.actions;

import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.algorithms.Algorithm.SearchType;
import de.uni_due.inf.ti.raven.algorithms.BackwardSimulationAlgorithm;
import de.uni_due.inf.ti.raven.algorithms.ForwardSimulationAlgorithm;
import de.uni_due.inf.ti.raven.algorithms.SimulationAlgorithm;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * Handles computation of the simulation relation for automata.
 * @author Christoph Blume
 * @version $Revision: 1.5 $
 *
 */
public class SimulationComputationAction extends ComputationAction<RavenBDD> {

	private Automaton firstAutomaton;
	private Automaton secondAutomaton;
	private SimulationAlgorithm simulationAlgorithm;
	private Logger logger;

	private RavenBDD simulation;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new {@code Action} to compute the simulation relation for the automata 
	 * {@code firstAutomaton} and {@code secondAutomaton}
	 * @param type the search type used by the algorithm, either {@code forward} or {@code backward}
	 * @param firstAutomaton the first automaton for which the simulation relation is computed
	 * @param secondAutomaton the second automaton for which the simulation relation is computed
	 */
	public SimulationComputationAction(SearchType type, Automaton firstAutomaton, 
			Automaton secondAutomaton) {
		if(type==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SEARCH_TYPE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(firstAutomaton==null) {
			throw new NullPointerException(
					RavenExceptionMessages.FIRST_AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		this.firstAutomaton = firstAutomaton;
		this.secondAutomaton = secondAutomaton;

		if (type.equals(SearchType.forward)){
			simulationAlgorithm = new ForwardSimulationAlgorithm();
		}
		else {
			simulationAlgorithm = new BackwardSimulationAlgorithm();	
		}
		logger = Logger.getLogger(getClass().getName());
	}

	/**
	 * Creates a new {@code Action} to compute the simulation relation for the automata 
	 * {@code firstAutomaton} and {@code secondAutomaton}
	 * @param firstAutomaton the first automaton for which the simulation relation is computed
	 * @param secondAutomaton the second automaton for which the simulation relation is computed
	 */
	public SimulationComputationAction(Automaton firstAutomaton, 
			Automaton secondAutomaton) {
		this(SearchType.forward, firstAutomaton, secondAutomaton);
	}
	
	/**
	 * Creates a new {@code Action} to compute the simulation relation for the automaton 
	 * {@code automaton}
	 * @param type the search type used by the algorithm, either {@code forward} or {@code backward}
	 * @param automaton the automaton for which the simulation relation is computed
	 */
	public SimulationComputationAction(SearchType type, Automaton automaton) {
		this(type, automaton, null);
	}
	
	/**
	 * Creates a new {@code Action} to compute the simulation relation for the automaton 
	 * {@code automaton}
	 * @param automaton the automaton for which the simulation relation is computed
	 */
	public SimulationComputationAction(Automaton automaton) {
		this(SearchType.forward, automaton);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public void execute() {
		//Notify that the computation has been started
		fireComputationStarted();
		
		logger.info("Compute simulation relation");

		if(secondAutomaton==null) {
			simulation = simulationAlgorithm.getSimulation(firstAutomaton);
		}
		else {
			simulation = simulationAlgorithm.getSimulation(firstAutomaton, secondAutomaton);
		}

		//Notify that the computation has been finished
		fireComputationFinished();
	}
	
	@Override
	public RavenBDD getResult() {
		return simulation;
	}

	@Override
	public void addComputationListener(ComputationListener l) {
		simulationAlgorithm.addComputationListener(l);
	}

	@Override
	public void removeComputationListener(ComputationListener l) {
		simulationAlgorithm.removeComputationListener(l);
	}

	@Override
	public void cancelComputation() {
		simulationAlgorithm.cancelComputation();
	}

}
