/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.actions;

import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.data.Signature;
import de.uni_due.inf.ti.raven.data.automata.AutomatonFactory;
import de.uni_due.inf.ti.raven.data.automata.VertexCountingAutomaton;
import de.uni_due.inf.ti.raven.encodings.DefaultVertexCountingEncoding;
import de.uni_due.inf.ti.raven.encodings.VertexCountingEncoding;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * Handles creation of a new vertex-counting automaton.
 * @author Christoph Blume
 * @version $Revision: 1.15 $
 *
 */
public class VertexCountingAutomatonAction extends ComputationAction<VertexCountingAutomaton> {

	private Signature signature;
	private AutomatonFactory factory;
	private VertexCountingEncoding encoding;
	private int remainder;
	private Logger logger;

	private VertexCountingAutomaton automaton;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new {@code Action} to create a new vertex-counting automaton with maximum 
	 * interface size {@code maximumInterface} over the signature {@code signature}, which accepts
	 * all graphs whose number of vertices is equal to {@code remainder} modulo {@code divisor} 
	 * using the default dominating set encoding.
	 * @param maximumInterface the maximum interface size
	 * @param innerInterface size of the inner interface
	 * @param outerInterface size of the outer interface
	 * @param signature the signature of the automaton
	 * @param remainder the remainder for the number of edges
	 * @param divisor the divisor modulo which edges are counted
	 * @see #VertexCountingAutomatonAction(int, int, int, Signature, int, int, String)
	 * @see AutomatonFactory
	 * @see VertexCountingAutomaton
	 * @see VertexCountingEncoding
	 * @see DefaultVertexCountingEncoding
	 */
	public VertexCountingAutomatonAction(int maximumInterface, int innerInterface, 
			int outerInterface, Signature signature, int divisor, int remainder) {
		if(maximumInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(outerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(outerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(signature==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SIGNATURE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(divisor<1) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.DIVISOR_MUST_BE_POSITIVE_EXCEPTION);
		}
		if(remainder<0 || remainder>=divisor) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_REMAINDER_EXCEPTION);
		}

		this.signature = signature;
		this.remainder = remainder;
		
		//Initialize factory
		factory = AutomatonFactory.getFactory();
		encoding = VertexCountingEncoding.getDefaultEncoding(maximumInterface, 
				innerInterface, outerInterface, divisor);
		//Initialize logger
		logger = Logger.getLogger(getClass().getName());
	}
	
	/**
	 * Creates a new {@code Action} to create a new vertex-counting automaton with maximum 
	 * interface size {@code maximumInterface} over the signature {@code signature}, which accepts
	 * all graphs whose number of vertices is equal to {@code remainder} modulo {@code divisor} 
	 * using the edge-counting encoding named by 
	 * {@code encodingName}.
	 * @param maximumInterface the maximum interface size
	 * @param innerInterface size of the inner interface
	 * @param outerInterface size of the outer interface
	 * @param signature the signature of the automaton
	 * @param remainder the remainder for the number of edges
	 * @param divisor the divisor modulo which edges are counted
	 * @param encodingName the name of the dominating set encoding
	 * @see #VertexCountingAutomatonAction(int, int, int, Signature, int, int)
	 * @see AutomatonFactory
	 * @see VertexCountingAutomaton
	 * @see VertexCountingEncoding
	 * @see DefaultVertexCountingEncoding
	 */
	public VertexCountingAutomatonAction(int maximumInterface, int innerInterface, 
			int outerInterface, Signature signature, int divisor, int remainder, 
			String encodingName) {
		if(maximumInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(outerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(outerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(signature==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SIGNATURE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(divisor<1) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.DIVISOR_MUST_BE_POSITIVE_EXCEPTION);
		}
		if(remainder<0 || remainder>=divisor) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_REMAINDER_EXCEPTION);
		}
		if(encodingName==null) {
			throw new NullPointerException(
					RavenExceptionMessages.ENCODING_NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(encodingName.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ENCODING_NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		
		this.signature = signature;
		//Initialize factory
		factory = AutomatonFactory.getFactory();
		encoding = VertexCountingEncoding.getEncoding(maximumInterface, innerInterface, outerInterface, divisor, encodingName);
		//Initialize logger
		logger = Logger.getLogger(getClass().getName());
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public void execute() {
		//Log entering 
		logger.entering(getClass().getName(), "execute");

		//Notify that the computation has been started
		fireComputationStarted();
		//Create new colorability automaton
		automaton = factory.createVertexCountingAutomaton(encoding, signature, remainder);
		//Notify that the computation has been finished
		fireComputationFinished();

		//Log exiting
		logger.exiting(getClass().getName(), "execute");
	}

	@Override
	public VertexCountingAutomaton getResult() {
		return automaton;
	}

	@Override
	public void addComputationListener(ComputationListener l) {
		factory.addComputationListener(l);
	}
	
	@Override
	public void removeComputationListener(ComputationListener l) {
		factory.removeComputationListener(l);
	}

	@Override
	public void cancelComputation() {
		// Do nothing (as long as the computation can not be cancelled reasonably
	}

}
