/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.algorithms;

import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.actions.ComputationListener;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.RuleCounterExample;
import de.uni_due.inf.ti.raven.data.automata.AutomatonFactory;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class implements the two antichain algorithms for checking invariants. The algorithms 
 * are based on the antichain algorithms for language inclusion by Henzinger et. al. More 
 * details about the invariant checking can be found in 
 * "Efficient Symbolic Implementation of Graph Automata with Applications to Invariant Checking"
 * <a href="http://link.springer.com/chapter/10.1007%2F978-3-642-33654-6_18">Springerlink</a>. The
 * antichain language inclusion algorithms can be found in "Antichains: A New Algorithm for Checking 
 * Universality of Finite Automata" 
 * <a href="http://link.springer.com/chapter/10.1007%2F11817963_5">Springerlink</a>.
 * 
 * @author Christoph Blume
 * @version $Revision: 1.23 $
 *
 */
public final class AntichainInvariantChecking implements InvariantCheckingAlgorithm {

	private AntichainLanguageInclusion languageInclusionAlgorithm;
	private final boolean computerCounterExample;
	private RuleCounterExample counterExample;
	private Logger logger;


	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new instance of the antichain-based invariant checking algorithm.
	 * @param type the search type of this instance
	 * @param computeCounterExample if {@code true} a counterexample will be computed 
	 * if the invariant check fails
	 * 
	 * @see Algorithm.SearchType
	 */
	public AntichainInvariantChecking(SearchType type, boolean computeCounterExample) {
		if(type==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SEARCH_TYPE_MUST_NOT_BE_NULL_EXCEPTION);
		}

		this.computerCounterExample = computeCounterExample;
		languageInclusionAlgorithm = new AntichainLanguageInclusion(type, computeCounterExample);

		// Register logger
		logger = Logger.getLogger(getClass().getName());
	}

	/**
	 * Creates a new instance of the antichain-based invariant checking algorithm.
	 * @param type the search type of this instance
	 * 
	 * @see Algorithm.SearchType
	 */
	public AntichainInvariantChecking(SearchType type) {
		this(type, false);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public void addComputationListener(ComputationListener l) {
		languageInclusionAlgorithm.addComputationListener(l);
	}

	@Override
	public void removeComputationListener(ComputationListener l) {
		languageInclusionAlgorithm.removeComputationListener(l);
	}

	@Override
	public boolean isInvariant(Automaton automaton,
			CospanDecomposition leftHandSide, CospanDecomposition rightHandSide) {
		if(automaton==null) {
			throw new NullPointerException(
					RavenExceptionMessages.AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(leftHandSide==null) {
			throw new NullPointerException(
					RavenExceptionMessages.LHS_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(rightHandSide==null) {
			throw new NullPointerException(
					RavenExceptionMessages.RHS_MUST_NOT_BE_NULL_EXCEPTION);
		}

		// Log entering
		logger.entering(getClass().getName(), "isInvariant", new Object[] {automaton,
			leftHandSide, rightHandSide});

		boolean isInvariant = false;
		Automaton leftHandSideAutomaton = AutomatonFactory.shiftAutomaton(automaton, leftHandSide);
		Automaton rightHandSideAutomaton = AutomatonFactory.shiftAutomaton(automaton, rightHandSide);

		isInvariant = languageInclusionAlgorithm.isIncluded(leftHandSideAutomaton, rightHandSideAutomaton);
		if(computerCounterExample && !isInvariant) {
			counterExample = new RuleCounterExample(leftHandSide, rightHandSide,
					languageInclusionAlgorithm.getCounterExample().getCospan());
		}

		// Log exiting
		logger.exiting(getClass().getName(), "isInvariant", isInvariant);

		return isInvariant;
	}

	@Override
	public RuleCounterExample getInvariantCounterExample() {
		return counterExample;
	}

	@Override
	public void cancelComputation() {
		languageInclusionAlgorithm.cancelComputation();
	}

}
