/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.algorithms;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.actions.ComputationEvent;
import de.uni_due.inf.ti.raven.actions.ComputationEvent.EventType;
import de.uni_due.inf.ti.raven.actions.ComputationListener;
import de.uni_due.inf.ti.raven.data.CospanCounterExample;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDFactory.BDDOperator;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDPairing;
import de.uni_due.inf.ti.raven.main.RavenMain;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class implements the two antichain algorithms deciding the language 
 * inclusion problem for (finite) automata. The algorithms are based on the work
 * of Henzinger et. al. More details about the algorithms can be found in 
 * "Antichains: A New Algorithm for Checking Universality of Finite Automata"
 * <a href="http://www.springerlink.com/content/250l02n76315n786/fulltext.pdf">Springerlink</a>.
 * 
 * @author Christoph Blume
 * @author Dominik Engelke
 * @version $Revision: 1.57 $
 *
 */
public class AntichainLanguageInclusion implements LanguageInclusionAlgorithm {

	/**
	 *
	 */
	private class BDDPair {

		private RavenBDD first;
		private RavenBDD second;
		private List<Operation> list;

		/*------------------------------------------------------------------------------------------*/
		/*		Constructors																		*/
		/*------------------------------------------------------------------------------------------*/

		public BDDPair(RavenBDD first, RavenBDD second, List<Operation> list) {
			this.first = first;
			this.second = second;
			this.list = list;
		}

		public BDDPair(RavenBDD first, RavenBDD second) {
			this(first, second, null);
		}

		/*------------------------------------------------------------------------------------------*/
		/*		Public Methods																		*/
		/*------------------------------------------------------------------------------------------*/

		@Override
		public boolean equals(Object o) {
			if(o==null) {
				return false;
			}
			else if(o==this) {
				return true;
			}
			else if(o.getClass().equals(getClass())) {
				BDDPair pair = (BDDPair) o;
				return first.equals(pair.first) && second.equals(pair.second);
			}
			else {
				return false;
			}
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((first == null) ? 0 : first.hashCode());
			result = prime * result + ((second == null) ? 0 : second.hashCode());
			return result;
		}

		@Override
		public String toString() {
			return "("+firstAutomaton.printStates(first)+", "+secondAutomaton.printStates(second)+", "+list+")";
		}

	}

	private final SearchType type;
	private final boolean computeCounterExample;
	private final Logger logger;

	private boolean verboseMode;
	private boolean isCancelled;
	private Automaton firstAutomaton;
	private Automaton secondAutomaton;
	private RavenBDDPairing pairing;
	private RavenBDDPairing currentSuccessorPairing;
	private List<ComputationListener> listener;
	private CospanCounterExample counterExample;

	int count = 0;

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new instance of the antichain-based language inclusion algorithm.
	 * @param type the search type of this instance
	 * @param computeCounterExample if {@code true} a counterexample will be computed 
	 * if the language inclusion check fails
	 * 
	 * @see Algorithm.SearchType
	 */
	public AntichainLanguageInclusion(SearchType type, boolean computeCounterExample) {
		if(type==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SEARCH_TYPE_MUST_NOT_BE_NULL_EXCEPTION);
		}

		this.type = type;
		this.computeCounterExample = computeCounterExample;
		listener = new LinkedList<>();

		// Register logger
		logger = Logger.getLogger(getClass().getName());
	}

	/**
	 * Creates a new instance of the antichain-based language inclusion algorithm.
	 * @param type the search type of this instance
	 * 
	 * @see Algorithm.SearchType
	 */
	public AntichainLanguageInclusion(SearchType type) {
		this(type, false);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public void addComputationListener(ComputationListener l) {
		listener.add(l);
	}

	@Override
	public void removeComputationListener(ComputationListener l) {
		listener.remove(l);
	}

	@Override
	public boolean isIncluded(Automaton first, Automaton second) {
		if(first==null) {
			throw new NullPointerException(
					RavenExceptionMessages.FIRST_AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(second==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SECOND_AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(first.getMaximumInterface()!=second.getMaximumInterface()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_INTERFACE_SIZES_MUST_NOT_BE_DIFFERENT_EXCEPTION);
		}
		if(first.getInnerInterface()!=second.getInnerInterface()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZES_MUST_NOT_BE_DIFFERENT_EXCEPTION);
		}
		if(first.getOuterInterface()!=second.getOuterInterface()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZES_MUST_NOT_BE_DIFFERENT_EXCEPTION);
		}

		// Log entering
		logger.entering(getClass().getName(), "isIncluded", new Object[] {first, second});
		fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_STARTED, "", 0, 0));

		isCancelled = false;
		this.firstAutomaton = first;
		this.secondAutomaton = second;
		verboseMode = Boolean.valueOf(RavenMain.getProperty(RavenMain.PROPERTY_VERBOSE));
		boolean result = false;

		switch (type) {
		case forward:
			result = performForwardSearch();
			break;

		case backward:
			pairing = Automaton.getInterfacePairing(first, second);
			currentSuccessorPairing = RavenBDDPairing.makePair();
			currentSuccessorPairing.set(secondAutomaton.getEncoding().getCurrentStateDomain(), 
					secondAutomaton.getEncoding().getSuccessorStateDomain());
			result = performBackwardSearch();
			break;

		default:
			throw new RuntimeException(RavenExceptionMessages.ILLEGAL_SEARCH_TYPE_EXCEPTION);
		}

		if(isCancelled) {
			fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_CANCELLED, "", 0, 0));
		}
		else {
			fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_FINISHED, "", 0, 0));
		}

		// Log exiting
		logger.exiting(getClass().getName(), "isIncluded", counterExample);
		return result;
	}

	@Override
	public CospanCounterExample getCounterExample() {
		return counterExample;
	}

	@Override
	public void cancelComputation() {
		//		isCancelled = true;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns {@code true} if and only if the language of the automaton {@code firstAutomaton} 
	 * is a subset of the language of the automaton {@code secondAutomaton}. The computation is 
	 * performed via a forward search based on the antichain algorithm by Henzinger et. al.
	 * @return {@code true} if the first language is included in the second language, 
	 * {@code false} otherwise
	 */
	private boolean performForwardSearch() {
		//Initialize sets F, Frontier and Target
		RavenBDD vars = firstAutomaton.getEncoding().getCurrentStateDomain().getDomainBDD();
		Set<BDDPair> F = new HashSet<>();
		Set<BDDPair> Frontier = new HashSet<>();

		// If the first automaton is in an accepting state and the second automaton is not,
		// return false since the empty word is a counter example
		if (firstAutomaton.isAcceptingStateSet(firstAutomaton .getInitialStates()) && 
				!secondAutomaton.isAcceptingStateSet(secondAutomaton.getInitialStates())) {
			if (computeCounterExample) {
				counterExample = new CospanCounterExample(new CospanDecomposition(
						firstAutomaton.getInnerInterface()));
			}
			return false;
		}

		// Create the initial state pairs
		for (RavenBDD bdd: firstAutomaton.getInitialStates().getSatAssignments(vars)) {
			BDDPair fPair = new BDDPair(bdd, secondAutomaton.getInitialStates());
			BDDPair frontierPair = new BDDPair(bdd, secondAutomaton.getInitialStates());

			if(computeCounterExample) {
				fPair.list = new LinkedList<>();
				frontierPair.list = new LinkedList<>();
			}

			F.add(fPair);
			Frontier.add(frontierPair);
		}

		//The Target-set is the set of pairs such that the first component is a final state 
		//of the first automaton and the set of all non-final states of the second automaton
		Set<BDDPair> target = new HashSet<>();
		for (RavenBDD bdd: firstAutomaton.getFinalStates().getSatAssignments(vars)) {
			target.add(new BDDPair(bdd, secondAutomaton.getNonFinalStates()));
		}

		int i = 0;
		while (!Frontier.isEmpty() && !isCancelled) {
			if(verboseMode) {
				i++;
				logger.info(LANGUAGE_INCLUSION_ITERATION_STEP+i);
			}

			Set<BDDPair> tempFrontier = new HashSet<>();
			Set<BDDPair> tempF = new HashSet<>();

			Set<BDDPair> results = fPost(Frontier);
			int maxRounds = results.size();
			int counter =0;

			//Compute new frontier
			for(BDDPair pair: results) {
				fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED, 
						"Forward Search", maxRounds, ++counter));

				boolean isIncomparable = true;
				//add the new pair if and only if it is minimal (according to the set F)
				for(BDDPair fPair: F) {
					if(fPair.first.equals(pair.first)) {
						if(isSubset(fPair.second, pair.second)) {
							isIncomparable = false;
							break;
						}
					}
				}

				if(isIncomparable) {
					//Check whether the current pair is a counterexample
					for(BDDPair targetPair: target) {
						if(pair.first.equals(targetPair.first)) {
							//If a counterexample can be found, return false
							if(isSubset(pair.second, targetPair.second)) {
								if(computeCounterExample) {
									CospanDecomposition cospan = new CospanDecomposition(
											firstAutomaton.getInnerInterface(), pair.list);
									counterExample = new CospanCounterExample(cospan);
								}
								return false;
							}
						}
					}
					//Add the pair to the new frontier (and prepare the new F)
					tempFrontier.add(pair);
					tempF.add(pair);
				}
			}
			Frontier = tempFrontier;

			//Compute new F
			for (BDDPair pair: F) {
				boolean isIncomparable = true;

				for (BDDPair frontierPair: Frontier) {
					if(pair.first.equals(frontierPair.first) && 
							isSubset(frontierPair.second, pair.second)) {
						isIncomparable = false;
						pair.first.free();
						pair.second.free();
						break;
					}
				}

				if(isIncomparable) {
					tempF.add(pair);
				}
			}
			F = tempF;
		}

		return true;
	}

	/**
	 * Returns {@code true} if and only if the language of the automaton {@code firstAutomaton} 
	 * is a subset of the language of the automaton {@code secondAutomaton}. The computation is 
	 * performed via a backward search based on the antichain algorithm by Henzinger et. al.
	 * @return {@code true} if the first language is included in the second language, 
	 * {@code false} otherwise
	 */
	private boolean performBackwardSearch() {
		//Initialize sets F, Frontier and Target
		RavenBDD vars = firstAutomaton.getEncoding().getCurrentStateDomain().getDomainBDD();
		Set<BDDPair> F = new HashSet<>();
		Set<BDDPair> Frontier = new HashSet<>();
		for (RavenBDD bdd: firstAutomaton.getFinalStates().getSatAssignments(vars)) {
			BDDPair fPair = new BDDPair(bdd, secondAutomaton.getNonFinalStates());
			BDDPair frontierPair = new BDDPair(bdd, secondAutomaton.getNonFinalStates());

			if(computeCounterExample) {
				fPair.list = new LinkedList<>();
				frontierPair.list = new LinkedList<>();
			}

			F.add(fPair);
			Frontier.add(frontierPair);
		}

		//The Target-set is the set of pairs such that the first component is a final state 
		//of the first automaton and the set of all non-final states of the second automaton
		Set<BDDPair> target = new HashSet<>();

		for (RavenBDD state: firstAutomaton.getInitialStates().getSatAssignments(vars)) {
			target.add(new BDDPair(state, secondAutomaton.getInitialStates()));
		}
		
		int i = 0;
		while (!Frontier.isEmpty() && !isCancelled) {
			if(verboseMode) {
				i++;
				logger.info(LANGUAGE_INCLUSION_ITERATION_STEP+i);
			}

			Set<BDDPair> tempFrontier = new HashSet<>();
			Set<BDDPair> tempF = new HashSet<>();

			Set<BDDPair> results = CPre(Frontier);
			int maxRounds = results.size();
			int counter = 0;

			//Compute new frontier
			for(BDDPair pair: results) {
				fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED, 
						"Backward Search", maxRounds, ++counter));

				boolean isIncomparable = true;
				//add the new pair if and only if it is minimal (according to the set F)
				for(BDDPair fPair: F) {
					if(fPair.first.equals(pair.first)) {
						if(isSubset(pair.second, fPair.second)) {
							isIncomparable = false;
							break;
						}
					}
				}

				if(isIncomparable) {
					//Check whether the current pair is a counterexample
					for(BDDPair targetPair: target) {
						if(pair.first.equals(targetPair.first)) {
							//If a counterexample can be found, return false
							if(isSubset(targetPair.second, pair.second)) {
								if(computeCounterExample) {
									CospanDecomposition cospan = new CospanDecomposition(
											firstAutomaton.getInnerInterface(), pair.list);
									counterExample = new CospanCounterExample(cospan);
								}
								return false;
							}
						}
					}
					//Add the pair to the new frontier (and prepare the new F)
					tempFrontier.add(pair);
					tempF.add(pair);
				}
				else {
					pair.first.free();
					pair.second.free();
				}
			}
			Frontier = tempFrontier;

			//Compute new F
			for (BDDPair pair: F) {
				boolean isIncomparable = true;

				for (BDDPair frontierPair: Frontier) {
					if(frontierPair.first.equals(pair.first) && 
							isSubset(pair.second, frontierPair.second)) {
						isIncomparable = false;
						pair.first.free();
						pair.second.free();
						break;
					}
				}

				if(isIncomparable) {
					tempF.add(pair);
				}
			}
			F = tempF;
		}

		return true;
	}

	/**
	 * Computes the CPre-function for some antichain {@code antichain}. The resulting set
	 * contains all pairs of "successors" which are maximum according to the antichain order.
	 * For more details about the CPre-function and the antichain order see the paper 
	 * "Antichains: A New Algorithm for Checking Universality of Finite Automata"
	 * <a href="http://www.springerlink.com/content/250l02n76315n786/fulltext.pdf">Springerlink</a>.
	 * 
	 * @param antichain the antichain containing the "predecessor pairs"
	 * @return the antichain containing the maximum of all "successor pairs"
	 */
	private Set<BDDPair> CPre(Set<BDDPair> antichain) {
		Set<BDDPair> result = new HashSet<>();
		RavenBDD firstcurrentStateDomain = firstAutomaton.getEncoding().getCurrentStateDomain().getDomainBDD();
		RavenBDD secondSuccessorDomain = secondAutomaton.getEncoding().getSuccessorStateDomain().getDomainBDD();
		int rounds = firstAutomaton.getAlphabet().size() * antichain.size();
		int counter = 0;

		for (Operation letter: firstAutomaton.getAlphabet()) {
			for (BDDPair pair: antichain) {
				if(isCancelled) {
					return result;
				}

				//Get the first automaton's predecessors
				RavenBDD predecessorsFirst = firstAutomaton.getPredecessorStates(pair.first, letter);

				//Get the second automaton's predecessors
				RavenBDD predecessorsSecond = secondAutomaton.getTransitionRelation(letter);
				predecessorsSecond = predecessorsSecond.applyAll(
						pair.second.replace(currentSuccessorPairing), BDDOperator.imp, 
						secondSuccessorDomain);
				predecessorsSecond = predecessorsSecond.and(secondAutomaton.getStates());

				//If the first automaton can perform a "backward" step
				if (!predecessorsFirst.isZero()) {
					for (RavenBDD state: predecessorsFirst.getSatAssignments(firstcurrentStateDomain)) {
						RavenBDD interfaceSizeBDD = firstAutomaton.getCurrentInterfaceSize(state);
						interfaceSizeBDD = interfaceSizeBDD.replace(pairing);
						BDDPair predecessorPair = new BDDPair(state.clone(), 
								predecessorsSecond.clone().and(interfaceSizeBDD));

						//If enabled update counter example
						if(computeCounterExample) {
							predecessorPair.list = new LinkedList<>();
							predecessorPair.list.add(letter);
							predecessorPair.list.addAll(pair.list);
						}

						result = addMaximumElement(result, predecessorPair);
						count++;
					}
				}

				//Free variables which are not used any longer
				predecessorsFirst.free();
				predecessorsSecond.free();
				fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED, 
						"CPre", rounds, ++counter));
			}
		}
		//Free variables which are not used any longer
		firstcurrentStateDomain.free();
		secondSuccessorDomain.free();

		return result;
	}

	/**
	 * Computes the Post-function for some (dual) antichain {@code antichain}. The resulting 
	 * set contains all pairs of "successors" which are minimum according to the (dual) 
	 * antichain order. For more details about the Post-function and the dual antichain order 
	 * see the paper "Antichains: A New Algorithm for Checking Universality of Finite Automata"
	 * <a href="http://www.springerlink.com/content/250l02n76315n786/fulltext.pdf">Springerlink</a>.
	 * 
	 * @param antichain the antichain containing the "predecessor pairs"
	 * @return the antichain containing the minimum of all "successor pairs"
	 */
	private Set<BDDPair> fPost(Set<BDDPair> antichain) {
		Set<BDDPair> result = new HashSet<>();
		RavenBDD vars = firstAutomaton.getEncoding().getCurrentStateDomain().getDomainBDD();
		int rounds = firstAutomaton.getAlphabet().size() * antichain.size();
		int counter = 0;

		for (Operation letter: firstAutomaton.getAlphabet()) {
			for (BDDPair pair: antichain) {			
				if(isCancelled) {
					return result;
				}

				//Get the automata's successors
				RavenBDD successorsFirst = firstAutomaton.getSuccessorStates(pair.first, letter);
				RavenBDD successorsSecond = secondAutomaton.getSuccessorStates(pair.second, letter);

				if (!successorsFirst.isZero()) {
					for (RavenBDD state: successorsFirst.getSatAssignments(vars)) {
						BDDPair successorPair = new BDDPair(state.clone(), successorsSecond.clone());

						//If enabled update counter example
						if(computeCounterExample) {
							successorPair.list = new LinkedList<>(pair.list);
							successorPair.list.add(letter);
						}

						result = addMinimumElement(result, successorPair);
						count++;
					}
				}

				//Free variables which are not used any longer
				successorsFirst.free();
				successorsSecond.free();
				fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED, 
						"fPost", rounds, ++counter));
			}
		}
		//Free variables which are not used any longer
		vars.free();

		return result;
	}

	/**
	 * Adds the pair {@code newPair} to the antichain {@code antichain} and computes
	 * the maximum of all pairs of the extended antichain.
	 * @param antichain the antichain in which the new pair will be added
	 * @param newPair the new pair which will be added to the antichain
	 * @return the antichain which contains only maximum elements 
	 */
	private static Set<BDDPair> addMaximumElement(Set<BDDPair> antichain, BDDPair newPair) {
		Set<BDDPair> result = new HashSet<>();

		for (BDDPair antichainPair: antichain) {
			if(antichainPair.first.equals(newPair.first)) {
				if(isSubset(newPair.second, antichainPair.second)) {
					newPair.first.free();
					newPair.second.free();
					return antichain;
				}

				if(!isSubset(antichainPair.second, newPair.second)) {
					result.add(antichainPair);
				}
				else {
					antichainPair.first.free();
					antichainPair.second.free();
				}
			} else {
				result.add(antichainPair);
			}
		}

		result.add(newPair);
		return result;
	}

	/**
	 * Adds the pair {@code newPair} to the (dual) antichain {@code antichain} and 
	 * computes the minimum of all pairs of the extended (dual) antichain.
	 * @param antichain the (dual) antichain in which the new pair will be added
	 * @param newPair the new pair which will be added to the (dual) antichain
	 * @return the (dual) antichain which contains only minimum elements
	 */
	private static Set<BDDPair> addMinimumElement(Set<BDDPair> antichain, BDDPair newPair) {
		Set<BDDPair> result = new HashSet<>();

		for (BDDPair antichainPair: antichain) {
			if(antichainPair.first.equals(newPair.first)) {
				if(isSubset(antichainPair.second, newPair.second)) {
					newPair.first.free();
					newPair.second.free();
					return antichain;
				}

				if(!isSubset(newPair.second, antichainPair.second)) {
					result.add(antichainPair);
				}
				else {
					antichainPair.first.free();
					antichainPair.second.free();
				}
			} else {
				result.add(antichainPair);
			}
		}

		result.add(newPair);
		return result;
	}

	/**
	 * Returns {@code true} if and only if the first set (represented by the BDD) 
	 * {@code firstSet} is a subset of the second set (represented by the BDD) 
	 * {@code secondSet} 
	 * @param firstSet the first set to compare
	 * @param secondSet the second set to compare
	 * @return {@code true} if the first set is a subset of the second set, 
	 * {@code false} otherwise
	 */
	private static boolean isSubset(RavenBDD firstSet, RavenBDD secondSet) {
		RavenBDD tempBDD = firstSet.clone().imp(secondSet.clone());
		boolean result = tempBDD.isOne();
		tempBDD.free();

		return result;
	}

	/**
	 * Fires an computation event notifying all listeners.
	 * @param e the computation event
	 */
	private void fireComputationEvent(ComputationEvent e) {
		for(ComputationListener l: listener) {
			l.computationAdvanded(e);
		}
	}

}
