/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.algorithms;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.actions.ComputationEvent;
import de.uni_due.inf.ti.raven.actions.ComputationEvent.EventType;
import de.uni_due.inf.ti.raven.actions.ComputationListener;
import de.uni_due.inf.ti.raven.data.CospanCounterExample;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.Operation.Connect;
import de.uni_due.inf.ti.raven.data.Operation.Fuse;
import de.uni_due.inf.ti.raven.data.Operation.Perm;
import de.uni_due.inf.ti.raven.data.Operation.Res;
import de.uni_due.inf.ti.raven.data.Operation.Trans;
import de.uni_due.inf.ti.raven.data.Operation.Vertex;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.encodings.BDDEncoding.StateType;
import de.uni_due.inf.ti.raven.main.RavenMain;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class implements the two antichain algorithms deciding the universality 
 * problem for (finite) automata. The algorithms are based on the work of 
 * Henzinger et. al. More details about the algorithms can be found in 
 * "Antichains: A New Algorithm for Checking Universality of Finite Automata"
 * <a href="http://www.springerlink.com/content/250l02n76315n786/fulltext.pdf">Springerlink</a>. 
 * 
 * @author Christoph Blume
 * @author Dominik Engelke
 * @version $Revision: 1.45 $
 *
 */
public final class AntichainUniversality implements UniversalityAlgorithm {

	/**
	 * A pair of BDDs. The first component represents a single state (of the first automaton),
	 * whereas the second component represents a set of states (of the second automaton).
	 * @author Christoph Blume
	 * @author Dominik Engelke
	 * @version $Revision: 1.45 $
	 *
	 */
	private class BDDPair {

		private RavenBDD bdd;
		private int size;
		private List<Operation> list;

		/*------------------------------------------------------------------------------------------*/
		/*		Constructors																		*/
		/*------------------------------------------------------------------------------------------*/
		
		public BDDPair(RavenBDD first, int size, List<Operation> list) {
			this.bdd = first;
			this.size = size;
			this.list = list;
		}

		public BDDPair(RavenBDD first, int size) {
			this(first, size, null);
		}

		/*------------------------------------------------------------------------------------------*/
		/*		Public Methods																		*/
		/*------------------------------------------------------------------------------------------*/
		
		@Override
		public boolean equals(Object o) {
			if(o==null) {
				return false;
			}
			else if(o==this) {
				return true;
			}
			else if(o.getClass().equals(getClass())) {
				BDDPair pair = (BDDPair) o;
				return bdd.equals(pair.bdd);
			}

			return false;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			return size * prime + ((bdd == null) ? 0 : bdd.hashCode());
		}

	}

	private final SearchType type;
	private final boolean computeCounterExample; 
	private final Logger logger;

	private boolean verboseMode;
	private boolean isCancelled;
	private Automaton automaton;
	private List<ComputationListener> listener;
	private CospanCounterExample counterExample;

	int count = 0;

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new instance of the antichain-based universality algorithm.
	 * @param type the search type of this instance
	 * @param computeCounterExample if {@code true} a counterexample will be computed 
	 * if the input language is not universal
	 * 
	 * @see Algorithm.SearchType
	 */
	public AntichainUniversality(SearchType type, boolean computeCounterExample) {
		if(type==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SEARCH_TYPE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		this.type = type;
		this.computeCounterExample = computeCounterExample;
		listener = new LinkedList<>();

		// Register logger
		logger = Logger.getLogger(getClass().getName());
	}

	/**
	 * Creates a new instance of the antichain-based universality algorithm.
	 * @param type the search type of this instance
	 * 
	 * @see Algorithm.SearchType
	 */
	public AntichainUniversality(SearchType type) {
		this(type, false);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public void addComputationListener(ComputationListener l) {
		listener.add(l);
	}

	@Override
	public void removeComputationListener(ComputationListener l) {
		listener.remove(l);
	}

	@Override
	public boolean isUniversal(Automaton automaton) {
		if(automaton==null) {
			throw new NullPointerException(
					RavenExceptionMessages.AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		// Log entering
		logger.entering(getClass().getName(), "isUniversal", automaton);
		verboseMode = Boolean.valueOf(RavenMain.getProperty(RavenMain.PROPERTY_VERBOSE));
		fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_STARTED, "", 0, 0));
		
		isCancelled = false;
		this.automaton = automaton;
		boolean result;
		
		switch (type) {
		case forward:
			result = performForwardSearch();
			break;

		case backward:
			result = performBackwardSearch();
			break;

		default:
			throw new RuntimeException(RavenExceptionMessages.ILLEGAL_SEARCH_TYPE_EXCEPTION);
		}

		if(isCancelled) {
			fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_CANCELLED, "", 0, 0));
		}
		else {
			fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_FINISHED, "", 0, 0));
		}
		
		// Log exiting
		logger.exiting(getClass().getName(), "isUniversal", counterExample);
		return result;
	}

	@Override
	public CospanCounterExample getCounterExample() {
		return counterExample;
	}
	
	@Override
	public void cancelComputation() {
		isCancelled = true;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns {@code true} if and only if the language of the automaton {@code automaton} 
	 * is universal. The computation is performed via a forward search based on the antichain 
	 * algorithm by Henzinger et. al. 
	 * @return {@code true} if the language of the automaton is universal,  {@code false} otherwise
	 */
	private boolean performForwardSearch() {
		//Initialize sets F, Frontier and Target
		BDDPair fInitialPair = new BDDPair(automaton.getInitialStates(),0);
		BDDPair frontierInitialPair = new BDDPair(automaton.getInitialStates(),0);
		if(computeCounterExample) {
			fInitialPair.list = new LinkedList<>();
			frontierInitialPair.list =	new LinkedList<>();
		}

		Set<BDDPair> F = new HashSet<>();
		Set<BDDPair> Frontier = new HashSet<>();
		F.add(fInitialPair);
		Frontier.add(frontierInitialPair);
		
		//The Target-set is the set of pairs such that the BDD component is a subset of the 
		//non-final states of the 0-interface of the automaton
		Set<BDDPair> target = new HashSet<>();
		target.add(new BDDPair(automaton.getNonFinalStates(),0));

		int i = 0;
		while (!Frontier.isEmpty() && !isCancelled) {
			if(verboseMode) {
				i++;
				logger.info(UNIVERSALITY_ITERATION_STEP+i);
			}

			Set<BDDPair> tempFrontier = new HashSet<>();
			Set<BDDPair> tempF = new HashSet<>();
			
			Set<BDDPair> results = fPost(Frontier);
			int maxRounds = results.size();
			int counter =0;
			
			//Compute new frontier
			for(BDDPair frontierPair: results) {
				fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED, 
						"Forward Search", maxRounds, ++counter));
				boolean isIncomparable = true;
				//add the new pair if and only if it is minimal (according to the set F)
				for(BDDPair fPair: F) {
					if(frontierPair.size==fPair.size && 
							isSubset(fPair.bdd, frontierPair.bdd)) {
						isIncomparable = false;
						break;
					}
				}
				
				if(isIncomparable) {
					//Check whether the current pair is a counterexample
					for(BDDPair targetPair: target) {
						//If a counterexample can be found, return false
						if(frontierPair.size==targetPair.size && 
								isSubset(frontierPair.bdd, targetPair.bdd)) {
							counterExample = new CospanCounterExample(
									new CospanDecomposition(frontierPair.list));
							return false;
						}
					}
					//Add the pair to the new frontier (and prepare the new F)
					tempFrontier.add(frontierPair);
					tempF.add(frontierPair);
				}
			}
			Frontier = tempFrontier;

			//Compute new F
			for (BDDPair fPair: F) {
				boolean isIncomparable = true;

				for (BDDPair frontierPair: Frontier) {
					if(frontierPair.size==fPair.size && 
							isSubset(frontierPair.bdd, fPair.bdd)) {
						isIncomparable = false;
						fPair.bdd.free();
						break;
					}
				}

				if(isIncomparable) {
					tempF.add(fPair);
				}
			}
			F = tempF;
		}

		return true;
	}

	/**
	 * Returns {@code true} if and only if the language of the automaton {@code automaton} 
	 * is universal. The computation is performed via a backward search based on the antichain 
	 * algorithm by Henzinger et. al. 
	 * @return {@code true} if the language of the automaton is universal,  {@code false} otherwise
	 */
	private boolean performBackwardSearch() {
		//Initialize sets F, Frontier and Target
		BDDPair fInitialPair = new BDDPair(automaton.getNonFinalStates(),0);
		BDDPair frontierInitialPair = new BDDPair(automaton.getNonFinalStates(),0);
		if(computeCounterExample) {
			fInitialPair.list = new LinkedList<>();
			frontierInitialPair.list =	new LinkedList<>();
		}

		Set<BDDPair> F = new HashSet<>();
		Set<BDDPair> Frontier = new HashSet<>();
		F.add(fInitialPair);
		Frontier.add(frontierInitialPair);

		//The Target-set is the set of pairs such that the first component is a final state 
		//of the first automaton and the set of all non-final states of the second automaton
		Set<BDDPair> target = new HashSet<>();
		target.add(new BDDPair(automaton.getInitialStates(),0));

		int i = 0;
		while (!Frontier.isEmpty() && !isCancelled) {
			if(verboseMode) {
				i++;
				logger.info(UNIVERSALITY_ITERATION_STEP+i);
			}

			Set<BDDPair> tempFrontier = new HashSet<>();
			Set<BDDPair> tempF = new HashSet<>();
			
			Set<BDDPair> results = CPre(Frontier);
			int maxRounds = results.size();
			int counter =0;
			
			//Compute new frontier
			for(BDDPair frontierPair: results) {
				fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED, 
						"Backward Search", maxRounds, ++counter));
				
				boolean isIncomparable = true;
				//add the new pair if and only if it is minimal (according to the set F)
				for(BDDPair fPair: F) {
					if(frontierPair.size==fPair.size && 
							isSubset(frontierPair.bdd, fPair.bdd)) {
						isIncomparable = false;
						break;
					}
				}

				if(isIncomparable) {
					//Check whether the current pair is a counterexample
					for(BDDPair targetPair: target) {
						//If a counterexample can be found, return false
						if(frontierPair.size==targetPair.size && 
								isSubset(targetPair.bdd, frontierPair.bdd)) {
							if (computeCounterExample){
								counterExample = new CospanCounterExample(
										new CospanDecomposition(frontierPair.list));
							}
							return false;
						}
					}
					//Add the pair to the new frontier (and prepare the new F)
					tempFrontier.add(frontierPair);
					tempF.add(frontierPair);
				}
			}
			Frontier = tempFrontier;

			//Compute new F
			for (BDDPair fPair: F) {
				boolean isIncomparable = true;

				for (BDDPair frontierPair: Frontier) {
					if(frontierPair.size==fPair.size && 
							isSubset(fPair.bdd, frontierPair.bdd)) {
						isIncomparable = false;
						fPair.bdd.free();
						break;
					}
				}

				if(isIncomparable) {
					tempF.add(fPair);
				}
			}
			F = tempF;
		}

		return true;
	}

	/**
	 * Computes the CPre-function for some antichain {@code antichain}. The resulting set
	 * contains all pairs of "successors" which are maximum according to the antichain order.
	 * For more details about the CPre-function and the antichain order see the paper 
	 * "Antichains: A New Algorithm for Checking Universality of Finite Automata"
	 * <a href="http://www.springerlink.com/content/250l02n76315n786/fulltext.pdf">Springerlink</a>.
	 * 
	 * @param antichain the antichain containing the "predecessor pairs"
	 * @return the antichain containing the maximum of all "successor pairs"
	 */
	private Set<BDDPair> CPre(Set<BDDPair> antichain) {
		Set<BDDPair> result = new HashSet<>();
		int rounds = automaton.getAlphabet().size() * antichain.size();
		int counter = 0;

		for (Operation letter: automaton.getAlphabet()) {
			for (BDDPair pair: antichain) {
				if(isCancelled) {
					return result;
				}
				
				//Get the current states' predecessors
				if(isReverseApplicable(letter, pair)) {
					int size = getPredecessorSize(letter, pair);
					RavenBDD predecessors = getCPredecessorStates(pair.bdd, letter);
					predecessors = predecessors.and(
							automaton.getEncoding().getInterfaceSize(StateType.current, size));

					//Create new BDDPair
					BDDPair predecessorPair = new BDDPair(predecessors, size);
					if(computeCounterExample) {
						predecessorPair.list = new LinkedList<>();
						predecessorPair.list.add(letter);
						predecessorPair.list.addAll(pair.list);
					}

					//Make sure that the result set only contains maximum elements
					result = addMaximumElement(result, predecessorPair);
					count++;

					fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED, 
							"CPre", rounds, ++counter));
				}
			}

		}
		
		return result;
	}

	/**
	 * Computes all predecessor states from which only states of the set {@code states} are 
	 * reachable by reading the letter {@code letter}.
	 * @param states the set of (successor) state which must be reachable from the predecessor states
	 * @param letter the letter whose transition function will be used
	 * @return the set of c-predecessor states
	 */
	private RavenBDD getCPredecessorStates(RavenBDD states, Operation letter) {
		RavenBDD result = automaton.getEncoding().zero();
		RavenBDD predecessors = automaton.getPredecessorStates(states, letter);
		RavenBDD vars = automaton.getEncoding().getCurrentStateDomain().getDomainBDD();

		if(!predecessors.isZero()) {
			for(RavenBDD bdd: predecessors.getSatAssignments(vars)) {
				if(isSubset(automaton.getSuccessorStates(bdd, letter), states)) {
					result = result.or(bdd);
				}
			}
		}

		result = result.or(automaton.getStatesWithoutSuccessorFor(letter));
		vars.free();
		
		return result;
	}

	/**
	 * Computes the Post-function for some (dual) antichain {@code antichain}. The resulting 
	 * set contains all pairs of "successors" which are minimum according to the (dual) 
	 * antichain order. For more details about the Post-function and the dual antichain order 
	 * see the paper "Antichains: A New Algorithm for Checking Universality of Finite Automata"
	 * <a href="http://www.springerlink.com/content/250l02n76315n786/fulltext.pdf">Springerlink</a>.
	 * 
	 * @param antichain the antichain containing the "predecessor pairs"
	 * @return the antichain containing the minimum of all "successor pairs"
	 */
	private Set<BDDPair> fPost(Set<BDDPair> antichain) {
		Set<BDDPair> result = new HashSet<>();
		int rounds = automaton.getAlphabet().size() * antichain.size();
		int counter = 0;
		
		for (Operation operation: automaton.getAlphabet()) {
			for (BDDPair pair: antichain) {
				if(isCancelled) {
					return result;
				}
				
				//Get the current states' successors
				if(isApplicable(operation, pair)) {
					RavenBDD successors = automaton.getSuccessorStates(pair.bdd, operation);
					BDDPair successorPair = new BDDPair(successors, 
							getSuccessorSize(operation, pair));

					//Create new BDDPair
					if(computeCounterExample) {
						successorPair.list = new LinkedList<>(pair.list);
						successorPair.list.add(operation);
					}
					
					//Make sure that the result set only contains minimum elements
					result = addMinimumElement(result, successorPair);
					count++;

					fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED,
							"fPost", rounds, ++counter));
				}
			}
		}

		return result;
	}

	/**
	 * Adds the pair {@code newPair} to the antichain {@code antichain} and computes
	 * the maximum of all pairs of the extended antichain.
	 * @param antichain the antichain in which the new pair will be added
	 * @param newPair the new pair which will be added to the antichain
	 * @return the antichain which contains only maximum elements 
	 */
	private static Set<BDDPair> addMaximumElement(Set<BDDPair> antichain, BDDPair newPair) {
		Set<BDDPair> result = new HashSet<>();

		for (BDDPair antichainPair: antichain) {
			if(antichainPair.size==newPair.size && 
					isSubset(newPair.bdd, antichainPair.bdd)) {
				return antichain;
			}

			if(antichainPair.size!=newPair.size || 
					!isSubset(antichainPair.bdd, newPair.bdd)) {
				result.add(antichainPair);
			}
		}

		result.add(newPair);
		return result;
	}

	/**
	 * Adds the pair {@code newPair} to the (dual) antichain {@code antichain} and 
	 * computes the minimum of all pairs of the extended (dual) antichain.
	 * @param antichain the (dual) antichain in which the new pair will be added
	 * @param newPair the new pair which will be added to the (dual) antichain
	 * @return the (dual) antichain which contains only minimum elements
	 */
	private static Set<BDDPair> addMinimumElement(Set<BDDPair> antichain, BDDPair newPair) {
		Set<BDDPair> result = new HashSet<>();

		for (BDDPair antichainPair: antichain) {
			if(antichainPair.size==newPair.size && 
					isSubset(antichainPair.bdd, newPair.bdd)) {
				return antichain;
			}

			if(antichainPair.size!=newPair.size || 
					!isSubset(newPair.bdd, antichainPair.bdd)) {
				result.add(antichainPair);
			}

		}

		result.add(newPair);
		return result;
	}

	/**
	 * Returns {@code true} if and only if the first set (represented by the BDD) 
	 * {@code firstSet} is a subset of the second set (represented by the BDD) 
	 * {@code secondSet} 
	 * @param firstSet the first set to compare
	 * @param secondSet the second set to compare
	 * @return {@code true} if the first set is a subset of the second set, 
	 * {@code false} otherwise
	 */
	private static boolean isSubset(RavenBDD firstSet, RavenBDD secondSet) {
		RavenBDD tempBDD = firstSet.clone().imp(secondSet.clone());
		boolean result = tempBDD.isOne();
		tempBDD.free();

		return result;
	}

	/**
	 * Returns {@code true} if and only if the operation {@code letter} is 
	 * applicable to the pair {@code pair} (i.e. if it is applicable to the 
	 * state set of the pair).
	 * @param letter the operation which is to be applied
	 * @param pair the pair to which the operation is applied
	 * @return {@code true} if the operation can be applied, {@code false}
	 * otherwise
	 */
	private boolean isApplicable(Operation letter, BDDPair pair) {
		boolean isApplicable = false;

		if(letter instanceof Connect) {
			isApplicable = ((Connect) letter).getArity()<=pair.size;
		}
		else if(letter instanceof Fuse || 
				letter instanceof Perm || 
				letter instanceof Trans) {
			isApplicable = pair.size>=2;
		}
		else if(letter instanceof Res) {
			isApplicable = pair.size>=1;
		}
		else if(letter instanceof Vertex) {
			isApplicable = pair.size<automaton.getMaximumInterface();
		}
		return isApplicable;
	}

	/**
	 * Returns {@code true} if and only if the pair could be reached by the 
	 * application of the operation {@code letter} to some other pair (i.e. 
	 * if the operation could be "reverse applied" to the state set of the 
	 * pair).
	 * @param letter the operation which is to be "reverse applied"
	 * @param pair the pair to which the operation is "reverse applied"
	 * @return {@code true} if the operation can be "reverse applied", 
	 * {@code false} otherwise
	 */
	private boolean isReverseApplicable(Operation letter, BDDPair pair) {
		boolean isApplicable = false;

		if(letter instanceof Connect) {
			isApplicable = ((Connect) letter).getArity()<=pair.size;
		}
		else if(letter instanceof Fuse) {
			isApplicable = pair.size>=1 && pair.size<automaton.getMaximumInterface();
		}
		else if(letter instanceof Perm || 
				letter instanceof Trans) {
			isApplicable = pair.size>=2;
		}
		else if(letter instanceof Res) {
			isApplicable = pair.size<automaton.getMaximumInterface();
		}
		else if(letter instanceof Vertex) {
			isApplicable = pair.size>0;
		}
		return isApplicable;
	}

	/**
	 * Returns the size of the interface of the state set which is reachable by 
	 * "reverse applying" the operation {@code letter} to the pair {@code pair}.
	 * @param letter the operation which is to be "reverse applied"
	 * @param pair the pair to which the operation is "reverse applied"
	 * @return the size of the "predecessor" interface
	 */
	private static int getPredecessorSize(Operation letter, BDDPair pair) {
		int size = pair.size;

		if(letter instanceof Fuse || letter instanceof Res) {
			size += 1;
		}
		else if(letter instanceof Vertex) {
			size -= 1;
		}

		return size;
	}

	/**
	 * Returns the size of the interface of the state set which is reachable by 
	 * applying the operation {@code letter} to the pair {@code pair}.
	 * @param letter the operation which is to be applied
	 * @param pair the pair to which the operation is applied
	 * @return the size of the "successor" interface
	 */
	private static int getSuccessorSize(Operation letter, BDDPair pair) {
		int size = pair.size;

		if(letter instanceof Fuse || letter instanceof Res) {
			size -= 1;
		}
		else if(letter instanceof Vertex) {
			size += 1;
		}

		return size;
	}

	/**
	 * Fires an computation event notifying all listeners.
	 * @param e the computation event
	 */
	private void fireComputationEvent(ComputationEvent e) {
		for(ComputationListener l: listener) {
			l.computationAdvanded(e);
		}
	}

}
