/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.algorithms;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.actions.ComputationEvent;
import de.uni_due.inf.ti.raven.actions.ComputationEvent.EventType;
import de.uni_due.inf.ti.raven.actions.ComputationListener;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDDomain;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDFactory.BDDOperator;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDPairing;
import de.uni_due.inf.ti.raven.main.RavenMain;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class implements the algorithm to compute the <i>backward simulation relation<i>.
 * @author Christoph Blume
 * @version $Revision: 1.20 $
 *
 * @see ForwardSimulationAlgorithm
 */
public class BackwardSimulationAlgorithm implements SimulationAlgorithm {

	private final Logger logger;

	private Map<Automaton, Map<Automaton, RavenBDD>> simulationRelations;
	private List<ComputationListener> listener;
	private boolean verboseMode;
	private boolean isCancelled;

	private RavenBDD secondStateSet;
	private List<RavenBDD> secondStateSets;
	private RavenBDD secondInitialStateSet;	
	private Map<Operation, RavenBDD> firstTransitionRelation;
	private Map<Operation, RavenBDD> secondTransitionRelation;

	private RavenBDDDomain firstTempDomain;
	private RavenBDDDomain secondTempDomain;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new instance of the default algorithm for computing the simulation order.
	 */
	public BackwardSimulationAlgorithm() {
		listener = new LinkedList<>();
		simulationRelations = new HashMap<>();

		// Register logger
		logger = Logger.getLogger(getClass().getName());
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public void addComputationListener(ComputationListener l) {
		listener.add(l);
	}

	@Override
	public void removeComputationListener(ComputationListener l) {
		listener.remove(l);
	}

	@Override
	public boolean isSimulatedBy(Automaton firstAutomaton, 
			RavenBDD firstStateSet, RavenBDD secondStateSet) {
		// Log entering
		logger.entering(getClass().getName(), "isSimulatedBy", new Object[] {
			firstAutomaton, firstStateSet, secondStateSet});

		RavenBDD simulation = getSimulation(firstAutomaton);
		simulation = simulation.and(firstStateSet.clone());
		simulation = simulation.and(secondStateSet.clone());	
		boolean result = !simulation.isZero();
		simulation.free();

		// Log exiting
		logger.exiting(getClass().getName(), "isSimulatedBy", result);
		return result;
	}

	@Override
	public boolean isSimulatedBy(Automaton firstAutomaton, Automaton secondAutomaton, 
			RavenBDD firstStateSet, RavenBDD secondStateSet) {
		// Log entering
		logger.entering(getClass().getName(), "isSimulatedBy", new Object[] {
			firstAutomaton, secondAutomaton, firstStateSet, secondStateSet});

		RavenBDD simulation = getSimulation(firstAutomaton, secondAutomaton);
		simulation = simulation.and(firstStateSet.clone());
		simulation = simulation.and(secondStateSet.clone());	
		boolean result = !simulation.isZero();
		simulation.free();

		// Log exiting
		logger.exiting(getClass().getName(), "isSimulatedBy", result);
		return result;
	}

	@Override
	public RavenBDD getSimulation(Automaton automaton) {
		// Log entering
		logger.entering(getClass().getName(), "getSimulation", automaton);
		
		isCancelled = false;
		if(!simulationRelations.containsKey(automaton)) {
			simulationRelations.put(automaton, new HashMap<Automaton, RavenBDD>()); 
		}
		if(!simulationRelations.get(automaton).containsKey(automaton)) {
			simulationRelations.get(automaton).put(automaton, 
					getSimulation(automaton, automaton));
		}

		RavenBDD simulation = simulationRelations.get(automaton).get(automaton).clone();
		// Log exiting
		logger.exiting(getClass().getName(), "getSimulation", simulation);
		return simulation;
	}

	@Override
	public RavenBDD getSimulation(Automaton firstAutomaton, Automaton secondAutomaton) {
		// Log entering
		logger.entering(getClass().getName(), "getSimulation", new Object[]{firstAutomaton,
			secondAutomaton});
		fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_STARTED, "", 0, 0));
		
		if(!firstAutomaton.getAlphabet().equals(secondAutomaton.getAlphabet())) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.DIFFERENT_ALPHABETS_EXCEPTION);
		}
		else if(firstAutomaton.getMaximumInterface()!=secondAutomaton.getMaximumInterface()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.DIFFERENT_MAXIMUM_INTERFACES_EXCEPTION);
		}

		if(!simulationRelations.containsKey(firstAutomaton)) {
			simulationRelations.put(firstAutomaton, new HashMap<Automaton, RavenBDD>()); 
		}
		if(!simulationRelations.get(firstAutomaton).containsKey(secondAutomaton)) {
			RavenBDD simulation = computeSimulationRelation(firstAutomaton, secondAutomaton);

			if(firstAutomaton!=secondAutomaton) {
				simulation = simulation.or(getSimulation(firstAutomaton));
				simulation = simulation.or(getSimulation(secondAutomaton));
			}

			simulationRelations.get(firstAutomaton).put(secondAutomaton, simulation);
		}
		
		RavenBDD simulation = simulationRelations.get(firstAutomaton).get(secondAutomaton).clone();
		if(isCancelled) {
			fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_CANCELLED, "", 0, 0));
		}
		else {
			fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_FINISHED, "", 0, 0));
		}
		
		// Log exiting
		logger.exiting(getClass().getName(), "getSimulation", simulation);		
		return simulation;
	}

	@Override
	public void cancelComputation() {
		isCancelled = true;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Computes the simulation relation for the automaton {@code automaton}.
	 * @param firstAutomaton the automaton for which the simulation relation is computed
	 * @return the simulation relation
	 */
	private RavenBDD computeSimulationRelation(Automaton firstAutomaton, Automaton secondAutomaton) {
		verboseMode = Boolean.valueOf(RavenMain.getProperty(RavenMain.PROPERTY_VERBOSE));
		RavenBDD currentSimulation;
		RavenBDD previousSimulation;

		Repository repo = Repository.getRepository();
		logger.info("Compute simulation relation for "+repo.getName(firstAutomaton)+" and "+repo.getName(secondAutomaton));
		initialize(firstAutomaton, secondAutomaton);
		
		RavenBDDPairing simulationPairing = RavenBDDPairing.makePair();
		RavenBDDPairing reversePairing = RavenBDDPairing.makePair();
		simulationPairing.set(firstAutomaton.getEncoding().getCurrentStateDomain(), firstTempDomain);
		simulationPairing.set(secondAutomaton.getEncoding().getSuccessorStateDomain(), 
				secondTempDomain);
		reversePairing.set(firstTempDomain, firstAutomaton.getEncoding().getCurrentStateDomain());
		reversePairing.set(secondTempDomain, secondAutomaton.getEncoding().getSuccessorStateDomain());

		//Initialize the simulation relation
		currentSimulation = firstAutomaton.getInitialStates().and(secondInitialStateSet);
		for(int size=0;size<=firstAutomaton.getMaximumInterface();size++) {
			RavenBDD nonInitialStates = firstAutomaton.getStates(size).and(
					firstAutomaton.getInitialStates().not());
			currentSimulation = currentSimulation.or(nonInitialStates.and(
					secondStateSets.get(size).clone()));
		}

		List<Operation> operationList = new LinkedList<>(firstAutomaton.getAlphabet().getOperations());

		int i = 0;
		do {
			if(verboseMode) {
				i++;
				logger.info("[simulation] Iteration step: " + i);
			}

			previousSimulation = currentSimulation;
			RavenBDD tempBDD = firstAutomaton.getEncoding().one();
			RavenBDD replacedSimulation = currentSimulation.clone().replace(simulationPairing);

			int letterCounter = 0;
			for(Operation letter: operationList) {
				letterCounter++;
				
				//Get transition BDD
				RavenBDD firstTransBDD = firstTransitionRelation.get(letter).clone();
				RavenBDD secondTransBDD = secondTransitionRelation.get(letter).clone();
				
				//For all predecessors of the first state, a predecessor for the second state must 
				//exist
				//Existential quantification over the predecessor states (of the second state in the 
				//backward simulation)
				RavenBDD existentialBDD = secondTransBDD.applyEx(currentSimulation.clone(), 
						BDDOperator.and, 
						secondAutomaton.getEncoding().getSuccessorStateDomain().getDomainBDD());
				secondTransBDD.free();
				
				//Build implication (if a predecessor for the first state was found 
				//then there exist a predecessor for the second state)
				tempBDD = tempBDD.and(firstTransBDD.imp(existentialBDD));

				fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED, 
						"Process '"+letter+"'-operation", firstAutomaton.getAlphabet().size(), 
						letterCounter));
				if(verboseMode) {
					logger.info("Process '"+letter+"'-operation");
				}

			}
			
			//Universal quantification over all successor states (of the first state in the simulation relation)
			currentSimulation = replacedSimulation.clone().and(tempBDD.forAll(
					firstAutomaton.getEncoding().getCurrentStateDomain().getDomainBDD()));
			currentSimulation = currentSimulation.replace(reversePairing);
		}
		while(!currentSimulation.equals(previousSimulation) && !isCancelled);
		
		return currentSimulation;
	}

	/**
	 * Initializes the simulation algorithms by preparing data structures which will be used
	 * to temporarily store states, state sets and transitions of the two involved automata.
	 * @param firstAutomaton the first automaton
	 * @param secondAutomaton the second automaton
	 */
	private void initialize(Automaton firstAutomaton, Automaton secondAutomaton) {
		RavenBDDPairing currentTempPairing = RavenBDDPairing.makePair();
		RavenBDDPairing successorTempPairing = RavenBDDPairing.makePair();
		RavenBDDPairing currentSuccessorPairing = RavenBDDPairing.makePair();
		RavenBDDPairing successorCurrentPairing = RavenBDDPairing.makePair();

		//Prepare temporary BDD Domains
		firstTempDomain = RavenBDDDomain.createDomain(firstAutomaton.getEncoding().getLength());
		secondTempDomain = RavenBDDDomain.createDomain(secondAutomaton.getEncoding().getLength());
		
		//Prepare BDD pairings
		currentTempPairing.set(firstAutomaton.getEncoding().getSuccessorStateDomain(),
				firstTempDomain);
		currentSuccessorPairing.set(secondAutomaton.getEncoding().getCurrentStateDomain(), 
				secondAutomaton.getEncoding().getSuccessorStateDomain());
		successorCurrentPairing.set(firstAutomaton.getEncoding().getSuccessorStateDomain(), 
				firstAutomaton.getEncoding().getCurrentStateDomain());
		
		firstTransitionRelation = new HashMap<>();
		secondTransitionRelation = new HashMap<>();
		if(firstAutomaton==secondAutomaton) {
			successorTempPairing.set(secondAutomaton.getEncoding().getSuccessorStateDomain(),
					secondTempDomain);
			
			//Prepare (initial) state sets
			secondStateSet = secondAutomaton.getStates();
			secondStateSet = secondStateSet.replace(currentSuccessorPairing);
			secondInitialStateSet = secondAutomaton.getInitialStates();
			secondInitialStateSet = secondInitialStateSet.replace(currentSuccessorPairing);

			secondStateSets = new ArrayList<>();
			for(int i=0;i<=secondAutomaton.getMaximumInterface();i++) {
				secondStateSets.add(i, secondAutomaton.getStates(i).replace(currentSuccessorPairing));
			}
			
			//Prepare transition relations
			for(Operation letter: firstAutomaton.getAlphabet()) {
				firstTransitionRelation.put(letter, firstAutomaton.getTransitionRelation(letter).replace(
						currentTempPairing));
				secondTransitionRelation.put(letter, secondAutomaton.getTransitionRelation(letter).replace(
						successorTempPairing).replace(currentSuccessorPairing));
			}
		}
		else {
			successorTempPairing.set(secondAutomaton.getEncoding().getCurrentStateDomain(),
					secondTempDomain);
			
			//Prepare (initial) state sets
			secondStateSet = secondAutomaton.getStates();
			secondInitialStateSet = secondAutomaton.getInitialStates();
			secondStateSets = new ArrayList<>();
			for(int i=0;i<=secondAutomaton.getMaximumInterface();i++) {
				secondStateSets.add(i, secondAutomaton.getStates(i));
			}
			
			//Prepare transition relations
			for(Operation letter: firstAutomaton.getAlphabet()) {
				firstTransitionRelation.put(letter, 
						firstAutomaton.getTransitionRelation(letter).replace(currentTempPairing));
				secondTransitionRelation.put(letter, 
						secondAutomaton.getTransitionRelation(letter).replace(
								successorTempPairing).replace(currentSuccessorPairing));
			}
		}
	}

	/**
	 * Fires an computation event notifying all listeners.
	 * @param e the computation event
	 */
	private void fireComputationEvent(ComputationEvent e) {
		for(ComputationListener l: listener) {
			l.computationAdvanded(e);
		}
	}
	
}
