/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.algorithms;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.actions.ComputationEvent;
import de.uni_due.inf.ti.raven.actions.ComputationEvent.EventType;
import de.uni_due.inf.ti.raven.actions.ComputationListener;
import de.uni_due.inf.ti.raven.data.CospanCounterExample;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.main.RavenMain;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class implements the bisimulation up to congruence algorithm for both deciding the 
 * language equivalence and the language  inclusion problem for (finite) automata. The algorithm 
 * is based on the work of Bonchi and Pous. More details about the algorithm can be found in 
 * "Checking NFA equivalence with bisimulations up to congruence" 
 * <a href="http://hal.archives-ouvertes.fr/docs/00/71/66/89/PDF/hkc.pdf">HAL</a>.
 * 
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.22 $
 *
 */
public class BisimulationUpToCongruenceLanguageEquivalence implements LanguageEquivalenceAlgorithm {

	private final class BDDPair {

		private RavenBDD first;
		private RavenBDD second;
		private List<Operation> list;

		/**
		 * Creates a new BDD Pair.
		 * @param first the pair's first component
		 * @param second the pair's second component
		 * 
		 */
		public BDDPair(RavenBDD first, RavenBDD second) {
			this.first = first;
			this.second = second;
			this.list = new LinkedList<>();
		}

		/**
		 * Returns the BDD of the first component.
		 * @return the first component's BDD
		 */
		public RavenBDD getFirst() {
			return first.clone();
		}


		/**
		 * Returns the BDD of the second component.
		 * @return the second component's BDD
		 */
		public RavenBDD getSecond() {
			return second.clone();
		}

		@Override
		public int hashCode() {
			return first.hashCode() + 47 * second.hashCode();
		}

		@Override
		public boolean equals(Object o) {
			if(!(o instanceof BDDPair)) {
				return false;
			}

			BDDPair p = (BDDPair) o;
			return first.equals(p.first) && second.equals(p.second);

		}

		@Override
		public String toString() {
			return "({"+firstAutomaton.printStates(first)+"}, {"+secondAutomaton.printStates(second)+"}, "+list+")";
		}

	}

	private final Logger logger;
	private final boolean computerCounterExample;
	private List<ComputationListener> listener;

	private boolean verboseMode;
	private boolean isCancelled;
	private Automaton firstAutomaton;
	private Automaton secondAutomaton;
	private CospanCounterExample counterExample;
	private SearchType type;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new instance of the bisimulation-up-to-congruence algorithm.
	 * @param type the search type of this instance
	 * @param computeCounterExample if {@code true} a counterexample will be computed 
	 * if the language equivalence/inclusion check fails
	 */
	public BisimulationUpToCongruenceLanguageEquivalence(SearchType type, 
			boolean computeCounterExample) {
		this.type = type;
		this.computerCounterExample = computeCounterExample;
		listener = new LinkedList<>();

		// Register logger
		logger = Logger.getLogger(getClass().getName());
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public void addComputationListener(ComputationListener l) {
		listener.add(l);
	}

	@Override
	public void removeComputationListener(ComputationListener l) {
		listener.remove(l);
	}

	@Override
	public boolean isEquivalent(Automaton first, Automaton second) {
		if(first==null) {
			throw new NullPointerException(
					RavenExceptionMessages.FIRST_AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(second==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SECOND_AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(first.getMaximumInterface()!=second.getMaximumInterface()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_INTERFACE_SIZES_MUST_NOT_BE_DIFFERENT_EXCEPTION);
		}
		if(first.getInnerInterface()!=second.getInnerInterface()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZES_MUST_NOT_BE_DIFFERENT_EXCEPTION);
		}
		if(first.getOuterInterface()!=second.getOuterInterface()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZES_MUST_NOT_BE_DIFFERENT_EXCEPTION);
		}
		
		// Log entering
		logger.entering(getClass().getName(), "isEquivalent", new Object[]{first, second});
		verboseMode = Boolean.valueOf(RavenMain.getProperty(RavenMain.PROPERTY_VERBOSE));
		fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_STARTED, "", 0, 0));
		
		isCancelled = false;
		firstAutomaton = first;
		secondAutomaton = second;
		boolean isEquivalent = true;

		switch (type) {
		case forward:
			isEquivalent = performForwardSearch();
			break;

		case backward:
			isEquivalent = performBackwardSearch();
			break;

		default:
			throw new RuntimeException(RavenExceptionMessages.ILLEGAL_SEARCH_TYPE_EXCEPTION);
		}
		
		if(isCancelled) {
			fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_CANCELLED, "", 0, 0));
		}
		else {
			fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_FINISHED, "", 0, 0));
		}		

		// Log exiting
		logger.exiting(getClass().getName(), "isEquivalent", isEquivalent);
		return isEquivalent;
	}

	@Override
	public CospanCounterExample getLanguageEquivalenceCounterExample() {
		return counterExample;
	}
	
	@Override
	public void cancelComputation() {
//		isCancelled = true;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns {@code true} if and only if the language of the automaton {@code firstAutomaton} 
	 * is a subset of the language of the automaton {@code secondAutomaton}. The computation is 
	 * performed via a forward search based on the bisimulation up-to congruence algorithm by 
	 * Bonchi and Pous.
	 * @return {@code true} if the first language is included in the second language, 
	 * {@code false} otherwise
	 */
	private boolean performForwardSearch() {
		boolean isEquivalent = true;
		Set<BDDPair> R = new HashSet<>();
		List<BDDPair> todo = new LinkedList<>();
		todo.add(new BDDPair(firstAutomaton.getInitialStates(), secondAutomaton.getInitialStates()));

		int i = 0;
		while (!todo.isEmpty() && !isCancelled) {
			
			if(verboseMode) {
				i++;
				logger.info(LANGUAGE_EQUIVALENCE_ITERATION_STEP+i);
			}
			BDDPair pair = todo.remove(0);

			Set<BDDPair> union = new HashSet<>(R);
			union.addAll(todo);
			
			//Check whether the pair is already contained in the bisimulation up to congruence
			if(compareNormalform(pair, union)) {
				pair.first.free();
				pair.second.free();
				continue;
			}
			
			//If the acceptance behavior of the two state sets of the pair is different
			//the state sets can be divided -> the two automata can not be equivalent
			if(!checkAcceptanceEquivalence(pair, firstAutomaton, secondAutomaton)) {
				isEquivalent = false;
				if(computerCounterExample) {
					CospanDecomposition cospan = new CospanDecomposition(
							firstAutomaton.getInnerInterface(), pair.list);
					counterExample = new CospanCounterExample(cospan);
				}
				break;
			}
			
			//Compute and add all successor pairs, i.e. the pairs containing the successor state sets
			//of the current pair's state sets
			for(Operation letter: firstAutomaton.getAlphabet()) {
				RavenBDD firstSuccessors = firstAutomaton.getSuccessorStates(pair.first, letter);
				RavenBDD secondSuccessors = secondAutomaton.getSuccessorStates(pair.second, letter);

				if(!firstSuccessors.isZero() || !secondSuccessors.isZero()) {
					BDDPair p = new BDDPair(firstSuccessors, secondSuccessors);
					if(computerCounterExample) {
						p.list = new LinkedList<>(pair.list);
						p.list.add(letter);
					}

					if(!union.contains(p)) {
						todo.add(p);
					}
					else {
						p.first.free();
						p.second.free();
					}
				}
			}
			R.add(pair);
		}
		return isEquivalent;
	}
	
	/**
	 * Returns {@code true} if and only if the language of the automaton {@code firstAutomaton} 
	 * is a subset of the language of the automaton {@code secondAutomaton}. The computation is 
	 * performed via a backward search based on the bisimulation up-to congruence algorithm by 
	 * Bonchi and Pous.
	 * @return {@code true} if the first language is included in the second language, 
	 * {@code false} otherwise
	 */
	private boolean performBackwardSearch() {
		boolean isEquivalent = true;
		Set<BDDPair> R = new HashSet<>();
		List<BDDPair> todo = new LinkedList<>();
		todo.add(new BDDPair(firstAutomaton.getFinalStates(), secondAutomaton.getFinalStates()));

		int i = 0;
		while (!todo.isEmpty() && !isCancelled) {
			if(verboseMode) {
				i++;
				logger.info(LANGUAGE_EQUIVALENCE_ITERATION_STEP+i);
			}
			BDDPair pair = todo.remove(0);

			Set<BDDPair> union = new HashSet<>(R);
			union.addAll(todo);

			//Check whether the pair is already contained in the bisimulation up to congruence
			if(compareNormalform(pair, union)) {
				pair.first.free();
				pair.second.free();
				continue;
			}

			//If the acceptance behavior of the two state sets of the pair is different
			//the state sets can be divided -> the two automata can not be equivalent
			if(!checkInitialEquivalence(pair, firstAutomaton, secondAutomaton)) {
				isEquivalent = false;
				if(computerCounterExample) {
					CospanDecomposition cospan = new CospanDecomposition(
							firstAutomaton.getInnerInterface(), pair.list);
					counterExample = new CospanCounterExample(cospan);
				}
				break;
			}

			//Compute and add all predecessor pairs, i.e. the pairs containing the predecessor state sets
			//of the current pair's state sets
			for(Operation letter: firstAutomaton.getAlphabet()) {
				RavenBDD firstPredecessors = firstAutomaton.getPredecessorStates(pair.first, letter);
				RavenBDD secondPredecessors = secondAutomaton.getPredecessorStates(pair.second, letter);

				if(!firstPredecessors.isZero() || !secondPredecessors.isZero()) {
					BDDPair p = new BDDPair(firstPredecessors, secondPredecessors);
					if(computerCounterExample) {
						p.list = new LinkedList<>();
						p.list.add(letter);
						p.list.addAll(pair.list);
					}

					if(!union.contains(p)) {
						todo.add(p);
					}
					else {
						p.first.free();
						p.second.free();
					}
				}
			}
			R.add(pair);
		}
		return isEquivalent;
	}
	
	/**
	 * Creates and compares the normal form for the first and second component BDDs of the pair 
	 * {@code pair} with respect to the relation {@code relation}. Returns {@code true} if the 
	 * normal forms of both BDDs are equal.
	 * @param pair the pair for which the normal form is computed
	 * @param relation the relation used for the computation of the normal form
	 * @return {@code true} if the normal forms of both components are equal, {@code false}
	 * otherwise
	 */
	private boolean compareNormalform(BDDPair pair, Set<BDDPair> relation) {
		RavenBDD first = pair.getFirst();
		RavenBDD second = pair.getSecond();
		RavenBDD temp;
		int maximum = relation.size();
		int counter = 0;

		//TODO If "pair" (i.e. first or second) has been changed start a 
		// new iteration over all elements (which have not been used yet) of "relation"
		boolean repeat = true;
		computeNormalform:
		while (repeat && !isCancelled) {
			repeat = false;
			for (BDDPair element : relation) {
				if (!element.second.isZero()) {
					temp = element.getFirst().imp(pair.getFirst());
					if (temp.isOne()) {
						first = first.or(element.getSecond());
						repeat = true;
						break computeNormalform;
					}
					temp.free();

					temp = element.getFirst().imp(pair.getSecond());
					if (temp.isOne()) {
						second = second.or(element.getSecond());
						repeat = true;
						break computeNormalform;
					}
					temp.free();
				}

				if (!element.first.isZero()) {
					temp = element.getSecond().imp(pair.getFirst());
					if (temp.isOne()) {
						first = first.or(element.getFirst());
						repeat = true;
						break computeNormalform;
					}
					temp.free();

					temp = element.getSecond().imp(pair.getSecond());
					if (temp.isOne()) {
						second = second.or(element.getFirst());
						repeat = true;
						break computeNormalform;
					}
					temp.free();
				}

				fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED, 
						"normalform", maximum, ++counter));
			}
		}

		boolean result = first.equals(second);
		first.free();
		second.free();

		return result;
	}

	/**
	 * Checks whether the first and second component BDDs of the pair {@code pair} are 
	 * accepting state equivalent, i.e. whether both state sets, the one given by the first 
	 * BDD and the one given by the second BDD, contain at least one final state of the 
	 * first and second automaton respectively, or whether both states do not contain a
	 * final state
	 * @param pair the pair whose acceptance behavior is checked
	 * @param firstAutomaton the first automaton
	 * @param secondAutomaton the second automaton
	 * @return {@code true} if either both state sets are accepting or both state sets are rejecting, 
	 * {@code false} otherwise
	 */
	private boolean checkAcceptanceEquivalence(BDDPair pair, Automaton firstAutomaton,
			Automaton secondAutomaton) {
		RavenBDD firstStateSet = pair.getFirst();
		RavenBDD secondStateSet = pair.getSecond();
		boolean result;

		if(firstStateSet.equals(secondStateSet)) {
			result = true;
		}
		else {
			result = ((firstAutomaton.isAcceptingStateSet(firstStateSet) && 
					secondAutomaton.isAcceptingStateSet(secondStateSet)) ||
					(!firstAutomaton.isAcceptingStateSet(firstStateSet) &&
							!secondAutomaton.isAcceptingStateSet(secondStateSet)));
		}

		firstStateSet.free();
		secondStateSet.free();

		return result;
	}
	
	/**
	 * 
	 * Checks whether the first and second component BDDs of the pair {@code pair} are 
	 * initial state equivalent, i.e. whether both state sets, the one given by the first 
	 * BDD and the one given by the second BDD, are subsets of the initial states of the 
	 * first and second automaton respectively, or whether both states are no subsets of 
	 * the initial state sets 
	 * @param pair the pair whose acceptance behavior is checked
	 * @param firstAutomaton the first automaton
	 * @param secondAutomaton the second automaton
	 * @return {@code true} if either both state sets are initial or both state sets are 
	 * non-initial, {@code false} otherwise
	 */
	private boolean checkInitialEquivalence(BDDPair pair, Automaton firstAutomaton,
			Automaton secondAutomaton) {
		RavenBDD firstStateSet = pair.getFirst();
		RavenBDD secondStateSet = pair.getSecond();
		boolean result;

		if(firstStateSet.equals(secondStateSet)) {
			result = true;
		}
		else {
			result = ((firstAutomaton.isInitialStateSet(firstStateSet) && 
					secondAutomaton.isInitialStateSet(secondStateSet)) ||
					(!firstAutomaton.isInitialStateSet(firstStateSet) &&
							!secondAutomaton.isInitialStateSet(secondStateSet)));
		}

		firstStateSet.free();
		secondStateSet.free();

		return result;
	}

	
	/**
	 * Fires an computation event notifying all listeners.
	 * @param e the computation event
	 */
	private void fireComputationEvent(ComputationEvent e) {
		for(ComputationListener l: listener) {
			l.computationAdvanded(e);
		}
	}
	
}
