/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.algorithms;

import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.actions.ComputationListener;
import de.uni_due.inf.ti.raven.data.CospanCounterExample;
import de.uni_due.inf.ti.raven.data.automata.AutomatonFactory;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class implements the bisimulation up to congruence algorithm for both deciding the 
 * language equivalence and the language  inclusion problem for (finite) automata. The algorithm 
 * is based on the work of Bonchi and Pous. More details about the algorithm can be found in 
 * "Checking NFA equivalence with bisimulations up to congruence" 
 * <a href="http://hal.archives-ouvertes.fr/docs/00/71/66/89/PDF/hkc.pdf">HAL</a>.
 * 
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.8 $
 *
 */
public class BisimulationUpToCongruenceLanguageInclusion implements LanguageInclusionAlgorithm {

	private final Logger logger;

	private BisimulationUpToCongruenceLanguageEquivalence languageEquivalenceAlgorithm;
	private final boolean computeCounterExample;
	private CospanCounterExample counterExample;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new instance of the bisimulation-up-to-congruence algorithm.
	 * @param type the search type of this instance
	 * @param computeCounterExample if {@code true} a counterexample will be computed 
	 * if the language equivalence/inclusion check fails
	 */
	public BisimulationUpToCongruenceLanguageInclusion(SearchType type, boolean computeCounterExample) {
		if(type==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SEARCH_TYPE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		// Register logger
		logger = Logger.getLogger(getClass().getName());
		this.computeCounterExample = computeCounterExample;
		
		languageEquivalenceAlgorithm = new BisimulationUpToCongruenceLanguageEquivalence(type, 
				computeCounterExample);
	}

	/**
	 * Creates a new instance of the bisimulation-up-to-congruence algorithm.
	 * @param type the search type of this instance
	 */
	public BisimulationUpToCongruenceLanguageInclusion(SearchType type) {
		this(type, false);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public void addComputationListener(ComputationListener l) {
		languageEquivalenceAlgorithm.addComputationListener(l);
	}

	@Override
	public void removeComputationListener(ComputationListener l) {
		languageEquivalenceAlgorithm.removeComputationListener(l);
	}

	@Override
	public boolean isIncluded(Automaton first, Automaton second) {
		// Log entering
		logger.entering(getClass().getName(), "isIncluded", new Object[]{first, second});

		AutomatonFactory factory = AutomatonFactory.getFactory();
		Automaton unionAutomaton = factory.createUnionAutomaton(first, second);
		boolean isEquivalent = languageEquivalenceAlgorithm.isEquivalent(unionAutomaton, second);
		if(computeCounterExample) {
			counterExample = languageEquivalenceAlgorithm.getLanguageEquivalenceCounterExample();
		}

		// Log exiting
		logger.exiting(getClass().getName(), "isIncluded", isEquivalent);
		return isEquivalent;
	}

	@Override
	public CospanCounterExample getCounterExample() {
		return counterExample;
	}

	@Override
	public void cancelComputation() {
		languageEquivalenceAlgorithm.cancelComputation();
	}
	
}
