/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.algorithms;

import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import de.uni_due.inf.ti.raven.actions.ComputationEvent;
import de.uni_due.inf.ti.raven.actions.ComputationEvent.EventType;
import de.uni_due.inf.ti.raven.actions.ComputationListener;
import de.uni_due.inf.ti.raven.data.Cospan;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Edge;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.Node;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class implements an algorithm which is used to decompose a given {@code Cospan} object into 
 * an <i>atomic cospan decomposition</i>, represented by a {@code CospanDecomposition} object.
 * The algorithm does the following:
 * <ul>
 * 		<li>The first step is to add a vertex-operation for each node in the middle graph of the 
 * 		given cospan. Note that there are additional nodes which are accessible from the inner 
 * 		interface (of the given cospan).</li>
 * 		<li>In the second step all edges of the middle graph are added by permuting the incident 
 * 		nodes to the last position of the outer interface with the help of shift- and trans-operations 
 * 		and a subsequent connect-operation, where  is just a placeholder for the corresponding 
 * 		label.</li>
 * 		<li>Next, in step three, the nodes added in step one are fused with the already existing 
 * 		nodes according to the inner interface (of the given cospan). At this point the middle graph
 * 		obtained by the atomic cospan decomposition matches the middle graph of the given cospan.</li>
 * 		<li>Hence, in step four, we can permute the outer interface (of the atomic cospan 
 * 		decomposition) by adding shift- and trans-operations repeatedly such that the nodes which 
 * 		are not accessible by the outer interface of the given cospan allocate the last positions of 
 * 		the outer interface of the atomic cospan decomposition.</li>
 * 		<li>Then, the fifth and last step is to remove these last nodes from the outer interface by
 * 		adding enough res-operations.</li>
 * </ul>
 * @author Christoph Blume, Dennis Nolte
 * @version $Revision: 1.11 $
 *
 */
public final class DefaultCospanDecompositionAlgorithm implements CospanDecompositionAlgorithm {

	private final List<ComputationListener> listener;
	
	private List<Node> innerInterface;
	private List<Node> outerInterface;
	private List<Node> currentInterface;
	private List<Operation> operations;
	private List<Node> reserveNodes;

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Searches the edge whose number of missing incident nodes is the smallest
	 * @param interfaceList the current interface
	 * @param edgeList the edges to be checked
	 * @return the edge which should be added next
	 */
	private static Edge findFavoriteNewEdge(List<Node> interfaceList, Set<Edge> edgeList) {
		Edge favorite = null;
		int favoritSize = 0;
		for(Edge edge: edgeList) {

			//if no favorite edge exists right now use the first edge found
			if(favorite == null) {
				favorite = edge;
				favoritSize = edge.getIncidentNodes().size();
			}
			
			List<Node> needed = new LinkedList<>(); 
			
			//search for the edge which needs the minimum number of new nodes in the current interface
			for(Node node:	edge.getIncidentNodes()) {
				needed.add(node);
			}
			needed.removeAll(interfaceList);
			
			if(needed.size() < favoritSize) {
				favorite = edge;
				favoritSize = needed.size();
			}
		}
		return favorite;
	}
	
	/**
	 * Checks if the given edge {@code edge} consists of edges
	 * @param edge the edge to be checked
	 * @return true if the edge consists of loops, false otherwise
	 */
	private static boolean hasLoops(Edge edge) {
		List<Node> differentNodes = new LinkedList<>();
		
		for(Node node: edge.getIncidentNodes()) {
			if(!differentNodes.contains(node)) {
				differentNodes.add(node);
			}
			else {
				return true;
			}
		}

		return false;
	}

	/**
	 * Checks if the given node {@code node} has copies in the current interface
	 * @param interfaceList the current interface
	 * @param node the node to be checked
	 * @return the number of copies
	 */
	private static int numberOfCopyNodes(List<Node> interfaceList, Node node) {
		int startIndex = interfaceList.indexOf(node);
		int number = 0;
		
		//starting from the node to be checked search the rest of the list for copies
		for(int j=startIndex+1;j<interfaceList.size();j++) {
			if(interfaceList.get(startIndex).equals(interfaceList.get(j))) {
				number++;
			}
		}
		
		return number;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * 
	 */
	public DefaultCospanDecompositionAlgorithm() {
		listener = new LinkedList<>();
	}	

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public CospanDecomposition decompose(Cospan cospan) {
		if(cospan==null) {
			throw new NullPointerException(RavenExceptionMessages.COSPAN_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		Graph middleGraph = cospan.getGraph();
		
		innerInterface = cospan.getInnerInterface();
		outerInterface = cospan.getOuterInterface();
		currentInterface = cospan.getInnerInterface();
		operations = new LinkedList<>();
		reserveNodes = new LinkedList<>();

		fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_STARTED, "", 0, 0));
		
		//Step 1: extract possible copies from the current interface
		extractCopys();
		
		//Step 2: create the reserve nodes by adding all nodes from the middle graph which aren't in the inner interface list
		for(Node node: cospan.getNodes()) {	
			if(!innerInterface.contains(node)) {
				reserveNodes.add(node);
			}
		}
		
		Set<Edge> edgeList = middleGraph.getEdges();
		List<Edge> completeEdge;
		
		//Step 3: add edges on the fly while using the reserve nodes to complete the required incident nodes for the edges
		while(!(reserveNodes.size() == 0 && edgeList.size()==0)) {
			
			completeEdge = new LinkedList<>();
			List<Edge> possibleEdges = new LinkedList<>(edgeList);
			
			//check if there exist edges that can be added
			for(Edge edge: possibleEdges) {
				if(currentInterface.containsAll(edge.getIncidentNodes())) {
					completeEdge.add(edge);
					edgeList.remove(edge);
				}
			}
			
			//if there are complete edges
			if(completeEdge.size()>0) {

				List<Node> sortList;
				List<Node> edgeNodes;
				
				//for each completed edge create an interface order to add the connect letter
				for(Edge edge: completeEdge) {
					
					sortList = new LinkedList<>(currentInterface);
					edgeNodes = new LinkedList<>(edge.getIncidentNodes());
					
					for(Node node: edgeNodes) {
						sortList.remove(node);
						sortList.add(node);
					}
					if(!currentInterface.equals(sortList)) {
						sortToInterfaceOrder(currentInterface, sortList);
					}
					
					//if the edge consists of loops new copies in the current interface are created
					if(hasLoops(edge)) {
						int arity = edgeNodes.size();
						int numberOfNeededPerm = currentInterface.size();
						
						//if there are more nodes in the current interface than the size of the edges arity
						if(numberOfNeededPerm >= arity) {
							int permCounter = 0;
							//the last positions of the current interface (length of arity) need to be the same 
							//as the edges required node list
							for(int i= arity-1; i >= 0; i--) {
								Node lastNode = currentInterface.get(currentInterface.size()-1);
								Node edgeNode = edgeNodes.get(i);
								if(lastNode.equals(edgeNode)) {
									addPerm();
									permCounter++;
								}
								else {
									operations.add(Operation.getVertex());
									currentInterface.add(edgeNode);
									addPerm();
								}
								
							}
							
							//add perm letters until the edge nodes are at the last position again
							while(permCounter < numberOfNeededPerm) {
								addPerm();
								permCounter++;
							}								
						}
						//if there only exists one node in the current interface just add vertex operations arity-1 times
						else if(numberOfNeededPerm == 1) {
							Node edgeNode = edgeNodes.get(0);
							for(int i=1; i<arity; i++) {
								operations.add(Operation.getVertex());
								currentInterface.add(edgeNode);
							}
						}
						//if the current interface size is less than the arity of the added loop
						else if(numberOfNeededPerm < arity) {
							
							List<Node> seenNodes = new LinkedList<>(currentInterface);
							
							//prevent seeing the same nodes more than once by using an alternate list
							for(int i= arity-1; i>= 0; i--) {
								Node lastNode = seenNodes.get(seenNodes.size()-1);
								Node edgeNode = edgeNodes.get(i);
								
								if(lastNode.equals(edgeNode)) {
									addPerm();
									seenNodes.remove(seenNodes.size()-1);
								}
								else {
									operations.add(Operation.getVertex());
									currentInterface.add(edgeNode);
									addPerm();
								}
								
								//when the alternate list is empty just fill in the edge nodes to the current interface
								if(seenNodes.isEmpty()) {
									for(int j = i-1; j >= 0; j--) {
										Node restNode = edgeNodes.get(j);
										operations.add(Operation.getVertex());
										currentInterface.add(restNode);
										addPerm();
									}
									break;
								}
								
							}
						}
						
						addConnect(edge);
						extractCopys();
					}
					//if the complete edge doesn't have loops just add the connect letter
					else {
						addConnect(edge);								
					}
				}
				
				//clean the current interface and only keep nodes for further edges and the outer interface list
				List<Node> neededList = new LinkedList<>(outerInterface);
				for(Edge edge: edgeList) {
					for(Node node : edge.getIncidentNodes()) {
						if(!neededList.contains(node)) {
							neededList.add(node);
						}
					}
				}
				
				if(!neededList.containsAll(currentInterface)) {
					cleanInterface(currentInterface, neededList);
				}
				
			}
			//if there are no complete edges at all use reserve nodes to fulfill more requirements
			else {
				//if there are any more edges available
				if(edgeList.size() > 0) {
					//search for the next edge to be added
					Edge edge = findFavoriteNewEdge(currentInterface, edgeList);
					
					//add the required nodes to the current interface
					for(Node node: edge.getIncidentNodes()) {
						if(!(currentInterface.contains(node))) {
							addVertex(reserveNodes.indexOf(node));
						}
					}						
				}
				//no more edges are needed but there are a few more nodes 
				else {
					while(!(reserveNodes.size() == 0)) {
						addVertex(0);
					}
				}
			}
			
		}
		
		//Step 4 clean the rest and permute to the outer interface
		if(!outerInterface.containsAll(currentInterface)) {
			cleanInterface(currentInterface, outerInterface);
		}
		
		if(!currentInterface.equals(outerInterface)) {
			sortToInterfaceOrder(currentInterface, outerInterface);
		}
		
		fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_FINISHED, "", 0, 0));
		
		return new CospanDecomposition(innerInterface.size(), operations);
	}
	

	@Override
	public void addComputationListener(ComputationListener l) {
		listener.add(l);
	}

	@Override
	public void removeComputationListener(ComputationListener l) {
		listener.remove(l);
	}

	@Override
	public void cancelComputation() {
		// TODO Auto-generated method stub

	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Fires a computation event notifying all listeners.
	 * 
	 * @param e the computation event
	 */
	private void fireComputationEvent(ComputationEvent e) {
		for (ComputationListener l : listener) {
			l.computationAdvanded(e);
		}
	}
	
	/**
	 * Clears the inner interface from multiple node names
	 */
	private void extractCopys() {
		List<Node> singleCopyList = new LinkedList<>();
		for(Node node: currentInterface) {
			if(!singleCopyList.contains(node)) {
				singleCopyList.add(node);
			}
		}
		
		for(Node node: singleCopyList) {
			int copyCount = numberOfCopyNodes(currentInterface, node);
			if(copyCount > 0) {	
				eliminateCopys(currentInterface, node, copyCount);
			}				
		}
	}
	
	/**
	 * Removes the number of copies {@code numberOfCopys} of the node {@code keeperNode} and other copies
	 * by chance from the current interface.
	 * @param interfaceList TODO
	 * @param node the node which copies are removed
	 * @param numberOfCopys the number of copies
	 */
	private void eliminateCopys(List<Node> interfaceList, Node node, int numberOfCopys) {
		while(numberOfCopys > 0) {
			
			//Shift the keeperNode to the second position of the current interface
			while(!interfaceList.get(1).equals(node)) {
				addPerm();	
			}
			
			//Search for a copy and shift it to the first position of the list to create a pair with the keeper node
			while(!interfaceList.get(0).equals(node)) {
				
				while(interfaceList.get(interfaceList.size()-1).equals(
						interfaceList.get(interfaceList.size()-2))) {					
					// If we found an expected copy add a fuse letter and remove one copy
					if(interfaceList.get(interfaceList.size()-1).equals(node)) {
						addFuse();
						numberOfCopys--;
					}
					// If we found a copy of anything else by chance add a fuse letter to the decomposition
					else {
						addFuse();
					}
					
					if(numberOfCopys == 0){
						return;
					}
				}
				
				addTrans();
				addPerm();
			}
			
			int counter = interfaceList.size()-2;
			
			//Shift the found pair to the last positions of the list, again remove other copies by chance
			for(int i=0;i<counter;i++) {
				
				if(interfaceList.get(interfaceList.size()-1).equals(
						interfaceList.get(interfaceList.size()-2))) {
					// If we found an expected copy decrease the number of searched copies 
					if(interfaceList.get(interfaceList.size()-1).equals(node)) {
						numberOfCopys--;
					}
					addFuse();
				}
				else {
					addPerm();						
				}
			}
			
			//Remove the found pair
			addFuse();
			numberOfCopys--;
		}
	}
	
	
	/**
	 * Sorts the current interface to be equal to the {@code sortList}
	 * @param sortList defines the order the current interface is sorted to
	 */
	private void sortToInterfaceOrder(List<Node> interfaceList, List<Node> sortList) {
		Node keeperNode;
		Node keeperSuccessor;
		int keeperIndex = sortList.size()-2;
		
		//if there are more than two nodes in the current interface sort the nodes to the sortedList order
		while(keeperIndex > 0) {
			keeperNode = sortList.get(keeperIndex);
			keeperSuccessor = sortList.get(keeperIndex+1);
			
			while(!(interfaceList.get(1).equals(keeperNode))) {
				addPerm();
			}
			
			while(!interfaceList.get(0).equals(keeperSuccessor)) {
				addTrans();
				addPerm();
			}
			
			addTrans();
			keeperIndex--;
		}
		
		//use perm letters until the correct order is shifted
		while(!interfaceList.equals(sortList)) {
			addPerm();
		}
	}
	
	/**
	 * Deletes all node names which don't occur in the {@code neededList} interface
	 * @param interfaceList the current interface
	 * @param neededList the list of nodes which have to be kept in the interface
	 */
	private void cleanInterface(List<Node> interfaceList, List<Node> neededList) {
		int lastIndex = interfaceList.size()-1;
		int indexOfFirstNode = 0;
		
		//Search for the first node that needs to be removed
		for(int i=0; i< interfaceList.size(); i++) {
			if(!neededList.contains(interfaceList.get(i))) {
				indexOfFirstNode = i;
				break;
			}
		}
		
		Node firstNode = interfaceList.get(indexOfFirstNode);
		Node lastNode = interfaceList.get(lastIndex);
		
		//Remove all other removable nodes until the previous found node is at the last position
		while(!lastNode.equals(firstNode)) {
			
			if(neededList.contains(lastNode)) {
				addPerm();
			}
			else {
				addRes();
				lastIndex--;
			}
			
			lastNode = interfaceList.get(lastIndex);
		}
		
		addRes();
	}
	
	
	/**
	 * Adds the perm letter to the decomposition
	 */
	private void addPerm() {
		if(currentInterface.size()<2) {
			throw new RuntimeException(
					RavenExceptionMessages.INTERFACE_SIZE_MUST_BE_GREATER_THAN_ONE_EXCEPTION);
		}
		
		//If the current interface is equal to 2 use the TRANS-operation, 
		//otherwise use the PERM-Operation
		operations.add(currentInterface.size()==2 ? Operation.getTrans() : Operation.getPerm());
		currentInterface.add(0, currentInterface.remove(currentInterface.size()-1));
	}
	
	/**
	 * Adds the fuse letter to the decomposition
	 */
	private void addFuse() {
		if(currentInterface.size()<2) {
			throw new RuntimeException(
					RavenExceptionMessages.INTERFACE_SIZE_MUST_BE_GREATER_THAN_ONE_EXCEPTION);
		}
		
		operations.add(Operation.getFuse());
		currentInterface.remove(currentInterface.size()-1);
	}
	
	/**
	 * Adds the trans letter to the decomposition
	 */	
	private void addTrans() {
		if(currentInterface.size()<2) {
			throw new RuntimeException(
					RavenExceptionMessages.INTERFACE_SIZE_MUST_BE_GREATER_THAN_ONE_EXCEPTION);
		}
		
		operations.add(Operation.getTrans());
		currentInterface.add(0, currentInterface.remove(1));
	}
	
	/**
	 * Adds the res letter to the decomposition
	 */
	private void addRes() {
		if(currentInterface.size()<2) {
			throw new RuntimeException(
					RavenExceptionMessages.INTERFACE_SIZE_MUST_BE_GREATER_THAN_ZERO_EXCEPTION);
		}
		
		operations.add(Operation.getRes());
		currentInterface.remove(currentInterface.size()-1);
	}
	
	/**
	 * Adds the vertex letter to the decomposition
	 * @param index the index of the node within the reserve nodes
	 */
	private void addVertex(int index) {
		operations.add(Operation.getVertex());
		currentInterface.add(reserveNodes.remove(index));
	}
	
	/**
	 * Adds the connect letter to the decomposition
	 * @param edge the edge which is added
	 */
	private void addConnect(Edge edge) {
		if(currentInterface.size()<2) {
			throw new RuntimeException(RavenExceptionMessages.ILLEGAL_ARITY_EXCEPTION);
		}
		
		operations.add(Operation.getConnect(edge.getLabel(), edge.getArity()));
	}

}
