/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.algorithms;

import java.util.LinkedList;
import java.util.List;
import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.actions.ComputationEvent;
import de.uni_due.inf.ti.raven.actions.ComputationListener;
import de.uni_due.inf.ti.raven.actions.ComputationEvent.EventType;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * Checks whether an automaton accepts the empty language.
 * @author Christoph Blume, Laura Steinert
 */
public class ForwardEmptinessCheck implements EmptinessCheckAlgorithm {

	private final Logger logger;
	private List<ComputationListener> listener;
	private boolean isCancelled;
	
	/**
	 * 
	 */
	public ForwardEmptinessCheck(){
		listener = new LinkedList<>();
		// Register logger
		logger = Logger.getLogger(getClass().getName());
	}
	

	/*------------------------------------------------------------------------------------------*/
	/* 		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public boolean acceptsEmptyLanguage(Automaton automaton) {
		if(automaton==null) {
			throw new NullPointerException(
					RavenExceptionMessages.AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}

		// Log entering
		logger.entering(getClass().getName(), "acceptsEmptyLanguage", automaton);
		fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_STARTED, "", 0, 0));
		
		RavenBDD examined = automaton.getInitialStates();
		RavenBDD prev = examined.clone();
		boolean isEmpty = true;
		
		// check whether an initial state is a final state
		// --> automaton does not accept empty language
		if(automaton.isAcceptingStateSet(examined)) {			
			isEmpty = false;
		}
		
		// while not all reachable states have been examined
		while(isEmpty && !prev.equals(examined) && !isCancelled) {
			prev = examined.clone();
			for (Operation letter : automaton.getAlphabet()) {
				examined = examined.or(automaton.getSuccessorStates(prev, letter));
				if (automaton.isAcceptingStateSet(examined)) {
					isEmpty = false;
				}

				fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED,
						"acceptsEmptyLanguage", automaton.getNumberOfStates(), examined.satCount(
								automaton.getEncoding().getCurrentStateDomain().getDomainBDD())));
			}
		}
		
		if(isCancelled) {
			fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_CANCELLED, "", 0, 0));
		}
		else {
			fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_FINISHED, "", 0, 0));
		}

		// Log exiting
		logger.exiting(getClass().getName(), "acceptsEmptyLanguage", isEmpty);
		return isEmpty;
	}

	@Override
	public void addComputationListener(ComputationListener l) {
		listener.add(l);
	}
	
	@Override
	public void removeComputationListener(ComputationListener l) {
		listener.remove(l);
	}
	
	@Override
	public void cancelComputation() {
		isCancelled = true;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/* 		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Fires an computation event notifying all listeners.
	 * @param e the computation event
	 */
	private void fireComputationEvent(ComputationEvent e) {
		for(ComputationListener l: listener) {
			l.computationAdvanded(e);
		}
	}

}
