/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.algorithms;

import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.RuleCounterExample;
import de.uni_due.inf.ti.raven.data.automata.Automaton;

/**
 * The algorithm interface for checking invariants. Each class which implements an invariant 
 * check must implements this interface.
 * @author Christoph Blume
 * @version $Revision: 1.16 $
 *
 */
public interface InvariantCheckingAlgorithm extends Algorithm {
	
	/**
	 * An enumeration of implementations of this algorithm. 
	 * @author Christoph Blume
	 *
	 */
	public enum Implementations {
		/** Antichain-based invariant checking algorithm */
		AntichainInvariantChecking,
		
		/** Simulation-based invariant checking algorithm */
		SimulationInvariantChecking,
		
		/** Bisimulation-up-to-congruence invariant checking algorithm */
		BisimulationUpToCongruence
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns {@code true} if and only if the language of the automaton {@code automaton} is an 
	 * invariant according to the graph transformation rule with the left-hand side {@code 
	 * leftHandSide} and the right-hand side {@code rightHandSide}.
	 * @param automaton the automaton representing the language
	 * @param leftHandSide the left-hand side of the transformation rule
	 * @param rightHandSide the right-hand side of the transformation rule
	 * @return {@code true} if the language is an invariant, {@code false} otherwise
	 */
	public boolean isInvariant(Automaton automaton, CospanDecomposition leftHandSide, 
			CospanDecomposition rightHandSide);
	
	/**
	 * Returns the counter example, if the language of the automaton is not 
	 * an invariant for the graph transformation rule, and {@code null} otherwise
	 * @return the counter example (if the invariant check failed) or {@code null}
	 * otherwise
	 */
	public RuleCounterExample getInvariantCounterExample();
	
}
