/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.algorithms;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.logging.Logger;

import nl.uu.cs.treewidth.algorithm.Permutation;
import nl.uu.cs.treewidth.input.GraphInput.InputData;
import nl.uu.cs.treewidth.ngraph.ListGraph;
import nl.uu.cs.treewidth.ngraph.ListVertex;
import nl.uu.cs.treewidth.ngraph.NGraph;
import nl.uu.cs.treewidth.ngraph.NVertex;
import nl.uu.cs.treewidth.ngraph.NVertexOrder;
import de.uni_due.inf.ti.raven.actions.ComputationEvent;
import de.uni_due.inf.ti.raven.actions.ComputationEvent.EventType;
import de.uni_due.inf.ti.raven.actions.ComputationListener;
import de.uni_due.inf.ti.raven.data.Edge;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.GraphDecomposition;
import de.uni_due.inf.ti.raven.data.Node;
import de.uni_due.inf.ti.raven.data.TreeDecomposition;

/**
 * This class is the abstract basis for all decomposition algorithms from LibTW library. This class 
 * implements 
 * @author Christoph Blume
 * @version $Revision: 1.13 $
 *
 */
public abstract class LibTWAlgorithm implements GraphDecompositionAlgorithm {

	private LinkedList<ComputationListener> listener;
	private Logger logger;
	private Permutation<InputData> algorithm;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new instance of a LibTW decomposition algorithm.
	 * @param algorithm the algorithm to be used
	 */	
	protected LibTWAlgorithm(Permutation<InputData> algorithm) {
		this.algorithm = algorithm;
		listener = new LinkedList<>();

		// Register logger
		logger = Logger.getLogger(getClass().getName());
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public void addComputationListener(ComputationListener l) {
		listener.add(l);
	}

	@Override
	public void removeComputationListener(ComputationListener l) {
		listener.remove(l);
	}
	
	@Override
	public GraphDecomposition decompose(Graph graph) {
		// Log entering
		logger.entering(getClass().getName(), "decompose", graph);
		
		fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_STARTED, "", 0, 0));

		NGraph<InputData> graphLibTW = new ListGraph<>();
		Map<Node, Integer> nodeMap = new HashMap<>();
		
		int i=0;
		for(Node node: graph.getNodes()) {
			nodeMap.put(node, i);
			graphLibTW.addVertex(new ListVertex<>(
					new InputData(i++, node.getName())));
		}
		
		for(Edge edge: graph.getEdges()) {
			int firstIndex = nodeMap.get(edge.getIncidentNodes().get(0));
			int secondIndex = nodeMap.get(edge.getIncidentNodes().get(1));
			NVertex<InputData> firstNode = graphLibTW.getVertex(firstIndex);
			NVertex<InputData> secondNode = graphLibTW.getVertex(secondIndex);
			
			graphLibTW.addEdge(firstNode, secondNode);
		}	

		algorithm.setInput(graphLibTW);
		algorithm.run();
		NVertexOrder<InputData> perm = algorithm.getPermutation();
		
		de.uni_due.inf.ti.raven.data.Permutation permutation = new de.uni_due.inf.ti.raven.data.Permutation(graph);
		for(NVertex<InputData> nVertex: perm.order) {
			permutation.addNode(graph.getNode(nVertex.data.name));
		}
		
		for(Node node: graph.getNodes()) {
			if(!permutation.contains(node)) {
				return null;
			}
		}
		
		TreeDecomposition decomposition = TreeDecomposition.getTreeDecomposition(permutation);
		
		fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_FINISHED, "", 0, 0));
		
		// Log exiting
		logger.exiting(getClass().getName(), "decompose", decomposition);
		
		return decomposition;
	}

	@Override
	public void cancelComputation() {
		// Do nothing (as long as the computation can not be cancelled reasonably
	}
	
	/**
	 * Fires a computation event notifying all listeners.
	 * 
	 * @param e the computation event
	 */
	private void fireComputationEvent(ComputationEvent e) {
		for (ComputationListener l : listener) {
			l.computationAdvanded(e);
		}
	}

}
