/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.algorithms;

import java.util.LinkedList;
import java.util.List;

import nl.uu.cs.treewidth.input.GraphInput.InputData;
import nl.uu.cs.treewidth.ngraph.NEdge;
import nl.uu.cs.treewidth.ngraph.NGraph;
import nl.uu.cs.treewidth.ngraph.NVertex;
import de.uni_due.inf.ti.raven.actions.ComputationListener;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.Node;

/**
 * Class to convert a LibTW graph (NGraph) into a (Raven) graph.
 * @author Laura Steinert
 * @version $Revision: 1.13 $
 *
 */
public class LibTWNGraphToGraphAlgorithm implements Algorithm {

	private static final String DEFAULT_LABEL = Graph.DEFAULT_EDGE_LABEL_STRING;

	private LinkedList<ComputationListener> listener;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Converts a LibTW graph (nGraph) into a (Raven) graph.
	 * @param graphName the name of the graph
	 * @param nGraph the LibTW graph to be converted
	 * @return the resulting (Raven) graph
	 */
	public static Graph convert(String graphName, NGraph<InputData> nGraph){
		Graph graph = new Graph(graphName);
		// add all nodes
		for (NVertex<InputData> v : nGraph){
			graph.addNode("n"+v.data.id);
		}
		
		// add all edges
		int i=0;
		for(NEdge<InputData> e : nGraph.edges()){
			List<Node> incidentNodes = new LinkedList<>();
			incidentNodes.add(graph.getNode("n" + e.a.data.id));
			incidentNodes.add(graph.getNode("n" + e.b.data.id));
			graph.addEdge("e"+i, DEFAULT_LABEL, incidentNodes);
			i++;
		}
		return graph;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * 
	 */
	public LibTWNGraphToGraphAlgorithm() {
		listener = new LinkedList<>();
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public void addComputationListener(ComputationListener l) {
		listener.add(l);
	}

	@Override
	public void removeComputationListener(ComputationListener l) {
		listener.remove(l);
	}

	@Override
	public void cancelComputation() {
		// Do nothing (as long as the computation can not be cancelled reasonably
	}

}
