/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.algorithms;

import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;

/**
 * The algorithm interface for the computation of the simulation relation. Each class which computes
 * the simulation relation must implement this interface.
 * @author Christoph Blume
 * @version $Revision: 1.19 $
 *
 */
public interface SimulationAlgorithm extends Algorithm {

	/**
	 * An enumeration of implementations of this algorithm. 
	 * @author Christoph Blume
	 *
	 */
	public enum Implementations {
		/** Forward Simulation Preorder */
		ForwardSimulationAlgorithm,
		
		/** Backward Simulation Preorder */
		BackwardSimulationAlgorithm,
		
//		ForwardBisimulationUpToCongruenceAlgorithm
	}
	
	/** Simulation Iteration Step Counter String Literal */
	public static final String SIMULATION_ITERATION_STEP = "[simulation] Iteration step: ";
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns {@code true} if and only if the state set {@code firstStateSet} of the automaton 
	 * {@code automaton} is simulated by the second state set {@code secondStateSet} of the 
	 * automaton {@code automaton}.
	 * @param automaton the automaton
	 * @param firstStateSet the set of states to be simulated
	 * @param secondStateSet the set of states simulate
	 * @return {@code true} if the first state set is simulated by the second state set, 
	 * {@code false} otherwise
	 */
	public boolean isSimulatedBy(Automaton automaton, RavenBDD firstStateSet,
			RavenBDD secondStateSet);

	/**
	 * Returns {@code true} if and only if the state set {@code firstStateSet} of the first 
	 * automaton {@code firstAutomaton} is simulated by the state set {@code secondStateSet}
	 * of the second automaton {@code secondAutomaton}.
	 * @param firstAutomaton the first automaton
	 * @param secondAutomaton the second automaton
	 * @param firstStateSet the set of states to be simulated
	 * @param secondStateSet the set of states simulate
	 * @return {@code true} if the first state set is simulated by the second state set, 
	 * {@code false} otherwise
	 */
	public boolean isSimulatedBy(Automaton firstAutomaton, Automaton secondAutomaton, 
			RavenBDD firstStateSet, RavenBDD secondStateSet);
	
	/**
	 * Returns the simulation relation on the states of the automaton {@code automaton}.
	 * @param automaton the automaton
	 * @return the simulation relation for the given automaton
	 */
	public RavenBDD getSimulation(Automaton automaton);
	
	/**
	 * Returns the simulation relation on the states of the automata {@code firstAutomaton} 
	 * and {@code secondAutomaton}.
	 * @param firstAutomaton the first automaton
	 * @param secondAutomaton the second automaton
	 * @return the simulation relation for the given automata
	 */
	public RavenBDD getSimulation(Automaton firstAutomaton, Automaton secondAutomaton);
	
}
