/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.algorithms;

import java.util.LinkedList;
import java.util.List;
import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.actions.ComputationListener;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.RuleCounterExample;
import de.uni_due.inf.ti.raven.data.automata.AutomatonFactory;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class implements the two antichain algorithms for checking invariants. The algorithms 
 * are based on the antichain algorithms for language inclusion by Henzinger et. al. More 
 * details about the algorithms can be found in 
 * "Efficient Symbolic Implementation of Graph Automata with Applications to Invariant Checking"
 * <a href="#">Springerlink</a>.
 * 
 * This class implements the two simulation-based antichain algorithms for checking invariants. 
 * The algorithms are based on the simulation-based antichain algorithms for language inclusion by 
 * Abdulla et. al. More details about the (latter) algorithms can be found in "When Simulation 
 * Meets Antichains (on Checking Language Inclusion on NFAs)" 
 * <a href="http://user.it.uu.se/~parosh/publications/papers/tacas10.pdf">Homepage</a> 
 * of Parosh Aziz Abdulla.
 * @author Christoph Blume
 * @version $Revision: 1.19 $
 *
 */
public final class SimulationAntichainInvariantChecking implements InvariantCheckingAlgorithm {

	private final Logger logger;
	private SearchType searchType;
	private boolean computeCounterExample;
	private List<ComputationListener> listener;
	private RuleCounterExample counterExample;
	private SimulationAntichainLanguageInclusion languageInclusionAlgorithm;

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new instance of the simulation-based invariant checking algorithm.
	 * @param searchType the search type used by the algorithm, either {@code forward} or 
	 * {@code backward}
	 * @param computeCounterExample if {@code true}, the algorithm will compute a counter-example
	 * @see Algorithm.SearchType
	 */
	public SimulationAntichainInvariantChecking(SearchType searchType, boolean computeCounterExample) {
		if(searchType==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SEARCH_TYPE_MUST_NOT_BE_NULL_EXCEPTION);
		}

		this.searchType = searchType;
		this.computeCounterExample = computeCounterExample;
		listener = new LinkedList<>();

		// Register logger
		logger = Logger.getLogger(getClass().getName());
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public void addComputationListener(ComputationListener l) {
		listener.add(l);
	}

	@Override
	public void removeComputationListener(ComputationListener l) {
		listener.remove(l);
	}

	@Override
	public boolean isInvariant(Automaton automaton,
			CospanDecomposition leftHandSide, CospanDecomposition rightHandSide) {
		if(automaton==null) {
			throw new NullPointerException(
					RavenExceptionMessages.AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(leftHandSide==null) {
			throw new NullPointerException(
					RavenExceptionMessages.LHS_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(rightHandSide==null) {
			throw new NullPointerException(
					RavenExceptionMessages.RHS_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		// Log entering
		logger.entering(getClass().getName(), "isInvariant", new Object[] {automaton,
			leftHandSide, rightHandSide});

		boolean result = false;
		Automaton leftHandSideAutomaton = AutomatonFactory.shiftAutomaton(automaton, leftHandSide);
		Automaton rightHandSideAutomaton = AutomatonFactory.shiftAutomaton(automaton, rightHandSide);

		languageInclusionAlgorithm = new SimulationAntichainLanguageInclusion(searchType, 
				computeCounterExample);
		result = languageInclusionAlgorithm.isIncluded(leftHandSideAutomaton, rightHandSideAutomaton);

		if(computeCounterExample && !result) {
			counterExample = new RuleCounterExample(leftHandSide, rightHandSide, 
					languageInclusionAlgorithm.getCounterExample().getCospan());
		}

		// Log exiting
		logger.exiting(getClass().getName(), "isInvariant", result);
		return result;


	}

	@Override
	public RuleCounterExample getInvariantCounterExample() {
		return counterExample;
	}

	@Override
	public void cancelComputation() {
		languageInclusionAlgorithm.cancelComputation();
	}

}
