/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.algorithms;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.actions.ComputationEvent;
import de.uni_due.inf.ti.raven.actions.ComputationEvent.EventType;
import de.uni_due.inf.ti.raven.actions.ComputationListener;
import de.uni_due.inf.ti.raven.data.CospanCounterExample;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.Operation.Connect;
import de.uni_due.inf.ti.raven.data.Operation.Fuse;
import de.uni_due.inf.ti.raven.data.Operation.Perm;
import de.uni_due.inf.ti.raven.data.Operation.Res;
import de.uni_due.inf.ti.raven.data.Operation.Trans;
import de.uni_due.inf.ti.raven.data.Operation.Vertex;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDDomain;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDFactory.BDDOperator;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDPairing;
import de.uni_due.inf.ti.raven.encodings.BDDEncoding.StateType;
import de.uni_due.inf.ti.raven.main.RavenMain;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class implements the two simulation-based antichain algorithms deciding
 * the language inclusion problem for (finite) automata. The algorithms are
 * based on the work of Abdulla et. al. More details about the algorithms can be
 * found in
 * "When Simulation Meets Antichains (on Checking Language Inclusion on NFAs)"
 * <a href="http://user.it.uu.se/~parosh/publications/papers/tacas10.pdf">
 * Homepage</a> of Parosh Aziz Abdulla.
 * 
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.60 $
 * 
 */
public final class SimulationAntichainLanguageInclusion implements
		LanguageInclusionAlgorithm {

	/**
	 * A pair of BDDs. The first component represents a single state (of the first automaton),
	 * whereas the second component represents a set of states (of the second automaton).
	 * @author Christoph Blume
	 * 
	 */
	private class BDDPair {

		private RavenBDD first;
		private RavenBDD second;
		private int size;
		private List<Operation> list;

		/*------------------------------------------------------------------------------------------*/
		/*		Constructors																		*/
		/*------------------------------------------------------------------------------------------*/

		public BDDPair(RavenBDD first, RavenBDD second, int size, List<Operation> list) {
			this.first = first;
			this.second = second;
			this.size = size;
			this.list = list;
		}

		public BDDPair(RavenBDD first, RavenBDD second, int size) {
			this(first, second, size, new LinkedList<Operation>());
		}

		/*------------------------------------------------------------------------------------------*/
		/* 		Public Methods																		*/
		/*------------------------------------------------------------------------------------------*/

		public RavenBDD getFirst() {
			return first.clone();
		}

		public RavenBDD getSecond() {
			return second.clone();
		}

		public List<Operation> getList() {
			return list;
		}

		@Override
		public boolean equals(Object o) {
			if (o==null) {
				return false;
			}
			else if (o==this) {
				return true;
			}
			else if (o.getClass().equals(getClass())) {
				BDDPair pair = (BDDPair) o;
				return first.equals(pair.first) && second.equals(pair.second);
			}
			else {
				return false;
			}
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + ((first == null) ? 0 : first.hashCode());
			result = prime * result
					+ ((second == null) ? 0 : second.hashCode());
			return result;
		}

		@Override
		public String toString() {
			return "("+firstAutomaton.printStates(first)+", "+secondAutomaton.printStates(second)+", "+list+")";
		}

	}

	private final IdentityRelationAlgorithm identityRelationAlgorithm;
	private final SimulationAlgorithm simulationAlgorithm;
	private final SearchType type;
	private final boolean computeCounterExample;

	private Map<Automaton, RavenBDD> orderingMap;
	private boolean verboseMode;
	private boolean isCancelled;
	private Automaton firstAutomaton;
	private Automaton secondAutomaton;
	private RavenBDDDomain firstCurrentStateDomain;
	private RavenBDDDomain firstSuccessorStateDomain;
	private RavenBDDDomain secondCurrentStateDomain;
	private RavenBDDDomain secondSuccessorStateDomain;
	
	private RavenBDDPairing firstAutomatonPairing;
	private RavenBDDPairing firstSwitchStatesPairing;
	private RavenBDDPairing secondAutomatonPairing;
	private RavenBDDPairing secondSwitchStatesPairing;
	private RavenBDDPairing currentToSuccessorPairing;
	
	private final Logger logger;
	private List<ComputationListener> listener;
	private CospanCounterExample counterExample;

	int count = 0;
	private Map<Automaton, RavenBDD> lexicographicOrderings;

	/*------------------------------------------------------------------------------------------*/
	/* 		Constructors 																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new instance of the antichain-based language inclusion
	 * algorithm.
	 * 
	 * @param searchType the search type of this instance
	 * @param computeCounterExample if {@code true} a counterexample will be computed if the 
	 * language inclusion check fails
	 * @see Algorithm.SearchType
	 */
	public SimulationAntichainLanguageInclusion(SearchType searchType,
			boolean computeCounterExample) {
		if(searchType==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SEARCH_TYPE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		this.type = searchType;
		this.computeCounterExample = computeCounterExample;
		
		orderingMap = new HashMap<>();
		identityRelationAlgorithm = new IdentityRelationAlgorithm();
		lexicographicOrderings = new HashMap<>();
		simulationAlgorithm = new ForwardSimulationAlgorithm();
		listener = new LinkedList<>();
		
		if(isCancelled) {
			fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_CANCELLED, "", 0, 0));
		}
		else {
			fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_FINISHED, "", 0, 0));
		}

		// Register logger
		logger = Logger.getLogger(getClass().getName());
	}

	/**
	 * Creates a new instance of the antichain-based language inclusion algorithm.
	 * @param searchType the search type of this instance
	 * @param orderingType the type of ordering which is to be used
	 * 
	 * @see Algorithm.SearchType
	 */
	public SimulationAntichainLanguageInclusion(SearchType searchType,
			OrderingType orderingType) {
		this(searchType, false);
	}

	/*------------------------------------------------------------------------------------------*/
	/* 		Public Methods 																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public void addComputationListener(ComputationListener l) {
		listener.add(l);
	}

	@Override
	public void removeComputationListener(ComputationListener l) {
		listener.remove(l);
	}

	@Override
	public boolean isIncluded(Automaton first, Automaton second) {
		if(first==null) {
			throw new NullPointerException(
					RavenExceptionMessages.FIRST_AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(second==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SECOND_AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(first.getMaximumInterface()!=second.getMaximumInterface()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_INTERFACE_SIZES_MUST_NOT_BE_DIFFERENT_EXCEPTION);
		}
		if(first.getInnerInterface()!=second.getInnerInterface()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZES_MUST_NOT_BE_DIFFERENT_EXCEPTION);
		}
		if(first.getOuterInterface()!=second.getOuterInterface()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZES_MUST_NOT_BE_DIFFERENT_EXCEPTION);
		}
		
		// Log entering
		logger.entering(getClass().getName(), "isIncluded", new Object[] {
				first, second });
		verboseMode = Boolean.valueOf(RavenMain.getProperty(RavenMain.PROPERTY_VERBOSE));

		isCancelled = false;
		firstAutomaton = first;
		secondAutomaton = second;
		boolean result = false;
		
		//Get state domains
		firstCurrentStateDomain = firstAutomaton.getEncoding().getCurrentStateDomain();
		firstSuccessorStateDomain = firstAutomaton.getEncoding().getSuccessorStateDomain();
		
		secondCurrentStateDomain = secondAutomaton.getEncoding().getCurrentStateDomain();
		secondSuccessorStateDomain = secondAutomaton.getEncoding().getSuccessorStateDomain();

		//Make pairings
		firstAutomatonPairing = RavenBDDPairing.makePair();
		firstAutomatonPairing.set(firstCurrentStateDomain, firstSuccessorStateDomain);

		secondAutomatonPairing = RavenBDDPairing.makePair();
		if(first.equals(second)) {
			secondAutomatonPairing.set(secondCurrentStateDomain, secondSuccessorStateDomain);
		}
		
		firstSwitchStatesPairing = RavenBDDPairing.makePair();
		firstSwitchStatesPairing.set(firstCurrentStateDomain, firstSuccessorStateDomain);
		firstSwitchStatesPairing.set(firstSuccessorStateDomain, firstCurrentStateDomain);
		
		secondSwitchStatesPairing = RavenBDDPairing.makePair();
		secondSwitchStatesPairing.set(secondCurrentStateDomain, secondSuccessorStateDomain);
		secondSwitchStatesPairing.set(secondSuccessorStateDomain, secondCurrentStateDomain);

		switch (type) {
		case forward:
			try {
				result = performForwardSearch();
			}
			catch (Exception e) {
				logger.throwing(getClass().getName(), "isIncluded", e);
			}
			break;

		case backward:
			secondAutomatonPairing = Automaton.getInterfacePairing(
					first, second);
			currentToSuccessorPairing = RavenBDDPairing.makePair();
			currentToSuccessorPairing.set(secondAutomaton.getEncoding().getCurrentStateDomain(), 
					secondAutomaton.getEncoding().getSuccessorStateDomain());
			result = performBackwardSearch();
			break;

		default:
			throw new RuntimeException(RavenExceptionMessages.ILLEGAL_SEARCH_TYPE_EXCEPTION);
		}

		// Log exiting
		logger.exiting(getClass().getName(), "isIncluded", counterExample);
		return result;
	}

	@Override
	public CospanCounterExample getCounterExample() {
		return counterExample;
	}

	@Override
	public void cancelComputation() {
//		isCancelled = true;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/* 		Private Methods 																	*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns {@code true} if and only if the language of the automaton {@code firstAutomaton} is 
	 * a subset of the language of the automaton {@code secondAutomaton}. The computation is 
	 * performed via a forward search based on the simulation antichain algorithm by Abdulla et. al.
	 * 
	 * @return {@code true} if the first language is included in the second language, 
	 * {@code false} otherwise
	 */
	private boolean performForwardSearch() {
		RavenBDD vars = firstAutomaton.getEncoding().getCurrentStateDomain().getDomainBDD();
		List<BDDPair> next = new LinkedList<>();
		Set<BDDPair> processed = new HashSet<>();

		// If the first automaton is in an accepting state and the second automaton is not,
		// return false since the empty word is a counter example
		if (firstAutomaton.isAcceptingStateSet(firstAutomaton .getInitialStates()) && 
				!secondAutomaton.isAcceptingStateSet(secondAutomaton.getInitialStates())) {
			if (computeCounterExample) {
				counterExample = new CospanCounterExample(new CospanDecomposition(
						firstAutomaton.getInnerInterface()));
			}
			return false;
		}

		// Create the initial state pairs
		for (RavenBDD bdd : firstAutomaton.getInitialStates().getSatAssignments(vars)) {
			BDDPair nextPair = new BDDPair(bdd.clone(), secondAutomaton.getInitialStates(), 0);
			next.add(reduce(nextPair));
			if (computeCounterExample) {
				nextPair.list = new LinkedList<>();
			}
		}
		vars.free();
		
		next = initialize(next);
		
		int i = 0;
		while (!next.isEmpty() && !isCancelled) {
			if(verboseMode) {
				i++;
				logger.info(LANGUAGE_INCLUSION_ITERATION_STEP+i);
			}

			// Pick and remove a BDD pair from next and move it to processed
			BDDPair pair = next.iterator().next();
			processed.add(pair);
			next.remove(pair);

			Set<BDDPair> results = computePost(pair);
			int maxRounds = results.size();
			int counter =0;
			
			//Compute new frontier
			for(BDDPair successorPair: results) {
				fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED, 
						"Forward Search", maxRounds, ++counter));
				
				if (isRejectingPairForwards(successorPair)) {
					counterExample = new CospanCounterExample(
							new CospanDecomposition(successorPair.getList()));
					return false;
				} 
				else if (!isPairOrdered(successorPair)) {
					Set<BDDPair> union = new HashSet<>();
					union.addAll(next);
					union.addAll(processed);

					boolean isIncomparable = true;
					for (BDDPair element : union) {						
						if (areStatesOrdered(successorPair, element)
								&& areStateSetsOrdered(element, successorPair)) {
							successorPair.first.free();
							successorPair.second.free();
							isIncomparable = false;
							break;
						}
					}

					if (isIncomparable) {
						Set<BDDPair> smallerElements = new HashSet<>();
						for (BDDPair element : union) {
							if (areStatesOrdered(element, successorPair)
									&& areStateSetsOrdered(successorPair, element)) {
								smallerElements.add(element);
							}
						}

						removePairs(smallerElements, next);
						removePairs(smallerElements, processed);
						next.add(successorPair);
					}
				}
			}
		}

		return true;
	}

	/**
	 * Returns {@code true} if and only if the language of the automaton
	 * {@code firstAutomaton} is a subset of the language of the automaton
	 * {@code secondAutomaton}. The computation is performed via a backward
	 * search based on the simulation antichain algorithm by Abdulla et. al.
	 * 
	 * @return {@code true} if the first language is included in the second
	 *         language, {@code false} otherwise
	 */
	@SuppressWarnings("static-method")
	private boolean performBackwardSearch() {
		// First, the theory needs to be checked
		return false;
	}

	/**
	 * Returns {@code true} if and only if the pair {@code pair} is rejecting,
	 * i.e. if the state (of the first automaton) is accepting and the state set
	 * (of the second automaton) is rejecting.
	 * 
	 * @param pair the pair to be checked
	 * @return {@code true} if the pair is rejecting, {@code false} otherwise
	 */
	private boolean isRejectingPairForwards(BDDPair pair) {
		if(pair.size!=firstAutomaton.getOuterInterface()) {
			return false;
		}

		return (firstAutomaton.isAcceptingStateSet(pair.getFirst()) && 
				pair.getSecond().and(secondAutomaton.getFinalStates()).isZero());
		
	}

	/**
	 * This method implements Optimization 1. It does the following: It returns
	 * the maximum subset of the given set of pairs, i.e. it is guaranteed that
	 * every two pairs are incomparable and that for every pair the state (of
	 * the first automaton) is not related to any state in the state set (of the
	 * second automaton)
	 * 
	 * (For more details about Optimization 1: see the paper mentioned above)
	 * 
	 * @param listOfPairs the set of pairs whose maximal subset is wanted
	 * @return the maximal subset of the set of pairs
	 * 
	 */
	private List<BDDPair> initialize(List<BDDPair> listOfPairs) {
		List<BDDPair> optimizedList = new LinkedList<>();
		
		for (BDDPair firstPair : listOfPairs) {
			boolean statesOrdered = false;
			boolean setsOrdered = false;
			boolean pairOrdered = false;

			// Implement the first part of Optimization 1
			// Checks that each two pairs (p,P), (q,Q) are pair-wise incomparable,
			// i.e. it holds either NOT(p <= q) or NOT(for all states s in Q
			// exists a state t in P s.t. s <= t)
			for (BDDPair secondPair : listOfPairs) {
				if (firstPair.equals(secondPair)) {
					continue;
				}

				statesOrdered = areStatesOrdered(firstPair, secondPair);
				setsOrdered = areStateSetsOrdered(firstPair, secondPair);

				if (statesOrdered & setsOrdered) {
					break;
				}
			}

			// Implement the second part of Optimization 1
			// Checks for each pair (p,P) that for all states s in P it holds
			// NOT(p <= s)
			pairOrdered = isPairOrdered(firstPair);

			if ((!statesOrdered || !setsOrdered) & !pairOrdered) {
				optimizedList.add(firstPair);
			}
		}
		
		listOfPairs.removeAll(optimizedList);
		for(BDDPair pair: listOfPairs) {
			pair.first.free();
			pair.second.free();
		}		

		return optimizedList;
	}

	/**
	 * This method implements Optimization 2. It does the following: it chooses
	 * a state <i>s</i> from the state set of the pair {@code pair}, removes
	 * <i>s</i> from the state set, if there exists a state <i>t</i> in the
	 * state set such state <i>s &lt;= t</i>, and then repeats the procedure
	 * until all of the states in the state set are processed. Afterwards the
	 * minimized pair is returned.
	 * 
	 * (Note: <i>&lt;</i> is the ordering which implies that the state <i>s</i>
	 * is simulated by the state <i>t</i>.)
	 * (For more details about Optimization 2: see the paper mentioned above)
	 * 
	 * @param pair the pair which is minimized
	 * @return the minimized pair
	 * 
	 */
	private BDDPair reduce(BDDPair pair) {
//		RavenBDDPairing automatonPairing = RavenBDDPairing.makePair();
//		automatonPairing.set(secondCurrentStateDomain, secondSuccessorStateDomain);
//
//		RavenBDD stateSetProduct = pair.getSecond().and(pair.getSecond().replace(automatonPairing));
//		RavenBDD tempBDD = stateSetProduct.applyEx(getOrdering(secondAutomaton), BDDOperator.and,
//				secondSuccessorStateDomain.getDomainBDD());		
//		tempBDD = pair.getSecond().and(tempBDD.not());
//
//		pair.second.free();
//		return new BDDPair(pair.first, tempBDD, pair.size, pair.list);
		return new BDDPair(pair.first.clone(), pair.second.clone(), pair.size, pair.list);
	}

	/**
	 * Computes the post-image pairs for the BDD pair {@code pair}, i.e. the
	 * following set will be computed Post(<i>pair = (s,S)</i>) = {<i>(t,T)</i>
	 * | exists letter <i>l</i> s.t. <i>t</i> is the <i>l</i>-successor of s and
	 * <i>T</i> is the set of <i>l</i>-successors of some state in <i>S</i>}.
	 * 
	 * @param pair the pair of a state and a state set whose "successor pairs" will be computed
	 * @return the set of "successor" pairs
	 */
	private Set<BDDPair> computePost(BDDPair pair) {
		Set<BDDPair> result = new HashSet<>();
		RavenBDD vars = firstCurrentStateDomain.getDomainBDD();
		int rounds = firstAutomaton.getAlphabet().size();
		int counter = 0;

		for (Operation letter : firstAutomaton.getAlphabet()) {
			if(isCancelled) {
				return result;
			}
			
			if (isApplicable(letter, pair)) {				
				//Get the automata's successors
				RavenBDD successorsFirst = firstAutomaton.getSuccessorStates(pair.getFirst(), letter);
				RavenBDD successorsSecond = secondAutomaton.getSuccessorStates(pair.getSecond(), letter);
				
				if (!successorsFirst.isZero()) {
					for (RavenBDD bdd : successorsFirst.getSatAssignments(vars)) {
						BDDPair successorPair = new BDDPair(bdd.clone(), successorsSecond.clone(), 
								getSuccessorSize(letter, pair));
						
						//If enabled update counter example
						if (computeCounterExample) {
							successorPair.list = new LinkedList<>(pair.list);
							successorPair.list.add(letter);
						}

						result.add(reduce(successorPair));
						count++;
					}
				}

				//Free variables which are not used any longer
				successorsFirst.free();
				successorsSecond.free();
				fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED,
						"fPost", rounds, ++counter));
			}
		}
		
		//Free variables which are not used any longer
		vars.free();

		return result;
	}

	/**
	 * Returns {@code true} if and only if the single state <i>s</i> of the
	 * first pair {@code firstePair} is related to the single state <i>t</i> of
	 * the second pair {@code secondPair}, i.e. <i>s &lt;= t</i> holds.
	 * 
	 * (Note: <i>&lt;=</i> is the ordering which implies that the state <i>s</i>
	 * is simulated by the state <i>t</i>.)
	 * 
	 * @param firstPair the first pair
	 * @param secondPair the second pair
	 * @return {@code true} if the state of the first pair is related to the
	 *         state of the second pair, {@code false} otherwise
	 */
	private boolean areStatesOrdered(BDDPair firstPair, BDDPair secondPair) {
		// Check whether p simulates q
		RavenBDD p = firstPair.getFirst();
		RavenBDD q = secondPair.getFirst();
		RavenBDD tempBDD;
		
		//Equal states are always ordered (due to reflexivity)
		if(p.equals(q)) {
			return true;
		}

		q = q.replace(firstAutomatonPairing);
		
		RavenBDD ordering = getOrdering(firstAutomaton);
		
		tempBDD = p.applyEx(ordering, BDDOperator.and, firstCurrentStateDomain.getDomainBDD());
		p.free();
		tempBDD = q.applyEx(tempBDD, BDDOperator.and, firstSuccessorStateDomain.getDomainBDD());
		q.free();
		
		//Free temporary BDD to release allocated resources
		boolean result = tempBDD.isOne();
		tempBDD.free();
		
		return result;
	}

	/**
	 * Returns {@code true} if and only if the first state set
	 * {@code firsteStateSet} is related to the second state set
	 * {@code secondStateSet}, i.e. for all states <i>s</i> of the first state
	 * set exist a state <i>t</i> in the second state set s.t. <i>s &lt;= t</i>.
	 * 
	 * (Note: <i>&lt;=</i> is the ordering which implies that the state <i>s</i>
	 * is simulated by the state <i>t</i>.)
	 * 
	 * @param firstPair the first pair
	 * @param secondPair the second pair
	 * @return {@code true} if the first state set is related to the second
	 *         state set, {@code false} otherwise
	 */
	private boolean areStateSetsOrdered(BDDPair firstPair, BDDPair secondPair) {
		RavenBDD P = firstPair.getSecond();
		RavenBDD Q = secondPair.getSecond();
		
		//Equal state sets are always ordered (due to reflexivity)
		if(P.equals(Q)) {
			return true;
		}
		
		Q = Q.replace(secondSwitchStatesPairing);
		
		RavenBDD ordering = getOrdering(secondAutomaton);
		RavenBDD tempBDD = Q.applyEx(ordering, BDDOperator.and, 
				secondSuccessorStateDomain.getDomainBDD());
		Q.free();
		tempBDD = P.applyAll(tempBDD, BDDOperator.imp,
				secondCurrentStateDomain.getDomainBDD());
		P.free();
		
		//Free temporary BDD to release allocated resources
		boolean result = tempBDD.isOne();
		tempBDD.free();
		
		return result;
	}

	/**
	 * Returns {@code true} if and only if the pair is ordered with respect to
	 * the used ordering, i.e. the state (in the first component) is related to
	 * the state set (in the second component)
	 * 
	 * @param pair the pair to be checked
	 * @return {@code true} if the pair is ordered, {@code false} otherwise
	 */
	private boolean isPairOrdered(BDDPair pair) {
		RavenBDD state = pair.getFirst();
		RavenBDD stateSet = pair.getSecond().replace(secondAutomatonPairing);

		RavenBDD tempBDD = getOrdering(firstAutomaton, secondAutomaton).and(state);
		
		//Free temporary BDD to release allocated resources
		boolean result = stateSet.applyEx(tempBDD, BDDOperator.and,
				secondSuccessorStateDomain.getDomainBDD()).isOne();
		stateSet.free();
		tempBDD.free();
		
		return result;
	}

	/**
	 * Returns the BDD representing the used ordering relation.
	 * 
	 * @return the ordering relation BDD
	 */
	private RavenBDD getOrdering(Automaton automaton) {
//		return identityRelationAlgorithm.getIdentityRelation(automaton);
		if(!orderingMap.containsKey(automaton)) {

			//Get the simulation preorder
			RavenBDD simulation = simulationAlgorithm.getSimulation(automaton);

			if(automaton==secondAutomaton) {
				// Get the identity relation (used to remove the "reflexive part" of the
				// language-inclusion-implying ordering)
				RavenBDD identity = identityRelationAlgorithm.getIdentityRelation(secondAutomaton);
				//					RavenBDD identity = secondAutomaton.getEncoding().zero();

				// Reverse the language-inclusion-implying ordering
				RavenBDD reversedSimulation = simulation.clone().replace(secondSwitchStatesPairing);

				simulation = simulation.and(identity.not().and(
						reversedSimulation).imp(getLexicographicOrdering(secondAutomaton)));
			}
			orderingMap.put(automaton, simulation);
		}

		return orderingMap.get(automaton).clone();
	}

	/**
	 * Returns the BDD representing the used ordering relation.
	 * 
	 * @return the ordering relation BDD
	 */
	private RavenBDD getOrdering(Automaton firstAutomaton,
			Automaton secondAutomaton) {
//		return firstAutomaton.getEncoding().zero();

		return simulationAlgorithm.getSimulation(firstAutomaton,
				secondAutomaton);
	}

	/**
	 * Returns the lexicographic ordering of the state sets of the automaton
	 * {@code automaton}. The states are first ordered by the "interface size"
	 * encoded in the bit string of each state. States with a smaller interface
	 * size are smaller than states with a larger interface size. If two states
	 * have the same interface size, then the states are ordered by the
	 * "state information" encoded in the bit string of each state.
	 * 
	 * @return the lexicographic ordering of the state set
	 */
	private RavenBDD getLexicographicOrdering(Automaton automaton) {
		if (!lexicographicOrderings.containsKey(automaton) || 
				lexicographicOrderings.get(automaton)==null) {
			RavenBDD ordering = automaton.getEncoding().zero();

			int interfaceLength = automaton.getEncoding().getInterfaceLength();
			int stateLength = automaton.getEncoding().getStateLength();
			RavenBDD tempBDD = automaton.getEncoding().one();
			for (int i = interfaceLength - 1; i >= 0; i--) {
				if (i < interfaceLength - 1) {
					tempBDD = tempBDD.and(
							automaton.getEncoding().getInterfaceBit(StateType.current, i + 1).biimp(
									automaton.getEncoding().getInterfaceBit(StateType.successor, i + 1)));
				}

				RavenBDD interfaceBDD = automaton.getEncoding().getInterfaceBit(StateType.current, i).not().and(
						automaton.getEncoding().getInterfaceBit(StateType.successor, i));
				ordering = ordering.or(interfaceBDD.and(tempBDD.clone()));
			}

			tempBDD = automaton.getEncoding().one();
			RavenBDD stateBDD = automaton.getEncoding().zero();
			for (int i = stateLength - 1; i >= 0; i--) {
				if (i < stateLength - 1) {
					tempBDD = tempBDD.and(
							automaton.getEncoding().getStateBit(StateType.current, i + 1).biimp(
									automaton.getEncoding().getStateBit(StateType.successor, i + 1)));
				}

				stateBDD = stateBDD.or(automaton.getEncoding().getStateBit(StateType.current, i).not().and(
						automaton.getEncoding().getStateBit(StateType.successor, i)).and(tempBDD.clone()));
			}

			RavenBDD interfaceBDD = automaton.getEncoding().one();
			for (int i = interfaceLength - 1; i >= 0; i--) {
				interfaceBDD = interfaceBDD.and(
						automaton.getEncoding().getInterfaceBit(StateType.current, i).biimp(
								automaton.getEncoding().getInterfaceBit(StateType.successor, i)));
			}

			ordering = ordering.or(interfaceBDD.and(stateBDD));
			
			lexicographicOrderings.put(automaton, ordering);
		}

		return lexicographicOrderings.get(automaton).clone();
	}

	/**
	 * Removes the pairs {@code elements} from the collection {@code listOfPairs} and
	 * frees the BDDs of the removed pairs.
	 * 
	 * @param elements the pairs which are marked for removal
	 * @param collectionOfPairs the collections of pairs from which the marked pairs will be removed
	 */
	private static void removePairs(Set<BDDPair> elements, Collection<BDDPair> collectionOfPairs) {
		for (BDDPair pair : elements) {
			if (collectionOfPairs.contains(pair)) {
				collectionOfPairs.remove(pair);
				pair.first.free();
				pair.second.free();
			}
		}
	}

	/**
	 * Returns {@code true} if and only if the operation {@code letter} is
	 * applicable to the pair {@code pair} (i.e. if it is applicable to the
	 * state set of the pair).
	 * 
	 * @param letter the operation which is to be applied
	 * @param pair the pair to which the operation is applied
	 * @return {@code true} if the operation can be applied, {@code false} otherwise
	 */
	private boolean isApplicable(Operation letter, BDDPair pair) {
		boolean isApplicable = true;

		if (letter instanceof Connect) {
			isApplicable = ((Connect) letter).getArity() <= pair.size;
		}
		else if (letter instanceof Fuse || letter instanceof Perm
				|| letter instanceof Trans) {
			isApplicable = pair.size >= 2;
		}
		else if (letter instanceof Res) {
			isApplicable = pair.size >= 1;
		}
		else if (letter instanceof Vertex) {
			isApplicable = pair.size < firstAutomaton.getMaximumInterface();
		}

		return isApplicable;
	}

	/**
	 * Returns the size of the interface of the state set which is reachable by
	 * applying the operation {@code letter} to the pair {@code pair}.
	 * 
	 * @param letter the operation which is to be applied
	 * @param pair the pair to which the operation is applied
	 * @return the size of the "successor" interface
	 */
	private static int getSuccessorSize(Operation letter, BDDPair pair) {
		int size = pair.size;

		if (letter instanceof Fuse || letter instanceof Res) {
			size -= 1;
		}
		else if (letter instanceof Vertex) {
			size += 1;
		}

		return size;
	}

	/**
	 * Fires a computation event notifying all listeners.
	 * 
	 * @param e the computation event
	 */
	private void fireComputationEvent(ComputationEvent e) {
		for (ComputationListener l : listener) {
			l.computationAdvanded(e);
		}
	}

}
