/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.algorithms;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.actions.ComputationEvent;
import de.uni_due.inf.ti.raven.actions.ComputationEvent.EventType;
import de.uni_due.inf.ti.raven.actions.ComputationListener;
import de.uni_due.inf.ti.raven.data.CospanCounterExample;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.Operation.Connect;
import de.uni_due.inf.ti.raven.data.Operation.Fuse;
import de.uni_due.inf.ti.raven.data.Operation.Perm;
import de.uni_due.inf.ti.raven.data.Operation.Res;
import de.uni_due.inf.ti.raven.data.Operation.Trans;
import de.uni_due.inf.ti.raven.data.Operation.Vertex;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDDomain;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDFactory.BDDOperator;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDPairing;
import de.uni_due.inf.ti.raven.encodings.BDDEncoding.StateType;
import de.uni_due.inf.ti.raven.main.RavenMain;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class implements the two simulation-based antichain algorithms deciding
 * the universality problem for (finite) automata. The algorithms are based on
 * the work of Abdulla et. al. More details about the algorithms can be found in
 * "When Simulation Meets Antichains (on Checking Language Inclusion on NFAs)"
 * <a href="http://user.it.uu.se/~parosh/publications/papers/tacas10.pdf">
 * Homepage</a> of Parosh Aziz Abdulla.
 * 
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.44 $
 * 
 */
public final class SimulationAntichainUniversality implements UniversalityAlgorithm {

	/**
	 * A pair of BDDs. The first component represents a single state (of the first automaton),
	 * whereas the second component represents a set of states (of the second automaton).
	 * @author Christoph Blume
	 * 
	 */
	private class BDDPair {

		private RavenBDD stateSet;
		private int size;
		private List<Operation> list;

		/*------------------------------------------------------------------------------------------*/
		/*		Constructors																		*/
		/*------------------------------------------------------------------------------------------*/

		public BDDPair(RavenBDD stateSet, int size, List<Operation> list) {
			this.stateSet = stateSet;
			this.size = size;
			this.list = list;
		}

		public BDDPair(RavenBDD stateSet, int size) {
			this(stateSet, size, null);
		}

		/*------------------------------------------------------------------------------------------*/
		/* 		Public Methods																		*/
		/*------------------------------------------------------------------------------------------*/

		public List<Operation> getList() {
			return list;
		}

		public RavenBDD getStateSet() {
			return stateSet.clone();
		}

		@Override
		public boolean equals(Object o) {
			if(o==null) {
				return false;
			}
			else if(o==this) {
				return true;
			}
			if (o.getClass().equals(getClass())) {
				BDDPair pair = (BDDPair) o;
				return size==pair.size && stateSet.equals(pair.stateSet);
			}

			return false;
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			return size * prime + ((stateSet == null) ? 0 : stateSet.hashCode());
		}

		@Override
		public String toString() {
			StringBuilder builder = new StringBuilder();
			builder.append("(" + automaton.printStates(stateSet) + ", "+size);
			if(list!=null) {
				builder.append(", "+list);
			}
			builder.append(")");

			return builder.toString();
		}

	}

	private final SearchType searchType;
	private final OrderingType orderingType;
	private final boolean computeCounterExample;
	private final Logger logger;

	private boolean verboseMode;
	private boolean isCancelled;
	private Automaton automaton;
	private CospanCounterExample counterExample;
	private List<ComputationListener> listener;

	private RavenBDDPairing automatonPairing;
	private RavenBDD orderingBDD;

	int count = 0;
	private IdentityRelationAlgorithm identityRelationAlgorithm;
	private SimulationAlgorithm forwardSimulationAlgorithm;
	private SimulationAlgorithm backwardSimulationAlgorithm;

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new instance of the antichain-based universality algorithm.
	 * @param searchType the search type used by the algorithm, either {@code forward} or 
	 * {@code backward}
	 * @param orderingType the ordering type used by the algorithm either {@code identityRelation},
	 * {@code forwardSimulation} or {@code backwardSimulation}
	 * @param computeCounterExample if {@code true} a counterexample will be computed if the input
	 * language is not universal
	 * 
	 * @see Algorithm.SearchType
	 * @see Algorithm.OrderingType
	 */
	public SimulationAntichainUniversality(SearchType searchType, OrderingType orderingType, 
			boolean computeCounterExample) {
		if(searchType==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SEARCH_TYPE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(orderingType==null) {
			throw new NullPointerException(
					RavenExceptionMessages.ORDERING_TYPE_MUST_NOT_BE_NULL_EXCEPTION);
		}

		this.searchType = searchType;
		this.orderingType = orderingType;
		this.computeCounterExample = computeCounterExample;

		identityRelationAlgorithm = new IdentityRelationAlgorithm();
		forwardSimulationAlgorithm = new ForwardSimulationAlgorithm();
		backwardSimulationAlgorithm = new BackwardSimulationAlgorithm();
		listener = new LinkedList<>();

		// Register logger
		logger = Logger.getLogger(getClass().getName());
	}

	/**
	 * Creates a new instance of the antichain-based universality algorithm.
	 * @param searchType the search type used by the algorithm, either {@code forward} or 
	 * {@code backward}
	 * @param orderingType the ordering type used by the algorithm either {@code identityRelation},
	 * {@code forwardSimulation} or {@code backwardSimulation}
	 * 
	 * @see Algorithm.SearchType
	 * @see Algorithm.OrderingType
	 */
	public SimulationAntichainUniversality(SearchType searchType, OrderingType orderingType) {
		this(searchType, orderingType, false);
	}

	/**
	 * Creates a new instance of the antichain-based universality algorithm.
	 * @param searchType the search type used by the algorithm, either {@code forward} or 
	 * {@code backward}
	 * 
	 * @see Algorithm.SearchType
	 */
	public SimulationAntichainUniversality(SearchType searchType) {
		this(searchType, OrderingType.identityRelation);
	}

	/*------------------------------------------------------------------------------------------*/
	/* 		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public void addComputationListener(ComputationListener l) {
		listener.add(l);
	}

	@Override
	public void removeComputationListener(ComputationListener l) {
		listener.remove(l);
	}

	@Override
	public boolean isUniversal(Automaton automaton) {
		if(automaton==null) {
			throw new NullPointerException(
					RavenExceptionMessages.AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}

		// Log entering
		logger.entering(getClass().getName(), "isUniversal", automaton);
		verboseMode = Boolean.valueOf(RavenMain.getProperty(RavenMain.PROPERTY_VERBOSE));
		fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_STARTED, "", 0, 0));
		
		isCancelled = false;
		this.automaton = automaton;
		boolean result;

		RavenBDDDomain currentStateDomain = automaton.getEncoding().getCurrentStateDomain();
		RavenBDDDomain successorStateDomain = automaton.getEncoding().getSuccessorStateDomain();
		automatonPairing = RavenBDDPairing.makePair();
		automatonPairing.set(currentStateDomain, successorStateDomain);
		orderingBDD = computeOrdering();

		switch (searchType) {
		case forward:
			result = performForwardSearch();
			break;

		case backward:
			result = performBackwardSearch();
			break;

		default:
			throw new RuntimeException(RavenExceptionMessages.ILLEGAL_SEARCH_TYPE_EXCEPTION);
		}

		if(isCancelled) {
			fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_CANCELLED, "", 0, 0));
		}
		else {
			fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_FINISHED, "", 0, 0));
		}

		// Log exiting
		logger.exiting(getClass().getName(), "isUniversal", counterExample);
		return result;
	}

	@Override
	public CospanCounterExample getCounterExample() {
		return counterExample;
	}
	
	@Override
	public void cancelComputation() {
		isCancelled = true;
	}

	/*------------------------------------------------------------------------------------------*/
	/* 		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Computes the chosen ordering which induces simulation. The ordering is
	 * represented as BDD.
	 * 
	 * @return the ordering BDD
	 */
	private RavenBDD computeOrdering() {
		switch (orderingType) {
		case identityRelation:
			RavenBDD identityRelation = identityRelationAlgorithm.getIdentityRelation(automaton);

			return identityRelation;

		case forwardSimulation:
			return forwardSimulationAlgorithm.getSimulation(automaton);
			
		case backwardSimulation:
			return backwardSimulationAlgorithm.getSimulation(automaton);

		default:
			return null;
		}
	}

	/**
	 * Returns {@code true} if and only if the language of the automaton {@code automaton} is 
	 * universal. The computation is performed via a forward search based on the simulation 
	 * antichain algorithm by Abdulla et. al.
	 * 
	 * @return {@code true} if the language of the automaton is universal, {@code false} otherwise
	 */
	private boolean performForwardSearch() {
		if (!automaton.isAcceptingStateSet(automaton.getInitialStates())) {
			return false;
		}

		Set<BDDPair> processed = new HashSet<>();
		Set<BDDPair> next = new HashSet<>();

		if (computeCounterExample) {
			next.add(new BDDPair(minimize(automaton.getInitialStates()), 
					automaton.getInnerInterface(), new LinkedList<Operation>()));
		} else {
			next.add(new BDDPair(minimize(automaton.getInitialStates()), 
					automaton.getInnerInterface()));
		}

		int i = 0;
		while (!next.isEmpty() && !isCancelled) {
			if(verboseMode) {
				i++;
				logger.info(UNIVERSALITY_ITERATION_STEP+i);
			}

			// Pick and remove a BDD pair from next and move it to processed
			BDDPair pair = next.iterator().next();
			processed.add(pair);
			next.remove(pair);

			Set<BDDPair> results = computePost(pair);
			int maxRounds = results.size();
			int counter =0;
			
			//Compute new frontier
			for(BDDPair p: results) {
				fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED, 
						"Forward Search", maxRounds, ++counter));	
				if (p.size==automaton.getOuterInterface() && 
						!automaton.isAcceptingStateSet(p.getStateSet())) {
					if (computeCounterExample) {
						counterExample = new CospanCounterExample(
								new CospanDecomposition(p.getList()));
					}
					return false;
				}

				Set<BDDPair> union = new HashSet<>(next);
				union.addAll(processed);

				boolean isComparable = false;
				for (BDDPair s : union) {
					if(areStateSetsOrdered(s, p)) {
						isComparable = true;
						break;
					}
				}

				if(!isComparable) {
					Set<BDDPair> tempSet = new HashSet<>();
					for (BDDPair s : union) {
						if (areStateSetsOrdered(p, s)) {
							tempSet.add(s);
						}
					}

					next.add(p);
					next.removeAll(tempSet);
					processed.removeAll(tempSet);
				}
			}
		}

		return true;
	}

	/**
	 * Returns {@code true} if and only if the language of the automaton {@code automaton} is 
	 * universal. The computation is performed via a backward search based on the simulation 
	 * antichain algorithm by Abdulla et. al.
	 * 
	 * @return {@code true} if the language of the automaton is universal, {@code false} otherwise
	 */
	private boolean performBackwardSearch() {
		Set<BDDPair> processed = new HashSet<>();
		Set<BDDPair> next = new HashSet<>();

		// Search backwards for a counter example, i.e. a path that leads from
		// an initial state to a non-final state
		RavenBDD init = automaton.getNonFinalStates().and(
				automaton.getStates(automaton.getInnerInterface()));

		if (!init.isZero() && automaton.isInitialStateSet(init)) {
			if (computeCounterExample) {
				counterExample = new CospanCounterExample(new CospanDecomposition(
						automaton.getInnerInterface()));
			}
			return false;
		}

		if (computeCounterExample) {
			next.add(new BDDPair(minimize(init), automaton.getInnerInterface(), 
					new LinkedList<Operation>()));
		} else {
			next.add(new BDDPair(minimize(init), automaton.getInnerInterface()));
		}

		int i = 0;
		while (!next.isEmpty() && !isCancelled) {
			if(verboseMode) {
				i++;
				logger.info(UNIVERSALITY_ITERATION_STEP+i);
			}

			// Pick and remove a BDD pair from next and move it to processed
			BDDPair pair = next.iterator().next();
			processed.add(pair);
			next.remove(pair);

			Set<BDDPair> results = CPre(pair);
			int maxRounds = results.size();
			int counter =0;
			
			//Compute new frontier
			for(BDDPair p: results) {
				fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED, 
						"Backward Search", maxRounds, ++counter));
				if (p.size==automaton.getInnerInterface() && 
						automaton.isInitialStateSet(p.getStateSet())) {
					if (computeCounterExample) {
						counterExample = new CospanCounterExample(
								new CospanDecomposition(automaton.getInnerInterface(), p.list));
					}
					return false;
				}
				Set<BDDPair> union = new HashSet<>(next);
				union.addAll(processed);

				boolean isComparable = false;
				for (BDDPair s : union) {
					if(areStateSetsOrdered(p, s)) {
						isComparable = true;
						break;
					}
				}

				if(!isComparable) {
					Set<BDDPair> tempSet = new HashSet<>();
					for (BDDPair s : union) {
						if (areStateSetsOrdered(p, s)) {
							tempSet.add(s);
						}
					}

					next.add(p);
					next.removeAll(tempSet);
					processed.removeAll(tempSet);
				}
			}
		}
		return true;
	}

	/**
	 * Computes the successor states for the state set of the BDD pair {@code pair}, i.e. the 
	 * following set will be computed Post(<i>pair = (S, ...)</i>) = {<i>(T, ...)</i> | exists 
	 * letter <i>l</i> s.t. <i>T</i> is the set of <i>l</i>-successors of some state in <i>S</i>}.
	 * 
	 * @param pair the pair of a state and a state set whose "successor pairs" will be computed
	 * @return the set of "successor" pairs
	 */
	private Set<BDDPair> computePost(BDDPair pair) {
		Set<BDDPair> result = new HashSet<>();
		int rounds = automaton.getAlphabet().size();
		int counter = 0;

		for (Operation letter : automaton.getAlphabet()) {
			if(isCancelled) {
				return result;
			}
			
			//Get the current states' successors
			if (isApplicable(letter, pair)) {
				RavenBDD successorsFirst = automaton.getSuccessorStates(pair.getStateSet(), letter);

				//Create new BDDPair
				BDDPair successorPair = new BDDPair(minimize(successorsFirst), 
						getSuccessorSize(letter, pair));
				if (computeCounterExample) {
					successorPair.list = new LinkedList<>(pair.list);
					successorPair.list.add(letter);
				}
				result.add(successorPair);

				successorsFirst.free();
				fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED, 
						"fPost", rounds, ++counter));
			}
		}

		return result;
	}

	/**
	 * Computes the CPre-function for some antichain {@code antichain}. The resulting set contains 
	 * all pairs of "predecessors" which are maximum according to the antichain order. For more 
	 * details about the CPre-function and the antichain order see the paper "Antichains: A New 
	 * Algorithm for Checking Universality of Finite Automata" <a href=
	 * "http://www.springerlink.com/content/250l02n76315n786/fulltext.pdf">Springerlink</a>.
	 * 
	 * @param pair the pair of a state and a state set whose "predecessor pairs" will be computed
	 * @return the antichain containing the maximum of all "successor pairs"
	 */
	private Set<BDDPair> CPre(BDDPair pair) {
		Set<BDDPair> result = new HashSet<>();
		int rounds = automaton.getAlphabet().size();
		int counter = 0;

		for (Operation operation : automaton.getAlphabet()) {
			 if(isCancelled) {
				 return result;
			 }
			
			 //Get the current states' predecessors
			if (isReverseApplicable(operation, pair)) {
				int size = getPredecessorSize(operation, pair);
				RavenBDD predecessors = getCPredecessorStates(pair.stateSet, operation);

				predecessors = minimize(predecessors);
				predecessors = predecessors.and(
						automaton.getEncoding().getInterfaceSize(StateType.current, size));

				//Create new BDDPair
				BDDPair predecessorPair = new BDDPair(predecessors, size);
				if (computeCounterExample) {
					predecessorPair.list = new LinkedList<>();
					predecessorPair.list.add(operation);
					predecessorPair.list.addAll(pair.list);
				}

				result.add(predecessorPair);
				count++;

				fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED, 
						"CPre", rounds, ++counter));
			}
		}


		return result;
	}

	/**
	 * Computes all predecessor states from which only states of the set {@code states} are 
	 * reachable by reading the letter {@code letter}.
	 * 
	 * @param states the set of (successor) state which must be reachable from the predecessor states
	 * @param letter the letter whose transition function will be used
	 * @return the set of c-predecessor states
	 */
	private RavenBDD getCPredecessorStates(RavenBDD states, Operation letter) {
		RavenBDD result = automaton.getEncoding().zero();
		RavenBDD predecessors = automaton.getPredecessorStates(states, letter);

		if (!predecessors.isZero()) {
			for (RavenBDD bdd : predecessors.getSatAssignments(
					automaton.getEncoding().getCurrentStateDomain().getDomainBDD())) {
				if (isSubset(automaton.getSuccessorStates(bdd, letter), states)) {
					result = result.or(bdd);
				}
			}
		}

		result = result.or(automaton.getStatesWithoutSuccessorFor(letter));
		return result;
	}

	/**
	 * This method implements Optimization 2. It does the following: it chooses
	 * a state <i>s</i> from the state set {@code stateSet}, removes <i>s</i>
	 * from the state set, if there exists a state <i>t</i> in the state set
	 * such state <i>s &lt;= t</i>, and then repeats the procedure until all of
	 * the states in the state set are processed.
	 * 
	 * (Note: <i>&lt;</i> is the ordering which implies that the state <i>s</i>
	 * is simulated by the state <i>t</i>.)
	 * 
	 * @param stateSet the state set which is minimized
	 * @return the minimized state set
	 */
	private RavenBDD minimize(RavenBDD stateSet) {
		// Iterate over all states of the state set (of the pair's second component)
		RavenBDDDomain currentStateDomain = automaton.getEncoding().getCurrentStateDomain();
		RavenBDDDomain successorStateDomain = automaton.getEncoding().getSuccessorStateDomain();

		RavenBDD vars = currentStateDomain.getDomainBDD();
		RavenBDD minimizedStateSet = stateSet.clone();
		RavenBDD modifiedStateSet;

		for (RavenBDD bdd : stateSet.getSatAssignments(vars)) {
			// Modify the state set such that it does not contain the currently chosen BDD
			modifiedStateSet = stateSet.clone().and(bdd.clone().not());
			RavenBDD tempBDD = modifiedStateSet.clone().replace(automatonPairing);
			tempBDD = tempBDD.applyEx(getOrdering().and(bdd), BDDOperator.and,
					successorStateDomain.getDomainBDD());

			if (tempBDD.isOne()) {
				minimizedStateSet = modifiedStateSet;
			}
		}
		return minimizedStateSet;
	}

	/**
	 * Returns {@code true} if and only if the the first state set {@code firsteStateSet} is related 
	 * to the second state set {@code secondStateSet}, i.e. for all states <i>s</i> of the first 
	 * state set exist a state <i>t</i> in the second state set s.t. <i>s &lt;= t</i>.
	 * 
	 * (Note: <i>&lt;=</i> is the ordering which implies that the state <i>s</i>
	 * is simulated by the state <i>t</i>.)
	 * 
	 * @param s the first state set
	 * @param p the second state set
	 * @return {@code true} if the first state set is related to the second state set, 
	 * {@code false} otherwise
	 */
	private boolean areStateSetsOrdered(BDDPair s, BDDPair p) {
		if(s.size!=p.size) {
			return false;
		}

		RavenBDDDomain currentStateDomain = automaton.getEncoding().getCurrentStateDomain();
		RavenBDDDomain successorStateDomain = automaton.getEncoding().getSuccessorStateDomain();
		RavenBDD fStateSet = s.getStateSet().clone();
		RavenBDD sStateSet = p.getStateSet().clone().replace(automatonPairing);

		RavenBDD tempBDD = sStateSet.applyEx(getOrdering(), BDDOperator.and, 
				successorStateDomain.getDomainBDD());
		tempBDD = fStateSet.applyAll(tempBDD, BDDOperator.imp,
				currentStateDomain.getDomainBDD());
		boolean result = tempBDD.isOne();
		tempBDD.free();

		return result;
	}

	/**
	 * Returns {@code true} if and only if the first set (represented by the BDD) {@code firstSet} 
	 * is a subset of the second set (represented by the BDD) {@code secondSet}
	 * 
	 * @param firstSet the first set to compare
	 * @param secondSet the second set to compare
	 * @return {@code true} if the first set is a subset of the second set, {@code false} otherwise
	 */
	private static boolean isSubset(RavenBDD firstSet, RavenBDD secondSet) {
		RavenBDD tempBDD = firstSet.clone().imp(secondSet.clone());
		boolean result = tempBDD.isOne();
		tempBDD.free();

		return result;
	}

	/**
	 * Returns the size of the interface of the state set which is reachable by
	 * "reverse applying" the operation {@code letter} to the pair {@code pair}.
	 * 
	 * @param letter the operation which is to be "reverse applied"
	 * @param pair the pair to which the operation is "reverse applied"
	 * @return the size of the "predecessor" interface
	 */
	private static int getPredecessorSize(Operation letter, BDDPair pair) {
		int size = pair.size;

		if (letter instanceof Fuse || letter instanceof Res) {
			size += 1;
		}
		else if (letter instanceof Vertex) {
			size -= 1;
		}

		return size;
	}

	/**
	 * Returns {@code true} if and only if the operation {@code letter} is applicable to the pair 
	 * {@code pair} (i.e. if it is applicable to the state set of the pair).
	 * 
	 * @param letter the operation which is to be applied
	 * @param pair the pair to which the operation is applied
	 * @return {@code true} if the operation can be applied, {@code false} otherwise
	 */
	private boolean isApplicable(Operation letter, BDDPair pair) {
		boolean isApplicable = false;

		if (letter instanceof Connect) {
			isApplicable = ((Connect) letter).getArity() <= pair.size;
		}
		else if (letter instanceof Fuse || letter instanceof Perm
				|| letter instanceof Trans) {
			isApplicable = pair.size >= 2;
		}
		else if (letter instanceof Res) {
			isApplicable = pair.size >= 1;
		}
		else if (letter instanceof Vertex) {
			isApplicable = pair.size < automaton.getMaximumInterface();
		}
		return isApplicable;
	}

	/**
	 * Returns {@code true} if and only if the pair could be reached by the application of the 
	 * operation {@code letter} to some other pair (i.e. if the operation could be "reverse applied" 
	 * to the state set of the pair).
	 * @param letter the operation which is to be "reverse applied"
	 * @param pair the pair to which the operation is "reverse applied"
	 * @return {@code true} if the operation can be "reverse applied", {@code false} otherwise
	 */
	private boolean isReverseApplicable(Operation letter, BDDPair pair) {
		boolean isApplicable = false;

		if (letter instanceof Connect) {
			isApplicable = ((Connect) letter).getArity() <= pair.size;
		}
		else if (letter instanceof Fuse) {
			isApplicable = pair.size >= 1
					&& pair.size < automaton.getMaximumInterface();
		}
		else if (letter instanceof Perm || letter instanceof Trans) {
			isApplicable = pair.size >= 2;
		}
		else if (letter instanceof Res) {
			isApplicable = pair.size < automaton.getMaximumInterface();
		}
		else if (letter instanceof Vertex) {
			isApplicable = pair.size > 0;
		}
		return isApplicable;
	}

	/**
	 * Returns the size of the interface of the state set which is reachable by 
	 * applying the operation {@code letter} to the pair {@code pair}.
	 * @param letter the operation which is to be applied
	 * @param pair the pair to which the operation is applied
	 * @return the size of the "successor" interface
	 */
	private static int getSuccessorSize(Operation letter, BDDPair pair) {
		int size = pair.size;

		if(letter instanceof Fuse || letter instanceof Res) {
			size -= 1;
		}
		else if(letter instanceof Vertex) {
			size += 1;
		}

		return size;
	}

	/**
	 * Returns the BDD representing the used ordering relation.
	 * @return the ordering relation BDD
	 */
	private RavenBDD getOrdering() {
		return orderingBDD.clone();
	}

	/**
	 * Fires an computation event notifying all listeners.
	 * @param e the computation event
	 */
	private void fireComputationEvent(ComputationEvent e) {
		for (ComputationListener l : listener) {
			l.computationAdvanded(e);
		}
	}

}
