/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.cli;

import java.io.File;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import de.uni_due.inf.ti.raven.actions.ColorabilityAutomatonAction;
import de.uni_due.inf.ti.raven.actions.ComputationAction;
import de.uni_due.inf.ti.raven.actions.DominatingSetAutomatonAction;
import de.uni_due.inf.ti.raven.actions.EdgeCountingAutomatonAction;
import de.uni_due.inf.ti.raven.actions.LinkAutomatonAction;
import de.uni_due.inf.ti.raven.actions.MaximumEdgeAutomatonAction;
import de.uni_due.inf.ti.raven.actions.MaximumVertexAutomatonAction;
import de.uni_due.inf.ti.raven.actions.MinimumEdgeAutomatonAction;
import de.uni_due.inf.ti.raven.actions.MinimumVertexAutomatonAction;
import de.uni_due.inf.ti.raven.actions.NoIsolatedNodesAutomatonAction;
import de.uni_due.inf.ti.raven.actions.PathAutomatonAction;
import de.uni_due.inf.ti.raven.actions.ProductAutomatonAction;
import de.uni_due.inf.ti.raven.actions.ShiftAutomatonAction;
import de.uni_due.inf.ti.raven.actions.SubgraphAutomatonAction;
import de.uni_due.inf.ti.raven.actions.UnionAutomatonAction;
import de.uni_due.inf.ti.raven.actions.VertexCountingAutomatonAction;
import de.uni_due.inf.ti.raven.actions.VertexCoverAutomatonAction;
import de.uni_due.inf.ti.raven.algorithms.CospanDecompositionAlgorithm;
import de.uni_due.inf.ti.raven.algorithms.DefaultCospanDecompositionAlgorithm;
import de.uni_due.inf.ti.raven.data.Cospan;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.Operation.Connect;
import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.data.Signature;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * Console command to handle the creation of (graph) automata.
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.50 $
 *
 */
public final class AutomatonCommand extends Command {

	private static final File AUTOMATON_DOCUMENTATION_FILE = new File(
			RavenMain.DOCUMENTATION_FOLDER + "AutomatonCommandDocumentation.txt");
	
	private static final String ARITY_TOKEN = Console.ARITY_TOKEN;
	private static final String NAME_PATTERN = Console.NAME_PATTERN;
	private static final String NUMBER_PATTERN = Console.NUMBER_PATTERN;
	private static final String IS_TOKEN = Console.IS_TOKEN;
	private static final String LABEL_PATTERN = "("+Console.LABEL_PATTERN+")" + ARITY_TOKEN + NUMBER_PATTERN;
	private static final String SEP_TOKEN = Console.SEP_TOKEN;
	private static final String SET_SEP_TOKEN = Console.SET_SEP_TOKEN;
	private static final String SET_OF_NUMBERS_PATTERN = Console.SET_OF_NUMBERS_PATTERN;
	
	private static final String AUTOMATON_KEY = Console.AUTOMATON_TOKEN;
	
	private static final String COLORABILITY_PARAMETER = "("+NAME_PATTERN+")" + IS_TOKEN + 
			Console.COLORABILTY_TOKEN + 
			"\\(("+NUMBER_PATTERN+")" + SEP_TOKEN + 
			"("+NUMBER_PATTERN+")" + SEP_TOKEN  + 
			"("+NUMBER_PATTERN+")" + SEP_TOKEN + 
			"("+NAME_PATTERN+")" + SEP_TOKEN + 
			"("+NUMBER_PATTERN+")\\)";
	
	private static final String DOMINATING_SET_PARAMETER = "("+NAME_PATTERN+")" + IS_TOKEN + 
			Console.DOMINATING_SET_TOKEN + 
			"\\(("+NUMBER_PATTERN+")" + SEP_TOKEN + 
			"("+NUMBER_PATTERN+")"+ SEP_TOKEN + 
			"("+NUMBER_PATTERN+")" + SEP_TOKEN + 
			"("+NAME_PATTERN+")" + SEP_TOKEN + 
			"("+NUMBER_PATTERN+")\\)";
	
	private static final String EDGE_COUNTING_PARAMETER = "("+NAME_PATTERN+")" + IS_TOKEN + 
			Console.EDGE_COUNTING_TOKEN + 
			"\\(("+NUMBER_PATTERN+")" + SEP_TOKEN + 
			"("+NAME_PATTERN+")" + SEP_TOKEN + 
			"("+NUMBER_PATTERN+")" + SEP_TOKEN + 
			"("+NUMBER_PATTERN+")(" + SEP_TOKEN + 
			"\\{("+LABEL_PATTERN+"("+SEP_TOKEN+LABEL_PATTERN+")*)\\})?\\)";
	
	private static final String LINK_PARAMETER = "("+NAME_PATTERN+")" + IS_TOKEN + 
			Console.LINK_TOKEN  + 
			"\\(("+NUMBER_PATTERN+")" + SEP_TOKEN +
			"("+NUMBER_PATTERN+")" + SEP_TOKEN + 
			"("+NUMBER_PATTERN+")" + SEP_TOKEN +
			"("+NAME_PATTERN+")\\)"; 

	private static final String MAXIMUM_EDGE_PARAMETER = "("+NAME_PATTERN+")" + IS_TOKEN + 
			Console.MAXIMUM_EDGE_TOKEN  + 
			"\\(("+NUMBER_PATTERN+")" + SEP_TOKEN + 
			"("+NUMBER_PATTERN+")"+ SEP_TOKEN + 
			"("+NUMBER_PATTERN+")" + SEP_TOKEN +
			"("+NAME_PATTERN+")" + SEP_TOKEN + 
			"("+NUMBER_PATTERN +")\\)";

	private static final String MAXIMUM_VERTEX_PARAMETER = "("+NAME_PATTERN+")" + IS_TOKEN + 
			Console.MAXIMUM_VERTEX_TOKEN  + 
			"\\(("+NUMBER_PATTERN+")" + SEP_TOKEN + 
			"("+NUMBER_PATTERN+")"+ SEP_TOKEN + 
			"("+NUMBER_PATTERN+")" + SEP_TOKEN +
			"("+NAME_PATTERN+")" + SEP_TOKEN + 
			"("+NUMBER_PATTERN +")\\)";

	private static final String MINIMUM_EDGE_PARAMETER = "("+NAME_PATTERN+")" + IS_TOKEN + 
			Console.MINIMUM_EDGE_TOKEN  + 
			"\\(("+NUMBER_PATTERN+")" + SEP_TOKEN + 
			"("+NUMBER_PATTERN+")"+ SEP_TOKEN + 
			"("+NUMBER_PATTERN+")" + SEP_TOKEN +
			"("+NAME_PATTERN+")" + SEP_TOKEN + 
			"("+NUMBER_PATTERN +")\\)";

	private static final String MINIMUM_VERTEX_PARAMETER = "("+NAME_PATTERN+")" + IS_TOKEN + 
			Console.MINIMUM_VERTEX_TOKEN  + 
			"\\(("+NUMBER_PATTERN+")" + SEP_TOKEN + 
			"("+NUMBER_PATTERN+")"+ SEP_TOKEN + 
			"("+NUMBER_PATTERN+")" + SEP_TOKEN +
			"("+NAME_PATTERN+")" + SEP_TOKEN + 
			"("+NUMBER_PATTERN +")\\)";
	 
	private static final String NO_ISOLATED_NODES_PARAMETER = "("+NAME_PATTERN+")" + IS_TOKEN + 
			Console.NO_ISOLATED_NODE_TOKEN  + 
			"\\(("+NUMBER_PATTERN+")" + SEP_TOKEN + 
			"("+NUMBER_PATTERN+")"+ SEP_TOKEN + 
			"("+NUMBER_PATTERN+")" + SEP_TOKEN +
			"("+NAME_PATTERN+")\\)";
	
	private static final String PATH_PARAMETER = "("+NAME_PATTERN+")" + IS_TOKEN + 
			Console.PATH_AUTOMATON_TOKEN + 
			"\\(("+NUMBER_PATTERN+")" + SEP_TOKEN + 
			"("+NUMBER_PATTERN+")"+ SEP_TOKEN + 
			"("+NUMBER_PATTERN+")" + SEP_TOKEN +
			"("+NAME_PATTERN+")" + SEP_TOKEN + 
			"("+SET_OF_NUMBERS_PATTERN+")" + SEP_TOKEN +
			"("+SET_OF_NUMBERS_PATTERN+")\\)"; 
	
	private static final String PRODUCT_PARAMETER = "("+NAME_PATTERN+")" + IS_TOKEN + 
			Console.PRODUCT_TOKEN + 
			"\\(("+NAME_PATTERN+")" + SEP_TOKEN + 
			"("+NAME_PATTERN+")\\)";
	
	private static final String SHIFT_PARAMETER = "("+NAME_PATTERN+")" + IS_TOKEN + 
			Console.SHIFT_TOKEN + "\\(("+NAME_PATTERN+")" + SEP_TOKEN + 
			"("+NAME_PATTERN+")\\)";
	
	private static final String SUBGRAPH_PARAMETER = "("+NAME_PATTERN+")" + IS_TOKEN + 
			Console.SUBGRAPH_TOKEN + 
			"\\(("+NUMBER_PATTERN+")" + SEP_TOKEN + 
			"("+NUMBER_PATTERN+")"+ SEP_TOKEN + 
			"("+NUMBER_PATTERN+")" + SEP_TOKEN +
			"("+NAME_PATTERN+")" + SEP_TOKEN + 
			"("+NAME_PATTERN+")\\)";
	
	private static final String UNION_PARAMETER = "("+NAME_PATTERN+")" + IS_TOKEN + 
			Console.UNION_TOKEN + 
			"\\(("+NAME_PATTERN+")" + SEP_TOKEN + 
			"("+NAME_PATTERN+")\\)";
	
	private static final String VERTEX_COUNTING_PARAMETER = "("+NAME_PATTERN+")" + IS_TOKEN + 
			Console.VERTEX_COUNTING_TOKEN + 
			"\\(("+NUMBER_PATTERN+")" + SEP_TOKEN +
			"("+NUMBER_PATTERN+")" + SEP_TOKEN + 
			"("+NUMBER_PATTERN+")" + SEP_TOKEN + 
			"("+NAME_PATTERN+")" + SEP_TOKEN + 
			"("+NUMBER_PATTERN+")" + SEP_TOKEN + 
			"("+NUMBER_PATTERN+")\\)";
	
	private static final String VERTEX_COVER_PARAMETER = "("+NAME_PATTERN+")" + IS_TOKEN + 
			Console.VERTEX_COVER_TOKEN + 
			"\\(("+NUMBER_PATTERN+")" + SEP_TOKEN + 
			"("+NUMBER_PATTERN+")" + SEP_TOKEN + 
			"("+NUMBER_PATTERN+")"+ SEP_TOKEN + 
			"("+NAME_PATTERN+")" + SEP_TOKEN + 
			"("+NUMBER_PATTERN+")\\)";
	
	private static AutomatonCommand command;

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the <tt>AUTOMATON</tt>-command.
	 * @return the <tt>AUTOMATON</tt>-command
	 */
	protected static AutomatonCommand getCommand() {
		if(command==null) {
			command = new AutomatonCommand();
		}

		return command;		
	}
	
	/**
	 * Parses a string containing a list of labels and returns the corresponding CONNECT-operations.
	 * @param labelString the label string
	 * @return the set of CONNECT-operations 
	 */
	private static Set<Connect> parseLabelSet(String labelString) {
		if(labelString==null) {
			return null;
		}
		
		Set<Connect> edgeSet = new HashSet<>();
		for(String label: labelString.split(SEP_TOKEN)) {
			String[] array = label.split(ARITY_TOKEN);
			
			edgeSet.add(Operation.getConnect(array[0], Integer.parseInt(array[1])));
		}
		
		return edgeSet;
	}
	
	/**
	 * Parses a string containing a set of nodes and returns the set.
	 * @param nodeString the node set string
	 * @return the set of nodes
	 */
	private static Set<Integer> parseNodeSet(String nodeString) {
		if(nodeString==null) {
			return null;
		}
		if(!nodeString.matches(SET_OF_NUMBERS_PATTERN)) {
			return null;
		}
		
		Set<Integer> nodeSet = new HashSet<>();
		nodeString = nodeString.substring(1, nodeString.length()-1);
		for(String node: nodeString.split(SET_SEP_TOKEN)) {
			nodeSet.add(Integer.parseInt(node));
		}
		
		return nodeSet;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates the <tt>AUTOMATON</tt>-command.
	 */
	private AutomatonCommand() {
		super(AUTOMATON_KEY, ConsoleMessages.AUTOMATON_DESCRIPTION_KEY, 
				AUTOMATON_DOCUMENTATION_FILE);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public boolean matchesParameter(String line) {
		return line.matches(COLORABILITY_PARAMETER) ||
				line.matches(DOMINATING_SET_PARAMETER) ||
				line.matches(EDGE_COUNTING_PARAMETER) ||
				line.matches(LINK_PARAMETER) ||
				line.matches(MAXIMUM_EDGE_PARAMETER) ||
				line.matches(MAXIMUM_VERTEX_PARAMETER) ||
				line.matches(MINIMUM_VERTEX_PARAMETER) ||
				line.matches(MINIMUM_EDGE_PARAMETER) ||
				line.matches(NO_ISOLATED_NODES_PARAMETER) ||
				line.matches(PATH_PARAMETER) ||
				line.matches(PRODUCT_PARAMETER) ||
				line.matches(SHIFT_PARAMETER) ||
				line.matches(SUBGRAPH_PARAMETER) ||
				line.matches(UNION_PARAMETER) ||
				line.matches(VERTEX_COUNTING_PARAMETER) ||
				line.matches(VERTEX_COVER_PARAMETER);
	}

	@Override
	public void execute(String parameter) {		
		if(!matchesParameter(parameter)) {
			throw new IllegalArgumentException(Console.messages.getString(
					ConsoleMessages.MISSING_ILLEGAL_ARGUMENT_DESCRIPTION_KEY, AUTOMATON_KEY));
		}

		String name;
		int maxInterface;
		int innerInterface;
		int outerInterface; 
		ComputationAction<? extends Automaton> action;
		if(parameter.matches(COLORABILITY_PARAMETER)) {
			//Scan and extract automaton properties from input string
			List<String> list = getProperties(parameter, COLORABILITY_PARAMETER);
			name = list.get(0);
			maxInterface = Integer.parseInt(list.get(1));
			innerInterface = Integer.parseInt(list.get(2));
			outerInterface = Integer.parseInt(list.get(3));
			String signatureName = list.get(4);
			int colors = Integer.parseInt(list.get(5));
			
			if(name==null) {
				throw new NullPointerException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_NULL_KEY));
			}
			if(name.isEmpty()) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_EMPTY_KEY));
			}
			if(!Console.repo.isSignature(signatureName)) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.SIGNATURE_NOT_FOUND_KEY, signatureName));
			}

			Signature signature = (Signature) Console.repo.get(signatureName);
			//Create and execute new automaton command
			action = new ColorabilityAutomatonAction(maxInterface, 
					innerInterface, outerInterface, signature, colors);
		}
		else if(parameter.matches(LINK_PARAMETER)) {
			//Scan and extract automaton properties from input string
			List<String> list = getProperties(parameter, LINK_PARAMETER);
			name = list.get(0);
			maxInterface = Integer.parseInt(list.get(1));
			innerInterface = Integer.parseInt(list.get(2));
			outerInterface = Integer.parseInt(list.get(3));
			String signatureName = list.get(4);
			
			if(name==null) {
				throw new NullPointerException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_NULL_KEY));
			}
			if(name.isEmpty()) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_EMPTY_KEY));
			}
			if(!Console.repo.isSignature(signatureName)) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.SIGNATURE_NOT_FOUND_KEY, signatureName));
			}

			Signature signature = (Signature) Console.repo.get(signatureName);
			//Create and execute new automaton command
			action = new LinkAutomatonAction(maxInterface, 
					innerInterface, outerInterface, signature);
		}
		else if(parameter.matches(DOMINATING_SET_PARAMETER)) {
			//Scan and extract automaton properties from input string
			List<String> list = getProperties(parameter, DOMINATING_SET_PARAMETER);
			name = list.get(0);
			maxInterface = Integer.parseInt(list.get(1));
			innerInterface = Integer.parseInt(list.get(2));
			outerInterface = Integer.parseInt(list.get(3));
			String signatureName = list.get(4);
			int dominatingSize = Integer.parseInt(list.get(5));
			
			if(name==null) {
				throw new NullPointerException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_NULL_KEY));
			}
			if(name.isEmpty()) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_EMPTY_KEY));
			}
			if(!Console.repo.isSignature(signatureName)) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.SIGNATURE_NOT_FOUND_KEY, signatureName));
			}

			Signature signature = (Signature) Console.repo.get(signatureName);
			//Create and execute new automaton command
			action = new DominatingSetAutomatonAction(maxInterface, 
					innerInterface, outerInterface, signature, dominatingSize);
		}
		else if(parameter.matches(EDGE_COUNTING_PARAMETER)) {
			//Scan and extract automaton properties from input string
			List<String> list = getProperties(parameter, EDGE_COUNTING_PARAMETER);
			name = list.get(0);
			maxInterface = Integer.parseInt(list.get(1));
			innerInterface = Integer.parseInt(list.get(2));
			outerInterface = Integer.parseInt(list.get(3));
			String signatureName = list.get(4);
			int divisor = Integer.parseInt(list.get(5));
			int remainder = Integer.parseInt(list.get(6));
			Set<Connect> edgeSet = parseLabelSet(list.get(7));
			
			if(name==null) {
				throw new NullPointerException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_NULL_KEY));
			}
			if(name.isEmpty()) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_EMPTY_KEY));
			}
			if(!Console.repo.isSignature(signatureName)) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.SIGNATURE_NOT_FOUND_KEY, signatureName));
			}
			if(divisor<1) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.DIVISOR_LESS_THAN_ONE_KEY));
			}
			if(remainder<0) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.REMAINDER_LESS_THAN_ZERO_KEY));
			}
			if(remainder>=divisor) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.REMAINDER_GREATER_OR_EQUAL_THAN_DIVISOR_KEY));
			}
			
			Signature signature = (Signature) Console.repo.get(signatureName);
			if(edgeSet==null) {
				edgeSet = new HashSet<>();
				
				for(Operation letter: signature) {
					if(letter instanceof Connect) {
						edgeSet.add((Connect) letter);
					}
				}
			}
			
			//Create and execute new automaton command
			action = new EdgeCountingAutomatonAction(maxInterface, innerInterface, outerInterface, 
					signature, divisor,  remainder, edgeSet);
		}
		else if(parameter.matches(MAXIMUM_EDGE_PARAMETER)) {
			//Scan and extract automaton properties from input string
			List<String> list = getProperties(parameter, MAXIMUM_EDGE_PARAMETER);
			name = list.get(0);
			maxInterface = Integer.parseInt(list.get(1));
			innerInterface = Integer.parseInt(list.get(2));
			outerInterface = Integer.parseInt(list.get(3));
			String signatureName = list.get(4);
			int maximum = Integer.parseInt(list.get(5));
			
			if(name==null) {
				throw new NullPointerException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_NULL_KEY));
			}
			if(name.isEmpty()) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_EMPTY_KEY));
			}
			if(!Console.repo.isSignature(signatureName)) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.SIGNATURE_NOT_FOUND_KEY, signatureName));
			}
			if(maximum<0) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.MAXIMUM_LESS_THAN_ZERO_KEY));
			}
			if(maximum>maxInterface) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.MAXIMUM_GREATER_THAN_MAXIMUM_INTERFACE_KEY));
			}

			Signature signature = (Signature) Console.repo.get(signatureName);			
			//Create and execute new automaton command
			action = new MaximumEdgeAutomatonAction(maxInterface, 
					innerInterface, outerInterface, signature, maximum);
		}
		else if(parameter.matches(MAXIMUM_VERTEX_PARAMETER)) {
			//Scan and extract automaton properties from input string
			List<String> list = getProperties(parameter, MAXIMUM_VERTEX_PARAMETER);
			name = list.get(0);
			maxInterface = Integer.parseInt(list.get(1));
			innerInterface = Integer.parseInt(list.get(2));
			outerInterface = Integer.parseInt(list.get(3));
			String signatureName = list.get(4);
			int maximum = Integer.parseInt(list.get(5));
			
			if(name==null) {
				throw new NullPointerException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_NULL_KEY));
			}
			if(name.isEmpty()) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_EMPTY_KEY));
			}
			if(!Console.repo.isSignature(signatureName)) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.SIGNATURE_NOT_FOUND_KEY, signatureName));
			}
			if(maximum<0) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.MAXIMUM_LESS_THAN_ZERO_KEY));
			}
			if(maximum>maxInterface) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.MAXIMUM_GREATER_THAN_MAXIMUM_INTERFACE_KEY));
			}

			Signature signature = (Signature) Console.repo.get(signatureName);			
			//Create and execute new automaton command
			action = new MaximumVertexAutomatonAction(maxInterface, innerInterface, outerInterface, 
					signature, maximum);
		}
		else if(parameter.matches(MINIMUM_EDGE_PARAMETER)) {
			//Scan and extract automaton properties from input string
			List<String> list = getProperties(parameter, MINIMUM_EDGE_PARAMETER);
			name = list.get(0);
			maxInterface = Integer.parseInt(list.get(1));
			innerInterface = Integer.parseInt(list.get(2));
			outerInterface = Integer.parseInt(list.get(3));
			String signatureName = list.get(4);
			int minimum = Integer.parseInt(list.get(5));
			
			if(name==null) {
				throw new NullPointerException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_NULL_KEY));
			}
			if(name.isEmpty()) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_EMPTY_KEY));
			}
			if(!Console.repo.isSignature(signatureName)) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.SIGNATURE_NOT_FOUND_KEY, signatureName));
			}
			if(minimum<0) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.MINIMUM_LESS_THAN_ZERO_KEY));
			}
			if(minimum>maxInterface) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.MINIMUM_GREATER_THAN_MAXIMUM_INTERFACE_KEY));
			}

			Signature signature = (Signature) Console.repo.get(signatureName);			
			//Create and execute new automaton command
			action = new MinimumEdgeAutomatonAction(maxInterface, innerInterface, outerInterface,
					signature, minimum);
		}
		else if(parameter.matches(MINIMUM_VERTEX_PARAMETER)) {
			//Scan and extract automaton properties from input string
			List<String> list = getProperties(parameter, MINIMUM_VERTEX_PARAMETER);
			name = list.get(0);
			maxInterface = Integer.parseInt(list.get(1));
			innerInterface = Integer.parseInt(list.get(2));
			outerInterface = Integer.parseInt(list.get(3));
			String signatureName = list.get(4);
			int minimum = Integer.parseInt(list.get(5));
			
			if(name==null) {
				throw new NullPointerException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_NULL_KEY));
			}
			if(name.isEmpty()) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_EMPTY_KEY));
			}
			if(!Console.repo.isSignature(signatureName)) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.SIGNATURE_NOT_FOUND_KEY, signatureName));
			}
			if(minimum<0) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.MINIMUM_LESS_THAN_ZERO_KEY));
			}
			if(minimum>maxInterface) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.MINIMUM_GREATER_THAN_MAXIMUM_INTERFACE_KEY));
			}

			Signature signature = (Signature) Console.repo.get(signatureName);			
			//Create and execute new automaton command
			action = new MinimumVertexAutomatonAction(maxInterface, innerInterface, outerInterface,
					signature, minimum);
		}
		else if(parameter.matches(NO_ISOLATED_NODES_PARAMETER)) {
			//Scan and extract automaton properties from input string
			List<String> list = getProperties(parameter, NO_ISOLATED_NODES_PARAMETER);
			name = list.get(0);
			maxInterface = Integer.parseInt(list.get(1));
			innerInterface = Integer.parseInt(list.get(2));
			outerInterface = Integer.parseInt(list.get(3));
			String signatureName = list.get(4);
			
			if(name==null) {
				throw new NullPointerException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_NULL_KEY));
			}
			if(name.isEmpty()) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_EMPTY_KEY));
			}
			if(!Console.repo.isSignature(signatureName)) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.SIGNATURE_NOT_FOUND_KEY, signatureName));
			}

			Signature signature = (Signature) Console.repo.get(signatureName);
			//Create and execute new automaton command
			action = new NoIsolatedNodesAutomatonAction(maxInterface, innerInterface, outerInterface, 
					signature);
		}
		else if(parameter.matches(PATH_PARAMETER)) {
			//Scan and extract automaton properties from input string
			List<String> list = getProperties(parameter, PATH_PARAMETER);
			name = list.get(0);
			maxInterface = Integer.parseInt(list.get(1));
			innerInterface = Integer.parseInt(list.get(2));
			outerInterface = Integer.parseInt(list.get(3));
			String signatureName = list.get(4);
			Set<Integer> sourceNodes = parseNodeSet(list.get(5));
			Set<Integer> targetNodes = parseNodeSet(list.get(7));
			
			if(name==null) {
				throw new NullPointerException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_NULL_KEY));
			}
			if(name.isEmpty()) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_EMPTY_KEY));
			}
			if(!Console.repo.isSignature(signatureName)) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.SIGNATURE_NOT_FOUND_KEY, signatureName));
			}

			Signature signature = (Signature) Console.repo.get(signatureName);
			//Create and execute new automaton command
			action = new PathAutomatonAction(maxInterface, innerInterface,
					outerInterface, signature, sourceNodes, targetNodes);
		}
		else if(parameter.matches(PRODUCT_PARAMETER)) {
			//Scan and extract automaton properties from input string
			List<String> list = getProperties(parameter, PRODUCT_PARAMETER);
			name = list.get(0);
			String firstName = list.get(1);
			String secondName = list.get(2);
			
			if(name==null) {
				throw new NullPointerException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_NULL_KEY));
			}
			if(name.isEmpty()) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_EMPTY_KEY));
			}
			if(!Console.repo.isAutomaton(firstName)) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NOT_FOUND_KEY, firstName));
			}
			else if(!Console.repo.isAutomaton(secondName)) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NOT_FOUND_KEY, firstName));
			}
			else {
				Automaton first = (Automaton) Console.repo.get(firstName);
				Automaton second = (Automaton) Console.repo.get(secondName);
				
				if(first.getMaximumInterface()!=second.getMaximumInterface()) {
					throw new IllegalArgumentException(Console.messages.getString(
							ConsoleMessages.DIFFERENT_MAXIMUM_INTERFACES_KEY));
				}
				else if(!first.getAlphabet().equals(second.getAlphabet())) {
					throw new IllegalArgumentException(Console.messages.getString(
							ConsoleMessages.DIFFERENT_SIGNATURES_KEY));
				}
				else {
					//Create and execute new automaton command
					maxInterface = first.getMaximumInterface();
					action = new ProductAutomatonAction(first, second);
				}
			}
		}
		else if(parameter.matches(SHIFT_PARAMETER)) {
			//Scan and extract automaton properties from input string
			List<String> list = getProperties(parameter, SHIFT_PARAMETER);
			name = list.get(0);
			String automatonName = list.get(1);
			String cospanOrDecompositionName = list.get(2);
			
			if(name==null) {
				throw new NullPointerException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_NULL_KEY));
			}
			if(name.isEmpty()) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_EMPTY_KEY));
			}
			if(!Console.repo.isAutomaton(automatonName)) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NOT_FOUND_KEY, automatonName));
			}
			else if(!Console.repo.isCospan(cospanOrDecompositionName) && 
					!Console.repo.isDecomposition(cospanOrDecompositionName)) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.COSPAN_OR_DECOMPOSITION_NOT_FOUND_KEY, cospanOrDecompositionName));
			}
			else {
				Automaton automaton = (Automaton) Console.repo.get(automatonName);
				
				CospanDecomposition decomposition = null;
				//If the user has entered the name of a cospan, it must be decomposed first
				if(Console.repo.isCospan(cospanOrDecompositionName)) {
					Cospan cospan = (Cospan) Console.repo.get(cospanOrDecompositionName);
					
					CospanDecompositionAlgorithm algorithm = new DefaultCospanDecompositionAlgorithm();
					decomposition = algorithm.decompose(cospan);
				}
				else {
					decomposition = (CospanDecomposition) Console.repo.get(cospanOrDecompositionName);
				}
				
				if(decomposition.getWidth()>automaton.getMaximumInterface()) {
					throw new IllegalArgumentException(Console.messages.getString(
							ConsoleMessages.WIDTH_GREATER_THAN_MAXIMUM_INTERFACE_KEY));
				}
				else if(automaton.getInnerInterface()!=decomposition.getInnerInterfaceSize()) {
					throw new IllegalArgumentException(Console.messages.getString(
							ConsoleMessages.DIFFERENT_INNER_INTERFACES_KEY));
				}
				else if(!automaton.getAlphabet().contains(decomposition.getSignature())) {
					throw new IllegalArgumentException(Console.messages.getString(
							ConsoleMessages.DIFFERENT_SIGNATURES_KEY));
				}
				else {
					//Create and execute new automaton command
					maxInterface = automaton.getMaximumInterface();
					action = new ShiftAutomatonAction(automaton, decomposition);
				}
			}
		}
		else if(parameter.matches(SUBGRAPH_PARAMETER)) {
			//Scan and extract automaton properties from input string
			List<String> list = getProperties(parameter, SUBGRAPH_PARAMETER);
			name = list.get(0);
			maxInterface = Integer.parseInt(list.get(1));
			innerInterface = Integer.parseInt(list.get(2));
			outerInterface = Integer.parseInt(list.get(3));
			String signatureName = list.get(4);
			String subgraphName = list.get(5);
			
			if(name==null) {
				throw new NullPointerException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_NULL_KEY));
			}
			if(name.isEmpty()) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_EMPTY_KEY));
			}
			if(!Console.repo.isSignature(signatureName)) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.SIGNATURE_NOT_FOUND_KEY, signatureName));
			}
			else if(!Console.repo.contains(subgraphName)) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.GRAPH_NOT_FOUND_KEY, subgraphName));
			}
			else {
				Signature signature = (Signature) Console.repo.get(signatureName);
				Graph subgraph = (Graph) Console.repo.get(subgraphName);
				//Create new automaton command
				action = new SubgraphAutomatonAction(maxInterface, innerInterface, outerInterface,
						signature, subgraph);
			}
		}
		else if(parameter.matches(UNION_PARAMETER)) {
			//Scan and extract automaton properties from input string
			List<String> list = getProperties(parameter, UNION_PARAMETER);
			name = list.get(0);
			String firstName = list.get(1);
			String secondName = list.get(2);
			
			if(name==null) {
				throw new NullPointerException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_NULL_KEY));
			}
			if(name.isEmpty()) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_EMPTY_KEY));
			}
			if(!Console.repo.isAutomaton(firstName)) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NOT_FOUND_KEY, firstName));
			}
			else if(!Console.repo.isAutomaton(secondName)) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NOT_FOUND_KEY, firstName));
			}
			else {
				Automaton first = (Automaton) Console.repo.get(firstName);
				Automaton second = (Automaton) Console.repo.get(secondName);
				
				if(first.getMaximumInterface()!=second.getMaximumInterface()) {
					throw new IllegalArgumentException(Console.messages.getString(
							ConsoleMessages.DIFFERENT_MAXIMUM_INTERFACES_KEY));
				}
				else if(!first.getAlphabet().equals(second.getAlphabet())) {
					throw new IllegalArgumentException(Console.messages.getString(
							ConsoleMessages.DIFFERENT_SIGNATURES_KEY));
				}
				else {
					//Create and execute new automaton command
					maxInterface = first.getMaximumInterface();
					action = new UnionAutomatonAction(first, second);
				}
			}
		}
		else if(parameter.matches(VERTEX_COUNTING_PARAMETER)) {
			//Scan and extract automaton properties from input string
			List<String> list = getProperties(parameter, VERTEX_COUNTING_PARAMETER);
			name = list.get(0);
			maxInterface = Integer.parseInt(list.get(1));
			innerInterface = Integer.parseInt(list.get(2));
			outerInterface = Integer.parseInt(list.get(3));
			String signatureName = list.get(4);
			int divisor = Integer.parseInt(list.get(5));
			int remainder = Integer.parseInt(list.get(6));
			
			if(name==null) {
				throw new NullPointerException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_NULL_KEY));
			}
			if(name.isEmpty()) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_EMPTY_KEY));
			}
			if(!Console.repo.isSignature(signatureName)) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.SIGNATURE_NOT_FOUND_KEY, signatureName));
			}
			if(divisor<1) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.DIVISOR_LESS_THAN_ONE_KEY));
			}
			if(remainder<0) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.REMAINDER_LESS_THAN_ZERO_KEY));
			}
			if(remainder>=divisor) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.REMAINDER_GREATER_OR_EQUAL_THAN_DIVISOR_KEY));
			}

			Signature signature = (Signature) Console.repo.get(signatureName);
			//Create and execute new automaton command
			action = new VertexCountingAutomatonAction(maxInterface, innerInterface, outerInterface,
					signature, divisor, remainder);
		}
		else if(parameter.matches(VERTEX_COVER_PARAMETER)) {
			//Scan and extract automaton properties from input string
			List<String> list = getProperties(parameter, VERTEX_COVER_PARAMETER);
			name = list.get(0);
			maxInterface = Integer.parseInt(list.get(1));
			innerInterface = Integer.parseInt(list.get(2));
			outerInterface = Integer.parseInt(list.get(3));
			String signatureName = list.get(4);
			int coverSize = Integer.parseInt(list.get(5));
			
			if(name==null) {
				throw new NullPointerException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_NULL_KEY));
			}
			if(name.isEmpty()) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NAME_IS_EMPTY_KEY));
			}
			if(!Console.repo.isSignature(signatureName)) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.SIGNATURE_NOT_FOUND_KEY, signatureName));
			}

			Signature signature = (Signature) Console.repo.get(signatureName);
			//Create and execute new automaton command
			action = new VertexCoverAutomatonAction(maxInterface, 
					innerInterface, outerInterface, signature, coverSize);
		}
		else {
			throw new IllegalArgumentException(Console.messages.getString(
					ConsoleMessages.MISSING_ILLEGAL_ARGUMENT_DESCRIPTION_KEY, AUTOMATON_KEY));
		}

		//Execute automaton command
		if(Console.progressThread!=null) {
			action.addComputationListener(Console.progressThread);
		}

		action.execute();
		
		//Print elapsed runtime (if time mode is activated)
		printRuntime(action.getElapsedTime());

		Repository.getRepository().put(name, action.getResult());
		Console.out.println(Console.messages.getString(
				ConsoleMessages.AUTOMATON_ADDED_KEY, name, String.valueOf(maxInterface)));

	}

}
