/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.cli;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Scanner;
import java.util.logging.Logger;
import java.util.regex.MatchResult;

/**
 * Represents the abstract super class of all console commands.
 * @author Christoph Blume
 * @version $Revision: 1.24 $
 *
 */
public abstract class Command {
	
	/** */
	protected final ConsoleMessages messages;
	
	private final String key;
	private final String description;
	private final String documentation;

	private Logger logger;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new command-line command.
	 * @param key the command key
	 * @param description the key for the short description of the command
	 * @param documentation the full documentation of the command's usage and function
	 */
	protected Command(String key, String description, String documentation) {
		if(key==null) {
			throw new NullPointerException(
					Console.messages.getString(ConsoleMessages.NULL_POINTER_KEY, "command key"));
		}
		if(key.isEmpty()) {
			throw new IllegalArgumentException(
					Console.messages.getString(ConsoleMessages.EMPTY_STRING_KEY, "command key"));
		}
		if(description==null) {
			throw new NullPointerException(
					Console.messages.getString(ConsoleMessages.NULL_POINTER_KEY, "description"));
		}
		if(description.isEmpty()) {
			throw new IllegalArgumentException(
					Console.messages.getString(ConsoleMessages.EMPTY_STRING_KEY, "description"));
		}
		if(documentation==null) {
			throw new NullPointerException(
					Console.messages.getString(ConsoleMessages.NULL_POINTER_KEY, "documentation"));
		}
		if(documentation.isEmpty()) {
			throw new IllegalArgumentException(
					Console.messages.getString(ConsoleMessages.EMPTY_STRING_KEY, "documentation"));
		}
		
		this.messages = ConsoleMessages.getMessages();
		this.key = key;
		this.description = description;
		this.documentation = documentation;
		
		//Get logger
		logger = Logger.getLogger(getClass().getName());
	}
	
	/**
	 * Creates a new command-line command.
	 * @param key the command key
	 * @param descriptionKey the key for the short description of the command
	 * @param documentationFile the file containing a full documentation of the command's usage and function
	 */
	protected Command(String key, String descriptionKey, File documentationFile) {	
		if(key==null) {
			throw new NullPointerException(
					Console.messages.getString(ConsoleMessages.NULL_POINTER_KEY, "command key"));
		}
		if(key.isEmpty()) {
			throw new IllegalArgumentException(
					Console.messages.getString(ConsoleMessages.EMPTY_STRING_KEY, "command key"));
		}
		if(descriptionKey==null) {
			throw new NullPointerException(
					Console.messages.getString(ConsoleMessages.NULL_POINTER_KEY, "description key"));
		}
		if(descriptionKey.isEmpty()) {
			throw new IllegalArgumentException(
					Console.messages.getString(ConsoleMessages.EMPTY_STRING_KEY, "description key"));
		}
//		if(documentationFile==null) {
//			throw new NullPointerException(
//					Console.messages.getString(ConsoleMessages.NULL_POINTER_KEY, "documentation file"));
//		}
		
		//Get logger
		logger = Logger.getLogger(getClass().getName());
		
		this.messages = ConsoleMessages.getMessages();
		this.key = key;
		this.description = Console.messages.getString(descriptionKey);

		StringBuilder documentation = new StringBuilder();
		if(documentationFile!=null) {
			try (BufferedReader reader = new BufferedReader(
					new FileReader(documentationFile))) {
				String line;
				while((line=reader.readLine())!=null) {
					documentation.append(line+System.lineSeparator());
				}
			} catch (FileNotFoundException e) {
				Console.out.print(Console.messages.getString(
						ConsoleMessages.FILE_NOT_FOUND_KEY, documentationFile.getName()));
				logger.throwing(getClass().getName(), "Command", e);
				documentation = new StringBuilder();
			} catch (IOException e) {
				Console.out.print(Console.messages.getString(
						ConsoleMessages.EXCEPTION_OCCURED_KEY));
				logger.throwing(getClass().getName(), "Command", e);
				documentation = new StringBuilder();
			} finally {
				this.documentation = documentation.toString();
			}
		}
		else {
			this.documentation = Console.messages.getString(ConsoleMessages.NO_DOCUMENTATION_KEY);
		}
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Abstract Methods																*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns {@code true} if and only if the string {@code line} is 
	 * a valid parameter for this command.
	 * @param line the input string
	 * @return {@code true} if {@code line} is a valid parameter, 
	 * {@code false} otherwise
	 */
	public abstract boolean matchesParameter(String line);
	
	/**
	 * Executes this command with the given parameter.
	 * @param parameter the parameter passed to the command
	 */
	public abstract void execute(String parameter);
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the command key of this command.
	 * @return the command key
	 */
	public String getKey() {
		return key;
	}
	
	/**
	 * Returns a short description of this command.
	 * @return a short informational description
	 */
	public String getDescription() {
		return description;
	}
	
	/**
	 * Returns a documentation of the usage of this command.
	 * @return a usage documentation
	 */
	public String getDocumentation() {
		return documentation;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Parses the properties from an input string {@code input} according to 
	 * a regular expression-based pattern {@code pattern} and returns the 
	 * properties as list in the order in which they occur in the pattern.
	 * @param input the input string containing the properties' values
	 * @param pattern the pattern to be used for the extraction of the values
	 * @return a list containing the values of the properties
	 */
	protected static List<String> getProperties(String input, String pattern) {
		List<String> propertiesList = new ArrayList<>();

		//Scan and extract properties from input string
		try(Scanner s = new Scanner(input)) {
			s.findInLine(pattern);
			MatchResult result = s.match();
			for(int i=1;i<=result.groupCount();i++) {
				propertiesList.add(result.group(i));
			}
		};

		return propertiesList;
	}
	
	/**
	 * @param seconds elapsed time in seconds
	 * 
	 */
	protected void printRuntime(long seconds) {
		if(Console.timeStatsActivated) {
			Console.out.println(messages.getString(
					ConsoleMessages.ELAPSED_TIME_KEY, String.valueOf(seconds)));
		}
	}
	
}
