/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.cli;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Logger;

import org.fusesource.jansi.Ansi;
import org.fusesource.jansi.Ansi.Color;

import de.uni_due.inf.ti.raven.actions.ComputationEvent;
import de.uni_due.inf.ti.raven.actions.ComputationListener;
import de.uni_due.inf.ti.raven.data.Cospan;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Decomposition;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.Operation.Connect;
import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.io.CommandReader;
import de.uni_due.inf.ti.raven.io.ConsolePrinter;
import de.uni_due.inf.ti.raven.io.RavenPrinter;
import de.uni_due.inf.ti.raven.main.RavenMain;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class implements a command-line interface and a command-line interpreter for Raven.
 * @author Christoph Blume
 * @version $Revision: 1.95 $
 *
 */
public final class Console implements Runnable {

	private static final boolean DEBUG_SHOW_PROGRESS_BAR = true;
	
	/**
	 * This class holds environmental settings.
	 * @author Christoph Blume
	 *
	 */
	protected static class Environment {

		/**
		 * Enumeration of the different environment types.
		 * @author Christoph Blume
		 *
		 */
		public enum Type {
			/** The Batch Mode Type */
			BATCH,
			
			/** The Cospan Define Mode Type */
			COSPAN,
			
			/** The Decomposition Define Mode Type */
			DECOMPOSITION, 
			
			/** The Default Interactive Mode Type */
			DEFAULT,
			
			/** The Formula Define Mode Type */
			FORMULA,
			
			/** The Graph Define Mode Type */
			GRAPH
		}

		private static final String DEFAULT_PROMPT = "> ";
		private static final String COSPAN_PROMPT = "$ ";
		private static final String DECOMP_PROMPT = "$ ";
		private static final String FORMULA_PROMPT = "$ ";
		private static final String GRAPH_PROMPT = "$ ";

		/** The list of commands usable in the default interactive mode */
		protected static final List<Command> DEFAULT_COMMAND_LIST = new LinkedList<>();
		
		/** The list of commands usable in the copsan define mode */
		protected static final List<Command> COSPAN_COMMAND_LIST= new LinkedList<>();
		
		/** The list of commands usable in the decomposition define mode */
		protected static final List<Command> DECOMP_COMMAND_LIST= new LinkedList<>();
		
		/** The list of commands usable in the formula define mode */
		protected static final List<Command> FORMULA_COMMAND_LIST = new LinkedList<>();
		
		/** The list of commands usable in the graph define mode */
		protected static final List<Command> GRAPH_COMMAND_LIST = new LinkedList<>();

		private static Type type;
		private static Type previousType;
		private static String prompt;
		
		/** The cospan defined in the cospan define mode */
		protected static Cospan cospan;
		
		/** The cospan decomposition defined in the decomposition define mode */
		protected static CospanDecomposition decomposition;
		
		/** The formula defined in the formula define mode */
		protected static String formula;
		
		/** The graph defined in the graph define mode */
		protected static Graph graph;
		
		/** The name of the defined cospan, cospan decomposition, formula or graph */
		protected static String name;

		/*------------------------------------------------------------------------------------------*/
		/*		Static Methods																		*/
		/*------------------------------------------------------------------------------------------*/
		
		/**
		 * Returns the previous environment mode type.
		 * @return the previous environment mode type
		 */
		public static Type getPreviousType() {
			return previousType;
		}

		/**
		 * Returns the current environment mode type.
		 * @return current environment mode type
		 */
		public static Type getType() {
			return type;
		}

		/**
		 * Sets the current environment mode type
		 * @param type the new environment mode type
		 */
		public static void setType(Type type) {
			previousType = Environment.type;
			Environment.type = type; 
			switch (type) {
			case BATCH:
			case DEFAULT:
				prompt = DEFAULT_PROMPT;
				break;
			
			case COSPAN:
				prompt = COSPAN_PROMPT;
				break;

			case DECOMPOSITION:
				prompt = DECOMP_PROMPT;
				break;

			case FORMULA:
				prompt = FORMULA_PROMPT;
				break;
				
			case GRAPH:
				prompt = GRAPH_PROMPT;
				break;

			default:
				break;
			}
		}

	}

	/**
	 * 
	 * @author Christoph Blume
	 *
	 */
	public class ProgressThread extends Thread implements ComputationListener {

		private int width;

		/*------------------------------------------------------------------------------------------*/
		/*		Constructors																		*/
		/*------------------------------------------------------------------------------------------*/
		
		/**
		 * Creates a new thread to display the computation progress.
		 * @param width the width of the progress bar
		 */
		public ProgressThread(int width) {
			this.width = width;
		}

		/*------------------------------------------------------------------------------------------*/
		/*		Public Methods																		*/
		/*------------------------------------------------------------------------------------------*/
		
		@Override
		public void run() {
			char[] signs = {'-', '\\', '|', '/'};

			try {
				while(true) {
					for(char sign: signs) {
						Thread.sleep(300);
						synchronized (out) {
							out.print(sign+"\r");
						}
					}
				}
			}
			//Fail silently
			catch (Exception e) {}
		}


		@Override
		public void computationAdvanded(ComputationEvent e) {
			int maximum = e.getMaximum();
			int value = e.getValue();
			float percentage = (float) value/(float) maximum;
			int length = Math.round(percentage*width);

			char[] filling = new char[length];
			Arrays.fill(filling, '=');
			char[] whiteSpace = new char[width-length];
			Arrays.fill(whiteSpace, ' ');

			String message = "Compute "+e.getName();
			out.print(message+" [" + String.valueOf(filling) + 
					String.valueOf(whiteSpace) + "] " + String.format("%3.2f", (percentage*100))+"%\r");
			if(value==maximum) {
				whiteSpace = new char[width+8];
				Arrays.fill(whiteSpace, ' ');
				out.println(message+" done"+String.valueOf(whiteSpace));
			}
		}

	};

	/** */
	private static final String COMMAND_PATTERN = "\\w+(\\s\\p{Graph}*)*|\\?";
	
	/** The Comment Token */
	public static final String COMMENT_TOKEN = "#";

	
	/** The ADD Token */
	protected static final String ADD_TOKEN = "add";
	
	/** The ALL START LEX BFS Algorithm Token */
	protected static final String ALL_START_LEX_BFS = "allstart";
	
	/** The ANTICHAIN Token */
	protected static final String ANTICHAIN_TOKEN = "antichain";
	
	/** The ARITY Token */
	public static final String ARITY_TOKEN = Connect.ARITY_TOKEN;
	
	/** The AUTOMATON Token */
	protected static final String AUTOMATON_TOKEN = "automaton";
	
	/** The AUTOMATON_RUN Token */
	protected static final String AUTOMATON_RUN_TOKEN = "run";
	
	/** The BACKWARD_SEARCH Token */
	protected static final String BACKWARD_TOKEN = "backward";
	
	/** The COLORABILITY AUTOMATON TYPE Token */
	protected static final String COLORABILTY_TOKEN = Automaton.Type.colorability.name();
	
	/** The COSPAN Token */
	protected static final String COSPAN_TOKEN = "cospan";
	
	/** The COSPAN Decomposition Type TOKEN */
	protected static final String COSPAN_DECOMPOSITION_TOKEN = Decomposition.Type.cospanDecomposition.toString();
	
	/** The DECOMPOSE Token */
	protected static final String DECOMPOSE_TOKEN = "decompose";
	
	/** The DECOMPOSITION Token */
	protected static final String DECOMPOSITION_TOKEN = "decomposition";
	
	/** The DEFINE Token */
	protected static final String DEFINE_TOKEN = "define";
	
	/** The DELETE Token */
	protected static final String DELETE_TOKEN = "delete";
	
	/** The DELIMITER Token */
	protected static final String DELIMITER_TOKEN = "[]";
	
	/** The DOMINATING SET AUTOMATON TYPE Token */
	protected static final String DOMINATING_SET_TOKEN = Automaton.Type.dominatingSet.name();
	
	/** The EDGE Token */
	protected static final String EDGE_TOKEN = "edge";
	
	/** The EDGE COUNTING AUTOMATON TYPE Token */
	protected static final String EDGE_COUNTING_TOKEN = Automaton.Type.edgeCounting.name();
	
	/** The EMPTINESS Token */
	protected static final String EMPTINESS_TOKEN = "emptiness";

	/** The END Token */
	protected static final String END_TOKEN = "end";
	
	/** The EVALUATION Token */
	protected static final String EVALUATION_TOKEN = "%";
	
	/** The EXIT Token */
	protected static final String EXIT_TOKEN = "exit";
	
	/** The EXPORT Token */
	protected static final String EXPORT_TOKEN = "export";

	/** The FALSE Token */
	protected static final String FALSE_TOKEN = "false";
	
	/** The FILE Token */
	protected static final String FILE_TOKEN = "file";
	
	/** The FINAL Token */
	protected static final String FINAL_TOKEN = "final";
	
	/** The FORMULA Token */
	protected static final String FORMULA_TOKEN = "formula";
	
	/** The FORWARD SEARCH Token */
	protected static final String FORWARD_TOKEN = "forward";
	
	/** The GREEDY DEGREE Algorithm Token */
	protected static final String GREEDY_DEGREE_TOKEN = "greedydegree";
	
	/** The GREEDY FILL-IN Algorithm Token */
	protected static final String GREEDY_FILL_IN_TOKEN = "greedyfillin";
		
	/** The GRAPH Token */
	protected static final String GRAPH_TOKEN = "graph";
	
	/** The HELP Token */
	protected static final String HELP_TOKEN = "help";
	
	/** The IMPORT Token */
	protected static final String IMPORT_TOKEN = "import";

	/** The INFO Token */
	protected static final String INFO_TOKEN = "info";
	
	/** The INITIAL Token */
	protected static final String INITIAL_TOKEN = "initial";
	
	/** The INNER Token */
	protected static final String INNER_TOKEN = "inner";
	
	/** The INTERFACE Token */
	protected static final String INTERFACE_TOKEN = "interface";
	
	/** The INVARIANT CHECK Token */
	protected static final String INVARIANT_TOKEN = "invariant";
	
	/** The IS Token */
	protected static final String IS_TOKEN = "\\s*=\\s*";
	
	/** The LANGUAGE INCLUSION Token */
	protected static final String LANGUAGE_INCLUSION_TOKEN = "langincl";
	
	/** The LEX BFS Algorithm Token */
	protected static final String LEX_BFS = "lexbfs";
	
	/** The Link Automaton Type Token */
	protected static final String LINK_TOKEN = Automaton.Type.link.name();

	/** The LOAD Token */
	protected static final String LOAD_TOKEN = "load";

	/** The MAXMIMUM EDGE AUTOMATON TYPE Token */
	protected static final String MAXIMUM_EDGE_TOKEN = Automaton.Type.maximumEdge.name();
	
	/** The MAXIMUM VERTEX AUTOMATON TYPE Token */
	protected static final String MAXIMUM_VERTEX_TOKEN = Automaton.Type.maximumVertex.name();
	
	/** The MEMBERSHIP Token */
	protected static final String MEMBERSHIP_TOKEN = "membership";
	
	/** The MINIMUM EDGE AUTOMATON TYPE Token */
	protected static final String MINIMUM_EDGE_TOKEN = Automaton.Type.minimumEdge.name();
	
	/** The MINIMUM VERTEX AUTOMATON TYPE Token */
	protected static final String MINIMUM_VERTEX_TOKEN = Automaton.Type.minimumVertex.name();
	
	/** THE NO Token */
	protected static final String NO_TOKEN = "n";
	
	/** The NO ISOLATED NODES AUTOMATON TYPE Token */
	protected static final String NO_ISOLATED_NODE_TOKEN = Automaton.Type.noIsolatedNodes.name();
	
	/** The NODE Token */
	protected static final String NODE_TOKEN = "node";
	
	/** The ON Token */
	protected static final String ON_TOKEN = "on";
	
	/** The OFF Token */
	protected static final String OFF_TOKEN = "off";
	
	/** The OUTER Token */
	protected static final String OUTER_TOKEN = "outer";
	
	/** The PATH AUTOMATON TYPE Token */
	protected static final String PATH_AUTOMATON_TOKEN = Automaton.Type.path.name();
	
	/** The PATH Decomposition Type TOKEN */
	protected static final String PATH_DECOMPOSITION_TOKEN = "path";
	
	/** The POWER SET Token */
	protected static final String POWER_SET_TOKEN = "powerset";
	
	/** The PRINT Token */
	protected static final String PRINT_TOKEN = "print";
	
	/** The PRINT DOT Token */
	protected static final String PRINT_DOT_TOKEN = "printdot";
	
	/** The PRODUCT AUTOMATON TYPE Token */
	protected static final String PRODUCT_TOKEN = Automaton.Type.product.name();
	
	/** The QUIT Token */
	protected static final String QUIT_TOKEN = "quit";

	/** The RANDOMGRAPH Token */
	protected static final String RANDOMGRAPH_TOKEN = "randomGraph";
	
	/** The REACHABILITY Token */
	protected static final String REACHABILITY_TOKEN = "reachability";
	
	/** The REMOVE Token */
	protected static final String REMOVE_TOKEN = "remove";
	
	/** The RENAME Token */
	protected static final String RENAME_TOKEN = "rename";
	
	/** The REPOSITORY Token */
	protected static final String REPOSITORY_TOKEN = "repository";
	
	/** The RUN Token */
	protected static final String RUN_TOKEN = "run";
	
	/** The SAVE Token */
	protected static final String SAVE_TOKEN = "save";
	
	/** The SEP Token */
	protected static final String SEP_TOKEN = "\\s*,\\s*";
	
	/** The SET Token */
	protected static final String SET_TOKEN = "set";
	
	/** The SEP Token */
	protected static final String SET_SEP_TOKEN = "\\s*;\\s*";
	
	/** The SHIFT Token */
	protected static final String SHIFT_TOKEN = "shift";
	
	/** The SHOW Token */
	protected static final String SHOW_TOKEN = "show";
	
	/** The STATES DD Token */
	protected static final String STATES_TOKEN = "states";
	
	/** The SUBGRAPH AUTOMATON TYPE Token */
	protected static final String SUBGRAPH_TOKEN = Automaton.Type.subgraph.name();
	
	/** The SIGNATURE Token */
	protected static final String SIGNATURE_TOKEN = "signature";
	
	/** The SIMULATION Token */
	protected static final String SIMULATION_TOKEN = "simulation";
	
	/** */
	protected static final String SYSTEM_VARIABLE_TOKEN = "_";
	
	/** The TEST Token */
	protected static final String TEST_TOKEN = "test";
	
	/** The TRANSITION Token */
	protected static final String TRANSITION_TOKEN = "transition";
	
	/** The TREE Decomposition Type TOKEN */
	protected static final String TREE_DECOMPOSITION_TOKEN = "tree";
	
	/** The TRUE Token */
	protected static final String TRUE_TOKEN = "true";
	
	/** The UNION AUTOMATON TYPE Token */
	protected static final String UNION_TOKEN = Automaton.Type.union.name();
	
	/** The UNIVERSALITY Token */
	protected static final String UNIVERSALITY_TOKEN = "universality";

	/** The UPDATE Token */
	protected static final String UPDATE_TOKEN = "update";
	
	/** The BISIMULATION-UP-TO-X Token */
	protected static final String UP_TO_TOKEN = "upTo";
	
	/** The VERTEX COUNTING AUTOMATON TYPE Token */
	protected static final String VERTEX_COUNTING_TOKEN = Automaton.Type.vertexCounting.name();
	
	/** The VERTEX COVER AUTOMATON TYPE Token */
	protected static final String VERTEX_COVER_TOKEN = Automaton.Type.vertexCover.name();
	
	/** The WHITE SPACE Token */
	protected static final String WHITE_SPACE_TOKEN = "\\s*";
	
	/** The YES Token */
	protected static final String YES_TOKEN = "y";
	
	
	/* Patterns */	
	/** The ACTIVATE Pattern */
	protected static final String ACTIVATE_PATTERN = ON_TOKEN+"|"+TRUE_TOKEN;
	
	/** The BOOLEAN Pattern */
	protected static final String BOOLEAN_PATTERN = TRUE_TOKEN+"|"+FALSE_TOKEN;
	
	/** The DEACTIVATE Pattern */
	protected static final String DEACTIVATE_PATTERN = FALSE_TOKEN+"|"+OFF_TOKEN;
	
	/** The FILENAME Pattern */
	protected static final String FILENAME_PATTERN = "(\\w+(\\\\|/))*(\\w+\\.\\w+||\\w+)";
	
	/** The EDGE LABEL Pattern */
	public static final String LABEL_PATTERN = "\\p{Alpha}+|" + Graph.DEFAULT_EDGE_LABEL_STRING;
	
	/** The NAME Pattern */
	protected static final String NAME_PATTERN = Repository.NAME_PATTERN;
	
	/** The NUMBER Pattern */
	public static final String NUMBER_PATTERN = "\\d+";
	
	/** The SEARCH TYPE Pattern */
	protected static final String SEARCH_TYPE_PATTERN = FORWARD_TOKEN + "|" + BACKWARD_TOKEN;
	
	/** The SET OF NUMBERS Pattern */
	protected static final String SET_OF_NUMBERS_PATTERN = "\\{"+NUMBER_PATTERN+"("+SET_SEP_TOKEN+NUMBER_PATTERN+")*\\}";
	
	/** */
	protected static final String VALUE_PATTERN = "\\p{Alnum}\\p{Print}*";

	/** */
	protected static final String VARIABLE_PATTERN = "\\p{Alnum}+";
	
	/* Letter Tokens */
	/** The Letter A Token */
	public static final String LETTER_A_TOKEN = "a";
	
	/** The Letter B Token */
	public static final String LETTER_B_TOKEN = "b";
	
	/** The Letter C Token */
	public static final String LETTER_C_TOKEN = "c";

	
	/* Graph Operation Patterns */
	/** The CONNECT Operation Token */
	public static final String CONNECT_TOKEN = Operation.CONNECT_TOKEN;
	
	/** The CONNECT parameter specifying the CONNECT Operation */
	public static final String CONNECT_PARAMETER = "\\(("+LABEL_PATTERN+")"+ARITY_TOKEN+"("+NUMBER_PATTERN+")\\)";
	
	/** The FUSE Operation Token */
	public static final String FUSE_TOKEN = Operation.FUSE_TOKEN;
	
	/** The Perm Operation Token */
	public static final String PERM_TOKEN = Operation.PERM_TOKEN;
	
	/** The Res Operation Token */
	public static final String RES_TOKEN = Operation.RES_TOKEN;
	
	/** The TRANS Operation Token */
	public static final String TRANS_TOKEN = Operation.TRANS_TOKEN;
	
	/** The VERTEX Operation Token */
	public static final String VERTEX_TOKEN = Operation.VERTEX_TOKEN;
	
	/** THE INNER INTERFACE SIZE Token */
	public static final String INTERFACE_SIZE_TOKEN = "interface";
	
	
	/** The TRANSITION Pattern */
	public static final String TRANSITION_PATTERN = CONNECT_TOKEN + "|" + FUSE_TOKEN + "|" + PERM_TOKEN + "|"
			+ RES_TOKEN + "|" + TRANS_TOKEN + "|" + VERTEX_TOKEN;
	
	/** The OPERATION Pattern */
	public static final String OPERATION_PATTERN = "(" + CONNECT_TOKEN + CONNECT_PARAMETER + "|" + 
			FUSE_TOKEN + "|" + PERM_TOKEN + "|" + RES_TOKEN + "|" + TRANS_TOKEN + "|" + VERTEX_TOKEN + ")";

	/** The INNER INTERFACE SIZE Pattern */
	public static final String INTERFACE_PATTERN = INTERFACE_SIZE_TOKEN + IS_TOKEN + "("+NUMBER_PATTERN+")";
	
	/* Console Commands */
	private final AddCommand ADD_CMD;
	private final AutomatonCommand AUTOMATON_CMD;
	private final DecomposeCommand DECOMPOSE_CMD;
	private final DefineCommand DEFINE_CMD;
	private final DeleteCommand DELETE_CMD;
	private final EdgeCommand EDGE_CMD;
	private final EmptinessCommand EMPTINESS_CMD;
	private final EndCommand END_CMD;
	private final ExitCommand EXIT_CMD;
	private final HelpCommand HELP_CMD;
	private final InvariantCommand INVARIANT_CMD;
	private final InfoCommand INFO_CMD;
	private final InnerInterfaceCommand INNER_INTERFACE_CMD;
	private final LanguageInclusionCommand LANGUAGE_INCLUSION_CMD;
	private final LoadCommand LOAD_CMD;
	private final MembershipCommand MEMBERSHIP_CMD;
	private final NodeCommand NODE_CMD;
	private final OuterInterfaceCommand OUTER_INTERFACE_CMD;
	private final PrintCommand PRINT_CMD;
	private final PrintDotCommand PRINT_DOT_CMD;
	private final QuitCommand QUIT_CMD;
	private final ReachabilityCommand REACHABILITY_CMD;
	private final RemoveCommand REMOVE_CMD;
	private final RenameCommand RENAME_CMD;
	private final RepositoryCommand REPOSITORY_CMD;
	private final RunCommand RUN_CMD;
	private final SaveCommand SAVE_CMD;
	private final SetCommand SET_CMD;
	private final ShowCommand SHOW_CMD;
	private final SignatureCommand SIGNATURE_CMD;
	private final SimulationCommand SIMULATION_CMD;
	private final UniversalityCommand UNIVERSALITY_CMD;
	private final UpdateCommand UPDATE_CMD;

	//Logger to log error messages
	private final Logger logger;
	
	/** Console Messages  */
	public static ConsoleMessages messages;
	
	/** Console Input  */
	protected static BufferedReader in;
	
	/** Console Output  */
	protected static RavenPrinter out;
	
	/** Repository containing all loaded objects */
	protected static Repository repo;
	
	/** Thread for displaying a progress bar */
	protected static ProgressThread progressThread;
	
	/** Flag for sloppy mode */
	protected static boolean sloppyActivated;
	
	/** Flag for verbose mode */
	protected static boolean verboseActivated;
	
	/** Flag for time statistic mode */
	protected static boolean timeStatsActivated;
	
	/** Flag for interactive mode */
	protected static boolean isInteractive;
	
	/** List of commands for the console-oriented mode */
	protected static List<Command> commandList;

	/** The console */
	private static Console console;


	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns a command-line console for Raven with the system input stream as input stream and 
	 * the default console printer as output stream.
	 * @return the console
	 */
	public static Console getConsole() {
		return getConsole(System.in, ConsolePrinter.getPrinter());
	}
	
	/**
	 * Returns a command-line console for Raven.
	 * @param in the input stream of the console
	 * @param out the output stream of the console
	 * @return the console
	 */
	public static Console getConsole(InputStream in, RavenPrinter out) {
		return getConsole(in, out, true);
	}
	
	/**
	 * Returns a command-line console for Raven.
	 * @param in the input stream of the console
	 * @param out the output stream of the console
	 * @param isInteractive indicates whether this console is interactive
	 * @return the console
	 */
	public static Console getConsole(InputStream in, RavenPrinter out, boolean isInteractive) {
		if(console==null) {
			messages = ConsoleMessages.getMessages();
			console = new Console(in, out);
			Console.isInteractive = isInteractive;
			//			Console.progressThread = console.new ProgressThread(40);
		}

		return console;
	}
	
	/**
	 * Executes the commands of the batch-file given by the filename {@code filename}. After the 
	 * execution Raven is terminated.
	 * @param filename the name of batch-file
	 * @throws FileNotFoundException thrown if the batch-file can not be found
	 */
	public static void runBatchFile(String filename) 
			throws FileNotFoundException {
		runBatchFile(filename, ConsolePrinter.getPrinter());
	}
	
	/**
	 * Executes the commands of the batch-file given by the filename {@code filename} and uses 
	 * {@code out} as output stream. After the execution Raven is terminated.
	 * @param filename the name of batch-file
	 * @param out the output stream of the console
	 * @throws FileNotFoundException thrown if the batch-file can not be found
	 */
	public static void runBatchFile(String filename, RavenPrinter out) 
			throws FileNotFoundException {
		console = null;
		try (FileInputStream stream = new FileInputStream(filename)) {
			console = getConsole(stream, out);
			Console.isInteractive = false;
			
			//Since Raven was started in batch mode, it is not necessary to start 
			//the console as (concurrent) thread -> use run() instead of start()
			console.run();
		} catch (IOException e) {
			Logger.getLogger(Console.class.getName()).throwing(Console.class.getName(), "runBatchFile", e);
			out.println(RavenExceptionMessages.IO_EXCEPTION_OCCURRED_EXCEPTION);
			RavenMain.exit(RavenMain.EXIT_CODE_UNEXPECTED_ERROR);
		}

		RavenMain.exit(RavenMain.EXIT_CODE_NORMAL_TERMINATION);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new console.
	 * @param in the console's input stream
	 * @param out the console's output printer
	 */
	private Console(InputStream in, RavenPrinter out) {
		Console.in = new CommandReader(in);
		Console.out = out;
		Console.repo = Repository.getRepository();
		Environment.setType(Environment.Type.DEFAULT);

		//Add default commands
		AUTOMATON_CMD = AutomatonCommand.getCommand();
		Environment.DEFAULT_COMMAND_LIST.add(AUTOMATON_CMD);
		DECOMPOSE_CMD = DecomposeCommand.getCommand();
		Environment.DEFAULT_COMMAND_LIST.add(DECOMPOSE_CMD);
		DEFINE_CMD = DefineCommand.getCommand();
		Environment.DEFAULT_COMMAND_LIST.add(DEFINE_CMD);
		DELETE_CMD = DeleteCommand.getCommand();
		Environment.DEFAULT_COMMAND_LIST.add(DELETE_CMD);
		EMPTINESS_CMD = EmptinessCommand.getCommand();
		Environment.DEFAULT_COMMAND_LIST.add(EMPTINESS_CMD);
		EXIT_CMD = ExitCommand.getCommand();
		Environment.DEFAULT_COMMAND_LIST.add(EXIT_CMD);
		HELP_CMD = HelpCommand.getCommand();
		Environment.DEFAULT_COMMAND_LIST.add(HELP_CMD);
		INFO_CMD = InfoCommand.getCommand();
		Environment.DEFAULT_COMMAND_LIST.add(INFO_CMD);
		INVARIANT_CMD = InvariantCommand.getCommand();
		Environment.DEFAULT_COMMAND_LIST.add(INVARIANT_CMD);
		LANGUAGE_INCLUSION_CMD = LanguageInclusionCommand.getCommand();
		Environment.DEFAULT_COMMAND_LIST.add(LANGUAGE_INCLUSION_CMD);
		LOAD_CMD = LoadCommand.getCommand();
		Environment.DEFAULT_COMMAND_LIST.add(LOAD_CMD);
		MEMBERSHIP_CMD = MembershipCommand.getCommand();
		Environment.DEFAULT_COMMAND_LIST.add(MEMBERSHIP_CMD);
		PRINT_CMD = PrintCommand.getCommand();
		Environment.DEFAULT_COMMAND_LIST.add(PRINT_CMD);
		PRINT_DOT_CMD = PrintDotCommand.getCommand();
		Environment.DEFAULT_COMMAND_LIST.add(PRINT_DOT_CMD);
		QUIT_CMD = QuitCommand.getCommand();
		Environment.DEFAULT_COMMAND_LIST.add(QUIT_CMD);
		REACHABILITY_CMD = ReachabilityCommand.getCommand();
		Environment.DEFAULT_COMMAND_LIST.add(REACHABILITY_CMD);
		REMOVE_CMD = RemoveCommand.getCommand();
		Environment.DEFAULT_COMMAND_LIST.add(REMOVE_CMD);
		RENAME_CMD = RenameCommand.getCommand();
		Environment.DEFAULT_COMMAND_LIST.add(RENAME_CMD);
		REPOSITORY_CMD = RepositoryCommand.getCommand();
		Environment.DEFAULT_COMMAND_LIST.add(REPOSITORY_CMD);
		RUN_CMD = RunCommand.getCommand();
		Environment.DEFAULT_COMMAND_LIST.add(RUN_CMD);
		SAVE_CMD = SaveCommand.getCommand();
		Environment.DEFAULT_COMMAND_LIST.add(SAVE_CMD);
		SET_CMD = SetCommand.getCommand();
		Environment.DEFAULT_COMMAND_LIST.add(SET_CMD);
		SHOW_CMD = ShowCommand.getCommand();
		Environment.DEFAULT_COMMAND_LIST.add(SHOW_CMD);
		SIGNATURE_CMD = SignatureCommand.getCommand();
		Environment.DEFAULT_COMMAND_LIST.add(SIGNATURE_CMD);
		SIMULATION_CMD = SimulationCommand.getCommand();
		Environment.DEFAULT_COMMAND_LIST.add(SIMULATION_CMD);
		UNIVERSALITY_CMD = UniversalityCommand.getCommand();
		Environment.DEFAULT_COMMAND_LIST.add(UNIVERSALITY_CMD);
		UPDATE_CMD = UpdateCommand.getCommand();
		Environment.DEFAULT_COMMAND_LIST.add(UPDATE_CMD);

		//TODO Remove the following line of code!
		Environment.DEFAULT_COMMAND_LIST.add(TestCommand.getCommand());
		
		//Add cospan definition mode commands
		EDGE_CMD = EdgeCommand.getCommand();
		Environment.COSPAN_COMMAND_LIST.add(EDGE_CMD);
		END_CMD = EndCommand.getCommand();
		Environment.COSPAN_COMMAND_LIST.add(END_CMD);
		Environment.COSPAN_COMMAND_LIST.add(HELP_CMD);
		INNER_INTERFACE_CMD = InnerInterfaceCommand.getCommand();
		Environment.COSPAN_COMMAND_LIST.add(INNER_INTERFACE_CMD);
		NODE_CMD = NodeCommand.getCommand();
		Environment.COSPAN_COMMAND_LIST.add(NODE_CMD);
		OUTER_INTERFACE_CMD = OuterInterfaceCommand.getCommand();
		Environment.COSPAN_COMMAND_LIST.add(OUTER_INTERFACE_CMD);
		
		//Add decomposition definition mode commands
		ADD_CMD = AddCommand.getCommand();
		Environment.DECOMP_COMMAND_LIST.add(ADD_CMD);
		Environment.DECOMP_COMMAND_LIST.add(END_CMD);
		Environment.DECOMP_COMMAND_LIST.add(HELP_CMD);

		//Add graph definition mode commands
		Environment.GRAPH_COMMAND_LIST.add(END_CMD);
		Environment.GRAPH_COMMAND_LIST.add(EDGE_CMD);
		Environment.GRAPH_COMMAND_LIST.add(HELP_CMD);
		Environment.GRAPH_COMMAND_LIST.add(NODE_CMD);
		
		//Add formula definition mode commands
		Environment.FORMULA_COMMAND_LIST.add(END_CMD);
		Environment.FORMULA_COMMAND_LIST.add(HELP_CMD);

		commandList = new LinkedList<>();
		commandList.addAll(Environment.COSPAN_COMMAND_LIST);
		commandList.addAll(Environment.DEFAULT_COMMAND_LIST);
		commandList.addAll(Environment.DECOMP_COMMAND_LIST);
		commandList.addAll(Environment.FORMULA_COMMAND_LIST);
		commandList.addAll(Environment.GRAPH_COMMAND_LIST);
		
		logger = Logger.getLogger(getClass().getName());
	}    

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public void run() {
		String line = null;

		if(isInteractive) {
			//Show informations about raven
			out.println(Ansi.ansi().render("@|white "+RavenMain.getASCIIArt()+"|@"));
			out.println(Ansi.ansi().render("@|green "+RavenMain.getInfo()+"|@"));
			
			out.println(messages.getString(
					ConsoleMessages.GET_HELP_DESCRIPTION_KEY, HELP_CMD.getKey()));
			progressThread = DEBUG_SHOW_PROGRESS_BAR ? new ProgressThread(40) : null;
		}

		try {
			do {
				if(isInteractive) {
					out.print(Environment.prompt);
					out.print(Ansi.ansi().fg(Color.CYAN));
				}

				line = in.readLine();
				if(line!=null && !line.matches("\\p{Space}*")) {
					if(isInteractive) {
						out.print(Ansi.ansi().reset());
					}
					parseLine(line.trim());
				}
			} while(line!=null);
		} catch (IOException e) {
			logger.throwing(getClass().getName(), "run", e);
			RavenMain.exit(RavenMain.EXIT_CODE_PROPERTIES_IO_EXCEPTION);
		} catch (OutOfMemoryError e) {
			logger.throwing(getClass().getName(), "run", e);
			RavenMain.exit(RavenMain.EXIT_CODE_OUT_OF_MEMORY);
		} catch (Exception e) {
			e.printStackTrace();
			logger.throwing(getClass().getName(), "run", e);
			RavenMain.exit(RavenMain.EXIT_CODE_UNEXPECTED_ERROR);
		}

	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Parses a command line and executes the specified command (if possible).
	 * @param line the command line to be executed
	 */
	private void parseLine(String line) {
		if(!line.matches(COMMAND_PATTERN)) {
			out.println("Syntactical error!");
			return;
		}

		String[] temp = line.split(" ", 2);
		String cmdString = temp[0];
		String paramString = temp.length==2 ? temp[1]: "";

		//Insert values for properties
		for(String property: RavenMain.getPropertyNames()) {
			paramString = paramString.replaceAll(EVALUATION_TOKEN+property, 
					RavenMain.getProperty(property));
		}
		//Check for unresolved variables (i.e. "subwords" starting with EVALUATION_TOKEN)
		if(paramString.contains(EVALUATION_TOKEN)) {
			int index = paramString.indexOf(EVALUATION_TOKEN);
			String variable = paramString.substring(index+1);
			variable = variable.replaceAll("[^"+VARIABLE_PATTERN+"].*", "");
			
			out.println(messages.getString(ConsoleMessages.UNKNOWN_VARIABLE_DESCRIPTION_KEY, variable));
			return;
		}
		
		
		try {
			switch (Environment.type) {
			case BATCH:
			case DEFAULT:
				for(Command cmd: Environment.DEFAULT_COMMAND_LIST) {
					if(cmdString.matches(cmd.getKey())) {
						cmd.execute(paramString);
						return;
					}
				}

				out.println(messages.getString(
						ConsoleMessages.UNKNOWN_COMMAND_DESCRIPTION_KEY, cmdString));
				break;

			case COSPAN:
				for(Command cmd: Environment.COSPAN_COMMAND_LIST) {
					if(cmdString.equals(cmd.getKey())) {
						try {
							cmd.execute(paramString);
						} catch (IllegalStateException e) {
							out.println(e.getMessage());
						}
						return;
					}
				}

				out.println(messages.getString(
						ConsoleMessages.UNKNOWN_COMMAND_DESCRIPTION_KEY, cmdString));
				break;
				
			case DECOMPOSITION:
				for(Command cmd: Environment.DECOMP_COMMAND_LIST) {
					if(cmdString.equals(cmd.getKey())) {
						try {
							cmd.execute(paramString);
						} catch (IllegalStateException e) {
							out.println(e.getMessage());
						}
						return;
					}
				}

				out.println(messages.getString(
						ConsoleMessages.UNKNOWN_COMMAND_DESCRIPTION_KEY, cmdString));
				break;

			case FORMULA:
				for(Command cmd: Environment.FORMULA_COMMAND_LIST) {
					if(cmdString.equals(cmd.getKey())) {
						try {
							cmd.execute(paramString);
						} catch (IllegalStateException e) {
							out.println(e.getMessage());
						}
						return;
					}
				}

				out.println(messages.getString(
						ConsoleMessages.UNKNOWN_COMMAND_DESCRIPTION_KEY, cmdString));
				break;
				
			case GRAPH:
				for(Command cmd: Environment.GRAPH_COMMAND_LIST) {
					if(cmdString.equals(cmd.getKey())) {
						try {
							cmd.execute(paramString);
						} catch (IllegalStateException e) {
							out.println(e.getMessage());
						}
						return;
					}
				}

				out.println(messages.getString(
						ConsoleMessages.UNKNOWN_COMMAND_DESCRIPTION_KEY, cmdString));
				break;

			default:
				out.println(messages.getString(
						ConsoleMessages.UNKNOWN_COMMAND_DESCRIPTION_KEY, cmdString));
				break;
			}
		} catch (RuntimeException e) {
			if(isInteractive) {
				out.println(Ansi.ansi().render("@|red "+e.getMessage()+"|@"));
			}
			logger.throwing(getClass().getName(), "parseLine", e);
		}
		finally {
			if(progressThread!=null) {
				progressThread.interrupt();
			}
		}

	}

}
