/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.cli;

import java.io.File;
import java.util.List;

import de.uni_due.inf.ti.raven.actions.AddToRepositoryAction;
import de.uni_due.inf.ti.raven.actions.ComputationAction;
import de.uni_due.inf.ti.raven.actions.CospanDecomposeAction;
import de.uni_due.inf.ti.raven.actions.LibTWAllStartLexBFSAction;
import de.uni_due.inf.ti.raven.actions.LibTWGreedyDegreeAction;
import de.uni_due.inf.ti.raven.actions.LibTWGreedyFillInAction;
import de.uni_due.inf.ti.raven.actions.LibTWLexBFSAction;
import de.uni_due.inf.ti.raven.data.Cospan;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Decomposition;
import de.uni_due.inf.ti.raven.data.Decomposition.Type;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.PathDecomposition;
import de.uni_due.inf.ti.raven.data.TreeDecomposition;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * Handles decomposition of graphs.
 * @author Christoph Blume
 * @version $Revision: 1.21 $
 *
 */
public final class DecomposeCommand extends Command {

	private enum InputType {
		/** */
		cospanInput,
		
		/** */
		graphInput
	}
	
	private static final File DECOMPOSE_DOCUMENTATION_FILE = new File(
			RavenMain.DOCUMENTATION_FOLDER + "DecomposeCommandDocumentation.txt");
	
	private static final String COSPAN_TOKEN = Console.COSPAN_TOKEN;
	private static final String GRAPH_TOKEN = Console.GRAPH_TOKEN;
	private static final String IS_TOKEN = Console.IS_TOKEN;
	private static final String NAME_PATTERN = Console.NAME_PATTERN;
	private static final String ALL_START_LEX_BFS_TOKEN = Console.ALL_START_LEX_BFS;
	private static final String GREEDY_DEGREE_TOKEN = Console.GREEDY_DEGREE_TOKEN;
	private static final String GREEDY_FILL_IN_TOKEN = Console.GREEDY_FILL_IN_TOKEN;
	private static final String LEX_BFS_TOKEN = Console.LEX_BFS;
	private static final String SEP_TOKEN = Console.SEP_TOKEN;

	private static final String COSPAN_DECOMPOSITION_TOKEN = Console.COSPAN_DECOMPOSITION_TOKEN;
	private static final String PATH_DECOMPOSITION_TOKEN = Console.PATH_DECOMPOSITION_TOKEN;
	private static final String TREE_DECOMPOSITION_TOKEN = Console.TREE_DECOMPOSITION_TOKEN;
	private static final String DECOMPOSITION_PATTERN = COSPAN_DECOMPOSITION_TOKEN +"|"+ 
												PATH_DECOMPOSITION_TOKEN + "|" + TREE_DECOMPOSITION_TOKEN;
	private static final String ALGORITHM_PATTERN = ALL_START_LEX_BFS_TOKEN + "|" + GREEDY_FILL_IN_TOKEN + "|" + 
												GREEDY_DEGREE_TOKEN + "|" + LEX_BFS_TOKEN;
	private static final String DEFAULT_DECOMPOSITION = COSPAN_DECOMPOSITION_TOKEN;
	private static final String DEFAULT_ALGORITHM = GREEDY_FILL_IN_TOKEN;

	private static final String DECOMPOSE_KEY = Console.DECOMPOSE_TOKEN;
	private static final String DEFAULT_DECOMPOSE_COSPAN_PARAMETER = "("+NAME_PATTERN+")" + IS_TOKEN + 
			COSPAN_TOKEN + "\\(" + "("+NAME_PATTERN+")" + "\\)";
	private static final String DEFAULT_DECOMPOSE_GRAPH_PARAMETER = "("+NAME_PATTERN+")" + IS_TOKEN + 
			GRAPH_TOKEN + "\\(" + "("+NAME_PATTERN+")" + "\\)";
	private static final String DEFAULT_DECOMPOSITION_DECOMPOSE_GRAPH_PARAMETER = "("+NAME_PATTERN+")" + IS_TOKEN + 
			GRAPH_TOKEN + "\\(" + "("+NAME_PATTERN+")" +SEP_TOKEN+ "("+ALGORITHM_PATTERN+")\\)";
	private static final String DEFAULT_ALGORITHM_DECOMPOSE_GRAPH_PARAMETER = "("+NAME_PATTERN+")" + IS_TOKEN + 
			GRAPH_TOKEN + "\\(" + "("+NAME_PATTERN+")" +SEP_TOKEN+ "("+DECOMPOSITION_PATTERN+")\\)";
	private static final String DECOMPOSE_GRAPH_PARAMETER = "("+NAME_PATTERN+")" + IS_TOKEN + 
			GRAPH_TOKEN + "\\(" + "("+NAME_PATTERN+")" +SEP_TOKEN+ "("+DECOMPOSITION_PATTERN+")" +SEP_TOKEN+ "("+ALGORITHM_PATTERN+")\\)";

	private static DecomposeCommand command;

	
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the <tt>INVARIANT</tt>-command.
	 * @return the <tt>INVARIANT</tt>-command
	 */
	protected static DecomposeCommand getCommand() {
		if(command==null) {
			command = new DecomposeCommand();
		}

		return command;		
	}


	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates the <tt>DECOMPOSE</tt>-command.
	 */
	private DecomposeCommand() {
		super(DECOMPOSE_KEY, ConsoleMessages.INVARIANT_DESCRIPTION_KEY, 
				DECOMPOSE_DOCUMENTATION_FILE);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public boolean matchesParameter(String line) {
		return line.matches(DEFAULT_DECOMPOSE_COSPAN_PARAMETER) ||
				line.matches(DEFAULT_DECOMPOSE_GRAPH_PARAMETER) || 
				line.matches(DEFAULT_ALGORITHM_DECOMPOSE_GRAPH_PARAMETER) || 
				line.matches(DEFAULT_DECOMPOSITION_DECOMPOSE_GRAPH_PARAMETER) || 
				line.matches(DECOMPOSE_GRAPH_PARAMETER);
	}

	@Override
	public void execute(String parameter) {
		if(!matchesParameter(parameter)) {
			throw new IllegalArgumentException(Console.messages.getString(
							ConsoleMessages.MISSING_ILLEGAL_ARGUMENT_DESCRIPTION_KEY, DECOMPOSE_KEY));
		}

		InputType type = null;
		List<String> list;
		String decompositionName = null;
		String cospanName = null;
		String graphName = null;
		Decomposition.Type decompositionType = null;
		String algorithmName = null;

		if(parameter.matches(DEFAULT_DECOMPOSE_COSPAN_PARAMETER)) {
			type = InputType.cospanInput;
			//Scan and extract parameter from input string
			list = getProperties(parameter, DEFAULT_DECOMPOSE_COSPAN_PARAMETER);
			decompositionName = list.get(0);
			cospanName = list.get(1);
		}
		else if(parameter.matches(DEFAULT_DECOMPOSE_GRAPH_PARAMETER)) {
			type = InputType.graphInput;
			//Scan and extract parameter from input string
			list = getProperties(parameter, DEFAULT_DECOMPOSE_GRAPH_PARAMETER);
			decompositionName = list.get(0);
			graphName = list.get(1);
			decompositionType = Type.valueOf(DEFAULT_DECOMPOSITION);
			algorithmName = DEFAULT_ALGORITHM;
		}
		else if(parameter.matches(DEFAULT_ALGORITHM_DECOMPOSE_GRAPH_PARAMETER)) {
			type = InputType.graphInput;
			//Scan and extract parameter from input string
			list = getProperties(parameter, DEFAULT_ALGORITHM_DECOMPOSE_GRAPH_PARAMETER);
			decompositionName = list.get(0);
			graphName = list.get(1);
			decompositionType = Type.valueOf(list.get(2));
			algorithmName = DEFAULT_ALGORITHM;
		}
		else if(parameter.matches(DEFAULT_DECOMPOSITION_DECOMPOSE_GRAPH_PARAMETER)) {
			type = InputType.graphInput;
			//Scan and extract parameter from input string
			list = getProperties(parameter, DEFAULT_DECOMPOSITION_DECOMPOSE_GRAPH_PARAMETER);
			decompositionName = list.get(0);
			graphName = list.get(1);
			decompositionType = Type.valueOf(DEFAULT_DECOMPOSITION);
			algorithmName = list.get(2);
		}
		else if(parameter.matches(DECOMPOSE_GRAPH_PARAMETER)) {
			type = InputType.graphInput;
			//Scan and extract parameter from input string
			list = getProperties(parameter, DECOMPOSE_GRAPH_PARAMETER);
			decompositionName = list.get(0);
			graphName = list.get(1);
			decompositionType = Type.valueOf(list.get(2));
			algorithmName = list.get(3);
		}

		if(type==InputType.cospanInput) {
			//Create automaton
			if(!Console.repo.isCospan(cospanName)) {
				Console.out.println(Console.messages.getString(ConsoleMessages.COSPAN_NOT_FOUND_KEY, cospanName));
				return;
			}
			
			Cospan cospan = (Cospan) Console.repo.get(cospanName);
			ComputationAction<CospanDecomposition> action = new CospanDecomposeAction(cospan);
			if(Console.progressThread!=null) {
				action.addComputationListener(Console.progressThread);
			}
			Console.out.println(Console.messages.getString(
					ConsoleMessages.COMPUTING_COSPAN_DECOMPOSITION_KEY, cospanName));
			action.execute();
			CospanDecomposition decomposition = action.getResult();
			new AddToRepositoryAction(decompositionName, decomposition).execute();
			Console.out.println(Console.messages.getString(
					ConsoleMessages.DECOMPOSITION_ADDED_KEY, decompositionName));
			
			//Print elapsed runtime (if time mode is activated)
			printRuntime(action.getElapsedTime());
		}
		else if(type==InputType.graphInput) {
			//Create automaton
			if(!Console.repo.isGraph(graphName)) {
				Console.out.println(Console.messages.getString(ConsoleMessages.GRAPH_NOT_FOUND_KEY, graphName));
				return;
			}
			Graph graph = (Graph) Console.repo.get(graphName);
			
			ComputationAction<TreeDecomposition> action = null;
			if(algorithmName.equals(ALL_START_LEX_BFS_TOKEN)) {
				//Compute the decomposition using the All Start Lex BFS decomposition algorithm
				action = new LibTWAllStartLexBFSAction(graph);
			}
			if(algorithmName.equals(GREEDY_DEGREE_TOKEN)) {
				//Compute the decomposition using the Greedy Degree decomposition algorithm
				action = new LibTWGreedyDegreeAction(graph);
			}
			else if(algorithmName.equals(GREEDY_FILL_IN_TOKEN)) {
				//Compute the decomposition using the Greedy Fill-in decomposition algorithm
				action = new LibTWGreedyFillInAction(graph);
				
			}
			else if(algorithmName.equals(LEX_BFS_TOKEN)) {
				//Compute the decomposition using the Lex BFS decomposition algorithm
				action = new LibTWLexBFSAction(graph);
				
			}
			
			if(Console.progressThread!=null) {
				action.addComputationListener(Console.progressThread);
			}
			Console.out.println(Console.messages.getString(
					ConsoleMessages.COMPUTING_GRAPH_DECOMPOSITION_KEY, graphName));
			action.execute();
			
			Decomposition<?> decomposition = null;
			String resultString = null;
			switch (decompositionType) {
			case treeDecomposition:
				decomposition = action.getResult();
				resultString = Console.messages.getString(
						ConsoleMessages.TREE_DECOMPOSITION_ADDED_KEY, decompositionName);
				
				break;
				
			case pathDecomposition:
				decomposition = PathDecomposition.getPathDecomposition(action.getResult());
				resultString = Console.messages.getString(
						ConsoleMessages.PATH_DECOMPOSITION_ADDED_KEY, decompositionName);
				break;
				
			case cospanDecomposition:
				decomposition = CospanDecomposition.getCospanDecomposition(
						PathDecomposition.getPathDecomposition(action.getResult()));
				resultString = Console.messages.getString(
						ConsoleMessages.DECOMPOSITION_ADDED_KEY, decompositionName);
				break;
				
			default:
				break;
			}

			new AddToRepositoryAction(decompositionName, decomposition).execute();
			Console.out.println(resultString);
			
			//Print elapsed runtime (if time mode is activated)
			printRuntime(action.getElapsedTime());
		}
		
	}

}
