/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.cli;

import java.io.File;
import java.util.List;

import de.uni_due.inf.ti.raven.actions.CreateRandomGraphAction;
import de.uni_due.inf.ti.raven.data.Cospan;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.Node;
import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * Console command to handle the definition of graphs or decompositions.
 * @author Christoph Blume
 * @version $Revision: 1.34 $
 *
 * @see AddCommand
 * @see EdgeCommand
 * @see EndCommand
 * @see NodeCommand
 */
public final class DefineCommand extends Command {

	private static final File DEFINE_DOCUMENTATION_FILE = new File(
			RavenMain.DOCUMENTATION_FOLDER + "DefineCommandDocumentation.txt");
	
	private static final String IS_TOKEN = Console.IS_TOKEN;
	private static final String COSPAN_TOKEN = Console.COSPAN_TOKEN;
	private static final String DECOMPOSITION_TOKEN = Console.DECOMPOSITION_TOKEN;
	private static final String GRAPH_TOKEN = Console.GRAPH_TOKEN;
	private static final String FORMULA_TOKEN = Console.FORMULA_TOKEN;
	private static final String RANDOMGRAPH_TOKEN = Console.RANDOMGRAPH_TOKEN;
	private static final String SEP_TOKEN = Console.SEP_TOKEN;
	
	private static final String NAME_PATTERN = Console.NAME_PATTERN;
	private static final String NUMBER_PATTERN = Console.NUMBER_PATTERN;
	
	private static final String DEFINE_KEY = Console.DEFINE_TOKEN;
	private static final String DEFINE_COSPAN_PARAMETER =  "("+NAME_PATTERN+")" + IS_TOKEN + COSPAN_TOKEN + 
			"\\(("+NUMBER_PATTERN+")\\)";
	private static final String DEFINE_DECOMP_PARAMETER =  "("+NAME_PATTERN+")" + IS_TOKEN + DECOMPOSITION_TOKEN;
	private static final String DEFINE_DEFAULT_DECOMP_BY_GRAPH_PARAMETER =  "("+NAME_PATTERN+")" + IS_TOKEN + DECOMPOSITION_TOKEN
			+ "\\(("+NAME_PATTERN+")\\)";
	private static final String DEFINE_FORMULA_PARAMETER =  "("+NAME_PATTERN+")" + IS_TOKEN + FORMULA_TOKEN;
	private static final String DEFINE_GRAPH_PARAMETER = "("+NAME_PATTERN+")" + IS_TOKEN + GRAPH_TOKEN;
	private static final String DEFINE_GRAPH_BY_DECOMP_PARAMETER = "("+NAME_PATTERN+")" + IS_TOKEN + GRAPH_TOKEN 
			+ "\\(("+NAME_PATTERN+")\\)";
	private static final String DEFINE_RANDOM_GRAPH_DEFAULT_PARAMETER = "("+NAME_PATTERN+")" + IS_TOKEN + RANDOMGRAPH_TOKEN + "\\(("+NUMBER_PATTERN+")"+ SEP_TOKEN + "("+NUMBER_PATTERN+")\\)";
	private static final String DEFINE_RANDOM_GRAPH_LIB_TW_PARAMETER = "("+NAME_PATTERN+")" + IS_TOKEN + RANDOMGRAPH_TOKEN + "\\(("+NUMBER_PATTERN+")\\)";
	
	private static DefineCommand command;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the <tt>DEFINE</tt>-command.
	 * @return the <tt>DEFINE</tt>-command
	 */
	protected static DefineCommand getCommand() {
		if(command==null) {
			command = new DefineCommand();
		}

		return command;		
	}
	
	/**
	 * Returns {@code true} if and only if the parameter {@code parameter} indicate the 
	 * DEFINE-*COSPAN* option.
	 * @param parameter the parameter passed to this action
	 * @return {@code true} if the parameter indicate the DEFINE-*COSPAN* option, {@code false} 
	 * otherwise
	 */
	private static boolean isCospanDefinition(String parameter) {
		return parameter.matches(DEFINE_COSPAN_PARAMETER);
	}
	
	/**
	 * Returns {@code true} if and only if the parameter {@code parameter} indicate the 
	 * DEFINE-*DECOMP* option.
	 * @param parameter the parameter passed to this action
	 * @return {@code true} if the parameter indicate the DEFINE-*DECOMP* option, {@code false} 
	 * otherwise
	 */
	private static boolean isDecompositionDefinition(String parameter) {
		return parameter.matches(DEFINE_DECOMP_PARAMETER) || 
				parameter.matches(DEFINE_DEFAULT_DECOMP_BY_GRAPH_PARAMETER);
	}
	
	/**
	 * Returns {@code true} if and only if the parameter {@code parameter} indicate the 
	 * DEFINE-*FORMULA* option.
	 * @param parameter the parameter passed to this action
	 * @return {@code true} if the parameter indicate the DEFINE-*FORMULA* option, {@code false} 
	 * otherwise
	 */
	private static boolean isFormulaDefinition(String parameter) {
		return parameter.matches(DEFINE_FORMULA_PARAMETER);
	}
	
	/**
	 * Returns {@code true} if and only if the parameter {@code parameter} indicate the 
	 * DEFINE-*GRAPH* option.
	 * @param parameter the parameter passed to this action
	 * @return {@code true} if the parameter indicate the DEFINE-*GRAPH* option, {@code false} 
	 * otherwise
	 */
	private static boolean isGraphDefinition(String parameter) {
		return parameter.matches(DEFINE_GRAPH_PARAMETER) ||
				parameter.matches(DEFINE_GRAPH_BY_DECOMP_PARAMETER);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates the <tt>DEFINE</tt>-command.
	 */
	private DefineCommand() {
		super(DEFINE_KEY, ConsoleMessages.DEFINE_DESCRIPTION_KEY, 
				DEFINE_DOCUMENTATION_FILE);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public boolean matchesParameter(String line) {
		return isCospanDefinition(line) ||
				isDecompositionDefinition(line) ||
				isFormulaDefinition(line) ||
				isGraphDefinition(line) ||
				line.matches(DEFINE_RANDOM_GRAPH_DEFAULT_PARAMETER) ||
				line.matches(DEFINE_RANDOM_GRAPH_LIB_TW_PARAMETER);
	}

	@Override
	public void execute(String parameter) {
		if(!matchesParameter(parameter)) {
			throw new IllegalArgumentException(Console.messages.getString(
					ConsoleMessages.MISSING_ILLEGAL_ARGUMENT_DESCRIPTION_KEY, DEFINE_KEY));
		}
		
		if(isCospanDefinition(parameter)) {
			List<String> list = getProperties(parameter, DEFINE_COSPAN_PARAMETER);
			int innerInterfaceSize = Integer.parseInt(list.get(1));
			Console.Environment.setType(Console.Environment.Type.COSPAN);
			Console.Environment.cospan = new Cospan(innerInterfaceSize);
			Console.Environment.name = list.get(0);
			
			List<Node> innerInterface = Console.Environment.cospan.getInnerInterface();
			Console.out.println(Console.messages.getString(
					ConsoleMessages.COSPAN_DEFINITION_MODE_STARTED_KEY, 
					String.valueOf(innerInterfaceSize),
					innerInterface.get(0).getName(),
					innerInterface.get(innerInterface.size()-1).getName(),
					EndCommand.getCommand().getKey()));
			Console.out.println(Console.messages.getString(
					ConsoleMessages.EXIT_DEFINITION_MODE_DESCRIPTION_KEY, EndCommand.getCommand().getKey()));
		}
		else if(isDecompositionDefinition(parameter)) {
			List<String> list = getProperties(parameter, DEFINE_DECOMP_PARAMETER);
			Console.Environment.setType(Console.Environment.Type.DECOMPOSITION);
			Console.Environment.decomposition = new CospanDecomposition();
			Console.Environment.name = list.get(0);
			Console.out.println(Console.messages.getString(
					ConsoleMessages.EXIT_DEFINITION_MODE_DESCRIPTION_KEY, EndCommand.getCommand().getKey()));
		}
		else if(isFormulaDefinition(parameter)) {
			List<String> list = getProperties(parameter, DEFINE_FORMULA_PARAMETER);
			Console.Environment.setType(Console.Environment.Type.FORMULA);
			Console.Environment.name = list.get(0);
			Console.out.println(Console.messages.getString(
					ConsoleMessages.EXIT_DEFINITION_MODE_DESCRIPTION_KEY, EndCommand.getCommand().getKey()));
		}
		else if(isGraphDefinition(parameter)) {
			if(parameter.matches(DEFINE_GRAPH_PARAMETER)) {
				List<String> list = getProperties(parameter, DEFINE_GRAPH_PARAMETER);
				Console.Environment.setType(Console.Environment.Type.GRAPH);
				Console.Environment.graph = new Graph(list.get(0));
				Console.Environment.name = list.get(0);
				Console.out.println(Console.messages.getString(
						ConsoleMessages.EXIT_DEFINITION_MODE_DESCRIPTION_KEY, EndCommand.getCommand().getKey()));
			}
			else if(parameter.matches(DEFINE_GRAPH_BY_DECOMP_PARAMETER)) {
				List<String> list = getProperties(parameter, DEFINE_GRAPH_BY_DECOMP_PARAMETER);
				String graphName = list.get(0);
				String decompositionName = list.get(1);
				Repository repo = Console.repo;
				
				if(!repo.isDecomposition(decompositionName)) {
					throw new IllegalArgumentException(Console.messages.getString(
							ConsoleMessages.DECOMPOSITION_NOT_FOUND_KEY, decompositionName));
				}
				
				//Create a new graph and add it to the repository
				CospanDecomposition decomposition = (CospanDecomposition) repo.get(decompositionName);
				boolean result = repo.put(graphName, decomposition.getGraph());			
				if(result){
					Console.out.println(Console.messages.getString(
							ConsoleMessages.GRAPH_ADDED_KEY, graphName));
				}
				else {
					Console.out.println(Console.messages.getString(
							ConsoleMessages.GRAPH_ALREADY_EXISTS_EXCEPTION, graphName));					
				}
			}
		}
		else if(parameter.matches(DEFINE_RANDOM_GRAPH_DEFAULT_PARAMETER)){
			List<String> list = getProperties(parameter, DEFINE_RANDOM_GRAPH_DEFAULT_PARAMETER);
			String graphName = list.get(0);
			int numberOfNodes = Integer.parseInt(list.get(1));
			int density = Integer.parseInt(list.get(2));
			Repository repo = Console.repo;
			
			if(density<0) {
				Console.out.println(Console.messages.getString(
						ConsoleMessages.DENSITY_MUST_NOT_BE_LESS_THAN_ZERO_KEY));
				return;
			}
			else if(density>100) {
				Console.out.println(Console.messages.getString(
						ConsoleMessages.DENSITY_MUST_NOT_BE_GREATER_THAN_HUNDRED_KEY));
				return;
			}
			
			CreateRandomGraphAction action = new CreateRandomGraphAction(graphName, numberOfNodes, density / 100f);
			action.execute();

			boolean result = repo.put(graphName, action.getResult());			
			if(result){
				Console.out.println(Console.messages.getString(
						ConsoleMessages.GRAPH_ADDED_KEY, graphName));
			}
			else {
				Console.out.println(Console.messages.getString(
						ConsoleMessages.GRAPH_ALREADY_EXISTS_EXCEPTION, graphName));					
			}
		}
		else if(parameter.matches(DEFINE_RANDOM_GRAPH_LIB_TW_PARAMETER)){
			List<String> list = getProperties(parameter, DEFINE_RANDOM_GRAPH_LIB_TW_PARAMETER);
			String graphName = list.get(0);
			int edgeCount = Integer.parseInt(list.get(1));
			Repository repo = Console.repo;
			
			CreateRandomGraphAction action = new CreateRandomGraphAction(graphName, edgeCount);
			action.execute();

			boolean result = repo.put(graphName, action.getResult());			
			if(result){
				Console.out.println(Console.messages.getString(
						ConsoleMessages.GRAPH_ADDED_KEY, graphName));
			}
			else {
				Console.out.println(Console.messages.getString(
						ConsoleMessages.GRAPH_ALREADY_EXISTS_EXCEPTION, graphName));					
			}
		}
		
	}

}
