/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.cli;

import java.io.File;
import java.util.LinkedList;
import java.util.List;

import de.uni_due.inf.ti.raven.cli.Console.Environment;
import de.uni_due.inf.ti.raven.data.Node;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * Console command to handle the addition of edges in cospan/graph mode.
 * @author Christoph Blume
 * @version $Revision: 1.20 $
 *
 * @see DefineCommand
 * @see NodeCommand
 * @see Console.Environment.Type
 */
public final class EdgeCommand extends Command {

	private static final File EDGE_DOCUMENTATION_FILE = new File(
			RavenMain.DOCUMENTATION_FOLDER + "EdgeCommandDocumentation.txt");
	
	private static final String IS_TOKEN = Console.IS_TOKEN;
	private static final String SEP_TOKEN = Console.SEP_TOKEN;
	
	private static final String LABEL_PATTERN = Console.LABEL_PATTERN;
	private static final String NAME_PATTERN = Console.NAME_PATTERN;
	private static final String NODE_LIST_PATTERN = NAME_PATTERN + "(" + SEP_TOKEN + NAME_PATTERN + ")*";
	private static final String EDGE_PATTERN = "("+NAME_PATTERN+")" + IS_TOKEN + "("+LABEL_PATTERN+")" +
			"\\((\\p{Graph}*)\\)";
	
	private static final String EDGE_KEY = Console.EDGE_TOKEN;
	
	private static EdgeCommand command;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the <tt>EDGE</tt>-command.
	 * @return the <tt>EDGE</tt>-command
	 */
	protected static EdgeCommand getCommand() {
		if(command==null) {
			command = new EdgeCommand();
		}

		return command;		
	}
	
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates the <tt>EDGE</tt>-command.
	 */
	private EdgeCommand() {
		super(EDGE_KEY, ConsoleMessages.EDGE_DESCRIPTION_KEY, 
				EDGE_DOCUMENTATION_FILE);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public boolean matchesParameter(String line) {
		return line.matches(EDGE_PATTERN);
	}

	@Override
	public void execute(String parameter) {
		if(!matchesParameter(parameter)) {
			throw new IllegalArgumentException(Console.messages.getString(
					ConsoleMessages.MISSING_ILLEGAL_ARGUMENT_DESCRIPTION_KEY, EDGE_KEY));
		}

		List<String> list = getProperties(parameter, EDGE_PATTERN);
		List<Node> incidentNodes = new LinkedList<>();
		String edgeName = list.get(0);
		String label = list.get(1);
		String nodeList = list.get(2);

		if(!nodeList.matches(NODE_LIST_PATTERN)) {
			throw new IllegalArgumentException(Console.messages.getString(
					ConsoleMessages.MISSING_ILLEGAL_ARGUMENT_DESCRIPTION_KEY, EDGE_KEY));
		}
		
		if(edgeName.isEmpty()) {
			Console.out.println(messages.getString(
					ConsoleMessages.NAME_MUST_NOT_BE_EMPTY_KEY));
		}
		else if(label.isEmpty()) {
			Console.out.println(messages.getString(
					ConsoleMessages.LABEL_MUST_NOT_BE_EMPTY_KEY));
		}
		else {
			for(String nodeName: nodeList.split(SEP_TOKEN)) {
				if(nodeName.isEmpty()) {
					Console.out.println(messages.getString(
							ConsoleMessages.NAME_MUST_NOT_BE_EMPTY_KEY));
					return;
				}

				if(Environment.getType()==Environment.Type.COSPAN) {
					if(!Environment.cospan.hasNode(nodeName)) {
						Console.out.println(messages.getString(
								ConsoleMessages.NODE_NOT_EXISTING_KEY, nodeName));
						return;
					}
					incidentNodes.add(Environment.cospan.getNode(nodeName));
				}
				else if(Environment.getType()==Environment.Type.GRAPH) {
					if(!Environment.graph.hasNode(nodeName)) {
						Console.out.println(messages.getString(
								ConsoleMessages.NODE_NOT_EXISTING_KEY, nodeName));
						return;
					}

					incidentNodes.add(Environment.graph.getNode(nodeName));
				}
			}
			
			if(Environment.getType()==Environment.Type.COSPAN) {
				Environment.cospan.addEdge(edgeName, label, incidentNodes);
			}
			else if(Environment.getType()==Environment.Type.GRAPH) {
				Environment.graph.addEdge(edgeName, label, incidentNodes);
			}
		}

	}

}
