/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.cli;

import java.io.File;
import java.util.List;

import de.uni_due.inf.ti.raven.actions.EmptinessCheckAction;
import de.uni_due.inf.ti.raven.algorithms.Algorithm.SearchType;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * Console command to handle the execution of the universality check algorithms.
 * 
 * @author Christoph Blume
 * @version $Revision: 1.13 $
 * 
 */
public final class EmptinessCommand extends Command {

	private static final File EMPTINESS_DOCUMENTATION_FILE = new File(
			RavenMain.DOCUMENTATION_FOLDER + "EmptinessCommandDocumentation.txt");

	private static final String NAME_PATTERN = Console.NAME_PATTERN;
	private static final String BACKWARD_TOKEN = Console.BACKWARD_TOKEN;
	private static final String FORWARD_TOKEN = Console.FORWARD_TOKEN;
	private static final String SEP_TOKEN = Console.SEP_TOKEN;

	private static final String EMPTINESS_KEY = Console.EMPTINESS_TOKEN;
	private static final String EMPTINESS_PARAMETER = "\\((" + FORWARD_TOKEN
			+ "|" + BACKWARD_TOKEN + ")" + SEP_TOKEN + "(" + NAME_PATTERN
			+ ")\\)";

	private static EmptinessCommand command;

	/*------------------------------------------------------------------------------------------*/
	/* 		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the <tt>UNIVERSALITY</tt>-command.
	 * 
	 * @return the <tt>UNIVERSALITY</tt>-command
	 */
	protected static EmptinessCommand getCommand() {
		if (command == null) {
			command = new EmptinessCommand();
		}

		return command;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates the <tt>UNIVERSALITY</tt>-command.
	 */
	private EmptinessCommand() {
		super(EMPTINESS_KEY, ConsoleMessages.UNIVERSALITY_DESCRIPTION_KEY,
				EMPTINESS_DOCUMENTATION_FILE);
	}

	/*------------------------------------------------------------------------------------------*/
	/* 		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public boolean matchesParameter(String line) {
		return line.matches(EMPTINESS_PARAMETER);
	}

	@Override
	public void execute(String parameter) {
		if (!matchesParameter(parameter)) {
			throw new IllegalArgumentException(
					Console.messages.getString(
							ConsoleMessages.MISSING_ILLEGAL_ARGUMENT_DESCRIPTION_KEY,
							EMPTINESS_KEY));
		}

		List<String> list;
		String automatonName = null;
		SearchType type = null;

		list = getProperties(parameter, EMPTINESS_PARAMETER);
		// Scan and extract parameter from input string
		type = SearchType.valueOf(list.get(0));
		automatonName = list.get(1);

		if (!Console.repo.isAutomaton(automatonName)
				&& !Console.repo.isAutomaton(automatonName)) {
			throw new IllegalArgumentException(Console.messages.getString(
					ConsoleMessages.AUTOMATON_NOT_FOUND_KEY, automatonName));
		}

		Automaton automaton = (Automaton) Console.repo.get(automatonName);

		EmptinessCheckAction action = new EmptinessCheckAction(type,
				automaton);
		if (Console.progressThread != null) {
			action.addComputationListener(Console.progressThread);
		}
		action.execute();

		String resultString;
		if (action.getResult()) {
			resultString = Console.messages.getString(
					ConsoleMessages.EMPTINESS_DOES_HOLD_KEY, automatonName);
		} else {
			resultString = Console.messages.getString(
					ConsoleMessages.EMPTINESS_DOES_NOT_HOLD_KEY, automatonName);
		}
		Console.out.println(resultString);
		
		//Print elapsed runtime (if time mode is activated)
		printRuntime(action.getElapsedTime());
	}

}
