/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.cli;

import java.io.File;

import de.uni_due.inf.ti.raven.actions.AddToRepositoryAction;
import de.uni_due.inf.ti.raven.cli.Console.Environment;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * Console command to handle the termination of decomposition or graph mode.
 * @author Christoph Blume
 * @version $Revision: 1.28 $
 * 
 * @see DefineCommand
 * @see Console.Environment.Type
 */
public final class EndCommand extends Command {

	private static final File END_DOCUMENTATION_FILE = new File(
			RavenMain.DOCUMENTATION_FOLDER + "EndCommandDocumentation.txt");
	
	private static final String END_KEY = Console.END_TOKEN;
	
	private static EndCommand command;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the <tt>END</tt>-command.
	 * @return the <tt>END</tt>-command
	 */
	protected static EndCommand getCommand() {
		if(command==null) {
			command = new EndCommand();
		}

		return command;		
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates the <tt>END</tt>-command.
	 */
	private EndCommand() {
		super(END_KEY, ConsoleMessages.END_DESCRIPTION_KEY, 
				END_DOCUMENTATION_FILE);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public boolean matchesParameter(String line) {
		return line==null || line.isEmpty();
	}

	@Override
	public void execute(String parameter) {
		if(!matchesParameter(parameter)) {
			throw new IllegalArgumentException(Console.messages.getString(
							ConsoleMessages.MISSING_ILLEGAL_ARGUMENT_DESCRIPTION_KEY, END_KEY));
		}
		
		AddToRepositoryAction action = null;
		String successMessage = null;
		String failureMessage = null;
		
		if(Environment.getType().equals(Environment.Type.COSPAN)) {
			//Create new cospan
			action = new AddToRepositoryAction(Environment.name, Environment.cospan);
			successMessage = Console.messages.getString(
					ConsoleMessages.COSPAN_ADDED_KEY, Environment.name);
			failureMessage = Console.messages.getString(
					ConsoleMessages.COSPAN_ALREADY_EXISTS_EXCEPTION_KEY);
		}
		else if(Environment.getType().equals(Environment.Type.DECOMPOSITION)) {
			//Create new decomposition
			action = new AddToRepositoryAction(Environment.name, Environment.decomposition);
			successMessage = Console.messages.getString(
					ConsoleMessages.DECOMPOSITION_ADDED_KEY, Environment.name);
			failureMessage = Console.messages.getString(
					ConsoleMessages.COSPAN_DECOMPOSITION_ALREADY_EXISTS_EXCEPTION_KEY);				
		}
		else if(Environment.getType().equals(Environment.Type.FORMULA)) {
			//Create new formula
		}
		else if(Environment.getType().equals(Environment.Type.GRAPH)) {
			//Create new graph
			action = new AddToRepositoryAction(Environment.name, Environment.graph);
			successMessage = Console.messages.getString(
					ConsoleMessages.GRAPH_ADDED_KEY, Environment.name);
			failureMessage = Console.messages.getString(
					ConsoleMessages.GRAPH_ALREADY_EXISTS_EXCEPTION);				
		}

		action.execute();
		boolean result = action.getResult();
		if(result){
			Console.out.println(successMessage);
		}
		else {
			Console.out.println(failureMessage);				
		}
		Environment.setType(Environment.getPreviousType());
		
	}

}
