/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.cli;

import java.io.File;
import java.util.List;

import de.uni_due.inf.ti.raven.cli.Console.Environment;
import de.uni_due.inf.ti.raven.io.TextTable;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * Console command to handle the presentation of additional help information.
 * @author Christoph Blume
 * @version $Revision: 1.23 $
 *
 */
public final class HelpCommand extends Command {

	private static final File HELP_DOCUMENTATION_FILE = new File(
			RavenMain.DOCUMENTATION_FOLDER + "HelpCommandDocumentation.txt");
	
	private static final String HELP_KEY = Console.HELP_TOKEN;

	private static HelpCommand command;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the <tt>HELP</tt>-command.
	 * @return the <tt>HELP</tt>-command
	 */	
	protected static HelpCommand getCommand() {
		if(command==null) {
			command = new HelpCommand();
		}

		return command;		
	}

	private static void addToTable(TextTable table, Command cmd) {
		String keyText = cmd.getKey();
		for(String string: cmd.getDescription().split("\\n")) {
			table.addContent(0, 0, keyText);
			table.addContent(0, 1, string);
			keyText = " ";
		}
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates the <tt>HELP</tt>-command.
	 */
	private HelpCommand() {
		super(HELP_KEY, ConsoleMessages.HELP_DESCRIPTION_KEY, 
				HELP_DOCUMENTATION_FILE);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public boolean matchesParameter(String line) {
		return (line==null || line.isEmpty() || line.matches("\\w+"));
	}

	@Override
	public void execute(String parameter) {
		if(!matchesParameter(parameter)) {
			throw new IllegalArgumentException(Console.messages.getString(
					ConsoleMessages.MISSING_ILLEGAL_ARGUMENT_DESCRIPTION_KEY, HELP_KEY));
		}
		
		if(parameter==null || parameter.isEmpty()) {
			Console.out.println(Console.messages.getString(
					ConsoleMessages.GET_ADDITIONAL_HELP_DESCRIPTION_KEY, HELP_KEY));

			TextTable table = new TextTable(' ', ' ', ' ');
			
			if(Environment.getType()==Environment.Type.DEFAULT) {
				for(Command cmd: Environment.DEFAULT_COMMAND_LIST) {
					addToTable(table, cmd);
				}
				
				Console.out.print(table.toString());
				
				Console.out.println();
			}
			
			if(Environment.getType()==Environment.Type.DEFAULT || 
					Environment.getType()==Environment.Type.COSPAN) {
				Console.out.println(messages.getString(
						ConsoleMessages.COSPAN_DEFINITION_MODE_COMMANDS_KEY)+":");
				
				table = new TextTable(' ', ' ', ' ');
				for(Command cmd: Environment.COSPAN_COMMAND_LIST) {
					addToTable(table, cmd);
				}
				
				Console.out.print(table.toString());
				
				Console.out.println();
			}
			
			if(Environment.getType()==Environment.Type.DEFAULT || 
					Environment.getType()==Environment.Type.DECOMPOSITION) {
				Console.out.println(messages.getString(
						ConsoleMessages.DECOMPOSITION_DEFINITION_MODE_COMMANDS_KEY)+":");
				
				table = new TextTable(' ', ' ', ' ');
				for(Command cmd: Environment.DECOMP_COMMAND_LIST) {
					addToTable(table, cmd);
				}
				
				Console.out.print(table.toString());
				
				Console.out.println();
			}
			
			if(Environment.getType()==Environment.Type.DEFAULT || 
					Environment.getType()==Environment.Type.GRAPH) {
				Console.out.println(messages.getString(
						ConsoleMessages.GRAPH_DEFINITION_MODE_COMMANDS_KEY)+":");
				
				table = new TextTable(' ', ' ', ' ');
				for(Command cmd: Environment.GRAPH_COMMAND_LIST) {
					addToTable(table, cmd);
				}
				
				Console.out.print(table.toString());
			}
		}
		else {
			List<Command> commandList = null;
			switch (Environment.getType()) {
				case COSPAN:
					commandList = Environment.COSPAN_COMMAND_LIST;
					break;
					
				case DECOMPOSITION:
					commandList = Environment.DECOMP_COMMAND_LIST;
					break;
					
				case FORMULA:
					commandList = Environment.FORMULA_COMMAND_LIST;
					break;
					
				case GRAPH:
					commandList = Environment.GRAPH_COMMAND_LIST;
					break;
					
				case BATCH:
				case DEFAULT:
					commandList = Console.commandList;
					break;
			}
			
			for(Command cmd: commandList) {
				if(cmd.getKey().equals(parameter)) {
					Console.out.println(cmd.getDocumentation());
					return;
				}
			}
			
			Console.out.println(messages.getString(
					ConsoleMessages.UNKNOWN_COMMAND_DESCRIPTION_KEY, parameter));
		}
	}
	
}
