/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.cli;

import java.io.File;
import java.util.List;

import de.uni_due.inf.ti.raven.actions.AntichainInvariantCheckingAction;
import de.uni_due.inf.ti.raven.actions.BisimulationUpToCongruenceInvariantCheckingAction;
import de.uni_due.inf.ti.raven.actions.CospanDecomposeAction;
import de.uni_due.inf.ti.raven.actions.InvariantCheckingAction;
import de.uni_due.inf.ti.raven.actions.SimulationAntichainInvariantCheckingAction;
import de.uni_due.inf.ti.raven.algorithms.Algorithm.SearchType;
import de.uni_due.inf.ti.raven.data.Cospan;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.RuleCounterExample;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * Console command to handle the execution of invariant checking algorithms.
 * @author Christoph Blume
 * @version $Revision: 1.38 $
 *
 */
public final class InvariantCommand extends Command {

	private static final File INVARIANT_DOCUMENTATION_FILE = new File(
			RavenMain.DOCUMENTATION_FOLDER + "InvariantCommandDocumentation.txt");
	
	private static final String ANTICHAIN_TOKEN = Console.ANTICHAIN_TOKEN;
	private static final String SEP_TOKEN = Console.SEP_TOKEN;
	private static final String SIMULATION_TOKEN = Console.SIMULATION_TOKEN;
	private static final String UP_TO_TOKEN = Console.UP_TO_TOKEN;
	private static final String BOOLEAN_PATTERN = Console.BOOLEAN_PATTERN;
	private static final String NAME_PATTERN = Console.NAME_PATTERN;
	private static final String SEARCH_TYPE_PATTERN = Console.SEARCH_TYPE_PATTERN;

	private static final String INVARIANT_KEY = Console.INVARIANT_TOKEN;
	private static final String INVARIANT_DEFAULT_PARAMETER = "\\(("+NAME_PATTERN+")" 
			+ SEP_TOKEN + "("+NAME_PATTERN+")" +SEP_TOKEN+ "("+NAME_PATTERN+")\\)";
	private static final String INVARIANT_ANTICHAIN_PARAMETER = ANTICHAIN_TOKEN 
			+ "\\((" + SEARCH_TYPE_PATTERN + ")" + SEP_TOKEN + "(" + NAME_PATTERN + ")" 
			+ SEP_TOKEN + "(" + NAME_PATTERN + ")" + SEP_TOKEN + "(" + NAME_PATTERN + ")" 
			+ SEP_TOKEN + "(" + BOOLEAN_PATTERN + ")\\)";
	private static final String INVARIANT_SIMULATION_PARAMETER = SIMULATION_TOKEN
			+ "\\((" + SEARCH_TYPE_PATTERN + ")" + SEP_TOKEN +  "(" + NAME_PATTERN + ")" 
			+ SEP_TOKEN + "(" + NAME_PATTERN + ")" + SEP_TOKEN + "(" + NAME_PATTERN + ")" 
			+ SEP_TOKEN + "(" + BOOLEAN_PATTERN + ")\\)";
	private static final String INVARIANT_UP_TO_PARAMETER = UP_TO_TOKEN + 
			"\\((" + SEARCH_TYPE_PATTERN + ")" + SEP_TOKEN + "(" + NAME_PATTERN + ")" 
			+ SEP_TOKEN + "(" + NAME_PATTERN + ")" + SEP_TOKEN + "(" + NAME_PATTERN + ")" 
			+ SEP_TOKEN + "(" + BOOLEAN_PATTERN + ")\\)";

	private static InvariantCommand command;

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the <tt>INVARIANT</tt>-command.
	 * @return the <tt>INVARIANT</tt>-command
	 */
	protected static InvariantCommand getCommand() {
		if(command==null) {
			command = new InvariantCommand();
		}

		return command;		
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates the <tt>INVARIANT</tt>-command.
	 */
	private InvariantCommand() {
		super(INVARIANT_KEY, ConsoleMessages.INVARIANT_DESCRIPTION_KEY, 
				INVARIANT_DOCUMENTATION_FILE);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public boolean matchesParameter(String line) {
		return line.matches(INVARIANT_DEFAULT_PARAMETER) ||
				line.matches(INVARIANT_ANTICHAIN_PARAMETER) ||
				line.matches(INVARIANT_SIMULATION_PARAMETER) ||
				line.matches(INVARIANT_UP_TO_PARAMETER);
	}

	@Override
	public void execute(String parameter) {
		if(!matchesParameter(parameter)) {
			throw new IllegalArgumentException(Console.messages.getString(
					ConsoleMessages.MISSING_ILLEGAL_ARGUMENT_DESCRIPTION_KEY, INVARIANT_KEY));
		}

		List<String> list;
		String algorithmName = null;
		String automatonName = null;
		String leftHandSideName = null;
		String rightHandSideName = null;
		SearchType type = null;
		boolean computeCounterExample = false;

		if(parameter.matches(INVARIANT_DEFAULT_PARAMETER)) {
			algorithmName = ANTICHAIN_TOKEN;
			type = SearchType.forward;
			computeCounterExample = false;
			//Scan and extract parameter from input string
			list = getProperties(parameter, INVARIANT_DEFAULT_PARAMETER);
			automatonName = list.get(0);
			leftHandSideName = list.get(1);
			rightHandSideName = list.get(2);
		}
		else if(parameter.matches(INVARIANT_ANTICHAIN_PARAMETER)) {
			algorithmName = ANTICHAIN_TOKEN;
			//Scan and extract parameter from input string
			list = getProperties(parameter, INVARIANT_ANTICHAIN_PARAMETER);
			type = SearchType.valueOf(list.get(0));
			automatonName = list.get(1);
			leftHandSideName = list.get(2);
			rightHandSideName = list.get(3);
			computeCounterExample = Boolean.parseBoolean(list.get(4));
		}
		else if(parameter.matches(INVARIANT_SIMULATION_PARAMETER)) {
			algorithmName = SIMULATION_TOKEN;
			//Scan and extract parameter from input string
			list = getProperties(parameter, INVARIANT_SIMULATION_PARAMETER);
			type = SearchType.valueOf(list.get(0));
			automatonName = list.get(1);
			leftHandSideName = list.get(2);
			rightHandSideName = list.get(3);
			computeCounterExample = Boolean.parseBoolean(list.get(4));
		}
		else if(parameter.matches(INVARIANT_UP_TO_PARAMETER)) {
			algorithmName = UP_TO_TOKEN;
			//Scan and extract parameter from input string
			list = getProperties(parameter, INVARIANT_UP_TO_PARAMETER);
			type = SearchType.valueOf(list.get(0));
			automatonName = list.get(1);
			leftHandSideName = list.get(2);
			rightHandSideName = list.get(3);
			computeCounterExample = Boolean.parseBoolean(list.get(4));
		}

		//Create automaton
		if(!Console.repo.isAutomaton(automatonName)) {
			//TODO Replace exception by console message
			throw new RuntimeException(INVARIANT_KEY+": "+Console.messages.getString(
					ConsoleMessages.AUTOMATON_NOT_FOUND_KEY, automatonName)+" "+
							Console.messages.getString(ConsoleMessages.COMPUTATION_INTERRUPTED_KEY));
		}
		Automaton automaton = (Automaton) Console.repo.get(automatonName);

		//Get decompositions of the left and the right hand-side of the transformation rule
		CospanDecomposition leftHandSide = null;
		CospanDecomposition rightHandSide = null;
		if(Console.repo.isCospan(leftHandSideName)) {
			Cospan cospan = (Cospan) Console.repo.get(leftHandSideName);
			CospanDecomposeAction action = new CospanDecomposeAction(cospan);
			action.execute();
			leftHandSide = action.getResult();
			
			if(leftHandSide==null) {
				Console.out.println(Console.messages.getString(
						ConsoleMessages.COSPAN_DECOMPOSITION_FAILED_EXCEPTION_KEY, leftHandSideName));
				Console.out.println(Console.messages.getString(
						ConsoleMessages.COMPUTATION_INTERRUPTED_KEY));
				return;
			}
		}
		else if(Console.repo.isDecomposition(leftHandSideName)) {
			leftHandSide = (CospanDecomposition) Console.repo.get(leftHandSideName);
		}
		else {
			Console.out.println(Console.messages.getString(
					ConsoleMessages.DECOMPOSITION_NOT_FOUND_KEY, leftHandSideName));
			Console.out.println(Console.messages.getString(
					ConsoleMessages.COMPUTATION_INTERRUPTED_KEY));
			return;
		}
		
		if(Console.repo.isCospan(rightHandSideName)) {
			Cospan cospan = (Cospan) Console.repo.get(rightHandSideName);
			CospanDecomposeAction action = new CospanDecomposeAction(cospan);
			action.execute();
			rightHandSide = action.getResult();
			
			if(rightHandSide==null) {
				Console.out.println(Console.messages.getString(
						ConsoleMessages.COSPAN_DECOMPOSITION_FAILED_EXCEPTION_KEY, rightHandSideName));
				Console.out.println(Console.messages.getString(
						ConsoleMessages.COMPUTATION_INTERRUPTED_KEY));
				return;
			}
		}
		else if(Console.repo.isDecomposition(rightHandSideName)) {
			rightHandSide = (CospanDecomposition) Console.repo.get(rightHandSideName);
		}
		else {
			Console.out.println(Console.messages.getString(
					ConsoleMessages.DECOMPOSITION_NOT_FOUND_KEY, rightHandSideName));
			Console.out.println(Console.messages.getString(
					ConsoleMessages.COMPUTATION_INTERRUPTED_KEY));
			return;
		}

		InvariantCheckingAction action = null;
		String resultString;
		if(algorithmName.equals(ANTICHAIN_TOKEN)) {
			//Perform the invariant check using the antichain language inclusion algorithm
			action = new AntichainInvariantCheckingAction(type, 
					computeCounterExample, automaton, leftHandSide, rightHandSide);
				
			if(Console.progressThread!=null) {
				action.addComputationListener(Console.progressThread);
			}
			Console.out.println(Console.messages.getString(ConsoleMessages.CHECK_INVARIANT_ANTICHAIN, type.toString()));
			action.execute();
		}
		else if(algorithmName.equals(SIMULATION_TOKEN)) {
			action = new SimulationAntichainInvariantCheckingAction(type, 
					computeCounterExample, automaton, leftHandSide, rightHandSide);
			if(Console.progressThread!=null) {
				action.addComputationListener(Console.progressThread);
			}
			Console.out.println(Console.messages.getString(ConsoleMessages.CHECK_INVARIANT_SIMULATION_ANTICHAIN));
			action.execute();
		}
		else if(algorithmName.equals(UP_TO_TOKEN)) {
			action = new BisimulationUpToCongruenceInvariantCheckingAction(type,
					computeCounterExample, automaton, leftHandSide, rightHandSide);
			if(Console.progressThread!=null) {
				action.addComputationListener(Console.progressThread);
			}
			Console.out.println(Console.messages.getString(ConsoleMessages.CHECK_INVARIANT_BISIMULATION_UP_TO));
			action.execute();
		}

		//Get result
		boolean isInvariant = action.getResult();
		if (isInvariant) {
			resultString = Console.messages.getString(ConsoleMessages.IS_AN_INVARIANT_KEY,
					automatonName);
		} else {
			resultString = Console.messages.getString(ConsoleMessages.IS_NOT_AN_INVARIANT_KEY,
					automatonName);
			
			resultString += Console.messages.getString(ConsoleMessages.COUNTER_EXAMPLE_FOUND_KEY,
					automatonName);
			RuleCounterExample counterExample = action.getCounterExample();
			Console.out.println(Console.messages.getString(ConsoleMessages.LEFT_HAND_SIDE_KEY));
			Console.out.print(counterExample.getSource());
			
			Console.out.println();
			Console.out.println(Console.messages.getString(ConsoleMessages.RIGHT_HAND_SIDE_KEY));
			Console.out.print(counterExample.getTarget());
		}
		Console.out.println(resultString);
		
		//Print elapsed runtime (if time mode is activated)
		printRuntime(action.getElapsedTime());
	}

}
