/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.cli;

import java.io.File;
import java.util.List;

import de.uni_due.inf.ti.raven.actions.AntichainLanguageInclusionAction;
import de.uni_due.inf.ti.raven.actions.BisimulationUpToCongruenceLanguageInclusionAction;
import de.uni_due.inf.ti.raven.actions.LanguageInclusionAction;
import de.uni_due.inf.ti.raven.actions.SimulationAntichainLanguageInclusionAction;
import de.uni_due.inf.ti.raven.algorithms.Algorithm.SearchType;
import de.uni_due.inf.ti.raven.data.CospanCounterExample;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * Console command to handle the execution of language inclusion
 * algorithms.
 * 
 * @author Christoph Blume
 * @version $Revision: 1.40 $
 * 
 */
public final class LanguageInclusionCommand extends Command {

	private static final File LANGUAGE_INCLUSION_DOCUMENTATION_FILE = new File(
			RavenMain.DOCUMENTATION_FOLDER + "LanguageInclusionCommandDocumentation.txt");

	private static final String ANTICHAIN_TOKEN = Console.ANTICHAIN_TOKEN;
	private static final String BACKWARD_TOKEN = Console.BACKWARD_TOKEN;
	private static final String FORWARD_TOKEN = Console.FORWARD_TOKEN;
	private static final String SIMULATION_TOKEN = Console.SIMULATION_TOKEN;
	private static final String SEP_TOKEN = Console.SEP_TOKEN;
	private static final String UP_TO_TOKEN = Console.UP_TO_TOKEN;
	private static final String BOOLEAN_PATTERN = Console.BOOLEAN_PATTERN;
	private static final String NAME_PATTERN = Console.NAME_PATTERN;
	private static final String SEARCH_TYPE_PATTERN = FORWARD_TOKEN + "|" + BACKWARD_TOKEN;

	private static final String LANGUAGE_INCLUSION_KEY = Console.LANGUAGE_INCLUSION_TOKEN;
	private static final String LANG_INC_DEFAULT_PARAMETER = "\\(("
			+ NAME_PATTERN + ")" + SEP_TOKEN + "(" + NAME_PATTERN + ")\\)";
	private static final String LANG_INC_ANTICHAIN_PARAMETER = 
			ANTICHAIN_TOKEN + "\\((" + SEARCH_TYPE_PATTERN + ")" + SEP_TOKEN + 
			"(" + NAME_PATTERN + ")" + SEP_TOKEN + "(" + NAME_PATTERN + ")" + SEP_TOKEN + 
			"(" + BOOLEAN_PATTERN + ")\\)";
	private static final String LANG_INC_SIM_BASED_ANTICHAIN_PARAMETER = 
			SIMULATION_TOKEN + "\\((" + SEARCH_TYPE_PATTERN + ")" + SEP_TOKEN
			+ "(" + NAME_PATTERN + ")" + SEP_TOKEN + "(" + NAME_PATTERN + ")" + SEP_TOKEN 
			+ "(" + BOOLEAN_PATTERN + ")\\)";
	private static final String LANG_INC_UP_TO_CONGRUENCE_PARAMETER = 
			UP_TO_TOKEN + "\\((" + SEARCH_TYPE_PATTERN + ")" + SEP_TOKEN + "(" + NAME_PATTERN + ")" 
			+ SEP_TOKEN + "(" + NAME_PATTERN + ")" + SEP_TOKEN + "(" + BOOLEAN_PATTERN + ")\\)";
	
	private static LanguageInclusionCommand command;

	/*------------------------------------------------------------------------------------------*/
	/* 		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the <tt>LANGUAGE_INCLUSION</tt>-command.
	 * 
	 * @return the <tt>LANGUAGE_INCLUSION</tt>-command
	 */
	protected static LanguageInclusionCommand getCommand() {
		if (command == null) {
			command = new LanguageInclusionCommand();
		}

		return command;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates the <tt>LANGUAGE_INCLUSION</tt>-command.
	 */
	private LanguageInclusionCommand() {
		super(LANGUAGE_INCLUSION_KEY,
				ConsoleMessages.LANGUAGE_INCLUSION_DESCRIPTION_KEY,
				LANGUAGE_INCLUSION_DOCUMENTATION_FILE);
	}

	/*------------------------------------------------------------------------------------------*/
	/* 		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public boolean matchesParameter(String line) {
		return line.matches(LANG_INC_DEFAULT_PARAMETER)
				|| line.matches(LANG_INC_ANTICHAIN_PARAMETER)
				|| line.matches(LANG_INC_SIM_BASED_ANTICHAIN_PARAMETER)
				|| line.matches(LANG_INC_UP_TO_CONGRUENCE_PARAMETER);
	}

	@Override
	public void execute(String parameter) {
		if (!matchesParameter(parameter)) {
			throw new IllegalArgumentException(Console.messages.getString(
					ConsoleMessages.MISSING_ILLEGAL_ARGUMENT_DESCRIPTION_KEY,
					LANGUAGE_INCLUSION_KEY));
		}

		List<String> list;
		String algorithmName = null;
		String firstAutomatonName = null;
		String secondAutomatonName = null;
		SearchType type = null;
		boolean computeCounterExample = false;
		LanguageInclusionAction action = null;

		if (parameter.matches(LANG_INC_DEFAULT_PARAMETER)) {
			algorithmName = ANTICHAIN_TOKEN;
			type = SearchType.forward;
			computeCounterExample = false;
			// Scan and extract parameter from input string
			list = getProperties(parameter, LANG_INC_DEFAULT_PARAMETER);
			firstAutomatonName = list.get(0);
			secondAutomatonName = list.get(1);
		} else if (parameter.matches(LANG_INC_ANTICHAIN_PARAMETER)) {
			algorithmName = ANTICHAIN_TOKEN;
			// Scan and extract parameter from input string
			list = getProperties(parameter, LANG_INC_ANTICHAIN_PARAMETER);
			type = SearchType.valueOf(list.get(0));
			firstAutomatonName = list.get(1);
			secondAutomatonName = list.get(2);
			computeCounterExample = Boolean.parseBoolean(list.get(3));
		} else if (parameter.matches(LANG_INC_SIM_BASED_ANTICHAIN_PARAMETER)) {
			algorithmName = SIMULATION_TOKEN;
			// Scan and extract parameter from input string
			list = getProperties(parameter,
					LANG_INC_SIM_BASED_ANTICHAIN_PARAMETER);
			type = SearchType.valueOf(list.get(0));
			firstAutomatonName = list.get(1);
			secondAutomatonName = list.get(2);
			computeCounterExample = Boolean.parseBoolean(list.get(3));
		} else if (parameter.matches(LANG_INC_UP_TO_CONGRUENCE_PARAMETER)) {
			algorithmName = UP_TO_TOKEN;
			// Scan and extract parameter from input string
			list = getProperties(parameter, LANG_INC_UP_TO_CONGRUENCE_PARAMETER);
			type = SearchType.valueOf(list.get(0));
			firstAutomatonName = list.get(1);
			secondAutomatonName = list.get(2);
			computeCounterExample = Boolean.parseBoolean(list.get(3));
		}

		if (!Console.repo.isAutomaton(firstAutomatonName)
				&& !Console.repo.isAutomaton(firstAutomatonName)) {
			throw new IllegalArgumentException(
					Console.messages.getString(
							ConsoleMessages.AUTOMATON_NOT_FOUND_KEY,
							firstAutomatonName));
		}
		if (!Console.repo.isAutomaton(secondAutomatonName)
				&& !Console.repo.isAutomaton(secondAutomatonName)) {
			throw new IllegalArgumentException(Console.messages.getString(
					ConsoleMessages.AUTOMATON_NOT_FOUND_KEY,
					secondAutomatonName));
		}

		Automaton firstAutomaton = (Automaton) Console.repo.get(firstAutomatonName);
		Automaton secondAutomaton = (Automaton) Console.repo.get(secondAutomatonName);

		String message = null;
		if (algorithmName.equals(ANTICHAIN_TOKEN)) {
			action = new AntichainLanguageInclusionAction(type,
					computeCounterExample, firstAutomaton, secondAutomaton);
			message = Console.messages.getString(ConsoleMessages.COMPUTING_LANGUAGE_INCLUSION_KEY,
					algorithmName+" "+type.toString(), firstAutomatonName, secondAutomatonName);
		} else if (algorithmName.equals(SIMULATION_TOKEN)) {
			action = new SimulationAntichainLanguageInclusionAction(type,
					computeCounterExample, firstAutomaton, secondAutomaton);
			message = Console.messages.getString(ConsoleMessages.COMPUTING_LANGUAGE_INCLUSION_KEY,
					algorithmName+" "+type.toString(), firstAutomatonName, secondAutomatonName);
		} else if (algorithmName.equals(UP_TO_TOKEN)) {
			action = new BisimulationUpToCongruenceLanguageInclusionAction(SearchType.forward, 
					computeCounterExample, firstAutomaton, secondAutomaton);
			message = Console.messages.getString(ConsoleMessages.COMPUTING_LANGUAGE_INCLUSION_KEY,
					algorithmName+" "+type.toString(), firstAutomatonName, secondAutomatonName);
		}

		if (Console.progressThread != null) {
			action.addComputationListener(Console.progressThread);
		}

		Console.out.println(message);
		action.execute();

		String resultString;
		boolean isIncluded = action.getResult();
		if (isIncluded) {
			resultString = Console.messages.getString(
					ConsoleMessages.LANGUAGE_INCLUSION_DOES_HOLD,
					firstAutomatonName, secondAutomatonName);
		} else {
			resultString = Console.messages.getString(
					ConsoleMessages.LANGUAGE_INCLUSION_DOES_NOT_HOLD,
					firstAutomatonName, secondAutomatonName);
		}

		Console.out.println(resultString);

		//Print the counterexample (if one is available)
		if(computeCounterExample && !isIncluded) {
			CospanCounterExample counterExample = action.getCounterExample();
			int size = counterExample.getCospan().getLength();
			Console.out.println("The counterexample consists of the " +
					"following operation"+(size<2 ? "" : "s")+":");
			int i = 1;
			int length = String.valueOf(size).length();
			for(Operation operation: counterExample.getCospan()) {
				StringBuilder output = new StringBuilder();
				for(int j=String.valueOf(i).length();j<length;j++) {
					output.append(" ");
				}
				output.append(i+": "+operation.toString());
				Console.out.println(output.toString());
				i++;
			}
		}

		//Print elapsed runtime (if time mode is activated)
		printRuntime(action.getElapsedTime());
	}

}
