/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.cli;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.List;
import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.actions.IOAction;
import de.uni_due.inf.ti.raven.actions.LoadAction;
import de.uni_due.inf.ti.raven.actions.LoadAction.LoadType;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * Console command to handle the loading of data.
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.37 $
 *
 */
public final class LoadCommand extends Command {

	private static final File LOAD_DOCUMENTATION_FILE = new File(
			RavenMain.DOCUMENTATION_FOLDER + "LoadCommandDocumentation.txt");
	
	private static final String AUTOMATON_TOKEN = Console.AUTOMATON_TOKEN;
	private static final String COSPAN_TOKEN = Console.COSPAN_TOKEN;
	private static final String DECOMPOSITION_TOKEN = Console.DECOMPOSITION_TOKEN;
	private static final String GRAPH_TOKEN = Console.GRAPH_TOKEN;
	private static final String IS_TOKEN = Console.IS_TOKEN;
	private static final String SIGNATURE_TOKEN = Console.SIGNATURE_TOKEN;

	private static final String FILENAME_PATTERN = Console.FILENAME_PATTERN;
	private static final String NAME_PATTERN = Console.NAME_PATTERN;

	private static final String LOAD_KEY = Console.LOAD_TOKEN;
	private static final String LOAD_AUTOMATON_PARAMETER = "("+NAME_PATTERN+")"+IS_TOKEN + AUTOMATON_TOKEN + "\\(("+FILENAME_PATTERN+")\\)";
	private static final String LOAD_COSPAN_PARAMETER = "("+NAME_PATTERN+")"+ IS_TOKEN + COSPAN_TOKEN + "\\(("+FILENAME_PATTERN+")\\)";
	private static final String LOAD_DECOMPOSITION_PARAMETER = "("+NAME_PATTERN+")"+ IS_TOKEN + DECOMPOSITION_TOKEN + "\\(("+FILENAME_PATTERN+")\\)";
	private static final String LOAD_GRAPH_PARAMETER = "("+NAME_PATTERN+")"+ IS_TOKEN + GRAPH_TOKEN + "\\(("+FILENAME_PATTERN+")\\)";
	private static final String LOAD_SIGNATURE_PARAMETER = "("+NAME_PATTERN+")"+ IS_TOKEN + SIGNATURE_TOKEN + "\\(("+FILENAME_PATTERN+")\\)";

	private static LoadCommand command;

	private Logger logger;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the <tt>LOAD</tt>-command.
	 * @return the <tt>LOAD</tt>-command
	 */
	protected static LoadCommand getCommand() {
		if(command==null) {
			command = new LoadCommand();
		}

		return command;		
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates the <tt>LOAD</tt>-command.
	 */	
	private LoadCommand() {
		super(LOAD_KEY, ConsoleMessages.LOAD_DESCRIPTION_KEY, 
				LOAD_DOCUMENTATION_FILE);
		
		logger = Logger.getLogger(getClass().getName());
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public boolean matchesParameter(String line) {
		return line.matches(LOAD_COSPAN_PARAMETER) ||
				line.matches(LOAD_DECOMPOSITION_PARAMETER) || 
				line.matches(LOAD_GRAPH_PARAMETER) || 
				line.matches(LOAD_AUTOMATON_PARAMETER) || 
				line.matches(LOAD_SIGNATURE_PARAMETER);
	}

	@Override
	public void execute(String parameter) {
		if(!matchesParameter(parameter)) {
			throw new IllegalArgumentException(Console.messages.getString(
					ConsoleMessages.MISSING_ILLEGAL_ARGUMENT_DESCRIPTION_KEY, LOAD_KEY));
		}
		
		IOAction action;

		if(parameter.matches(LOAD_COSPAN_PARAMETER)) {
			//Extract parameter from input string
			List<String> list = getProperties(parameter, LOAD_COSPAN_PARAMETER);
			String name = list.get(0);
			String filename = list.get(1);
			
			try {
				//Prepare the filename
				if (!filename.startsWith(RavenMain.SAVE_DIR)){
					filename = RavenMain.SAVE_DIR + filename;
				}
				if(!filename.endsWith("."+RavenMain.COSPAN_GXL_FILE_EXTENSION)) {
					filename += "." + RavenMain.COSPAN_GXL_FILE_EXTENSION;
				}
				
				//Execute load action
				action = new LoadAction(LoadType.COSPANGXL, name, filename);
				action.execute();
				
				//If no error occurred the decomposition has been successfully added
				if(!action.hasExceptionOccured()) {
					Console.out.println(Console.messages.getString(
							ConsoleMessages.COSPAN_ADDED_KEY, name));
				}
				//If an error occurred inform the user
				else {
					Console.out.println(Console.messages.getString(
							ConsoleMessages.EXCEPTION_OCCURED_KEY));
					action.getException().printStackTrace();

					logger.throwing(getClass().getName(), "execute", action.getException());
				}
			} catch (FileNotFoundException e) {
				Console.out.println(Console.messages.getString(
						ConsoleMessages.FILE_NOT_FOUND_KEY, filename));
			}
			
		}
		else if(parameter.matches(LOAD_DECOMPOSITION_PARAMETER)) {
			//Extract parameter from input string
			List<String> list = getProperties(parameter, LOAD_DECOMPOSITION_PARAMETER);
			String name = list.get(0);
			String filename = list.get(1);
			
			try {
				//Prepare the filename
				if (!filename.startsWith(RavenMain.SAVE_DIR)){
					filename = RavenMain.SAVE_DIR + filename;
				}
				if(!filename.endsWith("."+RavenMain.DECOMPOSITION_FILE_EXTENSION)) {
					filename += "." + RavenMain.DECOMPOSITION_FILE_EXTENSION;
				}
				
				//Execute load action
				action = new LoadAction(LoadType.DECOMPOSITION, name, filename);
				action.execute();
				
				//If no error occurred the decomposition has been successfully added
				if(!action.hasExceptionOccured()) {
					Console.out.println(Console.messages.getString(
							ConsoleMessages.DECOMPOSITION_ADDED_KEY, name));
				}
				//If an error occurred inform the user
				else {
					Console.out.println(Console.messages.getString(
							ConsoleMessages.EXCEPTION_OCCURED_KEY));
					action.getException().printStackTrace();

					logger.throwing(getClass().getName(), "execute", action.getException());
				}
			} catch (FileNotFoundException e) {
				Console.out.println(Console.messages.getString(
						ConsoleMessages.FILE_NOT_FOUND_KEY, filename));
			}
			
		}
		else if(parameter.matches(LOAD_GRAPH_PARAMETER)) {
			//Extract parameter from input string
			List<String> list = getProperties(parameter, LOAD_GRAPH_PARAMETER);
			String name = list.get(0);
			String filename = list.get(1);

			try {
				//Prepare the filename
				if (!filename.startsWith(RavenMain.SAVE_DIR)){
					filename = RavenMain.SAVE_DIR + filename;
				}
				if(!filename.endsWith("."+RavenMain.GRAPH_GXL_FILE_EXTENSION)) {
					filename += "." + RavenMain.GRAPH_GXL_FILE_EXTENSION;
				}

				//Execute load action
				action = new LoadAction(LoadType.GRAPHGXL, name, filename);
				action.execute();
				
				//If no error occurred the graph has been successfully added
				if(!action.hasExceptionOccured()) {
					Console.out.println(Console.messages.getString(
							ConsoleMessages.GRAPH_ADDED_KEY, name));
				}
				//If an error occurred inform the user
				else {
					Console.out.println(Console.messages.getString(
							ConsoleMessages.EXCEPTION_OCCURED_KEY));

					logger.throwing(getClass().getName(), "execute", action.getException());
				}
			} catch (FileNotFoundException e) {
				Console.out.println(Console.messages.getString(
					ConsoleMessages.FILE_NOT_FOUND_KEY, filename));
			}
		}		
		else if(parameter.matches(LOAD_AUTOMATON_PARAMETER)){
			if (!matchesParameter(parameter)) {
				throw new IllegalArgumentException(Console.messages.getString(
					ConsoleMessages.MISSING_ILLEGAL_ARGUMENT_DESCRIPTION_KEY, LOAD_KEY));
			}

			//Extract the parameters from the input string
			List<String> list = getProperties(parameter, LOAD_AUTOMATON_PARAMETER);
			String name = list.get(0);
			String filename = list.get(1);
			
			try {
				//Prepare the filename
				if (!filename.startsWith(RavenMain.SAVE_DIR)){
					filename = RavenMain.SAVE_DIR + File.separator + filename;
				}
				if(!filename.endsWith("."+RavenMain.AUTOMATON_FILE_EXTENSION)) {
					filename += "." + RavenMain.AUTOMATON_FILE_EXTENSION;
				}

				//Execute load action
				action = new LoadAction(LoadType.AUTOMATON, name, filename);
				action.execute();
				
				//If no error occurred the automaton has been successfully added
				if(!action.hasExceptionOccured()) {
					Automaton automaton = (Automaton) Console.repo.get(name);
					Console.out.println(Console.messages.getString(ConsoleMessages.AUTOMATON_ADDED_KEY, 
							name,String.valueOf(automaton.getMaximumInterface())));
				}
				//If an error occurred inform the user
				else {
					Console.out.println(Console.messages.getString(
							ConsoleMessages.EXCEPTION_OCCURED_KEY));
					logger.throwing(getClass().getName(), "execute", action.getException());
				}
			} catch (FileNotFoundException e) {
				Console.out.println(Console.messages.getString(
						ConsoleMessages.FILE_NOT_FOUND_KEY, filename));
			}
		}
		else if(parameter.matches(LOAD_SIGNATURE_PARAMETER)) {
			//Extract parameter from input string
			List<String> list = getProperties(parameter, LOAD_SIGNATURE_PARAMETER);
			String name = list.get(0);
			String filename = list.get(1);

			try {
				//Prepare the filename
				if (!filename.startsWith(RavenMain.SAVE_DIR)){
					filename = RavenMain.SAVE_DIR + filename;
				}
				if(!filename.endsWith("."+RavenMain.SIGNATURE_FILE_EXTENSION)) {
					filename += "." + RavenMain.SIGNATURE_FILE_EXTENSION;
				}

				//Execute load action
				action = new LoadAction(LoadType.SIGNATURE, name, filename);
				action.execute();
				
				//If no error occurred the graph has been successfully added
				if(!action.hasExceptionOccured()) {
					Console.out.println(Console.messages.getString(
							ConsoleMessages.SIGNATURE_ADDED_KEY, name));
				}
				//If an error occurred inform the user
				else {
					Console.out.println(Console.messages.getString(
							ConsoleMessages.EXCEPTION_OCCURED_KEY));

					logger.throwing(getClass().getName(), "execute", action.getException());
				}
			} catch (FileNotFoundException e) {
				Console.out.println(Console.messages.getString(
					ConsoleMessages.FILE_NOT_FOUND_KEY, filename));
			}
		}
	}
	


}
