/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.cli;

import java.io.File;
import java.util.List;

import de.uni_due.inf.ti.raven.actions.LibTWGreedyFillInAction;
import de.uni_due.inf.ti.raven.actions.MembershipAction;
import de.uni_due.inf.ti.raven.algorithms.DefaultCospanDecompositionAlgorithm;
import de.uni_due.inf.ti.raven.data.Cospan;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * Console command to handle the execution of membership check algorithms.
 * @author Christoph Blume
 * @version $Revision: 1.27 $
 *
 */
public final class MembershipCommand extends Command {

	private static final File MEMBERSHIP_DOCUMENTATION_FILE = new File(
			RavenMain.DOCUMENTATION_FOLDER + "MembershipCommandDocumentation.txt");
	
	private static final String SEP_TOKEN = Console.SEP_TOKEN;
	private static final String NAME_PATTERN = Console.NAME_PATTERN;
	
	private static final String MEMBERSHIP_KEY = Console.MEMBERSHIP_TOKEN;
	private static final String MEMBERSHIP_PARAMETER = "\\(("+NAME_PATTERN+")"+SEP_TOKEN+"("+NAME_PATTERN+")\\)";
	private static MembershipCommand command;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the <tt>MEMBERSHIP</tt>-command.
	 * @return the <tt>MEMBERSHIP</tt>-command
	 */
	protected static MembershipCommand getCommand() {
		if(command==null) {
			command = new MembershipCommand();
		}

		return command;		
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates the <tt>MEMBERSHIP</tt>-command.
	 */
	private MembershipCommand() {
		super(MEMBERSHIP_KEY, ConsoleMessages.MEMBERSHIP_DESCRIPTION_KEY, 
				MEMBERSHIP_DOCUMENTATION_FILE);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public boolean matchesParameter(String line) {
		return line.matches(MEMBERSHIP_PARAMETER);
	}

	@Override
	public void execute(String parameter) {
		if(!matchesParameter(parameter)) {
			throw new IllegalArgumentException(Console.messages.getString(
					ConsoleMessages.MISSING_ILLEGAL_ARGUMENT_DESCRIPTION_KEY, MEMBERSHIP_KEY));
		}
		
		//Scan and extract parameter from input string
		List<String> list = getProperties(parameter, MEMBERSHIP_PARAMETER);
		String automatonName = list.get(0);
		String objectName = list.get(1);

		if(!Console.repo.isAutomaton(automatonName)) {
			throw new IllegalArgumentException(Console.messages.getString(
					ConsoleMessages.AUTOMATON_NOT_FOUND_KEY, automatonName));
		}
		if(!Console.repo.isGraph(objectName) && 
				!Console.repo.isCospan(objectName) &&
				!Console.repo.isDecomposition(objectName)) {
			throw new IllegalArgumentException(Console.messages.getString(
							ConsoleMessages.DATA_NOT_FOUND_KEY, objectName));
		}

		Automaton automaton = (Automaton) Console.repo.get(automatonName);
		MembershipAction action = null;
		String dataString = null;
		CospanDecomposition decomposition = null;
		if(Console.repo.isGraph(objectName)) {
			Graph graph = (Graph) Console.repo.get(objectName);			
			LibTWGreedyFillInAction decompositionAction = new LibTWGreedyFillInAction(graph);
			decompositionAction.execute();			
			decomposition = CospanDecomposition.getCospanDecomposition(decompositionAction.getResult());
			dataString = "The graph '"+objectName+"'";
		}
		else if(Console.repo.isCospan(objectName)) {
			Cospan cospan = (Cospan) Console.repo.get(objectName);
			DefaultCospanDecompositionAlgorithm algorithm = new DefaultCospanDecompositionAlgorithm();
			decomposition = algorithm.decompose(cospan);
			dataString = "The cospan '"+objectName+"'";
		}
		else if(Console.repo.isDecomposition(objectName)) {
			decomposition = (CospanDecomposition) Console.repo.get(objectName);
			dataString = "The decomposition '"+objectName+"'";
		}
		action = new MembershipAction(automaton, decomposition);
		action.execute();
		Console.out.println(dataString+" has been " + (action.getResult() ? "accepted" : "rejected")+".");
		
		//Print elapsed runtime (if time mode is activated)
		printRuntime(action.getElapsedTime());
	}

}
