/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.cli;

import java.io.File;
import java.util.List;

import de.uni_due.inf.ti.raven.actions.PrintAction;
import de.uni_due.inf.ti.raven.data.Cospan;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Data;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.GraphDecomposition;
import de.uni_due.inf.ti.raven.data.Signature;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * Console command to handle the printout of data from the repository.
 * @author Christoph Blume
 * @version $Revision: 1.30 $
 *
 */
public final class PrintCommand extends Command {

	private static final File PRINT_DOCUMENTATION_FILE = new File(
			RavenMain.DOCUMENTATION_FOLDER + "PrintCommandDocumentation.txt");
	
	private static final String NAME_PATTERN = Console.NAME_PATTERN;
	
	private static final String PRINT_KEY = Console.PRINT_TOKEN;
	private static final String PRINT_PARAMETER = "("+NAME_PATTERN+")";

	private static PrintCommand command;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the <tt>PRINT</tt>-command.
	 * @return the <tt>PRINT</tt>-command
	 */
	protected static PrintCommand getCommand() {
		if(command==null) {
			command = new PrintCommand();
		}

		return command;		
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates the <tt>PRINT</tt>-command.
	 */
	private PrintCommand() {
		super(PRINT_KEY, ConsoleMessages.PRINT_DESCRIPTION_KEY, 
				PRINT_DOCUMENTATION_FILE);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public boolean matchesParameter(String line) {
		return line.matches(PRINT_PARAMETER);
	}

	@Override
	public void execute(String parameter) {
		if(!matchesParameter(parameter)) {
			throw new IllegalArgumentException(Console.messages.getString(
					ConsoleMessages.MISSING_ILLEGAL_ARGUMENT_DESCRIPTION_KEY, PRINT_KEY));
		}
		
		//Extract parameter from input string
		List<String> list = getProperties(parameter, PRINT_PARAMETER);
		String name = list.get(0);
		Data data = Console.repo.get(name);
		
		if(!Console.repo.contains(name)) {
			throw new IllegalArgumentException(Console.messages.getString(
					ConsoleMessages.DATA_NOT_FOUND_EXCEPTION_KEY, PRINT_KEY, name));
		}
		
		if(Console.repo.isAutomaton(name)) {			
			new PrintAction(Console.out, (Automaton) data).execute();
		}
		else if(Console.repo.isCospan(name)) {
			new PrintAction(Console.out, (Cospan) data).execute();
		}
		else if(Console.repo.isDecomposition(name)) {
			Data decomposition = Console.repo.get(name);
			
			if(decomposition instanceof CospanDecomposition) {
				new PrintAction(Console.out, (CospanDecomposition) data).execute();
			}
			else if(decomposition instanceof GraphDecomposition) {
				new PrintAction(Console.out, (GraphDecomposition) data).execute();
			}
		}
		else if(Console.repo.isGraph(name)) {
			new PrintAction(Console.out, (Graph) data).execute();
		}
		else if(Console.repo.isSignature(name)) {
			new PrintAction(Console.out, (Signature) data).execute();
		}
		else {
			Console.out.println(
					Console.messages.getString(ConsoleMessages.NO_PRINT_ACTION_AVAILABLE_DESCRIPTION_KEY, 
							Console.repo.get(name).getClass().getSimpleName()));
		}
	}

}
