/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.cli;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.List;

import de.uni_due.inf.ti.raven.actions.PrintAction;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Decomposition;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.GraphDecomposition;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.io.DotPrinter;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * Console command to handle the printout of BDDs or graphs in DOT-format.
 * @author Christoph Blume
 * @version $Revision: 1.34 $
 *
 */
public final class PrintDotCommand extends Command {

	private static final File PRINTDOT_DOCUMENTATION_FILE = new File(
			RavenMain.DOCUMENTATION_FOLDER + "PrintDotCommandDocumentation.txt");
	
	private static final String DECOMPOSITION_TOKEN = Console.DECOMPOSITION_TOKEN;
	private static final String FINAL_TOKEN = Console.FINAL_TOKEN;
	private static final String GRAPH_TOKEN = Console.GRAPH_TOKEN;
	private static final String INITIAL_TOKEN = Console.INITIAL_TOKEN;
	private static final String SEP_TOKEN = Console.SEP_TOKEN;
	private static final String STATES_TOKEN = Console.STATES_TOKEN;
	private static final String TRANSITION_TOKEN = Console.TRANSITION_TOKEN;

	private static final String NAME_PATTERN = Console.NAME_PATTERN;
	private static final String TRANSITION_PATTERN = Console.OPERATION_PATTERN;	

	private static final String PRINT_DOT_KEY = Console.PRINT_DOT_TOKEN;
	private static final String PRINT_DOT_DECOMPOSITION_PARAMETER = DECOMPOSITION_TOKEN+"\\(("+NAME_PATTERN+")\\)";
	private static final String PRINT_DOT_FINAL_STATES_PARAMETER = FINAL_TOKEN+"\\(("+NAME_PATTERN+")\\)";
	private static final String PRINT_DOT_GRAPH_PARAMETER = GRAPH_TOKEN+"\\(("+NAME_PATTERN+")\\)";
	private static final String PRINT_DOT_INITIAL_STATES_PARAMETER = INITIAL_TOKEN+"\\(("+NAME_PATTERN+")\\)";
	private static final String PRINT_DOT_STATES_PARAMETER = STATES_TOKEN+"\\(("+NAME_PATTERN+")\\)";
	private static final String PRINT_DOT_TRANSITION_PARAMETER = TRANSITION_TOKEN+"\\(("+NAME_PATTERN+")"+SEP_TOKEN+"("+TRANSITION_PATTERN+")\\)";

	private static PrintDotCommand command;

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the <tt>PRINTDOT</tt>-command.
	 * @return the <tt>PRINTDOT</tt>-command
	 */
	protected static PrintDotCommand getCommand() {
		if(command==null) {
			command = new PrintDotCommand();
		}

		return command;		
	}
	
	/**
	 * Returns {@code true} if and only if the parameter {@code parameter} indicate the 
	 * PRINT-DECOMPOSITION option.
	 * @param parameter the parameter passed to this action
	 * @return {@code true} if the parameter indicate the PRINT-DECOMPOSITION option, 
	 * {@code false} otherwise
	 */
	private static boolean isPrintDecompositionOption(String parameter) {
		return parameter.matches(PRINT_DOT_DECOMPOSITION_PARAMETER);
	}
	
	/**
	 * Returns {@code true} if and only if the parameter {@code parameter} indicate the 
	 * PRINT-GRAPH option.
	 * @param parameter the parameter passed to this action
	 * @return {@code true} if the parameter indicate the PRINT-GRAPH option, {@code false} otherwise
	 */
	private static boolean isPrintGraphOption(String parameter) {
		return parameter.matches(PRINT_DOT_GRAPH_PARAMETER);
	}

	/**
	 * Returns {@code true} if and only if the parameter {@code parameter} indicate the 
	 * PRINT-(INITIAL/FINAL/ALL)-STATES option.
	 * @param parameter the parameter passed to this action
	 * @return {@code true} if the parameter indicate the PRINT-(INITIAL/FINAL/ALL)-STATES 
	 * option, {@code false} otherwise
	 */
	private static boolean isPrintStatesOption(String parameter) {
		return parameter.matches(PRINT_DOT_FINAL_STATES_PARAMETER) ||
				parameter.matches(PRINT_DOT_INITIAL_STATES_PARAMETER) ||
				parameter.matches(PRINT_DOT_STATES_PARAMETER);
	}

	/**
	 * Returns {@code true} if and only if the parameter {@code parameter} indicate the 
	 * PRINT-TRANSITION option.
	 * @param parameter the parameter passed to this action
	 * @return {@code true} if the parameter indicate the PRINT-TRANSITION option, 
	 * {@code false} otherwise
	 */
	private static boolean isPrintTransitionOption(String parameter) {
		return parameter.matches(PRINT_DOT_TRANSITION_PARAMETER);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates the <tt>PRINTDOT</tt>-command.
	 */
	private PrintDotCommand() {
		super(PRINT_DOT_KEY, ConsoleMessages.PRINTDOT_DESCRIPTION_KEY, 
				PRINTDOT_DOCUMENTATION_FILE);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public boolean matchesParameter(String line) {
		return (line.matches(PRINT_DOT_DECOMPOSITION_PARAMETER) ||
				line.matches(PRINT_DOT_FINAL_STATES_PARAMETER) ||
				line.matches(PRINT_DOT_GRAPH_PARAMETER) || 
				line.matches(PRINT_DOT_INITIAL_STATES_PARAMETER) ||
				line.matches(PRINT_DOT_STATES_PARAMETER) ||
				line.matches(PRINT_DOT_TRANSITION_PARAMETER));
	}

	@Override
	public void execute(String parameter) {
		if(!matchesParameter(parameter)) {
			throw new IllegalArgumentException(Console.messages.getString(
					ConsoleMessages.MISSING_ILLEGAL_ARGUMENT_DESCRIPTION_KEY, PRINT_DOT_KEY));
		}

		if(isPrintDecompositionOption(parameter)) {
			//Extract parameter from input string
			List<String> list = getProperties(parameter, PRINT_DOT_DECOMPOSITION_PARAMETER);
			String decompositionName = list.get(0);
			if(Console.repo.isDecomposition(decompositionName)) {
				String filename = null;
				try {
					filename = RavenMain.DOT_SAVE_DIR + decompositionName+"."+RavenMain.DOT_FILE_EXTENSION;
					DotPrinter printer = new DotPrinter(filename);
					
					Decomposition<?> decomposition = (Decomposition<?>) Console.repo.get(decompositionName);
					if(decomposition instanceof CospanDecomposition) {
						new PrintAction(printer, (CospanDecomposition) decomposition).execute();
					}
					else if (decomposition instanceof GraphDecomposition) {
						new PrintAction(printer, (GraphDecomposition) decomposition).execute();
					}
					Console.out.println(messages.getString(
							ConsoleMessages.DECOMPOSITION_PRINTED_KEY, filename));
				} catch (FileNotFoundException e) {
					Console.out.println(Console.messages.getString(
							ConsoleMessages.FILE_NOT_FOUND_KEY, filename));
					return;
				}
			}
			else {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.DECOMPOSITION_NOT_FOUND_KEY, decompositionName));
			}
		}
		else if(isPrintGraphOption(parameter)) {
			//Extract parameter from input string
			List<String> list = getProperties(parameter, PRINT_DOT_GRAPH_PARAMETER);
			String graphName = list.get(0);
			if(Console.repo.isGraph(graphName)) {
				String filename = null;
				try {
					filename = RavenMain.DOT_SAVE_DIR + graphName+"."+RavenMain.DOT_FILE_EXTENSION;
					DotPrinter printer = new DotPrinter(filename);
					new PrintAction(printer, (Graph) Console.repo.get(graphName)).execute();
					Console.out.println(messages.getString(
							ConsoleMessages.GRAPH_PRINTED_KEY, filename));
				} catch (FileNotFoundException e) {
					Console.out.println(Console.messages.getString(
							ConsoleMessages.FILE_NOT_FOUND_KEY, filename));
					return;
				}
			}
			else {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.GRAPH_NOT_FOUND_KEY, graphName));
			}
		}
		else if(isPrintStatesOption(parameter)) {
			//Extract parameter from input string
			List<String> list = null;
			String stateType = null;

			if(parameter.matches(PRINT_DOT_FINAL_STATES_PARAMETER)) {
				list = getProperties(parameter, PRINT_DOT_FINAL_STATES_PARAMETER);
				stateType = FINAL_TOKEN;
			}
			else if(parameter.matches(PRINT_DOT_INITIAL_STATES_PARAMETER)) {
				list = getProperties(parameter, PRINT_DOT_INITIAL_STATES_PARAMETER);
				stateType = INITIAL_TOKEN;
			}
			else if(parameter.matches(PRINT_DOT_STATES_PARAMETER)) {
				list = getProperties(parameter, PRINT_DOT_STATES_PARAMETER);
				stateType = STATES_TOKEN;
			}
			
			String automatonName = list.get(0);
			if(Console.repo.isAutomaton(automatonName)) {
				String filename = null;
				try {
					filename = RavenMain.DOT_SAVE_DIR + automatonName+"_"+stateType+"."+RavenMain.DOT_FILE_EXTENSION;
					Automaton automaton = (Automaton) Console.repo.get(automatonName);
					RavenBDD statesBDD = null;
					
					if(stateType.equals(FINAL_TOKEN)) {
						statesBDD = automaton.getFinalStates();
					}
					else if(stateType.equals(INITIAL_TOKEN)) {
						statesBDD = automaton.getInitialStates();
					}
					else if(stateType.equals(STATES_TOKEN)) {
						statesBDD = automaton.getStates();
					}

					DotPrinter printer = new DotPrinter(filename);
					new PrintAction(printer, statesBDD).execute();
					Console.out.println(Console.messages.getString(
							ConsoleMessages.STATES_PRINTED_KEY, automatonName, filename));
				} catch (FileNotFoundException e) {
					Console.out.println(Console.messages.getString(
							ConsoleMessages.FILE_NOT_FOUND_KEY, filename));
					return;
				}
			}
			else {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NOT_FOUND_KEY, automatonName));
			}
		}
		else if(isPrintTransitionOption(parameter)) {
			//Extract parameter from input string
			List<String> list = null;
			if(parameter.matches(PRINT_DOT_TRANSITION_PARAMETER)) {
				list = getProperties(parameter, PRINT_DOT_TRANSITION_PARAMETER);
			}
			
			String automatonName = list.get(0);
			Operation operation = Operation.parse(list.get(1));
			if(Console.repo.isAutomaton(automatonName)) {
				Automaton automaton = (Automaton) Console.repo.get(automatonName);

				if(!automaton.getAlphabet().contains(operation)) {
					throw new IllegalArgumentException(Console.messages.getString(
									ConsoleMessages.ILLEGAL_OPERATION_KEY, 
									operation.toString(), automatonName));
				}

				String filename = "";
				try {
					File dir = new File(RavenMain.DOT_SAVE_DIR+automatonName);
					if(!dir.exists()) {
						if(!dir.mkdir()) {
							Console.out.println(Console.messages.getString(
									ConsoleMessages.DIRECTORY_NOT_CREATED_KEY, dir.getAbsolutePath()));
							return;
						}
					}
					
					filename = automatonName+"_"+operation+"."+RavenMain.DOT_FILE_EXTENSION;
					filename = dir.getAbsolutePath() + File.separator +filename.replace(":", ",");
					
					DotPrinter printer = new DotPrinter(filename);
					new PrintAction(printer, automaton.getTransitionRelation(operation)).execute();
					Console.out.println(Console.messages.getString(
							ConsoleMessages.TRANSITION_GRAPH_PRINTED_KEY, 
							operation.toString(), filename.replace(RavenMain.MAIN_DIR, "")));
				} catch (FileNotFoundException e) {
					Console.out.println(Console.messages.getString(
							ConsoleMessages.FILE_NOT_FOUND_KEY, filename));
					return;
				}
			}
			else {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NOT_FOUND_KEY, automatonName));
			}
		}
	}

}
