/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.cli;

import java.io.File;
import java.util.List;

import de.uni_due.inf.ti.raven.actions.ReachabilityAction;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * Console command to handle the execution of reachability algorithms.
 * @author Christoph Blume
 * @version $Revision: 1.27 $
 *
 */
public final class ReachabilityCommand extends Command {

	private static final File REACHABILITY_DOCUMENTATION_FILE = new File(
			RavenMain.DOCUMENTATION_FOLDER + "ReachabilityCommandDocumentation.txt");
	
	private static final String IS_TOKEN = Console.IS_TOKEN;
	private static final String SEP_TOKEN = Console.SEP_TOKEN;
	private static final String NAME_PATTERN = Console.NAME_PATTERN;

	private static final String REACHABILITY_KEY = Console.REACHABILITY_TOKEN;
	private static final String REACHABILITY_PARAMETER = "("+NAME_PATTERN+")" +IS_TOKEN+
			"\\(("+NAME_PATTERN+")" +SEP_TOKEN+ "("+NAME_PATTERN+")\\)";

	private static ReachabilityCommand command;

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the <tt>REACHABILITY</tt>-command.
	 * @return the <tt>REACHABILITY</tt>-command
	 */
	protected static ReachabilityCommand getCommand() {
		if(command==null) {
			command = new ReachabilityCommand();
		}

		return command;		
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates the <tt>REACHABILITY</tt>-command.
	 */
	private ReachabilityCommand() {
		super(REACHABILITY_KEY, ConsoleMessages.REACHABILITY_DESCRIPTION_KEY, 
				REACHABILITY_DOCUMENTATION_FILE);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public boolean matchesParameter(String line) {
		return line.matches(REACHABILITY_PARAMETER);
	}

	@Override
	public void execute(String parameter) {
		if(!matchesParameter(parameter)) {
			throw new IllegalArgumentException(Console.messages.getString(
					ConsoleMessages.MISSING_ILLEGAL_ARGUMENT_DESCRIPTION_KEY, REACHABILITY_KEY));
		}
		
		//Scan and extract automaton properties from input string
		List<String> list = getProperties(parameter, REACHABILITY_PARAMETER);
		String bddName = list.get(0);
		String automatonName = list.get(1);
		String decompositionName = list.get(2);
		ReachabilityAction action;
		if(!Console.repo.contains(automatonName) || !Console.repo.isAutomaton(automatonName)) {
			throw new IllegalArgumentException(Console.messages.getString(
					ConsoleMessages.AUTOMATON_NOT_FOUND_KEY, automatonName));
		}
		else if(!Console.repo.contains(decompositionName) || !Console.repo.isDecomposition(decompositionName)) {
			throw new IllegalArgumentException(Console.messages.getString(
					ConsoleMessages.DECOMPOSITION_NOT_FOUND_KEY, decompositionName));
		}
		else {
			Automaton automaton = (Automaton) Console.repo.get(automatonName);
			CospanDecomposition decomposition = (CospanDecomposition) Console.repo.get(decompositionName);
			//Create and execute new automaton command
			action = new ReachabilityAction(automaton, decomposition);
			action.execute();
			Repository.getRepository().put(bddName, action.getResult());
		}
		Console.out.println("Compute reachable states for automaton '" + automatonName + "' " +
				"and decomposition '" + decompositionName + "'.");

		action.execute();
		
		//Print elapsed runtime (if time mode is activated)
		printRuntime(action.getElapsedTime());
	}

}
