/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.cli;

import java.io.File;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import de.uni_due.inf.ti.raven.data.Data;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.GraphDecomposition;
import de.uni_due.inf.ti.raven.data.Signature;
import de.uni_due.inf.ti.raven.io.TextTable;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * Console command to handle the presentation of the repository.
 * @author Christoph Blume
 * @version $Revision: 1.21 $
 *
 */
public final class RepositoryCommand extends Command {

	private static final File REPOSITORY_DOCUMENTATION_FILE = new File(
			RavenMain.DOCUMENTATION_FOLDER + "RepositoryCommandDocumentation.txt");
	
	private static final String REPOSITORY_KEY = Console.REPOSITORY_TOKEN;

	private static RepositoryCommand command;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the <tt>REPOSITORY</tt>-command.
	 * @return the <tt>REPOSITORY</tt>-command
	 */
	protected static RepositoryCommand getCommand() {
		if(command==null) {
			command = new RepositoryCommand();
		}

		return command;		
	}
	
	/**
	 * Returns a string indicating which type of data the object {@code data} is.
	 * @param data the object
	 * @return the type string
	 */
	private static String getTypeString(Data data) {
		if(data instanceof CospanDecomposition || 
				data instanceof Graph ||
			data instanceof GraphDecomposition ||
				data instanceof Signature) { 
			return data.getClass().getSimpleName();
		}
		return data.toString();
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates the <tt>REPOSITORY</tt>-command.
	 */
	private RepositoryCommand() {
		super(REPOSITORY_KEY, ConsoleMessages.REPOSITORY_DESCRIPTION_KEY, 
				REPOSITORY_DOCUMENTATION_FILE);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public boolean matchesParameter(String line) {
		return (line==null || line.isEmpty());
	}

	@Override
	public void execute(String parameter) {
		if(!matchesParameter(parameter)) {
			throw new IllegalArgumentException(Console.messages.getString(
					ConsoleMessages.MISSING_ILLEGAL_ARGUMENT_DESCRIPTION_KEY, REPOSITORY_KEY));
		}
		
		TextTable table = new TextTable();
		table.addHeader(Console.messages.getString(
				ConsoleMessages.TABLE_HEADER_NAME_KEY));
		table.addHeader(Console.messages.getString(
				ConsoleMessages.TABLE_HEADER_TYPE_KEY));
		List<String> nameList = new LinkedList<>(Console.repo.getNames());
		Collections.sort(nameList);
		
		for(String name: nameList) {
			table.addContent(0, 0, name);
			table.addContent(0, 1, getTypeString(Console.repo.get(name)));
		}
		
		Console.out.print(table.toString());
	}

}
