/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.cli;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;

import de.uni_due.inf.ti.raven.io.CommandReader;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * Console command to handle the execution of command-line scripts.
 * @author Christoph Blume
 * @version $Revision: 1.21 $
 *
 */
public final class RunCommand extends Command {

	private static final File RUN_DOCUMENTATION_FILE = new File(
			RavenMain.DOCUMENTATION_FOLDER + "RunCommandDocumentation.txt");
	
	private static final String FILENAME_PATTERN = Console.FILENAME_PATTERN;
	private static final String RUN_KEY = Console.RUN_TOKEN;
	private static final String RUN_PARAMETER = FILENAME_PATTERN;
	private static RunCommand command;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the <tt>RUN</tt>-command.
	 * @return the <tt>RUN</tt>-command
	 */
	protected static RunCommand getCommand() {
		if(command==null) {
			command = new RunCommand();
		}

		return command;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates the <tt>RUN</tt>-command.
	 */
	private RunCommand() {
		super(RUN_KEY, ConsoleMessages.RUN_DESCRIPTION_KEY, 
				RUN_DOCUMENTATION_FILE);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public boolean matchesParameter(String line) {
		return line.matches(RUN_PARAMETER);
	}

	@Override
	public void execute(String parameter) {
		if(!matchesParameter(parameter)) {
			throw new IllegalArgumentException(Console.messages.getString(
					ConsoleMessages.MISSING_ILLEGAL_ARGUMENT_DESCRIPTION_KEY, RUN_KEY));
		}
		if(!Console.isInteractive) {
			throw new IllegalArgumentException(Console.messages.getString(
					ConsoleMessages.RUNNING_SCRIPTS_IN_SCRIPT_MODE_KEY));
		}

		BufferedReader reader = null;
		String filename = RavenMain.SCRIPT_DIR+parameter;
		final BufferedReader tempIn = Console.in;
		try {
			reader = new CommandReader(filename) {
				
				@Override
				public String readLine() throws IOException {
					String line = super.readLine();
					if(line==null) {
						try {
							close();
						} catch (IOException e) {} //Fail silently
						line = "";
					}
					
					return line;
					
				}
				
				@Override
				public void close() throws IOException {
					super.close();
					Console.isInteractive = true;
					Console.in = tempIn;
				}
				
			};

			Console.isInteractive = false;
			Console.in = reader;

		} catch (FileNotFoundException e) {
			Console.out.println(Console.messages.getString(
					ConsoleMessages.FILE_NOT_FOUND_KEY, filename));
		}
	}

}
