/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.cli;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.List;
import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.actions.Action;
import de.uni_due.inf.ti.raven.actions.SaveAction;
import de.uni_due.inf.ti.raven.actions.SaveAction.AutomatonPersistenceType;
import de.uni_due.inf.ti.raven.actions.SaveAction.GraphPersistenceType;
import de.uni_due.inf.ti.raven.data.Cospan;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.Signature;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * Console command to handle the storage of data.
 * 
 * @author Christoph Blume, Laura Steinert 
 * @version $Revision: 1.37 $
 * 
 */
public final class SaveCommand extends Command {

	private static final File SAVE_DOCUMENTATION_FILE = new File(
			RavenMain.DOCUMENTATION_FOLDER + "SaveCommandDocumentation.txt");

	private static final String NO_TOKEN = "n";
	private static final String SEP_TOKEN = Console.SEP_TOKEN;
	private static final String AUTOMATON_TOKEN = Console.AUTOMATON_TOKEN;
	private static final String COSPAN_TOKEN = Console.COSPAN_TOKEN;
	private static final String DECOMPOSITION_TOKEN = Console.DECOMPOSITION_TOKEN;
	private static final String GRAPH_TOKEN = Console.GRAPH_TOKEN;
	private static final String SAVE_KEY = Console.SAVE_TOKEN;
	private static final String SIGNATURE_TOKEN = Console.SIGNATURE_TOKEN;
	private static final String YES_TOKEN = "y";

	private static final String NAME_PATTERN = Console.NAME_PATTERN;
	private static final String FILENAME_PATTERN = Console.FILENAME_PATTERN;
	private static final String SAVE_PARAMETER = "\\(("+NAME_PATTERN+")" + SEP_TOKEN + "("+FILENAME_PATTERN+")\\)";
	private static final String SAVE_AUTOMATON_PARAMETER = AUTOMATON_TOKEN + "\\(("+NAME_PATTERN+")" + SEP_TOKEN + "("+FILENAME_PATTERN+")\\)";
	private static final String SAVE_COSPAN_PARAMETER = COSPAN_TOKEN + "\\((" + NAME_PATTERN + ")" + SEP_TOKEN + "(" + FILENAME_PATTERN + ")\\)";
	private static final String SAVE_DECOMPOSITION_PARAMETER = DECOMPOSITION_TOKEN + "\\(("+NAME_PATTERN+")" + SEP_TOKEN + "("+FILENAME_PATTERN+")\\)";
	private static final String SAVE_GRAPH_PARAMETER = GRAPH_TOKEN + "\\(("+NAME_PATTERN + ")" + SEP_TOKEN + "("+FILENAME_PATTERN+")\\)";
	private static final String SAVE_SIGNATURE_PARAMETER = SIGNATURE_TOKEN + "\\(("+NAME_PATTERN + ")" + SEP_TOKEN + "("+FILENAME_PATTERN+")\\)";

	private static SaveCommand command;

	private Logger logger;

	/*------------------------------------------------------------------------------------------*/
	/* 		Static Methods 																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the <tt>SAVE</tt>-command.
	 * 
	 * @return the <tt>SAVE</tt>-command
	 */
	protected static SaveCommand getCommand() {
		if (command == null) {
			command = new SaveCommand();
		}

		return command;
	}

	/**
	 * Checks whether file given by the filename {@code filename} is already existing. If it exists,
	 * the user is asked to overwrite the file.
	 * @param filename the filename indicating the file to be checked
	 * @return {@code true} if the file is either not existing or the the user permits to overwrite
	 * the file, {@code false} otherwise
	 * @throws IOException If an I/O error occurs while the user is asked to overwrite the file
	 */
	private static boolean checkAndOverwriteFile(String filename) throws IOException {
		File file = new File(filename);
		if (file.exists()) {
			Console.out.println();
			Console.out.print(Console.messages.getString(
					ConsoleMessages.FILE_ALREADY_EXISTING_KEY, file.getAbsolutePath()));
			Console.out.print(Console.messages.getString(
					ConsoleMessages.OVERWRITE_FILE_KEY, YES_TOKEN, NO_TOKEN));
			String line;
			while ((line = Console.in.readLine()) != null) {
				if (line.isEmpty()) {
					continue;
				} 
				else if (line.equals(YES_TOKEN)) {
					return true;
				}
				else if (line.equals(NO_TOKEN)) {
					return false;
				}
				else {
					Console.out.print(Console.messages.getString(
							ConsoleMessages.OVERWRITE_FILE_KEY, YES_TOKEN, NO_TOKEN));
				}
			}
		}

		return true;
	}

	/*------------------------------------------------------------------------------------------*/
	/* 		Constructors 																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates the <tt>SAVE</tt>-command.
	 */
	private SaveCommand() {
		super(SAVE_KEY, ConsoleMessages.SAVE_DESCRIPTION_KEY,
				SAVE_DOCUMENTATION_FILE);

		// Register logger
		logger = Logger.getLogger(getClass().getName());
	}

	/*------------------------------------------------------------------------------------------*/
	/* 		Public Methods 																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public boolean matchesParameter(String line) {
		return line.matches(SAVE_PARAMETER)
				|| line.matches(SAVE_AUTOMATON_PARAMETER)
				|| line.matches(SAVE_COSPAN_PARAMETER) 
				|| line.matches(SAVE_DECOMPOSITION_PARAMETER) 
				|| line.matches(SAVE_GRAPH_PARAMETER)
				|| line.matches(SAVE_SIGNATURE_PARAMETER);
	}

	@Override
	public void execute(String parameter) {
		if (!matchesParameter(parameter)) {
			throw new IllegalArgumentException(Console.messages.getString(
					ConsoleMessages.MISSING_ILLEGAL_ARGUMENT_DESCRIPTION_KEY, SAVE_KEY));
		}

		if (parameter.matches(SAVE_AUTOMATON_PARAMETER)) {
			List<String> list = getProperties(parameter, SAVE_AUTOMATON_PARAMETER);
			String name = list.get(0);
			String filename = list.get(1);

			if (!Console.repo.contains(name)) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.DATA_NOT_FOUND_EXCEPTION_KEY, SAVE_KEY, name));
			}

			if (!Console.repo.isAutomaton(name)) {
				Console.out.println(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NOT_FOUND_KEY, name));
			}
			else {
				Automaton automaton = (Automaton) Console.repo.get(name);

				if(!filename.startsWith(RavenMain.AUTOMATON_SAVE_DIR)) {
					filename = RavenMain.AUTOMATON_SAVE_DIR + filename;
				}

				if(!filename.endsWith("."+RavenMain.AUTOMATON_FILE_EXTENSION)) {
					filename = filename + "." + RavenMain.AUTOMATON_FILE_EXTENSION;
				}

				try {
					if(!checkAndOverwriteFile(filename)) {
						return;
					}

					Action action = new SaveAction(
							AutomatonPersistenceType.SERIALIZED_FORMAT, automaton, filename);
					action.execute();
					Console.out.println(Console.messages.getString(
							ConsoleMessages.AUTOMATON_SAVED_KEY, name, filename));
				} catch (FileNotFoundException e) {
					Console.out.println(Console.messages.getString(
							ConsoleMessages.FILE_NOT_FOUND_KEY, filename));
				} catch (IOException e) {
					logger.throwing(getClass().getName(), "execute", e);
					Console.out.println(Console.messages.getString(
							ConsoleMessages.SAVING_ERROR_OCCURED));
				}
			}
		}
		else if (parameter.matches(SAVE_COSPAN_PARAMETER)) {
			List<String> list = getProperties(parameter, SAVE_COSPAN_PARAMETER);
			String name = list.get(0);
			String filename = list.get(1);

			if (!Console.repo.contains(name)) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.DATA_NOT_FOUND_EXCEPTION_KEY, SAVE_KEY, name));
			}

			if (!Console.repo.isCospan(name)) {
				Console.out.println(Console.messages.getString(
						ConsoleMessages.COSPAN_NOT_FOUND_KEY, name));
			}
			else {
				Cospan cospan = (Cospan) Console.repo.get(name);

				if(!filename.startsWith(RavenMain.COSPAN_SAVE_DIR)) {
					filename = RavenMain.COSPAN_SAVE_DIR + filename;
				}

				if(!filename.endsWith("."+RavenMain.COSPAN_GXL_FILE_EXTENSION)) {
					filename += "." + RavenMain.COSPAN_GXL_FILE_EXTENSION;
				}

				try {
					if(!checkAndOverwriteFile(filename)) {
						return;
					}

					Action action = new SaveAction(cospan, filename);
					action.execute();
					Console.out.println(Console.messages.getString(
							ConsoleMessages.COSPAN_SAVED_KEY, name, filename));
				} catch (FileNotFoundException e) {
					Console.out.println(Console.messages.getString(
							ConsoleMessages.FILE_NOT_FOUND_KEY, filename));
				} catch (IOException e) {
					logger.throwing(getClass().getName(), "execute", e);
					Console.out.println(Console.messages.getString(
							ConsoleMessages.SAVING_ERROR_OCCURED));
				}

			}
		}
		else if (parameter.matches(SAVE_DECOMPOSITION_PARAMETER)) {
			List<String> list = getProperties(parameter, SAVE_DECOMPOSITION_PARAMETER);
			String name = list.get(0);
			String filename = list.get(1);

			if (!Console.repo.contains(name)) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.DATA_NOT_FOUND_EXCEPTION_KEY, SAVE_KEY, name));
			}

			if (!Console.repo.isDecomposition(name)) {
				Console.out.println(Console.messages.getString(
						ConsoleMessages.DECOMPOSITION_NOT_FOUND_KEY, name));
			} else {
				CospanDecomposition decomposition = (CospanDecomposition) Console.repo.get(name);

				if(!filename.startsWith(RavenMain.DECOMPOSITION_SAVE_DIR)) {
					filename = RavenMain.DECOMPOSITION_SAVE_DIR + filename;
				}

				if(!filename.endsWith("."+RavenMain.DECOMPOSITION_FILE_EXTENSION)) {
					filename += "." + RavenMain.DECOMPOSITION_FILE_EXTENSION;
				}

				try {
					if(!checkAndOverwriteFile(filename)) {
						return;
					}

					Action action = new SaveAction(decomposition, filename);
					action.execute();
					Console.out.println(Console.messages.getString(
							ConsoleMessages.DECOMPOSITION_SAVED_KEY, name, filename));
				} catch (FileNotFoundException e) {
					Console.out.println(Console.messages.getString(
							ConsoleMessages.FILE_NOT_FOUND_KEY, filename));
				} catch (IOException e) {
					logger.throwing(getClass().getName(), "execute", e);
					Console.out.println(Console.messages.getString(
							ConsoleMessages.SAVING_ERROR_OCCURED));
				}

			}
		} 
		else if (parameter.matches(SAVE_GRAPH_PARAMETER)) {
			List<String> list = getProperties(parameter, SAVE_GRAPH_PARAMETER);
			String name = list.get(0);
			String filename = list.get(1);

			if (!Console.repo.contains(name)) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.DATA_NOT_FOUND_EXCEPTION_KEY, SAVE_KEY, name));
			}

			if (!Console.repo.isGraph(name)) {
				Console.out.println(Console.messages.getString(
						ConsoleMessages.GRAPH_NOT_FOUND_KEY, name));
			} else {
				Graph graph = (Graph) Console.repo.get(name);

				if(!filename.startsWith(RavenMain.GRAPH_SAVE_DIR)) {
					filename = RavenMain.GRAPH_SAVE_DIR + filename;
				}

				if(!filename.endsWith("."+RavenMain.GRAPH_GXL_FILE_EXTENSION)) {
					filename += "." + RavenMain.GRAPH_GXL_FILE_EXTENSION;
				}

				try {
					if(!checkAndOverwriteFile(filename)) {
						return;
					}

					Action action = new SaveAction(GraphPersistenceType.GXL_FORMAT, graph, filename);
					action.execute();
					Console.out.println(Console.messages.getString(
							ConsoleMessages.GRAPH_SAVED_KEY, name, filename));
				} catch (FileNotFoundException e) {
					Console.out.println(Console.messages.getString(
							ConsoleMessages.FILE_NOT_FOUND_KEY, filename));
				} catch (IOException e) {
					logger.throwing(getClass().getName(), "execute", e);
					Console.out.println(Console.messages.getString(
							ConsoleMessages.SAVING_ERROR_OCCURED));
				}
			}
		}
		else if (parameter.matches(SAVE_SIGNATURE_PARAMETER)) {
			List<String> list = getProperties(parameter, SAVE_SIGNATURE_PARAMETER);
			String name = list.get(0);
			String filename = list.get(1);

			if (!Console.repo.contains(name)) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.DATA_NOT_FOUND_EXCEPTION_KEY, SAVE_KEY, name));
			}

			if (!Console.repo.isSignature(name)) {
				Console.out.println(Console.messages.getString(
						ConsoleMessages.SIGNATURE_NOT_FOUND_KEY, name));
			} else {
				Signature signature = (Signature) Console.repo.get(name);

				if(!filename.startsWith(RavenMain.SIGNATURE_SAVE_DIR)) {
					filename = RavenMain.SIGNATURE_SAVE_DIR + filename;
				}

				if(!filename.endsWith("."+RavenMain.SIGNATURE_FILE_EXTENSION)) {
					filename += "." + RavenMain.SIGNATURE_FILE_EXTENSION;
				}

				try {
					if(!checkAndOverwriteFile(filename)) {
						return;
					}

					Action action = new SaveAction(signature, filename);
					action.execute();
					Console.out.println(Console.messages.getString(
							ConsoleMessages.SIGNATURE_SAVED_KEY, name, filename));
				} catch (FileNotFoundException e) {
					Console.out.println(Console.messages.getString(
							ConsoleMessages.FILE_NOT_FOUND_KEY, filename));
				} catch (IOException e) {
					logger.throwing(getClass().getName(), "execute", e);
					Console.out.println(Console.messages.getString(
							ConsoleMessages.SAVING_ERROR_OCCURED));
				}

			}
		} 
		else {
			// Extract parameter from input string
			List<String> list = getProperties(parameter, SAVE_PARAMETER);
			String name = list.get(0);
			String filename = list.get(1);

			if (!Console.repo.contains(name)) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.DATA_NOT_FOUND_EXCEPTION_KEY, SAVE_KEY, name));
			}

			try {
				Action action = null;
				String resultString = null;
				if(!checkAndOverwriteFile(filename)) {
					return;
				}

				if (Console.repo.isAutomaton(name)) {
					action = new SaveAction(AutomatonPersistenceType.PLAIN_FORMAT,
							(Automaton) Console.repo.get(name), filename);
					resultString = Console.messages.getString(
							ConsoleMessages.AUTOMATON_SAVED_KEY, name, filename);
				}
				else if (Console.repo.isCospan(name)) {
					action = new SaveAction((Cospan) Console.repo.get(name), filename);
					resultString = Console.messages.getString(
							ConsoleMessages.COSPAN_SAVED_KEY, name, filename);
				}
				else if (Console.repo.isDecomposition(name)) {
					action = new SaveAction((CospanDecomposition) Console.repo.get(name), filename);
					resultString = Console.messages.getString(
							ConsoleMessages.DECOMPOSITION_SAVED_KEY, name, filename);
				}
				else if (Console.repo.isGraph(name)) {
					action = new SaveAction(GraphPersistenceType.GXL_FORMAT,
							(Graph) Console.repo.get(name), filename);
					resultString = Console.messages.getString(
							ConsoleMessages.GRAPH_SAVED_KEY, name, filename);
				}
				else if (Console.repo.isSignature(name)) {
					action = new SaveAction((Signature) Console.repo.get(name), filename);
					resultString = Console.messages.getString(
							ConsoleMessages.SIGNATURE_SAVED_KEY, name, filename);
				}
				else {
					Console.out.println(Console.messages.getString(
							ConsoleMessages.DATA_CAN_NOT_BE_SAVED_KEY, name));
				}

				action.execute();
				Console.out.println(resultString);
			} catch (IOException e) {
				Console.out.println(Console.messages.getString(ConsoleMessages.SAVING_ERROR_OCCURED, filename));
				logger.throwing(getClass().getName(), "execute", e);
			}
		}
	}

}
