/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.cli;

import java.io.File;
import java.util.List;

import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * Console command to handle the setting of command-line variables.
 * @author Christoph Blume
 * @version $Revision: 1.27 $
 *
 */
public final class SetCommand extends Command {

	private static final File SET_DOCUMENTATION_FILE = new File(
			RavenMain.DOCUMENTATION_FOLDER+"SetCommandDocumentation.txt");
	
	private static final String IS_TOKEN = Console.IS_TOKEN;
	private static final String ACTIVATE_PATTERN = Console.ACTIVATE_PATTERN;
	private static final String DEACTIVATE_PATTERN = Console.DEACTIVATE_PATTERN;
	private static final String SYSTEM_VARIABLE_TOKEN = Console.SYSTEM_VARIABLE_TOKEN;
	private static final String VARIABLE_PATTERN = Console.VARIABLE_PATTERN;
	private static final String VALUE_PATTERN = Console.VALUE_PATTERN;

	private static final String SET_KEY = Console.SET_TOKEN;
	private static final String SET_PARAMETER = "("+SYSTEM_VARIABLE_TOKEN+"?"+VARIABLE_PATTERN+")" + 
			IS_TOKEN + "("+VALUE_PATTERN+")";

	private static SetCommand command;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the <tt>SET</tt>-command.
	 * @return the <tt>SET</tt>-command
	 */
	protected static SetCommand getCommand() {
		if(command==null) {
			command = new SetCommand();
		}

		return command;		
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates the <tt>SET</tt>-command.
	 */
	private SetCommand() {
		super(SET_KEY, ConsoleMessages.SET_DESCRIPTION_KEY, 
				SET_DOCUMENTATION_FILE);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public boolean matchesParameter(String line) {
		return (line.isEmpty() || line.matches(SET_PARAMETER));
	}

	@Override
	public void execute(String parameter) {
		String activated = Console.messages.getString(ConsoleMessages.ACTIVATED_KEY);
		String deactivated = Console.messages.getString(ConsoleMessages.DEACTIVATED_KEY);
		
		if(!matchesParameter(parameter)) {
			throw new IllegalArgumentException(Console.messages.getString(
					ConsoleMessages.MISSING_ILLEGAL_ARGUMENT_DESCRIPTION_KEY, SET_KEY));
		}

		if(parameter.isEmpty()) {
			for(String property: RavenMain.getPropertyNames()) {				
				if(property.matches(RavenMain.PROPERTY_SLOPPY_MODE)) {
					Console.out.println(Console.messages.getString(ConsoleMessages.SLOPPY_MODE_SET_KEY, 
							(Console.sloppyActivated ? activated : deactivated)));
				}
				else if(property.matches(RavenMain.PROPERTY_TIME)) {
					Console.out.println(Console.messages.getString(ConsoleMessages.TIME_MODE_SET_KEY, 
							(Console.timeStatsActivated ? activated : deactivated)));
				}
				else if(property.matches(RavenMain.PROPERTY_VERBOSE)) {
					Console.out.println(Console.messages.getString(ConsoleMessages.VERBOSE_MODE_SET_KEY, 
							(Console.verboseActivated ? activated : deactivated)));
				}
				else {
					Console.out.println(property+" = "+RavenMain.getProperty(property));
				}
			}
		}
		else {
			//Scan and extract parameter from input string
			List<String> list = getProperties(parameter, SET_PARAMETER);
			String property = list.get(0);
			String value = list.get(1);
			value = value.matches(ACTIVATE_PATTERN) ? String.valueOf(true) : value;
			value = value.matches(DEACTIVATE_PATTERN) ? String.valueOf(false) : value;
			
			if(property.matches(RavenMain.PROPERTY_SLOPPY_MODE)) {
				Console.sloppyActivated = Boolean.parseBoolean(value);
				Console.out.println(Console.messages.getString(ConsoleMessages.SLOPPY_MODE_SET_KEY, 
						(Console.sloppyActivated ? activated : deactivated)));
			}
			else if(property.matches(RavenMain.PROPERTY_TIME)) {
				Console.timeStatsActivated = Boolean.parseBoolean(value);
				Console.out.println(Console.messages.getString(ConsoleMessages.TIME_MODE_SET_KEY, 
						(Console.timeStatsActivated ? activated : deactivated)));
			}
			else if(property.matches(RavenMain.PROPERTY_VERBOSE)) {
				Console.verboseActivated = Boolean.parseBoolean(value);
				Console.out.println(Console.messages.getString(ConsoleMessages.VERBOSE_MODE_SET_KEY, 
						(Console.verboseActivated ? activated : deactivated)));
			}
			
			RavenMain.setProperty(property, value);
		}
	}

}
