/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.cli;

import java.io.File;
import java.util.List;

import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.gui.GraphWindow;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * Console command to handle the presentation of data.
 * @author Christoph Blume
 * @version $Revision: 1.24 $
 *
 */
public final class ShowCommand extends Command {

	private static final File SHOW_DOCUMENTATION_FILE = new File(
			RavenMain.DOCUMENTATION_FOLDER+"ShowCommandDocumentation.txt");
	
	private static final String INITIAL_TOKEN = Console.INITIAL_TOKEN;
	private static final String FINAL_TOKEN = Console.FINAL_TOKEN;
	private static final String SEP_TOKEN = Console.SEP_TOKEN;
	private static final String AUTOMATON_RUN_TOKEN = Console.AUTOMATON_RUN_TOKEN;
	
	private static final String NAME_PATTERN = Console.NAME_PATTERN;
	
	private static final String SHOW_KEY = Console.SHOW_TOKEN;
	private static final String SHOW_REPOSITORY_PARAMETER = "("+NAME_PATTERN+")";
	private static final String SHOW_AUTOMATON_RUN_PARAMETER = AUTOMATON_RUN_TOKEN + 
			"\\(("+NAME_PATTERN+")" +SEP_TOKEN+ "("+NAME_PATTERN+")\\)";
	
	private static ShowCommand command;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the <tt>SHOW</tt>-command.
	 * @return the <tt>SHOW</tt>-command
	 */
	public static ShowCommand getCommand() {
		if(command==null) {
			command = new ShowCommand();
		}

		return command;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates the <tt>SHOW</tt>-command.
	 */
	private ShowCommand() {
		super(SHOW_KEY, ConsoleMessages.SHOW_DESCRIPTION_KEY, 
				SHOW_DOCUMENTATION_FILE);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public boolean matchesParameter(String line) {
		return line.matches(SHOW_REPOSITORY_PARAMETER) ||
				line.matches(SHOW_AUTOMATON_RUN_PARAMETER);
	}

	@Override
	public void execute(String parameter) {
		if(!matchesParameter(parameter)) {
			throw new IllegalArgumentException(Console.messages.getString(
					ConsoleMessages.MISSING_ILLEGAL_ARGUMENT_DESCRIPTION_KEY, SHOW_KEY));
		}
		
		if(parameter.matches(SHOW_REPOSITORY_PARAMETER)) {
			//Scan and extract automaton properties from input string
			List<String> list = getProperties(parameter, SHOW_REPOSITORY_PARAMETER);
			String name = list.get(0);
			
			if(!Console.repo.contains(name)) {
				Console.out.println(Console.messages.getString(
						ConsoleMessages.DATA_NOT_FOUND_KEY, name));
			}
		
			if (Console.repo.isGraph(name)) {
				GraphWindow.getWindow().print((Graph) Console.repo.get(name));
			}
			else {
				Console.out.println(Console.messages.getString(
						ConsoleMessages.GRAPH_NOT_FOUND_KEY, name));
			}
		}
		else if(parameter.matches(SHOW_AUTOMATON_RUN_PARAMETER)) {
			//Scan and extract automaton properties from input string
			List<String> list = getProperties(parameter, SHOW_AUTOMATON_RUN_PARAMETER);
			String name = list.get(0);
			String states = list.get(1);
			
			if(!Console.repo.contains(name) || !Console.repo.isAutomaton(name)) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NOT_FOUND_KEY, name));
			}

			Automaton automaton = (Automaton) Console.repo.get(name);
			if(states.equals(INITIAL_TOKEN)) {
//				window.print(automaton.getInitialStates());
			}
			else if(states.equals(FINAL_TOKEN)) {
//				window.print(automaton.getFinalStates());
			}
			else {
				if(!Console.repo.contains(states) || !Console.repo.isDecomposition(states)) {
					throw new IllegalArgumentException(Console.messages.getString(
							ConsoleMessages.DECOMPOSITION_NOT_FOUND_KEY, states));
				}
				
				CospanDecomposition decomposition = (CospanDecomposition) Console.repo.get(states);
				GraphWindow.getWindow().showRun(automaton, decomposition);
			}
			
		}
		
	}

}
