/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.cli;

import java.io.File;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import de.uni_due.inf.ti.raven.actions.SignatureAction;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * Console command to handle the creation of signatures.
 * @author Christoph Blume
 * @version $Revision: 1.22 $
 *
 */
public final class SignatureCommand extends Command {

	private static final File SIGNATURE_DOCUMENTATION_FILE = new File(
			RavenMain.DOCUMENTATION_FOLDER+"SignatureCommandDocumentation.txt");

	private static final String ARITY_TOKEN = Console.ARITY_TOKEN;
	private static final String CONNECT_TOKEN = Console.CONNECT_TOKEN+Console.CONNECT_PARAMETER;
	private static final String FUSE_TOKEN = Console.FUSE_TOKEN;
	private static final String NAME_PATTERN = Console.NAME_PATTERN;
	private static final String PERM_TOKEN = Console.PERM_TOKEN;
	private static final String RES_TOKEN = Console.RES_TOKEN;
	private static final String SEP_TOKEN = Console.SEP_TOKEN;
	private static final String TRANS_TOKEN = Console.TRANS_TOKEN;
	private static final String VERTEX_TOKEN = Console.VERTEX_TOKEN;
//	private static final String LETTER_PATTERN = Console.OPERATION_PATTERN;
	private static final String LETTER_PATTERN = "(" + CONNECT_TOKEN + "|" + 
			FUSE_TOKEN + "|" + PERM_TOKEN + "|" + RES_TOKEN + "|" + TRANS_TOKEN + "|" + VERTEX_TOKEN + ")";

	private static final String SIGNATURE_KEY = Console.SIGNATURE_TOKEN;
	private static final String SIGNATURE_PARAMETER = "("+NAME_PATTERN+")\\s*\\{"+ LETTER_PATTERN + 
			"(" + SEP_TOKEN + LETTER_PATTERN + ")*\\}";

	private static SignatureCommand command;

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the <tt>SIGNATURE</tt>-command.
	 * @return the <tt>SIGNATURE</tt>-command
	 */
	protected static SignatureCommand getCommand() {
		if(command==null) {
			command = new SignatureCommand();
		}

		return command;		
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates the <tt>SIGNATURE</tt>-command.
	 */
	private SignatureCommand() {
		super(SIGNATURE_KEY, ConsoleMessages.SIGNATURE_DESCRIPTION_KEY, 
				SIGNATURE_DOCUMENTATION_FILE);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public boolean matchesParameter(String line) {
		return line.matches(SIGNATURE_PARAMETER);
	}

	@Override
	public void execute(String parameter) {
		if(!matchesParameter(parameter)) {
			throw new IllegalArgumentException(Console.messages.getString(
					ConsoleMessages.MISSING_ILLEGAL_ARGUMENT_DESCRIPTION_KEY, SIGNATURE_KEY));
		}

		//Extract parameter from input string
		List<String> list = getProperties(parameter, SIGNATURE_PARAMETER);
		String name = list.get(0);

		Set<Operation> operations = new HashSet<>();
		String letters = parameter.substring(parameter.indexOf("{")+1, parameter.indexOf("}"));
		for(String string: letters.split(SEP_TOKEN)) {				
			if(string.matches(CONNECT_TOKEN)) {
				String properties = string.substring(string.indexOf("(")+1, 
						string.indexOf(")"));
				String[] array = properties.split(ARITY_TOKEN);
				operations.add(Operation.getConnect(array[0], Integer.parseInt(array[1])));
			}
			else if(string.matches(FUSE_TOKEN)) {
				operations.add(Operation.getFuse());
			}
			else if(string.matches(PERM_TOKEN)) {
				operations.add(Operation.getPerm());
			}
			else if(string.matches(RES_TOKEN)) {
				operations.add(Operation.getRes());
			}
			else if(string.matches(TRANS_TOKEN)) {
				operations.add(Operation.getTrans());
			}
			else if(string.matches(VERTEX_TOKEN)) {
				operations.add(Operation.getVertex());		
			}
		}

		//Create and execute new signature command
		new SignatureAction(name, operations).execute();
		Console.out.println(Console.messages.getString(
				ConsoleMessages.SIGNATURE_ADDED_KEY, name));
	}

}
