/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.cli;

import java.io.File;
import java.util.List;

import de.uni_due.inf.ti.raven.actions.SimulationComputationAction;
import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * Console command to handle the execution of simulation relation computation algorithms.
 * @author Christoph Blume
 * @version $Revision: 1.30 $
 *
 */
public final class SimulationCommand extends Command {

	private static final File SIMULATION_DOCUMENTATION_FILE = new File(
			RavenMain.DOCUMENTATION_FOLDER+"SimulationCommandDocumentation.txt");
	
	private static final String IS_TOKEN = Console.IS_TOKEN;
	private static final String SEP_TOKEN = Console.SEP_TOKEN;
	private static final String NAME_PATTERN = Console.NAME_PATTERN;
	
	private static final String SIMULATION_KEY = Console.SIMULATION_TOKEN;
	private static final String SIMULATION_SINGLE_AUTOMATON_PARAMETER = "("+NAME_PATTERN+")"
			+ IS_TOKEN + "\\(("+NAME_PATTERN+")\\)";
	private static final String SIMULATION_TWO_AUTOMATA_PARAMETER = "("+NAME_PATTERN+")"
			+ IS_TOKEN + "\\(("+NAME_PATTERN+")" + SEP_TOKEN + "("+NAME_PATTERN+")\\)";

	private static SimulationCommand command;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the <tt>SIMULATION</tt>-command.
	 * @return the <tt>SIMULATION</tt>-command
	 */
	protected static SimulationCommand getCommand() {
		if(command==null) {
			command = new SimulationCommand();
		}

		return command;		
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates the <tt>SIMULATION</tt>-command.
	 */
	private SimulationCommand() {
		super(SIMULATION_KEY, ConsoleMessages.SIMULATION_DESCRIPTION_KEY, 
				SIMULATION_DOCUMENTATION_FILE);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public boolean matchesParameter(String line) {
		return line.matches(SIMULATION_SINGLE_AUTOMATON_PARAMETER) ||
				line.matches(SIMULATION_TWO_AUTOMATA_PARAMETER);
	}

	@Override
	public void execute(String parameter) {
		if(!matchesParameter(parameter)) {
			throw new IllegalArgumentException(
					Console.messages.getString(
							ConsoleMessages.MISSING_ILLEGAL_ARGUMENT_DESCRIPTION_KEY, SIMULATION_KEY));
		}
		
		List<String> list;
		String firstAutomatonName = null;
		String secondAutomatonName = null;
		String simulationName = null;
		SimulationComputationAction action = null;
		
		if(parameter.matches(SIMULATION_SINGLE_AUTOMATON_PARAMETER)) {
			//Scan and extract parameter from input string
			list = getProperties(parameter, SIMULATION_SINGLE_AUTOMATON_PARAMETER);
			simulationName = list.get(0);
			firstAutomatonName = list.get(1);
		}
		else if(parameter.matches(SIMULATION_TWO_AUTOMATA_PARAMETER)) {
			//Scan and extract parameter from input string
			list = getProperties(parameter, SIMULATION_TWO_AUTOMATA_PARAMETER);
			simulationName = list.get(0);
			firstAutomatonName = list.get(1);
			secondAutomatonName = list.get(2);
		}
		
		if(!Console.repo.isAutomaton(firstAutomatonName)) {
			throw new IllegalArgumentException(Console.messages.getString(
					ConsoleMessages.AUTOMATON_NOT_FOUND_KEY, firstAutomatonName));
		}

		Automaton firstAutomaton = (Automaton) Console.repo.get(firstAutomatonName);
		//Compute simulation relation
		Console.out.println("");
		if(parameter.matches(SIMULATION_SINGLE_AUTOMATON_PARAMETER)) {
			action = new SimulationComputationAction(firstAutomaton);
			
		}
		else if(parameter.matches(SIMULATION_TWO_AUTOMATA_PARAMETER)) {
			if(!Console.repo.isAutomaton(secondAutomatonName)) {
				throw new IllegalArgumentException(Console.messages.getString(
						ConsoleMessages.AUTOMATON_NOT_FOUND_KEY, secondAutomatonName));
			}
			
			Automaton secondAutomaton = (Automaton) Console.repo.get(secondAutomatonName);
			action = new SimulationComputationAction(firstAutomaton, secondAutomaton);
			
		}
//		Console.out.println("The first state does "+(cmd.getResult() ? "" : "not ")+"simulate " +
//				"the second state.");

		if(Console.progressThread!=null) {
			action.addComputationListener(Console.progressThread);
		}
		action.execute();
		Repository.getRepository().put(simulationName, action.getResult());
		
		//Print elapsed runtime (if time mode is activated)
		printRuntime(action.getElapsedTime());
	}

}
