/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.cli;

import java.io.File;
import java.util.List;

import de.uni_due.inf.ti.raven.actions.AntichainUniversalityAction;
import de.uni_due.inf.ti.raven.actions.SimulationAntichainUniversalityAction;
import de.uni_due.inf.ti.raven.algorithms.Algorithm.SearchType;
import de.uni_due.inf.ti.raven.data.CospanCounterExample;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * Console command to handle the execution of the universality check
 * algorithms.
 * 
 * @author Christoph Blume
 * @version $Revision: 1.30 $
 * 
 */
public final class UniversalityCommand extends Command {

	private static final File UNIVERSALITY_DOCUMENTATION_FILE = new File(
			RavenMain.DOCUMENTATION_FOLDER + "UniversalityCommandDocumentation.txt");

	private static final String BOOLEAN_PATTERN = Console.BOOLEAN_PATTERN;
	private static final String NAME_PATTERN = Console.NAME_PATTERN;
	private static final String ANTICHAIN_TOKEN = Console.ANTICHAIN_TOKEN;
	private static final String SIMULATION_TOKEN = Console.SIMULATION_TOKEN;
	private static final String BACKWARD_TOKEN = Console.BACKWARD_TOKEN;
	private static final String FORWARD_TOKEN = Console.FORWARD_TOKEN;
	private static final String SEP_TOKEN = Console.SEP_TOKEN;

	private static final String UNIVERSALITY_KEY = Console.UNIVERSALITY_TOKEN;
	private static final String UNIVERSALITY_DEFAULT_PARAMETER = "\\(("
			+ NAME_PATTERN + ")\\)";
	private static final String UNIVERSALITY_ANTICHAIN_PARAMETER = 
			ANTICHAIN_TOKEN + "\\(("+ FORWARD_TOKEN + "|" + BACKWARD_TOKEN+ ")" + SEP_TOKEN
			+ "(" + NAME_PATTERN + ")" + SEP_TOKEN + "(" + BOOLEAN_PATTERN + ")\\)";
	private static final String UNIVERSALITY_SIMULATION_ANTICHAIN_PARAMETER = 
			SIMULATION_TOKEN + "\\((" + FORWARD_TOKEN + "|" + BACKWARD_TOKEN + ")" + SEP_TOKEN
			+ "(" + NAME_PATTERN + ")" + SEP_TOKEN + "(" + BOOLEAN_PATTERN + ")\\)";
	private static UniversalityCommand command;

	/*------------------------------------------------------------------------------------------*/
	/* 		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the <tt>UNIVERSALITY</tt>-command.
	 * 
	 * @return the <tt>UNIVERSALITY</tt>-command
	 */
	protected static UniversalityCommand getCommand() {
		if (command == null) {
			command = new UniversalityCommand();
		}

		return command;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates the <tt>UNIVERSALITY</tt>-command.
	 */
	private UniversalityCommand() {
		super(UNIVERSALITY_KEY, ConsoleMessages.UNIVERSALITY_DESCRIPTION_KEY,
				UNIVERSALITY_DOCUMENTATION_FILE);
	}

	/*------------------------------------------------------------------------------------------*/
	/* 		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public boolean matchesParameter(String line) {
		return line.matches(UNIVERSALITY_DEFAULT_PARAMETER)
				|| line.matches(UNIVERSALITY_ANTICHAIN_PARAMETER)
				|| line.matches(UNIVERSALITY_SIMULATION_ANTICHAIN_PARAMETER);
	}

	@Override
	public void execute(String parameter) {
		if (!matchesParameter(parameter)) {
			throw new IllegalArgumentException(Console.messages.getString(
					ConsoleMessages.MISSING_ILLEGAL_ARGUMENT_DESCRIPTION_KEY,
					UNIVERSALITY_KEY));
		}

		List<String> list;
		String algorithmName = null;
		String automatonName = null;
		SearchType type = null;
		boolean computeCounterExample = false;
		CospanCounterExample counterExample;

		if (parameter.matches(UNIVERSALITY_DEFAULT_PARAMETER)) {
			algorithmName = ANTICHAIN_TOKEN;
			type = SearchType.forward;
			computeCounterExample = false;
			// Scan and extract parameter from input string
			list = getProperties(parameter, UNIVERSALITY_DEFAULT_PARAMETER);
			automatonName = list.get(0);
		} else if (parameter.matches(UNIVERSALITY_ANTICHAIN_PARAMETER)) {
			list = getProperties(parameter, UNIVERSALITY_ANTICHAIN_PARAMETER);
			algorithmName = ANTICHAIN_TOKEN;
			// Scan and extract parameter from input string
			type = SearchType.valueOf(list.get(0));
			automatonName = list.get(1);
			computeCounterExample = Boolean.parseBoolean(list.get(2));
		}
		else if (parameter.matches(UNIVERSALITY_SIMULATION_ANTICHAIN_PARAMETER)) {
			list = getProperties(parameter, UNIVERSALITY_SIMULATION_ANTICHAIN_PARAMETER);
			algorithmName = SIMULATION_TOKEN;
			// Scan and extract parameter from input string
			type = SearchType.valueOf(list.get(0));
			automatonName = list.get(1);
			computeCounterExample = Boolean.parseBoolean(list.get(2));
		}

		if (!Console.repo.isAutomaton(automatonName)
				&& !Console.repo.isAutomaton(automatonName)) {
			throw new IllegalArgumentException(Console.messages.getString(
					ConsoleMessages.AUTOMATON_NOT_FOUND_KEY, automatonName));
		}

		Automaton automaton = (Automaton) Console.repo.get(automatonName);

		if (algorithmName.equals(ANTICHAIN_TOKEN)) {
			Console.out.println("Checking the universality for the automaton '"
					+ automatonName + "' using " + "the " + type
					+ " antichain universality algorithm.");

			AntichainUniversalityAction cmd = new AntichainUniversalityAction(
					type, computeCounterExample, automaton);
			if (Console.progressThread != null) {
				cmd.addComputationListener(Console.progressThread);
			}
			cmd.execute();

			String resultString;
			if (cmd.getResult()) {
				resultString = Console.messages.getString(
						ConsoleMessages.UNIVERSALITY_DOES_HOLD, automatonName);
			} else {
				resultString = Console.messages.getString(
						ConsoleMessages.UNIVERSALITY_DOES_NOT_HOLD,
						automatonName);
			}

			Console.out.println(resultString);
			//Print the counterexample (if one is available)			
			if(computeCounterExample) {
				counterExample = cmd.getCounterExample();
				if(counterExample!=null) {
					printCounterExample(counterExample);
				}
			}
		} else if (algorithmName.equals(SIMULATION_TOKEN)) {
			Console.out.println("Checking the universality for the automaton '"
					+ automatonName + "' using " + "the " + type
					+ " simulation antichain universality algorithm.");

			SimulationAntichainUniversalityAction cmd = new SimulationAntichainUniversalityAction(
					type, computeCounterExample, automaton);
			if (Console.progressThread != null) {
				cmd.addComputationListener(Console.progressThread);
			}
			cmd.execute();

			String resultString;
			if (cmd.getResult()) {
				resultString = Console.messages.getString(
						ConsoleMessages.UNIVERSALITY_DOES_HOLD, automatonName);
			} else {
				resultString = Console.messages.getString(
						ConsoleMessages.UNIVERSALITY_DOES_NOT_HOLD,
						automatonName);
			}

			Console.out.println(resultString);
			//Print the counterexample (if one is available)			
			if(computeCounterExample) {
				counterExample = cmd.getCounterExample();
				if(counterExample!=null) {
					printCounterExample(counterExample);
				}
			}
		}
	}

	/*------------------------------------------------------------------------------------------*/
	/* 		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	private static void printCounterExample(CospanCounterExample counterExample) {
		int size = counterExample.getCospan().getLength();

		Console.out.println("The counterexample consists of the following operation(s):");
		int i = 1;
		int length = String.valueOf(size).length();
		for(Operation operation: counterExample.getCospan()) {
			StringBuilder output = new StringBuilder();
			for(int j=String.valueOf(i).length();j<length;j++) {
				output.append(" ");
			}
			output.append(i+": "+operation.toString());
			Console.out.println(output.toString());
			i++;
		}

	}

}
