/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.cli;

import java.io.File;
import java.io.IOException;
import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.main.RavenMain;
import de.uni_due.inf.ti.raven.main.RavenUpdater;

/**
 * Handles the update process of Raven, i.e. looking whether new updates are available
 * and if so coordinating the download.
 * @author Christoph Blume
 * @version $Revision: 1.9 $
 *
 */
public class UpdateCommand extends Command {

	private static final File UPDATE_DOCUMENTATION_FILE = 
			new File(RavenMain.DOCUMENTATION_FOLDER+"UpdateCommandDocumentation.txt");

	private static final String UPDATE_KEY = Console.UPDATE_TOKEN;
	private static final String YES_TOKEN = Console.YES_TOKEN;
	private static final String NO_TOKEN = Console.NO_TOKEN;

	private static UpdateCommand command;

	private Logger logger;

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the <tt>UPDATE</tt>-command.
	 * @return the <tt>UPDATE</tt>-command
	 */
	protected static UpdateCommand getCommand() {
		if(command==null) {
			command = new UpdateCommand();
		}

		return command;		
	}


	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates the <tt>UPDATE</tt>-command.
	 */
	private UpdateCommand() {
		super(UPDATE_KEY, ConsoleMessages.UPDATE_DESCRIPTION_KEY, 
				UPDATE_DOCUMENTATION_FILE);

		logger = Logger.getLogger(getClass().getName());
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public boolean matchesParameter(String line) {
		return (line==null || line.isEmpty());
	}


	@Override
	public void execute(String parameter) {
		//The update process should only be initiated by the user himself/herself
		if(!Console.isInteractive) {
			return;
		}
		
		RavenMain.restart();
		
		//Look for updates (if updates are available the update-object is not null
		int updateStatus = RavenUpdater.updateAvailable();

		if(updateStatus==RavenUpdater.STATUS_CODE_UPDATE_AVAILABLE) {
			Console.out.println(Console.messages.getString(ConsoleMessages.UPDATE_AVAILABLE, 
					YES_TOKEN, NO_TOKEN));

			boolean performUpdate = false;
			try {
				while(true) {
					String line = Console.in.readLine();
					if(line.equals(YES_TOKEN)) {
						performUpdate = true;
						break;
					}
					else if(line.equals(NO_TOKEN)) {
						break;
					}
					else {
						Console.out.println(Console.messages.getString(
								ConsoleMessages.YES_OR_NO_OPTION, YES_TOKEN, NO_TOKEN));
					}
				} 
			} catch (IOException e) {
				logger.throwing(getClass().getName(), "execute", e);
				Console.out.println(ConsoleMessages.EXCEPTION_DURING_UPDATE_KEY);
				return;
			}

			//If the user wants to update Raven
			if(performUpdate) {
				//Try to download and install the update
				int status = RavenUpdater.downloadUpdate();
				if(status==RavenUpdater.STATUS_CODE_SUCCESS) {
					//If the update has been successfully installed, try to restart Raven
					Console.out.println(Console.messages.getString(
							ConsoleMessages.UPDATE_SUCCESSFUL));
					RavenMain.restart();
				}
				else {
					String message;
					switch (status) {
					case RavenUpdater.STATUS_CODE_DOWNLOAD_CORRUPTED:
						message = ConsoleMessages.UPDATE_DOWNLOAD_WAS_CORRUPTED_KEY;
						break;

					case RavenUpdater.STATUS_CODE_UPDATE_FILES_NOT_FOUND:
						message = ConsoleMessages.UPDATE_FILES_NOT_FOUND_KEY;
						break;

					case RavenUpdater.STATUS_CODE_IOERROR_OCCURRED:
						message = ConsoleMessages.UPDATE_IOERROR_OCCURRED_KEY;
						break;

					default:
						message = ConsoleMessages.UPDATE_FAILED_KEY;
						break;
					}
					//If the update could not be installed successfully, inform the user
					//about the reason
					Console.out.println(Console.messages.getString(message));
				}
			}
		}
		else {
			String message;
			switch (updateStatus) {
			case RavenUpdater.STATUS_CODE_NO_UPDATE_AVAILABLE:
				message = Console.messages.getString(ConsoleMessages.NO_UPDATE_AVAILABLE_KEY);
				break;

			case RavenUpdater.STATUS_CODE_UPDATE_SCHEMA_FILE_NOT_FOUND:
				message = Console.messages.getString(ConsoleMessages.UPDATE_SCHEMA_FILE_NOT_FOUND_KEY, 
						RavenMain.RAVEN_UPDATE_SCHEMA_FILENAME);
				break;
				
			case RavenUpdater.STATUS_CODE_UPDATE_SCHEMA_FILE_CORRUPTED:
				message = Console.messages.getString(ConsoleMessages.UPDATE_SCHEMA_FILE_CORRUPTED_KEY, 
						RavenMain.RAVEN_UPDATE_SCHEMA_FILENAME);
				break;
				
			case RavenUpdater.STATUS_CODE_BUILDING_PARSER_FAILED:
				message = Console.messages.getString(ConsoleMessages.UPDATE_BUILD_PARSER_FAILED_KEY);
				break;
				
			case RavenUpdater.STATUS_CODE_UPDATE_LOCATION_INVALID:
				message = Console.messages.getString(ConsoleMessages.UPDATE_LOCATION_INVALID_KEY);
				break;
				
			case RavenUpdater.STATUS_CODE_IOERROR_OCCURRED:
				message = Console.messages.getString(ConsoleMessages.UPDATE_IOERROR_OCCURRED_KEY);
				break;
				
			default:
				message = Console.messages.getString(ConsoleMessages.EXCEPTION_DURING_UPDATE_KEY);
				break;
			}
			//If no update could be found or an error occurred, inform the user
			//about the reason
			Console.out.println(message);
		}


	}

}
