/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.data;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class implements a <emph>cospan</emph> (of graphs), i.e. a pair of two 
 * (graph) morphisms with the same co-domain.
 * @author Christoph Blume, Dennis Nolte
 * @version $Revision: 1.15 $
 */
public class Cospan implements Data, Serializable {

	private static final long serialVersionUID = -6829214810891612111L;

	/**
	 * Enumeration of the different types of implemented cospans.
	 * @author Christoph Blume, Dennis Nolte
	 * @version $Revision: 1.15 $
	 *
	 */
	public enum Type implements Data.Type {
		/** Cospan Type */
		cospan;
	}

	private Graph graph;
	private List<Node> innerInterface;
	private List<Node> outerInterface;
	private int nodeCounter;

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new (empty) cospan with the discrete graph with {@code width} nodes
	 * as inner and outer interface.
	 * @param initialWidth the initial width of the cospan and the inner and outer interface
	 */
	public Cospan(int initialWidth) {
		if(initialWidth<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.WIDTH_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		
		graph = new Graph("");
		
		innerInterface = new ArrayList<>();
		outerInterface = new ArrayList<>();
		for(int i=0;i<initialWidth;i++) {
			String name = "n"+nodeCounter++;
			graph.addNode(name);
			
			innerInterface.add(graph.getNode(name));
		}
	}

	/**
	 * Creates a new (empty) cospan with an empty left interface.
	 */
	public Cospan() {
		this(0);
	}

	/**
	 * Creates a new cospan by copying the cospan decomposition {@code decomposition} information.
	 * @param decomposition the cospan decomposition to be copied
	 */
	public Cospan(CospanDecomposition decomposition) {
		if(decomposition==null) {
			throw new NullPointerException(
					RavenExceptionMessages.DECOMPOSITION_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		this.graph = decomposition.getGraph();
		this.innerInterface = new LinkedList<>(decomposition.getInnerInterface());
		this.outerInterface = new LinkedList<>(decomposition.getOuterInterface());
		this.nodeCounter = decomposition.getNodeCounter();
	}
	
	/**
	 * Creates a new cospan consisting of the middle graph {@code graph} and the 
	 * two interfaces {@code innerInterface} and {@code outerInterface}. 
	 * @param graph the middle graph
	 * @param innerInterface the inner interface
	 * @param outerInterface the outer interface
	 */
	public Cospan(Graph graph, List<Node> innerInterface, List<Node> outerInterface) {
		if(graph == null) {
			throw new NullPointerException(
					RavenExceptionMessages.GRAPH_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(innerInterface == null) {
			throw new NullPointerException(
					RavenExceptionMessages.INNER_INTERFACE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(outerInterface == null) {
			throw new NullPointerException(
					RavenExceptionMessages.OUTER_INTERFACE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(!graph.getNodes().containsAll(innerInterface)) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_MUST_BE_CONTAINED_IN_GRAPH_EXCEPTION);
		}
		if(!graph.getNodes().containsAll(outerInterface)) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_MUST_BE_CONTAINED_IN_GRAPH_EXCEPTION);
		}
		
		this.graph = graph;		
		this.innerInterface = new LinkedList<>();
		this.outerInterface = new LinkedList<>();
		
		for(Node node: innerInterface) {
			if(graph.contains(node)) {
				this.innerInterface.add(node);	
			}
		}
		for(Node node: outerInterface) {
			if(graph.contains(node)) {
				this.outerInterface.add(node);
			}
		}
	}

	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the graph which can be obtained from this cospan decomposition.
	 * @return the graph of this cospan decomposition
	 */
	public Graph getGraph() {
		return graph;
	}	

	/**
	 * Returns the size of the inner interface (of the cospan) of this decomposition.
	 * @return the inner interface size
	 */
	public int getInnerInterfaceSize() {
		return innerInterface.size();
	}

	/**
	 * Returns the size of the outer interface (of the cospan) of this decomposition.
	 * @return the outer interface size
	 */
	public int getOuterInterfaceSize() {
		return outerInterface.size();
	}
	
	
	/**
	 * Returns a list of the names of the nodes in the inner interface.
	 * @return the list of node names
	 */
	@Deprecated
	public List<String> getInnerInterfaceNodeNames() {
		List<String> names = new LinkedList<>();
		for(Node node: innerInterface) {
			names.add(node.getName());
		}

		return names;
	}

	/**
	 * Returns a list of the names of the nodes in the outer interface.
	 * @return the list of node names
	 */
	@Deprecated
	public List<String> getOuterInterfaceNodeNames() {
		List<String> names = new LinkedList<>();
		for(Node node: outerInterface) {
			names.add(node.getName());
		}

		return names;
	}
	
	/**
	 * Returns a list of the nodes in the inner interface of this cospan.
	 * @return the inner interface
	 */
	public List<Node> getInnerInterface() {
		return new LinkedList<>(innerInterface);
	}

	/**
	 * Returns the {@code i}-th node of the inner interface of this cospan.
	 * @param i the position of the interface node
	 * @return the {@code i}-th inner interface node
	 * @throws IllegalArgumentException thrown if the position is out of range, 
	 * i.e. if the position is less than zero or greater than or equal to the 
	 * size of the inner interface. 
	 */
	public Node getInnerInterfaceNode(int i) {
		if(i<0) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}
		if(i>=innerInterface.size()) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}
		
		return innerInterface.get(i);
	}

	/**
	 * Returns a list of the nodes in the outer interface of this cospan.
	 * @return the outer interface
	 */
	public List<Node> getOuterInterface() {
		return new LinkedList<>(outerInterface);
	}

	/**
	 * Returns the {@code i}-th node of the outer interface of this cospan.
	 * @param i the position of the interface node
	 * @return the {@code i}-th outer interface node
	 * @throws IllegalArgumentException thrown if the position is out of range, 
	 * i.e. if the position is less than zero or greater than or equal to the 
	 * size of the outer interface. 
	 */
	public Node getOuterInterfaceNode(int i) {
		if(i<0) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}
		if(i>=outerInterface.size()) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}
		
		return outerInterface.get(i);
	}
	
	/**
	 * Returns {@code true} if and only if this cospan contains
	 * a node with the name {@code nodeName}.
	 * @param nodeName the name of the node
	 * @return {@code true} if a node of the given name has 
	 * been found, {@code false} otherwise
	 * @throws NullPointerException thrown if {@code nodeName} is null
	 */
	public boolean hasNode(String nodeName) {
		return graph.hasNode(nodeName);
	}

	/**
	 * Returns the node with the name {@code nodeName} or {@code null} if no 
	 * node with the given name is existing.
	 * @param nodeName the name of the node
	 * @return the node or {@code null} if no node has been found
	 * 
	 * @throws NullPointerException thrown if {@code nodeName} is {@code null}
	 * @throws IllegalArgumentException thrown if {@code nodeName} is empty
	 */
	public Node getNode(String nodeName) {
		return graph.getNode(nodeName);
	}
	
	/**
	 * Returns the set of nodes of this cospan.
	 * @return the set of nodes
	 */
	public Set<Node> getNodes() {
		return graph.getNodes();
	}
	
	/**
	 * Adds a new node with node name {@code nodeName} and empty label to the cospan.
	 * @param nodeName the name of the new node
	 * @throws NullPointerException thrown if {@code nodeName} is null
	 * @throws IllegalArgumentException thrown if {@code nodeName} is empty or if 
	 * a node with name {@code nodeName} is already existing
	 */
	public void addNode(String nodeName) {
		graph.addNode(nodeName);
	}

	/**
	 * Returns {@code true} if and only if this cospan contains
	 * an edge with the name {@code nodeName}.
	 * @param edgeName the name of the node
	 * @return {@code true} if an edge of the given name has 
	 * been found, {@code false} otherwise
	 * @throws NullPointerException thrown if {@code edgeName} is null
	 */
	public boolean hasEdge(String edgeName) {
		return graph.hasEdge(edgeName);
	}
	
	/**
	 * Returns the edge with the name {@code edgeName} or {@code null} if no 
	 * edge with the given name is existing.
	 * @param edgeName the name of the edge
	 * @return the node or {@code null} if no node has been found
	 * 
	 * @throws NullPointerException thrown if {@code nodeName} is {@code null}
	 * @throws IllegalArgumentException thrown if {@code edgeName} is empty
	 */
	public Edge getEdge(String edgeName) {
		return graph.getEdge(edgeName);
	}
	
	/**
	 * Returns the set of edges of this cospan.
	 * @return the set of edges
	 */
	public Set<Edge> getEdges() {
		return graph.getEdges();
	}
	
	/**
	 * Adds a labeled edge to this cospan.
	 * @param edgeName the name of the new edge
	 * @param label the label of the new edge
	 * @param incidentNodes the incident nodes of the new edge
	 * @throws NullPointerException thrown if {@code edgeName}, {@code edgeLabel} 
	 * or {@code incidentNodes} is null
	 * @throws IllegalArgumentException thrown if {@code edgeName} or {@code edgeLabel} 
	 * are empty, at least one node of {@code incidentNodes} is not contained in 
	 * this graph or if an edge with name {@code edgeName} is already existing 
	 */
	public void addEdge(String edgeName, String label, List<Node> incidentNodes) {
		graph.addEdge(edgeName, label, incidentNodes);
	}
	
	/**
	 * Adds an existing node of the cospan to the last position of the inner 
	 * interface.
	 * @param nodeName the name of the node
	 * @throws NullPointerException thrown if {@code nodeName} is null
	 * @throws IllegalArgumentException thrown if {@code nodeName} is empty or 
	 * if a node with name {@code nodeName} is not existing
	 * 
	 * @see Cospan#addToInnerInterface(String, int)
	 * @see Cospan#addToOuterInterface(String)
	 * @see Cospan#addToOuterInterface(String, int)
	 */
	public void addToInnerInterface(String nodeName) {
		if(nodeName==null) {
			throw new NullPointerException(RavenExceptionMessages.NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(nodeName.isEmpty()) {
			throw new IllegalArgumentException(RavenExceptionMessages.NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		if(!graph.hasNode(nodeName)) {
			throw new IllegalArgumentException(RavenExceptionMessages.NODE_NOT_EXISTING_EXCEPTION);
		}
		
		innerInterface.add(graph.getNode(nodeName));
	}
	
	/**
	 * Adds an existing node of the cospan to the specified position of the 
	 * inner interface. <em>Please note, that the inner interface size must have
	 * at least the size of {@code position}!</em> 
	 * @param nodeName the name of the node
	 * @param position the position of the node in the interface
	 * @throws NullPointerException thrown if {@code nodeName} is null
	 * @throws IllegalArgumentException thrown if {@code nodeName} is empty, 
	 * if a node with name {@code nodeName} is not existing or if {@code position}
	 * is out of range, i.e. position is less than zero or greater than the size 
	 * of the inner interface. 
	 * 
	 * @see Cospan#addToInnerInterface(String)
	 * @see Cospan#addToOuterInterface(String)
	 * @see Cospan#addToOuterInterface(String, int)
	 */
	public void addToInnerInterface(String nodeName, int position) {
		if(nodeName==null) {
			throw new NullPointerException(RavenExceptionMessages.NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(nodeName.isEmpty()) {
			throw new IllegalArgumentException(RavenExceptionMessages.NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		if(!graph.hasNode(nodeName)) {
			throw new IllegalArgumentException(RavenExceptionMessages.NODE_NOT_EXISTING_EXCEPTION);
		}
		if(position<0) {
			throw new IllegalArgumentException(RavenExceptionMessages.POSITION_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface.size()<position) {
			throw new IllegalArgumentException(RavenExceptionMessages.POSITION_MUST_NOT_BE_LESS_OR_EQUAL_TO_INNER_INTERFACE_SIZE_EXCEPTION);
		}
		
		innerInterface.add(position, graph.getNode(nodeName));
		
	}

	/**
	 * Adds an existing node of the cospan to the last position of the outer 
	 * interface.
	 * @param nodeName the name of the node
	 * @throws NullPointerException thrown if {@code nodeName} is null
	 * @throws IllegalArgumentException thrown if {@code nodeName} is empty or 
	 * if a node with name {@code nodeName} is not existing
	 * 
	 * @see Cospan#addToInnerInterface(String)
	 * @see Cospan#addToInnerInterface(String, int)
	 * @see Cospan#addToOuterInterface(String, int)
	 */
	public void addToOuterInterface(String nodeName) {
		if(nodeName==null) {
			throw new NullPointerException(RavenExceptionMessages.NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(nodeName.isEmpty()) {
			throw new IllegalArgumentException(RavenExceptionMessages.NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		if(!graph.hasNode(nodeName)) {
			throw new IllegalArgumentException(RavenExceptionMessages.NODE_NOT_EXISTING_EXCEPTION);
		}
		
		outerInterface.add(graph.getNode(nodeName));
	}
	
	/**
	 * Adds an existing node of the cospan to the specified position of the 
	 * outer interface. <em>Please note, that the outer interface size must have
	 * at least the size of {@code position}!</em> 
	 * @param nodeName the name of the node
	 * @param position the position of the node in the interface
	 * @throws NullPointerException thrown if {@code nodeName} is null
	 * @throws IllegalArgumentException thrown if {@code nodeName} is empty, 
	 * if a node with name {@code nodeName} is not existing or if {@code position}
	 * is out of range, i.e. position is less than zero or greater than the size 
	 * of the outer interface.
	 * 
	 * @see Cospan#addToInnerInterface(String)
	 * @see Cospan#addToInnerInterface(String, int)
	 * @see Cospan#addToOuterInterface(String)
	 */
	public void addToOuterInterface(String nodeName, int position) {
		if(nodeName==null) {
			throw new NullPointerException(RavenExceptionMessages.NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(nodeName.isEmpty()) {
			throw new IllegalArgumentException(RavenExceptionMessages.NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		if(!graph.hasNode(nodeName)) {
			throw new IllegalArgumentException(RavenExceptionMessages.NODE_NOT_EXISTING_EXCEPTION);
		}
		if(position<0) {
			throw new IllegalArgumentException(RavenExceptionMessages.POSITION_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(outerInterface.size()<position) {
			throw new IllegalArgumentException(RavenExceptionMessages.POSITION_MUST_NOT_BE_LESS_OR_EQUAL_TO_OUTER_INTERFACE_SIZE_EXCEPTION);
		}
		
		outerInterface.add(position, graph.getNode(nodeName));
	}
	
	@Override
	public int hashCode() {
		return 13*innerInterface.hashCode()+31*graph.hashCode()+59*outerInterface.size();
	}

	@Override
	public boolean equals(Object o) {
		if(o==null) {
			return false;
		}
		else if(o==this) {
			return true;
		}
		else if(o.getClass().equals(getClass())) {
			Cospan cospan = (Cospan) o;
			return graph.equals(cospan.graph) && 
					innerInterface.equals(cospan.innerInterface) &&
					outerInterface.equals(cospan.outerInterface);
		}
		else {
			return false;
		}
	}

	@Override
	public Type getType() {
		return Cospan.Type.cospan;
	}

}
