/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.data;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import de.uni_due.inf.ti.raven.data.GraphDecomposition.Bag;
import de.uni_due.inf.ti.raven.data.Operation.Connect;
import de.uni_due.inf.ti.raven.data.Operation.Fuse;
import de.uni_due.inf.ti.raven.data.Operation.Perm;
import de.uni_due.inf.ti.raven.data.Operation.Res;
import de.uni_due.inf.ti.raven.data.Operation.Trans;
import de.uni_due.inf.ti.raven.data.Operation.Vertex;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class implements a sequence of composable cospans.
 * @author Christoph Blume
 * @version $Revision: 1.56 $
 *
 */
public class CospanDecomposition implements Decomposition<Operation>, Iterable<Operation> {

	private List<Operation> elements;
	private Graph graph;
	private Signature signature;
	private int maximumInterfaceSize;
	private List<Node> innerInterface;
	private List<Node> outerInterface;
	private int nodeCounter;

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns a cospan decomposition created from the graph decomposition {@code decomposition}
	 * @param decomposition the original path decomposition
	 * @return the created cospan decomposition
	 */
	public static CospanDecomposition getCospanDecomposition(GraphDecomposition decomposition) {
		if(decomposition instanceof PathDecomposition) {
			return getCospanDecomposition((PathDecomposition) decomposition);
		}
		else if(decomposition instanceof TreeDecomposition) {
			return getCospanDecomposition((TreeDecomposition) decomposition);
		}

		return null;
	}

	/**
	 * Returns a cospan decomposition created from the tree decomposition {@code decomposition}
	 * @param decomposition the original path decomposition
	 * @return the created cospan decomposition
	 */
	public static CospanDecomposition getCospanDecomposition(TreeDecomposition decomposition) {
		return getCospanDecomposition(PathDecomposition.getPathDecomposition(decomposition));
	}

	/**
	 * Returns a cospan decomposition created from the path decomposition {@code decomposition}
	 * @param pathDecomposition the original path decomposition
	 * @return the created cospan decomposition
	 */
	public static CospanDecomposition getCospanDecomposition(
			PathDecomposition pathDecomposition) {
		if(pathDecomposition==null) {
			throw new NullPointerException(
					RavenExceptionMessages.DECOMPOSITION_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		CospanDecomposition decomposition = new CospanDecomposition();
		Graph decompositionGraph = pathDecomposition.getGraph().clone();

		Set<Node> previousNodes = new HashSet<>();
		for(Bag bag: pathDecomposition) {
			previousNodes.removeAll(bag.getNodes());
			for(Node node: previousNodes) {
				decomposition.remove(decomposition.outerInterface.indexOf(node));
				decompositionGraph.removeNode(node);
			}

			Set<Node> nodeSet = new HashSet<>(bag.getNodes());
			nodeSet.removeAll(decomposition.outerInterface);
			for(Node node: nodeSet) {
				decomposition.addNode(node.getName());
			}

			//Add those edges which connect nodes contained in the previous bag and in the current bag
			for(Edge edge: decompositionGraph.getEdges()) {
				if(decomposition.outerInterface.containsAll(edge.getIncidentNodes())) {
					int offset = 0;

					List<Node> reversedList = new ArrayList<>(edge.getIncidentNodes());
					Collections.reverse(reversedList);
					for(Node node: reversedList) {
						int from = decomposition.outerInterface.indexOf(node);
						int to = decomposition.outerInterface.size()-offset-1;
						
						if(from!=to) {
							if(from<to) {
								decomposition.shiftRight(from, to-from);
							}
							else if(from>to) {
								decomposition.shiftLeft(from, from-to);
							}
						}
						offset++;
					}

					decomposition.addEdge(edge.getLabel(), edge.getArity());
					decompositionGraph.removeEdge(edge);
				}
			}

			previousNodes = new HashSet<>(bag.getNodes());
		}
		
		for(Node node: previousNodes) {
			decomposition.remove(decomposition.outerInterface.indexOf(node));
		}		
		
		return decomposition;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new (empty) cospan decomposition with the discrete graph with {@code width} nodes
	 * as inner and outer interface.
	 * @param initialWidth the initial width of the cospan decomposition and the inner and outer interface
	 */
	public CospanDecomposition(int initialWidth) {
		if(initialWidth<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.WIDTH_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		
		this.maximumInterfaceSize = initialWidth;
		elements = new LinkedList<>();
		signature = new Signature();
		graph = new Graph("");
		
		innerInterface = new ArrayList<>();
		outerInterface = new ArrayList<>();
		for(int i=0;i<initialWidth;i++) {
			String name = "n"+nodeCounter++;
			graph.addNode(name);
			
			innerInterface.add(graph.getNode(name));
			outerInterface.add(graph.getNode(name));
		}
	}

	/**
	 * Creates a new (empty) cospan decomposition with an empty left interface.
	 */
	public CospanDecomposition() {
		this(0);
	}

	/**
	 * Creates a new cospan decomposition by copying the cospan decomposition {@code decomposition}.
	 * @param decomposition the cospan decomposition to be copied
	 */
	public CospanDecomposition(CospanDecomposition decomposition) {
		if(decomposition==null) {
			throw new NullPointerException(
					RavenExceptionMessages.DECOMPOSITION_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		this.elements = new LinkedList<>(decomposition.elements);
		this.graph = decomposition.graph.clone();
		this.signature = new Signature(decomposition.signature);
		this.maximumInterfaceSize = decomposition.maximumInterfaceSize;
		this.innerInterface = new LinkedList<>(decomposition.innerInterface);
		this.outerInterface = new LinkedList<>(decomposition.outerInterface);
		this.nodeCounter = decomposition.nodeCounter;
	}

	/**
	 * Creates a cospan decomposition from the list {@code operationList} of operations with an empty 
	 * left interface.
	 * @param operationList the operation list
	 */
	public CospanDecomposition(List<Operation> operationList) {
		this(0, operationList);
	}
	
	/**
	 * Creates a cospan decomposition from the list {@code operationList} of operations with a left
	 * interface of width {@code width}.
	 * @param width the initial width of the cospan decomposition and the left interface
	 * @param operationList the operation list
	 */
	public CospanDecomposition(int width, List<Operation> operationList) {
		this(width);
		
		if(operationList==null) {
			throw new NullPointerException(
					RavenExceptionMessages.OPERATION_LIST_MUST_NOT_BE_NULL_EXCEPTION);
		}

		for(Operation operation: operationList) {
			add(operation);
		}
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Appends an operation at the end of this cospan decomposition.
	 * @param operation the new operation
	 * @throws UnsupportedOperationException thrown if the operation can not be applied to this decomposition 
	 */
	public void add(Operation operation) throws UnsupportedOperationException {
		if(operation instanceof Connect) {
			Connect connect = (Connect) operation;
			if(connect.getArity()>getOuterInterfaceSize()) {
				throw new UnsupportedOperationException(RavenExceptionMessages.ILLEGAL_ARITY_EXCEPTION);
			}

			addEdge(connect.getLabel(), connect.getArity());
		}
		else if(operation instanceof Fuse) {
			if(getOuterInterfaceSize()<2) {
				throw new UnsupportedOperationException(RavenExceptionMessages.INTERFACE_SIZE_MUST_BE_GREATER_THAN_TWO_EXCEPTION);
			}

			fuseNodes();
		}
		else if(operation instanceof Perm) {
			shift(1);
		}
		else if(operation instanceof Res) {
			if(getOuterInterfaceSize()<1) {
				throw new UnsupportedOperationException(RavenExceptionMessages.INTERFACE_SIZE_MUST_BE_GREATER_THAN_ONE_EXCEPTION);
			}
			remove(outerInterface.size()-1);
		}
		else if(operation instanceof Trans) {
			transpose();
		}
		else if(operation instanceof Vertex) {
			//TODO Substitute string literals by constants
			addNode("n"+nodeCounter++);
		}
		else {
			throw new UnsupportedOperationException(RavenExceptionMessages.OPERATION_CAN_NOT_BE_ADDED_EXCEPTION);
		}
	}

	/**
	 * Adds the cospan decomposition {@code decomposition} (seen as a list of operations) to this graph.
	 * @param decomposition the cospan decomposition to be applied
	 * @throws UnsupportedOperationException thrown if the at least one operation of the cospan 
	 * decomposition can not be applied to this graph
	 */
	public void add(CospanDecomposition decomposition) throws UnsupportedOperationException {
		for(Operation letter: decomposition) {
			add(letter);
		}
	}
	
	/**
	 * Appends all letters of the list {@code list} at the end of this cospan decomposition.
	 * @param list the list of letters to be added
	 */
	public void addAll(List<Operation> list) {
		for(Operation letter: list) {
			add(letter);
		}
	}

	/**
	 * Returns the list of operations used for this cospan decomposition
	 * @return the list of used operations
	 */
	public List<Operation> getOperations(){
		return new LinkedList<>(elements);
	}
	
	/**
	 * Returns the {@code index}-th operation of the cospan decomposition.
	 * @param index the index of the operation
	 * @return the {@code index}-th operation
	 */
	public Operation get(int index) {
		return elements.get(index);
	}

	/**
	 * Returns the underlying signature of this cospan decomposition
	 * @return the underlying signature
	 */
	public Signature getSignature() {
		return signature;
	}

	/**
	 * Returns the graph which can be obtained from this cospan decomposition.
	 * @return the graph of this cospan decomposition
	 */
	public Graph getGraph() {
		return graph;
	}

	@Override
	public Iterator<Operation> iterator() {
		return elements.listIterator();
	}

	@Override
	public int getWidth() {
		return maximumInterfaceSize;
	}
	

	/**
	 * Returns the size of the inner interface (of the cospan) of this decomposition.
	 * @return the inner interface size
	 */
	public int getInnerInterfaceSize() {
		return innerInterface.size();
	}

	/**
	 * Returns the size of the outer interface (of the cospan) of this decomposition.
	 * @return the outer interface size
	 */
	public int getOuterInterfaceSize() {
		return outerInterface.size();
	}
	
	/**
	 * Returns the node count of this decomposition.
	 * @return the node count
	 */
	public int getNodeCounter() {
		return nodeCounter;
	}
	
	@Override
	public boolean isEmpty() {
		return elements.isEmpty();
	}

	@Override
	public int getLength() {
		return elements.size();
	}

	/**
	 * Returns a list of the names of the nodes in the interface.
	 * @return the list of node names
	 */
	public List<String> getInterfaceNodeNames() {
		List<String> names = new LinkedList<>();
		for(Node node: outerInterface) {
			names.add(node.getName());
		}

		return names;
	}

	/**
	 * Returns a list of the nodes in the inner interface of this decomposition.
	 * @return the inner interface
	 */
	public List<Node> getInnerInterface() {
		return innerInterface;
	}

	/**
	 * Returns a list of the nodes in the outer interface of this decomposition.
	 * @return the outer interface
	 */
	public List<Node> getOuterInterface() {
		return outerInterface;
	}

	/**
	 * Returns {@code true} if and only if the operation {@code operation} can be added
	 * to this cospan decomposition.
	 * @param operation the operation to be added
	 * @return {@code true} if the operation can be added, {@code false} otherwise
	 * 
	 * @see #add(Operation)
	 * @see #add(CospanDecomposition)
	 */
	public boolean isApplicable(Operation operation) {
		if(operation instanceof Connect) {
			Connect connect = (Connect) operation;
			return outerInterface.size()>=connect.getArity();
		}
		else if(operation instanceof Fuse) {
			return outerInterface.size()>=2;
		}
		else if(operation instanceof Perm) {
			return outerInterface.size()>=2;
		}
		else if(operation instanceof Res) {
			return outerInterface.size()>=1;
		}
		else if(operation instanceof Trans) {
			return outerInterface.size()>=2;
		}
		else if(operation instanceof Vertex) {
			return true;
		}

		return false;
	}
	
	@Override
	public int hashCode() {
		return elements.hashCode() + 19*innerInterface.hashCode() + 37*outerInterface.hashCode();
	}
	
	@Override
	public boolean equals(Object o) {
		if(o==null) {
			return false;
		}
		else if(o==this) {
			return true;
		}
		else if(o.getClass().equals(getClass())) {
			CospanDecomposition decomposition = (CospanDecomposition) o;
			return graph.equals(decomposition.graph) && 
					innerInterface.equals(decomposition.innerInterface) &&
					outerInterface.equals(decomposition.outerInterface);
		}
		else {
			return false;
		}
	}

	@Override
	public Type getType() {
		return Decomposition.Type.cospanDecomposition;
	}

	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();
		for(Operation operation: this) {
			if(builder.length()>0) {
				builder.append("; ");
			}

			builder.append(operation.toString());
		}

		return builder.toString();
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	private void addOperation(Operation operation) throws UnsupportedOperationException {
		elements.add(operation);
		signature.add(operation);
		maximumInterfaceSize = Math.max(maximumInterfaceSize, outerInterface.size());
	}
	
	/**
	 * Adds a node to the middle graph.
	 * @param name the name of the node to be added
	 */
	private void addNode(String name) {
		addOperation(Operation.getVertex());
		graph.addNode(name);
		outerInterface.add(graph.getNode(name));
	}

	/**
	 * Adds an edge to the middle graph.
	 * @param label the label of the edge
	 * @param arity the arity of the edge
	 */
	private void addEdge(String label, int arity) {
		addOperation(Operation.getConnect(label, arity));
		
		List<Node> incidentNodes = new LinkedList<>();

		for(int i=getOuterInterfaceSize()-arity;i<getOuterInterfaceSize();i++) {
			incidentNodes.add(outerInterface.get(i));
		}			

		graph.addEdge("e"+graph.getEdgeCount(), label, incidentNodes);
	}
	
	/**
	 * 
	 */
	private void fuseNodes() {
		addOperation(Operation.getFuse());
		
		int lastIndex = getOuterInterfaceSize()-1;
		for(Edge edge: graph.getIncidentEdges(outerInterface.get(lastIndex))) {
			List<Node> incidentNodes = edge.getIncidentNodes();
			Node lastNode = outerInterface.get(lastIndex);
			Node nextToLastNode = outerInterface.get(lastIndex-1);
			
			int index = -1;
			while((index=incidentNodes.indexOf(lastNode))>-1) {
				incidentNodes.set(index, nextToLastNode);
			}
			
			graph.removeEdge(edge);
			graph.addEdge(edge.getName(), edge.getLabel(), incidentNodes);
		}
		int innerIndex = innerInterface.indexOf(outerInterface.get(lastIndex));
		if(innerIndex != -1) {
			innerInterface.set(innerIndex, outerInterface.get(lastIndex-1));			
		}
		
		graph.removeNode(outerInterface.get(lastIndex));
		outerInterface.remove(lastIndex);
	}
	
	/**
	 * Shifts the (outer) interface {@code length} steps to the right. 
	 * @param length the length of the shift
	 */
	private void shift(int length) {
		for(int i=0;i<length;i++) {
			permute();
			outerInterface.add(0, outerInterface.remove(outerInterface.size()-1));
		}
	}
	
	/**
	 * Shifts the {@code index}-th node of (outer) interface {@code length} steps to the left. 
	 * @param index the index of the node
	 * @param length the length of the shift
	 */
	private void shiftLeft(int index, int length) {		
		//Move the node which is permuted to the first interface position
		int steps = (getOuterInterfaceSize()-index)%getOuterInterfaceSize();
		shift(steps);
		
		for(int i=0;i<length;i++) {
			permute();
			addOperation(Operation.getTrans());

			outerInterface.add(0, outerInterface.remove(outerInterface.size()-1));
			Node firstNode = outerInterface.get(0);
			Node secondNode = outerInterface.get(1);
			outerInterface.set(0, secondNode);
			outerInterface.set(1, firstNode);
		}
		
		shift(index-length);
	}

	/**
	 * Shifts the {@code index}-th node of (outer) interface {@code length} steps to the right. 
	 * @param index the index of the node
	 * @param length the length of the shift
	 */
	private void shiftRight(int index, int length) {
		//Move the node which is permuted to the first interface position
		int steps = (getOuterInterfaceSize()-index)%getOuterInterfaceSize();
		shift(steps);

		for(int i=0;i<getOuterInterfaceSize()-length-1;i++) {
			permute();
			addOperation(Operation.getTrans());
			outerInterface.add(0, outerInterface.remove(outerInterface.size()-1));
			Node firstNode = outerInterface.get(0);
			Node secondNode = outerInterface.get(1);
			outerInterface.set(0, secondNode);
			outerInterface.set(1, firstNode);
		}

		shift(index+length);
	}

	/**
	 * Removes the {@code index}-th node from the interface.
	 * @param index the index of the node
	 */
	private void remove(int index) {
		for(int i=index;i<outerInterface.size()-1;i++) {
			permute();
			outerInterface.add(0, outerInterface.remove(outerInterface.size()-1));
		}

		addOperation(Operation.getRes());
		outerInterface.remove(outerInterface.size()-1);
	}
	
	private void permute() {
		if(getOuterInterfaceSize()>2) {
			addOperation(Operation.getPerm());
		}
		else {
			addOperation(Operation.getTrans());
		}
	}
	
	/**
	 * 
	 */
	private void transpose() {
		addOperation(Operation.getTrans());
		outerInterface.add(0, outerInterface.remove(1));
	}

}
