/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.data;

import java.io.Serializable;
import java.util.LinkedList;
import java.util.List;

/**
 * This class represents a single (hyper-)edge of some graph.
 * @author Christoph Blume
 * @version $Revision: 1.24 $
 *
 * @see Graph
 * @see Node
 */
public class Edge implements Data, Comparable<Edge>, Serializable {

	private static final long serialVersionUID = -1345415709807629054L;

	/**
	 * Enumeration of the different types of implemented edges.
	 * @author Christoph Blume
	 * @version $Revision: 1.24 $
	 *
	 */
	public enum Type implements Data.Type {
		/** */
		edge
	}
	
	private String name;
	private String label;
	private final List<Node> incidentNodes;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new edge with name {@code name}, label {@code label}
	 * @param name the name of the edge
	 * @param label the label of the edge
	 */
	public Edge(String name, String label) {
		this(name, label, new LinkedList<Node>());
	}
	
	/**
	 * Creates a new edge with name {@code name}, label {@code label} and incident nodes {@code incidentNodes}.
	 * @param name the name of the edge
	 * @param label the label of the edge
	 * @param incidentNodes the list of incident nodes
	 */
	public Edge(String name, String label, List<Node> incidentNodes) {
		this.name = new String(name);
		this.label = new String(label);
		this.incidentNodes = new LinkedList<>(incidentNodes);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the name of this edge.
	 * @return the edge's name
	 */
	public String getName() {
		return name;
	}
	
	/**
	 * Returns the label of this edge.
	 * @return the edge label
	 */
	public String getLabel() {
		return label;
	}
	
	/**
	 * Returns the arity of this edge
	 * @return the edge arity
	 */
	public int getArity() {
		return incidentNodes.size();
	}

	/**
	 * Returns the sequence of nodes incident to this edge
	 * @return the node sequence
	 */
	public List<Node> getIncidentNodes() {
		return new LinkedList<>(incidentNodes);
	}

	@Override
	public int hashCode() {
		return 41 * label.hashCode() + 91 * incidentNodes.hashCode();
	}
	
	@Override
	public boolean equals(Object o) {
		if(o==null) {
			return false;
		}
		else if(o==this) {
			return true;
		}
		else if(o.getClass().equals(getClass())) {
			Edge edge = (Edge) o;
			return label.equals(edge.getLabel()) && getArity()==edge.getArity() 
					&& name.equals(edge.getName());
		}
		else {
			return false;
		}
	}

	@Override
	public String toString() {
		return "Edge "+name;
	}

	@Override
	public int compareTo(Edge edge) {
		if(edge==null) {
			return Integer.MIN_VALUE;
		}

		if(name.length()==edge.getName().length()) {
			return name.compareTo(edge.getName());
		}
		
		return name.length()-edge.getName().length();
	}

	@Override
	public Type getType() {
		return Type.edge;
	}

}
