/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.data;

import java.io.Serializable;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class represents an edge-labeled <emph>hypergraph</emph>, i.e. a 4-tuple 
 * <i>G = (V<sub>G</sub>, E<sub>G</sub>, att<sub>G</sub>, lab<sub>G</sub>)</i> 
 * such that 
 * <ul>
 * 		<li><i>V<sub>G</sub></i> is a finite set of <emph>nodes</emph>,</li>
 * 		<li><i>E<sub>G</sub></i> is a finite set of <emph>(hyper-)edges</emph>,</li>
 * 		<li><i>att<sub>G</sub></i> is the <emph>attachment function</emph>, which 
 * 		assigns to each edge <i>e</i> a sequence of incident nodes 
 * 		<i>att<sub>G</sub>(e)</i> and</li>
 * 		<li><i>lab<sub>G</sub></i> is the <emph>labeling function</emph>, which 
 * 		assigns to each edge <i>e</i> label <i>lab<sub>G</sub>(e)</i>.</li>
 * </ul>
 * 
 * @author Christoph Blume
 * @version $Revision: 1.58 $
 *
 */
public class Graph implements Data, Cloneable, Serializable {

	private static final long serialVersionUID = 2940851421130492698L;

	/**
	 * Enumeration of the different types of implemented graphs.
	 * @author Christoph Blume
	 * @version $Revision: 1.58 $
	 *
	 */
	public enum Type implements Data.Type {
		/** Graph Type */
		graph;
	}

	/** Default Edge Name Prefix */
	public static final String DEFAULT_EDGE_NAME = "e";

	/** Default Node Name Prefix */
	public static final String DEFAULT_NODE_NAME = "n";

	/** Default edge label used if the label is not relevant */
	public static final String DEFAULT_EDGE_LABEL_STRING = "A";
	
	private String name;
	private Set<Node> nodes;
	private Set<Edge> edges;
	private Map<String, Integer> labels;

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new empty graph, i.e. the graph contains an empty node set and 
	 * an empty edge set.
	 * @param name the name of the graph
	 * 
	 * @see #Graph(String, Set, Set)
	 */
	public Graph(String name) {
		this(name, new HashSet<Node>(), new HashSet<Edge>());
	}

	/**
	 * Creates a new graph the node set {@code nodes} and edge set {@code edges}.
	 * @param name the name of the graph
	 * @param nodes the set of nodes
	 * @param edges the set of edges
	 * @throws NullPointerException if {@code nodes} or {@code edges} are null 
	 * 
	 * @see #Graph(String)
	 */
	private Graph(String name, Set<Node> nodes, Set<Edge> edges) throws NullPointerException {
		if(nodes==null) {
			throw new NullPointerException(
					RavenExceptionMessages.NODE_SET_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(edges==null) {
			throw new NullPointerException(
					RavenExceptionMessages.EDGE_SET_MUST_NOT_BE_NULL_EXCEPTION);
		}

		this.name = name;
		this.labels = new HashMap<>();
		this.nodes = nodes;
		this.edges = edges;
		for(Edge e: edges) {
			if(!labels.containsKey(e.getLabel())) {
				labels.put(e.getLabel(), e.getArity());
			}
			else if(labels.get(e.getLabel())!=e.getArity()) {
				throw new IllegalArgumentException(
						RavenExceptionMessages.ILLEGAL_EDGE_LABEL_ARITY_EXCEPTION);
			}
		}
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the name of the graph.
	 * @return the name of the graph
	 */
	public String getName() {
		return name;
	}

	/**
	 * Sets the name of this graph
	 * @param name the name of the graph
	 */
	public void setName(String name){
		this.name = name;
	}

	@Override
	public Graph clone() {
		Graph graph = new Graph(name);
		for(Node node: nodes) {
			graph.addNode(node.getName());
		}

		for(Edge edge: edges) {
			graph._addEdge(edge.getName(), edge.getLabel(), edge.getIncidentNodes());
		}
		return graph;
	}

	@Override
	public boolean equals(Object o) {
		if(o==null) {
			return false;
		}
		else if(o==this) {
			return true;
		}
		else if(o.getClass().equals(getClass())) {
			Graph graph = (Graph) o;
			return graph.nodes.equals(nodes) && graph.edges.equals(edges);
		}
		else {
			return false;
		}

	}

	@Override
	public int hashCode() {
		return nodes.hashCode() + 29 * edges.hashCode();
	}
	
	/**
	 * Returns {@code true} if and only if the graph is empty, i.e. it does neither contain any node 
	 * nor any edge.
	 * @return {@code true} if the graph is empty, {@code false} otherwise
	 */
	public boolean isEmpty() {
		return nodes.isEmpty() && edges.isEmpty();
	}

	/**
	 * Returns {@code true} if and only if this graph contains the node {@code node}.
	 * @param node the node to be checked
	 * @return {@code true} if the node is contained, {@code false} otherwise
	 * @throws NullPointerException thrown if {@code node} is null
	 */
	public boolean contains(Node node) {
		if(node==null) {
			throw new NullPointerException(
					RavenExceptionMessages.NODE_MUST_NOT_BE_NULL_EXCEPTION);
		}

		return getNodes().contains(node);
	}

	/**
	 * Returns {@code true} if and only if this graph contains
	 * a node with the name {@code nodeName}.
	 * @param nodeName the name of the node
	 * @return {@code true} if a node of the given name has 
	 * been found, {@code false} otherwise
	 * @throws NullPointerException thrown if {@code nodeName} is null
	 */
	public boolean hasNode(String nodeName) throws NullPointerException {
		if(nodeName==null) {
			throw new NullPointerException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}

		return getNode(nodeName)!=null ? true : false;
	}
	
	/**
	 * Returns {@code true} if and only if this graph contains
	 * a loop.
	 * @return {@code true} if a loop has 
	 * been found, {@code false} otherwise
	 */
	public boolean hasLoops() 
	{	
		Set<Edge> edgeSet = getEdges();
		Set<String> nodeNames = new HashSet<>();
		List<Node> nodeList;
		for(Edge edge: edgeSet) {
			nodeList = edge.getIncidentNodes();	
			for(Node node: nodeList) {	
				if(nodeNames.contains(node.getName())){
					return true;
				}

				nodeNames.add(node.getName());
			}
			nodeNames.clear();
		}
		return false;
	}

	/**
	 * Adds a new node with node name {@code nodeName} and label {@code label}
	 * to the graph.
	 * @param nodeName the name of the new node
	 * @param label the label of the new node
	 * @throws NullPointerException thrown if {@code nodeName} is null
	 * @throws IllegalArgumentException thrown if {@code nodeName} is empty or if 
	 * a node with name {@code nodeName} is already existing
	 */
	public void addNode(String nodeName, String label) 
			throws NullPointerException, IllegalArgumentException {
		if(nodeName==null) {
			throw new NullPointerException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(nodeName.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		if(label==null) {
			throw new NullPointerException(
					RavenExceptionMessages.LABEL_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(hasNode(nodeName)) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.DUPLICATE_NAME_EXCEPTION);
		}
		Node n = new Node(this, nodeName);
		nodes.add(n);
	}

	/**
	 * Adds a new node with node name {@code nodeName} and empty label to the graph.
	 * @param nodeName the name of the new node
	 * @throws NullPointerException thrown if {@code nodeName} is null
	 * @throws IllegalArgumentException thrown if {@code nodeName} is empty or if 
	 * a node with name {@code nodeName} is already existing
	 */
	public void addNode(String nodeName) 
			throws NullPointerException, IllegalArgumentException {
		addNode(nodeName, "");
	}

	/**
	 * Removes the node {@code node} from this graph.
	 * @param node the node to be removed
	 * 
	 * @throws NullPointerException thrown if {@code node} is {@code null}
	 * @throws IllegalArgumentException if {@code node} is not a node of this graph
	 */
	public void removeNode(Node node) 
			throws NullPointerException, IllegalArgumentException {
		if(node==null) {
			throw new NullPointerException(
					RavenExceptionMessages.NODE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(!nodes.contains(node)) {
			throw new IllegalArgumentException(RavenExceptionMessages.NODE_NOT_EXISTING_EXCEPTION);
		}

		for(Edge edge: getIncidentEdges(node)) {
			removeEdge(edge);
		}
		nodes.remove(node);
	}

	/**
	 * Removes the node with the name {@code nodeName} from this graph.
	 * @param nodeName the name of the node
	 * 
	 * @throws NullPointerException thrown if {@code nodeName} is {@code null}
	 * @throws IllegalArgumentException thrown if {@code nodeName} is empty or the node is not existing
	 */
	public void removeNode(String nodeName) 
			throws NullPointerException, IllegalArgumentException {
		if(nodeName==null) {
			throw new NullPointerException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(nodeName.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		if(!hasNode(nodeName)) {
			throw new IllegalArgumentException(RavenExceptionMessages.NODE_NOT_EXISTING_EXCEPTION);
		}

		removeNode(getNode(nodeName));
	}

	/**
	 * Returns the node with the name {@code nodeName} or {@code null} if no 
	 * node with the given name is existing.
	 * @param nodeName the name of the node
	 * @return the node or {@code null} if no node has been found
	 * 
	 * @throws NullPointerException thrown if {@code nodeName} is {@code null}
	 * @throws IllegalArgumentException thrown if {@code nodeName} is empty
	 */
	public Node getNode(String nodeName) 
			throws NullPointerException, IllegalArgumentException {
		if(nodeName==null) {
			throw new NullPointerException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(nodeName.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}

		for(Node node: nodes) {
			if(node.getName().equals(nodeName)) {
				return node;
			}
		}

		return null;
	}

	/**
	 * Returns the set of node names of this graph.
	 * @return the set of node names
	 */
	public Set<String> getNodeNames() {
		Set<String> names = new HashSet<>();
		for(Node node: nodes) {
			names.add(node.getName());
		}

		return names;
	}

	/**
	 * Returns the set of nodes of this graph.
	 * @return the set of nodes
	 */
	public Set<Node> getNodes() {
		return new HashSet<>(nodes);
	}
	
	/**
	 * Returns the map of labels of this graph.
	 * @return the set of labels
	 */
	public Map<String, Integer> getLabels() {
		return labels;
	}
	
	/**
	 * Returns {@code true} if and only if this graph contains the edge {@code edge}.
	 * @param edge the edge to be checked
	 * @return {@code true} if the edge is contained, {@code false} otherwise
	 * @throws NullPointerException thrown if {@code edge} is null
	 */
	public boolean contains(Edge edge) {
		if(edge==null) {
			throw new NullPointerException(
					RavenExceptionMessages.EDGE_MUST_NOT_BE_NULL_EXCEPTION);
		}

		return getEdges().contains(edge);
	}

	/**
	 * Returns {@code true} if and only if this graph contains
	 * an edge with the name {@code edgeName}.
	 * @param edgeName the name of the edge
	 * @return {@code true} if an edge of the given name has 
	 * been found, {@code false} otherwise
	 */
	public boolean hasEdge(String edgeName) {
		for(Edge edge: edges) {
			if(edge.getName().equals(edgeName)) {
				return true;
			}
		}

		return false;
	}

	/**
	 * Adds a labeled edge to this graph.
	 * @param edgeName the name of the new edge
	 * @param label the label of the new edge
	 * @param incidentNodes the incident nodes of the new edge
	 * @throws NullPointerException if {@code edgeName}, {@code edgeLabel} or 
	 * {@code incidentNodes} is null
	 * @throws IllegalArgumentException if {@code edgeName} or {@code edgeLabel} 
	 * are empty, at least one node of {@code incidentNodes} is not contained in 
	 * this graph or if an edge with name {@code edgeName} is already existing 
	 */
	public void addEdge(String edgeName, String label, List<Node> incidentNodes) 
			throws NullPointerException, IllegalArgumentException {
		if(edgeName==null) {
			throw new NullPointerException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(edgeName.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		if(label==null) {
			throw new NullPointerException(
					RavenExceptionMessages.LABEL_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(label.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.LABEL_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		if(labels.keySet().contains(label) && 
				labels.get(label)!=incidentNodes.size()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_EDGE_LABEL_ARITY_EXCEPTION);
		}
		if(incidentNodes==null) {
			throw new NullPointerException(
					RavenExceptionMessages.INCIDENT_NODES_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(!nodes.containsAll(incidentNodes)) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INCIDENT_NODES_MUST_BE_CONTAINED_EXCEPTION);
		}		
		if(hasEdge(edgeName)) {
			throw new IllegalArgumentException(RavenExceptionMessages.DUPLICATE_NAME_EXCEPTION);
		}

		_addEdge(edgeName, label, incidentNodes);
	}
	
	/**
	 * Adds an unlabeled edge to this graph.
	 * @param edgeName the name of the new edge
	 * @param incidentNodes the incident nodes of the new edge
	 * @throws NullPointerException if {@code edgeName} or {@code incidentNodes} is null
	 * @throws IllegalArgumentException if {@code edgeName} is empty, at least one node of 
	 * {@code incidentNodes} is not contained in this graph or if an edge with name 
	 * {@code edgeName} is already existing
	 */
	public void addEdge(String edgeName, List<Node> incidentNodes) {
		if(edgeName==null) {
			throw new NullPointerException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(edgeName.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		if(incidentNodes==null) {
			throw new NullPointerException(
					RavenExceptionMessages.INCIDENT_NODES_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(!nodes.containsAll(incidentNodes)) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INCIDENT_NODES_MUST_BE_CONTAINED_EXCEPTION);
		}		
		if(hasEdge(edgeName)) {
			throw new IllegalArgumentException(RavenExceptionMessages.DUPLICATE_NAME_EXCEPTION);
		}
		
		_addEdge(edgeName, DEFAULT_EDGE_LABEL_STRING, incidentNodes);
	}

	/**
	 * Removes the edge {@code edge} from this graph.
	 * @param edge the edge to be removed
	 * 
	 * @throws NullPointerException thrown if {@code edge} is {@code null}
	 * @throws IllegalArgumentException thrown if the {@code edge} is not an edge of this graph
	 */
	public void removeEdge(Edge edge) 
			throws NullPointerException, IllegalArgumentException {
		if(edge==null) {
			throw new NullPointerException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(!edges.contains(edge)) {
			throw new IllegalArgumentException(RavenExceptionMessages.EDGE_NOT_EXISTING_EXCEPTION);
		}

		edges.remove(edge);
	}

	/**
	 * Removes the edge with the name {@code edgeName} from this graph.
	 * @param edgeName the name of the edge
	 * 
	 * @throws NullPointerException thrown if {@code edgeName} is {@code null}
	 * @throws IllegalArgumentException thrown if {@code edgeName} is empty or the edge is not existing
	 */
	public void removeEdge(String edgeName) 
			throws NullPointerException, IllegalArgumentException {
		if(edgeName==null) {
			throw new NullPointerException(RavenExceptionMessages.NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(edgeName.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		if(!hasEdge(edgeName)) {
			throw new IllegalArgumentException(RavenExceptionMessages.EDGE_NOT_EXISTING_EXCEPTION);
		}

		removeEdge(getEdge(edgeName));
	}

	/**
	 * Returns the edge with the name {@code edgeName} or {@code null} if no 
	 * edge with the given name is existing.
	 * @param edgeName the name of the edge
	 * @return the node or {@code null} if no node has been found
	 * 
	 * @throws NullPointerException thrown if {@code nodeName} is {@code null}
	 * @throws IllegalArgumentException thrown if {@code edgeName} is empty
	 */
	public Edge getEdge(String edgeName) 
			throws NullPointerException, IllegalArgumentException {
		if(edgeName==null) {
			throw new NullPointerException(RavenExceptionMessages.NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(edgeName.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}

		for(Edge edge: edges) {
			if(edge.getName().equals(edgeName)) {
				return edge;
			}
		}

		return null;
	}


	/**
	 * Returns the set of edge names of this graph.
	 * @return the set of edge names
	 */
	public Set<String> getEdgeNames() {
		Set<String> names = new HashSet<>();
		for(Edge edge: edges) {
			names.add(edge.getName());
		}

		return names;
	}

	/**
	 * Returns the set of edges of this graph.
	 * @return the set of edges
	 */
	public Set<Edge> getEdges() {
		return new HashSet<>(edges);
	}

	/**
	 * Returns the set of edges incident to the node {@code node}
	 * @param node the node
	 * @return the set of incident edges
	 */
	public Set<Edge> getIncidentEdges(Node node) {
		Set<Edge> incidentEdges = new HashSet<>();
		for(Edge edge: edges) {
			if(edge.getIncidentNodes().contains(node)) {
				incidentEdges.add(edge);
			}
		}

		return incidentEdges;
	}

	/**
	 * Returns the set of edges incident to the node {@code nodeName}
	 * @param nodeName the name of the node
	 * @return the set of incident edges
	 */
	public Set<Edge> getIncidentEdges(String nodeName) {
		return getIncidentEdges(getNode(nodeName));
	}

	/**
	 * Returns the set of nodes incident to the node {@code node}
	 * @param node the node
	 * @return the set of incident nodes
	 */
	public Set<Node> getAdjacentNodes(Node node) {
		Set<Node> adjacentNodes = new HashSet<>();
		for(Edge edge: getIncidentEdges(node)) {
			adjacentNodes.addAll(edge.getIncidentNodes());
		}
		adjacentNodes.remove(node);

		return adjacentNodes;
	}

	/**
	 * Returns the set of nodes adjacent to the node {@code nodeName}
	 * @param nodeName the name of the node
	 * @return the set of adjacent nodes
	 */
	public Set<Node> getAdjacentNodes(String nodeName) {		
		return getAdjacentNodes(getNode(nodeName));
	}

	/**
	 * Returns the number of nodes adjacent to the node {@code node}
	 * @param node the node
	 * @return the number of adjacent nodes
	 */
	public int getNumberOfAdjacentNodes(Node node) {
		return getAdjacentNodes(node).size();
	}

	/**
	 * Returns the number of nodes adjacent to the node {@code node}
	 * @param nodeName the name of the node
	 * @return the number of adjacent nodes
	 */
	public int getNumberOfAdjacentNodes(String nodeName) {
		return getAdjacentNodes(nodeName).size();
	}

	/**
	 * Returns {@code true} if and only if the node {@code firstNode} is 
	 * adjacent to the node {@code secondNode} 
	 * @param firstNode the first node
	 * @param secondNode the second node
	 * @return {@code true} if the first node is adjacent to the second node, 
	 * {@code false} otherwise
	 */
	public boolean isAdjacentTo(Node firstNode, Node secondNode) {
		return getAdjacentNodes(firstNode).contains(secondNode);
	}

	/**
	 * Returns the size of this graph, i.e. the number of nodes.
	 * @return the graph size
	 */
	public int getSize() {
		return nodes.size()+edges.size();
	}

	/**
	 * Returns the number of nodes of this graph
	 * @return the number of nodes
	 */
	public int getNodeCount() {
		return nodes.size();
	}

	/**
	 * Returns the number of edges of this graph
	 * @return the number of edges
	 */
	public int getEdgeCount() {
		return edges.size();
	}

	@Override
	public Type getType() {
		return Type.graph;
	}
	
	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();
		
		List<Node> nodeList = new LinkedList<>(nodes);
		Collections.sort(nodeList);
		builder.append("Nodes:\n");
		for(Node node: nodeList) {
			builder.append("  - "+node.getName()+"\n");
		}
		builder.append("\n");
		
		List<Edge> edgeList = new LinkedList<>(edges);
		Collections.sort(edgeList);
		builder.append("Edges:\n");
		for(Edge edge: edgeList) {
			int arity = edge.getArity();
			builder.append("  - "+edge.getName() + "=" + edge.getLabel() +"(");
			int i=0;
			for(Node node: edge.getIncidentNodes()) {
				builder.append(node.getName());
				if(i<arity-1) {
					builder.append(",");
					i++;
				}
			}
			builder.append(")\n");
		}
		
		return builder.toString();
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	private void _addEdge(String edgeName, String label, List<Node> incidentNodes) 
			throws NullPointerException, IllegalArgumentException {
		if(edgeName==null) {
			throw new NullPointerException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(edgeName.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		if(label==null) {
			throw new NullPointerException(
					RavenExceptionMessages.LABEL_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(label.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.LABEL_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		if(labels.keySet().contains(label) && labels.get(label)!=incidentNodes.size()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_EDGE_LABEL_ARITY_EXCEPTION);
		}
		if(incidentNodes==null) {
			throw new NullPointerException(
					RavenExceptionMessages.INCIDENT_NODES_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(!nodes.containsAll(incidentNodes)) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INCIDENT_NODES_MUST_BE_CONTAINED_EXCEPTION);
		}		
		if(hasEdge(edgeName)) {
			throw new IllegalArgumentException(RavenExceptionMessages.DUPLICATE_NAME_EXCEPTION);
		}

		edges.add(new Edge(edgeName, label, incidentNodes));
		labels.put(label, incidentNodes.size());
	}

}
