/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.data;

import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

/**
 * The {@code GraphDecomposition} is the <i>parent interface</i> of the 
 * {@code PathDecomposition} and the {@code TreeDecomposition} interface. This 
 * interface represents decompositions of graphs in several bags (i.e. sets of 
 * nodes) which satisfy some specific criteria depending on the implementing 
 * classes.
 * @author Christoph Blume
 * @version $Revision: 1.16 $
 *
 * @see PathDecomposition
 * @see TreeDecomposition
 * @see Graph
 */
public interface GraphDecomposition extends Decomposition<GraphDecomposition.Bag> {
	
	/**
	 * This class represents a set of nodes used by several types of graph 
	 * decompositions.
	 * @author Christoph Blume
	 * @version $Revision: 1.16 $
	 *
	 * @see PathDecomposition
	 * @see TreeDecomposition
	 */
	public class Bag implements Iterable<Node> {
		
		private Set<Node> nodeSet;
		private Set<Bag> neighbors;
		private int identifier;
		
		/**
		 * Creates a new empty bag.
		 * @param identifier the identifier of the node
		 */
		public Bag(int identifier) {
			this.identifier = identifier;
			nodeSet = new HashSet<>();
			neighbors = new HashSet<>();
		}
		
		/**
		 * Returns the identifier of this bag.
		 * @return the bag's identifier
		 */
		public int getIdentifier() {
			return identifier;
		}
		
		/**
		 * Returns the size of this bag.
		 * @return the bag's size
		 */
		public int getSize() {
			return nodeSet.size();
		}
		
		/**
		 * Adds a node to this bag.
		 * @param node the node to be added
		 */
		public void addNode(Node node) {
			nodeSet.add(node);
		}
		
		/**
		 * Adds a set of nodes to this bag.
		 * @param nodes the node set to be added
		 */
		public void addNodeSet(Set<Node> nodes) {
			for(Node node: nodes) {
				addNode(node);
			}
		}
		
		/**
		 * Returns the set of nodes contained in this bag.
		 * @return an unmodifiable set of nodes
		 */
		public Set<Node> getNodes() {
			return Collections.unmodifiableSet(nodeSet);
		}
		
		/**
		 * Adds a neighbor to this bag
		 * @param neighbor the bag's neighbor
		 */
		public void addNeighbor(Bag neighbor) {
			neighbors.add(neighbor);
		}
		
		/**
		 * Returns the set of neighbors of this bag
		 * @return the bag's neighbors
		 */
		public Set<Bag> getNeighbors() {
			return neighbors;
		}
		
		/**
		 * Returns the number of neighbors of this bag.
		 * @return the number of neighbors
		 */
		public int getNumberOfNeighbors() {
			return neighbors.size();
		}

		@Override
		public Iterator<Node> iterator() {
			return nodeSet.iterator();
		}
		
		@Override
		public int hashCode() {
			int hashCode = 1;
			
			for(Node node: this) {
				hashCode = 29*hashCode + ((node==null) ? 0 : node.hashCode());
			}
			return hashCode;
		}
		
		@Override
		public boolean equals(Object o) {
			if(o==null) {
				return false;
			}
			else if(o==this) {
				return true;
			}
			else if(o.getClass().equals(getClass())) {
				Bag bag = (Bag) o;
				return identifier==bag.identifier && nodeSet.equals(bag.nodeSet);
			}
			else {
				return false;
			}
		}
		
		@Override
		public String toString() {
			return nodeSet.toString();
		}
		
	}

	/**
	 * Returns the bag with the identifier {@code identifier}.
	 * @param identifier the identifier of the bag
	 * @return the bag with the specified identifier
	 */
	public Bag getBag(int identifier);

	/**
	 * Returns the graph which can be obtained from this graph decomposition.
	 * @return the graph obtained from this graph decomposition
	 */
	public Graph getGraph();

	
}
