/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.data;

import java.util.HashSet;
import java.util.Set;

import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;


/**
 * This class is used to assign a name to some {@code data} object.
 * @author Christoph Blume
 * @version $Revision: 1.1 $
 *
 * @see Data
 */
public final class NameDataPair implements Comparable<NameDataPair> {

	private static final Set<String> NAMES_IN_USE_SET = new HashSet<>();
	
	private String name;
	private Data data;

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new {@code NameDataPair}.
	 * @param name the name of the {@code Data} object
	 * @param data the {@code Data} object
	 */
	public NameDataPair(String name, Data data) {
		if(name==null) {
			throw new NullPointerException(RavenExceptionMessages.NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(name.isEmpty()) {
			throw new IllegalArgumentException(RavenExceptionMessages.NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		if(NAMES_IN_USE_SET.contains(name)) {
			throw new RuntimeException(RavenExceptionMessages.NAME_ALREADY_IN_USE_EXCEPTION);
		}
		if(data==null) {
			throw new NullPointerException(RavenExceptionMessages.DATA_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		this.name = name;
		this.data = data;
		NAMES_IN_USE_SET.add(name);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/* 		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public boolean equals(Object o) {
		if(o==null) {
			return false; 
		}
		else if(o==this) {
			return true;
		}
		else if (!o.getClass().equals(getClass())) {
			return false;
		}
		else {
			NameDataPair p = (NameDataPair) o;
			return p.name.equals(name) && p.data.equals(data); 
		}
	}
	
	@Override
	public int hashCode() {
		return name.hashCode()*data.hashCode()*31;
	}

	/**
	 * Returns the name of the {@code Data} object
	 * @return the name of the object
	 */
	public String getName() {
		return name;
	}

	/**
	 * Returns the {@code Data} object
	 * @return the object
	 */
	public Data getData() {
		return data;
	}

	@Override
	public int compareTo(NameDataPair pair) {
		if(pair==this) {
			return 0;
		}
		else if(pair==null) {
			return -1;
		}
		else {
			return name.compareTo(pair.name);
		}
	}
	
	@Override
	public String toString() {
		return getName();
	}
	
}
