/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.data;

import java.io.Serializable;

import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class represents a single node of some hypergraph.
 * @author Christoph Blume
 * @version $Revision: 1.28 $
 * 
 * @see Graph
 * @see Edge
 *
 */
public class Node implements Data, Comparable<Node>, Serializable {

	private static final long serialVersionUID = 2412159522088632516L;

	/**
	 * Enumeration of the different types of implemented nodes.
	 * @author Christoph Blume
	 * @version $Revision: 1.28 $
	 *
	 */
	public enum Type implements Data.Type {
		/** Node Type */
		node
	}
	
	private String name;
	private String label;
	private Graph graph;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new node which is contained in the graph {@code graph}.
	 * @param graph the graph in which this node is contained
	 * @param name the name of the node
	 * @param label the node label
	 */

	protected Node(Graph graph, String name, String label) {
		if(graph==null) {
			throw new NullPointerException(
					RavenExceptionMessages.GRAPH_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(name==null) {
			throw new NullPointerException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(name.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		
		this.graph = graph;
		this.name = name;
		this.label = label;
	}
	
	/**
	 * Creates a new node which is contained in the graph {@code graph}.
	 * @param graph the graph in which this node is contained
	 * @param name the name of the node
	 */
	protected Node(Graph graph, String name) {
		this(graph, name, name);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the name of this node.
	 * @return the node's name
	 */
	public String getName() {
//		String text = graph!=null ? graph.getName()+"." : "";
//		return text+name;
		return name;
	}
	
//	/**
//	 * Sets the name of the node.
//	 * @param name the name of the node
//	 */
//	public void setName(String name){
//		this.name = name;
//	}

	/**
	 * Returns the graph to which this node belongs.
	 * @return the graph containing this node
	 */
	public Graph getGraph() {
		return graph;
	}
	
	@Override
	public String toString() {
		return label;
	}
	
	@Override
	public boolean equals(Object o) {
		if(o==null) {
			return false;
		}
		else if(o==this) {
			return true;
		}
		else if(o.getClass().equals(getClass())) {
			Node node = (Node) o;
			return name.equals(node.name);
		}
		else {
			return false;
		}
	}
		
	@Override
	public int hashCode() {
		return name.hashCode();
	}
	
	@Override
	public int compareTo(Node node) {
		if(node==null) {
			return Integer.MIN_VALUE;
		}

		if(name.length()==node.getName().length()) {
			return name.compareTo(node.getName());
		}

		return name.length()-node.getName().length();
	}

	@Override
	public Type getType() {
		return Type.node;
	}
	
}