/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.data;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Scanner;
import java.util.TreeMap;
import java.util.regex.MatchResult;

import de.uni_due.inf.ti.raven.cli.Console;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class is the abstract base class for the several operations which can be 
 * used to create graphs and cospans by means of the composition of these 
 * operations. 
 * @author Christoph Blume
 * @version $Revision: 1.42 $
 *
 */
public abstract class Operation implements Data, Serializable {

	private static final long serialVersionUID = -3460791386132048965L;


	/**
	 * Enumeration of the different types of implemented operations.
	 * @author Christoph Blume
	 * @version $Revision: 1.42 $
	 *
	 */
	public enum Type implements Data.Type {
		/** Operation */
		operation
	}

	/**
	 * This class imposes a total order on the set of (atomic cospan) operations.
	 * @author Christoph Blume
	 * @version $Revision: 1.42 $
	 *
	 */
	public final static class Comparator implements java.util.Comparator<Operation> {

		private Comparator() {}

		private static int getValue(Operation operation) {
			if(operation instanceof Connect) {
				return 1;
			}
			else if(operation instanceof Fuse) {
				return 2;
			}
			else if(operation instanceof Perm) {
				return 3;
			}
			else if(operation instanceof Res) {
				return 4;
			}
			else if(operation instanceof Trans) {
				return 5;
			}
			else if(operation instanceof Vertex) {
				return 6;
			}

			return 0;
		}

		@Override
		public int compare(Operation o1, Operation o2) {
			return getValue(o1)-getValue(o2);
		}

	}

	/**  */
	public static final Comparator COMPARATOR = new Comparator();

	/** The CONNECT Operation Token */
	public static final String CONNECT_TOKEN = "connect";

	/** The FUSE Operation Token */
	public static final String FUSE_TOKEN = "fuse";

	/** The PERM Operation Token */
	public static final String PERM_TOKEN = "perm";

	/** The RES Operation Token */
	public static final String RES_TOKEN = "res";

	/** The TRANS Operation Token */
	public static final String TRANS_TOKEN = "trans";

	/** The VERTEX Operation Token */
	public static final String VERTEX_TOKEN = "vertex";

	/** The ARITY Token */
	public static final String ARITY_TOKEN = ":";

	private static final String CONNECT_PATTERN = CONNECT_TOKEN + Console.CONNECT_PARAMETER;
	private static final String MAP_PATTERN = Console.CONNECT_PARAMETER;

	/**
	 * This class implements the CONNECT-Operation which is used to attach a new edge to the 
	 * last nodes in the interface of some graph (cospan). 
	 * @author Christoph Blume
	 * @version $Revision: 1.42 $
	 *
	 */
	public static final class Connect extends Operation {

		private static final long serialVersionUID = -1249425517447716540L;

		private String label;
		private int arity;

		/*------------------------------------------------------------------------------------------*/
		/*		Constructors																		*/
		/*------------------------------------------------------------------------------------------*/

		/**
		 * Creates a new {@code Connect}-operation which encodes the gluing with 
		 * a new edge labeled with {@code label} of arity {@code arity}.
		 * @param label the label of the new edge
		 * @param arity the arity of the new edge
		 */
		private Connect(String label, int arity) {
			super(Console.CONNECT_TOKEN + "(" + label + ARITY_TOKEN + String.valueOf(arity) + ")");
			this.label = label;
			this.arity = arity;
		}

		/*------------------------------------------------------------------------------------------*/
		/*		Public Methods																		*/
		/*------------------------------------------------------------------------------------------*/

		/**
		 * Returns the label of the edge.
		 * @return the label of the edge
		 */
		public String getLabel() {
			return label;
		}

		/**
		 * Returns the arity of the edge.
		 * @return the arity of the edge
		 */
		public int getArity() {
			return arity;
		}

		@Override
		public int hashCode() {
			return 43*arity + label.hashCode();
		}

		@Override
		public boolean equals(Object o) {
			if(o instanceof Connect) {
				Connect c = (Connect) o;
				return c.getArity()==arity && c.getLabel().equals(label);
			}

			return false;
		}

	}


	/**
	 * This class implements the FUSE-Operation which is used to fuse the last two nodes in the 
	 * interface of some graph (cospan).
	 * @author Christoph Blume
	 * @version $Revision: 1.42 $
	 *
	 */
	public static final class Fuse extends Operation {

		private static final long serialVersionUID = -3977468413270962136L;

		/*------------------------------------------------------------------------------------------*/
		/*		Constructors																		*/
		/*------------------------------------------------------------------------------------------*/

		/**
		 * 
		 */
		private Fuse() {
			super(Console.FUSE_TOKEN);
		}

		@Override
		public int hashCode() {
			return 47;
		}

		@Override
		public boolean equals(Object o) {
			if(o instanceof Fuse) {
				return true;
			}

			return false;
		}

	}

	/**
	 * This class implements the PERM-Operation which is used to cyclic shift the nodes in the 
	 * interface of some graph (cospan).
	 * @author Christoph Blume
	 * @version $Revision: 1.42 $
	 *
	 */
	public static final class Perm extends Operation {

		private static final long serialVersionUID = 5117855862329868739L;

		/*------------------------------------------------------------------------------------------*/
		/*		Constructors																		*/
		/*------------------------------------------------------------------------------------------*/


		/**
		 * 
		 */
		private Perm() {
			super(Console.PERM_TOKEN);
		}

		@Override
		public int hashCode() {
			return 53;
		}

		@Override
		public boolean equals(Object o) {
			if(o instanceof Perm) {
				return true;
			}

			return false;
		}

	}

	/**
	 * This class implements the RES-Operation which is used to restrict the last node in the 
	 * interface of some graph (cospan).
	 * @author Christoph Blume
	 * @version $Revision: 1.42 $
	 *
	 */
	public static final class Res extends Operation {

		private static final long serialVersionUID = -8862308621450129896L;

		/*------------------------------------------------------------------------------------------*/
		/*		Constructors																		*/
		/*------------------------------------------------------------------------------------------*/

		private Res() {
			super(Console.RES_TOKEN);
		}

		@Override
		public int hashCode() {
			return 59;
		}

		@Override
		public boolean equals(Object o) {
			if(o instanceof Res) {
				return true;
			}

			return false;
		}

	}

	/**
	 * This class implements the TRANS-Operation which is used to transpose the first two nodes in the 
	 * interface of some graph (cospan).
	 * @author Christoph Blume
	 * @version $Revision: 1.42 $
	 *
	 */
	public static final class Trans extends Operation {

		private static final long serialVersionUID = 4310660848654396834L;

		/*------------------------------------------------------------------------------------------*/
		/*		Constructors																		*/
		/*------------------------------------------------------------------------------------------*/

		private Trans() {
			super(Console.TRANS_TOKEN);
		}

		@Override
		public int hashCode() {
			return 61;
		}

		@Override
		public boolean equals(Object o) {
			if(o instanceof Trans) {
				return true;
			}

			return false;
		}

	}

	/**
	 * This class implements the VERTEX-Operation which is used to add a new node at the last 
	 * position of the interface of some graph (cospan).
	 * @author Christoph Blume
	 * @version $Revision: 1.42 $
	 *
	 */
	public static final class Vertex extends Operation {

		private static final long serialVersionUID = -8514378889588513185L;		

		/*------------------------------------------------------------------------------------------*/
		/*		Constructors																		*/
		/*------------------------------------------------------------------------------------------*/

		private Vertex() {
			super(Console.VERTEX_TOKEN);
		}

		@Override
		public int hashCode() {
			return 67;
		}

		@Override
		public boolean equals(Object o) {
			if(o instanceof Vertex) {
				return true;
			}

			return false;
		}

	}

	private static final Fuse fuse = new Fuse();
	private static final Perm perm = new Perm();
	private static final Res res = new Res();
	private static final Trans trans = new Trans();
	private static final Vertex vertex = new Vertex();

	/** The default edge label */
	public static final String DEFAULT_EDGE_LABEL = Graph.DEFAULT_EDGE_LABEL_STRING;

	private String name;

	private static Map<String, Connect> connectMap = new TreeMap<>(
			new java.util.Comparator<String>() {

				@Override
				public int compare(String op1, String op2) {
					List<String> parameter = parseParameter(op1);
					String label1 = parameter.get(0);
					String arity1 = parameter.get(1);
					parameter = parseParameter(op2);
					String label2 = parameter.get(0);
					String arity2 = parameter.get(1);

					int value = label1.compareTo(label2);
					if(value==0) {
						value = arity1.compareTo(arity2);
					}
					return value;
				}

				private List<String> parseParameter(String input) {
					List<String> propertiesList = new ArrayList<>();

					//Scan and extract properties from input string
					try(Scanner s = new Scanner(input)) {
						s.findInLine(MAP_PATTERN);
						MatchResult result = s.match();
						for(int i=1;i<=result.groupCount();i++) {
							propertiesList.add(result.group(i));
						}
					};

					return propertiesList;
				}

			});


	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Parses an operation from a text string.
	 * @param text the string to be parsed
	 * @return the parsed operation
	 */
	public static Operation parse(String text) {
		//TODO Parser fuer die Connect-Operation ueberarbeiten
		MatchResult result = null;
		
		if(text.matches(CONNECT_PATTERN)) {
			try(Scanner s = new Scanner(text)) {
				s.findInLine(CONNECT_PATTERN);
				result = s.match();
			};

			return getConnect(result.group(1), Integer.parseInt(result.group(2)));
		}
		else if(text.equals(getFuse().toString())) {
			return fuse;
		}
		else if(text.equals(getPerm().toString())) {
			return perm;
		}
		else if(text.equals(getRes().toString())) {
			return res;
		}
		else if(text.equals(getTrans().toString())) {
			return trans;
		}
		else if(text.equals(getVertex().toString())) {
			return vertex;
		}

		throw new IllegalArgumentException(
				RavenExceptionMessages.ILLEGAL_OPERATION_STRING_EXCEPTION);
	}


	/**
	 * Returns the CONNECT operation for the specified edge.
	 * @param label the edge label
	 * @param arity the edge arity
	 * @return the CONNECT operation
	 */
	public static Connect getConnect(String label, int arity) {
		String key = "("+label+ARITY_TOKEN+String.valueOf(arity)+")";
		if(!connectMap.containsKey(key)) {
			Connect c = new Connect(label, arity);
			connectMap.put(key, c);
		}

		return connectMap.get(key);
	}

	/**
	 * Checks whether the label is correct with respect to the pattern of allowed labels.
	 * @param label the label to be checked
	 * @return {@code true} if the name is valid, {@code false} otherwise
	 */
	public static boolean isLabelValid(String label) {
		if(label==null) {
			return false;
		}

		return label.matches(Console.LABEL_PATTERN);
	}

	/**
	 * Returns the FUSE operation.
	 * @return the FUSE operation
	 */
	public static final Fuse getFuse() {
		return fuse;
	}

	/**
	 * Returns the PERM operation.
	 * @return the PERM operation
	 */
	public static final Perm getPerm() {
		return perm;
	}

	/**
	 * Returns the RES operation.
	 * @return the RES operation
	 */
	public static final Res getRes() {
		return res;
	}

	/**
	 * Returns the TRANS operation.
	 * @return the TRANS operation
	 */
	public static final Trans getTrans() {
		return trans;
	}

	/**
	 * Returns the VERTEX operation.
	 * @return the VERTEX operation
	 */
	public static final Vertex getVertex() {
		return vertex;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new operation with name {@code name}.
	 * @param name the operation's name 
	 */
	protected Operation(String name) {
		this.name = name;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public String toString() {
		return name;
	}


	@Override
	public final Type getType() {
		return Type.operation;
	}

	/**
	 * Get the name of an operation
	 * @return the name/String of the operation
	 */
	public final String parse() {
		if(getClass().equals(Connect.class)) {
			Connect o = (Connect) this;
			return CONNECT_TOKEN+"(" + o.label + ARITY_TOKEN + o.arity + ")";
		}
		else if(getClass().equals(Fuse.class)) {
			return FUSE_TOKEN;
		}
		else if(getClass().equals(Perm.class)) {
			return PERM_TOKEN;
		}
		else if(getClass().equals(Res.class)) {
			return RES_TOKEN;
		}
		else if(getClass().equals(Trans.class)) {
			return TRANS_TOKEN;
		}
		else if(getClass().equals(Vertex.class)) {
			return VERTEX_TOKEN;
		}
		return "";
	}

}
