/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.data;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * A <emph>Path Decomposition</emph> of some graph <i>G</i> is a pair <i>&#x1D4AB; = 
 * (P,X)</i> where <i>P</i> is a path and <i>X</i> is a family of sets of nodes 
 * (of the graph <i>G</i>) which are indexed by (path) nodes of <i>P</i>, such 
 * that the following conditions are satisfied:
 * <ul>
 * <li>for each (graph) node <i>v &isin; G</i>, there exists a node <i>p</i> of 
 * <i>P</i> such that <i>v &isin; X<sub>p</sub></i></li>
 * <li>for each (graph) edge <i>e &isin; G</i>, there is a node <i>p</i> of 
 * <i>P</i> such that all nodes <i>v</i> attached to <i>e</i> are in 
 * <i>X<sub>p</sub></i></li>
 * <li>for each node <i>v &isin; G</i>, the graph induced by the nodes <i>{p 
 * &#8739 v &isin; X<sub>p</sub>}</i> is a subpath of <i>P</i>.</li>
 * </ul>
 * 
 * @author Christoph Blume
 * @version $Revision: 1.31 $
 *
 */
public class PathDecomposition implements GraphDecomposition {

	private static Graph decompositionGraph;

	private List<Bag> bagList;
	private Map<Node, Bag> nodeToBagMap = new HashMap<>();
	private Map<Integer, Bag> indexToBagMap = new HashMap<>();

	private int width;
	private Graph graph;

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new path decomposition from the permutation {@code permutation}.
	 * @param permutation the permutation
	 * @return the created path decomposition
	 */
	public static PathDecomposition getPathDecomposition(Permutation permutation) {
		if(permutation==null) {
			throw new NullPointerException(
					RavenExceptionMessages.PERMUTATION_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		decompositionGraph = permutation.getGraph().clone();
		PathDecomposition decomposition = new PathDecomposition();

		decomposition.width = 0;
		if(permutation.getSize()>0) {
			Bag previousBag = null;

			for(int i=permutation.getSize()-1;i>=0;i--) {
				Bag bag = previousBag==null ? decomposition.addBag() : decomposition.addBag(previousBag);
				decomposition.addNodeToBag(permutation.getNode(i), bag);

				for(Edge edge: decompositionGraph.getEdges()) {
					if(bag.getNodes().containsAll(edge.getIncidentNodes())) {
						decompositionGraph.removeEdge(edge);
					}
				}

				if(previousBag!=null) {
					for(Node node: previousBag.getNodes()) {
						if(decompositionGraph.contains(node)) {
							decomposition.addNodeToBag(node, bag);
						}
					}
				}

				for(Node node: bag.getNodes()) {
					if(bag.getNodes().containsAll(decompositionGraph.getAdjacentNodes(node))) {
						decompositionGraph.removeNode(node);
					}
				}
				decomposition.width = Math.max(decomposition.width, bag.getSize()-1);
				previousBag = bag;
			}
		}

		decomposition.graph = permutation.getGraph();
		return decomposition;
	}

	/**
	 * Creates a new path decomposition from the tree decomposition {@code decomposition}.
	 * @param treeDecomposition the tree decomposition
	 * @return the created path decomposition
	 */
	public static PathDecomposition getPathDecomposition(TreeDecomposition treeDecomposition) {
		if(treeDecomposition==null) {
			throw new NullPointerException(
					RavenExceptionMessages.DECOMPOSITION_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		PathDecomposition decomposition = new PathDecomposition();
		decomposition.graph = treeDecomposition.getGraph();
		decomposition.width = 0;

		Map<Bag,Bag> treeBagToPathBag = new HashMap<>();
		Map<Integer,Set<Bag>> bagNeighbors = new HashMap<>();

		//Create a "linearization" of the tree decomposition
		Bag prev = null;
		for(Bag treeBag: treeDecomposition) {
			Bag bag;

			if(prev==null) {
				bag = decomposition.addBag();
			}
			else {
				bag = decomposition.addBag(prev);
				prev.addNeighbor(bag);
			}
			bag.addNodeSet(treeBag.getNodes());
			treeBagToPathBag.put(treeBag, bag);
			bagNeighbors.put(bag.getIdentifier(), treeBag.getNeighbors());
			decomposition.width = Math.max(decomposition.width, bag.getSize()-1);
			prev = bag;
		}

		try {
			for(Bag bag: decomposition) {
				int index = decomposition.bagList.indexOf(bag);			

				for(Bag neighbor: bagNeighbors.get(bag.getIdentifier())) {
					Set<Node> commonNodes = new HashSet<>(bag.getNodes());
					commonNodes.retainAll(neighbor.getNodes());
					int neighborIndex = decomposition.bagList.indexOf(treeBagToPathBag.get(neighbor));

					for(int i=index+1;i<neighborIndex;i++) {
						decomposition.bagList.get(i).addNodeSet(commonNodes);
					}
				}
			}
		}
		catch (Exception e) {
			e.printStackTrace();
		}

		return decomposition;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new new empty path decomposition, i.e. a decomposition without any bag. 
	 */
	private PathDecomposition() {
		bagList = new ArrayList<>();
		nodeToBagMap = new HashMap<>();
		indexToBagMap = new HashMap<>();
		graph = new Graph("");
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Checks whether the bag {@code bag} is contained in this path decomposition.
	 * @param bag the bag to be checked
	 * @return {@code true} if the bag is contained, {@code false} otherwise
	 */
	public boolean contains(Bag bag) {
		return bagList.contains(bag);
	}

	@Override
	public int getWidth() {
		return width;
	}

	@Override
	public boolean isEmpty() {
		return bagList.isEmpty();
	}

	@Override
	public int getLength() {
		return bagList.size();
	}

	@Override
	public Iterator<Bag> iterator() {
		return bagList.listIterator();
	}

	@Override
	public Graph getGraph() {
		return graph;
	}

	@Override
	public Bag getBag(int index) {
		if(index<0 || index>=indexToBagMap.size()) {
			throw new IndexOutOfBoundsException();
		}
		
		return indexToBagMap.get(index);
	}

	@Override
	public int hashCode() {
		int hashCode = 1;

		for(Bag bag: this) {
			hashCode = 31*hashCode + ((bag==null) ? 0 : bag.hashCode());
		}
		return hashCode;
	}

	@Override
	public boolean equals(Object o) {
		if(o==null) {
			return false;
		}
		else if(o==this) {
			return true;
		}
		else if(o.getClass().equals(getClass())) {
			PathDecomposition decomposition = (PathDecomposition) o;
			if(bagList==null) {
				return decomposition.bagList==null;
			}

			return bagList.equals(decomposition.bagList);
		}
		else {
			return false;
		}
	}


	@Override
	public Data.Type getType() {
		return Type.pathDecomposition;
	}

	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();

		boolean isFirst = true;
		for(Bag bag: this) {
			if(isFirst) {
				isFirst = false;
			}
			else {
				builder.append(" -- ");
			}

			builder.append(bag.toString());
		}
		builder.append("\n");

		return builder.toString();
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Adds a new bag to this path decomposition with the bags {@code neighbors} as neighbors.
	 * @param neighbors the neighbors of the new bag
	 * @return the new bag
	 */
	private Bag addBag(Bag... neighbors) {		
		if(neighbors.length>2) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.NUMBER_OF_NEIGHBORS_MUST_BE_2_OR_LESS_EXCEPTION);
		}

		final int bagIndex = bagList.size();
		Bag bag = new Bag(bagIndex);

		bagList.add(bag);
		indexToBagMap.put(bagIndex, bag);

		for(Bag neighbor: neighbors) {
			if(neighbor==null) {
				throw new NullPointerException(
						RavenExceptionMessages.NO_NEIGHBOR_MUST_NOT_BE_NULL_EXCEPTION);
			}
			bag.addNeighbor(neighbor);
			indexToBagMap.get(bagIndex).addNeighbor(neighbor);
		}

		return bag;
	}

	/**
	 * Adds a node {@code node} to the bag {@code bag}.
	 * @param node the node to be added
	 * @param bag the bag to which the node is added
	 */
	private void addNodeToBag(Node node, Bag bag) {
		if(node==null) {
			throw new NullPointerException(
					RavenExceptionMessages.NODE_MUST_NOT_BE_NULL_EXCEPTION);
		}

		bag.addNode(node);
		if(!nodeToBagMap.containsKey(node)) {
			nodeToBagMap.put(node, bag);
		}
	}

}
