/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.data;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class implements a <i>permutation</i> on a set of nodes.
 * @author Christoph Blume
 * @version $Revision: 1.13 $
 *
 * @see Node
 * @see Graph
 */
public class Permutation implements Iterable<Node> {

	private final Graph graph;
	private final List<Node> nodeList;

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new permutation of the nodes of the graph {@code graph}.
	 * @param graph the graph of the permutation.
	 */
	public Permutation(Graph graph) {
		if(graph==null) {
			throw new NullPointerException(
					RavenExceptionMessages.GRAPH_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		this.graph = graph.clone();
		nodeList = new ArrayList<>();
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the graph of this permutation.
	 * @return the graph
	 */
	public Graph getGraph() {
		return graph;
	}

	/**
	 * Adds a node at the end of this permutation.
	 * @param node the node to be added
	 * 
	 * @throws NullPointerException thrown if the node {@code node} is {@code null}
	 * @throws IllegalArgumentException thrown if the node {@code node} does not belong to the 
	 * graph of this permutation
	 */
	public void addNode(Node node) {
		if(node==null) {
			throw new NullPointerException(
					RavenExceptionMessages.NODE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(!graph.contains(node)) {
			throw new IllegalArgumentException(RavenExceptionMessages.NODE_NOT_EXISTING_EXCEPTION);
		}
		if(nodeList.contains(node)) {
			throw new IllegalArgumentException(RavenExceptionMessages.NODE_ALREADY_ADDED_EXCEPTION);
		}
		
		nodeList.add(node);
	}
	

	/**
	 * Returns {@code true} if and only if the node {@code node} is contained in this permutation.
	 * @param node the node to be tested
	 * @return {@code true} if the node is contained, {@code false} otherwise
	 */
	public boolean contains(Node node) {
		return nodeList.contains(node);
	}
	
	/**
	 * Returns the {@code index}-th node of this permutation
	 * @param index the index of the node
	 * @return the {@code index}-th node
	 * @throws IndexOutOfBoundsException thrown if the index is less than 0 or greater or 
	 * equal to the permutation's size 
	 */
	public Node getNode(int index) throws IndexOutOfBoundsException {
		return nodeList.get(index);
	}
	
	/**
	 * Returns the size of the permutation, i.e. the number of nodes of this permutation.
	 * @return the size of the permutation
	 */
	public int getSize() {
		return nodeList.size();
	}
	
	/**
	 * Returns {@code true} if and only if the permutation is empty.
	 * @return {@code true} if the permutation is empty, {@code false} otherwise
	 */
	public boolean isEmpty() {
		return nodeList.isEmpty();
	}
	
	/**
	 * Returns the index of the node {@code node} or -1 if the node is not contained 
	 * in this permutation.
	 * @param node the node of which the index is returned
	 * @return the index of the node
	 */
	public int indexOf(Node node) {
		return nodeList.indexOf(node);
	}

	@Override
	public Iterator<Node> iterator() {
		return nodeList.listIterator();
	}
	
	@Override
	public String toString() {
		return nodeList.toString();
	}
	
}
