/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.data;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;

import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class implements an in-memory database which is used to store the 
 * {@code Data} objects available in Raven.
 * @author Christoph Blume
 * @version $Revision: 1.37 $
 *
 * @see Data
 * @see RepositoryListener
 */
public class Repository {

	/** The NAME Pattern of valid names */
	// If the NAME_PATTERN is changed, please also change the description "Message Properties Files".
	public static final String NAME_PATTERN = "\\w+";

	private static Repository repository;
	
	private Map<String, Data> objectMapping = null;
	private LinkedList<RepositoryListener> listenerList;

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the repository used for this instance of Raven. 
	 * @return the repository
	 */
	public static Repository getRepository() {
		if(repository==null) {
			repository = new Repository();
		}
		
		return repository;
	}
	
	/**
	 * Checks whether the name is correct with respect to the pattern of allowed names.
	 * @param name the name to be checked
	 * @return {@code true} if the name is valid, {@code false} otherwise
	 */
	public static boolean isNameValid(String name) {
		if(name==null) {
			return false;
		}
		
		return name.matches(NAME_PATTERN);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new empty repository.
	 */
	private Repository() {
		objectMapping = new HashMap<>();
		listenerList = new LinkedList<>();
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Adds the Repository Listener {@code l} to this repository.
	 * @param l the listener to receive repository events
	 */
	public void addRepositoryListener(RepositoryListener l) {
		listenerList.add(l);
	}
	
	/**
	 * Returns a set with the names of all automata contained in this repository.
	 * @return the names of all automata
	 */
	public Set<String> getAllAutomata() {		
		Set<String> automata = new HashSet<>();
		Set<String> keys = objectMapping.keySet();
		
		for(String key : keys){
			if(isAutomaton(key)){
				automata.add(key);				
			}
		}
		
		return automata;
	}
	
	/**
	 * Returns a set with the names of all graphs contained in this repository.
	 * @return the names of all graphs
	 */
	public Set<String> getAllGraphs() {		
		Set<String> graphs = new HashSet<>();
		Set<String> keys = objectMapping.keySet();
		
		for(String key : keys){
			if(isGraph(key)){
				graphs.add(key);				
			}
		}
		
		return graphs;
	}

	/**
	 * Returns the names of the data object contained in this repository.
	 * @return the names of the data objects
	 */
	public Set<String> getNames() {
		return objectMapping.keySet();
	}

	/**
	 * Returns {@code true} if and only if an object with name {@code name} is 
	 * contained in this repository.
	 * @param name the name of an object
	 * @return {@code true} if an object has been found, 
	 * {@code false} otherwise
	 */
	public boolean contains(String name) {
		return objectMapping.containsKey(name);
	}
	
	/**
	 * Returns {@code true} if and only if the object {@code data} is 
	 * contained in this repository.
	 * @param data the object
	 * @return {@code true} if the object has been found, 
	 * {@code false} otherwise
	 */
	public boolean contains(Data data) {
		return objectMapping.containsValue(data);
	}
	
	/**
	 * Returns {@code true} if and only if the object with 
	 * name {@code name} is of type #Automaton;
	 * @param name the name of the object
	 * @return {@code true} if the object is an automaton, 
	 * {@code false} otherwise
	 */
	public boolean isAutomaton(String name) {
		if(!contains(name)) {
			return false;
		}
		return get(name) instanceof Automaton;
	}
	
	/**
	 * Returns {@code true} if and only if the object with 
	 * name {@code name} is of type #Cospan;
	 * @param name the name of the object
	 * @return {@code true} if the object is an cospan, 
	 * {@code false} otherwise
	 */
	public boolean isCospan(String name) {
		if(!contains(name)) {
			return false;
		}
		return get(name) instanceof Cospan;
	}
	
	/**
	 * Returns {@code true} if and only if the object with 
	 * name {@code name} is of type #Decomposition;
	 * @param name the name of the object
	 * @return {@code true} if the object is a decomposition, 
	 * {@code false} otherwise
	 */
	public boolean isDecomposition(String name) {
		if(!contains(name)) {
			return false;
		}
		return get(name) instanceof Decomposition;
	}
	
	/**
	 * Returns {@code true} if and only if the object with 
	 * name {@code name} is of type #Graph;
	 * @param name the name of the object
	 * @return {@code true} if the object is a graph, 
	 * {@code false} otherwise
	 */
	public boolean isGraph(String name) {
		if(!contains(name)) {
			return false;
		}
		return get(name) instanceof Graph;
	}
	
	/**
	 * Returns {@code true} if and only if the object with 
	 * name {@code name} is of type #Signature;
	 * @param name the name of the object
	 * @return {@code true} if the object is a signature, 
	 * {@code false} otherwise
	 * 
	 */
	public boolean isSignature(String name) {
		if(!contains(name)) {
			return false;
		}
		return get(name) instanceof Signature;
	}
	
	/**
	 * Adds a new object {@code item} with the name {@code name}
	 * to this repository. 
	 * If the name is already used for another object, the other
	 * is replaced by the new object. If the new object is already 
	 * contained in this repository (with another name), the object 
	 * will <em>not</em> be added twice.
	 * 
	 * @param name the name of the new object
	 * @param item the new object
	 * @return {@code true} if the name-object pair has been added 
	 * successfully, {@code false} otherwise
	 * 
	 * @throws NullPointerException thrown if either the name
	 * or the object is {@code null}.
	 * @throws IllegalArgumentException thrown if the name is not valid
	 * @see #isNameValid(String)
	 */
	public boolean put(final String name, Data item) {
		if(name==null) {
			throw new NullPointerException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(item==null) {
			throw new NullPointerException(RavenExceptionMessages.DATA_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(!isNameValid(name)) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_NAME_EXCEPTION);
		}
//		if(contains(item)) {
//			return false;
//		}
		if(contains(name)){
			return false;
		}
		objectMapping.put(name, item);
		fireDataAddedEvent(name, item);
		return true;
	}
	
	/**
	 * Renames an existing data object with the name {@code name} to {@code replacement}. If the 
	 * operation has been successful, it returns {@code true}, otherwise, if no data object with the 
	 * given name can be found or the replacement name is already in use, {@code false} is returned.
	 * @param name the name of the data object to be renamed
	 * @param replacement the new name of the affected data set
	 * @return {@code true} if the operation has been successful, {@code false} otherwise
	 * 
	 * @throws NullPointerException thrown if either the name or the replacement is {@code null}.
	 * @throws IllegalArgumentException thrown if the name or the replacement is empty or the 
	 * replacement is not a valid name.
	 * @see #isNameValid(String)
	 */
	public boolean rename(String name, String replacement) {
		if(name==null) {
			throw new NullPointerException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(name.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		if(replacement==null) {
			throw new NullPointerException(
					RavenExceptionMessages.REPLACEMENT_NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(replacement.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.REPLACEMENT_NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		if(!isNameValid(replacement)) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_NAME_EXCEPTION);
		}
		
		if(!contains(name) || contains(replacement)) {
			return false;
		}
		
		objectMapping.put(replacement, objectMapping.remove(name));
		fireDataRenamedEvent(name, replacement);
		return true;
	}
	
	/**
	 * Returns the object with the name {@code name}.
	 * @param name the name of the requested object
	 * @return the requested object if an object of 
	 * this name is stored in the repository, {@code null}
	 * otherwise
	 */
	public Data get(String name) {
		return objectMapping.get(name);
	}
	
	/**
	 * Returns the name of object {@code data}.
	 * @param item the object whose name is requested 
	 * @return the name of the object if the object is 
	 * contained in this repository, {@code null} otherwise
	 */
	public String getName(Data item) {
		for(String name: getNames()) {
			if(get(name)==item) {
				return name;
			}
		}
		
		return null;
	}

	/**
	 * Removes the object with the name {@code name} 
	 * from the repository
	 * @param name the name of the object to be removed
	 */
	public void remove(String name) {
		if(name!=null && !name.isEmpty()) {
			Data data = objectMapping.remove(name);
			fireDataRemovedEvent(name, data);
		}
	}
	
	/**
	 * Sends a repository event to all registered listeners that a new data set has been added to 
	 * the repository.
	 * @param name the name of the affected data set
	 * @param data the data set which is affected
	 */
	private void fireDataAddedEvent(String name, Data data) {
		for(RepositoryListener l: listenerList) {
			l.dataAdded(name, data);
		}
	}
	
	/**
	 * Sends a repository event to all registered listeners that a data set has been removed from
	 * the repository.
	 * @param name the name of the affected data set
	 * @param data the data set which is affected
	 */
	private void fireDataRemovedEvent(String name, Data data) {
		for(RepositoryListener l: listenerList) {
			l.dataRemoved(name, data);
		}
	}
	
	/**
	 * Sends a repository event to all registered listeners to inform about 
	 * the renaming of a data object.
	 * @param name the name of the affected data set
	 * @param replacement the replacement name of data set which is affected
	 */
	private void fireDataRenamedEvent(String name, String replacement) {
		for(RepositoryListener l: listenerList) {
			l.dataRenamed(name, replacement);
		}
	}
	
}
