/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.data;

/**
 * A wrapper class for cospan decompositions that are used to give two 
 * graphs with the same context as a counter-example for an algorithm  
 * @author Christoph Blume, Dennis Nolte
 * @version $Revision: 1.7 $
 */

public class RuleCounterExample implements CounterExample {
	
	private CospanDecomposition source;
	private CospanDecomposition target;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new rule counter-example
	 * @param leftHandSide the cospan decomposition of the left hand-side of the rule
	 * @param rightHandSide the cospan decomposition of the right hand-side of the rule
	 * @param context the cospan decomposition of the context of the two graphs
	 */
	public RuleCounterExample(CospanDecomposition leftHandSide, CospanDecomposition rightHandSide,
			CospanDecomposition context) {
		source = new CospanDecomposition(leftHandSide);
		source.add(context);

		target = new CospanDecomposition(rightHandSide);
		target.add(context);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public int hashCode() {
		return source.hashCode() + 17*target.hashCode();
	}
	
	@Override
	public boolean equals(Object o) {
		if(o==null) {
			return false;
		}
		else if(o==this) {
			return true;
		}
		else if(o.getClass().equals(getClass())) {
			RuleCounterExample counterExample = (RuleCounterExample) o;
			return counterExample.source.equals(source) && counterExample.target.equals(target);
		}
		else {
			return false;
		}
	}
	
	/**
	 * Returns the cospan decomposition of the left hand-side used as a counter-example
	 * @return the cospan decomposition of the left hand-side used as a counter-example
	 */
	public CospanDecomposition getSource() {
		return new CospanDecomposition(source);
	}

	/**
	 * Returns the cospan decomposition of the right hand-side used as a counter-example
	 * @return the cospan decomposition of the right hand-side used as a counter-example
	 */
	public CospanDecomposition getTarget() {
		return new CospanDecomposition(target);
	}
	

	@Override
	public Type getType() {
		return Type.ruleCounterExample;
	}
	
	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();
		builder.append("source: "+source.toString()+System.lineSeparator());
		builder.append("target: "+target.toString()+System.lineSeparator());
		
		return builder.toString();
	}
	
}

