/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.data;

import java.io.Serializable;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class implements a <emph>signature</emph> which is used to define the 
 * alphabet of an automaton.  
 * @author Christoph Blume
 * @version $Revision: 1.31 $
 *
 * @see Automaton
 */
public class Signature implements Data, Iterable<Operation>, Serializable {

	private static final long serialVersionUID = -6906127082987537319L;

	/**
	 * Enumeration of the different types of implemented signatures.
	 * @author Christoph Blume
	 * @version $Revision: 1.31 $
	 *
	 */
	public enum Type implements Data.Type {
		/** Signature Type */
		signature;
	}
	
	private Set<Operation> letters;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new empty signature.
	 */
	public Signature() {
		letters = new HashSet<>();
	}
	
	/**
	 * Creates a new signature by copying all letters of the signature {@code signature}.
	 * @param signature the signature to be copied.
	 */
	public Signature(Signature signature) {
		letters = new HashSet<>(signature.letters);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Adds a new letter to this signature
	 * @param letter the letter to be added
	 */
	public void add(Operation letter) {
		letters.add(letter);
	}

	@Override
	public Iterator<Operation> iterator() {
		return letters.iterator();
	}
	
	/**
	 * Returns the set of letters of this signature
	 * @return the set of letters
	 */
	public Set<Operation> getOperations() {
		return new HashSet<>(letters);
	}

	/**
	 * Returns {@code true} if and only if the signature {@code signature} is
	 * contained in this signature, i.e. every letter of the signature {@code signature}
	 * is also an letter of this signature. 
	 * @param signature the signature to be checked
	 * @return {@code true} if the signature is contained, {@code false} otherwise
	 */
	public boolean contains(Signature signature) {
		if(signature==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SIGNATURE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		for(Operation letter: signature) {
			if(!letters.contains(letter)) {
				return false;
			}
		}

		return true;
	}

	/**
	 * Returns {@code true} if and only if letter {@code letter} is 
	 * contained in this signature.
	 * @param letter the letter to be checked
	 * @return {@code true} if the letter is contained, {@code false} otherwise
	 */
	public boolean contains(Operation letter) {		
		return letters.contains(letter);
	}

	/**
	 * Returns the number of letter of this signature.
	 * @return the number of letters
	 */
	public int size() {
		return letters.size();
	}
	
	@Override
	public int hashCode() {
		return 31*letters.hashCode();
	}

	@Override
	public boolean equals(Object o) {
		if(o==null) {
			return false;
		}
		else if(o==this) {
			return true;
		}
		else if(o.getClass().equals(getClass())) {
			Signature signature = (Signature) o;
			if(signature.size()!=size()) {
				return false;
			}
			
			for(Operation letter: signature){
				if(!contains(letter)) {
					return false;
				}
			}
			
			return true;
		}
		else {
			return false;
		}
	}
	
	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();
		for(Operation letter: this) {
			if(builder.length()!=0) {
				builder.append(", ");
			}
			
			builder.append(letter);
		}
		
		return builder.toString();
	}

	@Override
	public Type getType() {
		return Type.signature;
	}
	
}
