/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.data;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * A <emph>Tree Decomposition</emph> of some graph <i>G</i> is a pair <i>&#x1D4AF; = 
 * (T,X)</i> where <i>T</i> is a tree and <i>X</i> is a family of sets of nodes 
 * (of the graph <i>G</i>) which are indexed by (tree) nodes of <i>T</i>, such 
 * that the following conditions are satisfied:
 * <ul>
 * <li>for each (graph) node <i>v &isin; G</i>, there exists a node <i>t</i> of 
 * <i>T</i> such that <i>v &isin; X<sub>t</sub></i></li>
 * <li>for each (graph) edge <i>e &isin; G</i>, there is a node <i>t</i> of 
 * <i>T</i> such that all nodes <i>v</i> attached to <i>e</i> are in 
 * <i>X<sub>t</sub></i></li>
 * <li>for each node <i>v &isin; G</i>, the graph induced by the nodes <i>{t 
 * &#8739 v &isin; X<sub>t</sub>}</i> is a subtree of <i>T</i>.</li>
 * </ul>

 * @author Christoph Blume
 * @version $Revision: 1.34 $
 *
 */
public class TreeDecomposition implements GraphDecomposition, Cloneable {

	private Graph graph;
	private Bag root;
	private int width = 0;
	private List<Bag> bagList;
	private Map<Node, Bag> nodeToBagMap = new HashMap<>();
	private Map<Integer, Bag> indexToBagMap = new HashMap<>();

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new tree decomposition from the permutation {@code permutation}.
	 * @param permutation the permutation
	 * @return the created tree decomposition
	 */
	public static TreeDecomposition getTreeDecomposition(Permutation permutation) {
		if(permutation==null) {
			throw new NullPointerException(
					RavenExceptionMessages.PERMUTATION_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(permutation.isEmpty()) {
			return new TreeDecomposition();
		}
		
		TreeDecomposition decomposition = permutationToDecomposition(permutation,
				permutation.getGraph(), 0);
		decomposition.graph = permutation.getGraph();

		return decomposition;
	}

	/**
	 * Converts a node permutation {@code permutation} into a tree decomposition starting at the 
	 * {@code permIndex}-th node of the permutation.
	 * @param permutation the node permutation
	 * @param permIndex the index of the node to be used
	 * @return the tree decomposition
	 */
	private static TreeDecomposition permutationToDecomposition(Permutation permutation, 
			Graph graph, int permIndex) {
		int size = graph.getNodeCount();

		if (size == 0) {
			TreeDecomposition decomposition = new TreeDecomposition();
			decomposition.addBag();
			decomposition.width = 0;
			return decomposition;
		}
		else if (size == 1) {
			TreeDecomposition decomposition = new TreeDecomposition();
			Bag bag = decomposition.addBag();

			decomposition.width = 0;
			for(Node node: graph.getNodes()) {
				decomposition.addNodeToBag(node, bag);
			}
			return decomposition;
		}
		else {
			TreeDecomposition decomposition;
			Graph anotherGraph = graph.clone();
			Node node = permutation.getNode(permIndex);
			
			int numNeighbors = graph.getNumberOfAdjacentNodes(node);
			Set<Node> neighbors = graph.getAdjacentNodes(node);

			//Eliminate current node (i.e. the current node will be removed and 
			//its neighbors will become a clique)
			anotherGraph.removeNode(node);
			
			for(Node source: neighbors) {
				for(Node target: neighbors) {
					if(source!=target && !anotherGraph.isAdjacentTo(source, target)) {
						List<Node> adjacentNodes = new LinkedList<>();
						adjacentNodes.add(source);
						adjacentNodes.add(target);

						String edgeName = "e:"+source.getName()+"->"+target.getName();
						anotherGraph.addEdge(edgeName, adjacentNodes);
					}
				}
			}

			//go into recursion
			decomposition = permutationToDecomposition(permutation, anotherGraph, permIndex+1);
			//create a bag and add it to the decomposition
			Bag bag = decomposition.addBag();
			decomposition.addNodeToBag(node, bag);

			int lowestIndex = Integer.MAX_VALUE;
			Node lowestNeighbor = null;
			for (Node other: neighbors){
				decomposition.addNodeToBag(other, bag);
				int nodeIndex = permutation.indexOf(other);
				if (nodeIndex < lowestIndex) {
					lowestIndex = nodeIndex;
					lowestNeighbor = other;
				}
			}

			//create an edge between the bag and the neighborhood bag
			if(lowestNeighbor != null) {
				Bag neighborBag = decomposition.nodeToBagMap.get(lowestNeighbor);
				neighborBag.addNeighbor(bag);
				bag.addNeighbor(neighborBag);
			}
			//calculate the upper bound
			decomposition.width = Math.max(numNeighbors, decomposition.width);
			return decomposition;
		}
	}

	/**
	 * Traverses the (tree of the) tree decomposition (given by the bag {@code root}), starting
	 * at the root {@code root}. The linearization of the tree decomposition is saved in the 
	 * list {@code list}.
	 * @param root the root of the tree decomposition
	 * @param list the list in which the linearization of the tree decomposition is saved
	 * @return the linearization of the tree decomposition
	 */
	private static List<Bag> traverse(Bag root, List<Bag> list) {
		//Add root
		list.add(root);

		//Traverse neighbors
		for (Bag neighbor : root.getNeighbors()) {
			if(!list.contains(neighbor)) {
				traverse(neighbor, list);
			}
		}

		// end of current traversal
		return list;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a empty tree decomposition, i.e. a decomposition without any bag. 
	 */
	private TreeDecomposition() {
		bagList = new ArrayList<>();
		graph = new Graph("");
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public Iterator<Bag> iterator() {
		if(isEmpty()) {
			return Collections.<Bag>emptyList().iterator();
		}

		List<Bag> list = new LinkedList<>();
		List<Bag> remainingBags = new LinkedList<>(bagList);
		Bag bag;
		
		while(!remainingBags.isEmpty()) {
			bag = remainingBags.get(0);
			traverse(bag, list);
			remainingBags.removeAll(list);
		}
		
		return list.listIterator();
		
	}

	@Override
	public Graph getGraph() {
		return graph;
	}

	@Override
	public int getWidth() {
		return width;
	}

	@Override
	public boolean isEmpty() {
		return bagList.isEmpty();
	}

	@Override
	public int getLength() {
		return bagList.size();
	}

	@Override
	public Bag getBag(int index) {
		if(index<0 || index>=indexToBagMap.size()) {
			throw new IndexOutOfBoundsException();
		}
		
		return indexToBagMap.get(index);
	}

	@Override
	public int hashCode() {
		int hashCode = 1;

		for(Bag bag: this) {
			hashCode = 31*hashCode + ((bag==null) ? 0 : bag.hashCode());
		}
		return hashCode;
	}

	@Override
	public boolean equals(Object o) {
		if(o==null) {
			return false;
		}
		else if(o==this) {
			return true;
		}
		else if(o.getClass().equals(getClass())) {
			TreeDecomposition decomposition = (TreeDecomposition) o;
			if(root==null && decomposition.root==null) {
				return true;
			}
			else if(root!=null && decomposition.root!=null) {
				if(bagList==null) {
					traverse(root, bagList);
				}
				if(decomposition.bagList==null) {
					traverse(decomposition.root, decomposition.bagList);
				}
				
				return bagList.equals(decomposition.bagList);
			}
			else {
				return false;
			}
		}

		return false;
	}

	@Override
	public Data.Type getType() {
		return Type.treeDecomposition;
	}

	@Override
	public String toString() {
		StringBuilder bags = new StringBuilder("Bags:\n");
		StringBuilder links = new StringBuilder("\nLinks:\n");

		for(Bag bag: this) {
			bags.append(bag.getIdentifier()+": "+bag.toString()+"\n");
			links.append(bag.getIdentifier()+": ");

			boolean isFirst = true;
			for(Bag neighbor: bag.getNeighbors()) {
				if(isFirst) {
					isFirst = false;
				}
				else {
					links.append(", ");
				}

				links.append(neighbor.getIdentifier());
			}
			links.append("\n");
		}
		return bags.toString() + links.toString();
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	protected TreeDecomposition clone() {
		TreeDecomposition decomposition = new TreeDecomposition();
		decomposition.bagList = bagList;
		decomposition.graph = graph;
		decomposition.indexToBagMap = indexToBagMap;
		decomposition.nodeToBagMap = nodeToBagMap;
		decomposition.width = width;

		return decomposition;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Adds a new bag to this tree decomposition with the bags {@code neighbors} as neighbors.
	 * @param neighbors the neighbors of the new bag
	 * @return the new bag
	 */
	private Bag addBag(Bag... neighbors) {
		final int bagIndex = bagList.size();
		Bag bag = new Bag(bagIndex);
		if(root==null) {
			root = bag;
		}

		for(Bag neighbor: neighbors) {
			if(neighbor==null) {
				throw new NullPointerException(
						RavenExceptionMessages.NO_NEIGHBOR_MUST_NOT_BE_NULL_EXCEPTION);
			}

			bag.addNeighbor(neighbor);
			neighbor.addNeighbor(bag);
		}

		bagList.add(bag);
		indexToBagMap.put(bagIndex, bag);
		return bag;
	}

	/**
	 * Adds a node {@code node} to the bag {@code bag}.
	 * @param node the node to be added
	 * @param bag the bag to which the node is added
	 */
	private void addNodeToBag(Node node, Bag bag) {
		if(node==null) {
			throw new NullPointerException(
					RavenExceptionMessages.NODE_MUST_NOT_BE_NULL_EXCEPTION);
		}

		bag.addNode(node);
		if(!nodeToBagMap.containsKey(node)) {
			nodeToBagMap.put(node, bag);
		}
	}

}
