/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.data.automata;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.Serializable;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.logging.Logger;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import java.util.zip.ZipOutputStream;

import de.uni_due.inf.ti.raven.actions.ComputationEvent;
import de.uni_due.inf.ti.raven.actions.ComputationEvent.EventType;
import de.uni_due.inf.ti.raven.actions.ComputationListener;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Data;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.Operation.Connect;
import de.uni_due.inf.ti.raven.data.Operation.Fuse;
import de.uni_due.inf.ti.raven.data.Operation.Perm;
import de.uni_due.inf.ti.raven.data.Operation.Res;
import de.uni_due.inf.ti.raven.data.Operation.Trans;
import de.uni_due.inf.ti.raven.data.Operation.Vertex;
import de.uni_due.inf.ti.raven.data.Signature;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDDomain;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDFactory.BDDOperator;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDPairing;
import de.uni_due.inf.ti.raven.encodings.BDDEncoding;
import de.uni_due.inf.ti.raven.encodings.BDDEncoding.StateType;
import de.uni_due.inf.ti.raven.main.RavenMain;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This abstract class is the base class for all symbolic graph automata.
 * 
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.89 $
 * 
 */
public abstract class Automaton implements Data, Cloneable, Serializable {

	private static final long serialVersionUID = -8862823160145702503L;

	/** */
	public static final String AUTOMATON_MAIN_FILE = "general." + RavenMain.SERIALIZATION_FILE_EXTENSION;

	/** */
	public static final String STATES_FILE = "states." + RavenMain.BDD_FILE_EXTENSION;

	/** */
	public static final String INITIAL_STATES_FILE = "initial." + RavenMain.BDD_FILE_EXTENSION;

	/** */
	public static final String FINAL_STATES_FILE = "final." + RavenMain.BDD_FILE_EXTENSION;

	/** */
	public static final String NON_FINAL_STATES_FILE = "nonFinal." + RavenMain.BDD_FILE_EXTENSION;
	
	/** Automaton Type */
	protected final Type type;
	
	/** Automaton Encoding Name */
	protected final String encodingName;

	/** Automaton Encoding */
	transient protected BDDEncoding encoding;
	
	/** Automaton Encoding */
	protected final Signature signature;
	
	/** Automaton Maximum Interface Size */
	protected final int maximumInterface;
	
	/** The size of the inner interface */
	protected int innerInterface;
	
	/** The size of the outer interface */
	protected int outerInterface;

	/** Current State BDD Domain */
	transient protected RavenBDDDomain curStateDomain;

	/** Successor State BDD Domain */
	transient protected RavenBDDDomain sucStateDomain;

	/** State BDD */
	transient protected RavenBDD states;

	/** Initial State BDD */
	transient protected RavenBDD initialStates;

	/** Final State BDD */
	transient protected RavenBDD finalStates;

	/** Non-Final State BDD */
	transient protected RavenBDD nonFinalStates;

	/** Transition BDD Map */
	transient protected Map<Operation, RavenBDD> stateTransBDDs;

	/** The listener which have been registered to receive computation events of this automaton */
	transient protected LinkedList<ComputationListener> listener;
	
	transient private Logger logger;

	// needed number of variables
	private final int lengthEncoding;	

	private int startVariable;

	/*------------------------------------------------------------------------------------------*/
	/* 		Static Methods 																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns a BDD pairing which pairs the the interface bits of the two
	 * automata {@code first} and {@code second}.
	 * 
	 * @param first
	 *            the first automaton
	 * @param second
	 *            the second automaton
	 * @return the BDD pairing for the interfaces of the two automata
	 */
	public static final RavenBDDPairing getInterfacePairing(Automaton first,
			Automaton second) {
		if (first == null) {
			throw new NullPointerException(
					RavenExceptionMessages.AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if (second == null) {
			throw new NullPointerException(
					RavenExceptionMessages.AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if (first.getMaximumInterface() != second.getMaximumInterface()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.DIFFERENT_MAXIMUM_INTERFACES_EXCEPTION);
		}

		RavenBDDPairing pairing = RavenBDDPairing.makePair();
		for (int pos = 0; pos < first.encoding.getInterfaceLength(); pos++) {
			pairing.set(first.encoding.getBit(StateType.current, pos),
					second.encoding.getBit(StateType.current, pos));
		}

		return pairing;
	}

	/**
	 * Returns the natural number which is encoded by the BDD {@code bdd}.
	 * 
	 * @param bdd the bdd encoding the natural number
	 * @return the natural number encoded by the bdd or {@code -1} if the bdd is either 
	 * {@code Zero-BDD} or the {@code One-BDD}
	 */
	protected static final int getEncodedNumber(RavenBDD bdd) {
		int number = 0;
		int power = 1;

		if (!bdd.isOne() && !bdd.isZero()) {
			while (!bdd.isOne()) {
				if (!bdd.high().isZero()) {
					number += power;
					bdd = bdd.high();
				} else {
					bdd = bdd.low();
				}
				power <<= 1;
			}
			return number;
		}
		
		return -1;
	}
	
	/**
	 * Creates a new entry {@code name} in the zip stream {@code zos} and writes the content of 
	 * the file specified by {@code source} to the entry
	 * @param zos the {@code ZipOutputStream} of the underlying zip file
	 * @param name the name of the entry in the underlying zip file
	 * @param source the name of the source file
	 * @throws IOException thrown if an I/O error occurred
	 */
	private static void createZipEntry(ZipOutputStream zos, String name, String source)
			throws IOException {
		// put a new ZipEntry in the ZipOutputStream
		zos.putNextEntry(new ZipEntry(name));
		try(InputStream is = new FileInputStream(source)) {
			int size = 0;
			byte[] buffer = new byte[1024];
			
			// read data to the end of the source file and write it to the zip
			// output stream.
			while ((size = is.read(buffer, 0, buffer.length)) > 0) {
				zos.write(buffer, 0, size);
			}
			
			zos.closeEntry();
		};

		// delete the unzipped file
		File f = new File(source);
		if (!f.delete()) {
			f.deleteOnExit();
		}
	}

	/*------------------------------------------------------------------------------------------*/
	/* 		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new automaton.
	 * 
	 * @param type the type of the automaton
	 * @param encoding the encoding for the states
	 * @param signature the signature of this automaton
	 * 
	 * @see Automaton.Type
	 * @see ColorabilityAutomaton
	 * @see DominatingSetAutomaton
	 * @see NoIsolatedNodesAutomaton
	 * @see ProductAutomaton
	 * @see SubgraphAutomaton
	 * @see UnionAutomaton
	 * @see VertexCoverAutomaton
	 */
	protected Automaton(Type type, BDDEncoding encoding, Signature signature) {
		if(type==null) {
			throw new NullPointerException(
					RavenExceptionMessages.AUTOMATON_TYPE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(signature==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SIGNATURE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(encoding.getMaximumInterface()<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface>encoding.getMaximumInterface()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(outerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(outerInterface>encoding.getMaximumInterface()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		
		this.type = type;
		this.signature = signature;
		this.maximumInterface = encoding.getMaximumInterface();
		this.innerInterface = encoding.getInnerInterface();
		this.outerInterface = encoding.getOuterInterface();

		listener = new LinkedList<>();
		logger = Logger.getLogger(getClass().getName());

		this.encoding = encoding;
		encodingName = encoding.getClass().getName();
		lengthEncoding = encoding.getLength();

		curStateDomain = encoding.getCurrentStateDomain();
		sucStateDomain = encoding.getSuccessorStateDomain();

		startVariable = curStateDomain.getDomainBDD().var();

		states = encoding.zero();
		initialStates = encoding.zero();
		finalStates = encoding.zero();
		nonFinalStates = encoding.zero();

		stateTransBDDs = new HashMap<>();
		for (Operation operation : signature) {
			stateTransBDDs.put(operation, encoding.zero());
		}

		listener = new LinkedList<>();
		logger = Logger.getLogger(getClass().getName());
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Abstract Methods 																	*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public abstract Automaton clone() throws CloneNotSupportedException;

	/**
	 * Constructs BDDs which represent the set of initial and final states.
	 */
	protected abstract void createStates();

	/**
	 * Constructs BDDs which represent the set of initial and final states.
	 */
	protected abstract void createInitialAndFinalStates();

	/**
	 * Computes the BDDs which represent the transition relation of the
	 * automaton functor.
	 */
	protected abstract void createTransitions();

	/*------------------------------------------------------------------------------------------*/
	/* 		Public Methods 																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Enumeration of the different types of implemented automata.
	 * 
	 * @author Christoph Blume
	 * 
	 */
	public enum Type implements Data.Type {
		/** Colorability Automaton Type */
		colorability,
	
		/** Dominating Set Automaton Type */
		dominatingSet,

		/** Edge-counting Automaton Type */
		edgeCounting,
		
		/** Individual Automaton Type */
		individual,
		
		/** Link automaton Type */
		link,
		
		/** maximum-edge Automaton Type */
		maximumEdge,

		/** maximum-vertex Automaton Type */
		maximumVertex,
		
		/** minimum-edge Automaton Type */
		minimumEdge,
		
		/** minimum-vertex Automaton Type */
		minimumVertex,

		/** No isolated node Automaton Type */
		noIsolatedNodes,

		/** Path Automaton Type */
		path,
		
		/** Product Automaton Type */
		product,

		/** Subgraph Automaton Type */
		subgraph,

		/** Union Automaton Type */
		union,

		/** Vertex-counting Automaton Type */
		vertexCounting,

		/** Vertex Cover Automaton Type */
		vertexCover;
		
		
		/**
		 * Returns the automaton types of "basic" automata.
		 * 
		 * @return the basic automaton types
		 */
		public static Type[] getBasicTypes() {
			return new Type[] { colorability, dominatingSet, edgeCounting, link, 
					maximumEdge, maximumVertex, minimumEdge, minimumVertex, noIsolatedNodes, 
					path, product, subgraph, union, vertexCounting, vertexCover};
		}
		
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Adds the Computation Listener {@code l} to this automaton.
	 * 
	 * @param l the listener to receive computation events
	 */
	public void addComputationListener(ComputationListener l) {
		listener.add(l);
	}

	/**
	 * Remove the Computation Listener {@code l} from this automaton.
	 * @param l the listener which has been registered before
	 */
	public void removeComputationListener(ComputationListener l) {
		listener.remove(l);
	}
	
	@Override
	public final Type getType() {
		return type;
	}

	/**
	 * Returns the maximum interface of this automaton.
	 * @return the maximum interface
	 */
	public final int getMaximumInterface() {
		return maximumInterface;
	}
	
	/**
	 * Returns the size of the inner interface of this automaton.
	 * @return the size of the inner interface
	 */
	public final int getInnerInterface() {
		return innerInterface;
	}
	
	/**
	 * Returns the size of the outer interface of this automaton.
	 * @return the size of the outer interface
	 */
	public final int getOuterInterface() {
		return outerInterface;
	}

	/**
	 * Returns the alphabet of this automaton.
	 * @return the alphabet
	 */
	public final Signature getAlphabet() {
		return signature;
	}
	
	/**
	 * Returns the logger; For saving purposes only
	 * @return the logger of the automaton
	 */
	public final Logger getLogger() {
		return logger;
	}
	
	/**
	 * Returns the listener; For saving purposes only
	 * @return the listener of the automaton
	 */
	public final LinkedList<ComputationListener> getListener() {
		return listener;
	}
	
	@Override
	public final String toString() {
		return type.toString();
	}
	
	/**
	 * 
	 */
	public void initializeEncoding() {
		curStateDomain = encoding.getCurrentStateDomain();
		sucStateDomain = encoding.getSuccessorStateDomain();
	}

	/**
	 * Returns the BDD encoding used by this automaton.
	 * @return the BDD encoding
	 */
	public final BDDEncoding getEncoding() {
		return encoding;
	}

	/**
	 * Returns the size of the automaton (in the number of BDD nodes used to
	 * encode the state set)
	 * @return the size of the automaton
	 */
	public final int getSize() {
		return getStates().nodeCount();
	}

	/**
	 * This method is called when an object of the class Automaton shall be
	 * saved.
	 * @param filename the name of the file
	 * @throws IOException thrown if an I/O error occurred
	 */
	public final void writeAutomatonToZip(String filename) throws IOException {
		// serialize the automaton
		try(OutputStream fos = new FileOutputStream(RavenMain.SAVE_DIR + AUTOMATON_MAIN_FILE);
				ObjectOutputStream o = new ObjectOutputStream(fos);
				ZipOutputStream zos = new ZipOutputStream(new FileOutputStream(filename))) {
			o.writeObject(this);

			// put the serialized automaton into the zip archive and remove the
			// unzipped file
			String name = AUTOMATON_MAIN_FILE;
			createZipEntry(zos, name, RavenMain.SAVE_DIR + name);
			
			// save all BDDs into the zip archive
			RavenBDD.saveBDD(states, zos, STATES_FILE);
			RavenBDD.saveBDD(initialStates, zos, INITIAL_STATES_FILE);
			RavenBDD.saveBDD(finalStates, zos, FINAL_STATES_FILE);
			RavenBDD.saveBDD(nonFinalStates, zos, NON_FINAL_STATES_FILE);
			for (Operation letter : getAlphabet()) {
				RavenBDD bdd = stateTransBDDs.get(letter);
				RavenBDD.saveBDD(bdd, zos, letter.toString() + "." + RavenMain.BDD_FILE_EXTENSION);
			}
			
			// Finish zip process
			zos.closeEntry();
		};
	}

	/**
	 * Restores all transient attributes of the automaton.
	 * @param zipname the name of the zip archive
	 * @throws IOException thrown if an I/O error occurred
	 */
	public final void readAutomatonFromZip(String zipname) throws IOException {
		logger = Logger.getLogger(getClass().getName());
		listener = new LinkedList<>();

		if (!zipname.endsWith("."+RavenMain.AUTOMATON_FILE_EXTENSION)) {
			zipname = zipname + "." + RavenMain.AUTOMATON_FILE_EXTENSION;
		}

		//Restore the states from the zip archive
		try(ZipFile zipFile = new ZipFile(zipname)) {
			ZipEntry entry = zipFile.getEntry(STATES_FILE);		
			states = RavenBDD.loadBDD(zipFile.getInputStream(entry));
			
			entry = zipFile.getEntry(INITIAL_STATES_FILE);
			initialStates = RavenBDD.loadBDD(zipFile.getInputStream(entry));
			
			entry = zipFile.getEntry(FINAL_STATES_FILE);
			finalStates = RavenBDD.loadBDD(zipFile.getInputStream(entry));
			
			entry = zipFile.getEntry(NON_FINAL_STATES_FILE);
			nonFinalStates = RavenBDD.loadBDD(zipFile.getInputStream(entry));
			
			//Restore transitions from zip archive
			stateTransBDDs = new HashMap<>();
			for (Operation letter : getAlphabet()) {
				entry = zipFile.getEntry(letter + "." + RavenMain.BDD_FILE_EXTENSION);
				stateTransBDDs.put(letter, RavenBDD.loadBDD(zipFile.getInputStream(entry)));
			}
		};

		// ---------------------------------------------------------------------
		// the variable names need to be changed according to the new encoding
		// encoding was set during deserialization

		try {
			// Replace the variables in all BDDs to match the new encoding.
			RavenBDDPairing pairing = RavenBDDPairing.makePair();
			int[] oldVars = new int[2*lengthEncoding];
			int[] newVars = new int[2*lengthEncoding];
			for (int i = 0; i < oldVars.length; i++) {
				oldVars[i] = startVariable + i;
				if(i%2==0) {
					newVars[i] = curStateDomain.getVars()[i/2];
				}
				else {
					newVars[i] = sucStateDomain.getVars()[i/2];
				}
			}
			pairing.set(oldVars, newVars);
			startVariable = curStateDomain.getVars()[0];
			
			states = states.replace(pairing);
			initialStates = initialStates.replace(pairing);
			finalStates = finalStates.replace(pairing);
			nonFinalStates = nonFinalStates.replace(pairing);
			
			for (Operation op : signature){
				stateTransBDDs.put(op, stateTransBDDs.get(op).replace(pairing));
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Returns the state name of a single state encoded as bdd. The state name
	 * will be of the form <INTERFACE_SIZE>-<STATE_NUMBER>.
	 * @param state the bdd encoding the state
	 * @return the state name
	 */
	public final String getStateName(RavenBDD state) {
		if (state == null) {
			throw new NullPointerException(RavenExceptionMessages.BDD_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if (state.satCount(curStateDomain.getDomainBDD()) > 1) {
			throw new IllegalArgumentException(
					"The state-bdd must encode only a single state.");
		}

		RavenBDD stateNumberBDD = state.exists(encoding
				.getInterfaceEncodingBDD());
		int interfaceSize = getEncodedNumber(getCurrentInterfaceSize(state));
		int stateNumber = getEncodedNumber(stateNumberBDD);

		return interfaceSize + "-" + stateNumber;
	}

	/**
	 * Returns the state set encoded as an BDD.
	 * @return the state set
	 */
	public final RavenBDD getStates() {
		return states.clone();
	}

	/**
	 * Returns the state set for the interface of size {@code size} encoded as an BDD.
	 * @param size the size of the interface
	 * @return the {@code size}-th state set
	 */
	public final RavenBDD getStates(int size) {
		if (size < 0 || size > maximumInterface) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}

		return getStates().and(
				encoding.getInterfaceSize(StateType.current, size));
	}

	/**
	 * Returns the Transition BDD Map For saving purposes only
	 * @return the transition BDD map of the automaton
	 */
	public final Map<Operation, RavenBDD> getStateTransBDDs() {
		return stateTransBDDs;
	}

	/**
	 * Returns {@code true} if the state set {@code stateSet} is a subset of the
	 * initial state set.
	 * @param stateSet the set of states
	 * @return {@code true} if the state set contains only initial states,
	 *         {@code false} otherwise
	 * 
	 * @see #getInitialStates()
	 * @see #isAcceptingStateSet(RavenBDD)
	 */
	public final boolean isInitialStateSet(RavenBDD stateSet) {
		return !stateSet.isZero() && stateSet.clone().imp(getInitialStates()).isOne();
	}

	/**
	 * Returns the number of all states of this automaton.
	 * @return the number of states
	 */
	public final int getNumberOfStates() {
		return getStates().satCount(curStateDomain.getDomainBDD());
	}

	/**
	 * Returns the number of states of this automaton for the given interface
	 * size {@code size}.
	 * @param size the size of the interface
	 * @return the number of states
	 */
	public final int getNumberOfStates(int size) {
		return getStates(size).satCount(curStateDomain.getDomainBDD());
	}

	/**
	 * Returns the initial state set encoded as an BDD.
	 * @return the initial states
	 */
	public final RavenBDD getInitialStates() {
		return initialStates.clone();
	}

	/**
	 * Returns the number of initial states of this automaton.
	 * @return the number of initial states
	 */
	public final int getNumberOfInitialStates() {
		return getInitialStates().satCount(curStateDomain.getDomainBDD());
	}

	/**
	 * Returns {@code true} if the state set {@code stateSet} is a subset of the
	 * final state set.
	 * @param stateSet the set of states
	 * @return {@code true} if the state set contains only final states,
	 *         {@code false} otherwise
	 * 
	 * @see #getFinalStates()
	 * @see #isInitialStateSet(RavenBDD)
	 */
	public final boolean isAcceptingStateSet(RavenBDD stateSet) {
		RavenBDD tempBDD = stateSet.clone().and(getFinalStates());
		boolean result = tempBDD.isZero();
		tempBDD.free();

		return !result;
	}

	/**
	 * Returns {@code true} if the state set {@code stateSet} contains at least one
	 * initial state.
	 * @param stateSet the set of states
	 * @return {@code true} if the state set contains at least one initial state,
	 *         {@code false} otherwise
	 * 
	 * @see #getInitialStates()
	 * @see #isInitialStateSet(RavenBDD)
	 */
	public final boolean containsInitialState(RavenBDD stateSet) {
		RavenBDD tempBDD = stateSet.clone().and(getInitialStates());
		boolean result = tempBDD.isZero();
		tempBDD.free();

		return !result;
	}

	/**
	 * Returns the final state set encoded as an BDD.
	 * @return the final states
	 */
	public final RavenBDD getFinalStates() {
		return finalStates.clone();
	}

	/**
	 * Returns the number of final states of this automaton.
	 * @return the number of final states
	 */
	public final int getNumberOfFinalStates() {
		return getFinalStates().satCount(curStateDomain.getDomainBDD());
	}

	/**
	 * Returns the non-final state set encoded as an BDD.
	 * @return the non-final states
	 */
	public final RavenBDD getNonFinalStates() {
		return nonFinalStates.clone();
	}

	/**
	 * Returns the set of non-final states for the interface of size
	 * {@code size} encoded as an BDD.
	 * @param size the size of the interface
	 * @return the {@code size}-th set of non-final states
	 */
	public final RavenBDD getNonFinalStates(int size) {
		if (size < 0 || size > maximumInterface) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}

		return getNonFinalStates().and(
				encoding.getInterfaceSize(StateType.current, size));
	}

	/**
	 * Returns the number of non-final states of this automaton.
	 * @return the number of non-final states
	 */
	public final int getNumberOfNonFinalStates() {
		return getNonFinalStates().satCount(curStateDomain.getDomainBDD());
	}

	/**
	 * Returns the number of non-final states of this automaton for the given
	 * interface size {@code size}.
	 * @param size the size of the interface
	 * @return the number of non-final states
	 */
	public final int getNumberOfNonFinalStates(int size) {
		return getNonFinalStates(size).satCount(curStateDomain.getDomainBDD());
	}

	/**
	 * Returns {@code true} if the letter {@code letter} is applicable to the
	 * states {@code states}, i.e. there exists at least one successor state
	 * according to the transition relation for the letter {@code letter}.
	 * @param letter the input letter
	 * @param states the current states
	 * @return {@code true} if the letter is applicable, {@code false} otherwise
	 * 
	 * @see #getTransitionRelation(Operation)
	 * @see #getSuccessorStates(RavenBDD, Operation)
	 */
	public final boolean isApplicable(Operation letter, RavenBDD states) {
		if (!signature.contains(letter)) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INAPPLICABLE_OPERATION_EXCEPTION);
		}

		RavenBDD interfaceBDD = getCurrentInterfaceSize(states);
		RavenBDD tempBDD = getCurrentInterfaceSize(states);
		if (letter instanceof Connect) {
			int arity = ((Connect) letter).getArity();
			tempBDD = encoding.getInterfaceSizeIsGreaterOrEqual(
					StateType.current, arity);
		} else if (letter instanceof Fuse || letter instanceof Perm
				|| letter instanceof Trans) {
			tempBDD = encoding.getInterfaceSizeIsGreaterOrEqual(
					StateType.current, 2);
		} else if (letter instanceof Res) {
			tempBDD = encoding.getInterfaceSizeIsGreaterOrEqual(
					StateType.current, 1);
		} else if (letter instanceof Vertex) {
			tempBDD = encoding.getInterfaceSizeIsLessOrEqual(StateType.current,
					maximumInterface - 1);
		}
		return interfaceBDD.imp(tempBDD).isOne();
	}

	/**
	 * Returns the transition relation for the given letter
	 * @param letter the letter for which the transition relation is returned
	 * @return the transition relation represented as BDD
	 */
	public final RavenBDD getTransitionRelation(Operation letter) {
		if (!signature.contains(letter)) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INAPPLICABLE_OPERATION_EXCEPTION);
		}

		return stateTransBDDs.get(letter).clone();
	}

	/**
	 * Returns the interface size of the states in the given state set {@code states}.
	 * @param states the state set
	 * @return the interface size
	 */
	public final RavenBDD getCurrentInterfaceSize(RavenBDD states) {
		return states.clone().exists(encoding.getStateEncodingBDD());
	}

	/**
	 * Returns the set of reachable states for a given graph decomposition
	 * starting with the initial states.
	 * @param decomposition the input graph decomposition
	 * @return the set of states reachable by processing the graph decomposition
	 */
	public final RavenBDD getReachableStates(CospanDecomposition decomposition) {
		if(decomposition==null) {
			throw new NullPointerException(
					RavenExceptionMessages.DECOMPOSITION_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if (!signature.contains(decomposition.getSignature())) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.DECOMPOSITION_INCOMPATIBLE_TO_ALPHABET_EXCEPTION);
		}

		// Log entering
		logger.entering(getClass().getName(), "getReachableStates",
				decomposition);

		// We start with the initial states
		RavenBDD reachableStates = getInitialStates();
		int counter = 0;
		for (Operation letter : decomposition) {
			RavenBDD currentStates = reachableStates;
			reachableStates = getSuccessorStates(currentStates, letter);
			currentStates.free();
			//Fire computation event
			fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED, 
					"Process "+letter.toString(), decomposition.getLength(), ++counter));
		}

		// Log exiting
		logger.exiting(getClass().getName(), "getReachableStates", reachableStates);
		return reachableStates;
	}

	/**
	 * Returns the set of successor states for a given set of current states and
	 * and a given input letter (i.e. a graph operation).
	 * @param currentStates the current states
	 * @param letter the input letter
	 * @return the successor states (depeding on the current states and the
	 *         input letter)
	 */
	public final RavenBDD getSuccessorStates(RavenBDD currentStates, Operation letter) {
		if (!signature.contains(letter)) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INAPPLICABLE_OPERATION_EXCEPTION);
		}

		// Log entering
		logger.entering(getClass().getName(), "getReachableStates",
				new Object[] { currentStates, letter });

		// Get transition BDD for the next graph operation
		RavenBDD successorStates = getTransitionRelation(letter);
		successorStates = currentStates.applyEx(successorStates,
				BDDOperator.and, curStateDomain.getDomainBDD());

		// Nachfolge-Zustaende der ersten Relation ersetzen
		RavenBDDPairing pairing = RavenBDDPairing.makePair();
		pairing.set(sucStateDomain, curStateDomain);
		successorStates = successorStates.replace(pairing);

		// Log exiting
		logger.exiting(getClass().getName(), "getReachableStates",
				successorStates);
		return successorStates;
	}

	/**
	 * Returns the set of predecessor states for a given set of current states
	 * and and a given input letter (i.e. a graph operation).
	 * @param currentStates the current states
	 * @param letter the input letter
	 * @return the predecessor states (depeding on the current states and the
	 *         input letter)
	 */
	public final RavenBDD getPredecessorStates(RavenBDD currentStates, Operation letter) {
		if (!signature.contains(letter)) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INAPPLICABLE_OPERATION_EXCEPTION);
		}

		// Log entering
		logger.entering(getClass().getName(), "getPredecessorStates",
				new Object[] { currentStates, letter });

		// Switch the bdd nodes used to encode the "current" and the "successor"
		// state
		RavenBDDPairing pairing = RavenBDDPairing.makePair();
		pairing.set(curStateDomain, sucStateDomain);
		currentStates = currentStates.replace(pairing);

		// Get the transition relation (seen as relation between
		// "predecessor states" and
		// "current states") and quantify over the "current" states -> this
		// yields all predecessor states
		RavenBDD predecessorStates = getTransitionRelation(letter);
		predecessorStates = currentStates.applyEx(predecessorStates,
				BDDOperator.and, sucStateDomain.getDomainBDD());
		predecessorStates = predecessorStates.and(getStates());

		// Log exiting
		logger.exiting(getClass().getName(), "getPredecessorStates",
				predecessorStates);
		return predecessorStates;
	}

	/**
	 * Returns the set of states which have no successor state for the given
	 * {@code letter}
	 * @param letter the letter for which the transition function is checked
	 * @return the set of states without successors
	 */
	public final RavenBDD getStatesWithoutSuccessorFor(Operation letter) {
		if(letter==null) {
			throw new NullPointerException(
					RavenExceptionMessages.LETTER_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if (!signature.contains(letter)) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INAPPLICABLE_OPERATION_EXCEPTION);
		}

		// Log entering
		logger.entering(getClass().getName(), "getStatesWithoutSuccessorFor", letter);

		RavenBDD states = getTransitionRelation(letter).not();
		states = states.forAll(sucStateDomain.getDomainBDD());
		states = states.and(getStates());

		// Log exiting
		logger.exiting(getClass().getName(), "getStatesWithoutSuccessorFor", states);
		return states;
	}

	/**
	 * Prints a single state (seen as singleton state set) or a set of states.
	 * @param stateSet the state set to be printed
	 * @return the string representation of the state set
	 * 
	 * @see BDDEncoding#printStates(RavenBDD)
	 */
	public final String printStates(RavenBDD stateSet) {
		return encoding.printStates(stateSet);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods 																	*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Sets the initial states of this automaton to the states reachable from the cospan 
	 * decomposition {@code decomposition}.
	 * @param decomposition the cospan decomposition indicating the new initial states
	 * 
	 * @see #getInitialStates()
	 */
	protected final void setInitialStates(CospanDecomposition decomposition) {
		if(decomposition==null) {
			throw new NullPointerException(
					RavenExceptionMessages.DECOMPOSITION_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(!signature.contains(decomposition.getSignature())) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.DECOMPOSITION_INCOMPATIBLE_TO_ALPHABET_EXCEPTION);
		}
		if(decomposition.getWidth()>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.DECOMPOSITION_INCOMPATIBLE_EXCEPTION);
		}
		if(decomposition.getInnerInterfaceSize()!=innerInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.DECOMPOSITION_INCOMPATIBLE_EXCEPTION);
		}
				
		this.initialStates = getReachableStates(decomposition);
		innerInterface = decomposition.getOuterInterfaceSize();
	}

	/**
	 * Fires an computation event notifying all listeners.
	 * @param e the computation event
	 */
	protected final void fireComputationEvent(ComputationEvent e) {
		for (ComputationListener l : listener) {
			l.computationAdvanded(e);
		}
	}

	/*------------------------------------------------------------------------------------------*/
	/* 		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Writes this automaton to the specified ObjectOutputStream {@code out}.
	 * @param out the ObjectOutputStream to which the automaton is written
	 * @throws IOException thrown if an I/O error occurred
	 * 
	 * @see ObjectOutputStream#writeObject(Object)
	 */
	private void writeObject(ObjectOutputStream out) throws IOException {
		out.defaultWriteObject();
		out.writeObject(encoding.getClass().getName());
	}
}
