/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.data.automata;

import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.actions.ComputationListener;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Operation.Connect;
import de.uni_due.inf.ti.raven.data.Signature;
import de.uni_due.inf.ti.raven.encodings.BoundEncoding;
import de.uni_due.inf.ti.raven.encodings.ColorabilityEncoding;
import de.uni_due.inf.ti.raven.encodings.DominatingSetEncoding;
import de.uni_due.inf.ti.raven.encodings.EdgeCountingEncoding;
import de.uni_due.inf.ti.raven.encodings.LinkEncoding;
import de.uni_due.inf.ti.raven.encodings.NoIsolatedNodesEncoding;
import de.uni_due.inf.ti.raven.encodings.PathEncoding;
import de.uni_due.inf.ti.raven.encodings.ProductEncoding;
import de.uni_due.inf.ti.raven.encodings.SubgraphEncoding;
import de.uni_due.inf.ti.raven.encodings.UnionEncoding;
import de.uni_due.inf.ti.raven.encodings.VertexCountingEncoding;
import de.uni_due.inf.ti.raven.encodings.VertexCoverEncoding;
import de.uni_due.inf.ti.raven.main.RavenMain;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * Factory class for vending {@code Automaton} objects. This class provides methods for the creation
 * of the different graph automaton types implemented in Raven.
 * @author Christoph Blume
 * @version $Revision: 1.65 $
 * 
 * @see Automaton
 */
public class AutomatonFactory {

	private static AutomatonFactory factory;
	private static Logger logger;

	private List<ComputationListener> listener;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new factory for the creation of automata.
	 * @return the automaton factory
	 */
	public static AutomatonFactory getFactory() {
		if(factory==null) {
			factory = new AutomatonFactory();
			logger = Logger.getLogger(AutomatonFactory.class.getName());
		}
		
		return factory;
	}
	
	/**
	 * Creates a new automaton which is shifted over the cospan decomposition {@code decomposition}
	 * i.e. the initial states of the new automaton are states reachable by the original automaton
	 * after processing the cospan decomposition {@code decomposition}.
	 * 
	 * <i>Please note, that the original automaton is <em>NOT</em> altered.</i>
	 * @param automaton the automaton which is used for the shift
	 * @param decomposition the cospan decomposition over which the automaton is shifted
	 * @return the shifted automaton
	 */
	public static Automaton shiftAutomaton(final Automaton automaton, 
			final CospanDecomposition decomposition) {
		if(automaton==null) {
			throw new NullPointerException(
					RavenExceptionMessages.AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(decomposition==null) {
			throw new NullPointerException(
					RavenExceptionMessages.DECOMPOSITION_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(!automaton.getAlphabet().contains(decomposition.getSignature())) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.DECOMPOSITION_INCOMPATIBLE_TO_ALPHABET_EXCEPTION);
		}
		if(decomposition.getWidth()>automaton.getMaximumInterface()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.DECOMPOSITION_INCOMPATIBLE_EXCEPTION);
		}
		if(automaton.getInnerInterface()!=decomposition.getInnerInterfaceSize()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.DECOMPOSITION_INCOMPATIBLE_EXCEPTION);
		}
		
		Automaton shiftedAutomaton;
		try {
			shiftedAutomaton = automaton.clone();
			shiftedAutomaton.setInitialStates(decomposition);			
		} catch (CloneNotSupportedException e) {
			//Will not be thrown since the automaton class implements the cloneable interface
			throw new RuntimeException();
		}
		
		
		return shiftedAutomaton ;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new automaton factory.
	 */
	private AutomatonFactory() {
		listener = new LinkedList<>();
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new colorability automaton.
	 * @param encoding the colorability encoding on which this automaton is based on
	 * @param signature the signature on which this automaton is based on
	 * @return the automaton
	 */
	public ColorabilityAutomaton createColorabilityAutomaton(ColorabilityEncoding encoding, 
			Signature signature) {
		if(encoding==null) {
			throw new NullPointerException(
					RavenExceptionMessages.ENCODING_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(signature==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SIGNATURE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		//Log entering 
		logger.entering(getClass().getName(), "createColorabilityAutomaton", 
				new Object[]{encoding, signature});
		
		//Create automaton
		ColorabilityAutomaton automaton = new ColorabilityAutomaton(encoding, signature);
		initialize(automaton);
		logger.info("Colorability Automaton successfully initialized");
		
		//Log exiting
		logger.exiting(getClass().getName(), "createColorabilityAutomaton", automaton);
		return automaton;
	}

	/**
	 * Creates a new dominating set automaton.
	 * @param encoding the dominating set encoding on which this 
	 * automaton is based on
	 * @param signature the signature on which this automaton 
	 * is based on
	 * @return the automaton
	 */
	public DominatingSetAutomaton createDominatingSetAutomaton(DominatingSetEncoding encoding, 
			Signature signature) {
		if(encoding==null) {
			throw new NullPointerException(
					RavenExceptionMessages.ENCODING_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(signature==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SIGNATURE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		//Log entering 
		logger.entering(getClass().getName(), "createDominatingSetAutomaton", 
				new Object[]{encoding, signature});
		
		//Create automaton
		DominatingSetAutomaton automaton = new DominatingSetAutomaton(encoding, signature);
		initialize(automaton);
		logger.info("Dominating Automaton successfully initialized");
		
		//Log exiting
		logger.exiting(getClass().getName(), "createDominatingSetAutomaton", automaton);
		
		return automaton;
	}
	
	/**
	 * Creates a new edge-counting automaton counting only those edges which 
	 * are contained in the set {@code edgeSet}.
	 * @param encoding the edge-counting encoding on which this automaton is based on
	 * @param signature the signature on which this automaton is based on
	 * @param remainder the remainder for the number of edges
	 * @param edgeSet the set of edges which will be counted
	 * @return the automaton
	 */
	public EdgeCountingAutomaton createEdgeCountingAutomaton(EdgeCountingEncoding encoding, 
			Signature signature, int remainder, Set<Connect> edgeSet) {
		if(encoding==null) {
			throw new NullPointerException(
					RavenExceptionMessages.ENCODING_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(signature==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SIGNATURE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(remainder<0 || remainder>=encoding.getDivisor()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_REMAINDER_EXCEPTION);
		}
		if(edgeSet==null) {
			throw new NullPointerException(
					RavenExceptionMessages.EDGE_SET_MUST_NOT_BE_NULL_EXCEPTION);
		}
		for(Connect connect: edgeSet) {
			if(!signature.contains(connect)) {
				throw new IllegalArgumentException(
						RavenExceptionMessages.ILLEGAL_LETTER_EXCEPTION);
			}
		}
		
		//Log entering 
		logger.entering(getClass().getName(), "createEdgeCountingAutomaton", 
				new Object[]{encoding, signature});
		
		//Create automaton
		EdgeCountingAutomaton automaton = new EdgeCountingAutomaton(encoding, signature,
				remainder, edgeSet);
		initialize(automaton);
		logger.info("Edge-counting Automaton successfully initialized");
		
		//Log exiting
		logger.exiting(getClass().getName(), "createEdgeCountingAutomaton", automaton);
		return automaton;
	}
	
	/**
	 * Creates a new link automaton.
	 * @param encoding the link encoding on which this automaton is based on
	 * @param signature the signature on which this automaton is based on
	 * @return the automaton
	 */
	public LinkAutomaton createLinkAutomaton(LinkEncoding encoding, Signature signature) {
		if(encoding==null) {
			throw new NullPointerException(
					RavenExceptionMessages.ENCODING_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(signature==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SIGNATURE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		//Log entering 
		logger.entering(getClass().getName(), "createLinkAutomaton", 
				new Object[]{encoding, signature});
		
		//Create automaton
		LinkAutomaton automaton = new LinkAutomaton(encoding, signature);
		initialize(automaton);
		logger.info("Link Automaton successfully initialized");
		
		//Log exiting
		logger.exiting(getClass().getName(), "createLinkAutomaton", automaton);
		return automaton;
	}

	/**
	 * Creates a new minimum-vertex automaton.
	 * @param encoding the minimum-vertex encoding on which this automaton is based on
	 * @param signature the signature on which this automaton is based on
	 * @return the automaton
	 */
	public MinimumVertexAutomaton createMinimumVertexAutomaton(BoundEncoding encoding, 
			Signature signature) {
		if(encoding==null) {
			throw new NullPointerException(
					RavenExceptionMessages.ENCODING_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(signature==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SIGNATURE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		//Log entering 
		logger.entering(getClass().getName(), "createMinimumVertexAutomaton", 
				new Object[]{encoding, signature});
		
		//Create automaton
		MinimumVertexAutomaton automaton = new MinimumVertexAutomaton(encoding, signature);
		initialize(automaton);
		logger.info("Minimum-vertex Automaton successfully initialized");
		
		//Log exiting
		logger.exiting(getClass().getName(), "createMinimumVertexAutomaton", automaton);
		return automaton;
	}
	
	/**
	 * Creates a new maximum-edge automaton.
	 * @param encoding the maximum-edge encoding on which this automaton is based on
	 * @param signature the signature on which this automaton is based on
	 * @return the automaton
	 */
	public MaximumEdgeAutomaton createMaximumEdgeAutomaton(BoundEncoding encoding, 
			Signature signature) {
		if(encoding==null) {
			throw new NullPointerException(
					RavenExceptionMessages.ENCODING_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(signature==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SIGNATURE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		//Log entering 
		logger.entering(getClass().getName(), "createMaximumEdgeAutomaton", 
				new Object[]{encoding, signature});
		
		//Create automaton
		MaximumEdgeAutomaton automaton = new MaximumEdgeAutomaton(encoding, signature);
		initialize(automaton);
		logger.info("Maximum-edge Automaton successfully initialized");
		
		//Log exiting
		logger.exiting(getClass().getName(), "createMaximumEdgeAutomaton", automaton);
		return automaton;
	}

	/**
	 * Creates a new maximum-vertex automaton.
	 * @param encoding the maximum-vertex encoding on which this automaton is based on
	 * @param signature the signature on which this automaton is based on
	 * @return the automaton
	 */
	public MaximumVertexAutomaton createMaximumVertexAutomaton(BoundEncoding encoding, 
			Signature signature) {
		if(encoding==null) {
			throw new NullPointerException(
					RavenExceptionMessages.ENCODING_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(signature==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SIGNATURE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		//Log entering 
		logger.entering(getClass().getName(), "createMaximumVertexAutomaton", 
				new Object[]{encoding, signature});
		
		//Create automaton
		MaximumVertexAutomaton automaton = new MaximumVertexAutomaton(encoding, signature);
		initialize(automaton);
		logger.info("Maximum-vertex Automaton successfully initialized");
		
		//Log exiting
		logger.exiting(getClass().getName(), "createMaximumVertexAutomaton", automaton);
		return automaton;
	}

	/**
	 * Creates a new minimum-edge automaton.
	 * @param encoding the minimum-edge encoding on which this automaton is based on
	 * @param signature the signature on which this automaton is based on
	 * @return the automaton
	 */
	public MinimumEdgeAutomaton createMinimumEdgeAutomaton(BoundEncoding encoding, 
			Signature signature) {
		if(encoding==null) {
			throw new NullPointerException(
					RavenExceptionMessages.ENCODING_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(signature==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SIGNATURE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		//Log entering 
		logger.entering(getClass().getName(), "createMinimumEdgeAutomaton", 
				new Object[]{encoding, signature});
		
		//Create automaton
		MinimumEdgeAutomaton automaton = new MinimumEdgeAutomaton(encoding, signature);
		initialize(automaton);
		logger.info("Minimum-edge Automaton successfully initialized");
		
		//Log exiting
		logger.exiting(getClass().getName(), "createMinimumEdgeAutomaton", automaton);
		return automaton;
	}
	
	/**
	 * Creates a new isolated nodes automaton.
	 * @param encoding the isolated nodes encoding on which this 
	 * automaton is based on
	 * @param signature the signature on which this automaton is based on
	 * @return the automaton
	 */
	public NoIsolatedNodesAutomaton createNoIsolatedNodesAutomaton(NoIsolatedNodesEncoding encoding, 
			Signature signature) {
		if(encoding==null) {
			throw new NullPointerException(
					RavenExceptionMessages.ENCODING_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(signature==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SIGNATURE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		//Log entering 
		logger.entering(getClass().getName(), "createNoIsolatedNodesAutomaton", 
				new Object[]{encoding, signature});
		
		//Create automaton
		NoIsolatedNodesAutomaton automaton = new NoIsolatedNodesAutomaton(encoding, signature);
		initialize(automaton);
		logger.info("Isolated Node Automaton successfully initialized");
		
		//Log exiting
		logger.exiting(getClass().getName(), "createNoIsolatedNodesAutomaton", automaton);
		return automaton;
	}
	
	/**
	 * Creates a new path automaton.
	 * @param encoding the path encoding on which this automaton is based on
	 * @param signature the signature on which this automaton is based on
	 * @param sourceNodes the source nodes
	 * @param targetNodes the target nodes
	 * @return the automaton
	 */
	public PathAutomaton createPathAutomaton(PathEncoding encoding, Signature signature, 
			Set<Integer> sourceNodes, Set<Integer> targetNodes) {
		if(encoding==null) {
			throw new NullPointerException(
					RavenExceptionMessages.ENCODING_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(signature==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SIGNATURE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		//Log entering 
		logger.entering(getClass().getName(), "createPathAutomaton", 
				new Object[]{encoding, signature});
		
		//Create automaton
		PathAutomaton automaton = new PathAutomaton(encoding, signature, sourceNodes, targetNodes);
		initialize(automaton);
		logger.info("Path Automaton successfully initialized");
		
		//Log exiting
		logger.exiting(getClass().getName(), "createPathAutomaton", automaton);
		return automaton;
	}
	
	/**
	 * Creates a new product automaton out of the automata {@code first} and {@code second}.
	 * @param first the first automaton to be used for the product automaton
	 * @param second the second automaton to be used for the product automaton 
	 * @return the product automaton
	 */
	public ProductAutomaton createProductAutomaton(Automaton first, 
			Automaton second) {
		if(first==null) {
			throw new NullPointerException(
					RavenExceptionMessages.FIRST_AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(second==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SECOND_AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(first==second) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.AUTOMATA_MUST_NOT_BE_EQUAL_EXCEPTION);
		}
		if(!first.getAlphabet().equals(second.getAlphabet())) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ALPHABETS_MUST_BE_EQUAL_EXCEPTION);
		}
		if(first.getMaximumInterface()!=second.getMaximumInterface()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_INTERFACE_SIZES_MUST_NOT_BE_DIFFERENT_EXCEPTION);
		}
		if(first.getInnerInterface()!=second.getInnerInterface()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZES_MUST_NOT_BE_DIFFERENT_EXCEPTION);
		}
		if(first.getOuterInterface()!=second.getOuterInterface()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZES_MUST_NOT_BE_DIFFERENT_EXCEPTION);
		}
		
		//Log entering 
		logger.entering(getClass().getName(), "createProductAutomaton", 
				new Object[]{first, second});
		
		//Create automaton
		ProductEncoding encoding = ProductEncoding.getEncoding(first, second);
		ProductAutomaton automaton = new ProductAutomaton(encoding);
		initialize(automaton);
		logger.info("Product Automaton successfully initialized");
		
		//Log exiting
		logger.exiting(getClass().getName(), "createProductAutomaton", automaton);
		return automaton;
	}
	
	/**
	 * Creates a new subgraph automaton.
	 * @param encoding the subgraph encoding on which this automaton is based on
	 * @param signature the signature on which this automaton is based on
	 * @return the automaton
	 */
	public SubgraphAutomaton createSubgraphAutomaton(SubgraphEncoding encoding, 
			Signature signature) {
		if(encoding==null) {
			throw new NullPointerException(
					RavenExceptionMessages.ENCODING_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(signature==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SIGNATURE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		//Log entering 
		logger.entering(getClass().getName(), "createSubgraphAutomaton", 
				new Object[]{encoding, signature});
		
		//Create automaton
		boolean sloppyMode = Boolean.parseBoolean(
				RavenMain.getProperty(RavenMain.PROPERTY_SLOPPY_MODE));
		SubgraphAutomaton automaton = new SubgraphAutomaton(encoding, signature, sloppyMode);
		initialize(automaton);
		logger.info("Subgraph Automaton successfully initialized");
		
		//Log exiting
		logger.exiting(getClass().getName(), "createSubgraphAutomaton", automaton);
		return automaton;
	}
	
	/**
	 * Creates a new union automaton out of the automata {@code first} and {@code second}.
	 * @param first the first automaton to be used for the union automaton
	 * @param second the second automaton to be used for the union automaton 
	 * @return the union automaton
	 */
	public UnionAutomaton createUnionAutomaton(Automaton first, Automaton second) {
		if(first==null) {
			throw new NullPointerException(
					RavenExceptionMessages.FIRST_AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(second==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SECOND_AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(first==second) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.AUTOMATA_MUST_NOT_BE_EQUAL_EXCEPTION);
		}
		if(!first.getAlphabet().equals(second.getAlphabet())) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ALPHABETS_MUST_BE_EQUAL_EXCEPTION);
		}
		if(first.getMaximumInterface()!=second.getMaximumInterface()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_INTERFACE_SIZES_MUST_NOT_BE_DIFFERENT_EXCEPTION);
		}
		if(first.getInnerInterface()!=second.getInnerInterface()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZES_MUST_NOT_BE_DIFFERENT_EXCEPTION);
		}
		if(first.getOuterInterface()!=second.getOuterInterface()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZES_MUST_NOT_BE_DIFFERENT_EXCEPTION);
		}

		//Log entering 
		logger.entering(getClass().getName(), "createUnionAutomaton", 
				new Object[]{first, second});
		//Create automaton
		UnionEncoding encoding = UnionEncoding.getEncoding(first, second);
		UnionAutomaton automaton = new UnionAutomaton(encoding);
		try {
			initialize(automaton);
		}
		catch (Exception e) {
			e.printStackTrace();
		}
		logger.info("Union Automaton successfully initialized");
		
		//Log exiting
		logger.exiting(getClass().getName(), "createUnionAutomaton", automaton);
		return automaton;
	}
	
	/**
	 * Creates a new vertex-counting automaton.
	 * @param encoding the vertex-counting encoding on which this automaton is based on
	 * @param signature the signature on which this automaton is based on
	 * @param remainder the remainder for the number of vertices
	 * @return the automaton
	 */
	public VertexCountingAutomaton createVertexCountingAutomaton(VertexCountingEncoding encoding, 
			Signature signature, int remainder) {
		if(encoding==null) {
			throw new NullPointerException(
					RavenExceptionMessages.ENCODING_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(signature==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SIGNATURE_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(remainder<0 ||remainder>=encoding.getDivisor()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_REMAINDER_EXCEPTION);
		}
		
		//Log entering 
		logger.entering(getClass().getName(), "createVertexCountingAutomaton", 
				new Object[]{encoding, signature});
		
		//Create automaton
		VertexCountingAutomaton automaton = new VertexCountingAutomaton(encoding, signature, 
				remainder);
		initialize(automaton);
		logger.info("Vertex-counting Automaton successfully initialized");
		
		//Log exiting
		logger.exiting(getClass().getName(), "createVertexCountingAutomaton", automaton);
		return automaton;
	}
	
	/**
	 * Creates a new vertex cover automaton.
	 * @param encoding the vertex cover encoding on which this automaton is based on
	 * @param signature the signature on which this automaton is based on
	 * @return the automaton
	 */
	public VertexCoverAutomaton createVertexCoverAutomaton(VertexCoverEncoding encoding, 
			Signature signature) {
		if(encoding==null) {
			throw new NullPointerException(
					RavenExceptionMessages.ENCODING_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(signature==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SIGNATURE_MUST_NOT_BE_NULL_EXCEPTION);
		}

		//Log entering 
		logger.entering(getClass().getName(), "createVertexCoverAutomaton", 
				new Object[]{encoding, signature});
		
		//Create automaton
		VertexCoverAutomaton automaton = new VertexCoverAutomaton(encoding, signature);
		initialize(automaton);
		logger.info("Vertex Cover Automaton successfully initialized");
		
		//Log exiting
		logger.exiting(getClass().getName(), "createVertexCoverAutomaton", automaton);
		return automaton;
	}
	
	/**
	 * Adds the Computation Listener {@code l} to this algorithm.
	 * @param l the listener to receive computation events
	 */
	public void addComputationListener(ComputationListener l) {
		if(!listener.contains(l)) {
			listener.add(l);
		}
	}
	
	/**
	 * Removes the Computation Listener {@code l} from this algorithm.
	 * @param l the listener to be removed
	 */
	public void removeComputationListener(ComputationListener l) {
		listener.remove(l);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Initializes the automaton {@code automaton}, i.e. creates the states, initial states, 
	 * final states and transition BDDs.
	 * @param automaton the automaton to be initialized
	 */
	private void initialize(Automaton automaton) {
		//Add property listener
		for(ComputationListener l: listener) {
			automaton.addComputationListener(l);
		}

		boolean verboseMode = Boolean.valueOf(RavenMain.getProperty(RavenMain.PROPERTY_VERBOSE));
		automaton.createStates();
		if(verboseMode) {
			logger.info("State set successfully created");
		}
	
		automaton.createInitialAndFinalStates();
		if(verboseMode) {
			logger.info("Initial and final state sets successfully created");
		}
		
		try {
			automaton.createTransitions();
			if(verboseMode) {
				logger.info("Transitions successfully created");
			}
		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}
	
}
