/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.data.automata;

import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.actions.ComputationEvent;
import de.uni_due.inf.ti.raven.actions.ComputationEvent.EventType;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.Operation.Connect;
import de.uni_due.inf.ti.raven.data.Operation.Fuse;
import de.uni_due.inf.ti.raven.data.Operation.Perm;
import de.uni_due.inf.ti.raven.data.Operation.Res;
import de.uni_due.inf.ti.raven.data.Operation.Trans;
import de.uni_due.inf.ti.raven.data.Operation.Vertex;
import de.uni_due.inf.ti.raven.data.Signature;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.encodings.BDDEncoding.StateType;
import de.uni_due.inf.ti.raven.encodings.ColorabilityEncoding;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class implements a graph automaton accepting the language of all graphs which are 
 * <i>k</i>-colorable. A graph <i>G</i> is said to be <i>k-colorable</i> if each node can be 
 * colored by one of the <i>k</i> colors such that each two adjacent nodes have different colors.
 * Furthermore the inner and outer interface sized have to be fulfilled.
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.59 $
 *
 * @see ColorabilityEncoding
 */
public class ColorabilityAutomaton extends Automaton {

	private static final long serialVersionUID = -7054354624628063455L;

	transient private ColorabilityEncoding encoding;
	private final int colors;
	private final int colorBitLength;
	transient private Logger logger;	

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new colorability automaton.
	 * @param encoding the encoding for the states
	 * @param signature the signature of this automaton
	 */
	protected ColorabilityAutomaton(ColorabilityEncoding encoding, Signature signature) {
		super(Type.colorability, encoding, signature);
		this.encoding = encoding;
		colors = encoding.getColorCount();
		colorBitLength = encoding.getColorBitLength();
		
		//Register logger
		logger = Logger.getLogger(getClass().getName());
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the number of permitted colors.
	 * @return the number of colors
	 */
	public int getColorCount() {
		return colors;
	}

	/**
	 * Returns the color of the {@code node}-th interface node in the state {@code state}. 
	 * @param state the state, either <i>current state</i> or <i>successor state</i>
	 * @param node the index of the interface node
	 * @return a BDD encoding the color of the {@code node}-th node
	 */
	public int getNodeColor(RavenBDD state, int node) {
		if(state==null) {
			throw new NullPointerException(RavenExceptionMessages.BDD_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(node<0 || node>=maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}
		
		RavenBDD temp = encoding.getInterfaceEncodingBDD();
		for(int i=0;i<maximumInterface;i++) {
			if(i!=node) {
				for(int j=0;j<colorBitLength;j++) {
					temp = temp.and(encoding.getColorBit(StateType.current, i, j));
				}
			}
		}

		return getEncodedNumber(state.exists(temp));
	}
	
	@Override
	public Automaton clone() throws CloneNotSupportedException {
		Automaton clone = new ColorabilityAutomaton(encoding, signature);
		clone.initialStates = initialStates.clone();
		clone.finalStates = finalStates.clone();
		clone.nonFinalStates = nonFinalStates.clone();
		clone.states = states.clone();
		
		for(Operation operation: stateTransBDDs.keySet()) {
			clone.stateTransBDDs.put(operation, stateTransBDDs.get(operation).clone());
		}

		return clone;
	}
	
	@Override
	public void initializeEncoding() {
		encoding = ColorabilityEncoding.getEncoding(maximumInterface, innerInterface, 
				outerInterface, colors, encodingName);
		super.encoding = encoding;
		super.initializeEncoding();
	}

	
	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	protected void createInitialAndFinalStates() {
		//Log entering
		logger.entering(getClass().getName(), "createInitialAndFinalStates");
		
		initialStates = encoding.getInterfaceSize(StateType.current, innerInterface);
		for(int i=0;i<innerInterface;i++) {
			initialStates = initialStates.and(getHasBeenColored(StateType.current, i));
		}
		for(int i=innerInterface;i<maximumInterface;i++) {
			initialStates = initialStates.and(getHasNoColor(StateType.current, i));
		}
		
		finalStates = encoding.getInterfaceSize(StateType.current, outerInterface);
		for(int i=0;i<outerInterface;i++) {
			finalStates = finalStates.and(getHasBeenColored(StateType.current, i));
		}
		for(int i=outerInterface;i<maximumInterface;i++) {
			finalStates = finalStates.and(getHasNoColor(StateType.current, i));
		}

		nonFinalStates = states.clone().and(finalStates.clone().not());

		//Log exiting
		logger.exiting(getClass().getName(), "createInitialAndFinalStates");
	}

	@Override
	protected void createStates() {
		//Log entering
		logger.entering(getClass().getName(), "createStates");

		states = encoding.zero();
		for(int i=0;i<=maximumInterface;i++) {
			RavenBDD temp = encoding.getInterfaceSize(StateType.current, i);

			for(int node=0;node<maximumInterface;node++) {
				if(node<i) {
					temp = temp.and(getHasBeenColored(StateType.current, node));
				}
				else {
					temp = temp.and(getHasNoColor(StateType.current, node));
				}
			}

			states = states.or(temp);
		}

		//Log exiting
		logger.exiting(getClass().getName(), "createStates");
	}

	@Override
	protected void createTransitions() {
		//Log entering
		logger.entering(getClass().getName(), "createTransitions");

		for (Operation operation: getAlphabet()) {
			RavenBDD transBDD = stateTransBDDs.get(operation);
			RavenBDD tempBDD;

			if (operation instanceof Connect) {
				int arity = ((Connect) operation).getArity();

				//If the arity of the edge is greater than the number of available colors
				if(arity>colors) {
					//there is no valid coloring
					transBDD = encoding.zero();
				}
				//else compute all possible colorings
				else {
					//The CONNECT-operation is applicable only if the arity of the edge 
					//is greater or equal to the current interface size.
					for(int size=arity;size<=maximumInterface;size++) {
						//Check that the current interface size is i and that the interface 
						//is not changed by the CONNECT-operation
						tempBDD = encoding.getInterfaceSize(StateType.current, size);
						tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size));

						//Check that the color of each interface node is not changed 
						//by the CONNECT-operation
						for(int node=0;node<maximumInterface;node++) {
							if(node<size) {
								tempBDD = tempBDD.and(constructColorNotChangedBDD(node));
							}
							else {
								tempBDD = tempBDD.and(
										getHasNoColor(StateType.current, node).and(
												getHasNoColor(StateType.successor, node)));
							}
						}


						//Check that the all nodes which are adjacent to the new edge
						//have the same color, otherwise the colorability is not valid any more
						for(int j=size-arity;j<size;j++) {
							for(int k=j+1;k<size;k++) {
								tempBDD = tempBDD.and(constructNodeColorIsDifferentBDD(j, k));
							}
						}
						transBDD = transBDD.or(tempBDD);
						fireComputationEvent(new ComputationEvent(this, 
								EventType.COMPUTATION_ADVANCED, operation.toString(), maximumInterface, size));
					}
				}
			}
			else if (operation instanceof Fuse) {
				//The FUSE-operation is applicable only if at least two nodes are "touchable" 
				//through the interface, i.e. the interface size has to be at least 2
				for(int size=2;size<=maximumInterface;size++) {
					//Check that the current interface size is i and that the successor 
					//interface is decreased by one
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size-1));

					//Check that the color of each interface node is not changed 
					//by the FUSE-operation
					for(int node=0;node<maximumInterface;node++) {
						tempBDD = tempBDD.and(constructFuseNodeBDD(size, node));
					}

					transBDD = transBDD.or(tempBDD);
					fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED,
							operation.toString(), maximumInterface, size));
				}
			}
			else if (operation instanceof Perm) {				
				//The PERM-operation is applicable only if at least two nodes are "touchable" 
				//through the interface, i.e. the interface size has to be at least 2
				for(int size=2;size<=maximumInterface;size++) {
					//Check that the current interface size is i and that the interface 
					//is not changed by the PERM-operation
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size));

					//Check that the color of each interface node is permuted in a way 
					//that the color of the j-th node is shifted to the (j+1)-th node
					for(int node=0;node<maximumInterface;node++) {
						if(node<size) {
							tempBDD = tempBDD.and(constructColorPermutedBDD(size, node));
						}
						else {
							tempBDD = tempBDD.and(
									getHasNoColor(StateType.current, node).and(
											getHasNoColor(StateType.successor, node)));
						}
					}

					transBDD = transBDD.or(tempBDD);
					fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED,
							operation.toString(), maximumInterface, size));
				}
			}
			else if (operation instanceof Res) {
				//The RES-operation is applicable only if at least one node is "touchable"
				//through the interface, i.e. the interface size has to be at least 2
				for(int size=1;size<=maximumInterface;size++) {
					//Check that the current interface size is i and that the interface 
					//size of the successor interface is i-1
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size-1));

					//For all nodes which could possibly occur in the interface, check that ...
					for(int node=0;node<maximumInterface;node++) {
						//the color of each interface node is not changed if the node is not restricted
						if(node<(size-1)) {
							tempBDD = tempBDD.and(constructColorNotChangedBDD(node));
						}
						//the color of the restricted node will be deleted
						else if(node==(size-1)) {
							tempBDD = tempBDD.and(getHasNoColor(StateType.successor, node));
						}
						//the nodes which are not in the interface are not colored 
						else {
							tempBDD = tempBDD.and(
									getHasNoColor(StateType.current, node).and(
											getHasNoColor(StateType.successor, node)));
						}
					}

					transBDD = transBDD.or(tempBDD);
					fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED,
							operation.toString(), maximumInterface, size));
				}
			}
			else if (operation instanceof Trans) {				
				//The TRANS-operation is applicable only if at least two nodes are "touchable" 
				//through the interface, i.e. the interface size has to be at least 2
				for(int size=2;size<=maximumInterface;size++) {
					//Check that the current interface size is i and that the interface 
					//is not changed by the TRANS-operation
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size));

					//Check that the color of the first two interface node are transposed
					//and that the color of the other nodes will not be changed
					for(int node=0;node<maximumInterface;node++) {
						if(node<size) {
							tempBDD = tempBDD.and(constructTransposeNodeBDD(node));
						}
						else {
							tempBDD = tempBDD.and(
									getHasNoColor(StateType.current, node).and(
											getHasNoColor(StateType.successor, node)));
						}
					}

					transBDD = transBDD.or(tempBDD);
					fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED,
							operation.toString(), maximumInterface, size));
				}
			}
			else if (operation instanceof Vertex) {
				//The VERTEX-operation is applicable only if at least one node could be added
				//to the interface, i.e. the interface size has to be at most (maxSize-1)
				for(int size=0;size<maximumInterface;size++) {
					//Check that the current interface size is i and that the interface 
					//size of the successor interface is i+1
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size+1));

					//For all nodes which could possibly occur in the interface, check that ...
					for(int node=0;node<maximumInterface;node++) {
						//the color of each (existing) interface node is not changed and that the new 
						//interface node has been added correctly, i.e. the interface node has not 
						//been colored so far and will be colored now by a "valid" color
						if(node<=size) {
							tempBDD = tempBDD.and(constructAddNodeBDD(size, node));
						}
						//the nodes which are not in the interface are not colored 
						else {
							tempBDD = tempBDD.and(
									getHasNoColor(StateType.current, node).and(
											getHasNoColor(StateType.successor, node)));
						}
					}

					transBDD = transBDD.or(tempBDD);
					fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED,
							operation.toString(), maximumInterface, size));
				}
			}

			stateTransBDDs.put(operation, transBDD);
		}
		
		//Log exiting
		logger.exiting(getClass().getName(), "createTransitions");
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Constructs a BDD which is {@code true} if and only if in the state {@code state} 
	 * the {@code node}-th interface node has the color {@code color}.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param node the index of the interface node
	 * @param color the color of the node
	 * @return a BDD encoding that the {@code node}-th node has been colored correctly
	 */
	private RavenBDD getHasColor(StateType type, int node, int color) {
		if(node<0 || node>=maximumInterface) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}
		if(color<=0 || color>colors) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_COLOR_EXCEPTION);
		}

		return encoding.getBDDForColor(type, node, color);
	}
	
	/**
	 * Constructs a BDD which is {@code true} if and only if the 
	 * {@code node}-th interface node has been colored by some valid color.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i> 
	 * @param node the index of the interface node
	 * @return a BDD encoding that the {@code node}-th node has been colored
	 */
	private RavenBDD getHasBeenColored(StateType type, int node) {
		if(node<0 || node>=maximumInterface) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}
		
		RavenBDD bdd = encoding.zero();
		for(int color=1;color<=colors;color++) {
			bdd = bdd.or(getHasColor(StateType.current, node, color));
		}
		return bdd;
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if the color of 
	 * the {@code node}-th interface node has not been set.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param node the index of the interface node
	 * @return a BDD encoding that the {@code node}-th node has not been colored
	 */
	private RavenBDD getHasNoColor(StateType type, int node) {
		if(node<0 || node>=maximumInterface) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}

		return encoding.getBDDForColor(type, node, 0);
	}
	
	/**
	 * Constructs a BDD which is {@code true} if and only if the colors of 
	 * the {@code i}-th and the {@code j}-th interface node are different. 
	 * @param i the index of the first interface node which is compared
	 * @param j the index of the second interface node which is compared
	 * @return a BDD encoding the comparison
	 */
	private RavenBDD constructNodeColorIsDifferentBDD(int i, int j) {
		RavenBDD resultBDD = encoding.one();
		for(int k = 0;k<colorBitLength;k++) {
			resultBDD = resultBDD.and(
					encoding.getColorBit(StateType.current, i, k).biimp(
							encoding.getColorBit(StateType.successor, j, k)));
		}

		return resultBDD.not();
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if the color of 
	 * the {@code i}-th interface node has not been changed by the state 
	 * transition, i.e. the interface has the same color in both the current 
	 * state and the successor state.
	 * @param i the index of the interface node
	 * @return a BDD encoding the comparison
	 */
	private RavenBDD constructColorNotChangedBDD(int i) {
		RavenBDD resultBDD = encoding.one();
		for(int j=0;j<colorBitLength;j++) {
			resultBDD = resultBDD.and(
					encoding.getColorBit(StateType.current, i, j).biimp(
							encoding.getColorBit(StateType.successor, i, j)));
		}

		return resultBDD;
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if 
	 * <ol>
	 *     <li>the color of the first {@code i-1} interface nodes have not been changed,</li>
	 *     <li>the color of the {@code i-1}-th interface node equals the color 
	 *     of the {@code i}-th interface node and</li> 
	 *     <li>the color of the color {@code i}-th interface in the successor state has been removed.</li>
	 * </ol>
	 * @param i the current interface size
	 * @param j the index of the interface node
	 * @return a BDD encoding that the {@code i}-th interface node has not been changed or has been fused
	 */
	private RavenBDD constructFuseNodeBDD(int i, int j) {

		if(j<i-1) {
			return constructColorNotChangedBDD(j);
		}
		else if(j==i-1) {
			RavenBDD bdd = encoding.one();
			for(int k = 0;k<colorBitLength;k++) {
				bdd = bdd.and(
						encoding.getColorBit(StateType.current, j, k).biimp(
								encoding.getColorBit(StateType.current, j-1, k)));
				bdd = bdd.and(
						encoding.getColorBit(StateType.successor, j, k).not());
			}
			return bdd;
		}
		else {
			return getHasNoColor(StateType.current, j).and(
					getHasNoColor(StateType.successor, j));
		}

	}

	/**
	 * Constructs a BDD which is {@code true} if and only if the color of 
	 * the {@code i}-th interface node of the current state's interface is 
	 * permuted to the {@code i+1}-th interface node of the interface of the
	 * sucessor state.
	 * 
	 * <emph>Note:</emph> The color of the "last" interface node of the 
	 * current interface is permuted to the "first" node of the successor 
	 * interface   
	 * @param i the current interface size
	 * @param j the index of the interface node
	 * @return a BDD encoding that the color of {@code j}-th node has been permuted
	 */
	private RavenBDD constructColorPermutedBDD(int i, int j) {
		RavenBDD resultBDD = encoding.one();
		for(int k=0;k<colorBitLength;k++) {
			resultBDD = resultBDD.and(
					encoding.getColorBit(StateType.current, j, k).biimp(
							encoding.getColorBit(StateType.successor, (j+1)%i, k)));
		}

		return resultBDD;
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if 
	 * <ol>
	 *     <li>the color of the first interface node of the current state's 
	 *     interface is transposed to the second interface node of the 
	 *     successor state's interface or</li> 
	 *     <li>the color of the second interface node of the current state's 
	 *     interface is transposed to the first interface node of the
	 *     successor state's interface or</li>
	 *     <li>the color of the other interface nodes have not been changed</li>
	 * </ol>
	 * @param i the index of the interface node
	 * @return a BDD encoding that the color of {@code i}-th has been transposed 
	 */
	private RavenBDD constructTransposeNodeBDD(int i) {
		RavenBDD resultBDD = null; // der gesuchte BDD

		if(i<2) {
			int j = (i==0) ? 1 : 0;
			resultBDD = encoding.one();
			for(int k = 0;k<colorBitLength;k++) {
				resultBDD = resultBDD.and(
						encoding.getColorBit(StateType.current, i, k).biimp(
								encoding.getColorBit(StateType.successor, j, k)));
			}
		}
		else {
			resultBDD = constructColorNotChangedBDD(i);
		}

		return resultBDD;
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if the current interface 
	 * size is {@code i} and
	 * <ol>
	 *     <li>the color of the {@code j}-th interface node has not been changed 
	 *     if {@code j} is less or equal to {@code i}, or</li> 
	 *     <li>the color of the {@code j}-th interface node has not been set so far,
	 *     but will be set now a "valid" color</li>
	 * </ol> 
	 * @param i the size of the current interface
	 * @param j the index of the interface node
	 * @return a BDD encoding that the new node has been added correctly 
	 */
	private RavenBDD constructAddNodeBDD(int i, int j) {
		RavenBDD resultBDD = null; // der gesuchte BDD

		if(j<i) {
			return constructColorNotChangedBDD(j);
		}

		resultBDD = encoding.zero();
		for(int k=1;k<=colors;k++) {
			resultBDD = resultBDD.or(getHasColor(StateType.successor, j, k));
		}

		return resultBDD.and(getHasNoColor(StateType.current, i));
	}
	
}
