/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.data.automata;

import java.io.IOException;
import java.io.ObjectOutputStream;
import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.actions.ComputationEvent;
import de.uni_due.inf.ti.raven.actions.ComputationEvent.EventType;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.Operation.Connect;
import de.uni_due.inf.ti.raven.data.Operation.Fuse;
import de.uni_due.inf.ti.raven.data.Operation.Perm;
import de.uni_due.inf.ti.raven.data.Operation.Res;
import de.uni_due.inf.ti.raven.data.Operation.Trans;
import de.uni_due.inf.ti.raven.data.Operation.Vertex;
import de.uni_due.inf.ti.raven.data.Signature;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.encodings.BDDEncoding.StateType;
import de.uni_due.inf.ti.raven.encodings.DominatingSetEncoding;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class implements a graph automaton accepting the language of all graphs which have a 
 * dominating set of size at most <i>k</i>, i.e. a set <i>D</i> of nodes with size at most <i>k</i> 
 * such that each node of the graph is either in <i>D</i> or adjacent to a node in <i>D</i>.
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.53 $
 * 
 */
public class DominatingSetAutomaton extends Automaton {

	private static final long serialVersionUID = -187077827757412143L;

	transient private DominatingSetEncoding encoding;
	private int dominationNumber = 0;
	private int dominationBitLength = 0;
	transient private Logger logger;

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new dominating set automaton.
	 * @param encoding the encoding for the states
	 * @param signature the signature of this automaton
	 */
	protected DominatingSetAutomaton(DominatingSetEncoding encoding, Signature signature) {
		super(Type.dominatingSet, encoding, signature);
		this.encoding = encoding;
		dominationNumber = encoding.getDominationNumber();
		dominationBitLength = encoding.getDominationBitLength();

		// Register logger
		logger = Logger.getLogger(getClass().getName());
	}

	/*------------------------------------------------------------------------------------------*/
	/* 		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public Automaton clone() throws CloneNotSupportedException {
		Automaton clone = new DominatingSetAutomaton(encoding, signature);
		clone.initialStates = initialStates.clone();
		clone.finalStates = finalStates.clone();
		clone.nonFinalStates = nonFinalStates.clone();
		clone.states = states.clone();

		for (Operation operation : stateTransBDDs.keySet()) {
			clone.stateTransBDDs.put(operation, stateTransBDDs.get(operation).clone());
		}

		return clone;
	}
	
	/**
	 * Returns the maximum size of the dominating set.
	 * 
	 * @return the maximum dominating set size
	 */
	public int getMaximumDominatingSetSize() {
		return dominationNumber;
	}

	/**
	 * Returns {@code true} if and only if the interface at position {@code node} is a node of the 
	 * dominating set.
	 * @param state the state, either <i>current state</i> or <i>successor state</i>
	 * @param node the index of the node
	 * @return {@code true} if the interface node is node of the dominating set, {@code false} 
	 * otherwise
	 */
	public boolean isInDominatingSet(RavenBDD state, int node) {
		if (state == null) {
			throw new NullPointerException(
					RavenExceptionMessages.BDD_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if (node < 0 || node >= maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}

		RavenBDD temp = state.clone().restrict(
				encoding.getMembershipBit(StateType.current, node));
		temp = temp.exists(curStateDomain.getDomainBDD());
		return temp.isOne();
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	protected void createStates() {
		// Log entering
		logger.entering(getClass().getName(), "createStates");

		states = encoding.zero();
		for (int i = 0; i <= maximumInterface; i++) {
			RavenBDD temp = encoding.getInterfaceSize(StateType.current, i);
			for (int j = 0; j < i; j++) {
				temp = temp.and(
						encoding.getMembershipBit(StateType.current, j).imp(
								encoding.getDominationBit(StateType.current, j)));
			}

			for (int j = i; j < maximumInterface; j++) {
				temp = temp.and(encoding.getUndefinedNode(StateType.current, j));
			}

			temp = temp.and(getDominationSizeIsLessOrEqual(StateType.current, dominationNumber));
			states = states.or(temp);
		}

		// Log exiting
		logger.exiting(getClass().getName(), "createStates");
	}

	@Override
	protected void createInitialAndFinalStates() {
		// Log entering
		logger.entering(getClass().getName(), "createInitialAndFinalStates");

		initialStates = encoding.getInterfaceSize(StateType.current, innerInterface);
		initialStates = initialStates.and(createInitialStateSet(0, 0));
		for (int i = innerInterface;i<maximumInterface; i++) {
			initialStates = initialStates.and(encoding.getUndefinedNode(StateType.current, i));
		}
		
		finalStates = encoding.getInterfaceSize(StateType.current, outerInterface);
		finalStates = finalStates.and(createFinalStateSet(0, 0));
		for (int i = outerInterface;i<maximumInterface; i++) {
			finalStates = finalStates.and(encoding.getUndefinedNode(StateType.current, i));
		}

		nonFinalStates = states.clone().and(finalStates.clone().not());

		// Log exiting
		logger.exiting(getClass().getName(), "createInitialAndFinalStates");
	}

	@Override
	protected void createTransitions() {
		// Log entering
		logger.entering(getClass().getName(), "createTransitions");

		// Iteriere ueber alle Graphoperationen
		for (Operation operation : getAlphabet()) {
			RavenBDD transBDD = stateTransBDDs.get(operation);
			RavenBDD tempBDD;

			if (operation instanceof Connect) {
				int arity = ((Connect) operation).getArity();

				// The CONNECT-operation is applicable only if the arity of the
				// edge is greater or equal to the current interface size.
				for (int size = arity; size <= maximumInterface; size++) {
					// Check that the current interface size is i and that the
					// interface
					// is not changed by the CONNECT-operation
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size));

					// Check that the membership bits of the existing interface
					// nodes and the dominated
					// bit of those nodes which are not incident to the new edge
					// have not been changed
					// and that all bits which do not correspond to the current
					// interface are "undefined"
					for (int node = 0; node < maximumInterface; node++) {
						if (node < size) {
							tempBDD = tempBDD.and(constructMembershipNotChangedBDD(node));
							if (node < size - arity) {
								tempBDD = tempBDD.and(constructAdjacencyNotChangedBDD(node));
							}
						} else {
							tempBDD = tempBDD.and(constructInterfaceUndefinedBDD(node));
						}
					}

					// Add the new node and update the dominated bit of the
					// incident nodes, i.e. if at least
					// one incident node is a member of the dominating set the
					// dominated bit of all incidents
					// is set, otherwise the dominated bit will not be changed,
					// furthermore check that the
					// size of the dominating set has not been changed and then
					// add the new transition
					tempBDD = tempBDD.and(addDominatedEdge(size, arity));
					tempBDD = tempBDD.and(addUndominatedEdge(size, arity));
					tempBDD = tempBDD.and(constructDominationSizeNotChangedBDD());
					transBDD = transBDD.or(tempBDD);
					fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED, 
							operation.toString(), maximumInterface, size));
				}
			} else if (operation instanceof Fuse) {
				// The FUSE-operation is applicable only if at least two nodes are "touchable" 
				// through the interface, i.e. the interface size has to be at least 2
				for (int size = 2; size <= maximumInterface; size++) {
					// Check that the current interface size is i and that the successor interface 
					// is decreased by one
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size - 1));

					// Check that the bits of the existing interface nodes (except for the last two 
					// interface nodes) have not been changed, that the last two interface nodes 
					// have been fused correctly, i.e. the last node is removed from the interface, 
					// the membership bit is set to the maximum and the dominated bit is set to the 
					// maximum of the two nodes, and check that all bits which do not correspond to 
					// the current interface are "undefined"
					for (int node = 0; node < maximumInterface; node++) {
						if (node < size) {
							tempBDD = tempBDD.and(fuseNode(size, node));
						} else {
							tempBDD = tempBDD.and(constructInterfaceUndefinedBDD(node));
						}
					}
					

					// Check that the size of the dominating set has not been
					// changed unless both were members (then it is decreased 
					// by one) and add the new transition
					RavenBDD decreaseBDD = encoding.zero();
					for (int j = 1; j <= dominationNumber; j++) {
						decreaseBDD = decreaseBDD.or(
								getDominationSizeEquals(StateType.current, j).and(
										getDominationSizeEquals(StateType.successor, j - 1)));
					}
					
					tempBDD = tempBDD.and(
							encoding.getMembershipBit(StateType.current, size-1).and(
									encoding.getMembershipBit(StateType.current, size-2)).imp(
											decreaseBDD));
				
					tempBDD = tempBDD.and(
							encoding.getMembershipBit(StateType.current, size-1).and(
									encoding.getMembershipBit(StateType.current, size-2)).not().imp(
											constructDominationSizeNotChangedBDD()));
					
					transBDD = transBDD.or(tempBDD);
					fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED,
							operation.toString(), maximumInterface, size));
				}
			} else if (operation instanceof Perm) {
				// The PERM-operation is applicable only if at least two nodes
				// are "touchable" through the interface, i.e. the interface 
				// size has to be at least 3
				for (int size = 3; size <= maximumInterface; size++) {
					// Check that the current interface size is i and that the
					// interface
					// is not changed by the PERM-operation
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size));

					// Check that the dominating set bits for the each interface
					// node are permuted correctly and that all bits which do 
					// not correspond to the current interface are "undefined"
					for (int node = 0; node < maximumInterface; node++) {
						if (node < size) {
							tempBDD = tempBDD.and(permuteNodes(size, node));
						} else {
							tempBDD = tempBDD.and(constructInterfaceUndefinedBDD(node));
						}
					}

					// Check that the size of the dominating set has not been
					// changed and then add the new transition
					tempBDD = tempBDD.and(constructDominationSizeNotChangedBDD());
					transBDD = transBDD.or(tempBDD);
					fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED,
							operation.toString(), maximumInterface, size));
				}
			} else if (operation instanceof Res) {
				// The RES-operation is applicable only if at least one node is "touchable"
				// through the interface, i.e. the interface size has to be at least 1
				for (int size = 1; size <= maximumInterface; size++) {
					// Check that the current interface size is i and that the interface
					// size of the successor interface is i-1
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size - 1));

					// Check that the bits of the existing interface nodes (except for the last
					// interface node) have not been changed, that the last interface node has
					// been removed correctly, i.e. remove it from the interface and check
					// that all bits which do not correspond to the current interface are "undefined"
					// Remark, to apply this operation, the last node must be dominated
					for (int node = 0; node < maximumInterface; node++) {
						if (node < size - 1) {
							tempBDD = tempBDD.and(constructMembershipNotChangedBDD(node));
							tempBDD = tempBDD.and(constructAdjacencyNotChangedBDD(node));
						} else if (node == size - 1) {
							tempBDD = tempBDD.and(removeNode(size));
						} else {
							tempBDD = tempBDD.and(constructInterfaceUndefinedBDD(node));
						}
					}

					transBDD = transBDD.or(tempBDD);
					fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED,
							operation.toString(), maximumInterface, size));
				}
			} else if (operation instanceof Trans) {
				// The TRANS-operation is applicable only if at least two nodes
				// are "touchable" through the interface, i.e. the interface 
				// size has to be at least 2
				for (int size = 2; size <= maximumInterface; size++) {
					// Check that the current interface size is i and that the
					// interface is not changed by the TRANS-operation
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size));

					// Check that the dominating set bits for the first two
					// interface node are transposed, that the bits of the 
					// other nodes have not been changed and that all bits 
					// which do not correspond to the current interface are 
					// "undefined"
					for (int node = 0; node < maximumInterface; node++) {
						if (node < size) {
							tempBDD = tempBDD.and(transposeNode(size, node));
						} else {
							tempBDD = tempBDD.and(constructInterfaceUndefinedBDD(node));
						}
					}

					// Check that the size of the dominating set has not been
					// changed and then add the new transition
					tempBDD = tempBDD.and(constructDominationSizeNotChangedBDD());
					transBDD = transBDD.or(tempBDD);
					fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED,
							operation.toString(), maximumInterface, size));
				}
			} else if (operation instanceof Vertex) {
				// The VERTEX-operation is applicable only if at least one node
				// could be added to the interface, i.e. the interface size has 
				// to be at most (maxSize-1)
				for (int size = 0; size < maximumInterface; size++) {
					// Check that the current interface size is i and that the
					// interface size of the successor interface is i+1
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size + 1));

					// Check that the bits of the existing interface nodes have
					// not been changed, that the new node has been added 
					// correctly, i.e. add it to the interface and decide 
					// whether it is a node of the dominating set and if so 
					// increment the dominating set size, and check that all 
					// bits which do not correspond to the current interface 
					// are "undefined"
					for (int node = 0; node < maximumInterface; node++) {
						if (node < size) {
							tempBDD = tempBDD.and(constructMembershipNotChangedBDD(node));
							tempBDD = tempBDD.and(constructAdjacencyNotChangedBDD(node));
						} else if (node == size) {
							tempBDD = tempBDD.and(addNode(size));
						} else {
							tempBDD = tempBDD.and(constructInterfaceUndefinedBDD(node));
						}
					}

					transBDD = transBDD.or(tempBDD);
					fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED,
							operation.toString(), maximumInterface, size));

				}
			}
			stateTransBDDs.put(operation, transBDD);
		}

		// Log exiting
		logger.exiting(getClass().getName(), "createTransitions");
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods 																	*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public void initializeEncoding() {
		encoding = DominatingSetEncoding.getEncoding(maximumInterface, innerInterface, 
				outerInterface, dominationNumber, encodingName);
		super.encoding = encoding;
		super.initializeEncoding();
	}

	/**
	 * Creates the initial state set. The BDD encoding the initial state set is build recursively,
	 * starting at the interface node {@code node} with the a dominating set containing currently 
	 * {@code dominatingSetSize} nodes.
	 * @param node the index of the node at which the recursion begins
	 * @param dominatingSetSize the size of the current dominating set
	 * @return the BDD encoding the initial state set
	 */
	private RavenBDD createInitialStateSet(int node, int dominatingSetSize) {
		if (node < 0 || node > innerInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}
		if (dominatingSetSize< 0 || dominatingSetSize > dominationNumber) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_DOMINATING_SET_SIZE_EXCEPTION);
		}
		
		if(node==innerInterface) {
			return getDominationSizeEquals(StateType.current, dominatingSetSize);
		}

		RavenBDD dominatedNodeBDD = encoding.zero();

		if(dominatingSetSize<dominationNumber) {
			dominatedNodeBDD = encoding.getMembershipBit(StateType.current, node).and(
					encoding.getDominationBit(StateType.current, node));
			dominatedNodeBDD = dominatedNodeBDD.and(createInitialStateSet(node+1, 
					dominatingSetSize+1));
		}
		
		RavenBDD undominatedNodeBDD = encoding.getMembershipBit(StateType.current, node).not().and(
				encoding.getDominationBit(StateType.current, node).not());
		undominatedNodeBDD = undominatedNodeBDD.and(createInitialStateSet(node+1, 
				dominatingSetSize));
		
		return dominatedNodeBDD.or(undominatedNodeBDD);
	}
	
	/**
	 * Creates the final state set. The BDD encoding the final state set is build recursively,
	 * starting at the interface node {@code node} with the a dominating set containing currently 
	 * {@code dominatingSetSize} nodes.
	 * @param node the index of the node at which the recursion begins
	 * @param dominatingSetSize the size of the current dominating set
	 * @return the BDD encoding the final state set
	 */
	private RavenBDD createFinalStateSet(int node, int dominatingSetSize) {	
		if (node < 0 || node > outerInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}
		if (dominatingSetSize< 0 || dominatingSetSize > dominationNumber) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_DOMINATING_SET_SIZE_EXCEPTION);
		}
		
		if(node==outerInterface) {
			RavenBDD bdd = encoding.zero();
			for(int i=dominatingSetSize;i<=dominationNumber;i++) {
				bdd = bdd.or(getDominationSizeEquals(StateType.current, i));
			}

			return bdd;
		}

		RavenBDD dominatingNodeBDD = encoding.zero();

		if(dominatingSetSize<dominationNumber) {
			dominatingNodeBDD = encoding.getMembershipBit(StateType.current, node).and(
					encoding.getDominationBit(StateType.current, node));
			dominatingNodeBDD = dominatingNodeBDD.and(createFinalStateSet(node+1, 
					dominatingSetSize+1));
		}
		
		RavenBDD undominatingNodeBDD = encoding.getMembershipBit(StateType.current, 
				node).not().and(encoding.getDominationBit(StateType.current, node));
		undominatingNodeBDD = undominatingNodeBDD.and(createFinalStateSet(node+1, 
				dominatingSetSize));
		
		return dominatingNodeBDD.or(undominatingNodeBDD);
	}
	
	/**
	 * Constructs a BDD which is {@code true} if and only if in the state
	 * {@code state} the size of the dominating set equals {@code size}.
	 * 
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param size the size of the dominating set
	 * @return a BDD encoding that the domination number equals {@code size}
	 */
	private RavenBDD getDominationSizeEquals(StateType type, int size) {
		if (size < 0 || size > dominationNumber) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_DOMINATING_SET_SIZE_EXCEPTION);
		}

		RavenBDD bdd = encoding.one();
		for (int i = 0; i < dominationBitLength; i++) {
			int b = size % 2;
			size >>= 1;

			if (b == 0) {
				bdd = bdd.and(encoding.getDominationSizeBit(type, i).not());
			} else {
				bdd = bdd.and(encoding.getDominationSizeBit(type, i));
			}
		}

		return bdd;
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if in the state
	 * {@code state} the size of the dominating set is less or equal
	 * {@code size}.
	 * 
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param size the size of the dominating set
	 * @return a BDD encoding that the domination number equals {@code size}
	 */
	private RavenBDD getDominationSizeIsLessOrEqual(StateType type, int size) {
		if (size < 0 || size > dominationNumber) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_DOMINATING_SET_SIZE_EXCEPTION);
		}

		RavenBDD bdd = encoding.zero();
		for (int i = 0; i <= size; i++) {
			bdd = bdd.or(getDominationSizeEquals(type, i));
		}

		return bdd;
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if the size of the
	 * dominating set has not been changed.
	 * 
	 * @return a BDD encoding that the number of domination set has not been
	 *         changed
	 */
	private RavenBDD constructDominationSizeNotChangedBDD() {
		RavenBDD resultBDD = encoding.one();
		for (int i = 0; i < dominationBitLength; i++) {
			resultBDD = resultBDD.and(encoding.getDominationSizeBit(
					StateType.current, i).biimp(
					encoding.getDominationSizeBit(StateType.successor, i)));
		}

		return resultBDD;
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if the membership bit
	 * of the {@code node}-th interface node has not been changed.
	 * 
	 * @param node the index of the node
	 * @return a BDD encoding that the {@code node}-th membership bit has not
	 *         been changed.
	 */
	private RavenBDD constructMembershipNotChangedBDD(int node) {
		if (node < 0 || node > maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}

		return encoding.getMembershipBit(StateType.current, node).biimp(
				encoding.getMembershipBit(StateType.successor, node));
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if the adjacency bit
	 * of the {@code node}-th interface node has not been changed.
	 * 
	 * @param node the index of the node
	 * @return a BDD encoding that the {@code node}-th adjacency bit has not
	 *         been changed.
	 */
	private RavenBDD constructAdjacencyNotChangedBDD(int node) {
		if (node < 0 || node > maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}

		return encoding.getDominationBit(StateType.current, node).biimp(
				encoding.getDominationBit(StateType.successor, node));
	}

	/**
	 * Constructs a BDD which checks that all adjacent nodes of the added edge are dominated if at 
	 * least one of the adjacent nodes is in the dominating set.
	 * 
	 * @param size the interface size
	 * @param arity the arity of the added edge
	 * @return a BDD encoding that the {@code node}-th membership of the adjacent nodes is updated 
	 * correctly
	 */
	private RavenBDD addDominatedEdge(int size, int arity) {
		if (size < 0 || size > maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}
		if (arity < 0 || arity > size) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_ARITY_EXCEPTION);
		}

		RavenBDD resultBDD = encoding.zero();
		RavenBDD tempBDD = encoding.one();
		for (int i = size - arity; i < size; i++) {
			resultBDD = resultBDD.or(encoding.getMembershipBit(StateType.current, i));
			tempBDD = tempBDD.and(encoding.getDominationBit(StateType.successor, i));
		}

		return resultBDD.imp(tempBDD);
	}

	/**
	 * Constructs a BDD which checks that the adjacency of no node is changed if the added edge are 
	 * dominated if at least one of the adjacent nodes is in the dominating set.
	 * 
	 * @param size the interface size
	 * @param arity the arity of the added edge
	 * @return a BDD encoding that the {@code node}-th membership of the adjacent nodes is updated 
	 * correctly
	 */
	private RavenBDD addUndominatedEdge(int size, int arity) {
		if (size < 0 || size > maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}
		if (arity < 0 || arity > size) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_ARITY_EXCEPTION);
		}

		RavenBDD resultBDD = encoding.one();
		RavenBDD tempBDD = encoding.one();
		for (int i = size - arity; i < size; i++) {
			resultBDD = resultBDD.and(encoding.getMembershipBit(
					StateType.current, i).not());
			tempBDD = tempBDD.and(constructAdjacencyNotChangedBDD(i));
		}

		return resultBDD.imp(tempBDD);
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if the interface node at position 
	 * {@code node} has been undefined in both the current and the successor state.
	 * 
	 * @param node the index of the node
	 * @return a BDD encoding that interface node has been undefined
	 */
	private RavenBDD constructInterfaceUndefinedBDD(int node) {
		if (node < 0 || node >= maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}

		return encoding.getUndefinedNode(StateType.current, node).and(
				encoding.getUndefinedNode(StateType.successor, node));
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if
	 * <ul>
	 * 		<li>the first {@code size-2} interface nodes (of the current state) have
	 * 		not been changed, and</li>
	 * 		<li>the next-to-last interface node (of the current state) is dominated
	 * 		in the successor state if either the next-to-last or the last interface
	 * 		node (of the current state) have been dominated in the current state, and</li>
	 * 		<li>the next-to-last interface node (of the current state) is a node of
	 * 		the dominating set if and only if the last or next-to-last node (of the current state) is
	 * 		a node of the dominating set, and</li>
	 * 		<li>the last node (of the current state) is removed from the interface of
	 * 		the successor state</li>
	 * </ul>
	 * 
	 * @param size the interface size
	 * @param node the index of the node
	 * @return a BDD encoding that the last two interface nodes have been fused correctly
	 */
	private RavenBDD fuseNode(int size, int node) {
		if (size < 0 || size > maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}
		if (size < 2) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INTERFACE_SIZE_MUST_BE_GREATER_THAN_ONE_EXCEPTION);
		}
		if (node < 0 || node >= size) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}

		if (node < size - 2) {
			return constructMembershipNotChangedBDD(node).and(
					constructAdjacencyNotChangedBDD(node));
		} else if (node == size - 2) {
			RavenBDD resultBDD = encoding.getDominationBit(StateType.current,
					node).or(encoding.getDominationBit(StateType.current, node + 1));
			resultBDD = resultBDD.biimp(encoding.getDominationBit(StateType.successor, node));

			RavenBDD temp = encoding.getMembershipBit(StateType.current, node).or(
					encoding.getMembershipBit(StateType.current, node+1));
			return resultBDD.and(temp.biimp(encoding.getMembershipBit(StateType.successor, node+1)));
		} else { // node == size-1
			RavenBDD resultBDD = encoding.getMembershipBit(StateType.successor, node).not();
			return resultBDD.and(encoding.getDominationBit(StateType.successor, node).not());
		}
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if
	 * <ul>
	 * 		<li>the first interface node of the current state is mapped to the second
	 * 		interface node of the successor state, and</li>
	 * 		<li>the second interface node of the current state is mapped to the third
	 * 		interface node of the successor state, and</li>
	 * 		</li>..., and</li>
	 * 		<li>the last interface node of the current state is mapped to the first
	 * 		interface node of the successor state</li>
	 * </ul>
	 * 
	 * @param size the interface size
	 * @param node the index of the node
	 * @return a BDD encoding that the interface nodes have been permuted correctly
	 */
	private RavenBDD permuteNodes(int size, int node) {
		if (size < 3 || size > maximumInterface) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}
		if (node < 0 || node >= size) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}

		RavenBDD resultBDD = encoding.getMembershipBit(StateType.current, node).biimp(
				encoding.getMembershipBit(StateType.successor, (node + 1) % size));
		return resultBDD.and(encoding.getDominationBit(StateType.current, node).biimp(
				encoding.getDominationBit(StateType.successor, (node + 1) % size)));
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if
	 * <ul>
	 * 		<li>the last interface node of the current state has been dominated</li>
	 * 		<li>the last interface node is removed from the successor state's
	 * 		interface, and</li>
	 * 		<li>the dominating set size has not been changed</li>
	 * </ul>
	 * 
	 * @param size the size of the interface of the current state
	 * @return a BDD encoding that the last node of the interface has been
	 *         removed correctly
	 */
	private RavenBDD removeNode(int size) {
		if (size < 1 || size > maximumInterface) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}

		RavenBDD resultBDD = encoding.getMembershipBit(StateType.successor, size - 1).not().and(
				encoding.getDominationBit(StateType.current, size - 1).and(
						encoding.getDominationBit(StateType.successor, size - 1).not()));

		RavenBDD tempBDD = encoding.zero();
		for (int j = 0; j <= dominationNumber; j++) {
			tempBDD = tempBDD.or(getDominationSizeEquals(StateType.current, j).and(
					getDominationSizeEquals(StateType.successor, j)));
		}

		return resultBDD.and(tempBDD);
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if
	 * <ul>
	 * 		<li>the first node of the current state's interface is mapped to the
	 * 		second node of the successors state's interface, and</li>
	 * 		<li>the second node of the current state's interface is mapped to the
	 * 		first node of the successors state's interface, and</li>
	 * 		<li>all other nodes are not permuted</li>
	 * </ul>
	 * 
	 * @param size the size of the interface of the current state
	 * @param node the index of the node
	 * @return a BDD encoding that the first two nodes of the interface are
	 *         transposed
	 */
	private RavenBDD transposeNode(int size, int node) {
		if (size < 2 || size > maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}
		if (node < 0 || node >= size) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}

		if (node == 0) {
			RavenBDD resultBDD = encoding.getMembershipBit(StateType.current, 0).biimp(
							encoding.getMembershipBit(StateType.successor, 1));
			return resultBDD.and(encoding.getDominationBit(StateType.current, 0).biimp(
							encoding.getDominationBit(StateType.successor, 1)));
		} else if (node == 1) {
			RavenBDD resultBDD = encoding.getMembershipBit(StateType.current, 1).biimp(
							encoding.getMembershipBit(StateType.successor, 0));
			return resultBDD.and(encoding.getDominationBit(StateType.current, 1).biimp(
							encoding.getDominationBit(StateType.successor, 0)));
		} else {
			return constructMembershipNotChangedBDD(node).and(
					constructAdjacencyNotChangedBDD(node));
		}
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if a new node is added to the last 
	 * position of interface (i.e. the {@code size+1}-th position) and
	 * <ul>
	 * 		<li>the new node has not been dominated before (it has been added) and</li>
	 * 		<li>the new node is dominated if and only if it belongs to the dominating set</li>
	 * </ul>
	 * 
	 * @param size the size of the interface of the current state
	 * @return a BDD encoding that a new node is added correctly to the last
	 *         position of the interface
	 */
	private RavenBDD addNode(int size) {
		if (size < 0 || size > maximumInterface) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}
		if (size == maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ADDING_NODE_NOT_POSSIBLE_EXCEPTION);
		}
		
		RavenBDD resultBDD = encoding.getUndefinedNode(StateType.current, size).and(
				encoding.getMembershipBit(StateType.successor, size).biimp(
						encoding.getDominationBit(StateType.successor, size)));
		
		RavenBDD tempBDD = encoding.zero();
		for (int j = 0; j < dominationNumber; j++) {
			tempBDD = tempBDD.or(
					getDominationSizeEquals(StateType.current, j).and(
							getDominationSizeEquals(StateType.successor, j + 1)));
		}
		
		resultBDD = resultBDD.and(encoding.getMembershipBit(StateType.successor, size).imp(tempBDD));

		tempBDD = encoding.zero();
		for (int j = 0; j <= dominationNumber; j++) {
			tempBDD = tempBDD.or(
					getDominationSizeEquals(StateType.current, j).and(
							getDominationSizeEquals(StateType.successor, j)));
		}

		return resultBDD.and(
				encoding.getMembershipBit(StateType.successor, size).not().imp(tempBDD));
	}

	/**
	 * This method is called when an object of this class shall be serialized.
	 * 
	 * @param out the ObjectOutputStream the object is written on
	 * @throws IOException thrown if if I/O errors occur while writing to the underlying 
	 * {@code OutputStream}
	 */
	@SuppressWarnings("static-method")
	private void writeObject(ObjectOutputStream out) throws IOException {
		out.defaultWriteObject();
	}
	
}
