/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.data.automata;

import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.actions.ComputationEvent;
import de.uni_due.inf.ti.raven.actions.ComputationEvent.EventType;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.Operation.Connect;
import de.uni_due.inf.ti.raven.data.Operation.Perm;
import de.uni_due.inf.ti.raven.data.Operation.Res;
import de.uni_due.inf.ti.raven.data.Operation.Trans;
import de.uni_due.inf.ti.raven.data.Operation.Vertex;
import de.uni_due.inf.ti.raven.data.Signature;
import de.uni_due.inf.ti.raven.data.Operation.Fuse;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.encodings.BDDEncoding.StateType;
import de.uni_due.inf.ti.raven.encodings.LinkEncoding;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class implements a graph automaton accepting all graphs that have an
 * edge between a node from the inner to a node from the outer interface.
 * Additionally the size outer interface that is given as a parameter has to be
 * fulfilled.
 * 
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.16 $
 * 
 */
public class LinkAutomaton extends Automaton {

	private static final long serialVersionUID = 3031400095389229862L;
	transient private LinkEncoding encoding;
	transient private Logger logger;

	/*------------------------------------------------------------------------------------------*/
	/* Constructors																				*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new link automaton
	 * @param encoding the encoding for the states
	 * @param signature the signature of this automaton
	 */
	protected LinkAutomaton(LinkEncoding encoding, Signature signature) {
		super(Type.link, encoding, signature);
		this.encoding = encoding;

		// Register logger
		logger = Logger.getLogger(getClass().getName());
	}

	/*------------------------------------------------------------------------------------------*/
	/* Public Methods																			*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public Automaton clone() throws CloneNotSupportedException {
		Automaton clone = new LinkAutomaton(encoding, signature);
		clone.initialStates = initialStates.clone();
		clone.finalStates = finalStates.clone();
		clone.nonFinalStates = nonFinalStates.clone();
		clone.states = states.clone();

		for (Operation operation : stateTransBDDs.keySet()) {
			clone.stateTransBDDs.put(operation, stateTransBDDs.get(operation)
					.clone());
		}

		return clone;
	}

	@Override
	public void initializeEncoding() {
		encoding = LinkEncoding.getEncoding(maximumInterface, innerInterface,
				outerInterface, encodingName);
		super.encoding = encoding;
		super.initializeEncoding();
	}

	/*------------------------------------------------------------------------------------------*/
	/* Protected Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	protected void createStates() {
		// Log entering
		logger.entering(getClass().getName(), "createStates");

		states = encoding.zero();
		for (int i = 0; i <= maximumInterface; i++) {
			states = states.or(encoding.getInterfaceSize(StateType.current, i));
		}

		// Log exiting
		logger.exiting(getClass().getName(), "createStates");
	}

	@Override
	protected void createInitialAndFinalStates() {
		// Log entering
		logger.entering(getClass().getName(), "createInitialAndFinalStates");

		initialStates = encoding.getInterfaceSize(StateType.current, innerInterface);
		// all nodes are marked as initial
		for (int i=0; i<innerInterface; i++){
			initialStates = initialStates.
					and(encoding.getInitialNodeBit(StateType.current, i).
					and(encoding.getConnectedNodeBit(StateType.current, i).not()));
		}

		// final states have the interface size 'outerInterfaceSize' and at
		// least one
		// marked node
		finalStates = encoding.zero();
		for (int i=0; i<outerInterface; i++){
			finalStates = finalStates.or(encoding.getConnectedNodeBit(StateType.current, i));
		}		
		finalStates = finalStates.and(encoding.getInterfaceSize(StateType.current,
				outerInterface));

		nonFinalStates = states.clone().and(finalStates.clone().not());

		// Log exiting
		logger.exiting(getClass().getName(), "createInitialAndFinalStates");
	}

	@Override
	protected void createTransitions() {
		// Log entering
		logger.entering(getClass().getName(), "createTransitions");

		// iterate over all operations
		for (Operation operation : getAlphabet()) {
			RavenBDD transBDD = stateTransBDDs.get(operation);
			RavenBDD tempBDD;

			if (operation instanceof Connect) {
				int arity = ((Connect) operation).getArity();
				for (int size = arity; size <= maximumInterface; size++) {
					// Check that the current interface size is i and that
					// the successor interface is still i
					tempBDD = encoding
							.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(
							StateType.successor, size));

					tempBDD = tempBDD.and(getMarkingForConnectOp(size, arity));

					transBDD = transBDD.or(tempBDD);
					fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED, 
							operation.toString(), maximumInterface, size));
				}
			}
			else if (operation instanceof Fuse) {
				for (int size = 2; size <= maximumInterface; size++) {
					// Check that the current interface size is i and that
					// the successor interface is i-1
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(
							StateType.successor, size - 1));
					
					// mark fused node as connected if one of the nodes 
					// to be fused was marked as connected
					// mark fused node as initial if one of the nodes 
					// to be fused was marked as initial
					tempBDD = tempBDD.and(getMarkingForFuseOp(size));
					transBDD = transBDD.or(tempBDD);
					
					fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED,
							operation.toString(), maximumInterface, size));
				}
			}
			else if (operation instanceof Perm) {
				for (int size = 2; size <= maximumInterface; size++) {
					// Check that the current interface size is i and that
					// the successor interface is still i
					tempBDD = encoding
							.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(
							StateType.successor, size));
					
					tempBDD = tempBDD.and(getMarkingForPermOp(size));
					
					transBDD = transBDD.or(tempBDD);
					fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED,
							operation.toString(), maximumInterface, size));
				}
			}
			else if (operation instanceof Res) {
				for (int size = 1; size <= maximumInterface; size++) {
					// Check that the current interface size is i and
					// that the successor interface is i-1
					tempBDD = encoding
							.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(
							StateType.successor, size - 1));
					
					tempBDD = tempBDD.and(getMarkingForResOp(size));
					
					transBDD = transBDD.or(tempBDD);
					fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED,
							operation.toString(), maximumInterface, size));
				}
			}
			else if (operation instanceof Trans) {
				for (int size = 2; size <= maximumInterface; size++) {
					// Check that the current interface size is i and that
					// the successor interface is still i
					tempBDD = encoding
							.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(
							StateType.successor, size));

					tempBDD = tempBDD.and(getMarkingForTransOp());

					transBDD = transBDD.or(tempBDD);
					fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED,
							operation.toString(), maximumInterface, size));
				}
			}
			else if (operation instanceof Vertex) {
				for (int size = 0; size < maximumInterface; size++) {
					// Check that the current interface size is i and that
					// the successor interface is i+1
					tempBDD = encoding
							.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(
							StateType.successor, size + 1));

					// The markings are not changed
					tempBDD = tempBDD.and(getMarkingForVertexOp(size));

					transBDD = transBDD.or(tempBDD);
					fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED,
							operation.toString(), maximumInterface, size));
				}
			}
			stateTransBDDs.put(operation, transBDD);
		}

		// Log exiting
		logger.exiting(getClass().getName(), "createTransitions");
	}

	/*------------------------------------------------------------------------------------------*/
	/* Private Methods */
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns a BDD encoding that the last #{@code arity} nodes have been marked as connected. All 
	 * other markings are not changed.
	 * @return a BDD encoding that the last #{@code arity} nodes have been marked as connected. All 
	 * other markings are not changed.
	 */
	private RavenBDD getMarkingForConnectOp(int interfaceSize, int arity) {
		if(interfaceSize<0 || interfaceSize>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}
		
		RavenBDD bdd = encoding.one();
		for (int i = 0; i < interfaceSize - arity; i++) {
			bdd = bdd.and(getNodeInitialBitUnchangedBDD(i));
			bdd = bdd.and(getNodeConnectedBitUnchangedBDD(i));
		}
		
		RavenBDD initialCheck = encoding.zero();
		RavenBDD setConnected = encoding.one();
		RavenBDD connectedUnchanged = encoding.one();
		for (int i = interfaceSize - arity; i < interfaceSize; i++) {
			// initial bits are not changed
			bdd = bdd.and(getNodeInitialBitUnchangedBDD(i));
			// check if one is initial
			initialCheck = initialCheck.or(encoding.getInitialNodeBit(StateType.current, i));
			setConnected = setConnected.and(encoding.getConnectedNodeBit(StateType.successor, i));
			connectedUnchanged = connectedUnchanged.and(
					encoding.getConnectedNodeBit(StateType.current, i).biimp(
							encoding.getConnectedNodeBit(StateType.successor, i)));
		}
		
		bdd = bdd.and(initialCheck.clone().imp(setConnected).and(
				initialCheck.not().imp(connectedUnchanged)));
		bdd = bdd.and(getNodeBitsUnsetBDD(interfaceSize));
		return bdd;		
	}
	
	/**
	 * Returns a BDD encoding that the last two nodes have been fused. The markings for these two 
	 * are merged in the marking for the second to last node and the last node is removed from the 
	 * interface. All other markings are not changed.
	 * @param size size the size of the current interface to determine the position of the last two 
	 * nodes
	 * @return Returns a BDD encoding that the last two nodes have been fused and all other nodes 
	 * are not changed
	 */
	private RavenBDD getMarkingForFuseOp(int size) {
		RavenBDD bdd = encoding.one();

		for (int i=0;i<size;i++) {
			if (i == size-2) {
				bdd = bdd.and(encoding.getInitialNodeBit(StateType.current, i).or(
							encoding.getInitialNodeBit(StateType.current, i+1)).biimp(
									encoding.getInitialNodeBit(StateType.successor, i)));
				bdd = bdd.and(
						encoding.getConnectedNodeBit(StateType.current, i).or(
								encoding.getConnectedNodeBit(StateType.current, i+1)).biimp(
										encoding.getConnectedNodeBit(StateType.successor, i)));
			}
			else if (i == size-1) {
				bdd = bdd.and(
						encoding.getInitialNodeBit(StateType.successor, i).not().and(
								encoding.getConnectedNodeBit(StateType.successor, i).not()));
			}
			else {
				bdd = bdd.and(getNodeInitialBitUnchangedBDD(i));
				bdd = bdd.and(getNodeConnectedBitUnchangedBDD(i));
			}					
		}
		
		// all other nodes are unset
		bdd = bdd.and(getNodeBitsUnsetBDD(size));

		return bdd;		
	}

	/**
	 * Returns a BDD encoding that the markings of all nodes were shifted by 2 bits
	 * while the last two became the first two. Otherwise the marking did not change.
	 * 
	 * @return a BDD encoding that the node markings have been permuted 
	 * according to the perm operation
	 */
	private RavenBDD getMarkingForPermOp(int interfaceSize){
		RavenBDD bdd = encoding.one();
		
		for (int i = 0; i < interfaceSize-1; i++) {
			bdd = bdd.and(
					encoding.getInitialNodeBit(StateType.current, i).biimp(
							encoding.getInitialNodeBit(StateType.successor, i+1)));
			bdd = bdd.and(
					encoding.getConnectedNodeBit(StateType.current, i).biimp(
							encoding.getConnectedNodeBit(StateType.successor, i+1)));
		}
		
		// switch the markings for the last 'real' node to the first position
		bdd = bdd.and(
				encoding.getInitialNodeBit(StateType.current, interfaceSize-1).biimp(
						encoding.getInitialNodeBit(StateType.successor, 0)));
		bdd = bdd.and(
				encoding.getConnectedNodeBit(StateType.current, interfaceSize-1).biimp(
						encoding.getConnectedNodeBit(StateType.successor, 0)));

		// all other nodes are unset
		bdd = bdd.and(getNodeBitsUnsetBDD(interfaceSize));
		
		return bdd;		
	}
	
	/**
	 * Returns a BDD encoding that the marking of the last node have 
	 * been reset to 0.
	 * @param size the size of the current interface to determine the position of the last node
	 * @return a BDD encoding that the marking of the last node have 
	 * been reset to 0.
	 */
	private RavenBDD getMarkingForResOp(int size){
		RavenBDD bdd = encoding.one();
		
		// all previous node markings are not changed
		for (int i = 0; i < size-1; i++) {
			bdd = bdd.and(getNodeInitialBitUnchangedBDD(i));
			bdd = bdd.and(getNodeConnectedBitUnchangedBDD(i));
		}
		
		bdd = bdd.and(encoding.getInitialNodeBit(StateType.successor, size-1).not());
		bdd = bdd.and(encoding.getConnectedNodeBit(StateType.successor, size-1).not());
		
		bdd = bdd.and(getNodeBitsUnsetBDD(size));

		return bdd;		
	}
	
	/**
	 * Returns a BDD encoding that the markings of the 1st and 2nd
	 * node have been switched, otherwise the marking did not change.
	 * 
	 * @return a BDD encoding that the markings have been switched for two 
	 * nodes, otherwise it was not changed
	 */
	private RavenBDD getMarkingForTransOp(){
		RavenBDD bdd = encoding.one();
		int first = 0;
		int second = 1;
		
		for (int i = 0; i < maximumInterface; i++) {
			if (i == first){
				bdd = bdd.and(
						encoding.getInitialNodeBit(StateType.current, i).biimp(
								encoding.getInitialNodeBit(StateType.successor, second)));	
				bdd = bdd.and(
						encoding.getConnectedNodeBit(StateType.current, i).biimp(
								encoding.getConnectedNodeBit(StateType.successor, second)));
			}
			else if (i == second){
				bdd = bdd.and(
						encoding.getInitialNodeBit(StateType.current, i).biimp(
								encoding.getInitialNodeBit(StateType.successor, first)));	
				bdd = bdd.and(
						encoding.getConnectedNodeBit(StateType.current, i).biimp(
								encoding.getConnectedNodeBit(StateType.successor, first)));				
			}
			else {
				bdd = bdd.and(getNodeInitialBitUnchangedBDD(i));
				bdd = bdd.and(getNodeConnectedBitUnchangedBDD(i));
			}
		}

		
		return bdd;		
	}
	
	/**
	 * Returns a BDD encoding that the markings of the first #{@code size} nodes have not been 
	 * changed and a new isolated node has been added at the last position of the interface.
	 * @param size the size of the current interface to determine the position of the new node
	 * @return a BDD encoding that the existing markings have not been changed and a new isolated 
	 * node has been added
	 */
	private RavenBDD getMarkingForVertexOp(int size) {
		RavenBDD bdd = encoding.one();

		for (int i = 0; i < size; i++) {
			bdd = bdd.and(getNodeInitialBitUnchangedBDD(i));
			bdd = bdd.and(getNodeConnectedBitUnchangedBDD(i));
		}
		
		bdd = bdd.and(encoding.getInitialNodeBit(StateType.current, size).not());
		bdd = bdd.and(encoding.getInitialNodeBit(StateType.successor, size).not());
		bdd = bdd.and(encoding.getConnectedNodeBit(StateType.current, size).not());
		bdd = bdd.and(encoding.getConnectedNodeBit(StateType.successor, size).not());

		bdd = bdd.and(getNodeBitsUnsetBDD(size+1));
		
		return bdd;
	}
	
	/**
	 * Returns a BDD encoding that marking indicating the membership to the initial node set for 
	 * the {@code node}-th node of the interface has not been changed.
	 * @param node the index of the node in the current interface
	 * @return a BDD encoding that initial-node-marking has not been changed
	 */
	private RavenBDD getNodeInitialBitUnchangedBDD(int node) {
		return encoding.getInitialNodeBit(StateType.current, node).biimp(
				encoding.getInitialNodeBit(StateType.successor, node));
	}
	
	/**
	 * Returns a BDD encoding that marking indicating the adjacency to an initial node for the 
	 * {@code node}-th node of the interface has not been changed.
	 * @param node the index of the node in the current interface
	 * @return a BDD encoding that adjacency-to-an-initial-node-marking has not been changed
	 */
	private RavenBDD getNodeConnectedBitUnchangedBDD(int node) {
		return encoding.getConnectedNodeBit(StateType.current, node).biimp(
				encoding.getConnectedNodeBit(StateType.successor, node));
	}
	
	/**
	 * Returns a BDD encoding that both the initial-node-marking and 
	 * adjacency-to-an-initial-node-marking are not set for all nodes not in the current interface.
	 * @param size the size of the current interface to determine the position of the bits not in 
	 * the current interface
	 * @return a BDD encoding that the bits currently not used are not set
	 */
	private RavenBDD getNodeBitsUnsetBDD(int size) {
		RavenBDD bdd = encoding.one();
		for (int i = size;i<maximumInterface; i++) {
			// Neither the initial nor connection bits have been set
			bdd = bdd.and(
					encoding.getInitialNodeBit(StateType.current, i).not().and(
							encoding.getInitialNodeBit(StateType.successor, i).not()));
			bdd = bdd.and(
					encoding.getConnectedNodeBit(StateType.current, i).not().and(
						encoding.getConnectedNodeBit(StateType.successor, i).not()));
		}
		
		return bdd;
	}
	
}
