/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.data.automata;

import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.actions.ComputationEvent;
import de.uni_due.inf.ti.raven.actions.ComputationEvent.EventType;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.Operation.Connect;
import de.uni_due.inf.ti.raven.data.Operation.Perm;
import de.uni_due.inf.ti.raven.data.Operation.Res;
import de.uni_due.inf.ti.raven.data.Operation.Trans;
import de.uni_due.inf.ti.raven.data.Operation.Vertex;
import de.uni_due.inf.ti.raven.data.Signature;
import de.uni_due.inf.ti.raven.data.Operation.Fuse;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.encodings.BDDEncoding.StateType;
import de.uni_due.inf.ti.raven.encodings.BoundEncoding;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class implements a graph automaton accepting the language of all graphs
 * whose number of nodes is at most <i>maximum</i>.
 * 
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.15 $
 * 
 */
public class MaximumVertexAutomaton extends Automaton {

	private static final long serialVersionUID = 3031400095389229862L;
	private final int maximum;
	private final int maximumBitLength;
	transient private BoundEncoding encoding;
	transient private Logger logger;

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors 																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new maximum vertex automaton
	 * 
	 * @param encoding the encoding for the states
	 * @param signature the signature of this automaton
	 */
	protected MaximumVertexAutomaton(BoundEncoding encoding, Signature signature) {
		super(Type.maximumVertex, encoding, signature);
		this.maximum = encoding.getBound();
		this.maximumBitLength = encoding.getBoundBitLength();
		this.encoding = encoding;

		// Register logger
		logger = Logger.getLogger(getClass().getName());
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public Automaton clone() throws CloneNotSupportedException {
		Automaton clone = new MaximumVertexAutomaton(encoding,
				signature);
		clone.initialStates = initialStates.clone();
		clone.finalStates = finalStates.clone();
		clone.nonFinalStates = nonFinalStates.clone();
		clone.states = states.clone();

		for (Operation operation : stateTransBDDs.keySet()) {
			clone.stateTransBDDs.put(operation, stateTransBDDs.get(operation).clone());
		}

		return clone;
	}

	/**
	 * Returns the maximum number of vertices.
	 * @return the vertex maximum
	 */
	public int getMaximum() {
		return maximum;
	}

	@Override
	public void initializeEncoding() {
		encoding = BoundEncoding.getEncoding(maximumInterface,
				innerInterface, outerInterface, maximum, BoundEncoding.Type.MAXIMUM, encodingName);
		super.encoding = encoding;
		super.initializeEncoding();
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	protected void createStates() {
		// Log entering
		logger.entering(getClass().getName(), "createStates");

		states = encoding.zero();
		for (int i = 0; i <= maximumInterface; i++) {
			RavenBDD temp = encoding.getInterfaceSize(StateType.current, i);
			for (int j = 0; j <= maximum+1; j++) {
				states = states.or(temp.clone().and(getCounterEquals(StateType.current, j)));
			}
		}

		// Log exiting
		logger.exiting(getClass().getName(), "createStates");

	}

	@Override
	protected void createInitialAndFinalStates() {
		// Log entering
		logger.entering(getClass().getName(), "createInitialAndFinalStates");

		// initial: counter and interface size are 0
		initialStates = encoding.getInterfaceSize(StateType.current, innerInterface);
		initialStates = initialStates.and(getCounterEquals(StateType.current, 0));

		finalStates = encoding.getInterfaceSize(StateType.current, outerInterface);
		RavenBDD temp = encoding.zero();
		for (int i=0; i<=maximum-outerInterface; i++){
			temp = temp.or(getCounterEquals(StateType.current, i));
		}
		finalStates = finalStates.and(temp);

		nonFinalStates = states.clone().and(finalStates.clone().not());

		// Log exiting
		logger.exiting(getClass().getName(), "createInitialAndFinalStates");
	}

	@Override
	protected void createTransitions() {
		// Log entering
		logger.entering(getClass().getName(), "createTransitions");

		// iterate over all operations
		for (Operation operation : getAlphabet()) {
			RavenBDD transBDD = stateTransBDDs.get(operation);
			RavenBDD tempBDD;

			if (operation instanceof Connect) {
				int arity = ((Connect) operation).getArity();
				for (int size = arity; size <= maximumInterface; size++) {
					// Check that the current interface size is i and that the successor interface 
					// is still i
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size));

					// The counter is not changed
					tempBDD = tempBDD.and(constructCounterUnchangedBDD());

					transBDD = transBDD.or(tempBDD);
					fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED,
							operation.toString(), maximumInterface, size));
				}
			}

			if (operation instanceof Trans || operation instanceof Perm) {
				for (int size = 2; size <= maximumInterface; size++) {
					// Check that the current interface size is i and that the successor interface 
					// is still i
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size));

					// The counter is not changed
					tempBDD = tempBDD.and(constructCounterUnchangedBDD());

					transBDD = transBDD.or(tempBDD);
					fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED, 
							operation.toString(), maximumInterface, size));
				}
			}
			if (operation instanceof Res) {
				for (int size = 1; size <= maximumInterface; size++) {
					for (int count = 0; count <= maximum; count++) {
						// Check that the current interface size is i and that the successor 
						// interface is i-1
						tempBDD = encoding.getInterfaceSize(StateType.current, size);
						tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size-1));

						// The counter is increased by 1
						tempBDD = tempBDD.and(constructCounterChangedBDD(count, count+1));

						transBDD = transBDD.or(tempBDD);
						fireComputationEvent(new ComputationEvent(this, 
								EventType.COMPUTATION_ADVANCED, operation.toString(), 
								maximumInterface, size));
					}
				}
			}
			if (operation instanceof Vertex) {
				for (int size = 0; size < maximumInterface; size++) {
					// Check that the current interface size is i and that the successor interface 
					// is i+1
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size + 1));

					// The node count is not changed because only the res operations are counted
					tempBDD = tempBDD.and(constructCounterUnchangedBDD());

					transBDD = transBDD.or(tempBDD);
					fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED, 
							operation.toString(), maximumInterface, size));
				}
			}
			if (operation instanceof Fuse) {
				for (int size = 2; size <= maximumInterface; size++) {
					// Check that the current interface size is i and that the successor interface 
					// is i-1
					tempBDD = encoding
							.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(
							StateType.successor, size - 1));

					// The node count remains unchanged because only the res operations are counted
					tempBDD = tempBDD.and(constructCounterUnchangedBDD());

					transBDD = transBDD.or(tempBDD);
					fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED, 
							operation.toString(), maximumInterface, size));
				}
			}

			stateTransBDDs.put(operation, transBDD);
		}

		// Log exiting
		logger.exiting(getClass().getName(), "createTransitions");
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns a BDD encoding that in the state {@code state} the size of the
	 * remainder is equal to {@code value} if and only if the BDD is
	 * {@code true}.
	 * 
	 * @param type
	 *            the state type, either <i>current state</i> or <i>successor
	 *            state</i>
	 * @param value
	 *            the value of the remainder
	 * @return a BDD encoding that the remainder is equal to {@code value}
	 */
	private RavenBDD getCounterEquals(StateType type, int value) {
		if (value < 0) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_COUNTER_EXCEPTION);
		}
		else if (value > maximum+1) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_COUNTER_EXCEPTION);
		}

		RavenBDD bdd = encoding.one();
		for (int i = 0; i < maximumBitLength; i++) {
			int b = value % 2;
			value >>= 1;

			if (b == 0) {
				bdd = bdd.and(encoding.getValueBit(type, i).not());
			} else {
				bdd = bdd.and(encoding.getValueBit(type, i));
			}
		}

		return bdd;
	}

	/**
	 * Returns a BDD encoding that the counter equals {@code before} in the
	 * current state and equals {@code after} in the successor state if and only
	 * if the BDD is {@code true}.
	 * 
	 * @return a BDD encoding that the counter has been changed correctly
	 */
	private RavenBDD constructCounterChangedBDD(int before, int after) {
		return getCounterEquals(StateType.current, before).and(
				getCounterEquals(StateType.successor, after));
	}

	/**
	 * Returns a BDD encoding that the counter has not been changed between the
	 * current and the successor state if and only if the BDD is {@code true}.
	 * 
	 * @return a BDD encoding that the counter has not been changed
	 */
	private RavenBDD constructCounterUnchangedBDD() {
		RavenBDD bdd = encoding.one();

		for (int i = 0; i < maximumBitLength; i++) {
			bdd = bdd.and(encoding.getValueBit(StateType.current, i).biimp(
					encoding.getValueBit(StateType.successor, i)));
		}

		return bdd;
	}

}
