/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.data.automata;

import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.Operation.Connect;
import de.uni_due.inf.ti.raven.data.Operation.Fuse;
import de.uni_due.inf.ti.raven.data.Operation.Perm;
import de.uni_due.inf.ti.raven.data.Operation.Res;
import de.uni_due.inf.ti.raven.data.Operation.Trans;
import de.uni_due.inf.ti.raven.data.Operation.Vertex;
import de.uni_due.inf.ti.raven.data.Signature;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.encodings.BDDEncoding.StateType;
import de.uni_due.inf.ti.raven.encodings.NoIsolatedNodesEncoding;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class implements a graph automaton accepting the language of all graphs which do not 
 * contain any isolated node.
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.31 $
 *
 */
public class NoIsolatedNodesAutomaton extends Automaton {

	private static final long serialVersionUID = -3410626841660318163L;

	transient private NoIsolatedNodesEncoding encoding;
	transient private Logger logger;
	
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new no-isolated-nodes automaton.
	 * @param encoding the encoding for the states
	 * @param signature the signature of this automaton
	 */
	protected NoIsolatedNodesAutomaton(NoIsolatedNodesEncoding encoding, Signature signature) {
		super(Type.noIsolatedNodes, encoding, signature);
		this.encoding = encoding;

		//Register logger
		logger = Logger.getLogger(getClass().getName());
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public Automaton clone() throws CloneNotSupportedException {
		Automaton clone = new NoIsolatedNodesAutomaton(encoding, signature);
		clone.initialStates = initialStates.clone();
		clone.finalStates = finalStates.clone();
		clone.nonFinalStates = nonFinalStates.clone();
		clone.states = states.clone();
		
		for(Operation operation: stateTransBDDs.keySet()) {
			clone.stateTransBDDs.put(operation, stateTransBDDs.get(operation).clone());
		}

		return clone;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	protected void createStates() {
		//Log entering
		logger.entering(getClass().getName(), "createStates");

		states = encoding.zero();
		for(int i=0;i<=maximumInterface;i++) {
			RavenBDD temp = encoding.getInterfaceSize(StateType.current, i);
			for(int j=i;j<maximumInterface;j++) {
				temp = temp.and(encoding.getUndefinedNode(StateType.current, j));
			}
			states = states.or(temp);
		}

		//Log exiting
		logger.exiting(getClass().getName(), "createStates");
	}

	@Override
	protected void createInitialAndFinalStates() {
		//Log entering
		logger.entering(getClass().getName(), "createInitialAndFinalStates");

		initialStates = encoding.getInterfaceSize(StateType.current, innerInterface);
		finalStates = encoding.getInterfaceSize(StateType.current, outerInterface);
		for(int i=0;i<maximumInterface;i++) {
			if(i<innerInterface) {
				initialStates = initialStates.and(encoding.getIsNodeIsolated(StateType.current, i));				
			}
			else {
				initialStates = initialStates.and(encoding.getUndefinedNode(StateType.current, i));
			}
			if(i<outerInterface) {
				finalStates = finalStates.and(encoding.getIsNodeIsolated(StateType.current, i).not());				
			}
			else {
				finalStates = finalStates.and(encoding.getUndefinedNode(StateType.current, i));
			}	
		}
		if(innerInterface == 0) {
			initialStates = initialStates.and(encoding.getHasIsolatedNodes(StateType.current).not());			
		}
		else {
			initialStates = initialStates.and(encoding.getHasIsolatedNodes(StateType.current));			
		}
		finalStates = finalStates.and(encoding.getHasIsolatedNodes(StateType.current).not());

		nonFinalStates = states.clone().and(finalStates.clone().not());
		
		//Log exiting
		logger.exiting(getClass().getName(), "createInitialAndFinalStates");
	}

	@Override
	protected void createTransitions() {
		//Log entering
		logger.entering(getClass().getName(), "createTransitions");

		// Iteriere ueber alle Graphoperationen
		for (Operation operation: getAlphabet()) {
			RavenBDD transBDD = stateTransBDDs.get(operation);
			RavenBDD tempBDD;

			// Falls die Operation die das Verkleben mit einer Hyperkante ist
			if (operation instanceof Connect) {
				// Eigenschaften der neuen Kante bestimmen
				int arity = ((Connect) operation).getArity();

				//The CONNECT-operation is applicable only if the arity of the edge 
				//is greater or equal to the current interface size.
				for(int size=arity;size<=maximumInterface;size++) {
					//Check that the current interface size is i and that the interface 
					//is not changed by the CONNECT-operation
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size));

					tempBDD = tempBDD.and(addEdge(size, arity));
					tempBDD = tempBDD.and(getInterfaceUndefined(size));
					transBDD = transBDD.or(tempBDD);
				}
			}
			else if (operation instanceof Fuse) {
				//The FUSE-operation is applicable only if at least two nodes are "touchable" 
				//through the interface, i.e. the interface size has to be at least 2
				for(int size=2;size<=maximumInterface;size++) {
					//Check that the current interface size is i and that the successor 
					//interface is decreased by one
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size-1));

					tempBDD = tempBDD.and(fuseNodes(size));
					tempBDD = tempBDD.and(getInterfaceUndefined(size));
					transBDD = transBDD.or(tempBDD);
				}
			}
			else if (operation instanceof Perm) {
				//The PERM-operation is applicable only if at least two nodes are "touchable" 
				//through the interface, i.e. the interface size has to be at least 2
				for(int size=2;size<=maximumInterface;size++) {
					//Check that the current interface size is i and that the interface 
					//is not changed by the PERM-operation
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size));

					tempBDD = tempBDD.and(permuteNodes(size));
					tempBDD = tempBDD.and(getInterfaceUndefined(size));
					transBDD = transBDD.or(tempBDD);
				}
			}
			else if (operation instanceof Res) {
				//The RES-operation is applicable only if at least one node is "touchable"
				//through the interface, i.e. the interface size has to be at least 2
				for(int size=1;size<=maximumInterface;size++) {
					//Check that the current interface size is i and that the interface 
					//size of the successor interface is i-1
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size-1));

					tempBDD = tempBDD.and(removeNode(size));
					tempBDD = tempBDD.and(getInterfaceUndefined(size));
					transBDD = transBDD.or(tempBDD);
				}
			}
			else if (operation instanceof Trans) {
				//The TRANS-operation is applicable only if at least two nodes are "touchable" 
				//through the interface, i.e. the interface size has to be at least 2
				for(int size=2;size<=maximumInterface;size++) {
					//Check that the current interface size is i and that the interface 
					//is not changed by the TRANS-operation
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size));

					tempBDD = tempBDD.and(transposeNodes(size));
					tempBDD = tempBDD.and(getInterfaceUndefined(size));
					transBDD = transBDD.or(tempBDD);
				}
			}
			// Transitionsfunktion fuer die VERTEX-Operation
			else if (operation instanceof Vertex) {
				//The VERTEX-operation is applicable only if at least one node could be added
				//to the interface, i.e. the interface size has to be at most (maxSize-1)
				for(int size=0;size<maximumInterface;size++) {
					//Check that the current interface size is i and that the interface 
					//size of the successor interface is i+1
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size+1));

					tempBDD = tempBDD.and(addNode(size));
					tempBDD = tempBDD.and(getInterfaceUndefined(size+1));
					transBDD = transBDD.or(tempBDD);

				}
			}
			stateTransBDDs.put(operation, transBDD);
		}

		//Log exiting
		logger.exiting(getClass().getName(), "createTransitions");
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public void initializeEncoding() {
		encoding = NoIsolatedNodesEncoding.getEncoding(maximumInterface, 0, 0, encodingName);
		super.encoding = encoding;
		super.initializeEncoding();
	}
	
	/**
	 * Constructs a BDD which is {@code true} if and only if in both the 
	 * current state and the successor state the interface node at position 
	 * {@code node} is undefined.
	 * @param node the index of the node
	 * @return a BDD encoding that the interface node is undefined
	 */
	private RavenBDD getInterfaceUndefined(int node) {
		RavenBDD resultBDD = encoding.one();
		for(int i=node;i<maximumInterface;i++) {
			resultBDD = resultBDD.and(
					encoding.getUndefinedNode(StateType.current, i).and(
							encoding.getUndefinedNode(StateType.successor, i)));
		}

		return resultBDD;
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if in both the 
	 * current state and the successor state occur isolated nodes or in 
	 * both no isolated node occurs.
	 * @return a BDD encoding that the (non)-existence of isolated nodes 
	 * has not been changed.
	 */
	private RavenBDD isolationStateNotChanged() {
		return encoding.getHasIsolatedNodes(StateType.current).biimp(
				encoding.getHasIsolatedNodes(StateType.successor));
	}
	
	/**
	 * Constructs a BDD which is {@code true} if and only if in both the 
	 * current state and the successor state the interface node at position 
	 * {@code node} is isolated or is not isolated.
	 * @param node the position of the interface node
	 * @return a BDD encoding that the "isolation state" the interface node 
	 * has not been changed
	 */
	private RavenBDD interfaceNotChanged(int node) {
		return encoding.getIsNodeIsolated(StateType.current, node).biimp(
				encoding.getIsNodeIsolated(StateType.successor, node));
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if a new edge of 
	 * arity {@code arity} is added and
	 * <ul>
	 * 		<li>the "isolation state" of the first {@code size-arity} 
	 * 		interface nodes have not been changed</li>
	 * 		<li>the last {@code arity} interface nodes have been marked as "not isolated"</li>
	 * 		<li>the "overall isolation state" not been changed</li>
	 * </ul>
	 * @param size the size of the current interface
	 * @param arity the arity of the new edge
	 * @return a BDD encoding that the new edge has been added correctly
	 */
	private RavenBDD addEdge(int size, int arity) {
		RavenBDD resultBDD = encoding.one();
		for(int node=0;node<size;node++) {
			if(node<size-arity) {
				resultBDD = resultBDD.and(interfaceNotChanged(node));
			}
			else {
				resultBDD = resultBDD.and(
						encoding.getIsNodeIsolated(StateType.successor, node).not());
			}
		}
		return resultBDD.and(isolationStateNotChanged());
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if 
	 * <ul>
	 * 		<li>the "isolation state" of the first {@code size-2} interface nodes 
	 * 		(of the current state) has not been changed, and</li>
	 * 		<li>the next-to-last interface node (of the current state) is isolated 
	 * 		in the successor state if the next-to-last and the last interface node 
	 * 		(of the current state) are isolated in the current state, and</li>
	 * 		<li>the last node (of the current state) is removed from the interface of 
	 * 		the successor state 
	 * 		<li>the "overall isolation state" has not been changed</li>
	 * </ul>
	 * 
	 * @param size the interface size
	 * @return a BDD encoding that the last two interface nodes have been fused correctly
	 */
	private RavenBDD fuseNodes(int size) {
		RavenBDD resultBDD = encoding.one();
		for(int node=0;node<size;node++) {
			if(node<size-2) {
				resultBDD = resultBDD.and(interfaceNotChanged(node));
			}
			else if(node==size-2) {
				RavenBDD tempBDD = encoding.getIsNodeIsolated(StateType.current, node).and(
						encoding.getIsNodeIsolated(StateType.current, node+1));
				tempBDD = tempBDD.biimp(encoding.getIsNodeIsolated(StateType.successor, node));
				resultBDD = resultBDD.and(tempBDD);
			}
			else if (node==size-1) {
				resultBDD = resultBDD.and(
						encoding.getIsNodeIsolated(StateType.successor, node).not());
			}
		}
		return resultBDD.and(isolationStateNotChanged());
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if 
	 * <ul>
	 * 		<li>the first interface node of the current state is mapped to 
	 * 		the second interface node of the successor state, and</li>
	 * 		<li>the second interface node of the current state is mapped to 
	 * 		the third interface node of the successor state, and</li>
	 * 		</li>..., and</li>
	 * 		<li>the last interface node of the current state is mapped to 
	 * 		the first interface node of the successor state</li>
	 * 		<li>the "overall isolation state" has not been changed</li>
	 * </ul>

	 * @param size the interface size
	 * @return a BDD encoding that the interface nodes have been permuted correctly
	 */
	private RavenBDD permuteNodes(int size) {
		RavenBDD resultBDD = encoding.one();
		for(int node=0;node<size;node++) {
			resultBDD = resultBDD.and(
					encoding.getIsNodeIsolated(StateType.current, node).biimp(
							encoding.getIsNodeIsolated(StateType.successor, (node+1)%size)));
		}
		return resultBDD.and(isolationStateNotChanged());
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if 
	 * <ul>
	 * 		<li>the first {@code size-1} interface nodes have not been changed</li>
	 * 		<li>the last interface node is removed from the successor state's interface, and</li>
	 * 		<li>the graph contains isolated nodes if and only if the graph has contained 
	 * 		isolated nodes before or if the removed node is isolated</li>
	 * </ul>
	 * 
	 * @param size the size of the interface of the current state
	 * @return a BDD encoding that the last node of the interface has been removed correctly
	 */
	private RavenBDD removeNode(int size) {
		if (size < 1 || size > maximumInterface) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}
		
		RavenBDD resultBDD = encoding.one();
		for(int node=0;node<size;node++) {
			if(node<size-1) {
				resultBDD = resultBDD.and(interfaceNotChanged(node));
			}
			else if(node==size-1) {
				RavenBDD tempBDD = encoding.getIsNodeIsolated(StateType.current, node).or(
						encoding.getHasIsolatedNodes(StateType.current));
				resultBDD = resultBDD.and(tempBDD.biimp(encoding.getHasIsolatedNodes(StateType.successor)));
				resultBDD = resultBDD.and(encoding.getIsNodeIsolated(StateType.successor, node).not());
			}
		}
//		return resultBDD.and(isolationStateNotChanged());
		return resultBDD;
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if 
	 *  <ul>
	 * 		<li>the first node of the current state's interface is mapped 
	 * 		to the second node of the successors state's interface, and</li>
	 * 		<li>the second node of the current state's interface is mapped 
	 * 		to the first node of the successors state's interface, and</li>
	 * 		<li>all other nodes are not permuted</li>
	 * </ul>
	 * @param size the size of the interface of the current state
	 * @return a BDD encoding that the first two nodes of the interface are transposed
	 */
	private RavenBDD transposeNodes(int size) {
		RavenBDD resultBDD = encoding.one();
		for(int node=0;node<size;node++) {
			if(node==0 || node==1) {
				resultBDD = resultBDD.and(
						encoding.getIsNodeIsolated(StateType.current, node).biimp(
								encoding.getIsNodeIsolated(StateType.successor, (node+1)%2)));
			}
			else {
				resultBDD = resultBDD.and(interfaceNotChanged(node));
			}
		}
		return resultBDD.and(isolationStateNotChanged());
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if a new node is added to the 
	 * last position of interface (i.e. the {@code size+1}-th position) and
	 * <ul>
	 * 		<li>the new interface node has been undefined in the current state</li>
	 * 		<li>the new interface node is isolated in the successor state</li>
	 * 		<li>the "overall isolation state" has not been changed</li>
	 * </ul>
	 * 
	 * @param size the size of the interface of the current state
	 * @return a BDD encoding that a new node is added correctly to the last position of the 
	 * interface
	 */
	private RavenBDD addNode(int size) {
		RavenBDD resultBDD = encoding.one();
		for(int node=0;node<size;node++) {
			resultBDD = resultBDD.and(interfaceNotChanged(node));
		}

		resultBDD = resultBDD.and(
				encoding.getIsNodeIsolated(StateType.current, size).not().and(
						encoding.getIsNodeIsolated(StateType.successor, size)));

		return resultBDD.and(isolationStateNotChanged());
	}	

}
