/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.data.automata;

import java.util.Set;
import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.actions.ComputationEvent;
import de.uni_due.inf.ti.raven.actions.ComputationEvent.EventType;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.Operation.Fuse;
import de.uni_due.inf.ti.raven.data.Operation.Perm;
import de.uni_due.inf.ti.raven.data.Operation.Res;
import de.uni_due.inf.ti.raven.data.Operation.Trans;
import de.uni_due.inf.ti.raven.data.Operation.Vertex;
import de.uni_due.inf.ti.raven.data.Signature;
import de.uni_due.inf.ti.raven.data.Operation.Connect;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.encodings.PathEncoding;
import de.uni_due.inf.ti.raven.encodings.BDDEncoding.StateType;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class implements a graph automaton accepting the language of all graphs
 * where certain nodes are connected by a path. Furthermore the inner and outer
 * interface sized have to be fulfilled.
 * 
 * @author Christoph Blume, Dennis Nolte
 * @version $Revision: 1.36 $
 * 
 * @see PathEncoding
 */

public class PathAutomaton extends Automaton {

	private static final long serialVersionUID = -8616849370942972892L;

	private final Set<Integer> sourceNodes;
	private final Set<Integer> targetNodes;

	transient private PathEncoding encoding;	
	transient private Logger logger;

	/*------------------------------------------------------------------------------------------*/
	/* Constructors */
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new path automaton.
	 * @param encoding the encoding for the states
	 * @param signature the signature of this automaton
	 * @param sourceNodes the source nodes
	 * @param targetNodes the target nodes
	 */
	protected PathAutomaton(PathEncoding encoding, Signature signature, Set<Integer> sourceNodes, 
			Set<Integer> targetNodes) {
		super(Type.path, encoding, signature);
		
		if(sourceNodes==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SOURCE_NODE_SET_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(sourceNodes.size()>innerInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.SOURCE_NODE_SIZE_MUST_NOT_BE_GREATER_THAN_INNER_INTERFACE_EXCEPTION);
		}
		for(int node: sourceNodes) {
			if(node < 0 || node >= innerInterface) {
				throw new IllegalArgumentException(
						RavenExceptionMessages.SOURCE_NODE_RANGE_EXCEPTION);
			}
		}
		if(targetNodes==null) {
			throw new NullPointerException(
					RavenExceptionMessages.TARGET_NODE_SET_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(targetNodes.size()>outerInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.TARGET_NODE_SIZE_MUST_NOT_BE_GREATER_THAN_OUTER_INTERFACE_EXCEPTION);
		}
		for(int node: targetNodes) {
			if(node < 0 || node >= outerInterface ) {
				throw new IllegalArgumentException(
						RavenExceptionMessages.TARGET_NODE_RANGE_EXCEPTION);
			}
		}
		
		this.encoding = encoding;
		this.sourceNodes = sourceNodes;
		this.targetNodes = targetNodes;
		
		// Register logger
		logger = Logger.getLogger(getClass().getName());
	}

	/*------------------------------------------------------------------------------------------*/
	/* Public Methods */
	/*------------------------------------------------------------------------------------------*/

	@Override
	public Automaton clone() throws CloneNotSupportedException {
		Automaton clone = new PathAutomaton(encoding, signature, sourceNodes, targetNodes);
		clone.initialStates = initialStates.clone();
		clone.finalStates = finalStates.clone();
		clone.nonFinalStates = nonFinalStates.clone();
		clone.states = states.clone();

		for (Operation operation : stateTransBDDs.keySet()) {
			clone.stateTransBDDs.put(operation, stateTransBDDs.get(operation)
					.clone());
		}

		return clone;
	}
	
	/**
	 * Returns the source nodes.
	 * @return the source nodes
	 */
	public final Set<Integer> getSourceNodes() {
		return sourceNodes;
	}
	
	/**
	 * Returns the target nodes.
	 * @return the target nodes
	 */
	public final Set<Integer> getTargetNodes() {
		return targetNodes;
	}

	/*------------------------------------------------------------------------------------------*/
	/* Protected Methods */
	/*------------------------------------------------------------------------------------------*/

	@Override
	protected void createStates() {
		// Log entering
		logger.entering(getClass().getName(), "createStates");

		states = encoding.zero();
		for (int i = 0; i <= maximumInterface; i++) {
			RavenBDD temp = encoding.getInterfaceSize(StateType.current, i);

			for (int node = 0; node < i; node++) {
				temp = temp.and(encoding.getPathBit(StateType.current,
						node, node));
				for (int bit = i; bit < maximumInterface; bit++) {
					temp = temp.and(encoding.getPathBit(StateType.current,
							node, bit).not());
				}
			}
			for (int node = i; node < maximumInterface; node++) {

				for (int bit = 0; bit < maximumInterface; bit++) {
					temp = temp.and(encoding.getPathBit(StateType.current,
							node, bit).not());
				}
			}
			for (int bit= i; bit<maximumInterface;bit++){
				temp = temp.and(encoding.getSourcePathBit(StateType.current, bit).not());
			}

			states = states.or(temp);
		}

		// Log exiting
		logger.exiting(getClass().getName(), "createStates");
	}

	@Override
	protected void createInitialAndFinalStates() {
		// Log entering
		logger.entering(getClass().getName(), "createInitialAndFinalStates");

		// The initial states consists of the matrix where only the diagonal is
		// set to 1 up to the inner interface size
		initialStates = encoding.getInterfaceSize(StateType.current, innerInterface);
		for (int node = 0; node < maximumInterface; node++) {
			for (int bit = 0; bit < maximumInterface; bit++) {
				if (node < innerInterface && bit == node) {
					initialStates = initialStates.and(
							encoding.getPathBit(StateType.current, node, bit));
				} else {
					initialStates = initialStates.and(
							encoding.getPathBit(StateType.current, node, bit).not());
				}
			}
		}
		RavenBDD tempBDD = encoding.one();
		for(int bit=0;bit<maximumInterface;bit++){
			if(sourceNodes.contains(bit)){
				tempBDD = tempBDD.and(encoding.getSourcePathBit(StateType.current, bit));
			}
			else{
				tempBDD = tempBDD.and(encoding.getSourcePathBit(StateType.current, bit).not());
			}
		}		
		
		initialStates = initialStates.and(tempBDD);

		// A final state has to fulfill the same conditions as an initial state
		// regarding the diagonal...
		finalStates = encoding.getInterfaceSize(StateType.current, outerInterface);
		for (int node = 0; node < outerInterface; node++) {
			finalStates = finalStates.and(encoding.getPathBit(StateType.current, node, node));
			for (int bit = outerInterface; bit < maximumInterface; bit++) {
				finalStates = finalStates.and(
						encoding.getPathBit(StateType.current, node, bit).not());
			}
		}
		for (int node = outerInterface; node < maximumInterface; node++) {
			for (int bit = 0; bit < maximumInterface; bit++) {
				finalStates = finalStates.and(
						encoding.getPathBit(StateType.current, node, bit).not());
			}
		}

		// ...and has an adjacency bit with 1 at (source,target) entry if there
		// exists a path from a source to a target
		tempBDD = encoding.zero();
		for(int node=0;node<maximumInterface;node++) {
			if(targetNodes.contains(node)) {
				tempBDD = tempBDD.or(encoding.getSourcePathBit(StateType.current, node));
			}
		}
		finalStates = finalStates.and(tempBDD);
		nonFinalStates = states.clone().and(finalStates.clone().not());

		// Log exiting
		logger.exiting(getClass().getName(), "createInitialAndFinalStates");
	}

	@Override
	protected void createTransitions() {
		// Log entering
		logger.entering(getClass().getName(), "createTransitions");
		
		// Iterate over all graph operations
		int letterCounter = 0;
		for (Operation letter : getAlphabet()) {
			RavenBDD transBDD = stateTransBDDs.get(letter);
			RavenBDD tempBDD;

			if (letter instanceof Connect) {
				// The CONNECT-operation is applicable only if the size of the
				// interface is greater or
				// equal to the arity of the new hyperedge added to the graph
				// i.e. the interface size has to be at most m
				int arity = ((Connect) letter).getArity();

				for (int size = arity; size <= maximumInterface; size++) {

					tempBDD = encoding.one();
					// For all nodes which could possibly occur in the
					// interface...
					for (int node = 0; node < size - arity; node++) {
						// ..check that the nodes before column k-m and bits
						// before row k remain the same...
						for (int bit = 0; bit < size; bit++) {
							if (bit == node) {
								tempBDD = tempBDD.and(
										createAdjacencyPreservedBDD(node, bit));
							} else {
								tempBDD = tempBDD.and(
										createAdjacencyUnchangedBDD(node, bit));
							}
						}
						// ...and all other bits of those columns are set to 0
						for (int bit = size; bit < maximumInterface; bit++) {
							tempBDD = tempBDD.and(createNoAdjacencyBDD(node, bit));
						}
					}

					// For all nodes which could possibly occur in the
					// interface...
					for (int node = size - arity; node < size; node++) {
						// ..check that the nodes between column k-m and k with
						// bits before row k-m remain the same...
						for (int bit = 0; bit < size - arity; bit++) {
							tempBDD = tempBDD.and(createAdjacencyUnchangedBDD(
									node, bit));

						}
						// ...the bits in those columns between row k-m and k
						// are set to 1...
						for (int bit = size - arity; bit < size; bit++) {
								tempBDD = tempBDD.and(
										encoding.getPathBit(StateType.successor, node, bit));
								tempBDD = tempBDD.and(
										encoding.getPathBit(StateType.current, node, bit));
						}
						// ...and all other bits of those columns are set to 0
						for (int bit = size; bit < maximumInterface; bit++) {
							tempBDD = tempBDD.and(createNoAdjacencyBDD(node, bit));
						}
					}

					// All other columns are set to 0
					for (int node = size; node < maximumInterface; node++) {
						for (int bit = 0; bit < maximumInterface; bit++) {
							tempBDD = tempBDD.and(createNoAdjacencyBDD(node, bit));
						}
					}					
					
					// All bits have to be recalculated due to the fact of
					// transivity in the adjacency matrix
					tempBDD = createSquareBDD(tempBDD.clone(),size);
					
					// After the square method the adjacency bits in the current interface between
					// k-m and k that were set to 1 aren't of any interest any more
					for (int node = size - arity; node < size; node++) {
						for (int bit = size - arity; bit < size; bit++) {
							if (bit != node) {
								tempBDD = tempBDD.clone().exists(
										encoding.getPathBit(StateType.current, node, bit));		
							}
						}						
					}
										
					// Check that the current interface size is i and that the
					// interface
					// is not changed by the CONNECT-operation
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.current, size));
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size));

					// The connection bits are recalculated by a multiplication between the connection list and the adjacency matrix
					RavenBDD equationBDD;
					for(int node=0;node<maximumInterface;node++){
						if(node<size){
							equationBDD = encoding.zero();
							for(int bit=0;bit<size;bit++){
								equationBDD = equationBDD.or(
										encoding.getSourcePathBit(StateType.current, bit).and(
												encoding.getPathBit(StateType.successor, node, bit)));
							}
							tempBDD = tempBDD.and(
									encoding.getSourcePathBit(StateType.successor, node).biimp(equationBDD));
						}
						else{
							tempBDD = tempBDD.and(createNoConnectionBDD(node));
						}	
					}
					
					transBDD = transBDD.or(tempBDD);
				}
			} else if (letter instanceof Fuse) {
				// The FUSE-operation is applicable only if at least two nodes
				// could be fused
				// from the interface, i.e. the interface size has to be at most
				// 2
				RavenBDD[][] resultBDD;
				
				for (int size = 2; size <= maximumInterface; size++) {
					resultBDD = new RavenBDD[size-1][size-1];
					tempBDD = encoding.one();

					//Building the formula Matrix for the first step
					//For all nodes before column k-2 and all bits before row k-2 all bits have to be recalculated 
					//except for the diagonal
					for(int node=0;node<size-2;node++){
						
						for(int bit=0;bit<size-2;bit++){
							resultBDD[node][bit] = encoding.getPathBit(StateType.current, node, bit);
						}
						
						resultBDD[node][size-2] = encoding.getPathBit(StateType.current, node, size-2).or(
								encoding.getPathBit(StateType.current, node, size-1));
					}
					
					//For all bits in column k-2 the bits have to be calculated by a logical OR with their k-1 correspondents
					for(int bit=0;bit<size-2;bit++){
						resultBDD[size-2][bit] = encoding.getPathBit(StateType.current, size-2, bit).or(
								encoding.getPathBit(StateType.current, size-1, bit));
					}
					resultBDD[size-2][size-2] = encoding.getPathBit(StateType.current, size-2, size-2);
					
					//Building the formula matrix via an iteration over squaring the matrix
					RavenBDD[][] oldResultBDD = resultBDD.clone();
					for(int iterate=0;iterate<maximumInterface;iterate++){

						//For all nodes of the matrix...
						for (int node = 0; node < size-1; node++) {
							//...and all bits of the matrix...
							for (int bit = 0; bit < size-1; bit++) {
									if(node!=bit){
										//...calculate the logical formula for a multiplication of the bits
										RavenBDD formulaBDD = encoding.zero();
										for (int z = 0; z < size-1; z++) {
											formulaBDD = formulaBDD.or(
													oldResultBDD[z][bit].clone().and(
															oldResultBDD[node][z].clone()));
											
										}
										resultBDD[node][bit] = formulaBDD.clone();										
									}
							}		
						}
						oldResultBDD = resultBDD.clone();
					}		
					
					//Building the formula calculated upper square in the tempBDD
					for(int node=0;node<size-1;node++){
						for(int bit=0;bit<size-1;bit++){
							if(node==bit){
								tempBDD = tempBDD.and(createAdjacencyPreservedBDD(node, bit));
							}
							else{
								tempBDD = tempBDD.and(
										encoding.getPathBit(StateType.successor, node, bit).biimp(
												resultBDD[node][bit]));
							}
						}
						for(int bit=size-1;bit<maximumInterface;bit++){
							if(bit!=size-1){
								tempBDD = tempBDD.and(
										encoding.getPathBit(StateType.current, node, bit).not());
							}
							tempBDD = tempBDD.and(
									encoding.getPathBit(StateType.successor, node, bit).not());
						}
					}
					
					//Building column k-1 in the adjacency matrix by describing that the bits have to be 0 afterward
					for(int bit=0;bit<maximumInterface;bit++){
						if(bit < size-1){
							tempBDD = tempBDD.and(
									encoding.getPathBit(StateType.successor, size-1, bit).not());
						}
						else if(bit == size-1){
							tempBDD = tempBDD.and(
									encoding.getPathBit(StateType.current, size-1, bit));
							tempBDD = tempBDD.and(
									encoding.getPathBit(StateType.successor, size-1, bit).not());
						}
						else{
							tempBDD = tempBDD.and(createNoAdjacencyBDD(size-1, bit));
						}
					}
					
					// For all nodes after the k-1 column make sure that a 0 is preserved
					for(int node=size;node<maximumInterface;node++){
						for(int bit = 0;bit<maximumInterface;bit++){
							tempBDD = tempBDD.and(createNoAdjacencyBDD(node, bit));
						}
					}
					
					//For symmetry purposes
					for (int node = 0; node < maximumInterface; node++) {
						for (int bit = 0; bit < maximumInterface; bit++) {			
								tempBDD.and(
										encoding.getPathBit(StateType.successor, node, bit).biimp(
												encoding.getPathBit(StateType.successor, bit, node)));
								tempBDD.and(
										encoding.getPathBit(StateType.current, node, bit).biimp(
												encoding.getPathBit(StateType.current, bit, node)));
						}
					}
					
					// Check that the current interface size is i and that the
					// interface
					// size of the successor interface is i-1
					tempBDD = tempBDD.and(encoding.getInterfaceSize(
							StateType.current, size));
					tempBDD = tempBDD.and(encoding.getInterfaceSize(
							StateType.successor, size - 1));

					//Preparing a formula array for the connection list.
					RavenBDD[] preCalcBDD = new RavenBDD[size-1];
					for(int bit=0;bit<size-2;bit++){
						preCalcBDD[bit] = encoding.getSourcePathBit(StateType.current, bit);
					}
					preCalcBDD[size-2] = encoding.getSourcePathBit(StateType.current, size-2).or(
							encoding.getSourcePathBit(StateType.current, size-1));
					
					//Recalculating the connection bits by multiplying the formula array with the adjacency matrix
					RavenBDD equationBDD;
					for(int node=0;node<maximumInterface;node++){
						if(node<size-1){
							equationBDD = encoding.zero();
							for(int bit=0;bit<size-1;bit++){
								equationBDD = equationBDD.or(preCalcBDD[bit].clone().and(
										encoding.getPathBit(StateType.successor, node, bit)));
							}
							tempBDD = tempBDD.and(
									encoding.getSourcePathBit(StateType.successor, node).biimp(
											equationBDD));
						}
						else if(node==size-1){
							tempBDD = tempBDD.and(
									encoding.getSourcePathBit(StateType.successor, node).not());
						}
						else{
							tempBDD = tempBDD.and(createNoConnectionBDD(node));
						}	
					}
						
					transBDD = transBDD.or(tempBDD);
				}
			} else if (letter instanceof Perm) {
				// The PERM-operation is applicable only if at least two nodes
				// already exist
				// in the interface, i.e. the interface size has to be at most 2
				for (int size = 2; size <= maximumInterface; size++) {
					// Check that the current interface size is i and that the
					// interface
					// size of the successor interface is i
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(
							StateType.successor, size));

					// For all nodes which could possibly occur in the
					// interface...
					for (int node = 0; node < size; node++) {
						// ..check that the nodes before column k and row k are
						// permuted...
						for (int bit = 0; bit < size; bit++) {
							if (bit == node) {
								tempBDD = tempBDD.and(createAdjacencyPreservedBDD(node, bit));
							} 
							else {
								//Calculating the permutation partners position
								int oldNode = (node-1)%size;
								oldNode = oldNode>=0 ? oldNode : oldNode+size;
								
								int oldBit = (bit-1)%size;
								oldBit = oldBit>=0 ? oldBit : oldBit+size;
								
								tempBDD = tempBDD.and(
										encoding.getPathBit(StateType.current, oldNode, oldBit).biimp(
												encoding.getPathBit(StateType.successor, node, bit)));
								
							}
						}
						// ...and for all other bits of the first k columns and below row k a 0 is preserved
						for (int bit = size; bit < maximumInterface; bit++) {
							tempBDD = tempBDD.and(createNoAdjacencyBDD(node,
									bit));
						}
					}

					// All other columns are set to 0
					for (int node = size; node < maximumInterface; node++) {
						for (int bit = 0; bit < maximumInterface; bit++) {
							tempBDD = tempBDD.and(createNoAdjacencyBDD(node,
									bit));
						}
					}
					
					//The connection bits are permutated up to the position of k and the rest is unchanged
					for(int bit=0;bit<size;bit++){
						int oldBit = (bit-1)%size;
						oldBit = oldBit>=0 ? oldBit : oldBit+size;
						
						tempBDD = tempBDD.and(
								encoding.getSourcePathBit(StateType.current, oldBit).biimp(
										encoding.getSourcePathBit(StateType.successor, bit)));
					}
					for(int bit = size;bit<maximumInterface;bit++){
						tempBDD = tempBDD.and(createNoConnectionBDD(bit));
					}
					
					transBDD = transBDD.or(tempBDD);
				}
			} else if (letter instanceof Res) {
				// The RES-operation is applicable only if at least one node
				// could be deleted
				// from the interface, i.e. the interface size has to be at most
				// 1
				for (int size = 1; size <= maximumInterface; size++) {
					// Check that the current interface size is i and that the
					// interface
					// size of the successor interface is i-1
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size - 1));

					// For all nodes which could possibly occur in the
					// interface...
					for (int node = 0; node < size - 1; node++) {
						// ..check that the nodes before column k-1 and row k-1
						// remain the same and bits on the diagonal are always
						// 1...
						for (int bit = 0; bit < size - 1; bit++) {
							if (bit == node) {
								tempBDD = tempBDD.and(createAdjacencyPreservedBDD(node,bit));
							} else {
								tempBDD = tempBDD.and(createAdjacencyUnchangedBDD(node,bit));
							}

						}
						// ...and for all other adjacency bits of these columns below row k-1 are
						// set to 0 afterwards
						for (int bit = size - 1; bit < maximumInterface; bit++) {
							if (bit == size - 1) {
								tempBDD = tempBDD.and(
										encoding.getPathBit(StateType.successor, node, bit).not());
							} else {
								tempBDD = tempBDD.and(createNoAdjacencyBDD(node, bit));
							}
						}
					}
					// The nodes after k-1 have to be 0
					for (int node = size; node < maximumInterface; node++) {
						for (int bit = 0; bit < maximumInterface; bit++) {
							tempBDD = tempBDD.and(createNoAdjacencyBDD(node, bit));
						}
					}

					// The node in column k-1 has 0 at each bit after the k-1
					// row
					for (int bit = 0; bit < maximumInterface; bit++) {
						if (bit < size - 1) {
							tempBDD = tempBDD.and(encoding.getPathBit(
									StateType.successor, size - 1, bit).not());
						} else if (bit == size - 1) {
							tempBDD = tempBDD.and(
									encoding.getPathBit(StateType.current, size - 1, bit).and(
											encoding.getPathBit(StateType.successor, size - 1, bit).not()));
						} else {
							tempBDD = tempBDD.and(
									encoding.getPathBit(StateType.current, size - 1, bit).not().and(
											encoding.getPathBit(StateType.successor, size - 1, bit).not()));
						}
					}
					
					//The k-1 connection bit is set to 0 the others are preserved
					for(int bit=0;bit<size;bit++){
						if(bit == size-1){
							tempBDD = tempBDD.and(
									encoding.getSourcePathBit(StateType.successor, bit).not());
						}
						else{
							tempBDD = tempBDD.and(createConnectionUnchangedBDD(bit));
						}
					}
					for(int bit=size;bit<maximumInterface;bit++){
						tempBDD = tempBDD.and(createNoConnectionBDD(bit));
					}
					
					transBDD = transBDD.or(tempBDD);
				}
			} else if (letter instanceof Trans) {
				// The TRANS-operation is applicable only if at least two nodes
				// already exist
				// in the interface, i.e. the interface size has to be at most 2
				for (int size = 2; size <= maximumInterface; size++) {
					// Check that the current interface size is i and that the
					// interface
					// size of the successor interface is i
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size));

					// For all nodes which could possibly occur in the
					// interface...
					for (int node = 0; node < 2; node++) {
						// ..check that the nodes before column 2 and row 2
						// remain the same and bits on the diagonal are always
						// 1...
						for (int bit = 0; bit < 2; bit++) {
							if (bit == node) {
								tempBDD = tempBDD.and(createAdjacencyPreservedBDD(node, bit));
							} else {
								tempBDD = tempBDD.and(createAdjacencyUnchangedBDD(node, bit));
							}
						}
						// ...and all bits of the first two columns are switched
						// up to the k-th row...
						for (int bit = 2; bit < size; bit++) {
							tempBDD = tempBDD.and(
									encoding.getPathBit(StateType.current, (node + 1) % 2, bit).biimp(
											encoding.getPathBit(StateType.successor, node, bit)));
						}
						// ...and all other bits of those columns are set to 0
						for (int bit = size; bit < maximumInterface; bit++) {
							tempBDD = tempBDD.and(createNoAdjacencyBDD(node, bit));
						}
					}

					// For all nodes which could possibly occur in the
					// interface...
					for (int node = 2; node < size; node++) {
						// ..check that the nodes before column k and row 2 are
						// switched...
						for (int bit = 0; bit < 2; bit++) {
							tempBDD = tempBDD.and(
									encoding.getPathBit(StateType.current, node, (bit + 1) % 2).biimp(
											encoding.getPathBit(StateType.successor, node, bit)));
						}
						// ...and all bits between k-th row and k-th column
						// remain the same...
						for (int bit = 2; bit < size; bit++) {
							if (bit == node) {
								tempBDD = tempBDD.and(createAdjacencyPreservedBDD(node, bit));
							} else {
								tempBDD = tempBDD.and(createAdjacencyUnchangedBDD(node, bit));
							}
						}
						// ...and all other bits of those columns are set to 0
						for (int bit = size; bit < maximumInterface; bit++) {
							tempBDD = tempBDD.and(createNoAdjacencyBDD(node,
									bit));
						}
					}

					// All other columns are set to 0
					for (int node = size; node < maximumInterface; node++) {
						for (int bit = 0; bit < maximumInterface; bit++) {
							tempBDD = tempBDD.and(createNoAdjacencyBDD(node,
									bit));
						}
					}
					
					//The first and second connection bit are switched and the rest is unchanged
					tempBDD = tempBDD.and(
							encoding.getSourcePathBit(StateType.current, 0).biimp(
									encoding.getSourcePathBit(StateType.successor, 1)));
					tempBDD = tempBDD.and(
							encoding.getSourcePathBit(StateType.current, 1).biimp(
									encoding.getSourcePathBit(StateType.successor, 0)));					
					for(int bit=2;bit<size;bit++){
						tempBDD = tempBDD.and(createConnectionUnchangedBDD(bit));
					}
					for(int bit=size;bit<maximumInterface;bit++){
						tempBDD = tempBDD.and(createNoConnectionBDD(bit));
					}
					
					transBDD = transBDD.or(tempBDD);
				}

			} else if (letter instanceof Vertex) {
				// The VERTEX-operation is applicable only if at least one node
				// could be added
				// to the interface, i.e. the interface size has to be at most
				// (maxSize-1)
				for (int size = 0; size < maximumInterface; size++) {
					// Check that the current interface size is i and that the
					// interface
					// size of the successor interface is i+1
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(
							encoding.getInterfaceSize(StateType.successor, size + 1));

					// For all nodes which could possibly occur in the
					// interface...
					for (int node = 0; node < size; node++) {
						// ..check that the nodes before column k and row k
						// remain the same and bits on the diagonal are always
						// 1...
						for (int bit = 0; bit < size; bit++) {
							if (bit == node) {
								tempBDD = tempBDD.and(createAdjacencyPreservedBDD(node, bit));
							} else {
								tempBDD = tempBDD.and(createAdjacencyUnchangedBDD(node, bit));
							}
						}
						// ...and all other adjacency bits of these columns are
						// set to 0
						for (int bit = size; bit < maximumInterface; bit++) {
							tempBDD = tempBDD.and(createNoAdjacencyBDD(node, bit));
						}
					}

					// The nodes after k have to be 0
					for (int node = size + 1; node < maximumInterface; node++) {
						for (int bit = 0; bit < maximumInterface; bit++) {
							tempBDD = tempBDD.and(createNoAdjacencyBDD(node, bit));
						}
					}

					// The node in column k has a 1 at the k-th position
					// afterwards and all other bits are set to 0
					for (int bit = 0; bit < maximumInterface; bit++) {
						if (bit == size) {
							tempBDD = tempBDD.and(
									encoding.getPathBit(StateType.current, size, bit).not().and(
											encoding.getPathBit(StateType.successor, size, bit)));
						} else {
							tempBDD = tempBDD.and(createNoAdjacencyBDD(size, bit));
						}
					}
					
					//The connection list bits remain the same
					for(int bit = 0; bit <size;bit++){
						tempBDD = tempBDD.and(createConnectionUnchangedBDD(bit));
					}
					for(int bit=size;bit<maximumInterface;bit++){
						tempBDD = tempBDD.and(createNoConnectionBDD(bit));
					}
					
					transBDD = transBDD.or(tempBDD);
				}
			}
			stateTransBDDs.put(letter, transBDD);
			fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED,
					letter.toString(), signature.size(), ++letterCounter));
		}

		// Log exiting
		logger.exiting(getClass().getName(), "createTransitions");
	}


	/*------------------------------------------------------------------------------------------*/
	/* 		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Constructs a BDD which is {@code true} if and only if the adjacency bits
	 * of the result are equal to the multiplication of the given
	 * {@code tempBDD} times {@code factorBDD} RavenBDD
	 * @param tempBDD the BDD to be multiplied
	 * @return a BDD encoding the result of the multiplication
	 */
	private RavenBDD createSquareBDD(RavenBDD tempBDD,int size) {
		RavenBDD multBDD = encoding.one();		
		RavenBDD[][] resultBDD = new RavenBDD[maximumInterface][maximumInterface];
		RavenBDD positionAnalyserBDD;	
		
		//Building the first result matrix
		for (int node = 0; node < maximumInterface; node++) {
			for (int bit = 0; bit < maximumInterface; bit++) {
				positionAnalyserBDD = encoding.existsPathBit(StateType.successor, tempBDD.clone(), node, bit);
				
				//If the analysed Position was a 1 or a 0 this bit must remain
				if(positionAnalyserBDD.clone().biimp(
						encoding.getPathBit(StateType.successor, node, bit)).isOne() ||
						positionAnalyserBDD.clone().biimp(
								encoding.getPathBit(StateType.successor, node, bit)).isZero()) {
					resultBDD[node][bit] = encoding.getPathBit(StateType.current, node, bit);
				}
				//All positions with undefined bits must be recalculated via multiplication
				else{
					RavenBDD helpBDD = encoding.zero();
					for (int z = 0; z < maximumInterface; z++) {
						helpBDD = helpBDD.or(encoding.getPathBit(
								StateType.current, z, bit).and(
								encoding.getPathBit(StateType.current, node,
										z)));
					}
					resultBDD[node][bit] = helpBDD.clone();
				}	
			}
		}


		//Building the formular matrix
		RavenBDD[][] oldResultBDD = resultBDD.clone();
		for(int iterate=0;iterate<maximumInterface;iterate++){

			for (int node = 0; node < size; node++) {
				for (int bit = 0; bit < size; bit++) {
					positionAnalyserBDD = encoding.existsPathBit(
							StateType.successor, tempBDD.clone(), node, bit);
					
					if(!positionAnalyserBDD.clone().biimp(
							encoding.getPathBit(StateType.successor, node, bit)).isOne() &&
							!positionAnalyserBDD.clone().biimp(
									encoding.getPathBit(StateType.successor, node, bit)).isZero()) {
					
						RavenBDD formulaBDD = encoding.zero();
						for (int z = 0; z < size; z++) {
							formulaBDD = formulaBDD.or(oldResultBDD[z][bit].clone().and(
									encoding.getPathBit(StateType.current, node, z)));
							
						}
						resultBDD[node][bit] = formulaBDD.clone();
					}		
				}
			}		
			oldResultBDD = resultBDD.clone();
		}
		
		//Building the result
		for (int node = 0; node < maximumInterface; node++) {
			for (int bit = 0; bit < maximumInterface; bit++) {
				multBDD = multBDD.and(
						encoding.getPathBit(StateType.successor, node, bit).biimp(
								resultBDD[node][bit].clone()));
			}
		}
		
		//For symmetry purposes
		for (int node = 0; node < maximumInterface; node++) {
			for (int bit = 0; bit < maximumInterface; bit++) {			
					multBDD.and(encoding.getPathBit(StateType.successor, node, bit).biimp(
							encoding.getPathBit(StateType.successor, bit, node)));
					multBDD.and(encoding.getPathBit(StateType.current, node, bit).biimp(
							encoding.getPathBit(StateType.current, bit, node)));
			}
		}
		
		multBDD = multBDD.and(tempBDD.clone().exists(sucStateDomain.getDomainBDD()));
		
		return multBDD;
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if the adjacency bit
	 * of the {@code node}-th and the {@code bit}-th interface node remains 0.
	 * @param node the index of the column which is compared
	 * @param bit the index of the target node which is compared
	 * @return a BDD encoding the information
	 */
	private RavenBDD createNoAdjacencyBDD(int node, int bit) {
		if (node < 0 || node >= maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}
		if (bit < 0 || bit >= maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}

		return encoding.getPathBit(StateType.current, node, bit).not().and(
				encoding.getPathBit(StateType.successor, node, bit).not());
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if the adjacency bit
	 * of the {@code node}-th and the {@code bit}-th interface node remains the
	 * same.
	 * @param node the index of the column which is compared
	 * @param bit the index of the target node which is compared
	 * @return a BDD encoding the information
	 */
	private RavenBDD createAdjacencyUnchangedBDD(int node, int bit) {
		if (node < 0 || node >= maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}
		if (bit < 0 || bit >= maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}

		return encoding.getPathBit(StateType.current, node, bit).biimp(
				encoding.getPathBit(StateType.successor, node, bit));
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if the adjacency bit
	 * of the {@code node}-th and the {@code bit}-th interface node remains 1.
	 * @param node the index of the column which is compared
	 * @param bit the index of the target node which is compared
	 * @return a BDD encoding the information
	 */
	private RavenBDD createAdjacencyPreservedBDD(int node, int bit) {
		if (node < 0 || node >= maximumInterface) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}
		if (bit < 0 || bit >= maximumInterface) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}

		return encoding.getPathBit(StateType.current, node, bit).and(
				encoding.getPathBit(StateType.successor, node, bit));
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if the connection bit
	 * of the the {@code bit}-th interface connection node remains the
	 * same.
	 * @param bit the index of the connection bit which is compared
	 * @return a BDD encoding the information
	 */
	private RavenBDD createConnectionUnchangedBDD(int bit) {
		return encoding.getSourcePathBit(StateType.current, bit).biimp(
				encoding.getSourcePathBit(StateType.successor, bit));
	}
	
	/**
	 * Constructs a BDD which is {@code true} if and only if the connection bit
	 * of the the {@code bit}-th interface connection node remains 0.
	 * @param bit the index of the connection bit which is compared
	 * @return a BDD encoding the information
	 */
	private RavenBDD createNoConnectionBDD(int bit) {
		return encoding.getSourcePathBit(StateType.current, bit).not().and(
				encoding.getSourcePathBit(StateType.successor, bit).not());
	}
}