/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.data.automata;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.data.Edge;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.Node;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.Operation.Connect;
import de.uni_due.inf.ti.raven.data.Operation.Fuse;
import de.uni_due.inf.ti.raven.data.Operation.Perm;
import de.uni_due.inf.ti.raven.data.Operation.Res;
import de.uni_due.inf.ti.raven.data.Operation.Trans;
import de.uni_due.inf.ti.raven.data.Operation.Vertex;
import de.uni_due.inf.ti.raven.data.Signature;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.encodings.BDDEncoding.StateType;
import de.uni_due.inf.ti.raven.encodings.SubgraphEncoding;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * 
 * This class implements a graph automaton accepting the language of all graphs which contain a 
 * specific graph as subgraph.
 * 
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.55 $
 * 
 * @see SubgraphEncoding
 */
public class SubgraphAutomaton extends Automaton {

	private static final long serialVersionUID = 8464565241210344185L;

	transient private SubgraphEncoding encoding;
	private boolean sloppyModeEnabled;
	private Graph subgraph;
	private List<Node> nodeOrder;
	private List<Edge> edgeOrder;
	private int nodeCount;
	private int edgeCount;
	transient private Logger logger;

	/*------------------------------------------------------------------------------------------*/
	/* 		Constructors 																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new subgraph automaton. To get a more efficient automaton, i.e. to minimize the 
	 * size of the Transition-BDDs, it is possible to disable some consistency checks of the 
	 * transitions. These checks ensure that ...
	 * <ul>
	 * <li>the interface nodes for interfaces greater than the current interface are set to 
	 * {@code false}</li>
	 * <li>every (graph) node with is in the interface has been recognized, i.e. the node bit has 
	 * been set to {@code true}</li>
	 * <li>the interface has only be changed at the position for the current interface size</li>
	 * <li>for every (recognized) edge all incident nodes have also been recognized</li>
	 * </ul>
	 * By disabling the consistency checks additional "invalid" states, i.e. a state which does not 
	 * satisfy the above conditions, will be introduced to minimize the size of the BDDs. But even 
	 * if the consistency checks are disabled, it is guaranteed that starting in any "valid" state 
	 * only "valid" states are reachable.
	 * @param encoding the encoding for the states
	 * @param signature the signature of this automaton
	 * @param sloppyModeEnabled if {@code true} some consistency checks will be disabled
	 */
	protected SubgraphAutomaton(SubgraphEncoding encoding, Signature signature,
			boolean sloppyModeEnabled) {
		super(Type.subgraph, encoding, signature);
		this.encoding = encoding;
		subgraph = encoding.getSubgraph();
		nodeOrder = new ArrayList<>(encoding.getNodeOrder());
		edgeOrder = new ArrayList<>(encoding.getEdgeOrder());
		nodeCount = subgraph.getNodeCount();
		edgeCount = subgraph.getEdgeCount();
		this.sloppyModeEnabled = sloppyModeEnabled;

		// Register logger
		logger = Logger.getLogger(getClass().getName());
	}
	
	private synchronized void writeObject(ObjectOutputStream s) throws IOException {
		s.writeObject(subgraph);
		s.writeObject(nodeOrder);
		s.writeObject(edgeOrder);
		s.writeObject(nodeCount);
		s.writeObject(edgeCount);
		s.writeObject(sloppyModeEnabled);
		
//		s.defaultWriteObject();
	}
	
	@SuppressWarnings("unchecked")
	private synchronized void readObject(ObjectInputStream s) throws ClassNotFoundException, 
	IOException {
		subgraph = (Graph) s.readObject();
		nodeOrder = (List<Node>) s.readObject();
		edgeOrder = (List<Edge>) s.readObject();
		nodeCount = (int) s.readObject();
		edgeCount = (int) s.readObject();
		sloppyModeEnabled = (boolean) s.readObject();
	}

	/*------------------------------------------------------------------------------------------*/
	/* 		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public Automaton clone() throws CloneNotSupportedException {
		Automaton clone = new SubgraphAutomaton(encoding, signature,
				sloppyModeEnabled);
		clone.initialStates = initialStates.clone();
		clone.finalStates = finalStates.clone();
		clone.nonFinalStates = nonFinalStates.clone();
		clone.states = states.clone();

		for (Operation operation : stateTransBDDs.keySet()) {
			clone.stateTransBDDs.put(operation, stateTransBDDs.get(operation).clone());
		}

		return clone;
	}
	
	/**
	 * Returns the graph which is used as a subgraph for this automaton.
	 * @return the wanted subgraph
	 */
	public Graph getSubgraph() {
		return subgraph;
	}
	
	/**
	 * Returns the node-order which is used for this automaton.
	 * @return the node-order
	 */
	public List<Node> getNodeOrder() {
		return nodeOrder;
	}

	/**
	 * Returns the edge-order which is used for this automaton.
	 * @return the edge-order
	 */
	public List<Edge> getEdgeOrder() {
		return edgeOrder;
	}

	/**
	 * Returns {@code true} if and only if in the state {@code state} the node with name 
	 * {@code nodeName} is in the interface at position {@code size}.
	 * @param state the encoding of the state
	 * @param nodeName the name of the node
	 * @param size the current interface size
	 * @return {@code true}, if the node is at the given position of the interface, {@code false} 
	 * otherwise
	 */
	public boolean isNodeInInterface(RavenBDD state, String nodeName, int size) {
		if (state == null) {
			throw new NullPointerException(RavenExceptionMessages.BDD_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if (state.isZero() || state.isOne()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.BDD_MUST_NEITHER_BE_ZERO_NOR_ONE_EXCEPTION);
		}
		if (nodeName == null) {
			throw new NullPointerException(RavenExceptionMessages.NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if (nodeName.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		} 
		if (!subgraph.hasNode(nodeName)) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.NODE_NOT_EXISTING_EXCEPTION);
		}
		if (size <= 0 || size > maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}

		int index = encoding.getNodeOrder().indexOf(subgraph.getNode(nodeName));
		RavenBDD tempBDD = state.restrict(encoding.getMappingBit(StateType.current, index, size - 1));
		tempBDD = tempBDD.exists(curStateDomain.getDomainBDD());

		return tempBDD.isOne() && hasRecognizedNode(state, nodeName);
	}

	/**
	 * Returns {@code true} if and only if in the state {@code state} the edge
	 * with name {@code edgeName} has been recognized.
	 * 
	 * @param state the encoding of the state
	 * @param edgeName the name of the edge
	 * @return {@code true}, if the edge has already been recognized in the
	 *         state, {@code false} otherwise
	 */
	public boolean hasRecognizedEdge(RavenBDD state, String edgeName) {
		if (state == null) {
			throw new NullPointerException(RavenExceptionMessages.BDD_MUST_NOT_BE_NULL_EXCEPTION);
		} 
		if (state.isZero() || state.isOne()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.BDD_MUST_NEITHER_BE_ZERO_NOR_ONE_EXCEPTION);
		} 
		if (edgeName == null) {
			throw new NullPointerException(RavenExceptionMessages.NAME_MUST_NOT_BE_NULL_EXCEPTION);
		} 
		if (!subgraph.hasEdge(edgeName)) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.EDGE_NOT_EXISTING_EXCEPTION);
		}

		int index = encoding.getEdgeOrder().indexOf(subgraph.getEdge(edgeName));
		RavenBDD tempBDD = state.restrict(encoding.getEdgeBit(StateType.current, index));
		tempBDD = tempBDD.exists(encoding.getCurrentStateDomain().getDomainBDD());

		return tempBDD.isOne();
	}
	
	@Override
	public void initializeEncoding() {
		encoding = SubgraphEncoding.getEncoding(maximumInterface, 0, 0, subgraph, 
				nodeOrder, edgeOrder, encodingName);
		super.encoding = encoding;
		super.initializeEncoding();
	}
	
	/*------------------------------------------------------------------------------------------*/
	/* 		Protected Methods 																	*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	protected void createStates() {
		// Log entering
		logger.entering(getClass().getName(), "createStates");

		//Compute a BDD representing the set of all states
		states = encoding.zero();
		for (int size = 0; size <= maximumInterface; size++) {
			RavenBDD tempBDD = encoding.getInterfaceSize(StateType.current,
					size);

			for (int node = 0; node < nodeCount; node++) {
				for (int i = size; i < maximumInterface; i++) {
					tempBDD = tempBDD.and(encoding.getMappingBit(
							StateType.current, node, i).not());
				}
			}
			states = states.or(tempBDD);
		}

		// Log exiting
		logger.exiting(getClass().getName(), "createStates");
	}
	
	@Override
	protected void createInitialAndFinalStates() {
		// Log entering
		logger.entering(getClass().getName(), "createInitialAndFinalStates");

		initialStates = encoding.getInterfaceSize(StateType.current, innerInterface);
		finalStates = encoding.getInterfaceSize(StateType.current, outerInterface);
 
		//Calculate the conditions for the inner Interface and append them to the initialStatesBDD
		initialStates = initialStates.and(createSubgraphInitialStatesConditionBDD(innerInterface));
		
		//In the finalStates each node was recognized
		for (int node = 0; node < nodeCount; node++) {
			finalStates = finalStates.and(encoding.getNodeBit(StateType.current, node));			
		}
		
		//Calculate the conditions for the outer Interface and append them to the finalStatesBDD
		finalStates = finalStates.and(createSubgraphFinalStatesConditionBDD(outerInterface,sloppyModeEnabled));

		// Apply an edge bit for each edge
		for (int edge = 0; edge < edgeCount; edge++) {
			initialStates = initialStates.and(encoding.getEdgeBit(StateType.current, edge).not());
			finalStates = finalStates.and(encoding.getEdgeBit(StateType.current, edge));
		}

		// Compute set of non-final states
		nonFinalStates = states.clone().and(finalStates.clone().not());

		// Log exiting
		logger.exiting(getClass().getName(), "createInitialAndFinalStates");
	}


	@Override
	protected void createTransitions() {
		// Log entering
		logger.entering(getClass().getName(), "createTransitions");

		// Iteriere ueber alle Graphoperationen
		for (Operation operation : getAlphabet()) {
			RavenBDD transBDD = stateTransBDDs.get(operation);
			RavenBDD tempBDD;

			// Falls die Operation die das Verkleben mit einer Hyperkante ist
			if (operation instanceof Connect) {
				//Get the properties of the current edge-operation
				int arity = ((Connect) operation).getArity();
				String label = ((Connect) operation).getLabel();

				//For every interface size for which the operation is applicable
				for (int size = arity; size <= maximumInterface; size++) {
					//The interface size must not be changed
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size));

					tempBDD = tempBDD.and(addEdge(size, label, arity));

					//The nodes must not be changed
					for (int node = 0; node < nodeCount; node++) {
						tempBDD = tempBDD.and(nodeNotChanged(node, size));
					}

					//Add the current transition
					transBDD = transBDD.or(tempBDD);

					// Falls der Debug-Modus aktiviert wurde
					if (!sloppyModeEnabled) {
						// wird ein Konsistenz-Check durchgefuehrt
						transBDD = transBDD.and(getConsistencyCheck(size));
					}
				}
			} else if (operation instanceof Fuse) {
				//For every interface size for which the operation is applicable
				for (int size = 2; size <= maximumInterface; size++) {
					// BDD vorbereiten
					tempBDD = encoding.zero();

					// Durchlaufe alle Knoten
					for (int node = 0; node < nodeCount; node++) {
						// und konstruiere einen BDD, der ueberprueft, der
						// aktuelle Knoten ge-fuse-t wird
						tempBDD = tempBDD.or(fuseNodes(node, size));
					}

					//The interface size must be decreased by 1
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.current, size));
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size - 1));

					//The edges must not be changed
					for (int edge = 0; edge < edgeCount; edge++) {
						tempBDD = tempBDD.and(edgeNotChanged(edge));
					}

					tempBDD = tempBDD.and(getInterfaceUndefined(size));

					//Add the current transition
					transBDD = transBDD.or(tempBDD);

					if (!sloppyModeEnabled) {
						// wird ein Konsistenz-Check durchgefuehrt
						transBDD = transBDD.and(getConsistencyCheck(size));
					}
				}
			} else if (operation instanceof Perm) {
				//For every interface size for which the operation is applicable
				for (int size = 3; size <= maximumInterface; size++) {
					//The interface size must not be changed
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size));

					//The edges must not be changed
					for (int edge = 0; edge < edgeCount; edge++) {
						tempBDD = tempBDD.and(edgeNotChanged(edge));
					}
					// Durchlaufe alle Knoten
					for (int node = 0; node < nodeCount; node++) {
						// und konstruiere einen BDD, der ueberprueft, dass nur
						// die ersten beiden Knoten getauscht werden
						tempBDD = tempBDD.and(permuteNodes(node, size));
					}

					tempBDD = tempBDD.and(getInterfaceUndefined(size));
					//Add the current transition
					transBDD = transBDD.or(tempBDD);

					if (!sloppyModeEnabled) {
						// wird ein Konsistenz-Check durchgefuehrt
//						transBDD = transBDD.and(getConsistencyCheck(size));
						RavenBDD checkBDD = constructBDDForInterfaceSizeCheck(size);
						checkBDD = checkBDD.and(constructBDDForVertexExistenceCheck());
						checkBDD = checkBDD.and(constructBDDForEdgeCheck());
						transBDD = transBDD.and(checkBDD);
					}
				}
			} else if (operation instanceof Res) {
				//For every interface size for which the operation is applicable
				for (int size = 1; size <= maximumInterface; size++) {
					//The interface size must be decreased by 1
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size - 1));

					//The edges must not be changed
					for (int edge = 0; edge < edgeCount; edge++) {
						tempBDD = tempBDD.and(edgeNotChanged(edge));
					}

					// Durchlaufe alle Knoten
					for (int node = 0; node < nodeCount; node++) {
						// und konstruiere einen BDD, der ueberprueft, dass nur
						// die ersten beiden Knoten getauscht werden
						tempBDD = tempBDD.and(restrictNode(node, size));
					}

					tempBDD = tempBDD.and(getInterfaceUndefined(size));
					//Add the current transition
					transBDD = transBDD.or(tempBDD);

					if (!sloppyModeEnabled) {
						// wird ein Konsistenz-Check durchgefuehrt
						transBDD = transBDD.and(getConsistencyCheck(size));
					}
				}
			} else if (operation instanceof Trans) {
				//For every interface size for which the operation is applicable
				for (int size = 2; size <= maximumInterface; size++) {
					//The interface size must not be changed
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size));

					//The edges must not be changed
					for (int edge = 0; edge < edgeCount; edge++) {
						tempBDD = tempBDD.and(edgeNotChanged(edge));
					}

					// Durchlaufe alle Knoten
					for (int node = 0; node < nodeCount; node++) {
						// und konstruiere einen BDD, der ueberprueft, dass nur
						// die ersten beiden Knoten getauscht werden
						tempBDD = tempBDD.and(transposeNodes(node, size));
					}

					tempBDD = tempBDD.and(getInterfaceUndefined(size));
					//Add the current transition
					transBDD = transBDD.or(tempBDD);

					if (!sloppyModeEnabled) {
						// wird ein Konsistenz-Check durchgefuehrt
						RavenBDD checkBDD = constructBDDForInterfaceSizeCheck(size);
						checkBDD = checkBDD.and(constructBDDForVertexExistenceCheck());
						checkBDD = checkBDD.and(constructBDDForEdgeCheck());
						transBDD = transBDD.and(checkBDD);
					}
				}
			}
			// Transitionsfunktion fuer die VERTEX-Operation
			else if (operation instanceof Vertex) {
				//For every interface size for which the operation is applicable
				for (int size = 0; size < maximumInterface; size++) {
					//The interface size must be increased by 1
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size + 1));

					tempBDD = tempBDD.and(addNode(size));

					//The edges must not be changed
					for (int edge = 0; edge < edgeCount; edge++) {
						tempBDD = tempBDD.and(edgeNotChanged(edge));
					}

					tempBDD = tempBDD.and(getInterfaceUndefined(size + 1));
					//Add the current transition
					transBDD = transBDD.or(tempBDD);

					// Falls der Debug-Modus aktiviert wurde
					if (!sloppyModeEnabled) {
						// wird ein Konsistenz-Check durchgefuehrt
						tempBDD = constructBDDForInterfaceSizeCheck(size + 1);
						transBDD = transBDD.and(tempBDD);

						tempBDD = constructBDDForVertexExistenceCheck();
						transBDD = transBDD.and(tempBDD);

						if(size>0) {
							tempBDD = constructBDDForInterfaceUniquenessCheck(size);
							transBDD = transBDD.and(tempBDD);
						}

						tempBDD = constructBDDForEdgeCheck();
						transBDD = transBDD.and(tempBDD);
					}
				}
			}
			stateTransBDDs.put(operation, transBDD);
		}

		// Log exiting
		logger.exiting(getClass().getName(), "createTransitions");
	}

	/*------------------------------------------------------------------------------------------*/
	/* 		Private Methods 																	*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns {@code true} if and only if in the state {@code state} the name
	 * with name {@code nameName} has been recognized.
	 * @param state the encoding of the state
	 * @param nodeName the name of the node
	 * @return {@code true}, if the node has already been recognized in the state, {@code false} 
	 * otherwise
	 */
	private boolean hasRecognizedNode(RavenBDD state, String nodeName) {
		if (state == null) {
			throw new NullPointerException(RavenExceptionMessages.BDD_MUST_NOT_BE_NULL_EXCEPTION);
		} 
		if (state.isZero() || state.isOne()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.BDD_MUST_NEITHER_BE_ZERO_NOR_ONE_EXCEPTION);
		}
		if (nodeName == null) {
			throw new NullPointerException(RavenExceptionMessages.NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if (nodeName.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		} 
		if (!subgraph.hasNode(nodeName)) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.NODE_NOT_EXISTING_EXCEPTION);
		}

		int index = encoding.getNodeOrder().indexOf(subgraph.getNode(nodeName));
		RavenBDD tempBDD = state.clone().restrict(
				encoding.getNodeBit(StateType.current, index));
		tempBDD = tempBDD.exists(curStateDomain.getDomainBDD());

		return tempBDD.isOne();
	}
	
	/**
	 * Constructs a BDD which is {@code true} if and only if the transition
	 * function of the automaton for the current interface is consistent., i.e.
	 * the following conditions are satisfied:
	 * 
	 * <ul>
	 * <li>the interface nodes for interfaces greater than the current interface
	 * are set to {@code false}</li>
	 * <li>every (graph) node with is in the interface has been recognized, i.e.
	 * the node bit has been set to {@code true}</li>
	 * <li>the interface has only be changed at the position for the current
	 * interface size</li>
	 * <li>for every (recognized) edge all incident nodes have also been
	 * recognized</li>
	 * </ul>
	 * 
	 * @param size the size of the current interface
	 * @return a BDD encoding the consistency check
	 */
	private RavenBDD getConsistencyCheck(int size) {
		if (size < 0 || size > maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}
		
		RavenBDD checkBDD = constructBDDForInterfaceSizeCheck(size);
		checkBDD = checkBDD.and(constructBDDForVertexExistenceCheck());
		if(size>0) {
			checkBDD = checkBDD.and(constructBDDForInterfaceUniquenessCheck(size));
		}
		return checkBDD.and(constructBDDForEdgeCheck());
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if in both the current
	 * state and the successor state no node is in the interface of size
	 * {@code size} or greater.
	 * 
	 * @param size the size of the current interface
	 * @return a BDD encoding that no node is in the interface
	 */
	private RavenBDD getInterfaceUndefined(int size) {
		RavenBDD resultBDD = encoding.one();

		for (int node = 0; node < nodeCount; node++) {
			for (int i = size; i < maximumInterface; i++) {
				resultBDD = resultBDD.and(
						encoding.getMappingBit(StateType.current, node, i).not().and(
						encoding.getMappingBit(StateType.successor, node, i).not()));
			}
		}

		return resultBDD;
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if a new edge of arity
	 * {@code arity} is added and
	 * <ul>
	 * <li>if the edge is not part of the subgraph, no (recognized) part of the
	 * subgraph has been changed</li>
	 * <li>if the edge is part of the subgraph, the edge has been added
	 * correctly, i.e. the recognized parts of the subgraph have been updated</li>
	 * </ul>
	 * 
	 * @param size the size of the current interface
	 * @param label the label of the new edge
	 * @param arity the arity of the new edge
	 * @return a BDD encoding that the new edge has been added correctly
	 */
	private RavenBDD addEdge(int size, String label, int arity) {
		if (size < 0 || size > maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}
		if (label == null) {
			throw new IllegalArgumentException(
					"The edge label must not be null.");
		}
		if (arity < 0 || size < arity) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_ARITY_EXCEPTION);
		}
		RavenBDD resultBDD = encoding.one();

		// 1st case - Edge is not part of the subgraph: Build BDD to check that
		// no edge has been changed
		for (int edge = 0; edge < edgeCount; edge++) {
			resultBDD = resultBDD.and(edgeNotChanged(edge));
		}

		// 2nd case - Edge is part of the subgraph: Build BDD to check that the
		// edge is added correctly
		for (int edge = 0; edge < edgeCount; edge++) {
			Edge e = edgeOrder.get(edge);
			if (e.getArity() == arity && e.getLabel().equals(label)) {
				resultBDD = resultBDD.or(addNewEdge(edge, size));
			}
		}

		return resultBDD;
	}

	/**
	 * Konstruiert einen BDD, der genau dann wahr liefert, wenn die Kante
	 * {@code edgeName} entsprechend der Bedingungen der EDGE-Operation
	 * hinzugefuegt wurde. Dabei wird davon ausgegangen, dass alle adjazenten
	 * Knoten der Kante bereits ueber VERTEX-Operationen hinzugefuegt wurden
	 * 
	 * @param edge die neu hinzuzufuegende Kante
	 * @param size die aktuelle Interfacegroesse
	 * @return der zu konstruierende BDD
	 */
	private RavenBDD addNewEdge(int edge, int size) {
		if (edge < 0 || edge >= edgeCount) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_EDGE_NUMBER_EXCEPTION);
		}
		if (size < 0 || size > maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}

		// BDDs vorbereiten
		RavenBDD resultBDD = encoding.one();
		// Durchlaufe alle Kanten
		for (int anotherEdge = 0; anotherEdge < edgeCount; anotherEdge++) {
			// Falls die Kante nicht die neu hinzugefuegte Kante ist
			if (anotherEdge != edge) {
				// BDD konstruieren, der ueberprueft, dass sich das
				// (Nicht-)Vorhanden sein der Kante nicht aendert
				resultBDD = resultBDD.and(encoding.getEdgeBit(StateType.current, anotherEdge).biimp(
						encoding.getEdgeBit(StateType.successor, anotherEdge)));
			}
			// ansonsten handelt es sich um die neue Kante
			else {
				// BDD konstruieren, der ueberprueft, dass die neue Kante
				// wirklich
				// hinzugefuegt wird
				resultBDD = resultBDD.and(encoding.getEdgeBit(StateType.current, edge).not().and(
						encoding.getEdgeBit(StateType.successor, edge)));
			}
		}

		Edge e = edgeOrder.get(edge); // Kanten-Objekt
		int arity = e.getArity();
		// Durchlaufe alle Kanten
		for (int node = 0; node < nodeCount; node++) {
			Node n = nodeOrder.get(node);
			RavenBDD tempBDD;
			// Falls der Knoten adjazent zur Kante ist
			if (e.getIncidentNodes().contains(n)) {
				// muss der Knoten auch existieren
				tempBDD = encoding.getNodeBit(StateType.current, node).and(
						encoding.getNodeBit(StateType.successor, node));
			}
			// falls nicht
			else {
				// darf sich der Knoten nicht aendern
				tempBDD = encoding.getNodeBit(StateType.current, node).biimp(
						encoding.getNodeBit(StateType.successor, node));
			}

			// Durchlaufe das gesamte Interface
			for (int i = 0; i < size; i++) {
				// Im aktuellen Interface
				if (i < size - arity) {
					// darf sich nichts aendern
					tempBDD = tempBDD.and(encoding.getMappingBit(StateType.current, node, i).biimp(
							encoding.getMappingBit(StateType.successor, node, i)));
				}
				// im erweiterten Interface
				else if (i < size) {
					// falls der aktuelle Knoten an der (i-ifaceSize-1)-ten
					// Position der Kante vorkommt
					if (e.getIncidentNodes().get(i - (size - arity)).equals(n)) {
						// muss der Knoten auch an der i-ten Position im
						// Interface vorhanden sein
						tempBDD = tempBDD.and(encoding.getMappingBit(StateType.current, node, i).and(
								encoding.getMappingBit(StateType.successor, node, i)));
					}
					// falls nicht
					else {
						// darf sich das Interface an der i-ten Stelle nicht
						// aendern
						tempBDD = tempBDD.and(encoding.getMappingBit(
								StateType.current, node, i).biimp(
								encoding.getMappingBit(StateType.successor,
										node, i)));
					}
				}
			}
			resultBDD = resultBDD.and(tempBDD);
		}

		return resultBDD;
	}

	/**
	 * Konstruiert einen BDD, der genau dann wahr liefert, wenn der Knoten
	 * {@code vertexName} entsprechend der Bedingungen der FUSE-Operation
	 * verscholzen wurde.
	 * 
	 * @param node der zu fusende Knoten
	 * @param size die aktuelle Interfacegroesse
	 * @return der zu konstruierende BDD
	 */
	private RavenBDD fuseNodes(int node, int size) {
		if (node < 0 || node >= nodeCount) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}
		if (size < 2 || size > maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}

		// Ueberpruefe, dass kein Knoten hinzugefuegt bzw. geloescht wurde
		// Dies ist genau dann der Fall, wenn die Knotenbits gilt: v <-> v',
		// wobei
		// v das Bit fuer den "Vorher"-Knoten und v' das Bit fuer den
		// "Nachher"-Knoten bezeichnet
		RavenBDD resultBDD = encoding.getNodeBit(StateType.current, node)
				.biimp(encoding.getNodeBit(StateType.successor, node));

		for (int i = 0; i < size; i++) {
			if (i < size - 1) {
				// Diese Knoten duerfen sich durch die FUSE-Operation nicht
				// aendern
				resultBDD = resultBDD.and(encoding.getMappingBit(
						StateType.current, node, i).biimp(
						encoding.getMappingBit(StateType.successor, node, i)));
			} else if (i == size - 1) {
				// Diese Knoten duerfen sich durch die FUSE-Operation nicht
				// aendern
				resultBDD = resultBDD.and(encoding.getMappingBit(
						StateType.current, node, i - 1).biimp(
						encoding.getMappingBit(StateType.current, node, i)));
				resultBDD = resultBDD.and(encoding.getMappingBit(
						StateType.successor, node, i).not());
			}
		}

		// Iteriere ueber alle Knoten, die nicht ge-fuse-t werden sollen
		for (int anotherNode = 0; anotherNode < nodeCount; anotherNode++) {
			// Ueberpruefe, dass es wirklich ein Knoten ist, der NICHT
			// ge-fuse-t werden soll
			if (anotherNode != node) {
				// Ueberpruefe, dass kein Knoten hinzugefuegt bzw. geloescht
				// wurde
				resultBDD = resultBDD.and(encoding.getNodeBit(
						StateType.current, anotherNode).biimp(
						encoding.getNodeBit(StateType.successor, anotherNode)));

				for (int i = 0; i < size; i++) {
					if (i < size - 2) {
						// Diese Knoten duerfen sich durch die FUSE-Operation
						// nicht aendern
						resultBDD = resultBDD.and(encoding.getMappingBit(
								StateType.current, anotherNode, i).biimp(
								encoding.getMappingBit(StateType.successor,
										anotherNode, i)));
					} else {
						// Da durch die FUSE-Operation die "letzten beiden"
						// Knoten des Interfaces
						// ge-fuse-t werden, duerfen hier keine anderen Knoten
						// vorhanden sein
						// Ausserdem duerfen auch an Positionen ausserhalb des
						// aktuellen Interfaces
						// keine Knoten vorhanden sein
						resultBDD = resultBDD.and(encoding
								.getMappingBit(StateType.current, anotherNode,
										i)
								.not()
								.and(encoding.getMappingBit(
										StateType.successor, anotherNode, i)
										.not()));
					}
				}

			}
		}
		return resultBDD;
	}

	/**
	 * Konstruiert einen BDD, der genau dann wahr liefert, wenn der Knoten
	 * {@code vertexName} entsprechend der Bedingungen der PERMUTE-Operation
	 * permutiert wurde.
	 * 
	 * @param node der zu permutierende Knoten
	 * @param size die aktuelle Interfacegroesse
	 * @return der zu konstruierende BDD
	 */
	private RavenBDD permuteNodes(int node, int size) {
		if (node < 0 || node >= nodeCount) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}
		if (size < 3 || size > maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}

		// Ueberpruefe, dass kein Knoten hinzugefuegt bzw. geloescht wurde
		// Dies ist genau dann der Fall, wenn die Knotenbits gilt: v <-> v',
		// wobei
		// v das Bit fuer den "Vorher"-Knoten und v' das Bit fuer den
		// "Nachher"-Knoten bezeichnet
		RavenBDD resultBDD = encoding.getNodeBit(StateType.current, node).biimp(
				encoding.getNodeBit(StateType.successor, node));

		for (int i = 0; i < size; i++) {
			//Der Wert des Bits, das den Interfaceknoten (nach der PERM-Operation) 
			//repraesentiert, haengt von dem Wert des Bits ab, das den Vorgaenger-Interfaceknoten
			//(vor der PERM-Operation) repraesentiert
			resultBDD = resultBDD.and(
					encoding.getMappingBit(StateType.current, node, i).biimp(
							encoding.getMappingBit(StateType.successor, node, (i + 1) % size)));
		}
		return resultBDD;
	}

	/**
	 * Konstruiert einen BDD, der genau dann wahr liefert, wenn der Knoten
	 * {@code vertexName} entsprechend der Bedingungen der RES-Operation aus dem
	 * Interface entfernt wurde.
	 * 
	 * @param node der zu entfernende Knoten
	 * @param size die aktuelle Interfacegroesse
	 * @return der zu konstruierende BDD
	 */
	private RavenBDD restrictNode(int node, int size) {
		if (node < 0 || node >= nodeCount) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}
		if (size < 1 || size > maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}

		// Ueberpruefe, dass kein Knoten hinzugefuegt bzw. geloescht wurde
		// Dies ist genau dann der Fall, wenn die Knotenbits gilt: v <-> v',
		// wobei
		// v das Bit fuer den "Vorher"-Knoten und v' das Bit fuer den
		// "Nachher"-Knoten bezeichnet
		RavenBDD resultBDD = encoding.getNodeBit(StateType.current, node).biimp(
				encoding.getNodeBit(StateType.successor, node));

		// Zwei Faelle sind zu unterscheiden:
		// 1. Fall: Dieser Knoten soll verschattet werden. Dann darf der
		// Knoten nur noch an der Position ifaceSize
		// im Interface vorhanden sein und sonst nicht mehr
		for (int i = 0; i < size; i++) {
			if (i < size - 1) {
				// Diese Knoten duerfen nicht im Interface sein und auch
				// nicht durch die RES-Operation geaendert werden
				resultBDD = resultBDD.and(encoding.getMappingBit(StateType.current, node, i).not().and(
						encoding.getMappingBit(StateType.successor, node, i).not()));
			} else {
				// Dieser Knoten wird durch die RES-Operation aus dem
				// Interface entfernt
				resultBDD = resultBDD.and(encoding.getMappingBit(StateType.current, node, i).and(
						encoding.getMappingBit(StateType.successor, node, i).not()));
			}
		}

		// 2. Fall: Dieser Knoten soll nicht verschattet werden. Dann darf
		// sich das Interface nicht aendern, aber der Knoten darf
		// auch nicht an der letzten Position im Interface vorhanden sein
		RavenBDD tempBDD = encoding.getNodeBit(StateType.current, node).biimp(
				encoding.getNodeBit(StateType.successor, node));
		for (int i = 0; i < size; i++) {
			if (i < size - 1) {
				// Diese Knoten duerfen sich durch die RES-Operation nicht
				// aendern
				tempBDD = tempBDD.and(encoding.getMappingBit(StateType.current, node, i).biimp(
						encoding.getMappingBit(StateType.successor, node, i)));
			} else {
				// Da dieser Knoten nicht entfernt werden soll, darf er
				// auch nicht an letzen Position des Interfaces stehen
				tempBDD = tempBDD.and(encoding.getMappingBit(StateType.current, node, i).not().and(
						encoding.getMappingBit(StateType.successor, node, i).not()));
			}
		}
		return resultBDD.or(tempBDD);
	}

	/**
	 * Konstruiert einen BDD, der genau dann wahr liefert, wenn der Knoten
	 * {@code vertexName} entsprechend der Bedingungen der TRANS-Operation
	 * transponiert wird.
	 * 
	 * @param node der zu transponierende Knoten
	 * @param size die aktuelle Interfacegroesse
	 * @return der zu konstruierende BDD
	 */
	private RavenBDD transposeNodes(int node, int size) {
		if (node < 0 || node >= nodeCount) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}
		if (size < 2 || size > maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}

		// Ueberpruefe, dass kein Knoten hinzugefuegt bzw. geloescht wurde
		// Dies ist genau dann der Fall, wenn die Knotenbits gilt: v <-> v',
		// wobei
		// v das Bit fuer den "Vorher"-Knoten und v' das Bit fuer den
		// "Nachher"-Knoten bezeichnet
		RavenBDD resultBDD = encoding.getNodeBit(StateType.current, node).biimp(
				encoding.getNodeBit(StateType.successor, node));

		for (int i = 0; i < size; i++) {
			if (i == 0 || i == 1) {
				// Der Wert des Bits, das den ersten Interfaceknoten (nach der
				// TRANS-Operation) repraesentiert, haengt von
				// dem Wert des Bits ab, das den zweiten Interfaceknoten (vor
				// der TRANS-Operation) repraesentiert
				resultBDD = resultBDD.and(encoding.getMappingBit(StateType.current, node, i).biimp(
						encoding.getMappingBit(StateType.successor, node, (i + 1) % 2)));
			} else {
				// Diese Knoten duerfen sich durch die TRANS-Operation nicht
				// aendern
				resultBDD = resultBDD.and(encoding.getMappingBit(StateType.current, node, i).biimp(
						encoding.getMappingBit(StateType.successor, node, i)));
			}
		}

		return resultBDD;
	}

	/**
	 * Konstruiert einen BDD, der genau dann wahr liefert, wenn sich die
	 * vorgegebene Kante des Graphen nicht aendert.
	 * 
	 * @param edge der Name der zu ueberpruefenden Kante
	 * @return der zu konstruierende BDD
	 */
	private RavenBDD edgeNotChanged(int edge) {
		if (edge < 0 || edge >= edgeCount) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_EDGE_NUMBER_EXCEPTION);
		}

		return encoding.getEdgeBit(StateType.current, edge).biimp(
				encoding.getEdgeBit(StateType.successor, edge));
	}

	/**
	 * Konstruiert einen BDD, der genau dann wahr liefert, wenn ein Knoten
	 * entsprechend der Bedingungen der VERTEX-Operation hinzugefuegt wird.
	 * 
	 * @param size die aktuelle Interfacegroesse
	 * @return der zu konstruierende BDD
	 */
	private RavenBDD addNode(int size) {
		if (size < 0 || size >= maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}

		// 1. FALL: Es wird ein Knoten hinzugefuegt, der nicht zu dem gesuchten
		// Subgraphen gehoert
		// Durchlaufe alle Knoten
		RavenBDD resultBDD = encoding.one();
		for (int node = 0; node < nodeCount; node++) {
			resultBDD = resultBDD.and(nodeNotChanged(node, size));
		}

		// 2. FALL: Es wird ein neuer Knoten hinzufuegt, der bisher nicht
		// vorhanden war (Annahme: es wird ein Knoten aus dem gesuchten
		// Subgraphen hinzugefuegt)
		RavenBDD tempBDD = encoding.zero();
		for (int node = 0; node < nodeCount; node++) {
			tempBDD = tempBDD.or(addNewNode(node, size));
		}

		resultBDD = resultBDD.or(tempBDD);

		// 3. FALL: Es wird ein Knoten hinzugefuegt, der bereits im Interface
		// vorhanden war (Annahme: es wird ein Knoten aus dem gesuchten
		// Subgraphen hinzugefuegt)
		tempBDD = encoding.zero();
		if (size > 0) {
			for (int node = 0; node < nodeCount; node++) {
				tempBDD = tempBDD.or(addExistingNode(node, size));
			}
			resultBDD = resultBDD.or(tempBDD);
		}

		return resultBDD;
	}

	/**
	 * Konstruiert einen BDD, der genau dann wahr liefert, wenn der Knoten
	 * {@code vertexName} nicht veraendert wurde und die aktuelle
	 * Interfacegroesse {@code ifaceSize} ist.
	 * 
	 * @param node Name des unveraenderten Knotens
	 * @param size aktuelle Interfacegroesse
	 * @return der zu konstruierende BDD
	 */
	private RavenBDD nodeNotChanged(int node, int size) {
		if (node < 0 || node >= nodeCount) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}
		if (size < 0 || size > maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}
		// Unabhaengig davon, ob der Graphknoten vorhanden ist, oder nicht
		// darf sich an diesem Knoten nichts aendern
		RavenBDD resultBDD = encoding.getNodeBit(StateType.current, node)
				.biimp(encoding.getNodeBit(StateType.successor, node));

		// Iteriere ueber alle BDD-Knoten die das gesamte Interface
		// repraesentieren
		for (int i = 0; i < maximumInterface; i++) {
			// Falls der aktuelle Knoten innerhalb des aktuellen Interfaces
			// liegt
			if (i < size) {
				// Das aktuelle Interface darf von der Operation nicht
				// veraendert werden
				resultBDD = resultBDD.and(encoding.getMappingBit(
						StateType.current, node, i).biimp(
						encoding.getMappingBit(StateType.successor, node, i)));
			}
			// Falls der aktuelle Knoten ausserhalb des aktuellen Interfaces
			// liegt
			else {
				// Es darf kein Interface-Knoten gesetzt sein, da es sich um zu
				// grosse Interfaces handelt
				resultBDD = resultBDD.and(encoding
						.getMappingBit(StateType.current, node, i)
						.not()
						.and(encoding.getMappingBit(StateType.successor, node,
								i).not()));
			}
		}

		return resultBDD;
	}

	/**
	 * Konstruiert einen BDD, der genau dann wahr liefert, wenn der Knoten
	 * {@code vertexName} hinzugefuegt wurde und die aktuelle Interfacegroesse
	 * {@code ifaceSize} ist.
	 * 
	 * @param node Name des neu hinzugefuegten Knotens
	 * @param size aktuelle Interfacegroesse
	 * @return der zu konstruierende BDD
	 */
	private RavenBDD addNewNode(int node, int size) {
		if (node < 0 || node >= nodeCount) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}
		if (size < 0 || size >= maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}

		// Unabhaengig davon, ob der Graphknoten vorhanden ist, oder nicht
		// darf sich an diesem Knoten nichts aendern
		RavenBDD resultBDD = encoding.getNodeBit(StateType.current, node).not()
				.and(encoding.getNodeBit(StateType.successor, node));

		// Iteriere ueber alle BDD-Knoten die das aktuelle Interface
		// repraesentieren
		for (int i = 0; i <= size; i++) {
			if (i < size) {
				// Das aktuelle Interface darf von der Operation nicht
				// veraendert werden
				resultBDD = resultBDD.and(encoding
						.getMappingBit(StateType.current, node, i)
						.not()
						.and(encoding.getMappingBit(StateType.successor, node,
								i).not()));
			} else {
				// Das aktuelle Interface darf von der Operation nicht
				// veraendert werden
				resultBDD = resultBDD.and(encoding
						.getMappingBit(StateType.current, node, i)
						.not()
						.and(encoding.getMappingBit(StateType.successor, node,
								i)));
			}
		}

		// Durchlaufe alle Knoten
		for (int anotherNode = 0; anotherNode < nodeCount; anotherNode++) {
			// Falls der aktuelle Knoten nicht der neue Knoten ist
			if (anotherNode != node) {
				resultBDD = resultBDD.and(nodeNotChanged(anotherNode, size));
			}
		}

		return resultBDD;
	}

	/**
	 * Konstruiert einen BDD, der genau dann wahr liefert, wenn der Knoten
	 * {@code vertexName} bereits vorhanden war und die aktuelle
	 * Interfacegroesse {@code ifaceSize} ist.
	 * 
	 * @param node Name des hinzugefuegten Knotens
	 * @param size aktuelle Interfacegroesse
	 * @return der zu konstruierende BDD
	 */
	private RavenBDD addExistingNode(int node, int size) {
		if (node < 0 || node >= nodeCount) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}
		if (size < 1 || size >= maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}

		// Der Knoten muss vorher und nachher vorhanden sein
		RavenBDD resultBDD = encoding.getNodeBit(StateType.current, node).and(
				encoding.getNodeBit(StateType.successor, node));

		// Iteriere ueber alle BDD-Knoten die das Interface des aktuellen
		// Knoten repraesentieren, um zu ueberpruefen, ...
		for (int i = 0; i < size; i++) {
			// dass das Interface durch die Operation nicht veraendert wird
			resultBDD = resultBDD.and(encoding.getMappingBit(StateType.current,
					node, i).biimp(
					encoding.getMappingBit(StateType.successor, node, i)));
		}

		// BDD vorbereiten
		RavenBDD tempBDD = encoding.zero();
		// Iteriere ueber alle BDD-Knoten die Interface des aktuellen Knoten
		// repraesentieren, um zu ueberpruefen, ...
		for (int i = 0; i < size; i++) {
			// dass der Knoten, der "hinzugefuegt" werden soll, bereits im
			// Interface liegt
			tempBDD = tempBDD.or(encoding.getMappingBit(StateType.current,
					node, i));
		}
		// der Knoten wird nun (neu) an die erste Stelle des erweiterten
		// Interfaces aufgenommen
		resultBDD = resultBDD.and(tempBDD);
		resultBDD = resultBDD.and(encoding
				.getMappingBit(StateType.current, node, size).not()
				.and(encoding.getMappingBit(StateType.successor, node, size)));

		// Durchlaufe alle Knoten
		tempBDD = encoding.one();
		for (int anotherNode = 0; anotherNode < nodeCount; anotherNode++) {
			// Falls der aktuelle Knoten nicht der neue Knoten ist
			if (anotherNode != node) {
				tempBDD = tempBDD.and(nodeNotChanged(anotherNode, size));
			}
		}
		resultBDD = resultBDD.and(tempBDD);

		return resultBDD;
	}

	/**
	 * Konstruiert einen BDD, der ueberprueft, ob alle BDD-Knoten aus zu grossen
	 * Interfaces (Knoten fuer die Interfaces @code{maxInterface - ifaceSize})
	 * auf false gesetzt sind.
	 * 
	 * @param size die aktuelle Interfacegroesse
	 * @return der zu konstruierende BDD
	 */
	private RavenBDD constructBDDForInterfaceSizeCheck(int size) {
		if (size < 0 || size > maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}

		// BDD vorbereiten
		RavenBDD resultBDD = encoding.one();
		// Durchlaufe alle Knoten
		for (int node = 0; node < nodeCount; node++) {
			// Durchlaufe alle Interface-Bits fuer Knoten ausserhalb des
			// erweiterten Interfaces
			for (int i = size; i < maximumInterface; i++) {
				// Diese Bits duerfen nicht gesetzt sein, weil das Interface
				// nicht gross genug ist
				resultBDD = resultBDD.and(encoding.getMappingBit(StateType.current, node, i).not());
			}
		}
		return resultBDD;
	}

	/**
	 * Konstruiert einen BDD, der ueberprueft, ob jeder Graph-Knoten, der auch
	 * im (Bild des) Interface liegt, im Graphen vorhanden ist
	 * 
	 * @return der zu konstruierende BDD
	 */
	private RavenBDD constructBDDForVertexExistenceCheck() {
		// BDD vorbereiten
		RavenBDD resultBDD = encoding.one();
		// Durchlaufe alle Knoten
		for (int node = 0; node < nodeCount; node++) {
			// BDD vorbereiten
			RavenBDD tempBDD = encoding.zero();
			// Iteriere ueber alle Interfaceknoten
			for (int i = 0; i < maximumInterface; i++) {
				// ver-oder-e jeden Interfaceknoten (aus dem neuen Interface)
				tempBDD = tempBDD.or(encoding.getMappingBit(StateType.current, node, i));
			}
			// Falls der aktuelle Vertex als Bild im Interface auftaucht, muss
			// dieser Vertex auch wirklich vorhanden sein
			resultBDD = resultBDD.and(tempBDD.imp(encoding.getNodeBit(StateType.current, node)));
		}

		return resultBDD;
	}

	/**
	 * Konstruiert einen BDD, der ueberprueft, ob das aktuelle Interface an
	 * allen Position unveraendert bleibt, an denen die Knoten nicht veraendert
	 * wurden.
	 * 
	 * @param node the index of the node
	 * @param size the current interface size
	 * @return der zu konstruierende BDD
	 */
	private RavenBDD constructBDDForInterfaceUniquenessCheck(int node, int size) {
		if (node < 0 || node >= nodeCount) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}
		if (size < 0 || size > maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}

		try {
			// BDD vorbereiten
			RavenBDD resultBDD = encoding.one();
			
			// Durchlaufe alle Knoten
			for (int anotherNode = 0; anotherNode < nodeCount; anotherNode++) {
				// Falls der aktuelle Knoten nicht der neu hinzugefuegte ist
				if (anotherNode != node) {
					// fordere, dass der Interfaceknoten unveraendert bleibt
					resultBDD = resultBDD.and(
							encoding.getMappingBit(StateType.current, anotherNode, size-1).biimp(
							encoding.getMappingBit(StateType.successor, anotherNode, size-1)));
				}
				// und fordere, falls der Interfaceknoten veraendert wurde,
				// muessen die anderen Knoten unveraendert bleiben
			}
			
			RavenBDD tempBDD = encoding.getMappingBit(StateType.current, node, size-1).xor(
					encoding.getMappingBit(StateType.successor, node, size-1));
			resultBDD = tempBDD.imp(resultBDD);
			return resultBDD;
		}
		catch (Exception e) {
			e.printStackTrace();
		}
		
		return null;
	}

	/**
	 * Konstruiert einen BDD, der ueberprueft, ob das aktuelle Interface an
	 * allen Position unveraendert bleibt, an denen die Knoten nicht veraendert
	 * wurden.
	 * 
	 * @param size the current interface size
	 * @return der zu konstruierende BDD
	 */
	private RavenBDD constructBDDForInterfaceUniquenessCheck(int size) {
		if (size < 0 || size > maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}

		// BDD vorbereiten
		RavenBDD resultBDD = encoding.one();
		// Durchlaufe alle Knoten
		for (int node = 0; node < nodeCount; node++) {
			resultBDD = resultBDD.and(constructBDDForInterfaceUniquenessCheck(
					node, size));
		}

		return resultBDD;
	}

	/**
	 * Konstruiert einen BDD, der ueberprueft, ob alle Knoten vorhanden sind,
	 * die zu einer Kante gehoeren, falls die Kante vorhanden ist.
	 * 
	 * @return der zu konstruierende BDD
	 */
	private RavenBDD constructBDDForEdgeCheck() {
		// BDD vorbereiten
		RavenBDD resultBDD = encoding.one();
		// Durchlaufe alle Kanten
		for(int edge = 0; edge < edgeCount; edge++) {
			Edge e = edgeOrder.get(edge);
			List<Node> adjacentNodes = e.getIncidentNodes();

			// BDD vorbereiten
			RavenBDD tempBDD = encoding.one();
			// Durchlaufe alle adjazenten Knoten der aktuellen Kante
			for (Node n : adjacentNodes) {
				// Fuer jede Kanten muessen alle adjazenten Knoten vorhanden
				// sein
				tempBDD = tempBDD.and(encoding.getNodeBit(StateType.current, nodeOrder.indexOf(n)));
			}
			// Falls die Kante vorhanden ist, muessen auch alle adjazenten
			// Knoten
			// vorhanden sein
			resultBDD = resultBDD.and(encoding.getEdgeBit(StateType.current, edge).imp(tempBDD));
		}
		return resultBDD;
	}
	
	/**
	 * Constructs a BDD which defines the conditions that must be fullfilled 
	 * by the initial states.
	 * 
	 * @param interfaceSize the size for the inner- or outer interface
	 * @return The BDD describing the legal initial states
	 */
	private RavenBDD createSubgraphInitialStatesConditionBDD(int interfaceSize) {
		
		RavenBDD finalResultBDD = encoding.one();
		
		//For each node in the graph...
		for (int node = 0; node < nodeCount; node++) {
			
			//...the mapping bits beyond the given interfaceSize has to be 0
			for(int size=interfaceSize; size < maximumInterface;size++){
				finalResultBDD = finalResultBDD.and(encoding.getMappingBit(
						StateType.current, node, size).not());
			}
			
			//In addition each vector for the mapping bits of a node bit set to 1 has to consist of at least one 1 bit
			RavenBDD columnBDD = encoding.zero();
			for(int size = 0; size < interfaceSize; size++) {
				columnBDD = columnBDD.or(encoding.getMappingBit(StateType.current, node, size));
			}

			RavenBDD helpBDD = encoding.getNodeBit(StateType.current, node).biimp(columnBDD);
			
			finalResultBDD = finalResultBDD.and(helpBDD.clone());
		}
		
		//Each interface position may only be mapped by exactly one node
		for(int size = 0; size <interfaceSize; size++) {
			RavenBDD resultBDD = encoding.zero();
			for(int pos = 0; pos < nodeCount;pos++) {
				
				RavenBDD rowBDD = encoding.one();
				for(int node = 0; node<nodeCount; node++) {
					if(node == pos){
						rowBDD = rowBDD.and(encoding.getMappingBit(StateType.current, node, size));
					}
					else {
						rowBDD = rowBDD.and(encoding.getMappingBit(StateType.current, node, size).not());
					}
				}
				
				resultBDD = resultBDD.or(rowBDD.clone());
				
			}
			
			finalResultBDD = finalResultBDD.and(resultBDD.clone());
		}
		return finalResultBDD;
	}
	
	/**
	 * Constructs a BDD which defines the conditions that must be fullfilled 
	 * by the final states.
	 * 
	 * @param interfaceSize the size for the inner- or outer interface
	 * @return The BDD describing the legal final states
	 */
	private RavenBDD createSubgraphFinalStatesConditionBDD(int interfaceSize, boolean isSloppy) {
		RavenBDD finalResultBDD = encoding.one();
		
		//For each node in the graph...
		for (int node = 0; node < nodeCount; node++) {
			
			//...the mapping bits beyond the given interfaceSize have to be 0
			for(int size=interfaceSize; size < maximumInterface;size++){
				finalResultBDD = finalResultBDD.and(
						encoding.getMappingBit(StateType.current, node, size).not());
			}
		}
		
		if(!isSloppy) {
			//Each interface position may only be mapped by exactly one node
			for(int size = 0; size <interfaceSize; size++) {
				RavenBDD resultBDD = encoding.zero();
				for(int pos = 0; pos < nodeCount;pos++) {
					
					RavenBDD rowBDD = encoding.one();
					for(int node = 0; node<nodeCount; node++) {
						if(node == pos){
							rowBDD = rowBDD.and(encoding.getMappingBit(StateType.current, node, size));
						}
						else {
							rowBDD = rowBDD.and(encoding.getMappingBit(StateType.current, node, size).not());
						}
					}
					
					resultBDD = resultBDD.or(rowBDD);
					
				}
				
				finalResultBDD = finalResultBDD.and(resultBDD);
			}
		}
		return finalResultBDD;
		
	}


}