/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.data.automata;

import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.actions.ComputationEvent;
import de.uni_due.inf.ti.raven.actions.ComputationEvent.EventType;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.Signature;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.encodings.UnionEncoding;

/**
 * This class implements a graph automaton accepting the language which is obtained by the 
 * union of the languages of the two underlying graph automata. 
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.40 $
 *
 */
public class UnionAutomaton extends Automaton {

	private static final long serialVersionUID = 5902616799156961718L;

	transient private UnionEncoding encoding;
	private Automaton first;
	private Automaton second;
	transient private Logger logger;

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new union automaton.
	 * @param encoding the encoding for the states
	 * @param signature the signature of this automaton
	 */
	private UnionAutomaton(UnionEncoding encoding, Signature signature) {
		super(Type.union, encoding, signature);
		this.encoding = encoding;
		this.first = encoding.getFirstAutomaton();
		this.second = encoding.getSecondAutomaton();
		
		//Register logger
		logger = Logger.getLogger(getClass().getName());
	}
	
	/**
	 * Creates a new union automaton based on the two automata {@code first} and {@code second}.
	 * @param encoding the encoding for the states
	 */
	protected UnionAutomaton(UnionEncoding encoding) {
		this(encoding, encoding.getFirstAutomaton().getAlphabet());
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public Automaton clone() throws CloneNotSupportedException {
		Automaton clone = new UnionAutomaton(encoding, signature);
		clone.initialStates = initialStates.clone();
		clone.finalStates = finalStates.clone();
		clone.nonFinalStates = nonFinalStates.clone();
		clone.states = states.clone();
		
		for(Operation operation: stateTransBDDs.keySet()) {
			clone.stateTransBDDs.put(operation, stateTransBDDs.get(operation).clone());
		}

		return clone;
	}
	
	@Override
	public void initializeEncoding() {
		// TODO This does not work for saving/loading automata 
		encoding = UnionEncoding.getEncoding(first, second);
		super.encoding = encoding;
		super.initializeEncoding();
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	protected void createStates() {
		states = encoding.convertStates(first).or(encoding.convertStates(second));
	}

	@Override
	protected void createInitialAndFinalStates() {
		//Log entering
		logger.entering(getClass().getName(), "createInitialAndFinalStates");
		
		//Determine the initial states and inner interface size
		initialStates = encoding.convertInitialStates(first).or(
				encoding.convertInitialStates(second));
		innerInterface = first.getInnerInterface();
		
		//Determine the final states and the outer interface size
		finalStates = encoding.convertFinalStates(first).or(encoding.convertFinalStates(second));
		outerInterface = first.getOuterInterface();		
		
		nonFinalStates = encoding.convertNonFinalStates(first).or(
				encoding.convertNonFinalStates(second));
		
		//Log exiting
		logger.exiting(getClass().getName(), "createInitialAndFinalStates");
	}

	@Override
	protected void createTransitions() {
		//Log entering
		logger.entering(getClass().getName(), "createTransitions");
		
		// Iteriere ueber alle Graphoperationen
		for (Operation letter: getAlphabet()) {
			fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED, 
					letter.toString(), 2, 0));
			
			RavenBDD firstTransBDD = encoding.convertTransitions(first, letter);
			
			fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED, 
					letter.toString(), 2, 1));
			
			RavenBDD secondTransBDD = encoding.convertTransitions(second, letter);
			
			fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED, 
					letter.toString(), 2, 2));
			stateTransBDDs.put(letter, firstTransBDD.or(secondTransBDD));
		}
		
		//Log exiting
		logger.exiting(getClass().getName(), "createTransitions");
	}

}
