/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.data.automata;

import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.actions.ComputationEvent;
import de.uni_due.inf.ti.raven.actions.ComputationEvent.EventType;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.Operation.Connect;
import de.uni_due.inf.ti.raven.data.Operation.Perm;
import de.uni_due.inf.ti.raven.data.Operation.Res;
import de.uni_due.inf.ti.raven.data.Operation.Trans;
import de.uni_due.inf.ti.raven.data.Operation.Vertex;
import de.uni_due.inf.ti.raven.data.Signature;
import de.uni_due.inf.ti.raven.data.Operation.Fuse;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.encodings.BDDEncoding.StateType;
import de.uni_due.inf.ti.raven.encodings.VertexCountingEncoding;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class implements a graph automaton accepting the language of all graphs whose number of 
 * nodes is equal to <i>remainder</i> modulo <i>divisor</i>. The remainder and the divisor are 
 * not part of the input but given at creation time of the graph automaton.
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.20 $
 * 
 */
public class VertexCountingAutomaton extends Automaton {

	private static final long serialVersionUID = 3031400095389229862L;
	transient private Logger logger;
	private final int remainder;
	private final int divisor;
	private final int divisorBitLength;
	transient private VertexCountingEncoding encoding;

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new vertex-counting automaton counting the number of vertices modulo a divisor.
	 * @param encoding the encoding for the states
	 * @param signature the signature of this automaton
	 * @param remainder the remainder for the number of vertices
	 */
	protected VertexCountingAutomaton(VertexCountingEncoding encoding,
			Signature signature, int remainder) {
		super(Type.vertexCounting, encoding, signature);
		this.encoding = encoding;
		this.remainder = remainder;
		divisor = encoding.getDivisor();
		divisorBitLength = encoding.getRemainderBitLength();

		// Register logger
		logger = Logger.getLogger(getClass().getName());
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public Automaton clone() throws CloneNotSupportedException {
		Automaton clone = new VertexCountingAutomaton(encoding, signature, remainder);
		clone.initialStates = initialStates.clone();
		clone.finalStates = finalStates.clone();
		clone.nonFinalStates = nonFinalStates.clone();
		clone.states = states.clone();

		for (Operation operation : stateTransBDDs.keySet()) {
			clone.stateTransBDDs.put(operation, stateTransBDDs.get(operation)
					.clone());
		}

		return clone;
	}
	
	/**
	 * Returns the remainder for the number of vertices
	 * @return the remainder
	 */
	public int getRemainder(){
		return remainder;
	}

	/**
	 * Returns the divisor modulo which vertices are counted
	 * @return the divisor
	 */
	public int getDivisor(){
		return divisor;
	}
	
	@Override
	public void initializeEncoding(){
		encoding = VertexCountingEncoding.getEncoding(maximumInterface,
				0, 0, divisor, encodingName);
		super.encoding = encoding;
		super.initializeEncoding();
	}


	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	protected void createStates() {
		// Log entering
		logger.entering(getClass().getName(), "createStates");

		states = encoding.zero();
		for (int i = 0; i <= maximumInterface; i++) {
			RavenBDD temp = encoding.getInterfaceSize(StateType.current, i);

			for (int count = 0; count < divisor; count++) {
				states = states.or(temp.clone().and(
						getRemainderEquals(StateType.current, count)));
			}
		}

		// Log exiting
		logger.exiting(getClass().getName(), "createStates");

	}

	@Override
	protected void createInitialAndFinalStates() {
		// Log entering
		logger.entering(getClass().getName(), "createInitialAndFinalStates");

		initialStates = encoding.getInterfaceSize(StateType.current, innerInterface);
		initialStates = initialStates.and(getRemainderEquals(
				StateType.current, innerInterface%divisor));

		finalStates = encoding.getInterfaceSize(StateType.current, outerInterface);
		finalStates = finalStates.and(getRemainderEquals(StateType.current,
				remainder));

		nonFinalStates = states.clone().and(finalStates.clone().not());

		// Log exiting
		logger.exiting(getClass().getName(), "createInitialAndFinalStates");
	}

	@Override
	protected void createTransitions() {
		// Log entering
		logger.entering(getClass().getName(), "createTransitions");

		// iterate over all operations
		for (Operation operation : getAlphabet()) {
			RavenBDD transBDD = stateTransBDDs.get(operation);
			RavenBDD tempBDD;
			// iterate over all possible node counts

			if (operation instanceof Connect) {
				int arity = ((Connect) operation).getArity();
				for (int size = arity; size <= maximumInterface; size++) {
					// Check that the current interface size is i and that
					// the successor
					// interface is still i
					tempBDD = encoding
							.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(
							StateType.successor, size));

					// The node count is not changed
					tempBDD = tempBDD.and(constructNodeCountUnchangedBDD());

					transBDD = transBDD.or(tempBDD);
					fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED,
							operation.toString(), maximumInterface, size));
				}
			}

			if (operation instanceof Trans || operation instanceof Perm) {
				for (int size = 2; size <= maximumInterface; size++) {
					// Check that the current interface size is i and that
					// the successor
					// interface is still i
					tempBDD = encoding
							.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(
							StateType.successor, size));

					// The node count is not changed
					tempBDD = tempBDD.and(constructNodeCountUnchangedBDD());
					transBDD = transBDD.or(tempBDD);
					fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED,
							operation.toString(), maximumInterface, size));
				}
			}
			if (operation instanceof Fuse) {
				for (int nodeCount = 0; nodeCount < divisor; nodeCount++) {
					for (int size = 2; size <= maximumInterface; size++) {
						// Check that the current interface size is i and that
						// the successor
						// interface is i-1
						tempBDD = encoding.getInterfaceSize(StateType.current,
								size);
						tempBDD = tempBDD.and(encoding.getInterfaceSize(
								StateType.successor, size - 1));

						// The node count is decreased by 1
						tempBDD = tempBDD.and(constructNodeCountChangedBDD(
								nodeCount, modulo(nodeCount - 1)));

						transBDD = transBDD.or(tempBDD);
						fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED,
								operation.toString(), maximumInterface, size));
					}
				}
			}
			if (operation instanceof Res) {
				for (int size = 1; size <= maximumInterface; size++) {
					// Check that the current interface size is i and that
					// the successor
					// interface is i-1
					tempBDD = encoding
							.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(
							StateType.successor, size - 1));

					// The node count is not changed
					tempBDD = tempBDD.and(constructNodeCountUnchangedBDD());

					transBDD = transBDD.or(tempBDD);
					fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED,
							operation.toString(), maximumInterface, size));
				}
			}
			if (operation instanceof Vertex) {
				for (int nodeCount = 0; nodeCount < divisor; nodeCount++) {
					for (int size = 0; size <maximumInterface; size++) {
						// Check that the current interface size is i and that
						// the successor
						// interface is i+1
						tempBDD = encoding.getInterfaceSize(StateType.current,
								size);
						tempBDD = tempBDD.and(encoding.getInterfaceSize(
								StateType.successor, size + 1));

						// The node count is increased by 1
						tempBDD = tempBDD.and(constructNodeCountChangedBDD(
								nodeCount, modulo(nodeCount+1)));

						transBDD = transBDD.or(tempBDD);
						fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_ADVANCED,
								operation.toString(), maximumInterface, size));
					}
				}
			}

			stateTransBDDs.put(operation, transBDD);
		}

		// Log exiting
		logger.exiting(getClass().getName(), "createTransitions");
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the remainder of {@code number} modulo the divisor of this automaton.
	 * @param number the number of the remainder is computed
	 * @return the remainder
	 */
	private int modulo(int number) {
		if(number>=0) {
			return number % divisor;
		}

		while(number<0) {
			number += divisor;
		}
		return number;
	}
	
	/**
	 * Returns a BDD encoding that in the state {@code state} the size of the 
	 * remainder is equal to {@code value} if and only if the BDD is {@code true}.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param value the value of the remainder
	 * @return a BDD encoding that the remainder is equal to {@code value}
	 */
	private RavenBDD getRemainderEquals(StateType type, int value) {
		if(value<0 || value>=divisor) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_REMAINDER_EXCEPTION);
		}
		
		RavenBDD bdd = encoding.one();
		for(int i=0;i<divisorBitLength;i++) {
			int b = value%2;
			value >>= 1;
		
			if(b==0) {
				bdd = bdd.and(encoding.getRemainderBit(type, i).not());
			}
			else {
				bdd = bdd.and(encoding.getRemainderBit(type, i));
			}
		}
		
		return bdd;
	}

	/**
	 * Returns a BDD encoding that the size of the remainder equals {@code before} in the 
	 * current state and equals {@code after} in the successor state if and only if the 
	 * BDD is {@code true}.
	 * @return a BDD encoding that the remainder has been changed correctly
	 */
	private RavenBDD constructNodeCountChangedBDD(int before, int after) {
		return getRemainderEquals(StateType.current, before).and(
				getRemainderEquals(StateType.successor, after));
	}
	
	/**
	 * Returns a BDD encoding that the size of the remainder has not been changed between 
	 * the current and the successor state if and only if the BDD is {@code true}.
	 * @return a BDD encoding that the remainder has not been changed
	 */
	private RavenBDD constructNodeCountUnchangedBDD() {
		RavenBDD bdd = encoding.one();
		
		for(int i=0;i<divisorBitLength;i++) {
			bdd = bdd.and(encoding.getRemainderBit(StateType.current, i).biimp(
					encoding.getRemainderBit(StateType.successor, i)));
		}
		
		return bdd;
	}
}
