/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.data.automata;

import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.Operation.Connect;
import de.uni_due.inf.ti.raven.data.Operation.Fuse;
import de.uni_due.inf.ti.raven.data.Operation.Perm;
import de.uni_due.inf.ti.raven.data.Operation.Res;
import de.uni_due.inf.ti.raven.data.Operation.Trans;
import de.uni_due.inf.ti.raven.data.Operation.Vertex;
import de.uni_due.inf.ti.raven.data.Signature;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.encodings.BDDEncoding.StateType;
import de.uni_due.inf.ti.raven.encodings.VertexCoverEncoding;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class implements a graph automaton accepting the language of all graphs which have a 
 * vertex cover of size at most <i>k</i>, i.e. a set <i>C</i> of nodes with size at most <i>k</i> 
 * such that each edge of the graph is incident to at least one node of <i>C</i>.
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.48 $
 *
 */
public class VertexCoverAutomaton extends Automaton {	

	private static final long serialVersionUID = 740795299485953890L;

	transient private VertexCoverEncoding encoding;

	private int coverSize;
	private int coverSizeBitLength;
	private int innerInterfaceSize, outerInterfaceSize;

	transient private Logger logger;

	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new vertex cover automaton.
	 * @param encoding the encoding for the states
	 * @param signature the signature of this automaton
	 */
	protected VertexCoverAutomaton(VertexCoverEncoding encoding, Signature signature) {
		super(Type.vertexCover, encoding, signature);
		this.encoding = encoding;
		coverSize = encoding.getVertexCoverSize();
		coverSizeBitLength = encoding.getVertexCoverBitLength();
		innerInterfaceSize = encoding.getInnerInterface();
		outerInterfaceSize = encoding.getOuterInterface();

		//Register logger
		logger = Logger.getLogger(this.getClass().getName());
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public Automaton clone() throws CloneNotSupportedException {
		Automaton clone = new VertexCoverAutomaton(encoding, signature);
		clone.initialStates = initialStates.clone();
		clone.finalStates = finalStates.clone();
		clone.nonFinalStates = nonFinalStates.clone();
		clone.states = states.clone();
		
		for(Operation operation: stateTransBDDs.keySet()) {
			clone.stateTransBDDs.put(operation, stateTransBDDs.get(operation).clone());
		}

		return clone;
	}
	
	/**
	 * Returns the maximum size of the vertex cover.
	 * @return the maximum vertex cover size
	 */
	public int getMaximumVertexCoverSize() {
		return coverSize;
	}
	
	/**
	 * Returns {@code true} if and only if in the state {@code state} the {@code node}-th
	 * interface is in the vertex cover.
	 * @param state the state, either <i>current state</i> or <i>successor state</i>
	 * @param node the index of the interface node
	 * @return {@code true} if the node is in the vertex cover, {@code false} otherwise
	 */
	public boolean isInVertexCover(RavenBDD state, int node) {
		if(state==null) {
			throw new NullPointerException(RavenExceptionMessages.BDD_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(node<1 || node>maximumInterface) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}
		
		RavenBDD temp = state.clone().restrict(
				encoding.getMembershipBit(StateType.current, node-1));
		temp = temp.exists(curStateDomain.getDomainBDD());
		return temp.isOne();
	}
	
	@Override
	public void initializeEncoding() {
		encoding = VertexCoverEncoding.getEncoding(maximumInterface, innerInterfaceSize, 
				outerInterfaceSize, coverSize, encodingName);
		super.encoding = encoding;
		super.initializeEncoding();
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	protected void createStates() {
		//Log entering
		logger.entering(getClass().getName(), "createStates");

		states = encoding.zero();
		for(int i=0;i<=maximumInterface;i++) {
			RavenBDD temp = encoding.one();
			for(int j=i;j<maximumInterface;j++) {
				temp = temp.and(encoding.getUndefinedNode(StateType.current, j));
			}
				
			temp = temp.and(encoding.getInterfaceSize(StateType.current, i));
			temp = temp.and(getCoverSizeIsLessOrEqual(StateType.current, coverSize));
			states = states.or(temp);
		}

		//Log exiting
		logger.exiting(getClass().getName(), "createStates");
	}

	@Override
	protected void createInitialAndFinalStates() {
		//Log entering
		logger.entering(getClass().getName(), "createInitialAndFinalStates");
		
		initialStates = encoding.getInterfaceSize(StateType.current, innerInterfaceSize);
		initialStates = initialStates.and(createInitialStateSet(innerInterfaceSize, 0, 0));
		for(int i=innerInterfaceSize;i<maximumInterface;i++) {
			initialStates = initialStates.and(
					encoding.getMembershipBit(StateType.current, i).not());
		}
		
		finalStates = encoding.getInterfaceSize(StateType.current, outerInterfaceSize);
		finalStates = finalStates.and(getCoverSizeIsLessOrEqual(StateType.current, coverSize));
		
		nonFinalStates = states.clone().and(finalStates.clone().not());

		//Log exiting
		logger.exiting(getClass().getName(), "createInitialAndFinalStates");
	}

	@Override
	protected void createTransitions() {
		//Log entering
		logger.entering(getClass().getName(), "createTransitions");

		// Iteriere ueber alle Graphoperationen
		for (Operation operation: getAlphabet()) {
			RavenBDD transBDD = stateTransBDDs.get(operation);
			RavenBDD tempBDD;

			// Falls die Operation die das Verkleben mit einer Hyperkante ist
			if (operation instanceof Connect) {
				// Eigenschaften der neuen Kante bestimmen
				int arity = ((Connect) operation).getArity();

				//The CONNECT-operation is applicable only if the arity of the edge 
				//is greater or equal to the current interface size.
				for(int size=arity;size<=maximumInterface;size++) {
					//Check that the current interface size is i and that the interface 
					//is not changed by the CONNECT-operation
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size));

					//Check that the vertex cover is not changed by the CONNECT-operation
					for(int node=0;node<maximumInterface;node++) {
						if(node<size) {
							tempBDD = tempBDD.and(getCoverNotChangedBDD(node));
						}
						else {
							tempBDD = tempBDD.and(getInterfaceUndefinedBDD(node));
						}
					}

					//Check that the size of the vertex cover has not been changed
					tempBDD = tempBDD.and(getCoverSizeNotChanged());

					//Check that at least one node which is adjacent
					//to the new edge is in the vertex cover
					tempBDD = tempBDD.and(getAdjacentNodesInCoverBDD(size, arity));

					//Add transition BDD for the i-th interface
					transBDD = transBDD.or(tempBDD);
				}
			}
			else if (operation instanceof Fuse) {
				//The FUSE-operation is applicable only if at least two nodes are "touchable" 
				//through the interface, i.e. the interface size has to be at least 2
				for(int size=2;size<=maximumInterface;size++) {
					//Check that the current interface size is i and that the successor 
					//interface is decreased by one
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size-1));

					//For all nodes which could possibly occur in the interface, check that ...
					for(int node=0;node<maximumInterface;node++) {
						//the first i-1 nodes are not changed by the FUSE-operation
						if(node<size-1) {
							tempBDD = tempBDD.and(getCoverNotChangedBDD(node));
						}
						//the i-th node is fused with the (i-1)-th node
						else if(node==size-1) {
							tempBDD = tempBDD.and(
									encoding.getMembershipBit(StateType.current, node-1).biimp(
											encoding.getMembershipBit(StateType.current, node)));
							tempBDD = tempBDD.and(
									encoding.getMembershipBit(StateType.successor, node).not());
						}
						//the nodes outside the interface have not been set
						else {
							tempBDD = tempBDD.and(getInterfaceUndefinedBDD(node));
						}
					}

					//Check that the size of the vertex cover has not been changed
					tempBDD = tempBDD.and(getCoverSizeNotChanged());

					//Add transition BDD for the i-th interface
					transBDD = transBDD.or(tempBDD);
				}
			}
			else if (operation instanceof Perm) {
				//The PERM-operation is applicable only if at least two nodes are "touchable" 
				//through the interface, i.e. the interface size has to be at least 3
				for(int size=3;size<=maximumInterface;size++) {
					//Check that the current interface size is i and that the interface 
					//is not changed by the PERM-operation
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size));

					//Check that the vertex cover is permuted in a way that the cover of the 
					//j-th node is shifted to the (j+1)-th node
					for(int node=0;node<maximumInterface;node++) {
						if(node<size) {
							tempBDD = tempBDD.and(permuteNodes(size, node));
						}
						else {
							tempBDD = tempBDD.and(getInterfaceUndefinedBDD(node));
						}
					}

					//Check that the size of the vertex cover has not been changed
					tempBDD = tempBDD.and(getCoverSizeNotChanged());

					//Add transition BDD for the i-th interface
					transBDD = transBDD.or(tempBDD);
				}
			}
			else if (operation instanceof Res) {
				//The RES-operation is applicable only if at least one node is "touchable"
				//through the interface, i.e. the interface size has to be at least 2
				for(int size=1;size<=maximumInterface;size++) {
					//Check that the current interface size is i and that the interface 
					//size of the successor interface is i-1
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size-1));

					//For all nodes which could possibly occur in the interface, check that ...
					for(int node=0;node<maximumInterface;node++) {
						//the first i-1 nodes are not changed by the RES-operation
						if(node<size-1) {
							tempBDD = tempBDD.and(getCoverNotChangedBDD(node));
						}
						//the i-th node is restricted
						else if(node==size-1) {
							tempBDD = tempBDD.and(removeNode(node));
						}
						//the nodes outside the interface have not been set
						else {
							tempBDD = tempBDD.and(getInterfaceUndefinedBDD(node));
						}
					}

					//Add transition BDD for the i-th interface
					transBDD = transBDD.or(tempBDD);
				}
			}
			else if (operation instanceof Trans) {
				//The TRANS-operation is applicable only if at least two nodes are "touchable" 
				//through the interface, i.e. the interface size has to be at least 2
				for(int size=2;size<=maximumInterface;size++) {
					//Check that the current interface size is i and that the interface 
					//is not changed by the TRANS-operation
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size));

					//Check that the vertex cover is transposed in a way that the cover of the 
					//first two nodes are transposed and the other nodes are not changed
					for(int node=0;node<maximumInterface;node++) {
						if(node<size) {
							tempBDD = tempBDD.and(transposeNodes(size, node));
						}
						else {
							tempBDD = tempBDD.and(getInterfaceUndefinedBDD(node));
						}
					}

					//Check that the size of the vertex cover has not been changed
					tempBDD = tempBDD.and(getCoverSizeNotChanged());

					//Add transition BDD for the i-th interface
					transBDD = transBDD.or(tempBDD);
				}
			}
			else if (operation instanceof Vertex) {
				//The VERTEX-operation is applicable only if at least one node could be added
				//to the interface, i.e. the interface size has to be at most (maxSize-1)
				for(int size=0;size<maximumInterface;size++) {
					//Check that the current interface size is i and that the interface 
					//size of the successor interface is i+1
					tempBDD = encoding.getInterfaceSize(StateType.current, size);
					tempBDD = tempBDD.and(encoding.getInterfaceSize(StateType.successor, size+1));

					//For all nodes which could possibly occur in the interface, check that ...
					for(int node=0;node<maximumInterface;node++) {
						//the (existing) vertex cover has not changed and that the cover has been 
						//extended correctly, i.e. the interface node is either in the cover or not 
						if(node<size) {
							tempBDD = tempBDD.and(constructMembershipNotChangedBDD(node));
						}
						else if(node==size) {
							tempBDD = tempBDD.and(addNode(size));
						}
						//the nodes which are not in the interface are not in the cover 
						else {
							tempBDD = tempBDD.and(getInterfaceUndefinedBDD(node));
						}
					}
					
					transBDD = transBDD.or(tempBDD);
				}
			}
			stateTransBDDs.put(operation, transBDD);
		}
		
		//Log exiting
		logger.exiting(getClass().getName(), "createTransitions");
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	private RavenBDD createInitialStateSet(int interfaceSize, int node, int setSize) {
		if(node>=interfaceSize) {
			return getCoverSizeEquals(StateType.current, setSize);
		}

		RavenBDD coveredNodeBDD = encoding.zero();
		if(setSize<coverSize) {
			coveredNodeBDD = encoding.getMembershipBit(StateType.current, node).and(
					createInitialStateSet(interfaceSize, node+1, setSize+1));
		}
		
		RavenBDD uncoveredNodeBDD = encoding.getMembershipBit(StateType.current, node).not().and(
				createInitialStateSet(interfaceSize, node+1, setSize));
		
		return coveredNodeBDD.or(uncoveredNodeBDD);
	}
	
	/**
	 * Returns a BDD encoding that in the state {@code state} the size of the 
	 * vertex cover is equal to {@code size} if and only if the BDD is {@code true}.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param size the size of the vertex cover 
	 * @return a BDD encoding that the vertex cover is equal to {@code size}
	 */
	private RavenBDD getCoverSizeEquals(StateType type, int size) {
		if(size<0 || size>coverSize) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_VERTEX_COVER_SIZE_EXCEPTION);
		}

		RavenBDD bdd = encoding.one();
		for(int i=0;i<coverSizeBitLength;i++) {
			int b = size%2;
			size >>= 1;

			if(b==0) {
				bdd = bdd.and(encoding.getVertexCoverSizeBit(type, i).not());
			}
			else {
				bdd = bdd.and(encoding.getVertexCoverSizeBit(type, i));
			}
		}

		return bdd;
	}
	
	/**
	 * Returns a BDD encoding that in the state {@code state} the size of the 
	 * vertex cover is less or equal to {@code size} if and only if the BDD is {@code true}.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param size the maximum size of the vertex cover 
	 * @return a BDD encoding that the vertex cover is less or equal to {@code size}
	 */
	private RavenBDD getCoverSizeIsLessOrEqual(StateType type, int size) {
		if(size<0 || size>coverSize) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_VERTEX_COVER_SIZE_EXCEPTION);
		}

		RavenBDD bdd = encoding.zero();
		for(int i=0;i<=size;i++) {
			bdd = bdd.or(getCoverSizeEquals(type, i));
		}

		return bdd;
	}
	
	/**
	 * Constructs a BDD which is {@code true} if and only if the membership bit of the 
	 * {@code node}-th interface node has not been changed.
	 * @param node the index of the node
	 * @return a BDD encoding that the {@code node}-th membership bit has not been changed.
	 */
	private RavenBDD constructMembershipNotChangedBDD(int node) {
		if(node<0 || node>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}

		return encoding.getMembershipBit(StateType.current, node).biimp(
				encoding.getMembershipBit(StateType.successor, node));
	}
	
	/**
	 * Constructs a BDD which is {@code true} if and only if the size of the 
	 * vertex cover has not been changed by the transition from the current 
	 * to the successor state. 
	 * @return a BDD encoding that interface node has been undefined
	 */
	private RavenBDD getCoverSizeNotChanged() {
		RavenBDD resultBDD = encoding.zero();
		for (int j = 0; j <= coverSize; j++) {
			resultBDD = resultBDD.or(getCoverSizeEquals(StateType.current, j).and(
					getCoverSizeEquals(StateType.successor, j)));
		}

		return resultBDD;
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if a new node is added 
	 * to the last position of interface (i.e. the membership bits for the 
	 * interface nodes at position less or equal to {@code size} have not been 
	 * changed and the membership bits for nodes at positions greater than 
	 * {@code size} have not been set (in the current state).
	 * 
	 * <emph>Note:</emph> Whether the newly added node is member of the vertex 
	 * cover is chosen non-deterministically depending on the value of the 
	 * membership bit in the successor state. If the new interface node is a 
	 * node of the vertex cover, then the vertex cover size for the successor 
	 * state has been increased by 1. Otherwise the 
	 * vertex cover is not changed.
	 * 
	 * @param size the size of the interface of the current state
	 * @return a BDD encoding that a new node is added correctly to the last position of the 
	 * interface
	 */
	private RavenBDD addNode(int size) {
		if(size<0 || size>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}
		if(size==maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ADDING_NODE_NOT_POSSIBLE_EXCEPTION);
		}

		// before the node was undefined
		RavenBDD resultBDD = encoding.getUndefinedNode(StateType.current, size);
		
		// if the new node is in the cover, increase the cover size by one
		RavenBDD tempBDD = encoding.zero();
		for(int j=0;j<coverSize;j++) {
			tempBDD = tempBDD.or(getCoverSizeEquals(StateType.current, j).and(
					getCoverSizeEquals(StateType.successor, j+1)));
		}

		resultBDD = resultBDD.and(
				encoding.getMembershipBit(StateType.successor, size).imp(tempBDD));

		// if the new node is not in the cover, the cover size is unchanged
		tempBDD = encoding.zero();
		for(int j=0;j<=coverSize;j++) {
			tempBDD = tempBDD.or(getCoverSizeEquals(StateType.current, j).and(
					getCoverSizeEquals(StateType.successor, j)));
		}

		return resultBDD.and(
				encoding.getMembershipBit(StateType.successor, size).not().imp(tempBDD));
	}
	
	/**
	 * Constructs a BDD which is {@code true} if and only if 
	 * <ul>
	 * 		<li>the first interface node of the current state is mapped to 
	 * 		the second interface node of the successor state, and</li>
	 * 		<li>the second interface node of the current state is mapped to 
	 * 		the third interface node of the successor state, and</li>
	 * 		</li>..., and</li>
	 * 		<li>the last interface node of the current state is mapped to 
	 * 		the first interface node of the successor state</li>
	 * </ul>

	 * @param size the interface size
	 * @param node the index of the node
	 * @return a BDD encoding that the interface nodes have been permuted correctly
	 */
	private RavenBDD permuteNodes(int size, int node) {
		if (size < 3 || size > maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}
		if(node<0 || node>=size) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}

		return encoding.getMembershipBit(StateType.current, node).biimp(
				encoding.getMembershipBit(StateType.successor, (node+1)%size));
	}
	
	/**
	 * Constructs a BDD which is {@code true} if and only if 
	 * <ul>
	 * 		<li>the last interface node is removed from the successor state's interface, and</li>
	 * </ul>
	 * 
	 * @param node the index of the node which is to be removed
	 * @return a BDD encoding that the last node of the interface has been removed correctly
	 */
	private RavenBDD removeNode(int node) {
		if (node < 0 || node > maximumInterface) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}
		
		RavenBDD resultBDD = encoding.getMembershipBit(StateType.successor, node).not();
		return resultBDD.and(getCoverSizeNotChanged());
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if 
	 *  <ul>
	 * 		<li>the first node of the current state's interface is mapped 
	 * 		to the second node of the successors state's interface, and</li>
	 * 		<li>the second node of the current state's interface is mapped 
	 * 		to the first node of the successors state's interface, and</li>
	 * 		<li>all other nodes are not permuted</li>
	 * </ul>
	 * @param size the size of the interface of the current state
	 * @param node the index of the node
	 * @return a BDD encoding that the first two nodes of the interface are transposed
	 */
	private RavenBDD transposeNodes(int size, int node) {
		if(size<2 || size>maximumInterface) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}
		if(node<0 || node>=size) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}
		
		
		if(node==0) {
			return encoding.getMembershipBit(StateType.current, 0).biimp(
					encoding.getMembershipBit(StateType.successor, 1));
		}
		else if(node==1) {
			return encoding.getMembershipBit(StateType.current, 1).biimp(
					encoding.getMembershipBit(StateType.successor, 0));
		}
		else {
			return constructMembershipNotChangedBDD(node);
		}
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if at least one node incident to the edge 
	 * of arity {@code arity} is in the vertex cover. 
	 * @param size the size of the current interface
	 * @param arity the arity of the edge
	 * @return a BDD encoding that at least one incident node is in the vertex cover
	 */
	private RavenBDD getAdjacentNodesInCoverBDD(int size, int arity) {
		RavenBDD resultBDD = encoding.zero();
		for(int k=size-arity;k<size;k++) {
			resultBDD = resultBDD.or(encoding.getMembershipBit(StateType.current, k));	    
		}

		return resultBDD;
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if the membership node 
	 * at position {@code node} has been undefined in both the current and the 
	 * successor state. 
	 * @param node the index of the node
	 * @return a BDD encoding that interface node has been undefined
	 */
	private RavenBDD getInterfaceUndefinedBDD(int node) {
		return encoding.getMembershipBit(StateType.current, node).not().and(
				encoding.getMembershipBit(StateType.successor, node).not());
	}

	/**
	 * Constructs a BDD which is {@code true} if and only if the membership node 
	 * at position {@code node} has not been changed during the transition from 
	 * the current to the successor state. 
	 * @param node the index of the node
	 * @return a BDD encoding that membership node has not been changed
	 */
	private RavenBDD getCoverNotChangedBDD(int node) {
		return encoding.getMembershipBit(StateType.current, node).biimp(
				encoding.getMembershipBit(StateType.successor, node));
	}

}
