/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.data.bdd;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.io.StringWriter;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import net.sf.javabdd.BDD;
import net.sf.javabdd.BDD.BDDIterator;
import net.sf.javabdd.BDDFactory;
import net.sf.javabdd.BDDFactory.BDDOp;
import de.uni_due.inf.ti.raven.data.Data;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDFactory.BDDOperator;

/**
 * This class implements a BDD abstraction layer for the BDD implementation. The concrete BDD implementation and the 
 * BDD library are hidden from Raven through this class so the used BDD library can be changed easily without the 
 * need of changing the "abstract BDDs" used in Raven. 
 * @author Christoph Blume
 * @version $Revision: 1.38 $
 * 
 * @see RavenBDDFactory
 * @see RavenBDDDomain
 * @see RavenBDDPairing
 *
 */
public class RavenBDD implements Cloneable, Data {
	
	/**
	 * Enumeration of the different types of implemented BDDs.
	 * @author Christoph Blume
	 * @version $Revision: 1.38 $
	 *
	 */
	public enum Type implements Data.Type {
		/** BDD Type */
		bdd
	}
	
	/** The Buddy BDD */
	protected BDD bdd;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Converts the (satisfying) variable assignment (seen as byte array) to a n integer value
	 * @param assignment the (satisfying) variable assignment
	 * @return the bit string
	 */
	public static int toInteger(byte[] assignment) {
		int number = 0;

		for(int i=0;i<assignment.length;i++) {
			if(assignment[i]==1) {
				number += 1<<i;
			}
		}
		return number;
	}
	
	/**
	 * Converts the (satisfying) variable assignment (seen as byte array) to a bit string with
	 * least significant bit left.
	 * @param assignment the (satisfying) variable assignment
	 * @return the bit string
	 */
	public static String toBitString(byte[] assignment) {
		StringBuilder bitString = new StringBuilder();
		
		for(int i=0;i<assignment.length;i++) {
			bitString.append(assignment[i]);
		}
		
		return bitString.toString();
	}
	
	/**
	 * Writes the BDD {@code ravenBDD} to the stream {@code zos} and saves it to the zip file 
	 * named {@code name}.
	 * @param ravenBDD the BDD to be saved 
	 * @param zos the stream to the zip file
	 * @param name the name of the zip file entry
	 * @throws IOException if an I/O error has occurred
	 */
	public static void saveBDD(RavenBDD ravenBDD, ZipOutputStream zos, String name)
			throws IOException {		
		
		StringWriter writer = new StringWriter();
		String data = RavenBDDFactory.getFactory().save(writer, ravenBDD.bdd);
		
		try (InputStream is = new ByteArrayInputStream(data.getBytes())) {
			// put a new ZipEntry in the ZipOutputStream 
			zos.putNextEntry(new ZipEntry(name)); 
			
			int size = 0; 
			byte[] buffer = new byte[1024]; 
			
			// read data to the end of the source file and write it to the zip 
			// output stream. 
			while ((size = is.read(buffer, 0, buffer.length)) > 0) { 
				zos.write(buffer, 0, size); 
			} 
			
			zos.closeEntry();
		};
	}
	
	/**
	 * Loads a BDD from the input stream {@code is};
	 * @param is the input stream from which the BDD is read
	 * @return the BDD
	 * @throws IOException if an I/O error has occurred
	 */
	public static RavenBDD loadBDD(InputStream is) throws IOException {
		BDDFactory factory = RavenBDDFactory.getFactory().getBDDFactory();
		BufferedReader reader = new BufferedReader(new InputStreamReader(is));
		
		return new RavenBDD(factory.load(reader));
	}
	
	/**
	 * Converts the Raven BDD Operators to the BuDDy BDD Operators.
	 * @param operator the Raven BDD Operator
	 * @return the corresponding BuDDy BDD Operator
	 */
	private static BDDOp convertOperator(BDDOperator operator) {
		switch (operator) {
		case and:
			return BDDFactory.and;

		case biimp:
			return BDDFactory.biimp;
			
		case imp:
			return BDDFactory.imp;
	
		case or:
			return BDDFactory.or;
	
		case xor:
			return BDDFactory.xor;
			
		default:
			return null;
		}
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new {@code RavenBDD} using the BuDDy-BDD {@code bdd}. 
	 * @param bdd the used BuDDy-BDD {@code bdd}
	 */
	protected RavenBDD(BDD bdd) {
		this.bdd = bdd;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
		
	@Override
	public int hashCode() {
		return bdd.hashCode();
	}

	@Override
	public RavenBDD clone() {
		return new RavenBDD(bdd.id());
	}
	
	/**
	 * Returns a list of satisfying variable assignments (seen as BDDs) for the variables {@code var}. 
	 * @param var the variables for which a satisfying variable assignment is computed
	 * @return the list of satisfying variable assignments
	 */
	public List<RavenBDD> getSatAssignments(final RavenBDD var) {
		List<RavenBDD> list = new LinkedList<>();
		for(BDDIterator it = this.bdd.iterator(var.bdd);it.hasNext();) {
			list.add(new RavenBDD((BDD) it.next()));
		}		
		
		return list;
	}
	
	/**
	 * Frees this BDD.
	 */
	public void free() {
		bdd.free();
	}
	
	@Override
	public boolean equals(Object o) {
		if(o==null) {
			return false;
		}
		else if(o==this) {
			return true;
		}
		else if(o.getClass().equals(getClass())) {
			RavenBDD ravenBDD = (RavenBDD) o;
			return bdd.equals(ravenBDD.bdd);
		}
		else {
			return false;
		}
	}
	
	/**
	 * Returns the variable (in the root) of this BDD.
	 * @return the variable of this BDD
	 */
	public int var() {
		return bdd.var();
	}
	
	/**
	 * Returns the low successor of this BDD.
	 * @return the low successor
	 */
	public RavenBDD low() {
		return new RavenBDD(bdd.low());
	}
	
	/**
	 * Returns the high successor of this BDD.
	 * @return the high successor
	 */
	public RavenBDD high() {
		return new RavenBDD(bdd.high());
	}
	
	/**
	 * Returns a BDD which is the logical <tt>AND</tt> of this BDD and the 
	 * BDD {@code bdd}.
	 * @param bdd the BDD which will be applied to this BDD
	 * @return the resulting BDD
	 */
	public RavenBDD and(RavenBDD bdd) {
		return apply(BDDOperator.and, bdd);
	}

	/**
	 * Returns a BDD which is the logical <tt>BIIMPLICATION</tt> of this BDD and the 
	 * BDD {@code bdd}.
	 * @param bdd the BDD which will be applied to this BDD
	 * @return the resulting BDD
	 */
	public RavenBDD biimp(RavenBDD bdd) {
		return apply(BDDOperator.biimp, bdd);
	}

	/**
	 * Returns a BDD which is the logical <tt>IMPLICATION</tt> of the this BDD and the 
	 * BDD {@code bdd}.
	 * @param bdd the BDD which will be applied to this BDD
	 * @return the resulting BDD
	 */
	public RavenBDD imp(RavenBDD bdd) {
		return apply(BDDOperator.imp, bdd);
	}

	/**
	 * Returns a BDD which is the logical <tt>NOT</tt> of this BDD.
	 * @return the resulting BDD
	 */
	public RavenBDD not() {
		if(this.bdd.isOne()) {
			return RavenBDDFactory.getFactory().zero();
		}
		else if(this.bdd.isZero()) {
			return RavenBDDFactory.getFactory().one();
		}
		else {
			this.bdd = this.bdd.not();
			return new RavenBDD(this.bdd);
		}
	}

	/**
	 * Returns a BDD which is the logical <tt>OR</tt> of this BDD and the 
	 * BDD {@code bdd}.
	 * @param bdd the BDD which will be applied to this BDD
	 * @return the resulting BDD
	 */
	public RavenBDD or(RavenBDD bdd) {
		return apply(BDDOperator.or, bdd);
	}

	/**
	 * Returns a BDD which is the logical <tt>XOR</tt> of this BDD and the 
	 * BDD {@code bdd}.
	 * @param bdd the BDD which will be applied to this BDD
	 * @return the resulting BDD
	 */
	public RavenBDD xor(RavenBDD bdd) {
		return apply(BDDOperator.xor, bdd);
	}
	
	/**
	 * Performs an universal quantification of the BDD nodes
	 * specified by the given BDD.
	 * @param bdd the BDD to specify the BDD nodes to be quantified
	 * @return the resulting BDD
	 */
	public RavenBDD forAll(RavenBDD bdd) {
		return new RavenBDD(this.bdd.forAll(bdd.bdd));
	}

	/**
	 * Applies the operator {@code operator} to this BDD and 
	 * the BDD {@code bdd} and afterwards performs an universal 
	 * quantification of the BDD nodes specified by the BDD 
	 * {@code variables}.
	 * @param bdd the BDD to be applied with this bdd
	 * @param operator the operator to be applied to both bdds
	 * @param variables the BDD to specify the BDD nodes to be quantified
	 * @return the resulting BDD
	 */
	public RavenBDD applyAll(RavenBDD bdd, BDDOperator operator,
			RavenBDD variables) {
		return new RavenBDD(this.bdd.applyAll(bdd.bdd, 
				convertOperator(operator), variables.bdd));
	}
	
	/**
	 * Performs an existential quantification of the BDD nodes
	 * specified by the given BDD.
	 * @param bdd the BDD to specify the BDD nodes to be quantified
	 * @return the resulting BDD
	 */
	public RavenBDD exists(RavenBDD bdd) {
		return new RavenBDD(this.bdd.exist(bdd.bdd));
	}

	/**
	 * Applies the operator {@code operator} to this BDD and 
	 * the BDD {@code bdd} and afterwards performs an existential 
	 * quantification of the BDD nodes specified by the BDD 
	 * {@code variables}.
	 * @param bdd the BDD to be applied with this bdd
	 * @param operator the operator to be applied to both bdds
	 * @param variables the BDD to specify the BDD nodes to be quantified
	 * @return the resulting BDD
	 */
	public RavenBDD applyEx(RavenBDD bdd, BDDOperator operator,
			RavenBDD variables) {
		return new RavenBDD(this.bdd.applyEx(bdd.bdd, 
				convertOperator(operator), variables.bdd));
	}
	
	/**
	 * Restricts the BDD according to the given BDD.
	 * @param bdd the BDD to be used for the restriction
	 * @return the resulting BDD
	 */
	public RavenBDD restrict(RavenBDD bdd) {
		return apply(BDDOperator.restrict, bdd);
	}
	
	/**
	 * Replaces the BDD nodes according to the given BDD pairing.
	 * @param pairing the BDD pairing to be used for the replacement
	 * @return the resulting BDD
	 */
	public RavenBDD replace(RavenBDDPairing pairing) {
		return new RavenBDD(bdd.replace(pairing.pairing));
	}
	
	/**
	 * Returns the One-BDD.
	 * @return the One-BDD
	 */
	public RavenBDD getOne() {
		return new RavenBDD(bdd.getFactory().one());
	}
	
	/**
	 * Returns the Zero-BDD.
	 * @return the Zero-BDD
	 */
	public RavenBDD getZero() {
		return new RavenBDD(bdd.getFactory().zero());
	}
	
	/**
	 * Returns {@code true} if and only if the BDD is the ONE-BDD.
	 * @return {@code true} if the BDD is one, {@code false} otherwise
	 */
	public boolean isOne() {
		return bdd.isOne();
	}

	/**
	 * Returns {@code true} if and only if the BDD is the ZERO-BDD.
	 * @return {@code true} if the BDD is zero, {@code false} otherwise
	 */
	public boolean isZero() {
		return bdd.isZero();
	}

	/**
	 * Returns the number of nodes used to encode this BDD.
	 * @return the number of nodes
	 */
	public int nodeCount() {
		return bdd.nodeCount();
	}
	
	/**
	 * Returns the number of satisfying assignments to this BDD.
	 * @param vars a BDD encoding the fixed variable assignments
	 * @return the number of satisfying assignments
	 */
	public int satCount(RavenBDD vars) {
		return (int) bdd.satCount(vars.bdd);
	}
	
	/**
	 * Returns the number of paths leading to the true terminal.
	 * @return the number of paths to the true terminal
	 */
	public int pathCount() {
		return (int) bdd.pathCount();
	}
	
	/**
	 * Finds all satisfying variable assignments.
	 * @return the list of all satisfying variable assignments
	 */
	@SuppressWarnings("unchecked")
	public List<byte[]> allSat() {
		return bdd.allsat();
	}
	
	/**
	 * Returns the number of times each variable occurs in this BDD.
	 * @return the number of times each variable occurs in this BDD
	 */
	public int[] getVarProfile() {
		return bdd.varProfile();
	}

	/**
	 * Prints this BDD in DOT syntax to the output stream {@code out}.
	 * @param out the output stream to be used as output
	 * 
	 * For further informations about the DOT syntax see GraphViz 
	 * (<a href="http://www.graphviz.org/pdf/dotguide.pdf">Dot Guide</a>).
	 */
	public void printDot(PrintStream out) {
		out.println("digraph G {");
		out.println("0 [shape=box, label=\"0\", style=filled, shape=box, height=0.3, width=0.3];");
		out.println("1 [shape=box, label=\"1\", style=filled, shape=box, height=0.3, width=0.3];");

		boolean[] visited = new boolean[nodeCount()+2];
		visited[0] = true;
		visited[1] = true;
		HashMap<BDD, Integer> map = new HashMap<>();
		map.put(bdd.getFactory().zero(), new Integer(0));
		map.put(bdd.getFactory().one(), new Integer(1));
		printdot_rec(out, 1, visited, map);

		for (BDD bdd: map.keySet()) {
			bdd.free();
		}
		out.println("}");
	}
	
	@Override
	public String toString() {
		return "Raven BDD";
	}
	
	@Override
	public Type getType() {
		return Type.bdd;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Applies the operator {@code op} to this BDD and to the BDD 
	 * {@code bdd}.
	 * @param op the operator to be applied
	 * @param bdd the BDD to be used as second operand
	 * @return the resulting BDD
	 */
	protected RavenBDD apply(BDDOperator op, RavenBDD bdd) {
		switch (op) {
		case and:
			if(this.bdd.isOne()) {
				return bdd;
			}
			else if(this.bdd.isZero()) {
				return this;
			}
			else {
				this.bdd = this.bdd.andWith(bdd.bdd);
				return new RavenBDD(this.bdd);
			}

		case biimp:
			if(this.bdd.isOne()) {
				return bdd;
			}
			else if(this.bdd.isZero()) {
				return bdd.not();
			}
			else {
				this.bdd = this.bdd.biimpWith(bdd.bdd);
				return new RavenBDD(this.bdd);
			}

		case imp:
			if(this.bdd.isOne()) {
				return bdd;
			}
			else if(this.bdd.isZero()) {
				return RavenBDDFactory.getFactory().one();
			}
			else {
				this.bdd = this.bdd.impWith(bdd.bdd);
				return new RavenBDD(this.bdd);
			}

		case or:
			if(this.bdd.isOne()) {
				return this;
			}
			else if(this.bdd.isZero()) {
				return bdd;
			}
			else {
				this.bdd = this.bdd.orWith(bdd.bdd);
				return new RavenBDD(this.bdd);
			}
			
		case restrict:
			this.bdd = this.bdd.restrictWith(bdd.bdd);
			return new RavenBDD(this.bdd);

		case xor: 
			return biimp(bdd).not();

		default:
			return null;
		}
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Prints this BDD in DOT-format to the stream {@code out}. 
	 * @param out the ouput printstream
	 * @param current the current BDD node
	 * @param visited a list of visited BDD nodes
	 * @param map a mapping from BDDs to BDD nodes
	 * @return the visited BDD node
	 */
	private int printdot_rec(PrintStream out, int current, boolean[] visited, 
			HashMap<BDD, Integer> map) {
		Integer ri = map.get(bdd);
		if (ri == null) {
			map.put(bdd.id(), ri = new Integer(++current));
		}
		int r = ri.intValue();
		if (visited[r])
			return current;
		visited[r] = true;

		// TODO: support labelling of vars.
		out.println(r+" [label=\""+bdd.var()+"\"];");

		RavenBDD l = new RavenBDD(bdd.low());
		RavenBDD h = new RavenBDD(bdd.high());
		Integer li = map.get(l.bdd);
		if (li == null) {
			map.put(l.bdd.id(), li = new Integer(++current));
		}
		int low = li.intValue();
		Integer hi = map.get(h.bdd);
		if (hi == null) {
			map.put(h.bdd.id(), hi = new Integer(++current));
		}
		int high = hi.intValue();

		out.println(r+" -> "+low+" [style=dotted];");
		out.println(r+" -> "+high+" [style=filled];");

		current = l.printdot_rec(out, current, visited, map);
		l.bdd.free();
		current = h.printdot_rec(out, current, visited, map);
		h.bdd.free();
		return current;
	}

}
