/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.data.bdd;

import java.math.BigInteger;

import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;
import net.sf.javabdd.BDD;
import net.sf.javabdd.BDDDomain;
import net.sf.javabdd.BDDFactory;

/**
 * This class implements a BDD abstraction layer for the BDD Domain implementation. The concrete BDD Domain 
 * implementation and the BDD library are hidden from Raven through this class so the used BDD library can 
 * be changed easily without the need of changing the "abstract BDD Domains" used in Raven. 
 * @author Christoph Blume
 * @version $Revision: 1.22 $
 * 
 * @see RavenBDDFactory
 * @see RavenBDD
 * @see RavenBDDPairing
 *
 */
public class RavenBDDDomain{

	/** The BuDDy BDD Domain */
	protected final BDDDomain domain;

	private final BDDFactory factory;
	private final RavenBDD domainBDD;

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new BDD Domain with {@code length} bits.
	 * @param length the length of the BDD Domain (in bits)
	 * @return the new BDD Domain
	 */
	public static RavenBDDDomain createDomain(int length) {
		return new RavenBDDDomain(
				RavenBDDFactory.getFactory().factory.extDomain(
						BigInteger.valueOf(2).pow(length)));
	}
	
	/**
	 * Creates {@code amount} new BDD Domains with {@code length} bits.
	 * @param amount the number of new BDD Domains
	 * @param length the length of the BDD Domain (in bits)
	 * @return an array of new BDD Domains
	 * 
	 * <i>Note: The bits of the BDD Domains are interleaved.</i>
	 */
	public static RavenBDDDomain[] createDomains(int amount, int length) {
		RavenBDDDomain[] domains = new RavenBDDDomain[amount];
		
		BigInteger[] domainSizes = new BigInteger[amount];
		for(int i = 0;i<amount;i++) {
			domainSizes[i] = BigInteger.valueOf(2).pow(length);
		}
		
		int i = 0;
		for(BDDDomain bddDomain: RavenBDDFactory.getFactory().factory.extDomain(domainSizes)) {
			domains[i++] = new RavenBDDDomain(bddDomain);
		}
		
		return domains;
	}
	
	
	/**
	 * Returns the One-BDD.
	 * @return the One-BDD
	 */
	public static RavenBDD one() {
		return RavenBDDFactory.getFactory().one();
	}
	
	/**
	 * Returns the Zero-BDD.
	 * @return the Zero-BDD
	 */
	public static RavenBDD zero() {
		return RavenBDDFactory.getFactory().zero();
	}
	
	/**
	 * Returns the integer which is encoded (as bit string) by the BDD {@code bdd}. 
	 * If the bdd is not a singleton set, i.e. the number of satisfying assignments 
	 * is not equal to one, {@code -1} is returned.
	 * @param bdd the singleton set containing the encoded number (as bit string)
	 * @return the integer encoded by the bdd or {@code -1} if the bdd does not 
	 * encode a singleton set
	 */
	public static int getInteger(RavenBDD bdd) {
		if(bdd.pathCount()!=1) {
			return -1;
		}

		StringBuilder builder = new StringBuilder();
		while(!bdd.isOne()) {
			if(!bdd.high().isZero()) {
				builder.append(1);
				bdd = bdd.high();
			}
			else {
				builder.append(0);
				bdd = bdd.low();
			}
		}
		return Integer.valueOf(builder.reverse().toString(), 2);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new BDD domain using the BuDDy BDD Domain {@code domain}.
	 * @param domain the BuDDy BDD domain to be used
	 */
	protected RavenBDDDomain(BDDDomain domain) {
		this.domain = domain;
		factory = domain.getFactory();
		
		BDD bdd = factory.one();
		for(int var: domain.vars()) {
			bdd.andWith(factory.ithVar(var));
		}
		domainBDD = new RavenBDD(bdd);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the number of bits used to define this BDD domain.
	 * @return the length of the BDD domain
	 */
	public int getLength() {
		return domain.varNum();
	}
	
	/**
	 * Returns a BDD encoding all variables used by this BDD Domain
	 * @return the domain BDD
	 */
	public RavenBDD getDomainBDD() {
		return domainBDD.clone();
	}
	
	/**
	 * Return all variables of this domain.
	 * @return the variables defined in this domain
	 */
	public int[] getVars(){
		return domain.vars();
	}
	
	/**
     * Returns a BDD representing the {@code i}-th variable of the BDD-Domain. 
     * @param i index of the BDD-variable
     * @return the BDD-variable
     */
	public RavenBDD getIthBDD(int i) {
		return new RavenBDD(getIthBit(i));
	}

	/**
	 * Returns a BDD representing a (non-negative) integer value {@code number} as binary encoding 
	 * of length {@code bitLength}. 
	 * @param number number) to be represented
	 * @param bitLength length of the binary encoding
	 * @return the BDD
	 */
	public RavenBDD getBDDForInteger(int number, int bitLength) {
		// If number is negative throw an exception
		if(number<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.NUMBER_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		// If number is zero, build a BDD where all variables are set to zero, i.e. the only the 
		// low-edge leads to the terminal ONE.
		else if (number==0) {
			BDD bdd = factory.one();
			for (int i = bitLength - 1; i >= 0; i--) {
				bdd = bdd.andWith(getIthBit(i).not());
			}
			return new RavenBDD(bdd);
		}
		// In every other case
		else {
			BDD bdd = factory.one();
			int pos = 0;
			while (number > 0) {
				// Compute the remainder (modulo 2) and add the positive or negative BDD variable
				// for the current bit depending on the value of the remainder
				// Afterwards shift the number one to the right
				int rem = number % 2;
				number = number >> 1;
				if (rem == 0) {
					bdd = bdd.andWith(getIthBit(pos).not());
				}
				else {
					bdd = bdd.andWith(getIthBit(pos));
				}
				pos += 1;
			}

			// If needed insert "leading zeros" to "fill" the BDD to the desired bit length
			for (int i = pos; i < bitLength; i++) {
				bdd = bdd.andWith(getIthBit(i).not());
			}
			return new RavenBDD(bdd);
		}
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns a BDD representing the {@code i}-th bit of this BDD domain.
	 * @param i the index of the bit
	 * @return the BDD representing the {@code i}-th bit
	 */
	private BDD getIthBit(int i) {
		return factory.buildVector(domain).getBit(i).id();
	}
	
}
