/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.data.bdd;

import java.io.StringWriter;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import net.sf.javabdd.BDD;
import net.sf.javabdd.BDDDomain;
import net.sf.javabdd.BDDFactory;

/**
 * This class implements a BDD abstraction layer for the BDD Factory implementation. The concrete BDD Factory 
 * implementation and the BDD library are hidden from Raven through this class so the used BDD library can be 
 * changed easily without the need of changing the "abstract BDD Factory" used in Raven. 
 * @author Christoph Blume
 * @version $Revision: 1.24 $
 * 
 * @see RavenBDD
 * @see RavenBDDDomain
 * @see RavenBDDPairing
 *
 */
public class RavenBDDFactory {

	/**
	 * Enumeration of the different BDD operations.
	 * @author Christoph Blume
	 * @version $Revision: 1.24 $
	 *
	 */
	public enum BDDOperator {
		/** AND-Operator */
		and,
		
		/** BIIMPLICATION-Operator */
		biimp,
		
		/** IMPLICATION-Operator */
		imp,
		
		/** OR-Operator */
		or,
		
		/** EXCLUSIVE OR-Operator */
		xor,
		
		/** RESTRICTION-Operator */
		restrict
	}
	
	private static RavenBDDFactory ravenFactory;

	/** The underlying BuDDy BDD Factory */
	protected BDDFactory factory;

	private Logger logger;

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the BDD factory.
	 * @return the BDD factory
	 */
	public static RavenBDDFactory getFactory() {
		if(ravenFactory==null) {
			ravenFactory = new RavenBDDFactory(50000,5000);
		}

		return ravenFactory;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates an instance of the JavaBDD factory using the BuDDy library.
	 * @param nodenum the number of initially allocated BDD nodes 
	 * @param cachesize the initial size of the node cache
	 */
	private RavenBDDFactory(int nodenum, int cachesize) {
		logger = Logger.getLogger(getClass().getName());
		
		factory = BDDFactory.init(nodenum, cachesize);
		try {
			BuddyStatsPrinter printer = BuddyStatsPrinter.getPrinter();
			factory.registerGCCallback(printer, 
					BuddyStatsPrinter.class.getMethod("print", Object.class));
			factory.registerResizeCallback(printer, 
					BuddyStatsPrinter.class.getMethod("print", Object.class));
		} catch (SecurityException e) {
			logger.throwing(getClass().getName(), "RavenBDDFactory", e);
		} catch (NoSuchMethodException e) {
			logger.throwing(getClass().getName(), "RavenBDDFactory", e);
		}

	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Saves the BDD to a string.
	 * @param out the {@code StringWriter} to be used
	 * @param root the BDD to be saved
	 * @return the string representation of the BDD
	 */
	public String save(StringWriter out, BDD root) {
        if (root.isOne() || root.isZero()) {
            out.write("0 0 " + (root.isOne()?1:0) + "\n");
            return out.toString();
        }

        out.write(root.nodeCount() + " " + varNum(root) + "\n");

        for (int x = 0; x < varNum(root); x++)
            out.write(var2Level(x,root) + " ");
        out.write("\n");

        Map<BDD, Integer> visited = new HashMap<>();
        save_rec(out, visited, root);
        
        for (Iterator<BDD> it = visited.keySet().iterator(); it.hasNext(); ) {
            BDD b = it.next();
            if (b != root) b.free();
        }
        return out.toString();
    }
    
    /**
     * <p>Returns the number of defined variables in the BDD {@code bdd}.</p>
     * <p>Compare to bdd_varnum.</p>
     * @param bdd the BDD for which the number of variables is returned
     * @return  the number of defined variables
     */
    public int varNum(BDD bdd){
    	return factory.varNum();
    }
    
    /**
     * <p>Convert from a BDD variable to a BDD level.</p>
     * 
     * <p>Compare to bdd_var2level.</p>
     * @param var the BDD variable
     * @param bdd the BDD for which the variable is converted
     * @return the BDD level
     */
    public int var2Level(int var, BDD bdd){
    	return factory.var2Level(var);
	}

	
	/**
	 * Creates a list of {@code amount} finite BDD domains which each represents 
	 * {@code size} different values.
	 * @param amount the number of Finite BDD domains to be created
	 * @param size the size of each Finite BDD Domain
	 * @return a list of BDD domains
	 */
	public List<RavenBDDDomain> createDomains(int amount, BigInteger size) {
		//Log entering
		logger.entering(getClass().getName(), "createDomains", 
				new Object[]{size, amount});

		BigInteger[] sizes = new BigInteger[amount];
		Arrays.fill(sizes, size);
		ArrayList<RavenBDDDomain> list = new ArrayList<>();
		for(BDDDomain domain: factory.extDomain(sizes)) {
			list.add(new RavenBDDDomain(domain));
		}

		//Log exiting
		logger.exiting(getClass().getName(), "createDomains", list);

		return list;
	}
	
	/**
	 * Returns the One-BDD.
	 * @return the One-BDD
	 */
	public RavenBDD one() {
		return new RavenBDD(factory.one());
	}
	
	/**
	 * Returns the Zero-BDD.
	 * @return the Zero-BDD
	 */
	public RavenBDD zero() {
		return new RavenBDD(factory.zero());
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/


    /**
     * Helper function for save().
     * @param out the {@code StringWriter} to which the string representation of the BDD is written
     * @param visited a mapping of visited BDDs to BDD indices
     * @param root the BDD to be saved
     * @return the node index of the saved BDD
     */
    protected int save_rec(StringWriter out, Map<BDD,Integer> visited, BDD root) {
        if (root.isZero()) {
            root.free();
            return 0;
        }
        if (root.isOne()) {
            root.free();
            return 1;
        }
        Integer i = visited.get(root);
        if (i != null) {
            root.free();
            return i.intValue();
        }
        int v = visited.size() + 2;
        visited.put(root, new Integer(v));
        
        BDD l = root.low();
        int lo = save_rec(out, visited, l);
        
        BDD h = root.high();
        int hi = save_rec(out, visited, h);

        out.write(v + " ");
        out.write(root.var() + " ");
        out.write(lo + " ");
        out.write(hi + "\n");
        
        return v;
    }
	
	/**
	 * For saving purposes of BDDs only.
	 * @return the BuDDy BDD factory
	 */
	protected BDDFactory getBDDFactory(){
		return factory;
	}
	
}
