/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.data.bdd;

import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;
import net.sf.javabdd.BDDPairing;

/**
 * This class implements a BDD abstraction layer for the BDD Pairing implementation. The concrete BDD Pairing 
 * implementation and the BDD library are hidden from Raven through this class so the used BDD library can be 
 * changed easily without the need of changing the "abstract BDD Pairings" used in Raven. 
 * @author Christoph Blume
 * @version $Revision: 1.12 $
 * 
 * @see RavenBDDFactory
 * @see RavenBDD
 * @see RavenBDDDomain
 *
 */
public class RavenBDDPairing {

	private static RavenBDDFactory ravenFactory;
	
	/** The BuDDy BDD Pairing */
	protected BDDPairing pairing;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new BDD pairing.
	 * @return the new BDD pairing
	 */
	public static RavenBDDPairing makePair() {
		if(ravenFactory==null) {
			ravenFactory = RavenBDDFactory.getFactory();
		}
		
		return new RavenBDDPairing();
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new BDD pairing.
	 */
	private RavenBDDPairing() {
		pairing = ravenFactory.factory.makePair();
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Adds two BDDs as a pair to this BDD pairing.
	 * @param first the first BDD
	 * @param second the second BDD
	 */
	public void set(RavenBDD first, RavenBDD second) {
		if(first==null) {
			throw new NullPointerException(RavenExceptionMessages.BDD_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(second==null) {
			throw new NullPointerException(RavenExceptionMessages.BDD_MUST_NOT_BE_NULL_EXCEPTION);
		}

		pairing.set(first.bdd.var(), second.bdd.var());
	}
	
	/**
	 * Add the first int[] and the second int[] to this pairing.
	 * @param first the first int[]
	 * @param second the second int[]
	 */
	public void set(int[] first, int[] second) {
		pairing.set(first, second);
	}
	
	
	/**
	 * Adds two BDD domains as a pair to this BDD pairing.
	 * @param first the first BDD domain
	 * @param second the second BDD domain
	 */
	public void set(RavenBDDDomain first, RavenBDDDomain second) {
		if(first==null) {
			throw new NullPointerException(
					RavenExceptionMessages.BDD_DOMAIN_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(second==null) {
			throw new NullPointerException(
					RavenExceptionMessages.BDD_DOMAIN_MUST_NOT_BE_NULL_EXCEPTION);
		}

		pairing.set(first.domain.vars(), second.domain.vars());
	}
	
}
