/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.encodings;

import de.uni_due.inf.ti.raven.data.automata.PathAutomaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDDomain;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class provides a concrete implementation of an encoding for <i>path automata</i>.
 * The state encoding is as follows:
 * 
 * <i>b<sub>1</sub> ... b<sub>l</sub> (m<sub>0,0</sub> m<sub>0,1</sub> m<sub>1,1</sub> 
 * (m<sub>0,2</sub> ... m<sub>2,2</sub>) ... (m<sub>0,n-1</sub> 
 * ... m<sub>n-1,n-1</sub>)</i>, where <i>l = log<sub>2</sub>(n)</i> is the binary logarithm of the 
 * maximum interface size <i>n</i>. The block <i>b<sub>1</sub> ... b<sub>l</sub></i> encodes the 
 * interface size to which the current state belongs, the block <i>m<sub>0,j</sub> ... 
 * m<sub>j,j</sub></i> the adjacency bits of the <i>j</i>-th interface node in the current state.
 * @author Christoph Blume, Dennis Nolte
 * @version $Revision: 1.5 $
 *
 *	@see PathAutomaton
 */
public class AnotherPathEncoding extends PathEncoding {

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Constructs a new encoding for a <i>path automaton</i> with a maximum interface 
	 * size of {@code maximumInterface}.
	 * @param maximumInterface the maximum interface size
	 * @param innerInterface the size of the inner interface
	 * @param outerInterface the size of the outer interface
	 */
	protected AnotherPathEncoding(int maximumInterface, int innerInterface, int outerInterface) {
		super(maximumInterface, innerInterface, outerInterface, (maximumInterface*(maximumInterface+1))/2);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public RavenBDD getPathBit(StateType type, int source, int target) {
		if(source<0 || source>=maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}
		if(target<0 || target>=maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_BIT_POSITION_EXCEPTION);
		}
		if(target > source)
		{
			return getPathBit(type,target,source);
		}

		RavenBDDDomain domain = null;
		if(type==StateType.current) {
			domain = curStateDomain;
		}
		else {
			domain = sucStateDomain;
		}
		return domain.getIthBDD(interfaceLength + (source*(source+1))/2 + target);

	}

	@Override
	public RavenBDD getSourcePathBit(StateType type, int node) {
		if(node<0 || node>=maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_BIT_POSITION_EXCEPTION);
		}

		RavenBDDDomain domain = null;
		if(type==StateType.current) {
			domain = curStateDomain;
		}
		else {
			domain = sucStateDomain;
		}
		return domain.getIthBDD(interfaceLength + (maximumInterface*(maximumInterface+1)/2) + node);
	}

	@Override
	public String getBitType(int pos) {
		if(pos<0 || pos>=length) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_BIT_POSITION_EXCEPTION);
		}

		if(pos<interfaceLength) {
			return getInterfaceSizeType();
		}
		
		if(pos<interfaceLength+(maximumInterface*(maximumInterface+1))/2) {
			
			int from = 0;
			int to = 0;
			int previousSum =0;
			int nextSum = 0;
			for(int node=0;node<maximumInterface;node++)
			{
				previousSum = nextSum;
				nextSum = ((node+1)*(node+2))/2;
				if((pos-interfaceLength) < nextSum)
				{
					from = node;
					to = (pos-interfaceLength)-previousSum;
					node = maximumInterface;
				}
			}
			
			return getPathNodeType(from, to);
		}

		return getSourcePathNodeType(pos-interfaceLength-
				((maximumInterface*(maximumInterface+1))/2));
	}

	@Override
	public String printStates(RavenBDD stateSet) {
		// TODO Auto-generated method stub
		return null;
	}

}
