/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.encodings;

import java.math.BigInteger;
import java.util.List;

import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDDomain;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDFactory;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class represents the encoding of state space and the transition functions of some automaton. 
 * @author Christoph Blume
 * @version $Revision: 1.35 $
 * 
 * @see Automaton
 */
public abstract class BDDEncoding{
	
	/** The bit type indicating bits which encode undefined information */
	protected static final String UNDEFINED_BIT_TYPE = "undefined";

	/** The bit type indicating bits which encode interface size information */
	protected static final String INTERFACE_SIZE_BIT_TYPE = "interfaceBit";
	
	/** The bit type indicating bits which encode state information */
	protected static final String STATE_BIT_TYPE = "stateBit";
	
	/**
	 * Enumeration of state types to distinguish between bits used for different states 
	 * (e.g. in relations on the state set)
	 * @author Christoph Blume
	 * @version $Revision: 1.35 $
	 *
	 */
	public enum StateType {
		/** The Current State Type */
		current,
		
		/** The Successor State Type */
		successor
	}
	
	/** The length of the BDD encoding */
	protected final int length;
	
	/** The maximum interface size */
	protected final int maximumInterface;
	
	/** The inner interface size */
	protected final int innerInterface;
	
	/** The outer interface size */
	protected final int outerInterface;
	
	/** The bit length to encode the interface size */
	protected final int interfaceLength;
	
	/** The bit length to encode the states */
	protected final int stateLength;

	/** Current State BDD Domain */
	transient protected final RavenBDDDomain curStateDomain;
	
	/** Successor State BDD Domain */
	transient protected final RavenBDDDomain sucStateDomain;
	
	/** The BDD Factory */
	transient protected final RavenBDDFactory factory;

		
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the bit length of {@code number}.
	 * @param number the input number
	 * @return the bit length
	 */
	protected static final int getBitLength(int number) {
		return (int) Math.ceil(Math.log(number+1)/Math.log(2));
	}
	
	/**
	 * Returns the type string indicating the bits for the interface size.
	 * @return the interface size type string
	 */
	public static final String getInterfaceSizeType() {
		return INTERFACE_SIZE_BIT_TYPE;
	}
	
	/**
	 * Returns the type string indicating the bits for the state encoding.
	 * @return the state type string
	 */
	public static final String getStateType() {
		return STATE_BIT_TYPE;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new BDD encoding of length {@code length} encoding 
	 * state sets up to a maximum interface size of {@code maximumInterface}.
	 * @param length the length of the BDD encoding
	 * @param maximumInterface the maximum interface size
	 * @param innerInterface the inner interface size
	 * @param outerInterface the outer interface size
	 */
	protected BDDEncoding(int length, int maximumInterface, int innerInterface, int outerInterface) {
		if(length<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ENCODING_LENGTH_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(maximumInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(outerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(outerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		
		this.length = length;
		this.maximumInterface = maximumInterface;
		interfaceLength = getBitLength(maximumInterface);
		stateLength = length - interfaceLength;
		this.innerInterface = innerInterface;
		this.outerInterface = outerInterface;
		
		factory = RavenBDDFactory.getFactory();
		BigInteger size = BigInteger.valueOf(2).pow(length);
		List<RavenBDDDomain> list = factory.createDomains(2, size);
		curStateDomain = list.get(0);
		sucStateDomain = list.get(1);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Abstract Methods																*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns a string indicating which information is encoded by bit of the given position 
	 * {@code pos}. 
	 * @param pos the position of the bit
	 * @return the type of information encoded by the specified bit
	 */
	public abstract String getBitType(int pos);
	
	/**
	 * Prints a single state (seen as singleton state set) or a set of states.
	 * @param stateSet the state set to be printed
	 * @return the string representation of the state set
	 */
	public abstract String printStates(RavenBDD stateSet);
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the bit length of this BDD encoding
	 * @return the length of the encoding
	 */
	public final int getLength() {
		return length;
	}

	/**
	 * Returns a BDD which encodes in state {@code state} the {@code pos}-th bit the encoding. 
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param pos the position of the bit
	 * @return a BDD encoding a bit of the state encoding
	 */
	public final RavenBDD getBit(StateType type, int pos) {
		if(pos<0 || pos>=length) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_BIT_POSITION_EXCEPTION);
		}
		
		RavenBDDDomain domain;
		if(type==StateType.current) {
			domain = curStateDomain;
		}
		else {
			domain = sucStateDomain;
		}
		return domain.getIthBDD(pos);
	}
	
	/**
	 * Returns the maximum interface size.
	 * @return the maximum interface size
	 */
	public final int getMaximumInterface() {
		return maximumInterface;
	}
	
	/**
	 * Returns the inner interface size
	 * @return the inner interface size
	 */
	public final int getInnerInterface(){
		return innerInterface;
	}
	
	/**
	 * Returns the outer interface size
	 * @return the outer interface size
	 */
	public final int getOuterInterface(){
		return outerInterface;
	}

	/**
	 * Returns the bit length of the interface encoding.
	 * @return the length of the interface encoding
	 * 
	 * @see #getInterfaceBit(StateType, int)
	 */
	public final int getInterfaceLength() {
		return interfaceLength;
	}
	
	/**
	 * Returns the bit length of the state encoding.
	 * @return the length of the state encoding
	 * 
	 * @see #getStateBit(StateType, int)
	 */
	public final int getStateLength() {
		return stateLength;
	}
	
	/**
	 * Returns a BDD which encodes the {@code pos}-th interface bit for 
	 * the state type {@code type}.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param pos the index of the interface bit
	 * @return a BDD encoding the interface bit
	 * 
	 * @see #getInterfaceLength()
	 */
	public final RavenBDD getInterfaceBit(StateType type, int pos) {
		if(pos<0 || pos>=interfaceLength) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_BIT_POSITION_EXCEPTION);
		}

		return getBit(type, pos);
	}
	
	/**
	 * Returns a BDD which encodes the {@code pos}-th state bit for 
	 * the state type {@code type}.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param pos the index of the interface bit
	 * @return a BDD encoding the state bit
	 */
	public final RavenBDD getStateBit(StateType type, int pos) {
		if(pos<0 || pos>=stateLength) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_BIT_POSITION_EXCEPTION);
		}

		return getBit(type, interfaceLength+pos);
	}
	
	/**
	 * Returns a BDD which encodes the interface size {@code size} for 
	 * the state {@code state}.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param size the size of the interface
	 * @return a BDD encoding the interface size
	 * 
	 * @see #getInterfaceSizeIsGreaterOrEqual(StateType, int)
	 * @see #getInterfaceSizeIsLessOrEqual(StateType, int)
	 * @see BDDEncoding.StateType
	 */
	public final RavenBDD getInterfaceSize(StateType type, int size) {
		if(size<0 || size>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}

		RavenBDDDomain domain;
		if(type==StateType.current) {
			domain = curStateDomain;
		}
		else {
			domain = sucStateDomain;
		}
		return domain.getBDDForInteger(size, interfaceLength);
	}
	
	/**
	 * Returns a BDD encoding that in the state {@code state} the size of the 
	 * interface is less or equal to {@code size} if and only if the BDD is {@code true}.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param size the maximum size of the interface
	 * @return a BDD encoding that the interface is less or equal to {@code size}
	 * 
	 * @see #getInterfaceSizeIsGreaterOrEqual(StateType, int)
	 * @see #getInterfaceSize(StateType, int)
	 */
	public final RavenBDD getInterfaceSizeIsLessOrEqual(StateType type, int size) {
		if(size<0 || size>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}

		RavenBDD bdd = zero();
		for(int i=0;i<size;i++) {
			bdd = bdd.or(getInterfaceSize(type, i));
		}

		return bdd;
	}
	
	/**
	 * Returns a BDD encoding that in the state type {@code type} the size of the 
	 * interface is greater or equal to {@code size} if and only if the BDD is {@code true}.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param size the minimum size of the interface
	 * @return a BDD encoding that the interface is greater or equal to {@code size}
	 * 
	 * @see #getInterfaceSizeIsLessOrEqual(StateType, int)
	 * @see #getInterfaceSize(StateType, int)
	 */
	public final RavenBDD getInterfaceSizeIsGreaterOrEqual(StateType type, int size) {
		if(size<0 || size>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_SIZE_EXCEPTION);
		}

		RavenBDD bdd = getInterfaceSizeIsLessOrEqual(type, size).not();
		bdd = bdd.or(getInterfaceSize(type, size));

		return bdd;
	}
	
	/**
	 * Returns the {@code One}-BDD for this encoding.
	 * @return the {@code One}-BDD
	 * 
	 * @see #zero()
	 */
	public final RavenBDD one() {
		return factory.one();
	}
	
	/**
	 * Returns the {@code Zero}-BDD for this encoding.
	 * @return the {@code Zero}-BDD
	 * 
	 * @see #one()
	 */
	public final RavenBDD zero() {
		return factory.zero();
	}
	
	/**
	 * Returns the BDD domain for the (current) states 
	 * (of the transition relations)
	 * @return the (current) state BDD domain
	 * 
	 * @see #getSuccessorStateDomain()
	 */
	public final RavenBDDDomain getCurrentStateDomain() {
		return curStateDomain;
	}
	
	/**
	 * Returns the BDD domain for the successor states
	 * of the transition relations
	 * @return the successor state BDD domain
	 * 
	 * @see #getCurrentStateDomain()
	 */
	public final RavenBDDDomain getSuccessorStateDomain() {
		return sucStateDomain;
	}
	
	/**
	 * Returns a BDD encoding the bits of the interface 
	 * in the encoding for the current state.
	 * @return the bits if the interface encoding
	 * 
	 * @see #getStateEncodingBDD()
	 */
	public final RavenBDD getInterfaceEncodingBDD () {
		RavenBDD quantifierBDD = one();
		for(int i=0;i<getInterfaceLength();i++) {
			quantifierBDD = quantifierBDD.and(getBit(StateType.current, i));
		}
		
		return quantifierBDD;
	}
	
	/**
	 * Returns a BDD encoding the bits of the state information 
	 * in the encoding for the current state.
	 * @return the bits if the state information encoding
	 * 
	 * @see #getInterfaceEncodingBDD()
	 */
	public final RavenBDD getStateEncodingBDD () {
		RavenBDD quantifierBDD = one();
		for(int i=getInterfaceLength();i<getLength();i++) {
			quantifierBDD = quantifierBDD.and(getBit(StateType.current, i));
		}
		
		return quantifierBDD;
	}
	
	@Override
	public final String toString() {
		return getClass().getSimpleName().toString();
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the interface size of the state encoded by the BDD {@code bdd}. If the BDD does 
	 * not encode a singleton set {@code -1} will be returned.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param bdd the BDD encoding the state (seen as singleton set)
	 * @return the interface size of the state or {@code -1} if the BDD does not encode a 
	 * singleton state
	 */
	protected final int getInterfaceSize(StateType type, RavenBDD bdd) {
		RavenBDD vars;
		RavenBDDDomain domain = type==StateType.current ? curStateDomain : sucStateDomain;
		int interfaceSize = -1;

		vars = domain.getDomainBDD();
		if(bdd.satCount(vars)==1) {
			bdd = bdd.exists(getStateEncodingBDD());
			interfaceSize = RavenBDDDomain.getInteger(bdd);
		}
		return interfaceSize;
	}
	
}
