/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.encodings;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.data.automata.MaximumEdgeAutomaton;
import de.uni_due.inf.ti.raven.data.automata.MaximumVertexAutomaton;
import de.uni_due.inf.ti.raven.data.automata.MinimumEdgeAutomaton;
import de.uni_due.inf.ti.raven.data.automata.MinimumVertexAutomaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDDomain;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class represents the encoding of state space and the transition functions of 
 * minimum/maximum vertex/edge automata and provides methods to access the encoding.
 * 
 * <p><i>Note:</i> The concrete implementation of a minimum/maximum vertex/edge encoding
 * can be found in classes extending this abstract class.</p>
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.10 $
 *
 * @see DefaultMaximumBoundEncoding
 * @see DefaultMinimumBoundEncoding
 * @see MaximumEdgeAutomaton
 * @see MaximumVertexAutomaton
 * @see MinimumEdgeAutomaton
 * @see MinimumVertexAutomaton
 */
public abstract class BoundEncoding extends BDDEncoding {
	
	/**
	 * Enumeration of the different types of bound encodings.
	 * @author Christoph Blume, Laura Steinert
	 * @version $Revision: 1.10 $
	 */
	public static enum Type {
		/** The maximum bound type */
		MAXIMUM,
		
		/** The minimum bound type */
		MINIMUM;
	}
	
	private static final String BOUND_BIT_TYPE = "BoundType";
	
	private static final Logger logger = Logger.getLogger(
			BoundEncoding.class.getName());

	private final int bound;
	
	/** The number of bits needed to encode the bound */
	protected final int boundBitLength;

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the type string indicating the bound bits.
	 * @return the bound type string
	 */
	public static final String getBoundType() {
		return BOUND_BIT_TYPE;
	}
	
	/**
	 * 
	 * @param maximumInterface the maximum interface size
	 * @param innerInterface size of the inner interface
	 * @param outerInterface size of the outer interface
	 * @param bound the bound on the number of vertices/edges
	 * @return the default minimum encoding
	 */
	public static final BoundEncoding getDefaultMinimumEncoding(int maximumInterface, 
			int innerInterface, int outerInterface, int bound) {
		if(maximumInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(outerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(outerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(bound<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.BOUND_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}

		return new DefaultMinimumBoundEncoding(maximumInterface, innerInterface, outerInterface, 
				bound);
	}
	
	/**
	 * 
	 * @param maximumInterface the maximum interface size
	 * @param innerInterface size of the inner interface
	 * @param outerInterface size of the outer interface
	 * @param bound the bound on the number of vertices/edges
	 * @return the default maximum encoding
	 */
	public static final BoundEncoding getDefaultMaximumEncoding(int maximumInterface, int innerInterface, 
			int outerInterface, int bound) {
		if(maximumInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(outerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(outerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(bound<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.BOUND_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}

		return new DefaultMaximumBoundEncoding(maximumInterface, innerInterface, outerInterface, 
				bound);
	}
	
	/**
	 * Returns the BDD encoding for a bound instance given by the class name
	 * {@code name} with a maximum interface size of {@code maximumInterface} and a 
	 * modulo number of {@code modNumber}. If no vertex count encoding with the 
	 * given name can be found the default encoding is returned.
	 * @param maximumInterface size of the maximum interface
	 * @param innerInterface size of the inner interface
	 * @param outerInterface size of the outer interface
	 * @param bound the bound on the number of vertices/edges
	 * @param type minimum or maximum encoding
	 * @param encodingName the class name of the encoding class
	 * @return the encoding given by the class name or the default encoding if no 
	 * bound encoding class with the given name can be found 
	 */
	public static final BoundEncoding getEncoding(int maximumInterface, int innerInterface, 
			int outerInterface, int bound, Type type, String encodingName) {
		if(maximumInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(outerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(outerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(bound<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.BOUND_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(encodingName==null) {
			throw new NullPointerException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(encodingName.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		
		//Log entering 
		logger.entering(BoundEncoding.class.getName(), "getEncoding", 
				new Object[]{maximumInterface, innerInterface, outerInterface, bound, type, encodingName});

		BoundEncoding encoding = null;
		Class<?> encodingClass;
		try {
			//Create new vertex count encoding instance
			encodingClass = Class.forName(encodingName);
			Constructor<?> constructor = encodingClass.getDeclaredConstructor(int.class, int.class);
			encoding = (BoundEncoding) constructor.newInstance(maximumInterface, bound);
		} catch (ClassNotFoundException e) {
			//Log exception
			logger.throwing(BoundEncoding.class.getName(), "getEncoding", e);
		} catch (SecurityException e) {
			//Log exception
			logger.throwing(BoundEncoding.class.getName(), "getEncoding", e);
		} catch (NoSuchMethodException e) {
			//Log exception
			logger.throwing(BoundEncoding.class.getName(), "getEncoding", e);
		} catch (IllegalArgumentException e) {
			//Log exception
			logger.throwing(BoundEncoding.class.getName(), "getEncoding", e);
		} catch (InstantiationException e) {
			//Log exception
			logger.throwing(BoundEncoding.class.getName(), "getEncoding", e);
		} catch (IllegalAccessException e) {
			//Log exception
			logger.throwing(BoundEncoding.class.getName(), "getEncoding", e);
		} catch (InvocationTargetException e) {
			//Log exception
			logger.throwing(BoundEncoding.class.getName(), "getEncoding", e);
		}
		finally {
			//If no instance has been created fall back to the default encoding
			if(encoding==null) {
				if (type == Type.MINIMUM){
					encoding = getDefaultMinimumEncoding(maximumInterface, innerInterface, 
							outerInterface, bound);
				}
				else if (type == Type.MAXIMUM){
					encoding = getDefaultMaximumEncoding(maximumInterface, innerInterface, 
							outerInterface, bound);
				}
			}
		}

		//Log exiting
		logger.exiting(BoundEncoding.class.getName(), "getEncoding", encoding);		
		return encoding;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Constructs a new encoding for a <i>minimum/maximum vertex/edge automaton</i> with a maximum 
	 * interface size of {@code maximumInterface} and a bound of {@code bound}.
	 * @param maximumInterface size of the maximum interface
	 * @param innerInterface size of the inner interface
	 * @param outerInterface size of the outer interface
	 * @param bound the bound on the number of vertices/edges
	 */
	protected BoundEncoding(int maximumInterface, int innerInterface, int outerInterface, int bound) {
		super(getBitLength(maximumInterface) + getBitLength(bound), maximumInterface, 
				innerInterface, outerInterface);
		this.bound = bound;
		this.boundBitLength = getBitLength(bound);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Abstract Methods																*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns a BDD which encodes the {@code pos}-th bit of the value of the node/edge counter.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param pos the position of the value bit
	 * @return a BDD encoding the current value of the node/edge counter
	 */
	public abstract RavenBDD getValueBit(StateType type, int pos);

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the length of the encoding of the bound
	 * @return length of the encoding
	 */
	public final int getBoundBitLength(){
		return boundBitLength;
	}

	/**
	 * Returns the bound.
	 * @return the bound
	 */
	public int getBound(){
		return bound;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the current value of the node/edge counter in the state encoded by the BDD 
	 * {@code state}.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param state the BDD encoding the state
	 * @return the current value of the node/edge counter
	 */
	protected final int getValue(StateType type, RavenBDD state) {
		RavenBDD vars;
		RavenBDDDomain domain = type==StateType.current ? curStateDomain : sucStateDomain;
		int value = -1;
		
		vars = domain.getDomainBDD();
		if(state.satCount(vars)==1) {
			state = state.exists(getInterfaceEncodingBDD());
			
			for(int i=0;i<boundBitLength;i++) {
				state = state.exists(getValueBit(type, i));
			}
			
			value = RavenBDDDomain.getInteger(state);
		}
		return value;
	}
	
}
