/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.encodings;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.data.automata.ColorabilityAutomaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDDomain;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class represents the encoding of state space and the transition functions of colorability 
 * automata and provides methods to access the encoding.
 * 
 * <p><i>Note:</i> The concrete implementation of a colorability encoding can be found in classes 
 * extending this abstract class.</p>
 * @author Christoph Blume
 * @version $Revision: 1.29 $
 * 
 * @see DefaultColorabilityEncoding
 * @see ColorabilityAutomaton
 */
public abstract class ColorabilityEncoding extends BDDEncoding {
	
	/** */
	protected static final String NODE_BIT_TYPE = "NodeType";
	
	private static final Logger logger = Logger.getLogger(
			ColorabilityEncoding.class.getName());
	
	/** The number of colors */
	protected final int colors;
	
	/** The bit length to encode the number of colors */
	protected final int colorBitLength;


	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the default BDD encoding for a colorability instance with a maximum interface size 
	 * of {@code maximumInterface} and a number of colors of {@code colors}.
	 * @param maximumInterface the maximum interface size
	 * @param innerInterface size of the inner interface
	 * @param outerInterface size of the outer interface
	 * @param colors the number of colors
	 * @return the default encoding
	 */
	public static final ColorabilityEncoding getDefaultEncoding(int maximumInterface, 
			int innerInterface, int outerInterface, int colors) {
		if(maximumInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(outerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(outerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(colors<1) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.NUMBER_OF_COLORS_MUST_BE_POSITIVE_EXCEPTION);
		}

		return new DefaultColorabilityEncoding(maximumInterface, innerInterface, outerInterface, colors);
	}
	
	/**
	 * Returns the BDD encoding for a colorability instance given by the class name
	 * {@code name} with a maximum interface size of {@code maximumInterface} and a 
	 * number of colors of {@code colors}. If no colorability encoding with the 
	 * given name can be found the default encoding is returned.
	 * @param maximumInterface the maximum interface size
	 * @param innerInterface size of the inner interface
	 * @param outerInterface size of the outer interface
	 * @param colors the number of colors
	 * @param encodingName the class name of the dominating set encoding class
	 * @return the encoding given by the class name or the default encoding if no 
	 * colorability encoding class with the given can be found 
	 */
	public static final ColorabilityEncoding getEncoding(int maximumInterface, int innerInterface, 
			int outerInterface, int colors, String encodingName) {
		if(maximumInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(outerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(outerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(colors<1) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.NUMBER_OF_COLORS_MUST_BE_POSITIVE_EXCEPTION);
		}
		if(encodingName==null) {
			throw new NullPointerException(
					RavenExceptionMessages.ENCODING_NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(encodingName.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ENCODING_NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		
		//Log entering 
		logger.entering(ColorabilityEncoding.class.getName(), "getEncoding", 
				new Object[]{maximumInterface, innerInterface, outerInterface, colors, encodingName});

		ColorabilityEncoding encoding = null;
		Class<?> encodingClass;
		try {
			//Create new dominating set encoding instance
			encodingClass = Class.forName(encodingName);
			Constructor<?> constructor = encodingClass.getDeclaredConstructor(int.class, int.class, 
					int.class, int.class);
			encoding = (ColorabilityEncoding) constructor.newInstance(maximumInterface, colors);
		} catch (ClassNotFoundException e) {
			//Log exception
			logger.throwing(ColorabilityEncoding.class.getName(), "getEncoding", e);
		} catch (SecurityException e) {
			//Log exception
			logger.throwing(ColorabilityEncoding.class.getName(), "getEncoding", e);
		} catch (NoSuchMethodException e) {
			//Log exception
			logger.throwing(ColorabilityEncoding.class.getName(), "getEncoding", e);
		} catch (IllegalArgumentException e) {
			//Log exception
			logger.throwing(ColorabilityEncoding.class.getName(), "getEncoding", e);
		} catch (InstantiationException e) {
			//Log exception
			logger.throwing(ColorabilityEncoding.class.getName(), "getEncoding", e);
		} catch (IllegalAccessException e) {
			//Log exception
			logger.throwing(ColorabilityEncoding.class.getName(), "getEncoding", e);
		} catch (InvocationTargetException e) {
			//Log exception
			logger.throwing(ColorabilityEncoding.class.getName(), "getEncoding", e);
		}
		finally {
			//If no instance has been created fall back to the default encoding
			if(encoding==null) {
				encoding = getDefaultEncoding(maximumInterface, innerInterface, outerInterface, colors);
			}
		}

		//Log exiting
		logger.exiting(ColorabilityEncoding.class.getName(), "getEncoding", encoding);		
		return encoding;
	}
	
	/**
	 * Returns the type string indicating the color bits for the {@code node}-th node color.
	 * @param node the position of the node
	 * @return the color type string
	 */
	public static final String getColorNodeType(int node) {
		return NODE_BIT_TYPE+"-"+node;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Constructs a new encoding for a <i>colorability automaton</i> with a maximum interface 
	 * size of {@code maximumInterface} and {@code colors} different colors.
	 * @param maximumInterface the maximum interface size
	 * @param innerInterface size of the inner interface
	 * @param outerInterface size of the outer interface
	 * @param colors the number of different colors
	 */
	protected ColorabilityEncoding(int maximumInterface, int innerInterface, int outerInterface, 
			int colors) {
		super(getBitLength(maximumInterface) + maximumInterface*getBitLength(colors), 
				maximumInterface, innerInterface, outerInterface);
		this.colors = colors;
		this.colorBitLength = getBitLength(colors);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Abstract Methods																*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the {@code bit}-th BDD node of the color encoding of the {@code node}-th node 
	 * for the state {@code state}. 
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param node the position of the interface node
	 * @param bit the index of the bit in the color encoding
	 * @return the BDD node representing the bit of the color encoding
	 */
	public abstract RavenBDD getColorBit(StateType type, int node, int bit);
	
	/**
	 * Constructs a BDD which is {@code true} if and only if in the state {@code state} 
	 * the {@code node}-th interface node either has some color {@code color}.
	 * 
	 * <emph>Note:</emph> The "empty color", i.e. the color with value 0 is permitted.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param node the index of the interface node
	 * @param color the color of the node
	 * @return a BDD encoding that the {@code node}-th node has been colored
	 */
	public abstract RavenBDD getBDDForColor(StateType type, int node, int color);

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the number of colors.
	 * @return the number of colors
	 */
	public final int getColorCount() {
		return colors;
	}

	/**
	 * Returns the number of bits used to encode 
	 * the color of a single node.
	 * @return the number of bits of the color encoding
	 */
	public final int getColorBitLength() {
		return colorBitLength;
	}


	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the color of the {@code node}-th node in the state encoded by the BDD {@code state}.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param node the index of the node
	 * @param state the BDD encoding the state
	 * @return the color of the specified node
	 */
	protected final int getColor(StateType type, int node, RavenBDD state) {
		RavenBDD vars;
		RavenBDDDomain domain = type==StateType.current ? curStateDomain : sucStateDomain;
		int color = -1;
		
		vars = domain.getDomainBDD();
		if(state.satCount(vars)==1) {
			state = state.exists(getInterfaceEncodingBDD());
			
			for(int i=0;i<maximumInterface;i++) {
				if(i!=node) {
					state = state.exists(getColorEncodingBDD(type, i));
				}
			}
			
			color = RavenBDDDomain.getInteger(state);
		}
		return color;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the BDD containing the bits which encode the color of the {@code node}-th node. 
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param node the index of the node
	 * @return the color encoding BDD
	 */
	private final RavenBDD getColorEncodingBDD(StateType type, int node) {
		RavenBDD bdd;
		
		bdd = RavenBDDDomain.one();
		for(int i=0;i<colorBitLength;i++) {
			bdd = bdd.and(getColorBit(type, node, i));
		}
		
		return bdd;
	}
	
}
