/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.encodings;

import de.uni_due.inf.ti.raven.data.automata.EdgeCountingAutomaton;
import de.uni_due.inf.ti.raven.data.automata.VertexCountingAutomaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;

/**
 * This class represents the encoding of state space and the transition functions of (edge- or 
 * vertex-)counting automata and provides methods to access the encoding.
 * 
 * <i>Note:</i> The concrete implementation of a (edge- or vertex-)counting encoding can be 
 * found in classes extending this abstract class.
 * @author Christoph Blume
 * @version $Revision: 1.9 $
 *
 * @see EdgeCountingEncoding
 * @see VertexCountingEncoding
 * @see EdgeCountingAutomaton
 * @see VertexCountingAutomaton
 */
public abstract class CountingEncoding extends BDDEncoding {

	private static final String COUNTING_BIT_TYPE = "CountingType";
	
	/** The divisor with regard to which nodes are counted*/
	protected final int divisor;
	/** The bit length to encode the remainder */
	protected final int remainderBitLength;

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the type string indicating the counting bits.
	 * @return the counting type string
	 */
	public static final String getCountingType() {
		return COUNTING_BIT_TYPE;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Constructs a new encoding for an <i>edge-counting automaton</i> or a <i>vertex-counting 
	 * automaton</i> with a maximum interface size of {@code maximumInterface} and a number of edges 
	 * or nodes respectively equal to {@code remainder} modulo {@code divisor}.
	 * @param maximumInterface size of the maximum interface 
	 * @param innerInterface size of the inner interface
	 * @param outerInterface size of the outer interface
	 * @param divisor the divisor 
	 */
	protected CountingEncoding(int maximumInterface, int innerInterface, int outerInterface, 
			int divisor) {
		super(getBitLength(maximumInterface) + getBitLength(divisor), maximumInterface, 
				innerInterface, outerInterface);
		this.divisor = divisor;
		this.remainderBitLength = getBitLength(divisor);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Abstract Methods																*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns a BDD which encodes in state {@code state} the {@code pos}-th bit of the
	 * remainder.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param pos the index of the bit in the encoding
	 * @return the BDD node representing the bit of the encoding
	 */
	public abstract RavenBDD getRemainderBit(StateType type, int pos);

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the divisor.
	 * @return the divisor
	 */
	public final int getDivisor() {
		return divisor;
	}

	/**
	 * Returns the number of bits used to encode the remainder.
	 * @return the number of bits of the remainder
	 */
	public final int getRemainderBitLength() {
		return remainderBitLength;
	}

}