/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.encodings;

import de.uni_due.inf.ti.raven.data.automata.ColorabilityAutomaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDDomain;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class provides a concrete implementation of an encoding for <i>colorability automata</i>.
 * The state encoding is as follows:
 * 
 * <i>b<sub>1</sub> ... b<sub>l</sub> (c<sub>1,1</sub> ... c<sub>1,m</sub>) ... (c<sub>k,1</sub> 
 * ... c<sub>k,m</sub>)</i>, where <i>l = log<sub>2</sub>(k)</i> is the binary logarithm of the 
 * maximum interface size <i>k</i> and <i>m</i> is the binary logarithm of the number of colors 
 * (of the colorability automaton). The block <i>b<sub>1</sub> ... b<sub>l</sub></i> encodes the 
 * interface size to which the current state belongs, the block <i>c<sub>i,1</sub> ... 
 * c<sub>i,m</sub></i> the color of the <i>i</i>-th interface node in the current state.
 * @author Christoph Blume
 * @version $Revision: 1.20 $
 *
 *	@see ColorabilityAutomaton
 */
public class DefaultColorabilityEncoding extends ColorabilityEncoding {
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Constructs a new encoding for a <i>colorability automaton</i> with a maximum interface 
	 * size of {@code maximumInterface} and {@code colors} different colors.
	 * @param maximumInterface size of the maximum interface
	 * @param innerInterface size of the inner interface
	 * @param outerInterface size of the outer interface
	 * @param colors the number of different colors
	 */
	protected DefaultColorabilityEncoding(int maximumInterface, int innerInterface, int outerInterface, 
			int colors) {
		super(maximumInterface, innerInterface, outerInterface, colors);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public String getBitType(int pos) {
		if(pos<0 || pos>=length) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_BIT_POSITION_EXCEPTION);
		}
		
		if(pos<interfaceLength) {
			return getInterfaceSizeType();
		}

		return getColorNodeType((pos-interfaceLength) / colorBitLength);
	}

	@Override
	public RavenBDD getColorBit(StateType type, int node, int bit) {
		if(node<0 || node>=maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}
		if(bit<0 || bit>=colorBitLength) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_BIT_POSITION_EXCEPTION);
		}

		RavenBDDDomain domain = null;
		if(type==StateType.current) {
			domain = curStateDomain;
		}
		else {
			domain = sucStateDomain;
		}
		return domain.getIthBDD(interfaceLength+(node*colorBitLength)+bit);
	}

	@Override
	public RavenBDD getBDDForColor(StateType type, int node, int color) {
		RavenBDD bdd = one();
		for(int k=0;k<colorBitLength;k++) {
			int b = color%2;
			color >>= 1;

			if(b==0) {
				bdd = bdd.and(getColorBit(type, node, k).not());
			}
			else {
				bdd = bdd.and(getColorBit(type, node, k));
			}
		}

		return bdd;
	}

	@Override
	public String printStates(RavenBDD stateSet) {
		StringBuilder stateString = new StringBuilder();
		
		if(stateSet.satCount(curStateDomain.getDomainBDD())==1) {
			int interfaceSize = getInterfaceSize(StateType.current, stateSet);			
			
			stateString.append(interfaceSize);
			for(int i=0;i<maximumInterface;i++) {
				int color = getColor(StateType.current, i, stateSet);

				stateString.append("-");
				stateString.append(color);
			}
			
			return stateString.toString();
		}

		boolean isFirst = true;
		stateString.append("{");
		for(RavenBDD bdd: stateSet.getSatAssignments(curStateDomain.getDomainBDD())) {
			if(isFirst) {
				isFirst = false;
			}
			else {
				stateString.append(", ");
			}

			stateString.append(printStates(bdd));
		}
		stateString.append("}");
		
		return stateString.toString();
	}

}
