/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

 package de.uni_due.inf.ti.raven.encodings;

import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDDomain;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

 /**
  * This class represents the default encoding of state space and the transition functions of
  * link automata and provides methods to access the encoding.
  * Link automata accept graphs such that at least one node of the inner
  * interface is directly connected to a node of the outer interface.
  *
  * Encoding scheme:
  *--------------------------------------------------------------------------------------------
  *| interfacesize | node 1 initial | node 1 connected to initial node | node 2 initial | ... |
  *--------------------------------------------------------------------------------------------
  *
  * @author Laura Steinert
  * @version $Revision: 1.9 $
  *
  * @see LinkEncoding
  */
public class DefaultLinkEncoding extends LinkEncoding {

	/**
	 * @param maximumInterface the maximum interface size
	 * @param innerInterface the size of the inner interface
	 * @param outerInterface the size of the outer interface
	 */
	protected DefaultLinkEncoding(final int maximumInterface, final int innerInterface,
			final int outerInterface) {
		super(maximumInterface, innerInterface, outerInterface);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public String getBitType(int pos) {
		if(pos<0 || pos>=length) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_BIT_POSITION_EXCEPTION);
		}
		
		if(pos<interfaceLength) {
			return getInterfaceSizeType();
		}

		int mod = (pos-interfaceLength) % 2;
		int nodeNumber = (pos-interfaceLength) / 2;
		if (mod == 0){
			return getInitialNodeType(nodeNumber);
		}

		return getConnectedNodeType(nodeNumber);				
	}


	@Override
	public String printStates(RavenBDD stateSet) {
		// TODO
		return null;
	}
	
	/**
	 * Get the initial (i.e. 1st) bit of the {@code node}-th (starting with 0) node markings.
	 */
	@Override
	public RavenBDD getInitialNodeBit(StateType type, int node) {
		if(node<0 || node>=maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_BIT_POSITION_EXCEPTION);
		}
		
		RavenBDDDomain domain;
		if(type==StateType.current) {
			domain = curStateDomain;
		}
		else {
			domain = sucStateDomain;
		}
		
		return domain.getIthBDD(interfaceLength + (node)*2);
	}

	/**
	 * Get the connected (i.e. 2nd) bit of the {@code node}-th (starting with 0) node markings.
	 */
	@Override
	public RavenBDD getConnectedNodeBit(StateType type, int node) {
		if(node<0 || node>=maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_BIT_POSITION_EXCEPTION);
		}
		
		RavenBDDDomain domain;
		if(type==StateType.current) {
			domain = curStateDomain;
		}
		else {
			domain = sucStateDomain;
		}
		
		return domain.getIthBDD(interfaceLength + (node)*2+1);
	}

}
