/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.encodings;

import de.uni_due.inf.ti.raven.data.automata.MinimumEdgeAutomaton;
import de.uni_due.inf.ti.raven.data.automata.MinimumVertexAutomaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDDomain;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class provides a concrete implementation of an encoding for a <i>minimum edge automaton</i>
 * or <i>minimum vertex automaton</i> respectively. The state encoding is as follows:
 * 
 * <i>b<sub>1</sub> ... b<sub>l</sub> r<sub>1</sub> ... r<sub>d</sub></i>, where <i>l = 
 * log<sub>2</sub>(k)</i> is the binary logarithm of the maximum interface size <i>k</i> and 
 * <i>d</i> is the binary logarithm of the divisor (for the vertex-counting automaton). The 
 * block <i>b<sub>1</sub> ... b<sub>l</sub></i> encodes the interface size to which the current 
 * state belongs and the <i>r<sub>1</sub> ... r<sub>d</sub></i> encodes the current value of the 
 * remainder <i>#nodes<sub>G</sub></i> <tt>mod</tt> <i>divisor</i>, where 
 * <i>#nodes<sub>G</sub></i> denotes the number of nodes in the input graph <i>G</i>.
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.8 $
 *
 * @see MinimumEdgeAutomaton
 * @see MinimumVertexAutomaton
 */
public class DefaultMinimumBoundEncoding extends BoundEncoding {

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Constructs a new encoding for a <i>minimum vertex automaton</i> with a maximum interface 
	 * size of {@code maximumInterface} and a minimal vertex count.
	 * @param maximumInterface the maximal interface size
	 * @param innerInterface size of the inner interface
	 * @param outerInterface size of the outer interface
	 * @param bound the bound on the number of vertices/edges
	 */
	protected DefaultMinimumBoundEncoding(int maximumInterface, int innerInterface, 
			int outerInterface, int bound) {
		super(maximumInterface, innerInterface, outerInterface, bound);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public String getBitType(int pos) {
		if(pos<0 || pos>=length) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_BIT_POSITION_EXCEPTION);
		}
		
		if(pos<interfaceLength) {
			return getInterfaceSizeType();
		}

		return getBoundType();
	}

	@Override
	public RavenBDD getValueBit(StateType type, int pos) {
		if(pos<0 || pos>=boundBitLength) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_BIT_POSITION_EXCEPTION);
		}
		
		RavenBDDDomain domain;
		if(type==StateType.current) {
			domain = curStateDomain;
		}
		else {
			domain = sucStateDomain;
		}
		
		return domain.getIthBDD(interfaceLength + pos);
	}
	
	@Override
	public String printStates(RavenBDD stateSet) {		
		if(stateSet.satCount(curStateDomain.getDomainBDD())==1) {
			return getInterfaceSize(StateType.current, stateSet) + "-" + 
						getValue(StateType.current, stateSet);
		}

		StringBuilder stateString = new StringBuilder();
		boolean isFirst = true;
		stateString.append("{");
		for(RavenBDD bdd: stateSet.getSatAssignments(curStateDomain.getDomainBDD())) {
			if(isFirst) {
				isFirst = false;
			}
			else {
				stateString.append(", ");
			}

			stateString.append(printStates(bdd));
		}
		stateString.append("}");
		
		return stateString.toString();
	}

}
