/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.encodings;

import de.uni_due.inf.ti.raven.data.automata.NoIsolatedNodesAutomaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDDomain;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class provides a concrete implementation of an encoding for a <i>no isolated nodes 
 * automaton</i>. The state encoding is as follows:
 * 
 * <i>b<sub>1</sub> ... b<sub>l</sub> n<sub>1</sub> ... n<sub>k</sub> s</i>, where <i>l = 
 * log<sub>2</sub>(k)</i> is the binary logarithm of the maximum interface size <i>k</i>. The 
 * block <i>b<sub>1</sub> ... b<sub>l</sub></i> encodes the interface size to which the current 
 * state belongs and the bits <i>n<sub>1</sub> ... n<sub>d</sub></i> encodes whether the several 
 * interface nodes are isolated (if the bit is zero) or not (otherwise). The bit <i>s</i> indicates,
 * whether the input graph contains any isolated node which is no longer accessible by the interface.
 * 
 * @author Christoph Blume
 * @version $Revision: 1.16 $
 *
 * @see NoIsolatedNodesAutomaton
 */
public class DefaultNoIsolatedNodesEncoding extends NoIsolatedNodesEncoding {

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Constructs a new encoding for a <i>no-isolated-nodes automaton</i> with a maximum interface 
	 * size of {@code maximumInterface}.
	 * @param maximumInterface size of the maximum interface
	 * @param innerInterface size of the inner interface
	 * @param outerInterface size of the outer interface
	 */
	protected DefaultNoIsolatedNodesEncoding(int maximumInterface, int innerInterface, 
			int outerInterface) {
		super(maximumInterface, innerInterface, outerInterface);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public String getBitType(int pos) {
		if(pos<0 || pos>=length) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_BIT_POSITION_EXCEPTION);
		}
		
		if(pos<interfaceLength) {
			return getInterfaceSizeType();
		}
		else if(pos<interfaceLength+maximumInterface) {
			return getIsolatedNodeType(pos-interfaceLength);
		}
		else {
			return getIsolationStateType();
		}
	}
	
	@Override
	public RavenBDD getIsNodeIsolated(StateType type, int node) {
		if(node<0 || node>=maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}

		RavenBDDDomain domain = null;
		if(type==StateType.current) {
			domain = curStateDomain;
		}
		else {
			domain = sucStateDomain;
		}
		return domain.getIthBDD(interfaceLength + node);
	}
	
	@Override
	public RavenBDD getUndefinedNode(StateType type, int node) {
		return getIsNodeIsolated(type, node).not();
	}

	@Override
	public RavenBDD getHasIsolatedNodes(StateType type) {
		RavenBDDDomain domain = null;
		if(type==StateType.current) {
			domain = curStateDomain;
		}
		else {
			domain = sucStateDomain;
		}
		return domain.getIthBDD(interfaceLength + maximumInterface);
	}

	@Override
	public String printStates(RavenBDD stateSet) {
		// TODO Auto-generated method stub
		return null;
	}

}
