/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.encodings;

import java.util.List;

import de.uni_due.inf.ti.raven.data.Edge;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.Node;
import de.uni_due.inf.ti.raven.data.automata.SubgraphAutomaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDDomain;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class provides a concrete implementation of an encoding for <i>subgraph automata</i>.
 * The state encoding is as follows:
 * 
 * <i>b<sub>1</sub> ... b<sub>l</sub> e<sub>1</sub> ... e<sub>m</sub> (v<sub>1</sub> f<sub>1,1</sub>
 * ... f<sub>1,n</sub>)</i>, where <i>l = log<sub>2</sub>(k)</i> is the binary logarithm of the 
 * maximum interface size <i>k</i>, <i>m</i> is the number of edges of the wanted subgraph and 
 * <i>n</i> is the number of nodes of the wanted subgraph. The block <i>b<sub>1</sub> ... 
 * b<sub>l</sub></i> encodes the interface size to which the current state belongs, the bit 
 * <i>e<sub>i</sub></i> the (non-)existence of the <i>i</i>-th edge (of the wanted subgraph in the 
 * current state, the bit <i>v<sub>i</sub></i> the (non-)existence of the <i>i</i>-th node (of the 
 * wanted subgraph in the current state and the bit <i>f<sub>i,j</sub></i> encodes whether 
 * the <i>i</i>-th node (of the wanted subgraph) is accessible by the <i>j</i>-th outer interface 
 * node in the current state.
 * 
 * @author Christoph Blume
 * @version $Revision: 1.20 $
 * 
 * @see SubgraphAutomaton
 *
 */
public class DefaultSubgraphEncoding extends SubgraphEncoding {
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Constructs a new encoding for a <i>subgraph automaton</i> with a maximum interface 
	 * size of {@code maximumInterface}, a wanted subgraph {@code subgraph}, node ordering 
	 * {@code nodeOrder} and edge ordering {@code edgeOrder}.
	 * @param maximumInterface size of the maximum interface
	 * @param innerInterface size of the inner interface
	 * @param outerInterface size of the outer interface
	 * @param subgraph the wanted subgraph
	 * @param nodeOrder the node order
	 * @param edgeOrder the edge order
	 */
	protected DefaultSubgraphEncoding(int maximumInterface, int innerInterface, int outerInterface, 
			Graph subgraph, List<Node> nodeOrder, List<Edge> edgeOrder) {
		super(maximumInterface, subgraph, nodeOrder, edgeOrder, innerInterface, outerInterface);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public String getBitType(int pos) {
		if(pos<0 || pos>=length) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_BIT_POSITION_EXCEPTION);
		}
		
		if(pos<interfaceLength) {
			return getInterfaceSizeType();
		}
		else if(pos<interfaceLength+edgeCount) {
			return getEdgeType(pos-interfaceLength);
		}
		else {
			int node = (pos-interfaceLength-edgeCount)/(maximumInterface+1);
			int mapping = (pos-interfaceLength-edgeCount)%(maximumInterface+1);
			
			if(mapping==0) {
				return getNodeType(node);
			}

			return getMappingType(node, mapping-1);
		}
	}

	@Override
	public RavenBDD getNodeBit(StateType type, int node) {
		if(node<0 || node>=nodeCount) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_BIT_POSITION_EXCEPTION);
		}
		
		RavenBDDDomain domain;
		if(type==StateType.current) {
			domain = curStateDomain;
		}
		else {
			domain = sucStateDomain;
		}

		return domain.getIthBDD(interfaceLength + edgeCount + node*(maximumInterface+1));
	}

	@Override
	public RavenBDD getEdgeBit(StateType type, int edge) {
		if(edge<0 || edge>=edgeCount) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_BIT_POSITION_EXCEPTION);
		}
		
		RavenBDDDomain domain;
		if(type==StateType.current) {
			domain = curStateDomain;
		}
		else {
			domain = sucStateDomain;
		}

		return domain.getIthBDD(interfaceLength + edge);
	}

	@Override
	public RavenBDD getMappingBit(StateType type, int node, int size) {
		if(node<0 || node>=nodeCount) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_BIT_POSITION_EXCEPTION);
		}
		if(size<0 || size>=maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_BIT_POSITION_EXCEPTION);
		}
		
		RavenBDDDomain domain;
		if(type==StateType.current) {
			domain = curStateDomain;
		}
		else {
			domain = sucStateDomain;
		}

		return domain.getIthBDD(interfaceLength + edgeCount + 
				node*(maximumInterface+1) + 1 + size);
	}

	@Override
	public String printStates(RavenBDD stateSet) {
		StringBuilder stateString = new StringBuilder();
		boolean isFirst = true;
		
		if(stateSet.satCount(curStateDomain.getDomainBDD())==1) {
			int interfaceSize = getInterfaceSize(StateType.current, stateSet);			
			
			stateString.append(interfaceSize);
			stateString.append("-{");
			for(int i=0;i<edgeCount;i++) {
				if(hasEdge(StateType.current, i, stateSet)) {
					if(isFirst) {
						isFirst = false;
					}
					else {
						stateString.append(", ");
					}
					
					stateString.append(edgeOrder.get(i).getName());
				};

			}
			
			stateString.append("}-{");
			isFirst = true;
			for(int i=0;i<nodeCount;i++) {
				if(hasNode(StateType.current, i, stateSet)) {
					if(isFirst) {
						isFirst = false;
					}
					else {
						stateString.append(", ");
					}
					
					stateString.append(nodeOrder.get(i).getName());
				};

			}
			
			stateString.append("}-(");
			isFirst = true;
			for(int i=0;i<interfaceSize;i++) {
				if(isFirst) {
					isFirst = false;
				}
				else {
					stateString.append(", ");
				}
				
				int index = getSubgraphNode(StateType.current, i, stateSet);
				stateString.append(index>-1 ? nodeOrder.get(index).getName() : "_");
			}
			stateString.append(")");
			return stateString.toString();
		}

		isFirst = true;
		stateString.append("{");
		for(RavenBDD bdd: stateSet.getSatAssignments(curStateDomain.getDomainBDD())) {
			if(isFirst) {
				isFirst = false;
			}
			else {
				stateString.append(", ");
			}

			stateString.append(printStates(bdd));
		}
		stateString.append("}");
		
		return stateString.toString();
	}

}
