/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.encodings;

import de.uni_due.inf.ti.raven.data.automata.VertexCountingAutomaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDDomain;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class provides a concrete implementation of an encoding for <i>vertex cover automata</i>.
 * The state encoding is as follows:
 * 
 * <i>b<sub>1</sub> ... b<sub>l</sub> m<sub>1</sub> ... m<sub>k</sub> s<sub>1</sub> ... 
 * s<sub>t</sub></i>, where <i>k</i> is the maximum interface size, <i>l = log<sub>2</sub>(k)</i> 
 * is the binary logarithm of <i>k</i> and <i>t</i> is the is the binary logarithm of the maximum 
 * vertex cover size. The block <i>b<sub>1</sub> ... b<sub>l</sub></i> encodes the interface size 
 * to which the current state belongs, the block <i>m<sub>1</sub> ... m<sub>k</sub></i> which nodes 
 * currently in the interface belong to the vertex cover and the block <i>s<sub>1</sub> ... 
 * s<sub>t</sub></i> encodes the size of the vertex cover in the current state.
 * 
 * @author Christoph Blume
 * @version $Revision: 1.19 $
 *
 * @see VertexCountingAutomaton
 */
public class DefaultVertexCoverEncoding extends VertexCoverEncoding {

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Constructs a new encoding for a <i>vertex cover automaton</i> with a maximum interface 
	 * size of {@code maximumInterface} and a vertex cover of size at most {@code coverSize}.
	 * @param maximumInterface the maximum interface size
	 * @param innerInterfaceSize the size of the inner interface
	 * @param outerInterfaceSize the size of the outer interface
	 * @param coverSize the maximum size of the vertex cover
	 */
	protected DefaultVertexCoverEncoding(int maximumInterface, int innerInterfaceSize,
			int outerInterfaceSize, int coverSize) {
		super(maximumInterface, innerInterfaceSize, outerInterfaceSize, coverSize);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public String getBitType(int pos) {
		if(pos<0 || pos>=length) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_BIT_POSITION_EXCEPTION);
		}
		
		if(pos<interfaceLength) {
			return getInterfaceSizeType();
		}
		else if(pos<interfaceLength+maximumInterface) {
			return getMembershipNodeType(pos-interfaceLength);
		}
		else {
			
			return getVertexCoverSizeType(); 
		}
	}
	
	@Override
	public RavenBDD getMembershipBit(StateType type, int node) {
		if(node<0 || node>=maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}
		
		RavenBDDDomain domain;
		if(type==StateType.current) {
			domain = curStateDomain;
		}
		else {
			domain = sucStateDomain;
		}
		
		return domain.getIthBDD(interfaceLength + node);
	}
	
	@Override
	public RavenBDD getUndefinedNode(StateType type, int node) {
		if(node<0 || node>=maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}
		
		return getMembershipBit(type, node).not();
	}
	
	@Override
	public RavenBDD getVertexCoverSizeBit(StateType type, int pos) {
		if(pos<0 || pos>=vertexCoverBitLength) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}
		
		RavenBDDDomain domain;
		if(type==StateType.current) {
			domain = curStateDomain;
		}
		else {
			domain = sucStateDomain;
		}

		return domain.getIthBDD(interfaceLength + maximumInterface + pos);
	}

	@Override
	public String printStates(RavenBDD stateSet) {
		StringBuilder stateString = new StringBuilder();
		
		if(stateSet.satCount(curStateDomain.getDomainBDD())==1) {
			int interfaceSize = getInterfaceSize(StateType.current, stateSet);			
			
			stateString.append(interfaceSize);
			for(int i=0;i<maximumInterface;i++) {
				int membershipState = getMembershipState(StateType.current, i, stateSet);

				stateString.append("-");
				stateString.append(membershipState);
			}
			
			stateString.append("-");
			stateString.append(getVertexCoverSize(StateType.current, stateSet));
			
			return stateString.toString();
		}
		
		boolean isFirst = true;
		stateString.append("{");
		for(RavenBDD bdd: stateSet.getSatAssignments(curStateDomain.getDomainBDD())) {
			if(isFirst) {
				isFirst = false;
			}
			else {
				stateString.append(", ");
			}

			stateString.append(printStates(bdd));
		}
		stateString.append("}");
		
		return stateString.toString();
	}

}
