/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.encodings;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.data.automata.DominatingSetAutomaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDDomain;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class represents the encoding of state space and the transition functions of <i>dominating 
 * set automata</i> and provides methods to access the encoding.
 * 
 * <p><i>Note:</i> The concrete implementation of a dominating set encoding can be found in classes 
 * extending this abstract class.</p>
 * 
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.31 $
 *
 * @see DefaultDominatingSetEncoding
 * @see DominatingSetAutomaton
 */
public abstract class DominatingSetEncoding extends BDDEncoding {
	
	private static final String MEMBERSHIP_BIT_TYPE = "MembershipType";
	private static final String DOMINATION_BIT_TYPE = "DominationType";
	private static final String DOMINATION_SET_SIZE_BIT_TYPE = "DominationSetSizeType";

	private static final Logger logger = Logger.getLogger(
			DominatingSetEncoding.class.getName());
	
	/** The size of the dominating set */
	protected final int dominationNumber;
	
	/** The bit length to encode the size of the dominating set */
	protected final int dominationBitLength;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the default BDD encoding for a dominating set instance with a 
	 * maximum interface size of {@code maximumInterface} and a maximum dominating 
	 * set size of {@code dominating size}.
	 * @param maximumInterface the maximum interface size
	 * @param innerInterface the size of the inner interface
	 * @param outerInterface the size of the outer interface
	 * @param dominatingSize the maximum size of the dominating set
	 * @return the default encoding
	 */
	public static final DominatingSetEncoding getDefaultEncoding(int maximumInterface, 
			int innerInterface, int outerInterface, int dominatingSize) {
		if(maximumInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(outerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(outerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(dominatingSize<1) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_DOMINATING_SET_SIZE_MUST_BE_POSITIVE_EXCEPTION);
		}
		
		//Log entering 
		logger.entering(DominatingSetEncoding.class.getName(), "getEncoding", 
				new Object[]{maximumInterface, innerInterface, outerInterface, dominatingSize});

		DominatingSetEncoding encoding =  new DefaultDominatingSetEncoding(maximumInterface, 
				innerInterface, outerInterface, dominatingSize);
		
		//Log exiting
		logger.exiting(DominatingSetEncoding.class.getName(), "getEncoding", encoding);		
		return encoding;
	}
	
	/**
	 * Returns the BDD encoding for a dominating set instance given by the class name
	 * {@code name} with a maximum interface size of {@code maximumInterface} and a 
	 * maximum dominating set size of {@code dominating size}. If no dominating set 
	 * encoding with the given name can be found the default encoding is returned.
	 * @param maximumInterface the maximum interface size
	 * @param innerInterface the size of the inner interface
	 * @param outerInterface the size of the outer interface
	 * @param dominatingSize the maximum size of the dominating set
	 * @param encodingName the class name of the dominating set encoding class
	 * @return the encoding given by the class name or the default encoding if no 
	 * dominating set encoding class with the given can be found 
	 */
	public static final DominatingSetEncoding getEncoding(int maximumInterface, int innerInterface, 
			int outerInterface, int dominatingSize, String encodingName) {
		if(maximumInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(outerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(outerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(dominatingSize<1) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_DOMINATING_SET_SIZE_MUST_BE_POSITIVE_EXCEPTION);
		}
		if(encodingName==null) {
			throw new NullPointerException(
					RavenExceptionMessages.ENCODING_NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(encodingName.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ENCODING_NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		
		//Log entering 
		logger.entering(DominatingSetEncoding.class.getName(), "getEncoding", 
				new Object[]{maximumInterface, innerInterface, outerInterface, dominatingSize, 
								encodingName});

		
		DominatingSetEncoding encoding = null;
		Class<?> encodingClass;
		try {
			//Create new dominating set encoding instance
			encodingClass = Class.forName(encodingName);
			Constructor<?> constructor = encodingClass.getDeclaredConstructor(int.class, int.class, 
					int.class, int.class);
			encoding = (DominatingSetEncoding) constructor.newInstance(maximumInterface, 
					innerInterface, outerInterface, dominatingSize);
		} catch (ClassNotFoundException e) {
			//Log exception
			logger.throwing(DominatingSetEncoding.class.getName(), "getEncoding", e);
		} catch (SecurityException e) {
			//Log exception
			logger.throwing(DominatingSetEncoding.class.getName(), "getEncoding", e);
		} catch (NoSuchMethodException e) {
			//Log exception
			logger.throwing(DominatingSetEncoding.class.getName(), "getEncoding", e);
		} catch (IllegalArgumentException e) {
			//Log exception
			logger.throwing(DominatingSetEncoding.class.getName(), "getEncoding", e);
		} catch (InstantiationException e) {
			//Log exception
			logger.throwing(DominatingSetEncoding.class.getName(), "getEncoding", e);
		} catch (IllegalAccessException e) {
			//Log exception
			logger.throwing(DominatingSetEncoding.class.getName(), "getEncoding", e);
		} catch (InvocationTargetException e) {
			//Log exception
			logger.throwing(DominatingSetEncoding.class.getName(), "getEncoding", e);
		}
		finally {
			//If no instance has been created fall back to the default encoding
			if(encoding==null) {
				encoding = getDefaultEncoding(maximumInterface, innerInterface, outerInterface, 
						dominatingSize);
			}
		}

		//Log exiting
		logger.exiting(DominatingSetEncoding.class.getName(), "getEncoding", encoding);		
		return encoding;
	}
	
	/**
	 * Returns the type string indicating the membership bits for the {@code node}-th node.
	 * @param node the position of the node
	 * @return the membership type string
	 */
	public static final String getMembershipNodeType(int node) {
		return MEMBERSHIP_BIT_TYPE+"-"+node;
	}
	
	/**
	 * Returns the type string indicating the domination bits for the {@code node}-th node.
	 * @param node the position of the node
	 * @return the domination type string
	 */
	public static final String getDominationNodeType(int node) {
		return DOMINATION_BIT_TYPE+"-"+node;
	}
	
	/**
	 * Returns the type string indicating the bits for the domination set size.
	 * @return the domination set size type string
	 */
	public static final String getDominationSetSizeType() {
		return DOMINATION_SET_SIZE_BIT_TYPE;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Constructs a new encoding for an <i>dominating set automaton</i> with a maximum interface 
	 * size of {@code maximumInterface} and a dominating set of size at most {@code dominatingSize}.
	 * @param maximumInterface the maximum interface size
	 * @param innerInterfaceSize the size of the inner interface
	 * @param outerInterfaceSize the size of the outer interface
	 * @param dominatingSize the maximum size of the dominating set
	 */
	protected DominatingSetEncoding(int maximumInterface, int innerInterfaceSize, 
			int outerInterfaceSize, int dominatingSize) {
		super(getBitLength(maximumInterface) + 2*maximumInterface + 
				getBitLength(dominatingSize), maximumInterface, 
				innerInterfaceSize, outerInterfaceSize);
		this.dominationNumber = dominatingSize;
		this.dominationBitLength = getBitLength(dominatingSize);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Abstract Methods																*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns a BDD which encodes that the {@code node}-th node of the interface
	 * is a member of the dominating set.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param node the index of the interface node
	 * @return a BDD encoding the membership of the node
	 */
	public abstract RavenBDD getMembershipBit(StateType type, int node);
	
	/**
	 * Returns a BDD which encodes that in state {@code state} the {@code node}-th interface node 
	 * is dominated by some node of the dominating set if and only if the BDD is {@code true}.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param node the position of the interface node
	 * @return a BDD encoding that the interface node is dominated
	 */
	public abstract RavenBDD getDominationBit(StateType type, int node);
	
	/**
	 * Returns a BDD which encodes that in state {@code state} the {@code node}-th interface node 
	 * is not defined if and only if the BDD is {@code true}.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param node the position of the interface node
	 * @return a BDD encoding that the interface node is not defined
	 */
	public abstract RavenBDD getUndefinedNode(StateType type, int node);
	
	/**
	 * Returns a BDD which encodes in state {@code state} the {@code pos}-th bit of the
	 * current size of the dominating set. 
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param pos the position of the bit
	 * @return a BDD encoding a bit of the dominating set size
	 */
	public abstract RavenBDD getDominationSizeBit(StateType type, int pos);
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the maximum size of the dominating set.
	 * @return the maximum dominating set size
	 */
	public final int getDominationNumber() {
		return dominationNumber;
	}

	/**
	 * Returns the bit length used for the encoding of the dominating set size.
	 * @return the dominating set bit length
	 */
	public final int getDominationBitLength() {
		return dominationBitLength;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the membership state of the {@code node}-th in state encoded by the BDD {@code state}
	 * which is either "is member of the dominating set" or "is <i>not</i> member of the dominating set".
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param node the index of the node
	 * @param state the BDD encoding the state
	 * @return the membership state of the node, either {@code 1} (indicating the membership) or 
	 * {@code 0} (indicating the non-membership) 
	 */
	protected final int getMembershipState(StateType type, int node, RavenBDD state) {
		RavenBDD vars;
		RavenBDDDomain domain = type==StateType.current ? curStateDomain : sucStateDomain;
		int membershipState = -1;
		
		vars = domain.getDomainBDD();
		if(state.satCount(vars)==1) {
			if(state.clone().and(getMembershipBit(type, node)).isZero()) {
				membershipState = 0;
			}
			else {
				membershipState = 1;
			}
		}
		return membershipState;
	}
	
	/**
	 * Returns the domination state of the {@code node}-th in state encoded by the BDD {@code state}
	 * which is either "is dominated by some node" or "is <i>not</i> dominated by some node".
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param node the index of the node
	 * @param state the BDD encoding the state
	 * @return the membership state of the node, either {@code 1} (indicating the domination) or 
	 * {@code 0} (indicating the non-domination) 
	 */
	protected final int getDominationState(StateType type, int node, RavenBDD state) {
		RavenBDD vars;
		RavenBDDDomain domain = type==StateType.current ? curStateDomain : sucStateDomain;
		int dominationState = -1;
		
		vars = domain.getDomainBDD();
		if(state.satCount(vars)==1) {
			if(state.clone().and(getDominationBit(type, node)).isZero()) {
				dominationState = 0;
			}
			else {
				dominationState = 1;
			}
		}
		return dominationState;
	}

	/**
	 * Returns the size of the domination set in state encoded by the BDD {@code state}.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param state the BDD encoding the state
	 * @return the size of the domination set
	 */
	protected final int getDominationSetSize(StateType type, RavenBDD state) {
		RavenBDD vars;
		RavenBDDDomain domain = type==StateType.current ? curStateDomain : sucStateDomain;
		int dominationSetSize = -1;
		
		vars = domain.getDomainBDD();
		if(state.satCount(vars)==1) {
			state = state.exists(getInterfaceEncodingBDD());
			
			for(int i=0;i<maximumInterface;i++) {
				state = state.exists(getMembershipBit(type, i));
				state = state.exists(getDominationBit(type, i));
			}
			
			dominationSetSize = RavenBDDDomain.getInteger(state);
		}
		return dominationSetSize;
	}
	
}
