/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.encodings;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class represents the encoding of state space and the transition functions of edge-counting 
 * automata and provides methods to access the encoding.
 * 
 * <p><i>Note:</i> The concrete implementation of an edge-counting encoding can be found in classes 
 * extending this abstract class.</p>
 * @author Christoph Blume
 * @version $Revision: 1.11 $
  */
public abstract class EdgeCountingEncoding extends CountingEncoding {

	private static final Logger logger = Logger.getLogger(
			VertexCountingEncoding.class.getName());
	
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the default BDD encoding for a vertex count instance with a 
	 * maximum interface size of {@code maximumInterface} and a count of edges modulo 
	 * modNumber {@code modNumber}.
	 * @param maximumInterface size of the maximum interface
	 * @param innerInterface size of the inner interface
	 * @param outerInterface size of the outer interface
	 * @param divisor the divisor modulo which edges are counted
	 * @return the default encoding
	 */
	public static final EdgeCountingEncoding getDefaultEncoding(int maximumInterface, 
			int innerInterface, int outerInterface, int divisor) {
		if(maximumInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(outerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(outerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(divisor<1) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.DIVISOR_MUST_BE_POSITIVE_EXCEPTION);
		}

		return new DefaultEdgeCountingEncoding(maximumInterface, innerInterface, outerInterface, 
				divisor);
	}
	
	/**
	 * Returns the BDD encoding for a vertex count instance given by the class name
	 * {@code name} with a maximum interface size of {@code maximumInterface} and a 
	 * modulo number of {@code modNumber}. If no vertex count encoding with the 
	 * given name can be found the default encoding is returned.
	 * @param maximumInterface the maximum interface size
	 * @param innerInterface size of the inner interface
	 * @param outerInterface size of the outer interface
	 * @param divisor the divisor modulo which edges are counted
	 * @param encodingName the class name of the encoding class
	 * @return the encoding given by the class name or the default encoding if no 
	 * vertex count encoding class with the given name can be found 
	 */
	public static final EdgeCountingEncoding getEncoding(int maximumInterface, int innerInterface, 
			int outerInterface, int divisor, String encodingName) {
		if(maximumInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(outerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(outerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(divisor<1) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.DIVISOR_MUST_BE_POSITIVE_EXCEPTION);
		}
		if(encodingName==null) {
			throw new NullPointerException(
					RavenExceptionMessages.ENCODING_NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(encodingName.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ENCODING_NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		
		//Log entering 
		logger.entering(EdgeCountingEncoding.class.getName(), "getEncoding", 
				new Object[]{maximumInterface, innerInterface, outerInterface, divisor, 
			encodingName});

		EdgeCountingEncoding encoding = null;
		Class<?> encodingClass;
		try {
			//Create new vertex count encoding instance
			encodingClass = Class.forName(encodingName);
			Constructor<?> constructor = encodingClass.getDeclaredConstructor(int.class, int.class, 
					int.class, int.class);
			encoding = (EdgeCountingEncoding) constructor.newInstance(maximumInterface, 
					innerInterface, outerInterface, divisor);
		} catch (ClassNotFoundException e) {
			//Log exception
			logger.throwing(EdgeCountingEncoding.class.getName(), "getEncoding", e);
		} catch (SecurityException e) {
			//Log exception
			logger.throwing(EdgeCountingEncoding.class.getName(), "getEncoding", e);
		} catch (NoSuchMethodException e) {
			//Log exception
			logger.throwing(EdgeCountingEncoding.class.getName(), "getEncoding", e);
		} catch (IllegalArgumentException e) {
			//Log exception
			logger.throwing(EdgeCountingEncoding.class.getName(), "getEncoding", e);
		} catch (InstantiationException e) {
			//Log exception
			logger.throwing(EdgeCountingEncoding.class.getName(), "getEncoding", e);
		} catch (IllegalAccessException e) {
			//Log exception
			logger.throwing(EdgeCountingEncoding.class.getName(), "getEncoding", e);
		} catch (InvocationTargetException e) {
			//Log exception
			logger.throwing(EdgeCountingEncoding.class.getName(), "getEncoding", e);
		}
		finally {
			//If no instance has been created fall back to the default encoding
			if(encoding==null) {
				encoding = getDefaultEncoding(maximumInterface, innerInterface, outerInterface, 
						divisor);
			}
		}

		//Log exiting
		logger.exiting(EdgeCountingEncoding.class.getName(), "getEncoding", encoding);		
		return encoding;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Constructs a new encoding for a <i>vertex counting automaton</i> with a maximum interface 
	 * size of {@code maximumInterface} and a vertex count modulo{@code modNumber}.
	 * @param maximumInterface the maximum interface size
	 * @param innerInterface size of the inner interface
	 * @param outerInterface size of the outer interface
	 * @param divisor the divisor modulo which edges are counted
	 */
	protected EdgeCountingEncoding(int maximumInterface, int innerInterface, int outerInterface, 
			int divisor) {
		super(maximumInterface, innerInterface, outerInterface, divisor);
	}

}
