/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.encodings;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.data.automata.LinkAutomaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class represents the encoding of state space and the transition functions of <i>link 
 * automata</i> and provides methods to access the encoding.
 * 
 * <p><i>Note:</i> The concrete implementation of a link encoding can be found in classes 
 * extending this abstract class.</p>
 *
 * @author Laura Steinert
 * @version $Revision: 1.12 $
 * 
 * @see DefaultLinkEncoding
 * @see AnotherLinkEncoding
 * @see LinkAutomaton
 */
public abstract class LinkEncoding extends BDDEncoding {

	private static final String INITIAL_NODE_BIT_TYPE = "InitialNodeType";
	private static final String CONNECTED_NODE_BIT_TYPE = "ConnectedNodeType";

	private static final Logger logger = Logger.getLogger(
			LinkEncoding.class.getName());

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the default BDD encoding for a link encoding instance with a 
	 * maximum interface size of {@code maximumInterface}, an 
	 * inner interface size of {@code innerInterface} and an outer interface size 
	 * of {@code outerInterface}.
	 * @param maximumInterface the maximum interface size
	 * @param innerInterface the size of the inner interface
	 * @param outerInterface the size of the outer interface
	 * @return the default encoding
	 */
	public static final LinkEncoding getDefaultEncoding(
			int maximumInterface, int innerInterface, int outerInterface){
		if(maximumInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(outerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(outerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}

		return new AnotherLinkEncoding(maximumInterface, innerInterface, outerInterface);
	}
	
	/**
	 * Returns the BDD encoding for a link encoding instance given by the class name
	 * {@code name} with a maximum interface size of {@code maximumInterface}, an 
	 * inner interface size of {@code innerInterface} and an outer interface size 
	 * of {@code outerInterface}. If no link encoding with the given name can be 
	 * found the default encoding is returned.
	 * @param maximumInterface the maximum interface size
	 * @param innerInterface the size of the inner interface
	 * @param outerInterface the size of the outer interface
	 * @param encodingName the class name of the dominating set encoding class
	 * @return the encoding given by the class name or the default encoding if no 
	 * link encoding class with the given can be found 
	 */
	public static final LinkEncoding getEncoding(int maximumInterface, int innerInterface, 
			int outerInterface, String encodingName) {
		if(maximumInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(outerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(outerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(encodingName==null) {
			throw new NullPointerException(
					RavenExceptionMessages.ENCODING_NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(encodingName.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ENCODING_NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		
		//Log entering 
		logger.entering(LinkEncoding.class.getName(), "getEncoding", 
				new Object[]{encodingName, maximumInterface, innerInterface, outerInterface});

		LinkEncoding encoding = null;
		Class<?> encodingClass;
		try {
			encodingClass = Class.forName(encodingName);
			Constructor<?> constructor = encodingClass.getDeclaredConstructor(int.class, int.class, int.class);
			encoding = (LinkEncoding) constructor.newInstance(maximumInterface, innerInterface, outerInterface);
		} catch (ClassNotFoundException e) {
			//Log exception
			logger.throwing(LinkEncoding.class.getName(), "getEncoding", e);
		} catch (SecurityException e) {
			//Log exception
			logger.throwing(LinkEncoding.class.getName(), "getEncoding", e);
		} catch (NoSuchMethodException e) {
			//Log exception
			logger.throwing(LinkEncoding.class.getName(), "getEncoding", e);
		} catch (IllegalArgumentException e) {
			//Log exception
			logger.throwing(LinkEncoding.class.getName(), "getEncoding", e);
		} catch (InstantiationException e) {
			//Log exception
			logger.throwing(LinkEncoding.class.getName(), "getEncoding", e);
		} catch (IllegalAccessException e) {
			//Log exception
			logger.throwing(LinkEncoding.class.getName(), "getEncoding", e);
		} catch (InvocationTargetException e) {
			//Log exception
			logger.throwing(LinkEncoding.class.getName(), "getEncoding", e);
		}
		finally {
			//If no instance has been created fall back to the default encoding
			if(encoding==null) {
				encoding = getDefaultEncoding(maximumInterface, innerInterface, outerInterface);
			}
		}

		//Log exiting
		logger.exiting(LinkEncoding.class.getName(), "getEncoding", encoding);		
		return encoding;
	}
	
	/**
	 * Returns the type string indicating the initial node bits for the {@code node}-th node.
	 * @param node the position of the node
	 * @return the initial node type string
	 */
	public static final String getInitialNodeType(int node) {
		return INITIAL_NODE_BIT_TYPE+"-"+node;
	}
	
	/**
	 * Returns the type string indicating the connected node bits for the {@code node}-th node.
	 * @param node the position of the node
	 * @return the connected node type string
	 */
	public static final String getConnectedNodeType(int node) {
		return CONNECTED_NODE_BIT_TYPE+"-"+node;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * 
	 * @param maximumInterface the maximum interface size
	 * @param innerInterface the size of the inner interface
	 * @param outerInterface the size of the outer interface
	 */
	protected LinkEncoding(int maximumInterface, int innerInterface, int outerInterface) {
		super(getBitLength(maximumInterface) + maximumInterface*2, maximumInterface, 
				innerInterface, outerInterface);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Abstract Methods																*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns a BDD which encodes in state {@code state} whether the {@code pos}-th node of the 
	 * interface is an initial node or not.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param node the index of the node
	 * @return the initial node encoding BDD
	 */
	public abstract RavenBDD getInitialNodeBit(StateType type, int node);

	/**
	 * Returns a BDD which encodes in state {@code state} whether the {@code pos}-th node of the 
	 * interface is connected to an initial node or not.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param node the index of the node
	 * @return the node connection encoding BDD
	 */
	public abstract RavenBDD getConnectedNodeBit(StateType type, int node);

}