/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.encodings;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.data.automata.NoIsolatedNodesAutomaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class represents the encoding of state space and the transition functions of <i>no isolated 
 * nodes automata</i> and provides methods to access the encoding.
 * 
 * <p><i>Note:</i> The concrete implementation of a no isolated nodes encoding can be found in 
 * classes extending this abstract class.</p>
 * 
 * @author Christoph Blume
 * @version $Revision: 1.19 $
 *
 * @see DefaultNoIsolatedNodesEncoding
 * @see NoIsolatedNodesAutomaton
 */
public abstract class NoIsolatedNodesEncoding extends BDDEncoding {

	private static final String NODE_BIT_TYPE = "NodeType";
	private static final String ISOLATION_STATE_BIT_TYPE = "IsolationStateType";
	
	private static final Logger logger = Logger.getLogger(
			NoIsolatedNodesEncoding.class.getName());
	
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the default BDD encoding for a dominating set instance with a 
	 * maximum interface size of {@code maximumInterface} and a maximum dominating 
	 * set size of {@code dominating size}.
	 * @param maximumInterface size of the maximum interface
	 * @param innerInterface size of the inner interface
	 * @param outerInterface size of the outer interface
	 * @return the default encoding
	 */
	public static final NoIsolatedNodesEncoding getDefaultEncoding(int maximumInterface, 
			int innerInterface, int outerInterface) {
		if(maximumInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(outerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(outerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}

		return new DefaultNoIsolatedNodesEncoding(maximumInterface, innerInterface, outerInterface);
	}
	
	/**
	 * Returns the BDD encoding for a dominating set instance given by the class name
	 * {@code name} with a maximum interface size of {@code maximumInterface} and a 
	 * maximum dominating set size of {@code dominating size}. If no dominating set 
	 * encoding with the given name can be found the default encoding is returned.
	 * @param maximumInterface size of the maximum interface
	 * @param innerInterface size of the inner interface
	 * @param outerInterface size of the outer interface
	 * @param encodingName the class name of the dominating set encoding class
	 * @return the encoding given by the class name or the default encoding if no 
	 * dominating set encoding class with the given can be found 
	 */
	public static final NoIsolatedNodesEncoding getEncoding(int maximumInterface, int innerInterface, 
			int outerInterface, String encodingName) {
		if(maximumInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(outerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(outerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(encodingName==null) {
			throw new NullPointerException(
					RavenExceptionMessages.ENCODING_NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(encodingName.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ENCODING_NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		
		//Log entering 
		logger.entering(NoIsolatedNodesEncoding.class.getName(), "getEncoding", 
				new Object[]{maximumInterface, innerInterface, outerInterface, encodingName});

		NoIsolatedNodesEncoding encoding = null;
		Class<?> encodingClass;
		try {
			//Create new dominating set encoding instance
			encodingClass = Class.forName(encodingName);
			Constructor<?> constructor = encodingClass.getDeclaredConstructor(int.class, int.class);
			encoding = (NoIsolatedNodesEncoding) constructor.newInstance(maximumInterface);
		} catch (ClassNotFoundException e) {
			//Log exception
			logger.throwing(NoIsolatedNodesEncoding.class.getName(), "getEncoding", e);
		} catch (SecurityException e) {
			//Log exception
			logger.throwing(NoIsolatedNodesEncoding.class.getName(), "getEncoding", e);
		} catch (NoSuchMethodException e) {
			//Log exception
			logger.throwing(NoIsolatedNodesEncoding.class.getName(), "getEncoding", e);
		} catch (IllegalArgumentException e) {
			//Log exception
			logger.throwing(NoIsolatedNodesEncoding.class.getName(), "getEncoding", e);
		} catch (InstantiationException e) {
			//Log exception
			logger.throwing(NoIsolatedNodesEncoding.class.getName(), "getEncoding", e);
		} catch (IllegalAccessException e) {
			//Log exception
			logger.throwing(NoIsolatedNodesEncoding.class.getName(), "getEncoding", e);
		} catch (InvocationTargetException e) {
			//Log exception
			logger.throwing(NoIsolatedNodesEncoding.class.getName(), "getEncoding", e);
		}
		finally {
			//If no instance has been created fall back to the default encoding
			if(encoding==null) {
				encoding = getDefaultEncoding(maximumInterface, innerInterface, outerInterface);
			}
		}

		//Log exiting
		logger.exiting(NoIsolatedNodesEncoding.class.getName(), "getEncoding", encoding);		
		return encoding;
	}
	
	/**
	 * Returns the type string indicating the isolation state bits for the {@code node}-th node.
	 * @param node the position of the node
	 * @return the isolation state type string
	 */
	public static final String getIsolatedNodeType(int node) {
		return NODE_BIT_TYPE+"-"+node;
	}
	
	/**
	 * Returns the type string indicating the bit for the isolation state.
	 * @return the isolation state type string
	 */
	public static final String getIsolationStateType() {
		return ISOLATION_STATE_BIT_TYPE;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Constructs a new encoding for a <i>no-isolated-nodes automaton</i> with a maximum interface 
	 * size of {@code maximumInterface}.
	 * @param maximumInterface the maximum interface size
	 * @param innerInterface size of the inner interface
	 * @param outerInterface size of the outer interface
	 */
	protected NoIsolatedNodesEncoding(int maximumInterface, int innerInterface, int outerInterface) {
		super(getBitLength(maximumInterface)+maximumInterface+1, maximumInterface, innerInterface, 
				outerInterface);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Abstract Methods																*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns a BDD which encodes that in state {@code state} the {@code node}-th interface node 
	 * is isolated if and only if the BDD is {@code true}.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param node the position of the interface node
	 * @return a BDD encoding the node is isolated
	 */
	public abstract RavenBDD getIsNodeIsolated(StateType type, int node);
	
	/**
	 * Returns a BDD which encodes that in state {@code state} the {@code node}-th interface node 
	 * is not defined if and only if the BDD is {@code true}.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param node the position of the interface node
	 * @return a BDD encoding that the interface node is not defined
	 */
	public abstract RavenBDD getUndefinedNode(StateType type, int node);

	/**
	 * Returns a BDD which encodes that in state {@code state} isolated nodes are present 
	 * if and only if the BDD is {@code true}.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @return a BDD encoding in the state isolated nodes are present
	 */
	public abstract RavenBDD getHasIsolatedNodes(StateType type);

}
