/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.encodings;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.data.automata.PathAutomaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class represents the encoding of state space and the transition functions of path 
 * automata and provides methods to access the encoding.
 * 
 * <p><i>Note:</i> The concrete implementation of a path encoding can be found in classes 
 * extending this abstract class.</p>
 * @author Christoph Blume, Dennis Nolte
 * @version $Revision: 1.23 $
 * 
 * @see DefaultPathEncoding
 * @see PathAutomaton
 */

public abstract class PathEncoding extends BDDEncoding {

	private static final String PATH_BIT_TYPE = "PathType";
	private static final String SOURCE_PATH_BIT_TYPE = "SourcePathType";
	
	private static final Logger logger = Logger.getLogger(PathEncoding.class.getName());
	
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the default BDD encoding for a path instance with a 
	 * maximum interface size of {@code maximumInterface}.
	 * @param maximumInterface the maximum interface size
	 * @param innerInterface size of the inner interface
	 * @param outerInterface size of the outer interface
	 * @return the default encoding
	 */
	public static final PathEncoding getDefaultEncoding(int maximumInterface, int innerInterface, 
			int outerInterface) {
		if(maximumInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(outerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(outerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		
		//Log entering 
		logger.entering(PathEncoding.class.getName(), "getEncoding", 
				new Object[]{maximumInterface, innerInterface, outerInterface});
		
//		PathEncoding encoding = new DefaultPathEncoding(maximumInterface, innerInterface, 
//				outerInterface);
		PathEncoding encoding = new AnotherPathEncoding(maximumInterface, innerInterface, 
				outerInterface);
		
		//Log exiting
		logger.exiting(PathEncoding.class.getName(), "getEncoding", encoding);		
		return encoding;
	}
	
	/**
	 * Returns the BDD encoding for a path instance given by the class name {@code name} with a 
	 * maximum interface size of {@code maximumInterface} and a number of colors of {@code colors}. 
	 * If no path encoding with the given name can be found the default encoding is returned. 
	 * @param maximumInterface the maximum interface size
	 * @param innerInterface size of the inner interface
	 * @param outerInterface size of the outer interface
	 * @param encodingName the class name of the dominating set encoding class
	 * @return the encoding given by the class name or the default encoding if no 
	 * path encoding class with the given can be found 
	 */
	public static final PathEncoding getEncoding(int maximumInterface, int innerInterface, 
			int outerInterface, String encodingName) {
		if(maximumInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(outerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(outerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(encodingName==null) {
			throw new NullPointerException(
					RavenExceptionMessages.ENCODING_NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(encodingName.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ENCODING_NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		
		//Log entering 
		logger.entering(PathEncoding.class.getName(), "getEncoding", 
				new Object[]{maximumInterface, innerInterface, outerInterface, encodingName});

		PathEncoding encoding = null;
		Class<?> encodingClass;
		try {
			//Create new path encoding instance
			encodingClass = Class.forName(encodingName);
			Constructor<?> constructor = encodingClass.getDeclaredConstructor(int.class, int.class, 
					int.class);
			encoding = (PathEncoding) constructor.newInstance(maximumInterface, innerInterface,
					outerInterface);
		} catch (ClassNotFoundException e) {
			//Log exception
			logger.throwing(PathEncoding.class.getName(), "getEncoding", e);
		} catch (SecurityException e) {
			//Log exception
			logger.throwing(PathEncoding.class.getName(), "getEncoding", e);
		} catch (NoSuchMethodException e) {
			//Log exception
			logger.throwing(PathEncoding.class.getName(), "getEncoding", e);
		} catch (IllegalArgumentException e) {
			//Log exception
			logger.throwing(PathEncoding.class.getName(), "getEncoding", e);
		} catch (InstantiationException e) {
			//Log exception
			logger.throwing(PathEncoding.class.getName(), "getEncoding", e);
		} catch (IllegalAccessException e) {
			//Log exception
			logger.throwing(PathEncoding.class.getName(), "getEncoding", e);
		} catch (InvocationTargetException e) {
			//Log exception
			logger.throwing(PathEncoding.class.getName(), "getEncoding", e);
		}
		finally {
			//If no instance has been created fall back to the default encoding
			if(encoding==null) {
				encoding = getDefaultEncoding(maximumInterface, innerInterface, 
						outerInterface);
			}
		}

		//Log exiting
		logger.exiting(PathEncoding.class.getName(), "getEncoding", encoding);		
		return encoding;
	}
	
	/**
	 * Returns the type string indicating the (non-)existence of a path from the {@code from}-th 
	 * interface node to the {@code to}-th interface node.
	 * @param from the position of the source node
	 * @param to the position of the target node
	 * @return the path type string
	 */
	public static final String getPathNodeType(int from, int to) {
		return PATH_BIT_TYPE+"-"+to+"-"+from;
	}
	
	/**
	 * Returns the type string indicating the (non-)existence of a path from a source node to
	 * the {@code node}-th interface node.
	 * @param node the position of the node (in the interface)
	 * @return the source path type string
	 */
	public static final String getSourcePathNodeType(int node) {
		return SOURCE_PATH_BIT_TYPE+"-"+node;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Constructs a new encoding for a <i>path automaton</i> with a maximum interface size of 
	 * {@code maximumInterface}, a inner interface size of {@code innerInterface} and a outer 
	 * interface size of {@code outerInterface} checking the existence of a path from a node in 
	 * {@code sourceNodes} to a node in {@code targetNodes}.
	 * @param maximumInterface the maximum interface size
	 * @param innerInterfaceSize size of the inner interface
	 * @param outerInterfaceSize size of the outer interface
	 * @param reachabilityEncodingLength size of the encoding of the reachability information 
	 */
	protected PathEncoding(int maximumInterface, int innerInterfaceSize, int outerInterfaceSize, 
			int reachabilityEncodingLength) {
		super(getBitLength(maximumInterface) + reachabilityEncodingLength + maximumInterface, 
				maximumInterface, innerInterfaceSize, outerInterfaceSize);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Abstract Methods																*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns a BDD which encodes the (non)-existence of a path from the {@code from}-th node 
	 * in the interface to the {@code to}-th node of the interface. 
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param from the position of the interface node
	 * @param to the index of the bit in the path encoding
	 * @return the BDD node representing the bit of the path encoding
	 */
	public abstract RavenBDD getPathBit(StateType type, int from, int to);
		
	/**
	 * Returns the BDD which encodes whether the {@code node}-th interface node is connected to at 
	 * least one of the source nodes.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param node the index of the node in the path encoding
	 * @return the BDD node representing the connection bit of the path encoding
	 */
	public abstract RavenBDD getSourcePathBit(StateType type, int node);

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the BDD which describes the existence of an adjacency bit in the {@code type} domain 
	 * of the BDD
	 * @param type the StateType where the adjacency bit is extracted
	 * @param bdd the BDD which is observed
	 * @param source the position of the source not in the interface
	 * @param target the position of the target not in the interface
	 * @return the BDD representing the existence of the bit in the path encoding
	 */
	public RavenBDD existsPathBit(StateType type, RavenBDD bdd, int source, int target) {
		if(source<0 || source>=maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_INTERFACE_NODE_EXCEPTION);
		}
		if(target<0 || target>=maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_BIT_POSITION_EXCEPTION);
		}
		
		//Create a BDD where only the asked position isn't set to one
		RavenBDD temp = one();
		if(type==StateType.current) {
			for(int n=0;n<maximumInterface;n++) {
				for(int b=0;b<maximumInterface;b++) {
					if(n!=source || b!=target) {
						temp = temp.and(getPathBit(StateType.current,n,b));
					}
					temp = temp.and(getPathBit(StateType.successor,n,b));
				}
			}
			for(int b=0;b<maximumInterface;b++) {
				temp = temp.and(getSourcePathBit(StateType.current, b).and(
						getSourcePathBit(StateType.successor, b)));
			}

		}
		else {
			for(int n=0;n<maximumInterface;n++) {
				for(int b=0;b<maximumInterface;b++) {
					if(n!=source || b!=target) {
						temp = temp.and(getPathBit(StateType.successor,n,b));
					}
					temp = temp.and(getPathBit(StateType.current,n,b));
				}
			}
			for(int b=0;b<maximumInterface;b++){
				temp = temp.and(getSourcePathBit(StateType.current, b).and(
						getSourcePathBit(StateType.successor, b)));
			}
		}

		return bdd.clone().exists(temp); 
	}
}
