/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.encodings;

import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.data.automata.ProductAutomaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDDomain;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDPairing;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class provides a concrete implementation of an encoding for <i>intersection automata</i>.
 * The state encoding is as follows:
 * 
 * <i>b<sub>1</sub> ... b<sub>l</sub> (s<sub>1</sub> ... s<sub>m</sub>) (t<sub>1</sub> ... 
 * t<sub>n</sub>)</i>, where <i>l = log<sub>2</sub>(k)</i> is the binary logarithm of the maximum 
 * interface size <i>k</i>, <i>m</i> is the length of the state information bit sequence of the 
 * first underlying automaton and <i>n</i> is the length of the state information bit sequence of 
 * the second automaton. The block <i>b<sub>1</sub> ... b<sub>l</sub></i> encodes the interface size 
 * to which the current state belongs; the block <i>s<sub>1</sub> ... s<sub>m</sub></i> encodes the 
 * state information of the first underlying automaton; and the block <i>t<sub>1</sub> ... 
 * t<sub>n</sub></i> encodes the state information of the second underlying automata.
 * 
 * @author Christoph Blume
 * @version $Revision: 1.21 $
 *
 * @see ProductAutomaton
 */
public final class ProductEncoding extends BDDEncoding {

	private static final Logger logger = Logger.getLogger(
			VertexCoverEncoding.class.getName());
	
	private Automaton first;
	private Automaton second;
	private RavenBDDPairing pairing;

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the default BDD encoding for a product of two BDD encodings.
	 * @param first the first encoding to be used for the product encoding
	 * @param second the second encoding to be used for the product encoding
	 * @return the default encoding
	 */
	public static final ProductEncoding getEncoding(Automaton first, Automaton second) {
		if(first==null) {
			throw new NullPointerException(
					RavenExceptionMessages.FIRST_AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(second==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SECOND_AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(first==second) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.AUTOMATA_MUST_NOT_BE_EQUAL_EXCEPTION);
		}
		if(!first.getAlphabet().equals(second.getAlphabet())) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ALPHABETS_MUST_BE_EQUAL_EXCEPTION);
		}
		if(first.getMaximumInterface()!=second.getMaximumInterface()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_INTERFACE_SIZES_MUST_NOT_BE_DIFFERENT_EXCEPTION);
		}
		if(first.getInnerInterface()!=second.getInnerInterface()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZES_MUST_NOT_BE_DIFFERENT_EXCEPTION);
		}
		if(first.getOuterInterface()!=second.getOuterInterface()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZES_MUST_NOT_BE_DIFFERENT_EXCEPTION);
		}

		//Log entering 
		logger.entering(ProductEncoding.class.getName(), "getEncoding", 
				new Object[]{first, second});

		ProductEncoding encoding = new ProductEncoding(first, second);
		
		//Log exiting
		logger.exiting(ProductEncoding.class.getName(), "getEncoding", encoding);		
		return encoding;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new product encoding based on the two automata {@code first} and {@code second}.
	 * @param first the first automaton to be used for the product encoding
	 * @param second the second automaton to be used for the product encoding
	 */
	private ProductEncoding(Automaton first, Automaton second) {
		super(first.getEncoding().getLength() + second.getEncoding().getLength() - 
				second.getEncoding().interfaceLength, first.getMaximumInterface(), 
				first.getInnerInterface(), first.getOuterInterface());
		this.first = first;
		this.second = second;
		this.pairing = getPairing(first.getEncoding(), second.getEncoding());
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public String getBitType(int pos) {
		return UNDEFINED_BIT_TYPE;
	}

	@Override
	public String printStates(RavenBDD stateSet) {
		StringBuilder stateString = new StringBuilder();
		
		if(stateSet.satCount(curStateDomain.getDomainBDD())==1) {
			int value = RavenBDDDomain.getInteger(stateSet);
			int interfaceSize = value % (1 << interfaceLength);
			value >>= interfaceLength;
			
			stateString.append(interfaceSize+"-");
			for(int i=0;i<first.getEncoding().getStateLength();i++) {
				stateString.append((value % 2));
				value >>= 1;
			}
			
			stateString.append("-");
			for(int i=0;i<second.getEncoding().getStateLength();i++) {
				stateString.append((value % 2));
				value >>= 1;
			}
			return stateString.toString();
		}

		boolean isFirst = true;
		stateString.append("{");
		for(RavenBDD bdd: stateSet.getSatAssignments(curStateDomain.getDomainBDD())) {
			if(isFirst) {
				isFirst = false;
			}
			else {
				stateString.append(", ");
			}

			stateString.append(printStates(bdd));
		}
		stateString.append("}");
		
		return stateString.toString();
	}
	
	/**
	 * Returns the first automaton.
	 * @return the first automaton
	 */
	public Automaton getFirstAutomaton() {
		return first;
	}
	
	/**
	 * Returns the second automaton.
	 * @return the second automaton
	 */
	public Automaton getSecondAutomaton() {
		return second;
	}
	
	/**
	 * Converts the state set of the given automaton to the {@code UnionEncoding}.
	 * @param automaton the automaton whose states are to be converted (must either be the first or 
	 * the second automaton of the {@code UnionEncoding}
	 * @return a BDD representing the converted state set 
	 */
	public RavenBDD convertStates(Automaton automaton) {
		if(automaton==null) {
			throw new NullPointerException(RavenExceptionMessages.AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(automaton!=first && automaton!=second) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_AUTOMATON_EXCEPTION);
		}

		return convertBDD(automaton.getStates());
	}

	/**
	 * Converts the initial state set of the given automaton to the {@code UnionEncoding}.
	 * @param automaton the automaton whose states are to be converted (must either be the first or 
	 * the second automaton of the {@code UnionEncoding}
	 * @return a BDD representing the converted initial state set 
	 */
	public RavenBDD convertInitialStates(Automaton automaton) {
		if(automaton==null) {
			throw new NullPointerException(RavenExceptionMessages.AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(automaton!=first && automaton!=second) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_AUTOMATON_EXCEPTION);
		}
		
		return convertBDD(automaton.getInitialStates());
	}
	
	/**
	 * Converts the final state set of the given automaton to the {@code UnionEncoding}.
	 * @param automaton the automaton whose states are to be converted (must either be the first or 
	 * the second automaton of the {@code UnionEncoding}
	 * @return a BDD representing the converted final state set 
	 */
	public RavenBDD convertFinalStates(Automaton automaton) {
		if(automaton==null) {
			throw new NullPointerException(RavenExceptionMessages.AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(automaton!=first && automaton!=second) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_AUTOMATON_EXCEPTION);
		}
		
		return convertBDD(automaton.getFinalStates());
	}
	
	/**
	 * Converts the non-final state set of the given automaton to the {@code UnionEncoding}.
	 * @param automaton the automaton whose states are to be converted (must either be the first or 
	 * the second automaton of the {@code UnionEncoding}
	 * @return a BDD representing the converted non-final state set 
	 */
	public RavenBDD convertNonFinalStates(Automaton automaton) {
		if(automaton==null) {
			throw new NullPointerException(RavenExceptionMessages.AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(automaton!=first && automaton!=second) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_AUTOMATON_EXCEPTION);
		}
		
		return convertBDD(automaton.getNonFinalStates());
	}

	/**
	 * Converts the transitions of the given automaton and the given letter to the 
	 * {@code UnionEncoding}.
	 * @param automaton the automaton whose transitions are to be converted (must either be the first 
	 * or the second automaton of the {@code UnionEncoding}
	 * @param letter the letter which indicates the transition
	 * @return a BDD representing the converted non-final state set 
	 */
	public RavenBDD convertTransitions(Automaton automaton, Operation letter) {
		if(automaton==null) {
			throw new NullPointerException(RavenExceptionMessages.AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(automaton!=first && automaton!=second) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_AUTOMATON_EXCEPTION);
		}
		if(letter==null) {
			throw new NullPointerException(RavenExceptionMessages.LETTER_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(!first.getAlphabet().contains(letter)) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_LETTER_EXCEPTION);
		}
		
		return convertBDD(automaton.getTransitionRelation(letter));
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a BDD pairing which pairs the interface bits of the first encoding 
	 * {@code firstEncoding} and the interface bits of the second encoding {@code secondEncoding} 
	 * with the interface bits of product encoding, i.e. the interface bits of the first and second 
	 * encoding will be merged together. The state bits of the first and the state bits of the 
	 * second encoding will then be paired with different bits of the product encoding, thus they 
	 * will not be merged.
	 * @param firstEncoding the encoding of the first automaton
	 * @param secondEncoding the encoding of the second automaton
	 * @return a BDD pairing which pairs the encodings of the two "input" automata.
	 */
	private RavenBDDPairing getPairing(BDDEncoding firstEncoding, BDDEncoding secondEncoding) {
		int interfaceLength = getInterfaceLength();
		int offset = firstEncoding.getLength()-interfaceLength;
		RavenBDDPairing pairing = RavenBDDPairing.makePair();

		for(int i=0;i<interfaceLength;i++) {
			pairing.set(firstEncoding.getInterfaceBit(StateType.current, i), 
					getInterfaceBit(StateType.current, i));
			pairing.set(firstEncoding.getInterfaceBit(StateType.successor, i), 
					getInterfaceBit(StateType.successor, i));

			pairing.set(secondEncoding.getInterfaceBit(StateType.current, i), 
					getInterfaceBit(StateType.current, i));
			pairing.set(secondEncoding.getInterfaceBit(StateType.successor, i), 
					getInterfaceBit(StateType.successor, i));
		}

		for(int i=interfaceLength;i<firstEncoding.getLength();i++) {
			pairing.set(firstEncoding.getBit(StateType.current, i), 
					getBit(StateType.current, i));
			pairing.set(firstEncoding.getBit(StateType.successor, i), 
					getBit(StateType.successor, i));
		}

		for(int i=interfaceLength;i<secondEncoding.getLength();i++) {
			pairing.set(secondEncoding.getBit(StateType.current, i), 
					getBit(StateType.current, offset+i));
			pairing.set(secondEncoding.getBit(StateType.successor, i), 
					getBit(StateType.successor, offset+i));
		}

		return pairing;
	}
	
	private RavenBDD convertBDD(RavenBDD stateSet) {
		return stateSet.replace(pairing);
	}

}
