/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.encodings;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.List;
import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.data.Edge;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.Node;
import de.uni_due.inf.ti.raven.data.automata.SubgraphAutomaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDDomain;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;


/**
 * This class represents the encoding of state space and the transition functions of <i>subgraph 
 * automata</i> and provides methods to access the encoding.
 * 
 * <p><i>Note:</i> The concrete implementation of a subgraph encoding can be found in classes 
 * extending this abstract class.</p>
 * 
 * @author Christoph Blume
 * @version $Revision: 1.25 $
 *
 * @see DefaultSubgraphEncoding
 * @see SubgraphAutomaton
 */
public abstract class SubgraphEncoding extends BDDEncoding {

	private static final String EDGE_BIT_TYPE = "EdgeType";
	private static final String NODE_BIT_TYPE = "NodeType";
	private static final String MAPPING_BIT_TYPE = "MappingType";

	private static final Logger logger = Logger.getLogger(
			SubgraphEncoding.class.getName());

	/** The wanted subgraph */
	protected final Graph subgraph;

	/** The node order */
	protected final List<Node> nodeOrder;

	/** The edge order */
	protected final List<Edge> edgeOrder;

	/** The number of nodes */
	protected final int nodeCount;

	/** The number of edges */
	protected final int edgeCount; 

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the default BDD encoding for a subgraph instance with a 
	 * maximum interface size of {@code maximumInterface} and a wanted 
	 * subgraph {@code subgraph}.
	 * @param maximumInterface the maximum interface size
	 * @param innerInterface size of the inner interface
	 * @param outerInterface size of the outer interface
	 * @param subgraph the wanted subgraph
	 * @param nodeOrder the node order
	 * @param edgeOrder the edge order
	 * @return the default encoding
	 */
	public static final SubgraphEncoding getDefaultEncoding(int maximumInterface, int innerInterface, 
			int outerInterface, Graph subgraph, List<Node> nodeOrder, List<Edge> edgeOrder) {
		if(maximumInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(outerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(outerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(subgraph==null) {
			throw new NullPointerException(RavenExceptionMessages.GRAPH_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(nodeOrder==null) {
			throw new NullPointerException(
					RavenExceptionMessages.NODE_ORDER_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(!subgraph.getNodes().containsAll(nodeOrder) || 
				!nodeOrder.containsAll(subgraph.getNodes())) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_NODE_ORDER_EXCEPTION);
		}
		if(edgeOrder==null) {
			throw new NullPointerException(RavenExceptionMessages.EDGE_ORDER_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(!subgraph.getEdges().containsAll(edgeOrder) || 
				!edgeOrder.containsAll(subgraph.getEdges())) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_EDGE_ORDER_EXCEPTION);
		}

		//Log entering 
		logger.entering(SubgraphEncoding.class.getName(), "getEncoding", 
				new Object[]{maximumInterface, innerInterface, outerInterface, subgraph, nodeOrder, 
								edgeOrder});

		SubgraphEncoding encoding = new DefaultSubgraphEncoding(maximumInterface, innerInterface, 
				outerInterface, subgraph, nodeOrder, edgeOrder);

		//Log exiting
		logger.exiting(SubgraphEncoding.class.getName(), "getEncoding", encoding);		
		return encoding;
	}

	/**
	 * Returns the BDD encoding for a subgraph instance given by the class name
	 * {@code name} with a maximum interface size of {@code maximumInterface} and 
	 * a wanted subgraph {@code subgraph} If no subgraph encoding with the given 
	 * name can be found the default encoding is returned.
	 * @param maximumInterface the maximum interface size
	 * @param innerInterface size of the inner interface
	 * @param outerInterface size of the outer interface
	 * @param subgraph the wanted subgraph
	 * @param nodeOrder the node order
	 * @param edgeOrder the edge order
	 * @param encodingName the class name of the subgraph encoding class
	 * @return the encoding given by the class name or the default encoding if no 
	 * dominating set encoding class with the given can be found 
	 */
	public static final SubgraphEncoding getEncoding(int maximumInterface, int innerInterface, 
			int outerInterface, Graph subgraph, List<Node> nodeOrder, List<Edge> edgeOrder, 
			String encodingName) {
		if(maximumInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(innerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(outerInterface<0) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_NEGATIVE_EXCEPTION);
		}
		if(outerInterface>maximumInterface) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZE_MUST_NOT_BE_GREATER_THAN_THE_MAXIMUM_INTERFACE_EXCEPTION);
		}
		if(subgraph==null) {
			throw new NullPointerException(RavenExceptionMessages.GRAPH_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(nodeOrder==null) {
			throw new NullPointerException(
					RavenExceptionMessages.NODE_ORDER_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(!subgraph.getNodes().containsAll(nodeOrder) || 
				!nodeOrder.containsAll(subgraph.getNodes())) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_NODE_ORDER_EXCEPTION);
		}
		if(edgeOrder==null) {
			throw new NullPointerException(RavenExceptionMessages.EDGE_ORDER_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(!subgraph.getEdges().containsAll(edgeOrder) || 
				!edgeOrder.containsAll(subgraph.getEdges())) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_EDGE_ORDER_EXCEPTION);
		}
		if(encodingName==null) {
			throw new NullPointerException(
					RavenExceptionMessages.ENCODING_NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(encodingName.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ENCODING_NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}

		//Log entering 
		logger.entering(SubgraphEncoding.class.getName(), "getEncoding", 
				new Object[]{maximumInterface, innerInterface, outerInterface, subgraph, nodeOrder, 
								edgeOrder, encodingName});

		SubgraphEncoding encoding = null;
		Class<?> encodingClass;
		try {
			//Create new dominating set encoding instance
			encodingClass = Class.forName(encodingName);
			Constructor<?> constructor = encodingClass.getDeclaredConstructor(int.class, int.class);
			encoding = (SubgraphEncoding) constructor.newInstance(maximumInterface, subgraph);
		} catch (ClassNotFoundException e) {
			//Log exception
			logger.throwing(SubgraphEncoding.class.getName(), "getEncoding", e);
		} catch (SecurityException e) {
			//Log exception
			logger.throwing(SubgraphEncoding.class.getName(), "getEncoding", e);
		} catch (NoSuchMethodException e) {
			//Log exception
			logger.throwing(SubgraphEncoding.class.getName(), "getEncoding", e);
		} catch (IllegalArgumentException e) {
			//Log exception
			logger.throwing(SubgraphEncoding.class.getName(), "getEncoding", e);
		} catch (InstantiationException e) {
			//Log exception
			logger.throwing(SubgraphEncoding.class.getName(), "getEncoding", e);
		} catch (IllegalAccessException e) {
			//Log exception
			logger.throwing(SubgraphEncoding.class.getName(), "getEncoding", e);
		} catch (InvocationTargetException e) {
			//Log exception
			logger.throwing(SubgraphEncoding.class.getName(), "getEncoding", e);
		}
		finally {
			//If no instance has been created fall back to the default encoding
			if(encoding==null) {
				encoding = getDefaultEncoding(maximumInterface, innerInterface, outerInterface, 
						subgraph, nodeOrder, edgeOrder);
			}
		}

		//Log exiting
		logger.exiting(SubgraphEncoding.class.getName(), "getEncoding", encoding);		
		return encoding;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Constructs a new encoding for a <i>subgraph automaton</i> with a maximum interface 
	 * size of {@code maximumInterface}, a wanted subgraph {@code subgraph}, node ordering 
	 * {@code nodeOrder} and edge ordering {@code edgeOrder}.
	 * @param maximumInterface the maximum interface size
	 * @param subgraph the wanted subgraph
	 * @param nodeOrder the node order
	 * @param edgeOrder the edge order
	 * @param innerInterface size of the inner interface
	 * @param outerInterface size of the outer interface
	 */
	protected SubgraphEncoding(int maximumInterface, Graph subgraph, List<Node> nodeOrder, 
			List<Edge> edgeOrder, int innerInterface, int outerInterface) {
		super(getBitLength(maximumInterface) + subgraph.getEdgeCount() + 
				(maximumInterface+1)*subgraph.getNodeCount(), maximumInterface, innerInterface,
				outerInterface);
		this.subgraph = subgraph;
		this.nodeOrder = nodeOrder;
		this.edgeOrder = edgeOrder;
		nodeCount = subgraph.getNodeCount();
		edgeCount = subgraph.getEdgeCount();
	}

	/**
	 * Returns the type string indicating the existence bits for the {@code edge}-th edge.
	 * @param edge the position of the edge
	 * @return the edge existence type string
	 */
	public static final String getEdgeType(int edge) {
		return EDGE_BIT_TYPE+"-"+edge;
	}

	/**
	 * Returns the type string indicating the existence bits for the {@code node}-th node.
	 * @param node the position of the node
	 * @return the node existence type string
	 */
	public static final String getNodeType(int node) {
		return NODE_BIT_TYPE+"-"+node;
	}

	/**
	 * Returns the type string indicating the bit for the mapping of the {@code node}-th node 
	 * to the {@code pos}-th interface position.
	 * @param node the position of the node
	 * @param pos  the position in the interface
	 * @return the mapping type string
	 */
	public static final String getMappingType(int node, int pos) {
		return MAPPING_BIT_TYPE+"-"+node+"-"+pos;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Abstract Methods																*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns a BDD which encodes that in state {@code state} the {@code node}-th subgraph node 
	 * has been recognized if and only if the BDD is {@code true}.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param node the position of the subgraph node
	 * @return a BDD encoding that the subgraph node has been recognized
	 */
	public abstract RavenBDD getNodeBit(StateType type, int node);

	/**
	 * Returns a BDD which encodes that in state {@code state} the {@code edge}-th subgraph edge 
	 * has been recognized if and only if the BDD is {@code true}.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param edge the position of the subgraph edge
	 * @return a BDD encoding that the subgraph edge has been recognized
	 */
	public abstract RavenBDD getEdgeBit(StateType type, int edge);

	/**
	 * Returns a BDD which encodes that in state {@code state} the {@code node}-th subgraph node 
	 * is the image of {@code pos}-th interface node (of the function mapping the interface into 
	 * the subgraph node set) if and only if the BDD is {@code true}.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param node the position of the subgraph node
	 * @param pos the position of the interface node
	 * @return a BDD encoding that the interface node is mapped to the subgraph node
	 */
	public abstract RavenBDD getMappingBit(StateType type, int node, int pos);

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the wanted subgraph.
	 * @return the wanted subgraph
	 */
	public final Graph getSubgraph() {
		return subgraph;
	}

	/**
	 * Returns the number of nodes of the wanted subgraph.
	 * @return the number of nodes
	 */
	public final int getNodeCount() {
		return subgraph.getNodeCount();
	}

	/**
	 * Returns the number of edges of the wanted subgraph.
	 * @return the number of edges
	 */
	public final int getEdgeCount() {
		return subgraph.getEdgeCount();
	}

	/**
	 * Returns the node order of this encoding.
	 * @return the node order
	 */
	public final List<Node> getNodeOrder() {
		return nodeOrder;
	}

	/**
	 * Returns the edge order of this encoding.
	 * @return the edge order
	 */
	public final List<Edge> getEdgeOrder() {
		return edgeOrder;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns {@code true} if and only if the {@code edge}-th edge is set in the state encoded by 
	 * the BDD {@code state}.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param edge the index of the edge
	 * @param state the BDD encoding the state
	 * @return {@code true} if the edge is set, {@code false} otherwise
	 */
	protected final boolean hasEdge(StateType type, int edge, RavenBDD state) {
		RavenBDD vars;
		RavenBDDDomain domain = type==StateType.current ? curStateDomain : sucStateDomain;
		boolean isExisting = false;
		
		vars = domain.getDomainBDD();
		if(state.satCount(vars)==1) {
			state = state.clone().and(getEdgeBit(type, edge));
			isExisting = !state.isZero();
			state.free();
		}
		return isExisting;
	}
	
	/**
	 * Returns {@code true} if and only if the {@code node}-th node is set in the state encoded by 
	 * the BDD {@code state}.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param node the index of the node
	 * @param state the BDD encoding the state
	 * @return {@code true} if the node is set, {@code false} otherwise
	 */
	protected final boolean hasNode(StateType type, int node, RavenBDD state) {
		RavenBDD vars;
		RavenBDDDomain domain = type==StateType.current ? curStateDomain : sucStateDomain;
		boolean isExisting = false;
		
		vars = domain.getDomainBDD();
		if(state.satCount(vars)==1) {
			state = state.clone().and(getNodeBit(type, node));
			isExisting = !state.isZero();
			state.free();
		}
		return isExisting;
	}
	
	/**
	 * Returns the index of the subgraph node (according to the node order) to which the 
	 * {@code pos}-th interface node is mapped.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @param pos the position of the interface node
	 * @param state the BDD encoding the state
	 * @return the index of the subgraph node or {@code -1} if the interface node is not in the image
	 * of the function 
	 */
	protected final int getSubgraphNode(StateType type, int pos, RavenBDD state) {
		RavenBDD vars;
		RavenBDDDomain domain = type==StateType.current ? curStateDomain : sucStateDomain;
		int index = -1;
		
		vars = domain.getDomainBDD();
		if(state.satCount(vars)==1) {
			RavenBDD bdd;
			for(int i=0;i<nodeCount;i++) {
				bdd = state.clone().and(getMappingBit(type, i, pos));
				
				boolean result = bdd.isZero();
				bdd.free();
				if(!result) {
					index = i;
					break;
				};
			}
			
		}
		return index;
	}

}
