/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.encodings;

import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.data.automata.UnionAutomaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDDomain;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDPairing;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class provides a concrete implementation of an encoding for <i>union automata</i>.
 * The state encoding is as follows:
 * 
 * <i>b<sub>1</sub> ... b<sub>l</sub> d (s<sub>1</sub> ... s<sub>m</sub>)</i>, where <i>l = 
 * log<sub>2</sub>(k)</i> is the binary logarithm of the maximum interface size <i>k</i> and 
 * <i>m</i> is the maximum length of the state information bit sequences of both underlying 
 * automata. The block <i>b<sub>1</sub> ... b<sub>l</sub></i> encodes the interface size to which 
 * the current state belongs; the bit <i>p</i> is a distinction bit which is used to distinguish 
 * the state information of the first automaton from the state information of the second automaton; 
 * and the block <i>s<sub>1</sub>... s<sub>m</sub></i> encodes the state information of the two 
 * underlying automata in the current state.
 * 
 * @author Christoph Blume
 * @version $Revision: 1.23 $
 *
 * @see UnionAutomaton
 */
public final class UnionEncoding extends BDDEncoding {
	
	private static final Logger logger = Logger.getLogger(
			VertexCoverEncoding.class.getName());
	
	private Automaton first;
	private Automaton second;
	private RavenBDDPairing firstPairing;
	private RavenBDDPairing secondPairing;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the default BDD encoding for a union of two automata.
	 * @param first the first automaton to be used for the union encoding
	 * @param second the second automaton to be used for the union encoding
	 * @return the default encoding
	 */
	public static final UnionEncoding getEncoding(Automaton first, Automaton second) {
		if(first==null) {
			throw new NullPointerException(
					RavenExceptionMessages.FIRST_AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(second==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SECOND_AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(first==second) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.AUTOMATA_MUST_NOT_BE_EQUAL_EXCEPTION);
		}
		if(!first.getAlphabet().equals(second.getAlphabet())) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ALPHABETS_MUST_BE_EQUAL_EXCEPTION);
		}
		if(first.getMaximumInterface()!=second.getMaximumInterface()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.MAXIMUM_INTERFACE_SIZES_MUST_NOT_BE_DIFFERENT_EXCEPTION);
		}
		if(first.getInnerInterface()!=second.getInnerInterface()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INNER_INTERFACE_SIZES_MUST_NOT_BE_DIFFERENT_EXCEPTION);
		}
		if(first.getOuterInterface()!=second.getOuterInterface()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.OUTER_INTERFACE_SIZES_MUST_NOT_BE_DIFFERENT_EXCEPTION);
		}

		//Log entering 
		logger.entering(UnionEncoding.class.getName(), "getEncoding", 
				new Object[]{first, second});

		UnionEncoding encoding = new UnionEncoding(first, second);
		
		//Log exiting
		logger.exiting(UnionEncoding.class.getName(), "getEncoding", encoding);		
		return encoding;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new union encoding based on the two automata {@code first} and {@code second}.
	 * @param first the first automaton to be used for the union encoding
	 * @param second the second automaton to be used for the union encoding
	 */
	private UnionEncoding(Automaton first, Automaton second) {
		super(Math.max(first.getEncoding().getLength(), second.getEncoding().getLength())+1, 
				first.getMaximumInterface(), first.getInnerInterface(), first.getOuterInterface());
		
		this.first = first;
		this.second = second;
		this.firstPairing = getPairing(first.getEncoding());
		this.secondPairing = getPairing(second.getEncoding());
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns a BDD which encodes that in state {@code state} the separation bit has been set if 
	 * and only if the BDD is {@code true}.
	 * @param type the state type, either <i>current state</i> or <i>successor state</i>
	 * @return a BDD encoding that the separation bit has been set
	 */
	public RavenBDD getSeparationBit(StateType type) {
		RavenBDDDomain domain;
		
		if(type==StateType.current) {
			domain = curStateDomain;
		}
		else {
			domain = sucStateDomain;
		}		
		return domain.getIthBDD(interfaceLength);
	}
	
	@Override
	public String getBitType(int pos) {
		return UNDEFINED_BIT_TYPE;
	}

	@Override
	public String printStates(RavenBDD stateSet) {
		StringBuilder stateString = new StringBuilder();
		
		if(stateSet.satCount(curStateDomain.getDomainBDD())==1) {
			int value = RavenBDDDomain.getInteger(stateSet);
			int interfaceSize = value % (1 << interfaceLength);
			value >>= interfaceLength;
			
			stateString.append(interfaceSize+"-");
			stateString.append((value%2)+"-");
			value >>= 1;

			for(int i=0;i<length-(interfaceLength+1);i++) {
				stateString.append((value % 2));
				value >>= 1;
			}
			return stateString.toString();
		}

		boolean isFirst = true;		
		for(RavenBDD bdd: stateSet.getSatAssignments(curStateDomain.getDomainBDD())) {
			if(isFirst) {
				isFirst = false;
			}
			else {
				stateString.append(", ");
			}

			stateString.append(printStates(bdd));
		}
		
		return stateString.toString();
	}
	
	/**
	 * Returns the first automaton.
	 * @return the first automaton
	 */
	public Automaton getFirstAutomaton() {
		return first;
	}
	
	/**
	 * Returns the second automaton.
	 * @return the second automaton
	 */
	public Automaton getSecondAutomaton() {
		return second;
	}
	
	/**
	 * Converts the state set of the given automaton to the {@code UnionEncoding}.
	 * @param automaton the automaton whose states are to be converted (must either be the first or 
	 * the second automaton of the {@code UnionEncoding}
	 * @return a BDD representing the converted state set 
	 */
	public RavenBDD convertStates(Automaton automaton) {		
		if(automaton==null) {
			throw new NullPointerException(RavenExceptionMessages.AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(automaton!=first && automaton!=second) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_AUTOMATON_EXCEPTION);
		}
		
		return convertBDD(automaton, automaton.getStates());
	}

	/**
	 * Converts the initial state set of the given automaton to the {@code UnionEncoding}.
	 * @param automaton the automaton whose states are to be converted (must either be the first or 
	 * the second automaton of the {@code UnionEncoding}
	 * @return a BDD representing the converted initial state set 
	 */
	public RavenBDD convertInitialStates(Automaton automaton) {
		if(automaton==null) {
			throw new NullPointerException(RavenExceptionMessages.AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(automaton!=first && automaton!=second) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_AUTOMATON_EXCEPTION);
		}
		
		return convertBDD(automaton, automaton.getInitialStates());
	}
	
	/**
	 * Converts the final state set of the given automaton to the {@code UnionEncoding}.
	 * @param automaton the automaton whose states are to be converted (must either be the first or 
	 * the second automaton of the {@code UnionEncoding}
	 * @return a BDD representing the converted final state set 
	 */
	public RavenBDD convertFinalStates(Automaton automaton) {
		if(automaton==null) {
			throw new NullPointerException(RavenExceptionMessages.AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(automaton!=first && automaton!=second) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_AUTOMATON_EXCEPTION);
		}
		
		return convertBDD(automaton, automaton.getFinalStates());
	}
	
	/**
	 * Converts the non-final state set of the given automaton to the {@code UnionEncoding}.
	 * @param automaton the automaton whose states are to be converted (must either be the first or 
	 * the second automaton of the {@code UnionEncoding}
	 * @return a BDD representing the converted non-final state set 
	 */
	public RavenBDD convertNonFinalStates(Automaton automaton) {
		if(automaton==null) {
			throw new NullPointerException(RavenExceptionMessages.AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(automaton!=first && automaton!=second) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_AUTOMATON_EXCEPTION);
		}
		
		return convertBDD(automaton, automaton.getNonFinalStates());
	}

	/**
	 * Converts the transitions of the given automaton and the given letter to the 
	 * {@code UnionEncoding}.
	 * @param automaton the automaton whose transitions are to be converted (must either be the first 
	 * or the second automaton of the {@code UnionEncoding}
	 * @param letter the letter which indicates the transition
	 * @return a BDD representing the converted non-final state set 
	 */
	public RavenBDD convertTransitions(Automaton automaton, Operation letter) {
		if(automaton==null) {
			throw new NullPointerException(RavenExceptionMessages.AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(automaton!=first && automaton!=second) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_AUTOMATON_EXCEPTION);
		}
		if(letter==null) {
			throw new NullPointerException(RavenExceptionMessages.LETTER_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(!first.getAlphabet().contains(letter)) {
			throw new IllegalArgumentException(RavenExceptionMessages.ILLEGAL_LETTER_EXCEPTION);
		}
		
		RavenBDD result = automaton.getTransitionRelation(letter);
		if(automaton==first) {
			result = result.replace(firstPairing);
			result = result.and(getSeparationBit(StateType.current).not());
			result = result.and(getSeparationBit(StateType.successor).not());
		}
		else {
			result = result.replace(secondPairing);
			result = result.and(getSeparationBit(StateType.current));
			result = result.and(getSeparationBit(StateType.successor));
		}
		
		if(automaton.getEncoding().getStateLength()+1<getStateLength()) {
			for(int i=automaton.getEncoding().getStateLength()+1;i<getStateLength();i++) {
				result = result.and(getStateBit(StateType.current, i).not());
				result = result.and(getStateBit(StateType.successor, i).not());
			}
		}
		
		return result;
	}
	
	
	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a BDD pairing which pairs the interface bits of the template encoding 
	 * {@code templateEncoding} with the interface bits of union encoding. The state bits of the 
	 * template encoding will then be paired with the next "free" bits of the union encoding.
	 * @param templateEncoding the template encoding
	 * @return a BDD pairing which pairs the template encoding with the union encoding.
	 */
	private RavenBDDPairing getPairing(BDDEncoding templateEncoding) {
		int interfaceLength = getInterfaceLength();
		RavenBDDPairing pairing = RavenBDDPairing.makePair();

		for(int i=0;i<interfaceLength;i++) {
			pairing.set(templateEncoding.getInterfaceBit(StateType.current, i), 
					getInterfaceBit(StateType.current, i));
			pairing.set(templateEncoding.getInterfaceBit(StateType.successor, i), 
					getInterfaceBit(StateType.successor, i));
		}

		for(int i=0;i<templateEncoding.getStateLength();i++) {
			pairing.set(templateEncoding.getStateBit(StateType.current, i), 
					getStateBit(StateType.current, i+1));
			pairing.set(templateEncoding.getStateBit(StateType.successor, i), 
					getStateBit(StateType.successor, i+1));
		}
		
		return pairing;
	}

	private RavenBDD convertBDD(Automaton automaton, RavenBDD stateSet) {
		RavenBDDPairing pairing;
		RavenBDD seperationBit;
		if(automaton==first) {
			pairing = firstPairing;
			seperationBit = getSeparationBit(StateType.current).not();
		}
		else {
			pairing = secondPairing;
			seperationBit = getSeparationBit(StateType.current);
		}
		
		RavenBDD result = stateSet.replace(pairing);
		result = result.and(seperationBit);
		
		if(automaton.getEncoding().getStateLength()+1<getStateLength()) {
			for(int i=automaton.getEncoding().getStateLength()+1;i<getStateLength();i++) {
				result = result.and(getStateBit(StateType.current, i).not());
			}
		}
		
		return result;
	}
	
}
