/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.gui;

import java.awt.BorderLayout;
import java.awt.Window;
import java.io.File;
import java.io.IOException;

import javax.imageio.ImageIO;
import javax.swing.JFrame;
import javax.swing.JPanel;

import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.encodings.BDDEncoding;
import de.uni_due.inf.ti.raven.gui.graphs.BDDView;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * This class provides a dialog to show a given BDD. Please note that the dialog is modal, i.e. 
 * the {@code showDialog} method blocks the caller until the user's interaction is complete.
 * @author Christoph Blume
 * @version $Revision: 1.23 $
 *
 */
public class BDDDialog extends JFrame {

	private static final long serialVersionUID = -3427428915126561034L;
	private static GuiMessages messages;
	
	private final JPanel bddPanel;
	private BDDView bddView;

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Shows the dialog for displaying BDDs which encode transitions of the automaton 
	 * {@code automaton} for the letter {@code letter}.
	 * @param owner the parent window of the dialog
	 * @param automaton the automaton whose information will be displayed
	 * @param letter the letter whose transition will be displayed
	 */
	public static void showDialog(Window owner, Automaton automaton, Operation letter) {
		messages = GuiMessages.getMessages();
		
		BDDDialog dialog = new BDDDialog(owner);
		dialog.setTitle(messages.getString(GuiMessages.BDD_DIALOG_TITLE_KEY));
		dialog.displayBDD(automaton.getTransitionRelation(letter), automaton.getEncoding());
		dialog.pack();
		dialog.setLocationRelativeTo(null);
		dialog.setVisible(true);
	}

	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new dialog for displaying BDD information.
	 * @param owner the parent window of this dialog
	 */
	private BDDDialog(Window owner) {
		bddPanel = createBDDPanel();
		getRootPane().setLayout(new BorderLayout());
		getRootPane().add(bddPanel, BorderLayout.CENTER);

		try {
			setIconImage(ImageIO.read(new File(RavenMain.RAVEN_ICON_FILENAME)));
		} catch (IOException e) {
			//Do nothing
		}
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods 																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Displays the BDD {@code bdd} using the BDD Encoding {@code coding} to highlight 
	 * additional informations.
	 * @param bdd the BDD to be displayed
	 * @param encoding the encoding used to encode the BDD
	 */
	public void displayBDD(RavenBDD bdd, BDDEncoding encoding) {
		bddView = new BDDView(bddPanel, encoding);
		bddPanel.add(bddView, BorderLayout.CENTER);
		
		bddView.print(bdd);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods 																	*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates the panel on which the BDD will be displayed.
	 * @return the BDD panel
	 */
	private static JPanel createBDDPanel() {
		JPanel panel = new JPanel();
		panel.setLayout(new BorderLayout());

		return panel;
	}

}
