/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui;

import java.util.LinkedList;
import java.util.List;

import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.SwingWorker;
import javax.swing.SwingWorker.StateValue;

import com.jgoodies.forms.builder.DefaultFormBuilder;

import de.uni_due.inf.ti.raven.actions.ComputationEvent;
import de.uni_due.inf.ti.raven.actions.ComputationListener;

/**
 * This abstract class implements a panel for the invocation of different algorithms.
 * @author Christoph Blume
 * @version $Revision: 1.9 $
 *
 * @see ControlPanelFactory
 * @see DecompositionControlPanel
 * @see EmptinessControlPanel
 * @see InvariantCheckingControlPanel
 * @see LanguageInclusionControlPanel
 * @see MembershipControlPanel
 * @see SimulationControlPanel
 * @see UniversalityControlPanel
 */
public abstract class ControlPanel extends JPanel {

	/**
	 * Abstract Worker Thread for the execution of algorithms controlled by the control panel.
	 */
	public static abstract class WorkerThread extends SwingWorker<Void, Void> {}
	
	private static final long serialVersionUID = 7740580232384758309L;

	private static final String LABEL_SUFFIX = ": ";

	/** */
	protected static final GuiMessages messages = GuiMessages.getMessages();
	
	/** The list of listeners which are registered for events of the worker threads */
	protected List<ComputationListener> listeners = new LinkedList<>();

	private WorkerThread workerThread;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Adds a text label and component to the panel.
	 * @param builder the default form builder which is used to add the control components 
     * @param label  the label text
     * @param component the component to be added
	 */
	protected static void append(DefaultFormBuilder builder, String label, JComponent component) {
		builder.append(messages.getString(label) + LABEL_SUFFIX, component);
		builder.nextLine();
	}
	
	/**
	 * Adds a component to the panel that spans a specified number of columns
	 * @param builder the default form builder which is used to add the control components 
	 * @param component the component to be added
	 * @param columnSpan the size of the column span
	 */
	protected static void append(DefaultFormBuilder builder, JComponent component, int columnSpan) {
		builder.append(component, columnSpan);
		builder.nextLine();
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates the panel on which controls for the different algorithms will be displayed.
	 */
	protected ControlPanel() {
		setOpaque(true);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Abstract Methods																	*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Initializes the control panel and adds the control components using the default form builder
	 * {@code builder}
	 * @param builder the default form builder which is used to add the control components 
	 */
	protected abstract void initialize(DefaultFormBuilder builder);
	
	/**
	 * Returns a new instance of the worker thread which is used to invoke the underlying algorithms.
	 * @return a new worker thread instance
	 */
	protected abstract WorkerThread getWorkerThread();
	
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Adds the computation listener {@code listener} to the underlying control panel.
	 * @param listener the listener to receive computation events
	 */
	public void addComputationListener(ComputationListener listener) {
		listeners.add(listener);
	}
	
	/**
	 * Removes the computation listener {@code listener} from the underlying control panel.
	 * @param listener the listener to be removed
	 */
	public void removeComputationListener(ComputationListener listener) {
		listeners.remove(listener);
	}
	
	/**
	 * Registers the {@code ComputationListener} and starts the currently set worker thread.
	 */
	public void startWorkerThread() {
		if(workerThread!=null && workerThread.getState().equals(StateValue.STARTED)) {
			stopWorkerThread();
		}
		
		try {
			workerThread = getWorkerThread();
			workerThread.execute();
		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Stops the currently executed worker thread, if it is executed.
	 */
	public void stopWorkerThread() {
		if(workerThread!=null && !workerThread.isDone()) {
			workerThread.cancel(true);
		}
	}
	
	
	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Fires a computation event notifying all listeners.
	 * 
	 * @param e the computation event
	 */
	protected void fireComputationEvent(ComputationEvent e) {
		for (ComputationListener l : listeners) {
			l.computationAdvanded(e);
		}
	}
	
}
