/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui;

import com.jgoodies.forms.builder.DefaultFormBuilder;
import com.jgoodies.forms.factories.Borders;
import com.jgoodies.forms.layout.FormLayout;

/**
 * This class implements a factory for the creation of {@code ControlPanel}-objects for the 
 * different algorithms of Raven.
 * @author Christoph Blume
 * @version $Revision: 1.4 $
 *
 */
public final class ControlPanelFactory {
	
	private static final String LAYOUT_SPECIFICATION = "left:max(40dlu;pref), 3dlu, 40dlu:grow";
	
	private ControlPanelFactory() {}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns a new graph decomposition control panel created by this {@code ControlPanelFactory}.
	 * @return the graph decomposition control panel
	 */
	public static DecompositionControlPanel getDecompositionControlPanel() {
		DecompositionControlPanel panel = new DecompositionControlPanel();
		panel.setBackground(MainWindow.BACKGROUND_COLOR);
		
		FormLayout layout = new FormLayout(LAYOUT_SPECIFICATION, "");
		DefaultFormBuilder builder = new DefaultFormBuilder(layout, panel);
		builder.border(Borders.DLU4);
		panel.initialize(builder);
		
		return panel;		
	}

	/**
	 * Returns a new emptiness checking control panel created by this {@code ControlPanelFactory}.
	 * @return the emptiness checking control panel
	 */
	public static ControlPanel getEmptinessControlPanel() {
		EmptinessControlPanel panel = new EmptinessControlPanel();
		panel.setBackground(MainWindow.BACKGROUND_COLOR);
		
		FormLayout layout = new FormLayout(LAYOUT_SPECIFICATION, "");
		DefaultFormBuilder builder = new DefaultFormBuilder(layout, panel);
		builder.border(Borders.DLU4);
		panel.initialize(builder);
		
		return panel;
	}

	/**
	 * Returns a new invariant checking control panel created by this {@code ControlPanelFactory}.
	 * @return the invariant checking control panel
	 */
	public static ControlPanel getInvariantCheckingControlPanel() {
		InvariantCheckingControlPanel panel = new InvariantCheckingControlPanel();
		panel.setBackground(MainWindow.BACKGROUND_COLOR);
		
		FormLayout layout = new FormLayout(LAYOUT_SPECIFICATION, "");
		DefaultFormBuilder builder = new DefaultFormBuilder(layout, panel);
		builder.border(Borders.DLU4);
		panel.initialize(builder);
		
		return panel;
	}

	/**
	 * Returns a new language inclusion checking control panel created by this 
	 * {@code ControlPanelFactory}.
	 * @return the language inclusion checking control panel
	 */
	public static ControlPanel getLanguageInclusionControlPanel() {
		LanguageInclusionControlPanel panel = new LanguageInclusionControlPanel();
		panel.setBackground(MainWindow.BACKGROUND_COLOR);
		
		FormLayout layout = new FormLayout(LAYOUT_SPECIFICATION, "");
		DefaultFormBuilder builder = new DefaultFormBuilder(layout, panel);
		builder.border(Borders.DLU4);
		panel.initialize(builder);
		
		return panel;
	}

	/**
	 * Returns a new membership checking control panel created by this {@code ControlPanelFactory}.
	 * @return the membership checking control panel
	 */
	public static ControlPanel getMembershipControlPanel() {
		MembershipControlPanel panel = new MembershipControlPanel();
		panel.setBackground(MainWindow.BACKGROUND_COLOR);
		
		FormLayout layout = new FormLayout(LAYOUT_SPECIFICATION, "");
		DefaultFormBuilder builder = new DefaultFormBuilder(layout, panel);
		builder.border(Borders.DLU4);
		panel.initialize(builder);
		
		return panel;
	}

	/**
	 * Returns a new simulation computation control panel created by this {@code ControlPanelFactory}.
	 * @return the simulation computation control panel
	 */
	public static ControlPanel getSimulationControlPanel() {
		SimulationControlPanel panel = new SimulationControlPanel();
		panel.setBackground(MainWindow.BACKGROUND_COLOR);
		
		FormLayout layout = new FormLayout(LAYOUT_SPECIFICATION, "");
		DefaultFormBuilder builder = new DefaultFormBuilder(layout, panel);
		builder.border(Borders.DLU4);
		panel.initialize(builder);
		
		return panel;
	}

	/**
	 * Returns a new universality control panel created by this {@code ControlPanelFactory}.
	 * @return the universality control panel
	 */
	public static ControlPanel getUniversalityControlPanel() {
		UniversalityControlPanel panel = new UniversalityControlPanel();
		panel.setBackground(MainWindow.BACKGROUND_COLOR);
		
		FormLayout layout = new FormLayout(LAYOUT_SPECIFICATION, "");
		DefaultFormBuilder builder = new DefaultFormBuilder(layout, panel);
		builder.border(Borders.DLU4);
		panel.initialize(builder);
		
		return panel;
	}
	
}
