/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;
import javax.swing.SpringLayout;
import javax.swing.SwingConstants;

import com.jgoodies.forms.builder.DefaultFormBuilder;
import com.jgoodies.forms.factories.Borders;
import com.jgoodies.forms.layout.FormLayout;
import com.mxgraph.model.mxCell;

import de.uni_due.inf.ti.raven.actions.AddToRepositoryAction;
import de.uni_due.inf.ti.raven.data.Cospan;
import de.uni_due.inf.ti.raven.data.Edge;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.Node;
import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.gui.NewEdgeDialog.EdgeData;
import de.uni_due.inf.ti.raven.gui.graphs.GraphView;
import de.uni_due.inf.ti.raven.gui.graphs.VisualGraph;

/**
 * This class provides a dialog to create a new cospan.
 * @author Christoph Blume, Dennis Nolte
 * @version $Revision: 1.22 $
 * 
 */
public class CospanCreationDialog extends RavenDialog {

	private static final String LAYOUT_SPECIFICATION_LEFT = "40dlu:grow, 3dlu, center:pref";
	private static final String LAYOUT_SPECIFICATION_RIGHT = "center:pref, 3dlu, 40dlu:grow";

	/**
	 * This class provides an action to add a new node to the center graph of the cospan 
	 * created by the corresponding {@code CospanCreationDialog}.
	 * @author Christoph Blume, Dennis Nolte
	 *
	 */
	private class AddNodeAction extends AbstractAction {

		private static final long serialVersionUID = -5016215291184560913L;

		/**
		 * 
		 */
		public AddNodeAction() {
			super(messages.getString(GuiMessages.ADD_NODE_LABEL_KEY));
		}
		
		@Override
		public void actionPerformed(ActionEvent e) {
			graph.addNode(NODE_NAME_PREFIX + nodeCount);
			nodeCount++;
			removeNodeAction.setEnabled(true);
			
			new Thread(new Runnable() {

				@Override
				public void run() {
					graphView.print(graph);
				}
			
			}).start();
			
			VisualGraph visualGraph = (VisualGraph) graphView.getGraph();
			DefaultListModel<Node> model = (DefaultListModel<Node>) innerInterfaceNodesList.getModel();
			for(Node node: Collections.list(model.elements())) {
				addInnerInterfaceStyle(visualGraph.getVertexByName(node.getName()));
			}
			
			model = (DefaultListModel<Node>) outerInterfaceNodesList.getModel();
			for(Node node: Collections.list(model.elements())) {
				addOuterInterfaceStyle(visualGraph.getVertexByName(node.getName()));
			}
		}
		
	}
	
	/**
	 * This class provides an action to remove a node from the center graph of the copsan created 
	 * by the corresponding {@code CospanCreationDialog}.
	 */
	private class RemoveNodeAction extends AbstractAction {

		private static final long serialVersionUID = -5016215291184560913L;

		/**
		 * Creates a new {@code RemoveNodeAction}.
		 */
		public RemoveNodeAction() {
			super(messages.getString(GuiMessages.REMOVE_NODE_LABEL_KEY));
		}

		@Override
		public void actionPerformed(ActionEvent e) {
			//Get a list of selected nodes
			if(graphView.getGraph().getSelectionCount() > 0) {
				Object[] cells = graphView.getGraph().getSelectionCells();
				for(Object object: cells) {
					mxCell cell = (mxCell) object;
					if(cell.getStyle().equals(VisualGraph.HYPERGRAPH_VERTEX_STYLE)) {
						graph.removeNode(cell.getValue().toString());
					}
				}

				removeNodeAction.setEnabled(graph.getNodeCount()>0);

				new Thread(new Runnable() {

					@Override
					public void run() {
						graphView.print(graph);
					}

				}).start();
			}
			else {
				showWarningMessage(GuiMessages.NO_NODE_SELECTED_TITLE_EXCEPTION_KEY,
						GuiMessages.NO_NODE_SELECTED_TEXT_EXCEPTION_KEY);
			}

		}

	}
	
	/**
	 * This class provides an action to add a new edge to the center graph of the cospan created by 
	 * the corresponding {@code CospanCreationDialog}.
	 * @author Christoph Blume, Dennis Nolte
	 *
	 */
	private class AddEdgeAction extends AbstractAction {

		private static final long serialVersionUID = -5016215291184560913L;

		/**
		 * 
		 */
		public AddEdgeAction() {
			super(messages.getString(GuiMessages.ADD_EDGE_LABEL_KEY));
		}
		
		@Override
		public void actionPerformed(ActionEvent e) {
			//Get a list of select nodes
			List<Node> presetNodes = new LinkedList<>();
			VisualGraph visualGraph = (VisualGraph) graphView.getGraph();
			
			if(visualGraph.getSelectionCount() > 0) {
				Object[] cells = visualGraph.getSelectionCells();
				for(Object object: cells) {
					mxCell cell = (mxCell) object;
					if(visualGraph.isVertex(cell)) {
						presetNodes.add(graph.getNode(cell.getId()));
					}
				}						
			}
			
			//Show the new-edge-dialog
			EdgeData edgeData = NewEdgeDialog.showDialog(
					CospanCreationDialog.this, graph, presetNodes);

			//If the user has not cancelled the dialog, add the new edge
			if (edgeData!=null) {
				graph.addEdge(edgeData.getName(), edgeData.getLabel(), edgeData.getIncidentNodes());		
				removeEdgeAction.setEnabled(true);

				new Thread(new Runnable() {

					@Override
					public void run() {
						graphView.print(graph);
					}

				}).start();
				
				DefaultListModel<Node> model = (DefaultListModel<Node>) innerInterfaceNodesList.getModel();
				for(Node node: Collections.list(model.elements())) {
					addInnerInterfaceStyle(visualGraph.getVertexByName(node.getName()));
				}
				
				model = (DefaultListModel<Node>) outerInterfaceNodesList.getModel();
				for(Node node: Collections.list(model.elements())) {
					addOuterInterfaceStyle(visualGraph.getVertexByName(node.getName()));
				}
			}
		}
		
	}
	
	/**
	 * This class provides an action to remove an edge from the center graph of the cospan created 
	 * by the corresponding {@code CospanCreationDialog}.
	 */
	private class RemoveEdgeAction extends AbstractAction {

		private static final long serialVersionUID = -5016215291184560913L;

		/**
		 * Creates a new {@code RemoveEdgeAction}.
		 */
		public RemoveEdgeAction() {
			super(messages.getString(GuiMessages.REMOVE_EDGE_LABEL_KEY));
		}

		@Override
		public void actionPerformed(ActionEvent e) {
			//Show the new-edge-dialog
			Edge selectedEdge = RemoveEdgeDialog.showDialog(
					CospanCreationDialog.this, graph);

			//If the user has not cancelled the dialog, add the new edge
			if (selectedEdge!=null) {
				graph.removeEdge(selectedEdge);
				removeEdgeAction.setEnabled(graph.getEdgeCount()>0);

				new Thread(new Runnable() {

					@Override
					public void run() {
						graphView.print(graph);
					}

				}).start();
			}
		}

	}
	
	private static final long serialVersionUID = 746543425346637851L;

	private static final String NAME_FIELD_LABEL = "nameField";
	private static final String NODE_NAME_PREFIX = "n";
	// private static final String EDGE_NAME_PREFIX = "e";

	private static CospanCreationDialog dialog;

	private static GuiMessages messages;

	private static GraphView graphView;
	private static Graph graph;
	
	private static Cospan cospan;

	private int nodeCount;
	// private int edgeCount;

	private JList<Node> innerInterfaceNodesList;
	private JList<Node> outerInterfaceNodesList;
	
	private final RemoveNodeAction removeNodeAction;
	private final RemoveEdgeAction removeEdgeAction;

	/*------------------------------------------------------------------------------------------*/
	/* 		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Shows the dialog for the creation of new signatures.
	 * 
	 * @param owner
	 *            the parent window of the dialog
	 */
	public static void showDialog(Window owner) {
		messages = GuiMessages.getMessages();
		dialog = new CospanCreationDialog(owner);

		dialog.setLocationRelativeTo(owner);
		dialog.setVisible(true);
	}
	
	private static void addInnerInterfaceStyle(mxCell cell) {
		String style = null;
		switch (cell.getStyle()) {
		case VisualGraph.HYPERGRAPH_VERTEX_STYLE:
		case VisualGraph.HYPERGRAPH_INNER_VERTEX_STYLE:
			style = VisualGraph.HYPERGRAPH_INNER_VERTEX_STYLE;
			break;

		case VisualGraph.HYPERGRAPH_OUTER_VERTEX_STYLE:
		case VisualGraph.HYPERGRAPH_INNER_OUTER_VERTEX_STYLE:
			style = VisualGraph.HYPERGRAPH_INNER_OUTER_VERTEX_STYLE;
			break;
		}
		((VisualGraph) graphView.getGraph()).setVertexStyle(cell.getId(), style);
	}
	
	private static void addOuterInterfaceStyle(mxCell cell) {
		String style = null;
		switch (cell.getStyle()) {
		case VisualGraph.HYPERGRAPH_VERTEX_STYLE:
		case VisualGraph.HYPERGRAPH_OUTER_VERTEX_STYLE:
			style = VisualGraph.HYPERGRAPH_OUTER_VERTEX_STYLE;
			break;

		case VisualGraph.HYPERGRAPH_INNER_VERTEX_STYLE:
		case VisualGraph.HYPERGRAPH_INNER_OUTER_VERTEX_STYLE:
			style = VisualGraph.HYPERGRAPH_INNER_OUTER_VERTEX_STYLE;
			break;
		}
		((VisualGraph) graphView.getGraph()).setVertexStyle(cell.getId(), style);
	}
	
	private static void removeInnerInterfaceStyle(mxCell cell) {
		String style = null;
		switch (cell.getStyle()) {
		case VisualGraph.HYPERGRAPH_VERTEX_STYLE:
		case VisualGraph.HYPERGRAPH_INNER_VERTEX_STYLE:
			style = VisualGraph.HYPERGRAPH_VERTEX_STYLE;
			break;
			
		case VisualGraph.HYPERGRAPH_OUTER_VERTEX_STYLE:
		case VisualGraph.HYPERGRAPH_INNER_OUTER_VERTEX_STYLE:
			style = VisualGraph.HYPERGRAPH_OUTER_VERTEX_STYLE;
			break;
		}
		((VisualGraph) graphView.getGraph()).setVertexStyle(cell.getId(), style);
	}
	
	private static void removeOuterInterfaceStyle(mxCell cell) {
		String style = null;
		switch (cell.getStyle()) {
		case VisualGraph.HYPERGRAPH_VERTEX_STYLE:
		case VisualGraph.HYPERGRAPH_OUTER_VERTEX_STYLE:
			style = VisualGraph.HYPERGRAPH_VERTEX_STYLE;
			break;
			
		case VisualGraph.HYPERGRAPH_INNER_VERTEX_STYLE:
		case VisualGraph.HYPERGRAPH_INNER_OUTER_VERTEX_STYLE:
			style = VisualGraph.HYPERGRAPH_INNER_VERTEX_STYLE;
			break;
		}
		((VisualGraph) graphView.getGraph()).setVertexStyle(cell.getId(), style);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	private CospanCreationDialog(Window owner) {
		super(owner, messages.getString(GuiMessages.COSPAN_CREATION_DIALOG_TITLE_KEY),
				RavenDialog.ACCEPT_CANCEL_LAYOUT);

		AbstractAction acceptAction = new AbstractAction(
				messages.getString(GuiMessages.ACCEPT_LABEL_KEY)) {

			private static final long serialVersionUID = 2418165190397918534L;

			@Override
			public void actionPerformed(ActionEvent e) {
				String name = ((JTextField) Utilities.getComponentByName(
						NAME_FIELD_LABEL, CospanCreationDialog.this)).getText();
				if (name.isEmpty()) {
					showErrorMessage(GuiMessages.EMPTY_INPUTFIELD_TITLE_EXCEPTION_KEY,
							GuiMessages.EMPTY_INPUTFIELD_TEXT_EXCEPTION_KEY,
							messages.getString(GuiMessages.NAME_LABEL_KEY));
					return;
				}
				if(!Repository.isNameValid(name)){
					showErrorMessage(GuiMessages.INVALID_NAME_TITLE_EXCEPTION_KEY,
							GuiMessages.INVALID_NAME_TEXT_EXCEPTION_KEY, name);
					return;
				}
				if (Repository.getRepository().contains(name)) {
					showInformationMessage(GuiMessages.NAME_ALREADY_IN_USE_TITLE_EXCEPTION_KEY,
							GuiMessages.NAME_ALREADY_IN_USE_TEXT_EXCEPTION_KEY, name);
				} else {
					List<Node> innerInterface = Collections.list(
							((DefaultListModel<Node>) innerInterfaceNodesList.getModel()).elements());
					List<Node> outerInterface = Collections.list(
							((DefaultListModel<Node>) outerInterfaceNodesList.getModel()).elements());
					
					cospan = new Cospan(graph, innerInterface, outerInterface);
					new AddToRepositoryAction(name, cospan).execute();
					CospanCreationDialog.this.dispose();
				}
			}

		};

		removeNodeAction = new RemoveNodeAction();
		removeNodeAction.setEnabled(false);
		
		removeEdgeAction = new RemoveEdgeAction();
		removeEdgeAction.setEnabled(false);
		
		getContentPane().setBackground(MainWindow.BACKGROUND_COLOR);
		setMainPane(createMainPanel());
		setAction(ACCEPT_BUTTON, acceptAction);
		pack();

		nodeCount = 0;
		// edgeCount = 0;

	}

	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the panel which displays components for the creation of the
	 * graph.
	 * 
	 * @return the main panel
	 */
	private JPanel createMainPanel() {
		graph = new Graph("middleGraph");
		String labelSuffix = ": ";

		JLabel nameLabel = new JLabel(
				messages.getString(GuiMessages.NAME_LABEL_KEY) + labelSuffix);
		nameLabel.setOpaque(false);
		JTextField nameField = new JTextField(10);
		nameField.setOpaque(false);
		nameField.setName(NAME_FIELD_LABEL);

		SpringLayout layout = new SpringLayout();
		JPanel namePanel = new JPanel(layout);
		namePanel.setOpaque(false);
		namePanel.add(nameLabel);
		namePanel.add(nameField);

		Utilities.makeGridLayout(layout, namePanel, 2, 10, 10);

		final JPopupMenu popup = new JPopupMenu();
		popup.add(new AddNodeAction());
		popup.add(removeNodeAction);
		popup.add(new AddEdgeAction());
		popup.add(removeEdgeAction);

		MouseListener popupListener = new MouseAdapter() {
			
			@Override
			public void mouseReleased(MouseEvent e) {
				if (e.isPopupTrigger()) {
					popup.show(e.getComponent(), e.getX(), e.getY());
		        }
		    }
			
		};
		
		graphView = new GraphView(this);
		graphView.setOpaque(false);
		graphView.getViewport().setOpaque(false);
		graphView.setPreferredSize(new Dimension(600, 500));
		graphView.addMouseListener(popupListener);
		
		addMouseListener(popupListener);
		
		JPanel leftListPanel = createInnerInterfaceListPanel();
		leftListPanel.setPreferredSize(new Dimension(150, 500));
		leftListPanel.setOpaque(false);
		
		
		JPanel rightListPanel = createOuterInterfaceListPanel();
		rightListPanel.setPreferredSize(new Dimension(150, 500));
		rightListPanel.setOpaque(false);
		
		JPanel panel = new JPanel(new BorderLayout());
		panel.setOpaque(false);
		panel.add(namePanel, BorderLayout.NORTH);
		panel.add(graphView, BorderLayout.CENTER);
		panel.add(leftListPanel, BorderLayout.WEST);
		panel.add(rightListPanel, BorderLayout.EAST);
		
		
		JPanel buttonPanel = new JPanel();
		buttonPanel.setOpaque(false);
		buttonPanel.add(new JButton(new AddNodeAction()));
		buttonPanel.add(new JButton(removeNodeAction));
		buttonPanel.add(new JButton(new AddEdgeAction()));
		buttonPanel.add(new JButton(removeEdgeAction));
		panel.add(buttonPanel, BorderLayout.SOUTH);

		return panel;
	}

	
	/**
	 * Returns the panel which displays components for the creation of the
	 * inner interface.
	 * 
	 * @return the inner interface panel
	 */
	private JPanel createInnerInterfaceListPanel() {
		JScrollPane scrollpane;

		FormLayout layout = new FormLayout(LAYOUT_SPECIFICATION_LEFT, "");
		DefaultFormBuilder builder = new DefaultFormBuilder(layout);
		builder.border(Borders.DIALOG);
		
		builder.append(new JLabel(messages.getString(GuiMessages.INNER_INTERFACE_LABEL_KEY)), 3);
		builder.nextLine();

		innerInterfaceNodesList = new JList<>(new DefaultListModel<Node>());
		innerInterfaceNodesList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		
		scrollpane = new JScrollPane(innerInterfaceNodesList);
		//scrollpane.setPreferredSize(new Dimension(50, 450));
		builder.append(scrollpane);
		
		JButton shiftSourceLeftButton = new JButton(
				new AbstractAction(messages.getString(GuiMessages.SHIFT_LEFT_LABEL_KEY)) {
					
					private static final long serialVersionUID = -1131690046542140393L;
					
					@Override
					public void actionPerformed(ActionEvent e) {
						VisualGraph visualGraph = (VisualGraph) graphView.getGraph();
						
						if(visualGraph.getSelectionCount() > 0) {
							Object[] cells = visualGraph.getSelectionCells();
							for(Object object: cells) {
								mxCell cell = (mxCell) object;
								if(visualGraph.isVertex(cell)) {
									Node node = graph.getNode(cell.getId());
									((DefaultListModel<Node>) innerInterfaceNodesList.getModel()).addElement(node);
									addInnerInterfaceStyle(cell);
								}
							}
						}
					}
				});

		JButton shiftSourceRightButton = new JButton(
				new AbstractAction(messages.getString(GuiMessages.SHIFT_RIGHT_LABEL_KEY)) {

			private static final long serialVersionUID = 4005498419843856297L;

			@Override
			public void actionPerformed(ActionEvent e) {
				Node node = innerInterfaceNodesList.getSelectedValue();
				int selectedIndex = innerInterfaceNodesList.getSelectedIndex();
				if (node != null) {
					((DefaultListModel<Node>) innerInterfaceNodesList.getModel()).remove(selectedIndex);
					VisualGraph visualGraph = (VisualGraph) graphView.getGraph();
					removeInnerInterfaceStyle(visualGraph.getVertexByName(node.getName()));
				}
			}
		});
		
		JButton shiftSourceUpButton = new JButton(
				new AbstractAction(messages.getString(GuiMessages.SHIFT_UP_LABEL_KEY)) {

			private static final long serialVersionUID = -5503087052888200689L;

			@Override
			public void actionPerformed(ActionEvent e) {
				int selectedIndex = innerInterfaceNodesList.getSelectedIndex();
				if(selectedIndex > 0) {
					DefaultListModel<Node> model = (DefaultListModel<Node>) 
							innerInterfaceNodesList.getModel();
					Node shiftHelp = model.set(selectedIndex-1, model.get(selectedIndex));
					model.set(selectedIndex, shiftHelp);	
					innerInterfaceNodesList.setSelectedIndex(selectedIndex-1);
				}
					
			}
		});
		
		JButton shiftSourceDownButton = new JButton(
				new AbstractAction(messages.getString(GuiMessages.SHIFT_DOWN_LABEL_KEY)) {

			private static final long serialVersionUID = 8303456378437071918L;

			@Override
			public void actionPerformed(ActionEvent e) {
				int selectedIndex = innerInterfaceNodesList.getSelectedIndex();
				if(selectedIndex < innerInterfaceNodesList.getLastVisibleIndex()) {
					DefaultListModel<Node> model = (DefaultListModel<Node>) 
							innerInterfaceNodesList.getModel();
					Node shiftHelp = model.set(selectedIndex+1, model.get(selectedIndex));
					model.set(selectedIndex, shiftHelp);	
					innerInterfaceNodesList.setSelectedIndex(selectedIndex+1);
				}
					
			}
		});

		JPanel buttonPanel = new JPanel(new GridLayout(4, 1, 0, 5));
		buttonPanel.add(shiftSourceLeftButton);
		buttonPanel.add(shiftSourceRightButton);
		buttonPanel.add(shiftSourceUpButton);
		buttonPanel.add(shiftSourceDownButton);
		builder.append(buttonPanel);	
		
		return builder.getPanel();
	}
	
	/**
	 * Returns the panel which displays components for the creation of the
	 * outer interface.
	 * 
	 * @return the outer interface panel
	 */
	private JPanel createOuterInterfaceListPanel() {
		JScrollPane scrollpane;

		FormLayout layout = new FormLayout(LAYOUT_SPECIFICATION_RIGHT, "");
		DefaultFormBuilder builder = new DefaultFormBuilder(layout);
		builder.border(Borders.DIALOG);

		JLabel label = new JLabel(messages.getString(GuiMessages.OUTER_INTERFACE_LABEL_KEY));
		label.setHorizontalAlignment(SwingConstants.RIGHT);
		builder.append(label,3);
		builder.nextLine();

		outerInterfaceNodesList = new JList<>(new DefaultListModel<Node>());
		outerInterfaceNodesList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		
		JButton shiftSourceRightButton = new JButton(new AbstractAction(messages.getString(GuiMessages.SHIFT_RIGHT_LABEL_KEY)) {

			private static final long serialVersionUID = 8303456378437071918L;

			@Override
			public void actionPerformed(ActionEvent e) {
				VisualGraph visualGraph = (VisualGraph) graphView.getGraph();
				
				if(visualGraph.getSelectionCount() > 0) {
					DefaultListModel<Node> model = (DefaultListModel<Node>) outerInterfaceNodesList.getModel();
					Object[] cells = visualGraph.getSelectionCells();
					for(Object object: cells) {
						mxCell cell = (mxCell) object;
						if(visualGraph.isVertex(cell)) {
							Node node = graph.getNode(cell.getId());
							if(!model.contains(node)) {
								model.addElement(node);
							}
							addOuterInterfaceStyle(cell);
						}
					}
				}			
			}
		});

		JButton shiftSourceLeftButton = new JButton(new AbstractAction(
				messages.getString(GuiMessages.SHIFT_LEFT_LABEL_KEY)) {

			private static final long serialVersionUID = 2084231743392836166L;

			@Override
			public void actionPerformed(ActionEvent e) {
				Node node = outerInterfaceNodesList.getSelectedValue();
				if(node!=null) {
					((DefaultListModel<Node>) outerInterfaceNodesList.getModel()).removeElement(node);
					VisualGraph visualGraph = (VisualGraph) graphView.getGraph();
					removeOuterInterfaceStyle(visualGraph.getVertexByName(node.getName()));
				}
			}
			
		});
		
		JButton shiftSourceUpButton = new JButton(new AbstractAction(messages.getString(GuiMessages.SHIFT_UP_LABEL_KEY)) {

			private static final long serialVersionUID = 2239066926395988487L;

			@Override
			public void actionPerformed(ActionEvent e) {
				int selectedIndex = outerInterfaceNodesList.getSelectedIndex();
				if(selectedIndex > 0) {
					DefaultListModel<Node> model = (DefaultListModel<Node>) 
							outerInterfaceNodesList.getModel();
					Node shiftHelp = model.set(selectedIndex-1, model.get(selectedIndex));
					model.set(selectedIndex, shiftHelp);	
					outerInterfaceNodesList.setSelectedIndex(selectedIndex-1);
				}
					
			}
		});
		
		JButton shiftSourceDownButton = new JButton(new AbstractAction(messages.getString(GuiMessages.SHIFT_DOWN_LABEL_KEY)) {

			private static final long serialVersionUID = 5109082859972582809L;

			@Override
			public void actionPerformed(ActionEvent e) {
				int selectedIndex = outerInterfaceNodesList.getSelectedIndex();
				if(selectedIndex < outerInterfaceNodesList.getLastVisibleIndex()) {
					DefaultListModel<Node> model = (DefaultListModel<Node>) 
							outerInterfaceNodesList.getModel();
					Node shiftHelp = model.set(selectedIndex+1, model.get(selectedIndex));
					model.set(selectedIndex, shiftHelp);	
					outerInterfaceNodesList.setSelectedIndex(selectedIndex+1);
				}
					
			}
		});

		JPanel buttonPanel = new JPanel(new GridLayout(4, 1, 0, 5));
		buttonPanel.add(shiftSourceRightButton);
		buttonPanel.add(shiftSourceLeftButton);
		buttonPanel.add(shiftSourceUpButton);
		buttonPanel.add(shiftSourceDownButton);
		builder.append(buttonPanel);
		
		scrollpane = new JScrollPane(outerInterfaceNodesList);
		//scrollpane.setPreferredSize(new Dimension(50, 450));
		builder.append(scrollpane);
		
		return builder.getPanel();
	}
	
}
