/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.gui;

import java.awt.CardLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.AbstractAction;
import javax.swing.BorderFactory;
import javax.swing.ButtonGroup;
import javax.swing.DefaultListModel;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.ListCellRenderer;

import com.jgoodies.forms.builder.DefaultFormBuilder;
import com.jgoodies.forms.layout.FormLayout;

import de.uni_due.inf.ti.raven.data.Cospan;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.gui.graphs.CospanView;
import de.uni_due.inf.ti.raven.gui.graphs.DecompositionView;

/**
 * This class provides a builder which is used to  create an {@code InformationPanel} 
 * for displaying information about <i>cospan decompositions</i>.
 * @author Christoph Blume, Laura Steinert, Dennis Nolte
 * @version $Revision: 1.16 $
 *
 * @see CospanDecomposition
 */
public final class CospanDecompositionInformationPanelBuilder extends 
	InformationPanelBuilder<CospanDecomposition> {

	/**
	 * 
	 * @author Christoph Blume
	 *
	 */
	private static class CospanDecompositionInformationPanel extends 
		InformationPanel<CospanDecomposition> {

		private static final long serialVersionUID = 6860034365591736411L;

		private static final Repository repo = Repository.getRepository();
		
		/*------------------------------------------------------------------------------------------*/
		/* 		Public Methods																		*/
		/*------------------------------------------------------------------------------------------*/

		@Override
		public void setData(CospanDecomposition decomposition) {
			cospanNameField.setText(repo.getName(decomposition));
			decompositionLengthField.setText(String.valueOf(decomposition.getLength()));
			innerInterfaceSizeField.setText(String.valueOf(decomposition.getInnerInterface().size()));
			outerInterfaceSizeField.setText(String.valueOf(decomposition.getOuterInterface().size()));
			CospanDecompositionInformationPanelBuilder.decomposition = decomposition;
			// display the graph
			decompositionView.print(decomposition);
			cospanView.print(new Cospan(decomposition));
			
			DefaultListModel<Operation> model = (DefaultListModel<Operation>) operationList.getModel();
			model.clear();
			for(Operation operation: decomposition.getOperations()) {
				model.addElement(operation);
			}
		}
		
	}
	
	private static final String DECOMPOSITION_PANEL_LABEL = "decompositionPanel";
	private static final String COSPAN_PANEL_LABEL = "cospanPanel";

	private static CospanDecompositionInformationPanelBuilder panelBuilder;
	private static InformationPanel<CospanDecomposition> panel;
	private static JTextField cospanNameField;
	private static JTextField decompositionLengthField;
	private static JTextField innerInterfaceSizeField;
	private static JTextField outerInterfaceSizeField;
	private static JList<Operation> operationList;
	private static DecompositionView decompositionView;
	private static CospanView cospanView;
	private static CospanDecomposition decomposition;
	private static JRadioButton decompositionButton;
	private static JRadioButton cospanButton;
	
	private GuiMessages messages;	
	
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates an information panel displaying properties of a given cospan decomposition.
	 * @return the cospan decomposition information panel
	 */
	public static InformationPanel<CospanDecomposition> createPanel() {
		if(panelBuilder==null) {
			panel = new CospanDecompositionInformationPanel();
			DefaultFormBuilder builder = new DefaultFormBuilder(
					new FormLayout(LAYOUT_SPECIFICATION), panel);
			panelBuilder = new CospanDecompositionInformationPanelBuilder(builder);
		}
		
		
		return panel;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates the panel on which signature information will be displayed.
	 * @param formBuilder the form builder used to build the panel
	 */
	private CospanDecompositionInformationPanelBuilder(DefaultFormBuilder formBuilder) {
		super(formBuilder);		
		messages = GuiMessages.getMessages();

		formBuilder.getPanel().setOpaque(false);

		cospanNameField = new JTextField(10);
		cospanNameField.setEditable(false);
		cospanNameField.setOpaque(false);
		append(GuiMessages.NAME_LABEL_KEY, cospanNameField);
		
		appendSeparator();
		
		decompositionLengthField = new JTextField(10);
		decompositionLengthField.setEditable(false);
		decompositionLengthField.setOpaque(false);
		append(GuiMessages.DECOMPOSITION_LENGTH_LABEL_KEY, decompositionLengthField);

		innerInterfaceSizeField = new JTextField(10);
		innerInterfaceSizeField.setEditable(false);
		innerInterfaceSizeField.setOpaque(false);
		append(GuiMessages.INNER_INTERFACE_SIZE_LABEL_KEY, innerInterfaceSizeField);

		outerInterfaceSizeField = new JTextField(10);
		outerInterfaceSizeField.setEditable(false);
		outerInterfaceSizeField.setOpaque(false);
		append(GuiMessages.OUTER_INTERFACE_SIZE_LABEL_KEY, outerInterfaceSizeField);
		
		appendSeparator();
		
		operationList = new JList<>(new DefaultListModel<Operation>());
		operationList.setVisibleRowCount(8);
		operationList.setOpaque(false);
		operationList.setCellRenderer(new ListCellRenderer<Operation>() {

			@Override
			public Component getListCellRendererComponent(
					JList<? extends Operation> list, Operation value, int index,
					boolean isSelected, boolean cellHasFocus) {
				return new JLabel(value.toString());
			}
		});
		
		JScrollPane scrollpane = new JScrollPane(operationList);
		scrollpane.getViewport().setOpaque(false);
		scrollpane.setOpaque(false);
		
		append(GuiMessages.OPERATIONS_LABEL_KEY, scrollpane);

		appendSeparator();
		
		// preview of the resulting graph
		cospanView = new CospanView();
		cospanView.setBorder(BorderFactory.createEtchedBorder());
		cospanView.setOpaque(false);
		cospanView.getViewport().setOpaque(false);
		cospanView.setPreferredSize(new Dimension(0, 310));
		cospanView.getGraphControl().addMouseListener(new MouseAdapter() {			
			private JFrame window;
			
			@Override
			public void mouseClicked(MouseEvent e) {
				if(e.getClickCount()>=2 && window==null) {
					CospanView view = new CospanView();
					view.print(new Cospan(decomposition));
					view.setOpaque(false);
					view.getViewport().setOpaque(false);					
					window = new JFrame(Repository.getRepository().getName(decomposition));
					window.getContentPane().add(view);
					window.getContentPane().setBackground(MainWindow.BACKGROUND_COLOR);
					window.setSize(500, 500);
					window.setLocationRelativeTo(null);
					
					window.addWindowListener(new WindowAdapter() {
						
						@Override
						public void windowClosed(WindowEvent arg0) {
							window = null;							
						}
						@Override
						public void windowClosing(WindowEvent arg0) {
							window.dispose();							
						}
						
					});
					window.setVisible(true);
				}
			}
			
		});

		decompositionView = new DecompositionView();
		decompositionView.setBorder(BorderFactory.createEtchedBorder());
		decompositionView.setOpaque(false);
		decompositionView.getViewport().setOpaque(false);
		decompositionView.setPreferredSize(new Dimension(0, 310));
		decompositionView.getGraphControl().addMouseListener(new MouseAdapter() {			
			private JFrame window;
			
			@Override
			public void mouseClicked(MouseEvent e) {
				if(e.getClickCount()>=2 && window==null) {
					DecompositionView view = new DecompositionView();
					view.print(decomposition);
					view.setOpaque(false);
					view.getViewport().setOpaque(false);					
					window = new JFrame(Repository.getRepository().getName(decomposition));
					window.getContentPane().add(view);
					window.getContentPane().setBackground(MainWindow.BACKGROUND_COLOR);
					window.setSize(500, 500);
					window.setLocationRelativeTo(null);
					
					window.addWindowListener(new WindowAdapter() {
						
						@Override
						public void windowClosed(WindowEvent e) {
							window = null;							
						}
						@Override
						public void windowClosing(WindowEvent e) {
							window.dispose();							
						}

					});
					window.setVisible(true);
				}
			}

		});
		
		final JPanel viewPanel = new JPanel(new CardLayout());
		viewPanel.setOpaque(false);
		viewPanel.add(cospanView, COSPAN_PANEL_LABEL);
		viewPanel.add(decompositionView, DECOMPOSITION_PANEL_LABEL);
		
		append(viewPanel, 3);

		cospanButton = new JRadioButton(messages.getString(GuiMessages.COSPAN_LABEL_KEY));
		cospanButton.setName("");
		cospanButton.setOpaque(false);	
		cospanButton.setSelected(true);
		cospanButton.addActionListener(new AbstractAction() {
			
			private static final long serialVersionUID = 8061347003637341440L;
			
			@Override
			public void actionPerformed(ActionEvent e) {
				CardLayout layout = (CardLayout) (viewPanel.getLayout());
				layout.show(viewPanel, COSPAN_PANEL_LABEL);
			}
			
		});

		decompositionButton = new JRadioButton(messages.getString(GuiMessages.COSPAN_DECOMPOSITION_LABEL_KEY));
		decompositionButton.setName("");
		decompositionButton.setOpaque(false);
		decompositionButton.addActionListener(new AbstractAction() {

			private static final long serialVersionUID = 664866924215783647L;

			@Override
			public void actionPerformed(ActionEvent e) {
				CardLayout layout = (CardLayout) (viewPanel.getLayout());
				layout.show(viewPanel, DECOMPOSITION_PANEL_LABEL);
			}

		});
		
		ButtonGroup formatGroup = new ButtonGroup();
		formatGroup.add(cospanButton);
		formatGroup.add(decompositionButton);

		JPanel buttonPanel = new JPanel(new FlowLayout(FlowLayout.LEFT, 0, 0));
		buttonPanel.setOpaque(false);
		buttonPanel.add(cospanButton);
		buttonPanel.add(decompositionButton);
		append(buttonPanel, 3);
	}

}
