/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.gui;

import java.awt.BorderLayout;
import java.awt.CardLayout;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.IOException;

import javax.imageio.ImageIO;
import javax.swing.AbstractAction;
import javax.swing.BorderFactory;
import javax.swing.ButtonGroup;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JSplitPane;

import de.uni_due.inf.ti.raven.algorithms.LanguageInclusionAlgorithm;
import de.uni_due.inf.ti.raven.algorithms.UniversalityAlgorithm;
import de.uni_due.inf.ti.raven.data.Cospan;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.gui.graphs.CospanView;
import de.uni_due.inf.ti.raven.gui.graphs.DecompositionView;
import de.uni_due.inf.ti.raven.gui.graphs.GraphView;
import de.uni_due.inf.ti.raven.main.RavenMain;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class provides a window which is used to display the counterexamples 
 * computed by the different universality and language inclusion algorithms. In 
 * the first case a single cospan decomposition has to be provided, in the second
 * case two cospan decompositions have to be provided.
 * @author Christoph Blume, Dennis Nolte
 * @version $Revision: 1.8 $
 *
 * @see CospanDecomposition
 * @see UniversalityAlgorithm
 * @see LanguageInclusionAlgorithm
 */
public class CounterExampleWindow extends JFrame {

	private static final long serialVersionUID = -977721595405898943L;

	private static final String GRAPH_COUNTEREXAMPLE_PANEL_LABEL = "graphCounterExamplePanel";
	private static final String DECOMPOSITION_COUNTEREXAMPLE_PANEL_LABEL = "decompositionCounterExamplePanel";
	private static final String COSPAN_COUNTEREXAMPLE_PANEL_LABEL = "cospanCounterExamplePanel";

	private static CounterExampleWindow window;
	private GraphView graphView;
	private DecompositionView decompositionView;
	private CospanView cospanView;
	
	private GraphView graphViewRight;
	private DecompositionView decompositionViewRight;
	private CospanView cospanViewRight;
	
	private JRadioButton graphButton;
	private JRadioButton decompositionButton;
	private JRadioButton cospanButton;

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the counterexample window.
	 * @return the counterexample window
	 */
	public static CounterExampleWindow getWindow() {
		if(window==null) {
			window = new CounterExampleWindow();
		}
		
		return window;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new counterexample window.
	 */
	private CounterExampleWindow() {
		try {
			setIconImage(ImageIO.read(new File(RavenMain.RAVEN_ICON_FILENAME)));
		} catch (IOException e) {
			//Do nothing
		}
		
		addWindowListener(new WindowAdapter() {

			@Override
			public void windowClosing(WindowEvent e) {
				e.getWindow().dispose();
			}

		});
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Shows the chosen element {@code decomposition} in this window.
	 * @param element the decomposition, graph or cospan to be shown
	 * @param title the title of the window
	 */
	public void showCounterExampleChooser(CospanDecomposition element, String title) {
		if(element==null) {
			throw new NullPointerException(
					RavenExceptionMessages.DECOMPOSITION_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		initializeLeftViews(element);
		
		initializeButtons();
		
		final JPanel centerPanel = new JPanel(new CardLayout());
		centerPanel.setOpaque(false);
		centerPanel.add(graphView, GRAPH_COUNTEREXAMPLE_PANEL_LABEL);
		centerPanel.add(decompositionView, DECOMPOSITION_COUNTEREXAMPLE_PANEL_LABEL);
		centerPanel.add(cospanView, COSPAN_COUNTEREXAMPLE_PANEL_LABEL);
		
		graphButton.addActionListener(
				new AbstractAction() {

					private static final long serialVersionUID = 6719474644481154530L;

					@Override
					public void actionPerformed(ActionEvent e) {
						CardLayout layout = (CardLayout) (centerPanel.getLayout());
						layout.show(centerPanel, GRAPH_COUNTEREXAMPLE_PANEL_LABEL);							
					}								
				}
		);
		
		decompositionButton.addActionListener(
				new AbstractAction() {

					private static final long serialVersionUID = 664866924215783647L;

					@Override
					public void actionPerformed(ActionEvent e) {
						CardLayout layout = (CardLayout) (centerPanel.getLayout());
						layout.show(centerPanel, DECOMPOSITION_COUNTEREXAMPLE_PANEL_LABEL);							
					}								
				}
		);
		
		cospanButton.addActionListener(
				new AbstractAction() {

					private static final long serialVersionUID = 8061347003637341440L;

					@Override
					public void actionPerformed(ActionEvent e) {
						CardLayout layout = (CardLayout) (centerPanel.getLayout());
						layout.show(centerPanel, COSPAN_COUNTEREXAMPLE_PANEL_LABEL);							
					}								
				}
		);
		
		ButtonGroup formatCounterExample = new ButtonGroup();
		formatCounterExample.add(graphButton);
		formatCounterExample.add(decompositionButton);
		formatCounterExample.add(cospanButton);
		
		JPanel formatCounterPanel = new JPanel(new FlowLayout(FlowLayout.LEFT, 0, 0));
		formatCounterPanel.setOpaque(false);
		formatCounterPanel.add(graphButton);
		formatCounterPanel.add(decompositionButton);
		formatCounterPanel.add(cospanButton);
	
		JPanel panel = new JPanel(new BorderLayout());
		panel.setOpaque(false);
		panel.add(centerPanel, BorderLayout.CENTER);
		panel.add(formatCounterPanel, BorderLayout.SOUTH);
		
		setContentPane(panel);
		setExtendedState(JFrame.MAXIMIZED_BOTH);
		setTitle(title);
		setBackground(MainWindow.BACKGROUND_COLOR);
		setVisible(true);
	}
	
	
	/**
	 * Shows the chosen element {@code decomposition} in this window.
	 * @param left the cospan decomposition shown on the left
	 * @param right the cospan decomposition shown on the right
	 * @param title the title of the window
	 */
	public void showRuleCounterExampleChooser(CospanDecomposition left, CospanDecomposition right, 
			String title) {
		if(left==null) {
			throw new NullPointerException(
					RavenExceptionMessages.DECOMPOSITION_MUST_NOT_BE_NULL_EXCEPTION);
		}
		else if(right==null) {
			throw new NullPointerException(
					RavenExceptionMessages.DECOMPOSITION_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		initializeLeftViews(left);
		initializeRightViews(right);
		
		initializeButtons();
		
		final JPanel centerPanelLeft = new JPanel(new CardLayout());
		centerPanelLeft.setOpaque(false);
		centerPanelLeft.add(graphView, GRAPH_COUNTEREXAMPLE_PANEL_LABEL);
		centerPanelLeft.add(decompositionView, DECOMPOSITION_COUNTEREXAMPLE_PANEL_LABEL);
		centerPanelLeft.add(cospanView, COSPAN_COUNTEREXAMPLE_PANEL_LABEL);
		
		final JPanel centerPanelRight = new JPanel(new CardLayout());
		centerPanelRight.setOpaque(false);
		centerPanelRight.add(graphViewRight, GRAPH_COUNTEREXAMPLE_PANEL_LABEL);
		centerPanelRight.add(decompositionViewRight, DECOMPOSITION_COUNTEREXAMPLE_PANEL_LABEL);
		centerPanelRight.add(cospanViewRight, COSPAN_COUNTEREXAMPLE_PANEL_LABEL);
		
		graphButton.addActionListener(
				new AbstractAction() {

					private static final long serialVersionUID = 6719474644481154530L;

					@Override
					public void actionPerformed(ActionEvent e) {
						CardLayout layout = (CardLayout) (centerPanelLeft.getLayout());
						layout.show(centerPanelLeft, GRAPH_COUNTEREXAMPLE_PANEL_LABEL);	
						
						layout = (CardLayout) (centerPanelRight.getLayout());
						layout.show(centerPanelRight, GRAPH_COUNTEREXAMPLE_PANEL_LABEL);
					}								
				}
		);
		
		decompositionButton.addActionListener(
				new AbstractAction() {

					private static final long serialVersionUID = 664866924215783647L;

					@Override
					public void actionPerformed(ActionEvent e) {
						CardLayout layout = (CardLayout) (centerPanelLeft.getLayout());
						layout.show(centerPanelLeft, DECOMPOSITION_COUNTEREXAMPLE_PANEL_LABEL);		
						
						layout = (CardLayout) (centerPanelRight.getLayout());
						layout.show(centerPanelRight, DECOMPOSITION_COUNTEREXAMPLE_PANEL_LABEL);
					}								
				}
		);
		
		cospanButton.addActionListener(
				new AbstractAction() {

					private static final long serialVersionUID = 8061347003637341440L;

					@Override
					public void actionPerformed(ActionEvent e) {
						CardLayout layout = (CardLayout) (centerPanelLeft.getLayout());
						layout.show(centerPanelLeft, COSPAN_COUNTEREXAMPLE_PANEL_LABEL);	
						
						layout = (CardLayout) (centerPanelRight.getLayout());
						layout.show(centerPanelRight, COSPAN_COUNTEREXAMPLE_PANEL_LABEL);
					}								
				}
		);
		
		ButtonGroup formatCounterExample = new ButtonGroup();
		formatCounterExample.add(graphButton);
		formatCounterExample.add(decompositionButton);
		formatCounterExample.add(cospanButton);
		
		JPanel formatCounterPanel = new JPanel(new FlowLayout(FlowLayout.LEFT, 0, 0));
		formatCounterPanel.setOpaque(false);
		formatCounterPanel.add(graphButton);
		formatCounterPanel.add(decompositionButton);
		formatCounterPanel.add(cospanButton);
	
		JSplitPane splitpane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);
		splitpane.setBackground(MainWindow.BACKGROUND_COLOR);
		splitpane.setLeftComponent(centerPanelLeft);
		splitpane.setRightComponent(centerPanelRight);
		splitpane.setBorder(BorderFactory.createRaisedBevelBorder());
		splitpane.setResizeWeight(0.5);

		
		JPanel panel = new JPanel(new BorderLayout());
		panel.setOpaque(false);
		panel.add(splitpane, BorderLayout.CENTER);
		panel.add(formatCounterPanel, BorderLayout.SOUTH);
		
		setContentPane(panel);
		setExtendedState(JFrame.MAXIMIZED_BOTH);
		setTitle(title);
		setBackground(MainWindow.BACKGROUND_COLOR);
		setVisible(true);
	}
	
	
	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Initialize the main views
	 * @param element the cospan decomposition shown
	 */
	private void initializeLeftViews(CospanDecomposition element) {
		graphView = new GraphView(this);
		graphView.print(element.getGraph());
		graphView.setOpaque(false);
		graphView.getViewport().setOpaque(false);
		
		decompositionView = new DecompositionView();		
		decompositionView.print(element);
		decompositionView.setOpaque(false);
		decompositionView.getViewport().setOpaque(false);
		
		Cospan cospan = new Cospan(element);
		cospanView = new CospanView();
		cospanView.print(cospan);
		cospanView.setOpaque(false);
		cospanView.getViewport().setOpaque(false);	
	}
	
	/**
	 * Initialize the right hand side views used by a rule counter example
	 * @param element the cospan decomposition for the right hand side of the window
	 */
	private void initializeRightViews(CospanDecomposition element) {
		graphViewRight = new GraphView(this);
		graphViewRight.print(element.getGraph());
		graphViewRight.setOpaque(false);
		graphViewRight.getViewport().setOpaque(false);
		
		decompositionViewRight = new DecompositionView();		
		decompositionViewRight.print(element);
		decompositionViewRight.setOpaque(false);
		decompositionViewRight.getViewport().setOpaque(false);
		
		Cospan cospan = new Cospan(element);
		cospanViewRight = new CospanView();
		cospanViewRight.print(cospan);
		cospanViewRight.setOpaque(false);
		cospanViewRight.getViewport().setOpaque(false);	
	}
	
	/**
	 * Initialize the radio buttons
	 */
	private void initializeButtons() {
		graphButton = new JRadioButton(GuiMessages.getMessages().getString(GuiMessages.GRAPH_LABEL_KEY));
		graphButton.setName("");
		graphButton.setOpaque(false);
		graphButton.setSelected(true);
		
		decompositionButton = new JRadioButton(GuiMessages.getMessages().getString(GuiMessages.COSPAN_DECOMPOSITION_LABEL_KEY));
		decompositionButton.setName("");
		decompositionButton.setOpaque(false);
		
		cospanButton = new JRadioButton(GuiMessages.getMessages().getString(GuiMessages.COSPAN_LABEL_KEY));
		cospanButton.setName("");
		cospanButton.setOpaque(false);	
	}
	
}
