/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.gui;

import java.awt.Component;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import javax.swing.DefaultComboBoxModel;
import javax.swing.DefaultListCellRenderer;
import javax.swing.JComboBox;
import javax.swing.JList;

import de.uni_due.inf.ti.raven.data.Data;
import de.uni_due.inf.ti.raven.data.NameDataPair;
import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.data.RepositoryListener;

/**
 * This class represents a drop-down menu which can be used to select a {@code Data} object
 * from a list of various items. 
 * @author Christoph Blume
 * @version $Revision: 1.14 $
 * 
 * @see Data
 * @see Repository
 */
public class DataComboBox extends JComboBox<NameDataPair> implements RepositoryListener {
	
	private class DataListCellRenderer extends DefaultListCellRenderer {

		private static final long serialVersionUID = 1517061786907498408L;
		
		@Override
		public Component getListCellRendererComponent(JList<?> list, Object value, int index,
				boolean isSelected, boolean cellHasFocus) {
			String text = null; 
			if(value!=null) {
				NameDataPair pair = (NameDataPair) value;
				text = pair.getName() + " (" + messages.getString(
						GuiMessages.DATA_TYPE_PREFIX_KEY+pair.getData().getType())+")";
			}
			
			return super.getListCellRendererComponent(list, text, index, isSelected, cellHasFocus);
		}
		
	}
	
	private static final long serialVersionUID = 5413779080792042992L;
	private final Repository repo;
	private final List<Class<? extends Data>> dataClassList;
	private final GuiMessages messages; 

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new combo box presenting the (names of) data objects stored in
	 * the repository {@code repo} of the class {@code dataClass}.
	 * @param repo the repository to be used
	 */
	public DataComboBox(Repository repo) {
		super();

		this.repo = repo;
		repo.addRepositoryListener(this);
		dataClassList = new LinkedList<>();
		messages = GuiMessages.getMessages();

		// Create default model
		DefaultComboBoxModel<NameDataPair> model = new DefaultComboBoxModel<>();
		setModel(model);
		
		setRenderer(new DataListCellRenderer());
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods 																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * 
	 * @param dataClass the class of the data objects to be presented
	 */
	public void addDataClass(Class<? extends Data> dataClass) {
		dataClassList.add(dataClass);

		// Add elements
		DefaultComboBoxModel<NameDataPair> model = (DefaultComboBoxModel<NameDataPair>) dataModel;
		List<String> list = new LinkedList<>(repo.getNames());
		Collections.sort(list);
		
		for (String name : list) {
			Data data = repo.get(name);
			if (dataClass.isInstance(data)) {
				model.addElement(new NameDataPair(name, data));
			}
		}
	}

	/**
	 * Returns the name of the selected {@code Data}-object.
	 * @return the name of the selected {@code Data}-object
	 */
	public String getSelectedName() {
		NameDataPair pair = (NameDataPair) getSelectedItem();
		
		String name = null;
		if(pair!=null) {
			name = pair.getName();
		}
		
		return name;
	}
	
	/**
	 * Returns the selected {@code Data}-object.
	 * @return the selected {@code Data}-object
	 */
	public Data getSelectedData() {
		NameDataPair pair = (NameDataPair) getSelectedItem();
		
		Data data = null;
		if(pair!=null) {
			data = pair.getData();
		}
		
		return data;
	}
	
	@Override
	public void dataAdded(String name, Data data) {
		for (Class<? extends Data> dataClass : dataClassList) {
			if (dataClass.isInstance(data)) {
				DefaultComboBoxModel<NameDataPair> model = (DefaultComboBoxModel<NameDataPair>) dataModel;
				NameDataPair pair = new NameDataPair(name, data);
				int i = 0;
				
				while (i<model.getSize() && model.getElementAt(i).compareTo(pair)<=0) {
					i++;
				}
				model.insertElementAt(pair, i);
				break;
			}
		}
	}
	
	@Override
	public void dataRenamed(String name, String replacement) {
		DefaultComboBoxModel<NameDataPair> model = (DefaultComboBoxModel<NameDataPair>) dataModel;
		
		for(int i=0;i<dataModel.getSize();i++) {
			NameDataPair pair = dataModel.getElementAt(i);
			if(pair.getName().equals(name)) {
				model.removeElementAt(i);
				dataAdded(replacement, pair.getData());				
			}
		}
	}

	@Override
	public void dataRemoved(String name, Data data) {
		for (Class<? extends Data> dataClass : dataClassList) {
			if (dataClass.isInstance(data)) {
				((DefaultComboBoxModel<NameDataPair>) dataModel).removeElement(
						new NameDataPair(name, data));
				break;
			}
		}
	}
	
}
