/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui;

import java.awt.CardLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JComboBox;
import javax.swing.JPanel;
import javax.swing.JTextField;

import com.jgoodies.forms.builder.DefaultFormBuilder;

import de.uni_due.inf.ti.raven.actions.ComputationAction;
import de.uni_due.inf.ti.raven.actions.ComputationEvent;
import de.uni_due.inf.ti.raven.actions.ComputationEvent.EventType;
import de.uni_due.inf.ti.raven.actions.ComputationListener;
import de.uni_due.inf.ti.raven.actions.CospanDecomposeAction;
import de.uni_due.inf.ti.raven.actions.LibTWAllStartLexBFSAction;
import de.uni_due.inf.ti.raven.actions.LibTWGreedyDegreeAction;
import de.uni_due.inf.ti.raven.actions.LibTWGreedyFillInAction;
import de.uni_due.inf.ti.raven.actions.LibTWLexBFSAction;
import de.uni_due.inf.ti.raven.algorithms.CospanDecompositionAlgorithm;
import de.uni_due.inf.ti.raven.algorithms.GraphDecompositionAlgorithm;
import de.uni_due.inf.ti.raven.algorithms.GraphDecompositionAlgorithm.Implementations;
import de.uni_due.inf.ti.raven.data.Cospan;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.PathDecomposition;
import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.data.TreeDecomposition;

/**
 * This class implements a {@code ControlPanel} for the invocation of the different graph 
 * decomposition algorithms implemented in Raven.
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.11 $
 * 
 * @see GraphDecompositionAlgorithm
 */
public final class DecompositionControlPanel extends ControlPanel {

	private static final long serialVersionUID = -5656676400273284253L;

	/**
	 * This class is used as a worker thread to perform the computation of graph decomposition. 
	 */
	private class DecompositionWorkerThread extends WorkerThread {

		private String name;
		private ComputationAction<TreeDecomposition> graphAction;
		private ComputationAction<CospanDecomposition> cospanAction;

		@Override
		protected Void doInBackground() throws Exception {
			// Get the parameter for the invariant checking algorithm
			Implementations algorithm = (Implementations) graphDecompositionAlgorithmBox.getSelectedItem();
			
			// Store the names of the parameter (which will be used later on for the output)
			name = nameField.getText();

			//Check whether the name is valid
			if (name.isEmpty()) {
				Utilities.showWarningMessage(DecompositionControlPanel.this,
						GuiMessages.EMPTY_INPUTFIELD_TITLE_EXCEPTION_KEY,
						GuiMessages.EMPTY_INPUTFIELD_TEXT_EXCEPTION_KEY,
						messages.getString(GuiMessages.NAME_LABEL_KEY));
				cancel(true);
				fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_CANCELLED, "", 0, 0));
				return null;
			}
			if(!Repository.isNameValid(name)) {
				Utilities.showWarningMessage(DecompositionControlPanel.this,
						GuiMessages.INVALID_NAME_TITLE_EXCEPTION_KEY,
						GuiMessages.INVALID_NAME_TEXT_EXCEPTION_KEY,
						name);
				cancel(true);
				fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_CANCELLED, "", 0, 0));
				return null;
			}
			if(Repository.getRepository().contains(name)) {
				Utilities.showWarningMessage(DecompositionControlPanel.this,
						GuiMessages.NAME_ALREADY_IN_USE_TITLE_EXCEPTION_KEY,
						GuiMessages.NAME_ALREADY_IN_USE_TEXT_EXCEPTION_KEY,
						name);
				cancel(true);
				fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_CANCELLED, "", 0, 0));
				return null;
			}
			
			if(graphCospanBox.getSelectedData() instanceof Graph) {
				
				Graph graph = (Graph) (graphCospanBox.getSelectedData());
				// Check whether a graph has been selected
				if(graph==null) {
					Utilities.showWarningMessage(DecompositionControlPanel.this,
							GuiMessages.NO_GRAPH_SELECTED_TITLE_EXCEPTION_KEY,
							GuiMessages.NO_GRAPH_SELECTED_TEXT_EXCEPTION_KEY,
							messages.getString(GuiMessages.GRAPH_LABEL_KEY));
					cancel(true);
					fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_CANCELLED, "", 0, 0));
					return null;
				}
				
				if(graph.hasLoops())
				{
					Utilities.showWarningMessage(DecompositionControlPanel.this,
							GuiMessages.NO_LOOP_GRAPH_TITLE_EXCEPTION_KEY,
							GuiMessages.NO_LOOP_GRAPH_TEXT_EXCEPTION_KEY,
							messages.getString(GuiMessages.GRAPH_LABEL_KEY));
					cancel(true);
					fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_CANCELLED, "", 0, 0));
					return null;
				}

				// Create decomposition action
				graphAction = null;
				cospanAction = null;
				if (algorithm.equals(Implementations.ALL_START_LEX_BFS)) {
					graphAction = new LibTWAllStartLexBFSAction(graph);
				} else if (algorithm.equals(Implementations.GREEDY_DEGREE)) {
					graphAction = new LibTWGreedyDegreeAction(graph);
				} else if (algorithm.equals(Implementations.GREEDY_FILL_IN)) {
					graphAction = new LibTWGreedyFillInAction(graph);
				} else if (algorithm.equals(Implementations.LEX_BFS)) {
					graphAction = new LibTWLexBFSAction(graph);
				}
				
				// Register listeners
				for(ComputationListener listener : listeners) {
						graphAction.addComputationListener(listener);
				}


				// Start computation
				graphAction.execute();
			} 
			else {
				Cospan cospan = (Cospan) (graphCospanBox.getSelectedData());
				
				if(cospan==null) {
					Utilities.showWarningMessage(DecompositionControlPanel.this,
							GuiMessages.NO_COSPAN_SELECTED_TITLE_EXCEPTION_KEY,
							GuiMessages.NO_COSPAN_SELECTED_TEXT_EXCEPTION_KEY,
							messages.getString(GuiMessages.COSPAN_LABEL_KEY));
					cancel(true);
					fireComputationEvent(new ComputationEvent(
							this, EventType.COMPUTATION_CANCELLED, "", 0, 0));
					return null;
				}
				
				graphAction = null;
				cospanAction = null;
				
				cospanAction = new CospanDecomposeAction(cospan);
				
				// Register listeners
				for(ComputationListener listener : listeners) {
						cospanAction.addComputationListener(listener);
				}			
				
				// Start computation
				cospanAction.execute();
			}

			return null;
		}

		@Override
		protected void done() {
			//If the worker thread has been cancelled
			if(isCancelled()) {
				//also cancel the computation of the action
				if(graphAction!=null) {
					graphAction.cancelComputation();
				}
				
				if(cospanAction!=null) {
					cospanAction.cancelComputation();
				}

				name = null;
				graphAction = null;
				cospanAction = null;
			}
			//else bring up message dialog to display result
			else {
				if(graphCospanBox.getSelectedData() instanceof Graph) {
					TreeDecomposition treeDecomposition = graphAction.getResult(); 
					Repository.getRepository().put(name, CospanDecomposition.getCospanDecomposition(
							PathDecomposition.getPathDecomposition(treeDecomposition)));
					
				}
				else {
					CospanDecomposition cospanDecomposition = cospanAction.getResult();
					Repository.getRepository().put(name, cospanDecomposition);
				}
				
				Utilities.showInformationMessage(DecompositionControlPanel.this, 
						GuiMessages.DECOMPOSTION_ADDED_DIALOG_TITLE_KEY,
						GuiMessages.DECOMPOSTION_ADDED_DIALOG_TEXT_KEY,
						name);
				
				nameField.setText("");
			}
		}

	}

	/** Name Text Field Label (used for Testing) */
	protected static final String NAME_LABEL = "name";
	
	/** Graph Decomposition Algorithm Box Label (used for Testing) */
	protected static final String GRAPH_DECOMPOSITION_ALGORITHM_LABEL = "graphDecompositionAlgorithm";
	
	/** Cospan Decomposition Algorithm Box Label (used for Testing) */
	protected static final String COSPAN_DECOMPOSITION_ALGORITHM_LABEL = "graphDecompositionAlgorithm";
	
	/** Graph Box Label (used for Testing) */
	protected static final String GRAPH_LABEL = "graph";
	
	/** Decomposition Type Box Label (used for Testing) */
	protected static final String TYPE_LABEL = "decompositionType";
	
	private static final String GRAPH_DECOMPOSITION_ALGORITHM_CARD = "graphDecompositionAlgorithmBox";
	private static final String COSPAN_DECOMPOSITION_ALGORITHM_CARD = "cospanDecompositionAlgorithmBox";
	
	private static JTextField nameField;
	private static DataComboBox graphCospanBox;
	private static JComboBox<GraphDecompositionAlgorithm.Implementations> graphDecompositionAlgorithmBox;
	private static JComboBox<CospanDecompositionAlgorithm.Implementations> cospanDecompositionAlgorithmBox;
	private static JPanel decompositionAlgorithmPanel;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public void setEnabled(boolean isEnabled) {
		nameField.setEnabled(isEnabled);
		graphDecompositionAlgorithmBox.setEnabled(isEnabled);
		graphCospanBox.setEnabled(isEnabled);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	protected void initialize(DefaultFormBuilder builder) {
		Repository repo = Repository.getRepository();
		
		nameField = new JTextField(10);
		nameField.setName(NAME_LABEL);
		append(builder, GuiMessages.NAME_LABEL_KEY, nameField);
		
		graphCospanBox = new DataComboBox(repo);
		graphCospanBox.setName(GRAPH_LABEL);
		graphCospanBox.addDataClass(Graph.class);
		graphCospanBox.addDataClass(Cospan.class);
		append(builder, GuiMessages.GRAPH_COSPAN_LABEL_KEY, graphCospanBox);
		
		graphCospanBox.addActionListener(new ActionListener() {
			
			@Override
			public void actionPerformed(ActionEvent e) {
				CardLayout layout = (CardLayout) decompositionAlgorithmPanel.getLayout();
				String name = null;
				
				if(graphCospanBox.getSelectedData() instanceof Graph) {
					name = GRAPH_DECOMPOSITION_ALGORITHM_CARD;
				}
				else {
					name = COSPAN_DECOMPOSITION_ALGORITHM_CARD;
				}
				layout.show(decompositionAlgorithmPanel, name);
			}
		});

		graphDecompositionAlgorithmBox = new JComboBox<>(
				GraphDecompositionAlgorithm.Implementations.values());
		graphDecompositionAlgorithmBox.setName(GRAPH_DECOMPOSITION_ALGORITHM_LABEL);
		cospanDecompositionAlgorithmBox = new JComboBox<>(
				CospanDecompositionAlgorithm.Implementations.values());
		cospanDecompositionAlgorithmBox.setName(COSPAN_DECOMPOSITION_ALGORITHM_LABEL);
		
		decompositionAlgorithmPanel = new JPanel(new CardLayout());
		decompositionAlgorithmPanel.add(graphDecompositionAlgorithmBox, 
				GRAPH_DECOMPOSITION_ALGORITHM_CARD);
		decompositionAlgorithmPanel.add(cospanDecompositionAlgorithmBox, 
				COSPAN_DECOMPOSITION_ALGORITHM_CARD);
		
		append(builder, GuiMessages.ALGORITHM_LABEL_KEY, decompositionAlgorithmPanel);
		
		// set Color
		nameField.setBackground(getBackground());
		graphDecompositionAlgorithmBox.setBackground(getBackground());
		graphCospanBox.setBackground(getBackground());
	}

	@Override
	protected WorkerThread getWorkerThread() {
		return new DecompositionWorkerThread();
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	
	
}
