/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui;

import java.awt.BorderLayout;
import java.awt.CardLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.text.DecimalFormat;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.ButtonGroup;
import javax.swing.DefaultListCellRenderer;
import javax.swing.DefaultListModel;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JFormattedTextField;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;
import javax.swing.SpringLayout;

import com.jgoodies.forms.builder.DefaultFormBuilder;
import com.jgoodies.forms.factories.Borders;
import com.jgoodies.forms.layout.FormLayout;

import de.uni_due.inf.ti.raven.cli.Console;
import de.uni_due.inf.ti.raven.data.Cospan;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.NameDataPair;
import de.uni_due.inf.ti.raven.data.Node;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.data.Signature;
import de.uni_due.inf.ti.raven.gui.graphs.CospanView;
import de.uni_due.inf.ti.raven.gui.graphs.DecompositionView;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * This class provides a dialog to create a new decomposition. Please note that the dialog is 
 * modal, i.e. the {@code showDialog} method blocks the caller until the user's interaction 
 * is complete. 
 * @author Christoph Blume, Laura Steinert, Dennis Nolte
 * @version $Revision: 1.36 $
 * 
 */
public class DecompositionCreationDialog extends RavenDialog {

	/**
	 * This class provides a dialog to create a new label which can be used for 
	 * a new CONNECT-operation.
	 * @author Christoph Blume, Dennis Nolte
	 * @version $Revision: 1.36 $
	 *
	 * @see Operation
	 */
	private static class NewLabelDialog extends RavenDialog {

		private static final long serialVersionUID = -1735339928124098876L;

		private static NewLabelDialog dialog;

		private final JTextField LABEL_FIELD;
		private final JTextField ARITY_FIELD;

		/*------------------------------------------------------------------------------------------*/
		/*		Static Methods																		*/
		/*------------------------------------------------------------------------------------------*/

		/**
		 * Shows the dialog for the creation of new labels.
		 * @param owner the parent window of the dialog
		 */
		public static void showDialog(Window owner) {
			dialog = new NewLabelDialog(owner);			
			dialog.setLocationRelativeTo(owner);
			dialog.setVisible(true);
		}

		/*------------------------------------------------------------------------------------------*/
		/*		Constructors																		*/
		/*------------------------------------------------------------------------------------------*/

		/**
		 * Creates a new dialog for the creation of new labels.
		 * @param owner the parent window of this dialog
		 */
		private NewLabelDialog(Window owner) {
			super(owner, messages.getString(GuiMessages.NEW_LABEL_DIALOG_TITLE_KEY), RavenDialog.ACCEPT_CANCEL_LAYOUT);

			String labelSuffix = ": ";

			JLabel labelLabel = new JLabel(
					messages.getString(GuiMessages.LABEL_LABEL_KEY)+labelSuffix);
			LABEL_FIELD = new JTextField(10);

			JLabel arityLabel = new JLabel(
					messages.getString(GuiMessages.ARITY_LABEL_KEY)+labelSuffix);
			ARITY_FIELD = new JFormattedTextField(new DecimalFormat("0"));
			ARITY_FIELD.setColumns(10);

			SpringLayout layout = new SpringLayout();
			JPanel panel = new JPanel();
			panel.setBackground(MainWindow.BACKGROUND_COLOR);
			panel.add(labelLabel);
			panel.add(LABEL_FIELD);
			panel.add(arityLabel);
			panel.add(ARITY_FIELD);

			Utilities.makeGridLayout(layout, panel, 4, 10, 10);

			AbstractAction acceptAction = new AbstractAction(
					messages.getString(GuiMessages.ACCEPT_LABEL_KEY)) {

				private static final long serialVersionUID = 321250982627352495L;

				@Override
				public void actionPerformed(ActionEvent e) {
					String label = LABEL_FIELD.getText();
					String arityString = ARITY_FIELD.getText();

					if (!Operation.isLabelValid(label)) {
						showErrorMessage(GuiMessages.INVALID_LABEL_TITLE_EXCEPTION_KEY,
								GuiMessages.INVALID_LABEL_TEXT_EXCEPTION_KEY, label);
						return;
					}
					if(label.isEmpty()) {
						showErrorMessage(GuiMessages.EMPTY_INPUTFIELD_TITLE_EXCEPTION_KEY,
								GuiMessages.EMPTY_INPUTFIELD_TEXT_EXCEPTION_KEY,
								messages.getString(GuiMessages.LABEL_LABEL_KEY));
						return;
					}
					if(arityString.isEmpty()) {
						showErrorMessage(GuiMessages.EMPTY_INPUTFIELD_TITLE_EXCEPTION_KEY,
								GuiMessages.EMPTY_INPUTFIELD_TEXT_EXCEPTION_KEY,
								messages.getString(GuiMessages.ARITY_LABEL_KEY));
						return;
					}

					int arity = Integer.parseInt(arityString);
					if(arity<0) {
						showErrorMessage(GuiMessages.VALUE_MUST_BE_NON_NEGATIVE_TITLE_EXCEPTION_KEY,
								GuiMessages.VALUE_MUST_BE_NON_NEGATIVE_TEXT_EXCEPTION_KEY,
								messages.getString(GuiMessages.ARITY_LABEL_KEY));
						return;
					}

					operationListModel.addElement(Operation.getConnect(label, arity));
					List<Operation> sortedList = sortListModel(operationListModel);

					operationListModel.removeAllElements();
					for(Operation letter: sortedList) {
						operationListModel.addElement(letter);
					}

					NewLabelDialog.this.dispose();
				}
			};

			setMainPane(panel);
			setAction(ACCEPT_BUTTON, acceptAction);
			pack();
		}

	}

	/**
	 * 
	 * @author Christoph Blume, Dennis Nolte
	 *
	 */
	public static class SignatureDialog extends RavenDialog {

		private static final long serialVersionUID = -2434388572612819070L;

		private static SignatureDialog dialog;
		private DataComboBox signatureBox;
		private JList<Operation> selectionList;
		private DefaultListModel<Operation> selectedListModel;
		private JLabel sigLabel;
		private JLabel innerInterfaceLabel;
		private JFormattedTextField innerInterfaceSizeField;

		/*------------------------------------------------------------------------------------------*/
		/* 		Static Methods																		*/
		/*------------------------------------------------------------------------------------------*/

		/**
		 * 
		 * @param owner the parent window of the dialog
		 */
		public static void showDialog(Window owner) {
			dialog = new SignatureDialog(owner);
			dialog.setLocationRelativeTo(owner);
			dialog.setVisible(true);
		}		

		/**
		 * Returns the integer which has been typed in the specified input field.
		 * @param textfield the input field whose value will be read
		 * @return the integer which has been typed in or {@code null} if no integer has been typed in
		 */
		protected static Integer getInteger(JFormattedTextField textfield) {
			Object value = textfield.getValue();

			if(value!=null && value instanceof Number) {
				return ((Number) value).intValue();
			}

			return null;
		}

		/*------------------------------------------------------------------------------------------*/
		/*		Constructors																		*/
		/*------------------------------------------------------------------------------------------*/

		/**
		 * This class provides a dialog to ask the user to choose an existing signature which is 
		 * used to create a suitable cospan decomposition.
		 * @param owner the parent window of this dialog
		 */
		private SignatureDialog(Window owner) {
			super(owner, messages.getString(GuiMessages.SIGNATURE_SELECTION_TITLE_KEY), 
					ACCEPT_CANCEL_LAYOUT);
			getContentPane().setBackground(MainWindow.BACKGROUND_COLOR);
			AbstractAction acceptAction = new AbstractAction(
					messages.getString(GuiMessages.ACCEPT_LABEL_KEY)) {

				private static final long serialVersionUID = 2418165190397918534L;

				@Override
				public void actionPerformed(ActionEvent e) {
					NameDataPair chosenSignature = (NameDataPair) signatureBox.getSelectedItem();
					signature = (Signature) chosenSignature.getData();
					innerInterfaceSize = getInteger(innerInterfaceSizeField);
					if(innerInterfaceSize<0) {
						showWarningMessage(GuiMessages.VALUE_MUST_BE_NON_NEGATIVE_TITLE_EXCEPTION_KEY, 
								GuiMessages.VALUE_MUST_BE_NON_NEGATIVE_TEXT_EXCEPTION_KEY,
								messages.getString(GuiMessages.INNER_INTERFACE_SIZE_LABEL_KEY));
						return;
					}

					stopDecompositionCreation = false;
					dialog.dispose();
				}

			};

			AbstractAction cancelAction = new AbstractAction(
					messages.getString(GuiMessages.CANCEL_LABEL_KEY)) {

				private static final long serialVersionUID = -2235775756823745441L;

				@Override
				public void actionPerformed(ActionEvent e) {
					dialog.dispose();
				}
			};

			setAction(ACCEPT_BUTTON, acceptAction);
			setAction(CANCEL_BUTTON, cancelAction);
			setMainPane(createMainPanel());
			pack();
		}

		/*------------------------------------------------------------------------------------------*/
		/*		Private Methods																		*/
		/*------------------------------------------------------------------------------------------*/

		/**
		 * Creates the main panel for the New-Edge-Dialog.
		 * 
		 * @return the main panel
		 */
		private JPanel createMainPanel() {
			DefaultFormBuilder builder = new DefaultFormBuilder(new FormLayout(
					"left:max(40dlu;pref), 3dlu, 80dlu:grow", ""));
			builder.border(Borders.DIALOG);

			sigLabel = new JLabel(messages.getString(GuiMessages.SIGNATURE_LABEL_KEY)+":");
			sigLabel.setOpaque(false);
			signatureBox = new DataComboBox(Repository.getRepository());
			signatureBox.addDataClass(Signature.class);
			signatureBox.addItem(new NameDataPair(
					messages.getString(GuiMessages.DEFAULT_LABEL_KEY), DEFAULT_SIGNATURE));
			signatureBox.setName(SIGNATURE_BOX_LABEL);
			signatureBox.setBackground(MainWindow.BACKGROUND_COLOR);
			signatureBox.setOpaque(false);
			signatureBox.addItemListener(new ItemListener() {

				@Override
				public void itemStateChanged(ItemEvent e) {
					updateSelectionList((Signature) ((NameDataPair) signatureBox.getSelectedItem()).getData());				
				}
			});

			selectedListModel = new DefaultListModel<>();
			selectionList = new JList<>(selectedListModel);
			selectionList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
			selectionList.setCellRenderer(new DefaultListCellRenderer() {

				private static final long serialVersionUID = -7900279308667561952L;

				@Override
				public Component getListCellRendererComponent(JList<?> list, Object value, 
						int index, boolean isSelected, boolean cellHasFocus) {
					Component comp = super.getListCellRendererComponent(list, value, 
							index, false, false);
					return comp;

				}
			});

			updateSelectionList((Signature) signatureBox.getSelectedData());

			innerInterfaceLabel = new JLabel(messages.getString(GuiMessages.INNER_INTERFACE_SIZE_LABEL_KEY)+":");
			innerInterfaceLabel.setOpaque(false);

			innerInterfaceSizeField = new JFormattedTextField(new DecimalFormat("0"));
			innerInterfaceSizeField.setValue(0);
			innerInterfaceSizeField.setColumns(10);

			builder.append(innerInterfaceLabel, innerInterfaceSizeField);
			builder.nextLine();
			builder.append(sigLabel, signatureBox);
			builder.nextLine();
			builder.append(new JScrollPane(selectionList),3);

			return builder.getPanel();
		}

		/**
		 * Updates the selection list in the SignatureDialog
		 * 
		 * @param selectedSignature the selected signature
		 */
		private void updateSelectionList(Signature selectedSignature) {

			List<Operation> sortedList = new LinkedList<>(selectedSignature.getOperations());
			Collections.sort(sortedList, Operation.COMPARATOR);
			selectedListModel.removeAllElements();					
			for(Operation letter: sortedList) {
				selectedListModel.addElement(letter);
			}
		}

	}

	private static final long serialVersionUID = 746543425346637851L;

	private static final String SIGNATURE_BOX_LABEL = "signatureBox";
	private static final String LAYOUT_SPECIFICATION = "60dlu, 3dlu, left:pref, 3dlu, 60dlu, 3dlu, 60dlu, 250dlu:grow";

	private static final String DECOMPOSITION_PANEL_LABEL = "decompositionPanel";
	private static final String COSPAN_PANEL_LABEL = "cospanPanel";

	private static final Signature DEFAULT_SIGNATURE = new Signature();

	private static DecompositionCreationDialog dialog;
	private static Integer innerInterfaceSize;
	private static GuiMessages messages;
	private static DecompositionView decompositionView;
	private static CospanView cospanView;
	private static Signature signature;
	private static JList<Operation> signatureList;
	private static DefaultListModel<Operation> operationListModel;
	private static boolean stopDecompositionCreation;
	private static JRadioButton decompositionButton;
	private static JRadioButton cospanButton;
	private static CospanDecomposition decomposition;
	private static Cospan cospan;
	private static JList<Node> interfaceList;
	private static JPanel connectPanel;
	private JList<Operation> decompositionOperations;
	private JTextField nameTextField;

	static {
		DEFAULT_SIGNATURE.add(Operation.getFuse());
		DEFAULT_SIGNATURE.add(Operation.getPerm());
		DEFAULT_SIGNATURE.add(Operation.getRes());
		DEFAULT_SIGNATURE.add(Operation.getTrans());
		DEFAULT_SIGNATURE.add(Operation.getVertex());
	}

	/*------------------------------------------------------------------------------------------*/
	/* 		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Shows the dialog for the creation of new decompositions.
	 * 
	 * @param owner the parent window of the dialog
	 */
	public static void showDialog(Window owner) {
		messages = GuiMessages.getMessages();
		stopDecompositionCreation = true;

		SignatureDialog.showDialog(owner);

		if(!stopDecompositionCreation) {	
			dialog = new DecompositionCreationDialog(owner);
			dialog.setLocationRelativeTo(owner);
			dialog.setVisible(true);
		}
	}

	/**
	 * Returns the panel which displays graph corresponding to the current decomposition.
	 * @return the graph preview panel
	 */
	private static JPanel createPreviewPanel() {

		FormLayout layout = new FormLayout("1dlu:grow", "20dlu");
		DefaultFormBuilder builder = new DefaultFormBuilder(layout);

		builder.appendSeparator(messages.getString(GuiMessages.DECOMPOSITION_PREVIEW));
		decompositionView = new DecompositionView();
		decompositionView.setPreferredSize(new Dimension(500, 500));
		decompositionView.setOpaque(false);
		decompositionView.getViewport().setOpaque(false);
		decompositionView.print(decomposition);
		//		decompositionView.getGraphControl().addMouseListener(new MouseAdapter() {			
		//			private JFrame window;
		//			
		//			@Override
		//			public void mouseClicked(MouseEvent e) {
		//				if(e.getClickCount()>=2 && window==null) {
		//					DecompositionView view = new DecompositionView();
		//					view.print(decomposition);
		//					view.setOpaque(false);
		//					view.getViewport().setOpaque(false);					
		//					window = new JFrame(messages.getString(GuiMessages.DECOMPOSITION_LABEL_KEY));
		//					window.getContentPane().add(view);
		//					window.getContentPane().setBackground(MainWindow.BACKGROUND_COLOR);
		//					window.setSize(500, 500);
		//					window.setLocationRelativeTo(null);
		//					
		//					window.addWindowListener(new WindowAdapter() {
		//						
		//						@Override
		//						public void windowClosed(WindowEvent e) {
		//							window = null;							
		//						}
		//						@Override
		//						public void windowClosing(WindowEvent e) {
		//							window.dispose();							
		//						}
		//
		//					});
		//					window.setVisible(true);
		//				}
		//			}
		//
		//		});

		JPanel decompositionPreviewPanel = new JPanel(new BorderLayout());
		decompositionPreviewPanel.setOpaque(false);
		decompositionPreviewPanel.add(builder.getPanel(), BorderLayout.NORTH);
		decompositionPreviewPanel.add(decompositionView, BorderLayout.CENTER);

		builder = new DefaultFormBuilder(layout);

		builder.appendSeparator(messages.getString(GuiMessages.COSPAN_PREVIEW));
		cospanView = new CospanView();
		cospanView.setPreferredSize(new Dimension(500, 500));
		cospanView.setOpaque(false);
		cospanView.getViewport().setOpaque(false);
		cospanView.print(cospan);
		//		cospanView.getGraphControl().addMouseListener(new MouseAdapter() {			
		//			private JFrame window;
		//			
		//			@Override
		//			public void mouseClicked(MouseEvent e) {
		//				if(e.getClickCount()>=2 && window==null) {
		//					CospanView view = new CospanView();
		//					view.print(cospan);
		//					view.setOpaque(false);
		//					view.getViewport().setOpaque(false);					
		//					window = new JFrame(messages.getString(GuiMessages.COSPAN_LABEL_KEY));
		//					window.getContentPane().add(view);
		//					window.getContentPane().setBackground(MainWindow.BACKGROUND_COLOR);
		//					window.setSize(500, 500);
		//					window.setLocationRelativeTo(null);
		//					
		//					window.addWindowListener(new WindowAdapter() {
		//						
		//						@Override
		//						public void windowClosed(WindowEvent e) {
		//							window = null;							
		//						}
		//						@Override
		//						public void windowClosing(WindowEvent e) {
		//							window.dispose();							
		//						}
		//
		//					});
		//					window.setVisible(true);
		//				}
		//			}
		//
		//		});

		JPanel cospanPreviewPanel = new JPanel(new BorderLayout());
		cospanPreviewPanel.setOpaque(false);
		cospanPreviewPanel.add(builder.getPanel(), BorderLayout.NORTH);
		cospanPreviewPanel.add(cospanView, BorderLayout.CENTER);

		final JPanel centerPanel = new JPanel(new CardLayout());
		centerPanel.setOpaque(false);
		centerPanel.add(cospanPreviewPanel, COSPAN_PANEL_LABEL);
		centerPanel.add(decompositionPreviewPanel, DECOMPOSITION_PANEL_LABEL);

		cospanButton = new JRadioButton(messages.getString(GuiMessages.COSPAN_LABEL_KEY));
		cospanButton.setName("");
		cospanButton.setOpaque(false);	
		cospanButton.setSelected(true);
		cospanButton.addActionListener(new AbstractAction() {
			
			private static final long serialVersionUID = 8061347003637341440L;
			
			@Override
			public void actionPerformed(ActionEvent e) {
				CardLayout layout = (CardLayout) (centerPanel.getLayout());
				layout.show(centerPanel, COSPAN_PANEL_LABEL);
			}
			
		});

		decompositionButton = new JRadioButton(messages.getString(GuiMessages.COSPAN_DECOMPOSITION_LABEL_KEY));
		decompositionButton.setName("");
		decompositionButton.setOpaque(false);
		decompositionButton.addActionListener(new AbstractAction() {

			private static final long serialVersionUID = 664866924215783647L;

			@Override
			public void actionPerformed(ActionEvent e) {
				CardLayout layout = (CardLayout) (centerPanel.getLayout());
				layout.show(centerPanel, DECOMPOSITION_PANEL_LABEL);
			}

		});

		ButtonGroup formatGroup = new ButtonGroup();
		formatGroup.add(cospanButton);
		formatGroup.add(decompositionButton);

		JPanel buttonPanel = new JPanel(new FlowLayout(FlowLayout.LEFT, 0, 0));
		buttonPanel.setOpaque(false);
		buttonPanel.add(cospanButton);
		buttonPanel.add(decompositionButton);

		JPanel finalPanel = new JPanel(new BorderLayout());
		finalPanel.setOpaque(false);
		finalPanel.add(centerPanel, BorderLayout.CENTER);
		finalPanel.add(buttonPanel, BorderLayout.SOUTH);

		return finalPanel;
	}

	/**
	 * Sorts a DefaultListModel by converting it to a list and sort it by the operation comparator.
	 * @param unsorted the DefaultListModel that needs to be sorted
	 * @return the sorted list of operations
	 */
	private static List<Operation> sortListModel(DefaultListModel<Operation> unsorted) {
		List<Operation> sortedList = Collections.list(
				unsorted.elements());
		Collections.sort(sortedList, Operation.COMPARATOR);			
		return sortedList;		
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * 
	 * @param owner the parent window of the dialog
	 */
	private DecompositionCreationDialog(final Window owner) {
		super(owner, messages.getString(GuiMessages.DECOMPOSITION_CREATION_DIALOG_TITLE_KEY),
				RavenDialog.ACCEPT_CANCEL_LAYOUT);
		getContentPane().setBackground(MainWindow.BACKGROUND_COLOR);
		decomposition = new CospanDecomposition(innerInterfaceSize);
		cospan = new Cospan(decomposition);

		AbstractAction acceptAction = new AbstractAction(
				messages.getString(GuiMessages.ACCEPT_LABEL_KEY)) {

			private static final long serialVersionUID = 2418165190397918534L;

			@Override
			public void actionPerformed(ActionEvent e) {

				String name = nameTextField.getText();
				if (name.isEmpty()) {
					showErrorMessage(GuiMessages.EMPTY_INPUTFIELD_TITLE_EXCEPTION_KEY,
							GuiMessages.EMPTY_INPUTFIELD_TEXT_EXCEPTION_KEY, 
							messages.getString(GuiMessages.NAME_LABEL_KEY));
					return;
				}
				if(!Repository.isNameValid(name)){
					showErrorMessage(GuiMessages.INVALID_NAME_TITLE_EXCEPTION_KEY,
							GuiMessages.INVALID_NAME_TEXT_EXCEPTION_KEY, name);
					return;
				}
				if (Repository.getRepository().contains(name)) {
					showInformationMessage(GuiMessages.NAME_ALREADY_IN_USE_TITLE_EXCEPTION_KEY,
							GuiMessages.NAME_ALREADY_IN_USE_TEXT_EXCEPTION_KEY, name);
				} else {
					Repository.getRepository().put(name, decomposition);
					dialog.dispose();
				}
			}

		};

		setMainPane(createMainPanel());
		setAction(ACCEPT_BUTTON, acceptAction);
		pack();
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the panel which displays components for the creation of the
	 * signature.
	 * 
	 * @return the main panel
	 */
	private JPanel createMainPanel() {

		JPanel namePanel = new JPanel(new BorderLayout());
		namePanel.setOpaque(false);
		nameTextField = new JTextField(10);

		FormLayout layout = new FormLayout("80dlu, 3dlu, 40dlu:grow", "");
		DefaultFormBuilder builder = new DefaultFormBuilder(layout);
		builder.border(Borders.DIALOG);

		builder.append(messages.getString(GuiMessages.DECOMPOSITION_NAME_LABEL_KEY)+":", nameTextField);
		builder.nextLine();
		builder.appendSeparator();
		namePanel.add(builder.getPanel(),BorderLayout.CENTER);


		layout = new FormLayout(LAYOUT_SPECIFICATION, "10dlu, 100dlu, 20dlu, 10dlu, 300dlu:grow");
		builder = new DefaultFormBuilder(layout);
		builder.border(Borders.DIALOG);

		builder.append(messages.getString(GuiMessages.OPERATIONS_LABEL_KEY));
		builder.append("");
		builder.append(messages.getString(GuiMessages.DECOMPOSITION_LABEL_KEY));
		builder.append(messages.getString(GuiMessages.CURRENT_INTERFACE_LABEL_KEY));
		builder.nextLine();

		List<Operation> sortedList = new LinkedList<>(signature.getOperations());
		Collections.sort(sortedList, Operation.COMPARATOR);
		operationListModel = new DefaultListModel<>();
		for(Operation letter: sortedList) {
			operationListModel.addElement(letter);
		}
		signatureList = new JList<>(operationListModel);
		signatureList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

		builder.append(new JScrollPane(signatureList));

		JButton shiftRightButton = new JButton(new AbstractAction(
				messages.getString(GuiMessages.SHIFT_RIGHT_LABEL_KEY)) {

			private static final long serialVersionUID = -2632710493259765026L;

			@Override
			public void actionPerformed(ActionEvent e) {
				Operation op = signatureList.getSelectedValue();
				if (op == null) {
					showWarningMessage(GuiMessages.NO_OPERATION_SELECTED_TITLE_EXCEPTION_KEY,
							GuiMessages.NO_OPERATION_SELECTED_TEXT_EXCEPTION_KEY);
				} else {
					if (decomposition.isApplicable(op)) {
						addLetter(op);
					}
					else {
						// the operation can not be applied
						showInformationMessage(
								GuiMessages.OPERATION_NOT_APPLICABLE_TITLE_EXCEPTION_KEY,
								GuiMessages.OPERATION_NOT_APPLICABLE_TEXT_EXCEPTION_KEY, op.toString());
					}
				}
			}

		});

		JButton shiftLeftButton = new JButton(new AbstractAction(
				messages.getString(GuiMessages.SHIFT_LEFT_LABEL_KEY)) {

			private static final long serialVersionUID = -2632710493259765026L;

			@Override
			public void actionPerformed(ActionEvent e) {
				if(decompositionOperations.getModel().getSize()>0) {
					removeLetter();
				}
			}

		});


		JPanel buttonPanel = new JPanel(new GridLayout(3, 1, 0, 5));
		buttonPanel.setOpaque(false);
		buttonPanel.add(shiftRightButton);
		buttonPanel.add(shiftLeftButton);		
		builder.append(buttonPanel);

		decompositionOperations = new JList<>(new DefaultListModel<Operation>());
		decompositionOperations.setCellRenderer(new DefaultListCellRenderer() {

			private static final long serialVersionUID = 6442759808071453045L;

			@Override
			public Component getListCellRendererComponent(JList<?> list, Object value, 
					int index, boolean isSelected, boolean cellHasFocus) {
				Component comp = super.getListCellRendererComponent(list, value, 
						index, false, false);
				return comp;			
			}
		});
		builder.append(new JScrollPane(decompositionOperations));


		DefaultListModel<Node> interfaceListModel = new DefaultListModel<>();
		for(Node node: decomposition.getOuterInterface()) {
			interfaceListModel.addElement(node);
		}
		interfaceList = new JList<>(interfaceListModel);
		interfaceList.setCellRenderer(new DefaultListCellRenderer() {

			private static final long serialVersionUID = 6442759808071453045L;

			@Override
			public Component getListCellRendererComponent(JList<?> list, Object value,
					int index, boolean isSelected, boolean cellHasFocus) {
				JLabel label = (JLabel) super.getListCellRendererComponent(list, value, index, 
						isSelected, cellHasFocus);
				label.setText(String.valueOf(index) + ": " + label.getText());

				return label;
			}

		});
		builder.append(new JScrollPane(interfaceList));

		builder.nextLine();

		JButton edgeButton = new JButton(new AbstractAction("",
				new ImageIcon(RavenMain.PLUS_ICON_FILENAME)) {

			private static final long serialVersionUID = 6480710296758476298L;

			@Override
			public void actionPerformed(ActionEvent e) {
				NewLabelDialog.showDialog(DecompositionCreationDialog.this);
			}
		});

		connectPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
		connectPanel.add(new JLabel(messages.getString(GuiMessages.ADD_OPERATION_KEY, Console.CONNECT_TOKEN)));
		connectPanel.add(edgeButton);
		connectPanel.setOpaque(false);
		connectPanel.setVisible(signature.equals(DEFAULT_SIGNATURE));

		builder.append(connectPanel, 8);
		builder.nextLine();
		builder.nextLine();
		builder.append(createPreviewPanel(),8);

		JPanel mainPanel = new JPanel();
		mainPanel.setOpaque(false);
		mainPanel.setLayout(new BorderLayout());
		mainPanel.add(namePanel, BorderLayout.NORTH);
		mainPanel.add(builder.getPanel(), BorderLayout.CENTER);
		return mainPanel;
	}

	/**
	 * Applies the given operation {@code letter} and returns whether it could be applied. 
	 * @param letter the name of the operation which is to be applied
	 */
	private void addLetter(Operation letter) {
		DefaultListModel<Operation> decompositionModel = (DefaultListModel<Operation>) 
				decompositionOperations.getModel();
		DefaultListModel<Node> interfaceModel = (DefaultListModel<Node>) 
				interfaceList.getModel();

		//Add letter to decomposition
		decomposition.add(letter);
		decompositionModel.addElement(letter);

		//Refresh GUI elements
		interfaceModel.removeAllElements();
		for(Node node: decomposition.getOuterInterface()) {
			interfaceModel.addElement(node);
		}

		// visualize the new graph
		decompositionView.print(decomposition);

		cospan = new Cospan(decomposition);
		cospanView.print(cospan);
	}

	/**
	 * Removes the last letter from the decomposition. 
	 */
	private void removeLetter() {
		DefaultListModel<Operation> decompositionModel = (DefaultListModel<Operation>) 
				decompositionOperations.getModel();
		DefaultListModel<Node> interfaceModel = (DefaultListModel<Node>) 
				interfaceList.getModel();

		//Create a new decomposition without the last letter and overwrite the old one
		CospanDecomposition newDecomp = new CospanDecomposition(innerInterfaceSize);

		decompositionModel.remove(decompositionModel.getSize()-1);

		for(int i=0; i<decompositionModel.getSize(); i++) {
			newDecomp.add(decompositionModel.get(i));
		}

		decomposition = newDecomp;

		//Refresh GUI elements
		interfaceModel.removeAllElements();
		for(Node node: decomposition.getOuterInterface()) {
			interfaceModel.addElement(node);
		}

		// visualize the new graph
		decompositionView.print(decomposition);

		cospan = new Cospan(decomposition);
		cospanView.print(cospan);
	}

}
