/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui;

import java.awt.Dimension;
import java.awt.FlowLayout;

import javax.swing.ButtonGroup;
import javax.swing.JPanel;
import javax.swing.JRadioButton;

import com.jgoodies.forms.builder.DefaultFormBuilder;

import de.uni_due.inf.ti.raven.actions.ComputationEvent;
import de.uni_due.inf.ti.raven.actions.ComputationListener;
import de.uni_due.inf.ti.raven.actions.EmptinessCheckAction;
import de.uni_due.inf.ti.raven.actions.ComputationEvent.EventType;
import de.uni_due.inf.ti.raven.algorithms.Algorithm.SearchType;
import de.uni_due.inf.ti.raven.algorithms.EmptinessCheckAlgorithm;
import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.data.automata.Automaton;

/**
 * This class implements a {@code ControlPanel} for the invocation of the different emptiness 
 * checking algorithms implemented in Raven.
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.7 $
 * 
 * @see EmptinessCheckAlgorithm
 */
public final class EmptinessControlPanel extends ControlPanel {

	private static final long serialVersionUID = -2728127068888693659L;

	/**
	 * This class is used as a worker thread to perform the emptiness check. 
	 */
	private class EmptinessWorkerThread extends WorkerThread {

		private String automatonName;
		private EmptinessCheckAction action;

		@Override
		protected Void doInBackground() throws Exception {
			// Get the parameter for the invariant checking algorithm
			Automaton automaton = (Automaton) automatonBox.getSelectedData();

			// Store the names of the parameter (which will be used later on for the output)
			automatonName = automatonBox.getSelectedName();

			// Determine the search type to be used (either forward or backward)
			SearchType type = null;
			if (forwardButton.isSelected()) {
				type = SearchType.forward;
			}
			else if (backwardButton.isSelected()) {
				type = SearchType.backward;
			}

			if(automaton==null) {
				Utilities.showWarningMessage(EmptinessControlPanel.this,
						GuiMessages.NO_AUTOMATON_SELECTED_TITLE_EXCEPTION_KEY,
						GuiMessages.NO_AUTOMATON_SELECTED_TEXT_EXCEPTION_KEY,
						messages.getString(GuiMessages.AUTOMATON_LABEL_KEY));
				cancel(true);
				fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_CANCELLED, "", 0, 0));
				return null;
			}

			// Create emptiness action
			action = new EmptinessCheckAction(type, automaton);

			for(ComputationListener listener : listeners) {
				action.addComputationListener(listener);						
			}

			// Register listeners
			for(ComputationListener listener : listeners) {
				action.addComputationListener(listener);
			}

			// Start computation
			action.execute();
			return null;
		}

		@Override
		protected void done() {
			//If the worker thread has been cancelled
			if(isCancelled()) {
				if(action!=null) {
					//also cancel the computation of the action
					action.cancelComputation();
				}
				
				action = null;
				automatonName = null;
			}
			//else bring up message dialog to display result
			else {
				boolean isEmpty = action.getResult();
				if (isEmpty) {
					Utilities.showInformationMessage(EmptinessControlPanel.this,
							GuiMessages.EMPTINESS_DOES_HOLD_DIALOG_TITLE_KEY, 
							GuiMessages.EMPTINESS_DOES_HOLD_DIALOG_TEXT_KEY,
							automatonName);
				}
				else {
					Utilities.showInformationMessage(EmptinessControlPanel.this,
							GuiMessages.EMPTINESS_DOES_NOT_HOLD_DIALOG_TITLE_KEY,
							GuiMessages.EMPTINESS_DOES_NOT_HOLD_DIALOG_TEXT_KEY,
							automatonName);
				}
			}
		}

	}

	/** Automaton Box Label (used for Testing) */
	protected static final String AUTOMATON_BOX_NAME = "automatonBox";

	/** Forward Search Check Box Label (used for Testing) */
	protected static final String FORWARD_SEARCH_RADIO_BUTTON_NAME = "forwardSearch";

	/** Backward Search Check Box Label (used for Testing) */
	protected static final String BACKWARD_SEARCH_RADIO_BUTTON_NAME = "backwardSearch";

	private DataComboBox automatonBox;
	private JRadioButton forwardButton;
	private JRadioButton backwardButton;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public void setEnabled(boolean isEnabled) {
		automatonBox.setEnabled(isEnabled);
		forwardButton.setEnabled(isEnabled);
		backwardButton.setEnabled(isEnabled);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	protected void initialize(DefaultFormBuilder builder) {
		Repository repo = Repository.getRepository();

		automatonBox = new DataComboBox(repo);
		automatonBox.setName(AUTOMATON_BOX_NAME);
		automatonBox.addDataClass(Automaton.class);
		automatonBox.setMaximumSize(new Dimension(100, 25));
		append(builder, GuiMessages.AUTOMATON_LABEL_KEY, automatonBox);

		forwardButton = new JRadioButton(messages.getString(GuiMessages.FORWARD_LABEL_KEY));
		forwardButton.setName(FORWARD_SEARCH_RADIO_BUTTON_NAME);
		forwardButton.setOpaque(false);
		forwardButton.setSelected(true);

		backwardButton = new JRadioButton(messages.getString(GuiMessages.BACKWARD_LABEL_KEY));
		backwardButton.setName(BACKWARD_SEARCH_RADIO_BUTTON_NAME);
		backwardButton.setOpaque(false);

		ButtonGroup searchTypeGroup = new ButtonGroup();
		searchTypeGroup.add(forwardButton);
		searchTypeGroup.add(backwardButton);

		JPanel searchTypePanel = new JPanel(new FlowLayout(FlowLayout.LEFT, 0, 0));
		searchTypePanel.setOpaque(false);
		searchTypePanel.add(forwardButton);
		searchTypePanel.add(backwardButton);
		append(builder, GuiMessages.SEARCHTYPE_LABEL_KEY, searchTypePanel);

		// Set color
		forwardButton.setBackground(getBackground());
		backwardButton.setBackground(getBackground());
		automatonBox.setBackground(getBackground());
	}

	@Override
	protected WorkerThread getWorkerThread() {
		return new EmptinessWorkerThread();
	}

}
