/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.text.DecimalFormat;
import java.util.LinkedList;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.BorderFactory;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JFormattedTextField;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JTextField;
import javax.swing.SpringLayout;

import com.mxgraph.model.mxCell;

import de.uni_due.inf.ti.raven.actions.AddToRepositoryAction;
import de.uni_due.inf.ti.raven.data.Edge;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.Node;
import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.gui.NewEdgeDialog.EdgeData;
import de.uni_due.inf.ti.raven.gui.graphs.GraphView;
import de.uni_due.inf.ti.raven.gui.graphs.VisualGraph;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * This class provides a dialog to create a new graph. Please note that the dialog is modal, i.e. 
 * the {@code showDialog} method blocks the caller until the user's interaction is complete. 
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.46 $
 * 
 */
public class GraphCreationDialog extends RavenDialog {

	/**
	 * This class provides an action to create a new random graph.
	 */
	private class CreateRandomGraphAction extends AbstractAction {

		private static final long serialVersionUID = -5016215291184560913L;

		/**
		 * Creates a new {@code CreateRandomGraphAction}.
		 */
		public CreateRandomGraphAction() {
			super(messages.getString(GuiMessages.CREATE_RANDOM_GRAPH_LABEL_KEY));
		}

		@Override
		public void actionPerformed(ActionEvent e) {
			//Build the main panel
			JLabel iconLabel = new JLabel(new ImageIcon(
					RavenMain.DIALOG_QUESTION_ICON_FILENAME));
			iconLabel.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));

			JLabel messageLabel = new JLabel(messages.getString(
					GuiMessages.CREATE_RANDOM_GRAPH_DIALOG_TEXT_KEY));

			final JFormattedTextField inputField = new JFormattedTextField(
					new DecimalFormat("#"));

			JPanel messagePanel = new JPanel(new GridLayout(2, 1));
			messagePanel.add(messageLabel);
			messagePanel.add(inputField);
			messagePanel.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));

			JPanel mainPane = new JPanel(new BorderLayout());
			mainPane.add(iconLabel, BorderLayout.WEST);
			mainPane.add(messagePanel, BorderLayout.CENTER);

			//Create the dialog
			final RavenDialog dialog = new RavenDialog(GraphCreationDialog.this, 
					messages.getString(GuiMessages.CREATE_RANDOM_GRAPH_DIALOG_TITLE_KEY), 
					RavenDialog.ACCEPT_CANCEL_LAYOUT) {

				private static final long serialVersionUID = -1067570019014871553L;

			};
			dialog.setMainPane(mainPane);
			dialog.setAction(ACCEPT_BUTTON, new AbstractAction(
					messages.getString(GuiMessages.ACCEPT_LABEL_KEY)) {

				private static final long serialVersionUID = 3756730273954021809L;

				@Override
				public void actionPerformed(ActionEvent e) {
					//Get value and check it
					Object value = inputField.getValue();
					if(value!=null && !value.toString().isEmpty()) {
						int maximumEdges = Integer.parseInt(value.toString());
						if(maximumEdges>=0) {
							//If the value is valid a new random graph
							de.uni_due.inf.ti.raven.actions.CreateRandomGraphAction action;
							action = new de.uni_due.inf.ti.raven.actions.CreateRandomGraphAction("RandomGraph", maximumEdges);
							action.execute();
							graph = action.getResult();
							graphView.print(graph);
							dialog.dispose();
						}
						else {
							Utilities.showWarningMessage(dialog, 
									GuiMessages.VALUE_MUST_BE_NON_NEGATIVE_TITLE_EXCEPTION_KEY, 
									GuiMessages.VALUE_MUST_BE_NON_NEGATIVE_TEXT_EXCEPTION_KEY, "Maximum Number of Edges");
						}
					}
					else {
						Utilities.showWarningMessage(dialog, 
								GuiMessages.INPUT_MISSING_TITLE_EXCEPTION_KEY, 
								GuiMessages.INPUT_MISSING_TEXT_EXCEPTION_KEY);
					}

				}
			});


			dialog.pack();
			dialog.setLocationRelativeTo(GraphCreationDialog.this);
			dialog.setResizable(false);
			dialog.setVisible(true);

		}

	}

	/**
	 * This class provides an action to add a new node to the graph created by the 
	 * corresponding {@code GraphCreationDialog}.
	 */
	private class AddNodeAction extends AbstractAction {

		private static final long serialVersionUID = -5016215291184560913L;

		/**
		 * Creates a new {@code AddNodeAction}.
		 */
		public AddNodeAction() {
			super(messages.getString(GuiMessages.ADD_NODE_LABEL_KEY));
		}

		@Override
		public void actionPerformed(ActionEvent e) {
			graph.addNode(NODE_NAME_PREFIX + nodeCount);
			nodeCount++;
			removeNodeAction.setEnabled(true);

			new Thread(new Runnable() {

				@Override
				public void run() {
					graphView.print(graph);
				}

			}).start();
		}

	}

	/**
	 * This class provides an action to remove a node from the graph created by 
	 * the corresponding {@code GraphCreationDialog}.
	 */
	private class RemoveNodeAction extends AbstractAction {

		private static final long serialVersionUID = -5016215291184560913L;

		/**
		 * Creates a new {@code RemoveNodeAction}.
		 */
		public RemoveNodeAction() {
			super(messages.getString(GuiMessages.REMOVE_NODE_LABEL_KEY));
		}

		@Override
		public void actionPerformed(ActionEvent e) {
			//Get a list of selected nodes
			if(graphView.getGraph().getSelectionCount() > 0) {
				Object[] cells = graphView.getGraph().getSelectionCells();
				for(Object object: cells) {
					mxCell cell = (mxCell) object;
					if(cell.getStyle().equals(VisualGraph.HYPERGRAPH_VERTEX_STYLE)) {
						graph.removeNode(cell.getValue().toString());
					}
				}

				removeNodeAction.setEnabled(graph.getNodeCount()>0);

				new Thread(new Runnable() {

					@Override
					public void run() {
						graphView.print(graph);
					}

				}).start();
			}
			else {
				showWarningMessage(GuiMessages.NO_NODE_SELECTED_TITLE_EXCEPTION_KEY,
						GuiMessages.NO_NODE_SELECTED_TEXT_EXCEPTION_KEY);
			}

		}

	}

	/**
	 * This class provides an action to add a new edge to the graph created by the 
	 * corresponding {@code GraphCreationDialog}.
	 */
	private class AddEdgeAction extends AbstractAction {

		private static final long serialVersionUID = -5016215291184560913L;

		/**
		 * Creates a new {@code AddEdgeAction}.
		 */
		public AddEdgeAction() {
			super(messages.getString(GuiMessages.ADD_EDGE_LABEL_KEY));
		}

		@Override
		public void actionPerformed(ActionEvent e) {
			//Get a list of select nodes
			List<Node> presetNodes = new LinkedList<>();
			if(graphView.getGraph().getSelectionCount() > 0) {
				Object[] cells = graphView.getGraph().getSelectionCells();
				for(Object object: cells) {
					mxCell cell = (mxCell) object;
					if(cell.getStyle().equals(VisualGraph.HYPERGRAPH_VERTEX_STYLE)) {
						presetNodes.add(graph.getNode(cell.getValue().toString()));
					}
				}						
			}

			//Show the new-edge-dialog
			EdgeData edgeData = NewEdgeDialog.showDialog(
					GraphCreationDialog.this, graph, presetNodes);

			//If the user has not cancelled the dialog, add the new edge
			if (edgeData!=null) {
				graph.addEdge(edgeData.getName(), edgeData.getLabel(), edgeData.getIncidentNodes());
				removeEdgeAction.setEnabled(true);

				new Thread(new Runnable() {

					@Override
					public void run() {
						graphView.print(graph);
					}

				}).start();
			}
		}

	}

	/**
	 * This class provides an action to remove an edge from the graph created by 
	 * the corresponding {@code GraphCreationDialog}.
	 */
	private class RemoveEdgeAction extends AbstractAction {

		private static final long serialVersionUID = -5016215291184560913L;

		/**
		 * Creates a new {@code RemoveEdgeAction}.
		 */
		public RemoveEdgeAction() {
			super(messages.getString(GuiMessages.REMOVE_EDGE_LABEL_KEY));
		}

		@Override
		public void actionPerformed(ActionEvent e) {
			//Show the new-edge-dialog
			Edge selectedEdge = RemoveEdgeDialog.showDialog(
					GraphCreationDialog.this, graph);

			//If the user has not cancelled the dialog, add the new edge
			if (selectedEdge!=null) {
				graph.removeEdge(selectedEdge);
				removeEdgeAction.setEnabled(graph.getEdgeCount()>0);

				new Thread(new Runnable() {

					@Override
					public void run() {
						graphView.print(graph);
					}

				}).start();
			}
		}

	}

	private static final long serialVersionUID = 746543425346637851L;


	private static final String NAME_FIELD_LABEL = "nameField";
	private static final String NODE_NAME_PREFIX = "n";
	// private static final String EDGE_NAME_PREFIX = "e";

	private static GraphCreationDialog dialog;

	private static GuiMessages messages;

	private static GraphView graphView;
	private static Graph graph;

	private int nodeCount;
	// private int edgeCount;

	private final RemoveNodeAction removeNodeAction;
	private final RemoveEdgeAction removeEdgeAction;

	/*------------------------------------------------------------------------------------------*/
	/* 		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Shows the dialog for the creation of new signatures.
	 * @param owner the parent window of the dialog
	 */
	public static void showDialog(Window owner) {
		messages = GuiMessages.getMessages();
		dialog = new GraphCreationDialog(owner);

		dialog.setLocationRelativeTo(owner);
		dialog.setVisible(true);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new graph creation dialog.
	 * @param owner the parent window of this dialog
	 */
	private GraphCreationDialog(Window owner) {
		super(owner, messages.getString(GuiMessages.GRAPH_CREATION_DIALOG_TITLE_KEY),
				RavenDialog.ACCEPT_CANCEL_LAYOUT);

		AbstractAction acceptAction = new AbstractAction(
				messages.getString(GuiMessages.ACCEPT_LABEL_KEY)) {

			private static final long serialVersionUID = 2418165190397918534L;

			@Override
			public void actionPerformed(ActionEvent e) {
				JTextField nameField = (JTextField) Utilities.getComponentByName(
						NAME_FIELD_LABEL, GraphCreationDialog.this);
				String name = nameField.getText();
				if (name.isEmpty()) {
					showErrorMessage(GuiMessages.EMPTY_INPUTFIELD_TITLE_EXCEPTION_KEY,
							GuiMessages.EMPTY_INPUTFIELD_TEXT_EXCEPTION_KEY,
							messages.getString(GuiMessages.NAME_LABEL_KEY));
					nameField.requestFocus();
					return;
				}
				if(!Repository.isNameValid(name)){
					showErrorMessage(GuiMessages.INVALID_NAME_TITLE_EXCEPTION_KEY,
							GuiMessages.INVALID_NAME_TEXT_EXCEPTION_KEY, name);
					nameField.requestFocus();
					return;
				}
				if (Repository.getRepository().contains(name)) {
					showInformationMessage(GuiMessages.NAME_ALREADY_IN_USE_TITLE_EXCEPTION_KEY,
							GuiMessages.NAME_ALREADY_IN_USE_TEXT_EXCEPTION_KEY, name);
					nameField.requestFocus();
					return;
				}

				graph.setName(name);
				new AddToRepositoryAction(name, graph).execute();
				GraphCreationDialog.this.dispose();
			}

		};

		removeNodeAction = new RemoveNodeAction();
		removeNodeAction.setEnabled(false);
		
		removeEdgeAction = new RemoveEdgeAction();
		removeEdgeAction.setEnabled(false);

		getContentPane().setBackground(MainWindow.BACKGROUND_COLOR);
		setMainPane(createMainPanel());
		setAction(ACCEPT_BUTTON, acceptAction);
		pack();

		nodeCount = 0;
		// edgeCount = 0;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the panel which displays components for the creation of the signature.
	 * @return the main panel
	 */
	private JPanel createMainPanel() {
		graph = new Graph("");
		String labelSuffix = ": ";

		JLabel nameLabel = new JLabel(
				messages.getString(GuiMessages.NAME_LABEL_KEY) + labelSuffix);
		nameLabel.setOpaque(false);
		JTextField nameField = new JTextField(10);
		nameField.setOpaque(false);
		nameField.setName(NAME_FIELD_LABEL);

		SpringLayout layout = new SpringLayout();
		JPanel namePanel = new JPanel(layout);
		namePanel.setOpaque(false);
		namePanel.add(nameLabel);
		namePanel.add(nameField);

		Utilities.makeGridLayout(layout, namePanel, 2, 10, 10);

		final JPopupMenu popup = new JPopupMenu();
		popup.add(new AddNodeAction());
		popup.add(removeNodeAction);
		popup.add(new AddEdgeAction());
		popup.add(removeEdgeAction);

		MouseListener popupListener = new MouseAdapter() {

			@Override
			public void mouseReleased(MouseEvent e) {
				if (e.isPopupTrigger()) {
					popup.show(e.getComponent(), e.getX(), e.getY());
				}
			}

		};

		graphView = new GraphView(this);
		graphView.setOpaque(false);
		graphView.getViewport().setOpaque(false);
		graphView.setPreferredSize(new Dimension(500, 500));
		graphView.getGraphControl().addMouseListener(popupListener);

		JPanel panel = new JPanel(new BorderLayout());
		panel.setOpaque(false);
		panel.add(namePanel, BorderLayout.NORTH);
		panel.add(graphView, BorderLayout.CENTER);

		JPanel buttonPanel = new JPanel();
		buttonPanel.setOpaque(false);
		buttonPanel.add(new JButton(new CreateRandomGraphAction()));
		buttonPanel.add(new JButton(new AddNodeAction()));
		buttonPanel.add(new JButton(removeNodeAction));
		buttonPanel.add(new JButton(new AddEdgeAction()));
		buttonPanel.add(new JButton(removeEdgeAction));
		panel.add(buttonPanel, BorderLayout.SOUTH);

		return panel;
	}

}
