/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui;

import java.awt.Color;

import javax.swing.BorderFactory;
import javax.swing.JTextField;
import javax.swing.border.BevelBorder;

import com.jgoodies.forms.builder.DefaultFormBuilder;
import com.jgoodies.forms.layout.FormLayout;

import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.gui.graphs.GraphView;

/**
 * This class provides a builder which is used to  create an {@code InformationPanel} 
 * for displaying information about <i>graphs</i>.
 * @author Christoph Blume
 * @version $Revision: 1.17 $
 *
 * @see Graph
 */
public final class GraphInformationPanelBuilder extends InformationPanelBuilder<Graph> {

	/**
	 * 
	 * @author Christoph Blume
	 *
	 */
	private static class GraphInformationPanel extends InformationPanel<Graph> {

		private static final long serialVersionUID = -1821154740159142849L;

		private static final Repository repo = Repository.getRepository();
		private static final GuiMessages messages = GuiMessages.getMessages();

		/*------------------------------------------------------------------------------------------*/
		/* 		Public Methods																		*/
		/*------------------------------------------------------------------------------------------*/

		@Override
		public void setData(Graph graph) {
			graphNameField.setText(repo.getName(graph));

			int numberOf = graph.getNodeCount();
			String key = numberOf < 2 ? GuiMessages.NODE_LABEL_KEY : GuiMessages.NODES_LABEL_KEY;
			nodeNumberField.setText(numberOf+" "+messages.getString(key));

			numberOf = graph.getEdgeCount();
			key = numberOf < 2 ? GuiMessages.EDGE_LABEL_KEY : GuiMessages.EDGES_LABEL_KEY;
			edgeNumberField.setText(numberOf+" "+messages.getString(key));

			graphView.print(graph);
		}

	}

	private static GraphInformationPanelBuilder panelBuilder;
	private static InformationPanel<Graph> panel;
	private static GraphView graphView;
	private static JTextField graphNameField;
	private static JTextField nodeNumberField;
	private static JTextField edgeNumberField;

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates an information panel displaying properties of a given graph.
	 * @return the graph information panel
	 */
	public static InformationPanel<Graph> createPanel() {
		if(panelBuilder==null) {
			panel = new GraphInformationPanel();
			DefaultFormBuilder builder = new DefaultFormBuilder(
					new FormLayout(LAYOUT_SPECIFICATION), panel);
			panelBuilder = new GraphInformationPanelBuilder(builder);
		}

		return panel;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates the panel on which graph information will be displayed.
	 * @param formBuilder the form builder used to build the panel
	 */
	public GraphInformationPanelBuilder(DefaultFormBuilder formBuilder) {
		super(formBuilder);

		formBuilder.getPanel().setOpaque(false);

		graphNameField = new JTextField(10);
		graphNameField.setEditable(false);
		graphNameField.setOpaque(false);
		append(GuiMessages.NAME_LABEL_KEY, graphNameField);

		appendSeparator();

		nodeNumberField = new JTextField();
		nodeNumberField.setEditable(false);
		nodeNumberField.setColumns(10);
		nodeNumberField.setOpaque(false);
		append(GuiMessages.NODE_NUMBER_LABEL_KEY, nodeNumberField);

		edgeNumberField = new JTextField();
		edgeNumberField.setEditable(false);
		edgeNumberField.setColumns(10);
		edgeNumberField.setOpaque(false);
		append(GuiMessages.EDGE_NUMBER_LABEL_KEY, edgeNumberField);

		appendSeparator();

		graphView = new GraphView(panel);
		graphView.setSaveButtonVisible(true);
		graphView.setMaximizeViewButtonVisible(true);
		graphView.setBorder(BorderFactory.createLoweredBevelBorder());
		graphView.setBorder(BorderFactory.createBevelBorder(
				BevelBorder.LOWERED, Color.LIGHT_GRAY, Color.GRAY));

		graphView.setOpaque(false);
		graphView.getViewport().setOpaque(false);

		append(graphView, 3);

	}

}
