/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.IOException;
import java.util.LinkedList;
import java.util.List;

import javax.imageio.ImageIO;
import javax.swing.BorderFactory;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JSplitPane;
import javax.swing.JTextField;

import com.jgoodies.forms.builder.DefaultFormBuilder;
import com.jgoodies.forms.layout.FormLayout;
import com.mxgraph.model.mxCell;
import com.mxgraph.util.mxEventObject;
import com.mxgraph.util.mxEventSource.mxIEventListener;
import com.mxgraph.view.mxGraph;
import com.mxgraph.view.mxGraphSelectionModel;

import de.uni_due.inf.ti.raven.data.Cospan;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.gui.automaton.AutomatonRunView;
import de.uni_due.inf.ti.raven.gui.graphs.CospanView;
import de.uni_due.inf.ti.raven.gui.graphs.DecompositionView;
import de.uni_due.inf.ti.raven.gui.graphs.GraphView;
import de.uni_due.inf.ti.raven.io.GraphPrinter;
import de.uni_due.inf.ti.raven.main.RavenMain;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class provides a window which is used to display a single graph.
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.37 $
 *
 * @see Graph
 */
public class GraphWindow extends JFrame implements GraphPrinter {

	private static final long serialVersionUID = -977721595405898943L;

	private static GraphWindow window;

	private final GuiMessages messages; 
	private Repository repository;

	private JTextField automatonNameField;
	private JTextField typeField;
	private JTextField numberOfStatesField;
	private JTextField numberOfInitialStatesField;
	private JTextField numberOfFinalStatesField;
	private JTextField maximumInterfaceField;
	private JTextField innerInterfaceField;
	private JTextField outerInterfaceField;
	
	private CospanView stateInformationView;

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the graph window.
	 * @return the graph window
	 */
	public static GraphWindow getWindow() {
		if(window==null) {
			window = new GraphWindow();
		}

		return window;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new graph window.
	 */
	private GraphWindow() {
		repository = Repository.getRepository();
		messages = GuiMessages.getMessages();
		
		try {
			setIconImage(ImageIO.read(new File(RavenMain.RAVEN_ICON_FILENAME)));
		} catch (IOException e) {
			//Do nothing
		}

		addWindowListener(new WindowAdapter() {

			@Override
			public void windowClosing(WindowEvent e) {
				e.getWindow().dispose();
			}

		});

		getContentPane().setBackground(MainWindow.BACKGROUND_COLOR);
		setLocationRelativeTo(null);
		setExtendedState(JFrame.MAXIMIZED_BOTH);
		pack();
		setVisible(true);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public void print(Graph graph) {
		if(graph==null) {
			throw new NullPointerException(RavenExceptionMessages.GRAPH_MUST_NOT_BE_NULL_EXCEPTION);
		}

		GraphView graphView = new GraphView(this);
		graphView.getViewport().setOpaque(false);
		graphView.print(graph);

		getContentPane().add(graphView);
		setTitle(messages.getString(GuiMessages.GRAPH_LABEL_KEY)+" '"+repository.getName(graph)+"'");
		pack();
		setLocationRelativeTo(null);
		setVisible(true);
	}

	/**
	 * Shows the graph with the name {@code name} in this window.
	 * @param name the name of the graph to be shown
	 */
	public void showGraph(String name) {
		if(name==null) {
			throw new NullPointerException(
					RavenExceptionMessages.GRAPH_NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(!repository.isGraph(name)) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.GRAPH_NOT_FOUND_EXCEPTION);
		}

		print((Graph) repository.get(name));
	}

	//TEST PURPOSES
	/**
	 * Shows the Decomposition {@code decomposition} in this window.
	 * @param decomposition the decomposition to be shown
	 * @param title the title of the window
	 */
	public void showDecomposition(CospanDecomposition decomposition, String title) {
		if(decomposition==null) {
			throw new NullPointerException(
					RavenExceptionMessages.DECOMPOSITION_MUST_NOT_BE_NULL_EXCEPTION);
		}

		DecompositionView decompositionView = new DecompositionView();		
		decompositionView.print(decomposition);

		setContentPane(decompositionView);
		setExtendedState(JFrame.MAXIMIZED_BOTH);
		setTitle(title);
		setVisible(true);
		pack();
	}

	/**
	 * Shows the two graphs {@code left} and {@code right} in split view side-by-side.
	 * @param left the left graph
	 * @param right the right graph
	 * @param title the title of the window
	 */
	public void showTwoGraphs(Graph left, Graph right, String title) {
		if(left==null || right==null) {
			throw new NullPointerException(
					RavenExceptionMessages.GRAPH_MUST_NOT_BE_NULL_EXCEPTION);
		}

		GraphView graphViewLeft = new GraphView(this);
		graphViewLeft.print(left);
		GraphView graphViewRight = new GraphView(this);		
		graphViewRight.print(right);

		JSplitPane graphPanel = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, graphViewLeft, graphViewRight);

		setContentPane(graphPanel);
		setExtendedState(JFrame.MAXIMIZED_BOTH);
		setTitle(title);
		setVisible(true);
		pack();
	}

	/**
	 * Shows the run of the automaton {@code automaton} on the cospan decomposition {@code decomposition}.
	 * @param automaton the automaton used for the run
	 * @param decomposition the cospan decomposition which is the input for this run
	 */
	public void showRun(Automaton automaton, final CospanDecomposition decomposition) {
		if(automaton==null) {
			throw new NullPointerException(RavenExceptionMessages.AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(decomposition==null) {
			throw new NullPointerException(RavenExceptionMessages.DECOMPOSITION_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		JPanel automatonInformationPanel = createAutomatonInformationPanel(automaton);
		automatonInformationPanel.setBorder(BorderFactory.createCompoundBorder(
				BorderFactory.createTitledBorder(" "+messages.getString(
						GuiMessages.AUTOMATON_PROPERTIES_LABEL_KEY)+": "),
				BorderFactory.createEmptyBorder(5, 5, 5, 5)));
		
		stateInformationView = new CospanView();
		stateInformationView.setBorder(BorderFactory.createCompoundBorder(
				BorderFactory.createTitledBorder(" "+messages.getString(
						GuiMessages.AUTOMATON_PROPERTIES_LABEL_KEY)+": "),
				BorderFactory.createEmptyBorder(5, 5, 5, 5)));		
		
		JPanel informationPanel = new JPanel(new GridLayout(1, 2));
		informationPanel.add(automatonInformationPanel);
		informationPanel.add(stateInformationView);

		AutomatonRunView automatonRunView = new AutomatonRunView();
		automatonRunView.getViewport().setOpaque(false);
		automatonRunView.addSelectionListener(new mxIEventListener() {
			
			@Override
			public void invoke(Object sender, mxEventObject e) {
				mxCell cell = (mxCell) ((mxGraphSelectionModel) sender).getCell();
				if(cell!=null) {
//					int value = Integer.parseInt(cell.getValue().toString());
					List<Operation> list = new LinkedList<>();
					
					mxCell current = cell;
					while(current!=null) {
						mxCell previous = current;
						current = null;
						
						for(int i=0;i<previous.getEdgeCount();i++) {
							mxCell edge = (mxCell) previous.getEdgeAt(i);
							
							if(edge.getTarget().equals(previous)) {
								String label = edge.getValue().toString();
								list.add(0, Operation.parse(label));
								current = (mxCell) edge.getSource();
								break;
							}
						}
					}
					
					Cospan cospan = new Cospan(new CospanDecomposition(list));
					stateInformationView.print(cospan);
				}
				else {
					stateInformationView.setGraph(new mxGraph());
				}
			}
		});
		automatonRunView.setBorder(BorderFactory.createCompoundBorder(
				BorderFactory.createTitledBorder(" "+messages.getString(
						GuiMessages.AUTOMATON_PROPERTIES_LABEL_KEY)+": "),
				BorderFactory.createEmptyBorder(5, 5, 5, 5)));
		automatonRunView.print(automaton, decomposition);

		JPanel mainPanel = new JPanel(new BorderLayout());
		mainPanel.add(informationPanel, BorderLayout.NORTH);
		mainPanel.add(automatonRunView, BorderLayout.CENTER);

		Dimension automatonPanelSize = automatonInformationPanel.getPreferredSize();
		Dimension statePanelSize = stateInformationView.getPreferredSize();
		Dimension runViewSize = automatonRunView.getPreferredSize();
		
		Dimension size = new Dimension(
				Math.max(automatonPanelSize.width+statePanelSize.width, runViewSize.width), 
				Math.max(automatonPanelSize.height, statePanelSize.height)+runViewSize.height);
		setPreferredSize(size);

		getContentPane().removeAll();
		getContentPane().add(mainPanel);
		setExtendedState(JFrame.MAXIMIZED_BOTH);
		setTitle(messages.getString(GuiMessages.AUTOMATON_RUN_DIALOG_TITLE_KEY, 
				repository.getName(automaton), repository.getName(decomposition)));
		pack();
		setLocationRelativeTo(null);
		setVisible(true);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates an information panel for the automaton {@code automaton} which is used to display
	 * all important information of the automaton, i.e.
	 * <ul>
	 * 	<li>the name of the automaton,</li>
	 * 	<li>the type of the automaton,</li>, 
	 *  <li>the number of states of the automaton,</li>
	 *  <li>the number of initial states of the automaton,</li>
	 *  <li>the number of final states of the automaton,</li>
	 *  <li>the maximum interface size of the automaton,</li>
	 *  <li>the inner interface size of the automaton, and</li>
	 *  <li>the outer interface size of the automaton.</li>
	 *</ul>
	 * @param automaton the automaton for which an information panel is build
	 * @return the automaton information panel
	 */
	private JPanel createAutomatonInformationPanel(Automaton automaton) {
		DefaultFormBuilder builder = new DefaultFormBuilder(
				new FormLayout("left:max(40dlu;pref), 3dlu, 40dlu:grow"));
		automatonNameField = new JTextField(Repository.getRepository().getName(automaton), 10);
		automatonNameField.setEditable(false);
		automatonNameField.setOpaque(false);
		builder.append(messages.getString(GuiMessages.NAME_LABEL_KEY), automatonNameField);

		typeField = new JTextField(messages.getString(
				GuiMessages.DATA_TYPE_PREFIX_KEY+automaton.getType().toString()), 10);
		typeField.setEditable(false);
		typeField.setOpaque(false);
		builder.append(messages.getString(GuiMessages.TYPE_LABEL_KEY), typeField);

		builder.appendSeparator();
		builder.nextLine();
		
		numberOfStatesField = new JTextField(String.valueOf(automaton.getNumberOfStates()), 10);
		numberOfStatesField.setEditable(false);
		numberOfStatesField.setOpaque(false);
		builder.append(messages.getString(GuiMessages.NUMBER_OF_STATES_LABEL_KEY), numberOfStatesField);

		numberOfInitialStatesField = new JTextField(String.valueOf(automaton.getNumberOfInitialStates()), 10);
		numberOfInitialStatesField.setEditable(false);
		numberOfInitialStatesField.setOpaque(false);
		builder.append(messages.getString(GuiMessages.NUMBER_OF_INITIAL_STATES_LABEL_KEY), numberOfInitialStatesField);

		numberOfFinalStatesField = new JTextField(String.valueOf(automaton.getNumberOfFinalStates()), 10);
		numberOfFinalStatesField.setEditable(false);
		numberOfFinalStatesField.setOpaque(false);
		builder.append(messages.getString(GuiMessages.NUMBER_OF_FINAL_STATES_LABEL_KEY), numberOfFinalStatesField);
		
		builder.appendSeparator();
		builder.nextLine();
		
		maximumInterfaceField = new JTextField(String.valueOf(automaton.getMaximumInterface()), 10);
		maximumInterfaceField.setEditable(false);
		maximumInterfaceField.setOpaque(false);
		builder.append(messages.getString(GuiMessages.MAXIMUM_INTERFACE_LABEL_KEY), maximumInterfaceField);

		innerInterfaceField = new JTextField(String.valueOf(automaton.getInnerInterface()), 10);
		innerInterfaceField.setEditable(false);
		innerInterfaceField.setOpaque(false);
		builder.append(messages.getString(GuiMessages.INNER_INTERFACE_SIZE_LABEL_KEY), innerInterfaceField);

		outerInterfaceField = new JTextField(String.valueOf(automaton.getOuterInterface()), 10);
		outerInterfaceField.setEditable(false);
		outerInterfaceField.setOpaque(false);
		builder.append(messages.getString(GuiMessages.OUTER_INTERFACE_SIZE_LABEL_KEY), outerInterfaceField);
		
		JPanel automatonInformationPanel = builder.getPanel();
		return automatonInformationPanel;
	}	

}
