/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.gui;

import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JLabel;

import com.jgoodies.forms.builder.DefaultFormBuilder;
import com.jgoodies.forms.factories.Borders;

import de.uni_due.inf.ti.raven.data.Data;

/**
 * This class provides an abstract builder which is used to to create a general 
 * purpose {@code InformationPanel}. This panel can be used for displaying 
 * information about different {@code data} objects.
 * @author Christoph Blume
 * @version $Revision: 1.8 $
 * @param <T> The type of display-able data
 *
 * @see InformationPanel
 * @see Data
 */
public abstract class InformationPanelBuilder<T extends Data> {

	/** Default Information Panel Layout Specification */
	protected  static final String LAYOUT_SPECIFICATION = "left:max(40dlu;pref), 3dlu, 40dlu:grow";

	private static final String LABEL_SUFFIX = ": ";
	private static final GuiMessages messages = GuiMessages.getMessages();

	private DefaultFormBuilder builder;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new information panel builder.
	 * @param builder the default form builder used for the creation of the panel
	 */
	protected InformationPanelBuilder(DefaultFormBuilder builder) {
		this.builder = builder;
		builder.border(Borders.DLU4);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/* 		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Adds a separator with the text specified by the label key.
     * @param key the separator title text
     * @return the {@code JLabel} containing the label text
	 */
	protected JLabel append(String key) {
		JLabel label = builder.append(messages.getString(key) + LABEL_SUFFIX);
		builder.nextLine();
		
		return label;
	}
	
	/**
	 * Adds a text label specified by the label key and a component to the panel.
     * @param key the label key
     * @param component the component to be added
	 * @return the {@code JLabel} containing the label text
	 */
	protected JLabel append(String key, JComponent component) {
		JLabel label = builder.append(messages.getString(key) + LABEL_SUFFIX, component);
		builder.nextLine();
		
		return label;
	}
	
	/**
	 * Adds a component to the panel that spans a specified number of columns
	 * @param component the component to be added
	 * @param columnSpan the size of the column span
	 */
	protected void append(JComponent component, int columnSpan) {
		builder.append(component, columnSpan);
		builder.nextLine();
	}
	
	/**
	 * Adds a button to the panel with the label specified by the label key.
     * @param key the label key
     * @param button the button to be added
	 */
	protected void appendButton(String key, JButton button) {
		button.setText(messages.getString(key));
		builder.nextColumn();
		builder.append(button);
		builder.nextLine();
	}
	
	/**
	 * Adds a separator to the panel.
	 */
	protected void appendSeparator() {
		builder.appendSeparator();
		builder.nextLine();
	}
	
}
