/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui;

import java.awt.Component;
import java.awt.FlowLayout;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;

import javax.swing.ButtonGroup;
import javax.swing.DefaultListCellRenderer;
import javax.swing.JComboBox;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JRadioButton;

import com.jgoodies.forms.builder.DefaultFormBuilder;

import de.uni_due.inf.ti.raven.actions.AntichainInvariantCheckingAction;
import de.uni_due.inf.ti.raven.actions.BisimulationUpToCongruenceInvariantCheckingAction;
import de.uni_due.inf.ti.raven.actions.ComputationEvent;
import de.uni_due.inf.ti.raven.actions.ComputationListener;
import de.uni_due.inf.ti.raven.actions.InvariantCheckingAction;
import de.uni_due.inf.ti.raven.actions.SimulationAntichainInvariantCheckingAction;
import de.uni_due.inf.ti.raven.actions.ComputationEvent.EventType;
import de.uni_due.inf.ti.raven.algorithms.Algorithm.SearchType;
import de.uni_due.inf.ti.raven.algorithms.DefaultCospanDecompositionAlgorithm;
import de.uni_due.inf.ti.raven.algorithms.InvariantCheckingAlgorithm;
import de.uni_due.inf.ti.raven.algorithms.InvariantCheckingAlgorithm.Implementations;
import de.uni_due.inf.ti.raven.data.Cospan;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.data.RuleCounterExample;
import de.uni_due.inf.ti.raven.data.automata.Automaton;

/**
 * This class implements a {@code ControlPanel} for the invocation of the different invariant 
 * checking algorithms implemented in Raven.
 * @author Christoph Blume, Dennis Nolte
 * @version $Revision: 1.11 $
 * 
 */
public final class InvariantCheckingControlPanel extends ControlPanel {

	private static final long serialVersionUID = -5383018170781400670L;

	/**
	 * This class is used as a worker thread to perform the invariant check. 
	 */
	private class InvariantCheckingWorkerThread extends WorkerThread {

		private String automatonName;
		private String leftHandSideName;
		private String rightHandSideName;
		private InvariantCheckingAction action;

		@Override
		protected Void doInBackground() throws Exception {			
			// Get the parameter for the invariant checking algorithm
			Implementations algorithm = 
					(Implementations) invariantCheckingAlgorithmBox.getSelectedItem();
			Automaton automaton = (Automaton) automatonBox.getSelectedData();
			
			CospanDecomposition leftHandSide = null;
			CospanDecomposition rightHandSide = null;
			
			if(leftHandSideBox.getSelectedData() instanceof CospanDecomposition) {
				leftHandSide = (CospanDecomposition) leftHandSideBox.getSelectedData();				
			}
			else{
				Cospan cospan = (Cospan) leftHandSideBox.getSelectedData();
				try {
					leftHandSide = new DefaultCospanDecompositionAlgorithm().decompose(cospan);
				}
				catch (Exception e) {
					e.printStackTrace();
				}
			}
			
			if(rightHandSideBox.getSelectedData() instanceof CospanDecomposition) {
				rightHandSide = (CospanDecomposition) rightHandSideBox.getSelectedData();				
			}
			else{
				Cospan cospan = (Cospan) rightHandSideBox.getSelectedData();
				try {
					rightHandSide = new DefaultCospanDecompositionAlgorithm().decompose(cospan);
				}
				catch (Exception e) {
					e.printStackTrace();
				}
			}

			// Store the names of the parameter (which will be used later on for the output)
			automatonName = automatonBox.getSelectedName();
			leftHandSideName = leftHandSideBox.getSelectedName();
			rightHandSideName = rightHandSideBox.getSelectedName();

			// Determine the search type to be used (either forward or backward)
			SearchType type;
			if (forwardButton.isSelected()) {
				type = SearchType.forward;
			} 
			else {
				type = SearchType.backward;
			}

			// Determine whether to compute a counter-example (if the result is negative)
			boolean computeCounterEx;
			if (computeCE.isSelected()) {
				computeCounterEx = true;
			}
			else {
				computeCounterEx = false;
			}

			if(automaton==null) {
				Utilities.showWarningMessage(InvariantCheckingControlPanel.this,
						GuiMessages.NO_AUTOMATON_SELECTED_TITLE_EXCEPTION_KEY,
						GuiMessages.NO_AUTOMATON_SELECTED_TEXT_EXCEPTION_KEY,
						messages.getString(GuiMessages.AUTOMATON_LABEL_KEY));
				cancel(true);
				fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_CANCELLED, "", 0, 0));
				return null;
			}
			else if(leftHandSide==null) {
				Utilities.showWarningMessage(InvariantCheckingControlPanel.this,
						GuiMessages.NO_DECOMPOSITION_SELECTED_TITLE_EXCEPTION_KEY,
						GuiMessages.NO_DECOMPOSITION_SELECTED_TEXT_EXCEPTION_KEY,
						messages.getString(GuiMessages.LEFT_HAND_DECOMPOSITION_LABEL_KEY));
				cancel(true);
				fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_CANCELLED, "", 0, 0));
				return null;
			}
			else if(rightHandSide==null) {
				Utilities.showWarningMessage(InvariantCheckingControlPanel.this,
						GuiMessages.NO_DECOMPOSITION_SELECTED_TITLE_EXCEPTION_KEY,
						GuiMessages.NO_DECOMPOSITION_SELECTED_TEXT_EXCEPTION_KEY,
						messages.getString(GuiMessages.RIGHT_HAND_DECOMPOSITION_LABEL_KEY));
				cancel(true);
				fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_CANCELLED, "", 0, 0));
				return null;
			}

			// Create invariant checking action
			action = null;					
			if (algorithm == Implementations.AntichainInvariantChecking) {
				action = new AntichainInvariantCheckingAction(type, computeCounterEx, 
						automaton, leftHandSide, rightHandSide);
			} else if (algorithm == Implementations.SimulationInvariantChecking) {
				action = new SimulationAntichainInvariantCheckingAction(type, computeCounterEx, 
						automaton, leftHandSide, rightHandSide);
			} else if (algorithm == Implementations.BisimulationUpToCongruence) {
				action = new BisimulationUpToCongruenceInvariantCheckingAction(type, computeCounterEx,
						automaton, leftHandSide, rightHandSide);
			}

			// Register listeners
			for(ComputationListener listener : listeners) {
				action.addComputationListener(listener);
			}

			try {
				// Start computation
				action.execute();
			}
			catch (Exception e) {
				e.printStackTrace();
				System.exit(0);
			}
			return null;
		}

		@Override
		protected void done() {
			//If the worker thread has been cancelled
			if(isCancelled()) {
				if(action!=null) {
					//also cancel the computation of the action
					action.cancelComputation();
				}
				
				action = null;
				automatonName = null;
				leftHandSideName = null;
				rightHandSideName = null;
			}
			//else bring up message dialog to display result
			else {
				boolean isInvariant = action.getResult();
				if(isInvariant) {
					Utilities.showInformationMessage(InvariantCheckingControlPanel.this,
							GuiMessages.IS_INVARIANT_DIALOG_TITLE_KEY,
							GuiMessages.IS_INVARIANT_DIALOG_TEXT_KEY,
							leftHandSideName, rightHandSideName, automatonName);
				}
				else {
					RuleCounterExample counterExample = action.getCounterExample();
					
					Utilities.showInformationMessage(InvariantCheckingControlPanel.this,
							GuiMessages.IS_NOT_INVARIANT_DIALOG_TITLE_KEY, 
							GuiMessages.IS_NOT_INVARIANT_DIALOG_TEXT_KEY, 
							leftHandSideName, rightHandSideName, automatonName);
					
					if (counterExample != null) {
						CounterExampleWindow.getWindow().showRuleCounterExampleChooser(
								counterExample.getSource(), counterExample.getTarget(), 
								messages.getString(GuiMessages.COUNTEREXAMPLE_LABEL_KEY));
					}
				}
			}
		}

	}

	/** Algorithm Box Label (used for Testing) */
	protected static final String ALGORITHM_BOX_NAME = "algorithmBox";

	/** Automaton Box Label (used for Testing) */
	protected static final String AUTOMATON_BOX_NAME = "automatonBox";

	/** Left Hand-Side Box Label (used for Testing) */
	protected static final String LEFT_HAND_SIDE_BOX_NAME = "leftHandSideBox";

	/** Right Hand-Side Box Label (used for Testing) */
	protected static final String RIGHT_HAND_SIDE_BOX_NAME = "rightHandSideBox";

	/** Forward Search Radio Button Label (used for Testing) */
	protected static final String FORWARD_SEARCH_RADIO_BUTTON_NAME = "forwardSearch";

	/** Backward Search Radio Button Label (used for Testing) */
	protected static final String BACKWARD_SEARCH_RADIO_BUTTON_NAME = "backwardSearch";

	/** Compute Counter-Example Radio Button Label (used for Testing) */
	protected static final String COMPUTE_COUNTER_EXAMPLE_RADIO_BUTTON_NAME = "computeCounterExampleBox";

	/** Do Not Compute Counter-Example Radio Button Label (used for Testing) */
	protected static final String DO_NOT_COMPUTE_COUNTER_EXAMPLE_RADIO_BUTTON_NAME = "doNotComputeCounterExampleBox";

	private JComboBox<InvariantCheckingAlgorithm.Implementations> invariantCheckingAlgorithmBox;
	private DataComboBox automatonBox;
	private DataComboBox leftHandSideBox;
	private DataComboBox rightHandSideBox;
	private JRadioButton forwardButton;
	private JRadioButton backwardButton;
	private JRadioButton computeCE;
	private JRadioButton doNotComputeCE;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public void setEnabled(boolean isEnabled) {
		invariantCheckingAlgorithmBox.setEnabled(isEnabled);
		automatonBox.setEnabled(isEnabled);
		leftHandSideBox.setEnabled(isEnabled);
		rightHandSideBox.setEnabled(isEnabled);
		forwardButton.setEnabled(isEnabled);
		backwardButton.setEnabled(isEnabled);
		computeCE.setEnabled(isEnabled);
		doNotComputeCE.setEnabled(isEnabled);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	protected void initialize(DefaultFormBuilder builder) {
		Repository repo = Repository.getRepository();

		invariantCheckingAlgorithmBox = new JComboBox<>(Implementations.values());
		invariantCheckingAlgorithmBox.setName(ALGORITHM_BOX_NAME);
		invariantCheckingAlgorithmBox.setRenderer(new DefaultListCellRenderer() {

			private static final long serialVersionUID = 48576265719193199L;

			@Override
			public Component getListCellRendererComponent(
					JList<?> list, Object value, int index,
					boolean isSelected, boolean cellHasFocus) {
				return super.getListCellRendererComponent(
						list, messages.getString(GuiMessages.LABEL_PREFIX + value.toString()), 
						index, isSelected, cellHasFocus);
			}
		});
		invariantCheckingAlgorithmBox.addItemListener(new ItemListener() {
			
			@Override
			public void itemStateChanged(ItemEvent e) {
				boolean setEnabled = e.getItem().equals(
						Implementations.AntichainInvariantChecking);
				
				forwardButton.setEnabled(setEnabled);
				backwardButton.setEnabled(setEnabled);
			}
		});

		append(builder, GuiMessages.ALGORITHM_LABEL_KEY, invariantCheckingAlgorithmBox);

		automatonBox = new DataComboBox(repo);
		automatonBox.setName(AUTOMATON_BOX_NAME);
		automatonBox.addDataClass(Automaton.class);
		append(builder, GuiMessages.AUTOMATON_LABEL_KEY, automatonBox); 

		leftHandSideBox = new DataComboBox(repo);
		leftHandSideBox.setName(LEFT_HAND_SIDE_BOX_NAME);
		leftHandSideBox.addDataClass(CospanDecomposition.class);
		leftHandSideBox.addDataClass(Cospan.class);
		append(builder, GuiMessages.LEFT_HAND_DECOMPOSITION_LABEL_KEY, leftHandSideBox);

		rightHandSideBox = new DataComboBox(repo);
		rightHandSideBox.setName(RIGHT_HAND_SIDE_BOX_NAME);
		rightHandSideBox.addDataClass(CospanDecomposition.class);
		rightHandSideBox.addDataClass(Cospan.class);
		append(builder, GuiMessages.RIGHT_HAND_DECOMPOSITION_LABEL_KEY, rightHandSideBox);

		forwardButton = new JRadioButton(messages.getString(GuiMessages.FORWARD_LABEL_KEY));
		forwardButton.setName(FORWARD_SEARCH_RADIO_BUTTON_NAME);
		forwardButton.setOpaque(false);
		forwardButton.setSelected(true);

		backwardButton = new JRadioButton(messages.getString(GuiMessages.BACKWARD_LABEL_KEY));
		backwardButton.setName(BACKWARD_SEARCH_RADIO_BUTTON_NAME);
		backwardButton.setOpaque(false);

		ButtonGroup searchTypeGroup = new ButtonGroup();
		searchTypeGroup.add(forwardButton);
		searchTypeGroup.add(backwardButton);

		JPanel searchTypePanel = new JPanel(new FlowLayout(FlowLayout.LEFT, 0, 0));
		searchTypePanel.setOpaque(false);
		searchTypePanel.add(forwardButton);
		searchTypePanel.add(backwardButton);
		append(builder, GuiMessages.SEARCHTYPE_LABEL_KEY, searchTypePanel);

		computeCE = new JRadioButton(messages.getString(GuiMessages.YES_LABEL_KEY));
		computeCE.setName(COMPUTE_COUNTER_EXAMPLE_RADIO_BUTTON_NAME);
		computeCE.setOpaque(false);
		computeCE.setSelected(true);

		doNotComputeCE = new JRadioButton(messages.getString(GuiMessages.NO_LABEL_KEY));
		doNotComputeCE.setName(DO_NOT_COMPUTE_COUNTER_EXAMPLE_RADIO_BUTTON_NAME);
		doNotComputeCE.setOpaque(false);

		ButtonGroup computeCounterExample = new ButtonGroup();
		computeCounterExample.add(computeCE);
		computeCounterExample.add(doNotComputeCE);

		JPanel counterExamplePanel = new JPanel(new FlowLayout(FlowLayout.LEFT, 0, 0));
		counterExamplePanel.setOpaque(false);
		counterExamplePanel.add(computeCE);
		counterExamplePanel.add(doNotComputeCE);
		append(builder, GuiMessages.COMPUTE_COUNTER_EXAMPLE_LABEL_KEY, counterExamplePanel);

		// set Color
		invariantCheckingAlgorithmBox.setBackground(getBackground());
		automatonBox.setBackground(getBackground());
		leftHandSideBox.setBackground(getBackground());
		rightHandSideBox.setBackground(getBackground());
		forwardButton.setBackground(getBackground());
		backwardButton.setBackground(getBackground());
		computeCE.setBackground(getBackground());
		doNotComputeCE.setBackground(getBackground());
	}

	@Override
	protected WorkerThread getWorkerThread() {
		return new InvariantCheckingWorkerThread();
	}

}
