/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.gui;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Window;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;

import javax.swing.Action;
import javax.swing.BoxLayout;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.border.EmptyBorder;
import javax.swing.filechooser.FileFilter;
import javax.swing.plaf.basic.BasicFileChooserUI;

import de.uni_due.inf.ti.raven.actions.LoadAction;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * This class is the abstract base class for all {@code Data} load dialogs. It provides a simple
 * mechanism for the user to choose to {@code Data} file to be loaded.
 * @author Christoph Blume, Dennis Nolte, Laura Steinert
 * @version $Revision: 1.14 $
 *
 * @see LoadAction
 */
public abstract class LoadDataDialog extends RavenDialog {
	
	private static final long serialVersionUID = -2316808700155215140L;
	
	/** GUI Messages Resource Bundle */
	protected static final GuiMessages messages = GuiMessages.getMessages();

	private JLabel dataNameLabel;
	private JTextField dataNameField;
	private JFileChooser chooser;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a file filter with a description given by the {@code descriptionKey} based on the 
	 * file extension {@code fileExtension}.
	 * @param descriptionKey the GuiMessage key used for the description
	 * @param fileExtension the file extension used for the filer
	 * @return the created file filter
	 * 
	 * @see GuiMessages
	 */
	protected static FileFilter createFileFilter(final String descriptionKey, 
			final String fileExtension) {
		return new FileFilter() {

			@Override
			public boolean accept(File f) {
				return f.isDirectory() || f.getName().endsWith("."+fileExtension);
			}

			@Override
			public String getDescription() {
				return messages.getString(descriptionKey) + " (*." + fileExtension + ")";
			}
			
		};
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new dialog to load data into the repository.
	 * @param owner the parent window of this dialog
	 * @param title the title of the dialog
	 * @param defaultDirectory the default directory which is used for the file system view
	 */
	protected LoadDataDialog(Window owner, String title, File defaultDirectory) {
		super(owner, title, RavenDialog.LOAD_CANCEL_LAYOUT);

		getContentPane().add(createMainPanel(defaultDirectory));
		setLocationByPlatform(true);
		pack();
		
		setAction(RavenDialog.LOAD_BUTTON, getLoadAction());
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Abstract Protected Methods															*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the {@code Action} which will be performed whenever the user clicks the load button.
	 * @return the load action
	 */
	protected abstract Action getLoadAction();
	
	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the currently selected file
	 * @return the currently selected file
	 */
	protected File getSelectedFile() {
		BasicFileChooserUI ui = (BasicFileChooserUI) chooser.getUI();
		return new File(chooser.getCurrentDirectory(), ui.getFileName());
	}
	
	/**
	 * Returns the name for the data object entered by the user
	 * @return the name for the data object
	 */
	protected String getDataName() {
		return dataNameField.getText();
	}
	
	/**
	 * Sets the label for the data name textfield
	 * @param text the label of the textfield
	 */	
	protected void setDataNameLabel(String text) {
		dataNameLabel.setText(text+":");
		Dimension d = dataNameLabel.getPreferredSize();
		dataNameLabel.setPreferredSize(new Dimension(d.width + 5, d.height));
	}
	
	/**
	 * Sets the file filter which can be used by this dialog.
	 * @param fileFilters a list of available file filters
	 */
	protected void setFileFilters(FileFilter... fileFilters) {
		chooser.resetChoosableFileFilters();
		for(FileFilter filter: fileFilters) {
			chooser.addChoosableFileFilter(filter);
		}
	}
	
	/*------------------------------------------------------------------------------------------*/
	/* 		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the panel which displays the selection of an automaton from the repository.
	 * @param defaultDirectory the default directory which is used for the file system view
	 * @return the main panel
	 */
	private JPanel createMainPanel(File defaultDirectory) {
		dataNameLabel = new JLabel();

		dataNameField = new JTextField(10);
		dataNameLabel.setLabelFor(dataNameField);

		JPanel dataNamePanel = new JPanel();
		dataNamePanel.setLayout(new BoxLayout(dataNamePanel, BoxLayout.LINE_AXIS));
		dataNamePanel.setBorder(new EmptyBorder(12, 12, 0, 11));
		dataNamePanel.add(dataNameLabel);
		dataNamePanel.add(dataNameField);

		RavenDirectoryFileSystemView view = new RavenDirectoryFileSystemView(
				new File(RavenMain.SAVE_DIR), defaultDirectory);
		chooser = new JFileChooser(view);
		chooser.setCurrentDirectory(view.getDefaultDirectory());
		
		chooser.setControlButtonsAreShown(false);
		chooser.setAcceptAllFileFilterUsed(false);
		
		chooser.addPropertyChangeListener(JFileChooser.SELECTED_FILE_CHANGED_PROPERTY, 
				new PropertyChangeListener() {
			
			@Override
			public void propertyChange(PropertyChangeEvent e) {				
				JFileChooser chooser = (JFileChooser) e.getSource();

				if(chooser.getSelectedFile()!=null) {
					String filename = chooser.getSelectedFile().getName();				
					dataNameField.setText(filename.substring(0, filename.lastIndexOf(".")));
				}
			}
		});
		
		JPanel mainPanel = new JPanel(new BorderLayout());
		mainPanel.add(dataNamePanel, BorderLayout.NORTH);
		mainPanel.add(chooser, BorderLayout.CENTER);

		return mainPanel;
	}
	
}
