/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui;

import java.awt.Window;
import java.awt.event.ActionEvent;
import java.io.File;
import java.io.FileNotFoundException;

import javax.swing.AbstractAction;
import javax.swing.Action;

import de.uni_due.inf.ti.raven.actions.LoadAction;
import de.uni_due.inf.ti.raven.actions.LoadAction.LoadType;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.io.GXLReader;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * This class provides a simple mechanism for the user to load a graph file.
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.24 $
 *
 * @see Graph
 * @see GXLReader
 *
 */
public class LoadGraphDialog extends LoadDataDialog {

	private static final long serialVersionUID = -7472771725622723691L;

	private static LoadGraphDialog dialog;

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Brings up a dialog to load graphs into the repository.
	 * @param owner the parent window of this dialog
	 */
	public static void showDialog(Window owner) {
		dialog = new LoadGraphDialog(owner);
		dialog.setLocationRelativeTo(owner);
		dialog.setVisible(true);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new dialog to select a graph file which will be loaded
	 * @param owner the parent window of this dialog
	 */
	private LoadGraphDialog(Window owner) {
		super(owner, messages.getString(GuiMessages.LOAD_GRAPH_DIALOG_TITLE_KEY),
				new File(RavenMain.GRAPH_SAVE_DIR));

		setDataNameLabel(messages.getString(GuiMessages.GRAPH_NAME_LABEL_KEY));
		setFileFilters(RavenFileFilter.GRAPH_GXL_FILE_FILTER);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	protected Action getLoadAction() {
		return new AbstractAction(messages.getString(GuiMessages.LOAD_GRAPH_KEY)) {

			private static final long serialVersionUID = 5792355974515770837L;

			@Override
			public void actionPerformed(ActionEvent e) {
				String filename = getSelectedFile().getAbsolutePath();
				filename = filename.trim();

				String graphName = getDataName();
				graphName = graphName.trim();
				if(!Repository.isNameValid(graphName)){
					showErrorMessage(GuiMessages.INVALID_NAME_TITLE_EXCEPTION_KEY,
							GuiMessages.INVALID_NAME_TEXT_EXCEPTION_KEY, graphName);
					return;
				}

				//If either the file name is empty, return
				if (filename.isEmpty()) {
					showErrorMessage(GuiMessages.EMPTY_INPUTFIELD_TITLE_EXCEPTION_KEY,
							GuiMessages.EMPTY_INPUTFIELD_TEXT_EXCEPTION_KEY,
							messages.getString(GuiMessages.NAME_LABEL_KEY));
					return;
				}
				
				if(graphName.isEmpty()){
					graphName = getSelectedFile().getName();
					if(graphName.endsWith("."+RavenMain.GRAPH_GXL_FILE_EXTENSION)){
						graphName = graphName.substring(0, 
								graphName.length()-(RavenMain.GRAPH_GXL_FILE_EXTENSION.length()+1));
					}
				}

				//If the name is already in use show a message to the user
				if (Repository.getRepository().contains(graphName)) {
					showErrorMessage(GuiMessages.NAME_ALREADY_IN_USE_TITLE_EXCEPTION_KEY,
							GuiMessages.NAME_ALREADY_IN_USE_TEXT_EXCEPTION_KEY, graphName);
					return;
				}

				//Check whether the specified file can be found and is a valid graph file
				if (filename.endsWith("."+RavenMain.GRAPH_GXL_FILE_EXTENSION)) {
					try {
						new LoadAction(LoadType.GRAPHGXL, graphName, filename).execute();
					} catch (FileNotFoundException ex) {
						showErrorMessage(GuiMessages.FILE_NOT_FOUND_TITLE_EXCEPTION_KEY,
								GuiMessages.FILE_NOT_FOUND_TEXT_EXCEPTION_KEY, filename);
					}
					finally {
						dialog.dispose();
					}
				} else {
					showErrorMessage(GuiMessages.NO_GRAPH_FILE_SELECTED_TITLE_EXCEPTION_KEY,
							GuiMessages.NO_GRAPH_FILE_SELECTED_TEXT_EXCEPTION_KEY, filename);
				}
			}

		};
	}

}
