/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui;

import java.awt.AlphaComposite;
import java.awt.BorderLayout;
import java.awt.CardLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Rectangle;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.logging.Logger;

import javax.help.HelpBroker;
import javax.help.HelpSet;
import javax.imageio.ImageIO;
import javax.swing.BorderFactory;
import javax.swing.ImageIcon;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTable;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.JToolBar;
import javax.swing.ListSelectionModel;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.table.DefaultTableModel;

import de.uni_due.inf.ti.raven.cli.Console;
import de.uni_due.inf.ti.raven.data.Cospan;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Data;
import de.uni_due.inf.ti.raven.data.Decomposition;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.PathDecomposition;
import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.data.RepositoryListener;
import de.uni_due.inf.ti.raven.data.Signature;
import de.uni_due.inf.ti.raven.data.TreeDecomposition;
import de.uni_due.inf.ti.raven.data.automata.ColorabilityAutomaton;
import de.uni_due.inf.ti.raven.data.automata.DominatingSetAutomaton;
import de.uni_due.inf.ti.raven.data.automata.EdgeCountingAutomaton;
import de.uni_due.inf.ti.raven.data.automata.LinkAutomaton;
import de.uni_due.inf.ti.raven.data.automata.MaximumEdgeAutomaton;
import de.uni_due.inf.ti.raven.data.automata.MaximumVertexAutomaton;
import de.uni_due.inf.ti.raven.data.automata.MinimumEdgeAutomaton;
import de.uni_due.inf.ti.raven.data.automata.MinimumVertexAutomaton;
import de.uni_due.inf.ti.raven.data.automata.NoIsolatedNodesAutomaton;
import de.uni_due.inf.ti.raven.data.automata.PathAutomaton;
import de.uni_due.inf.ti.raven.data.automata.ProductAutomaton;
import de.uni_due.inf.ti.raven.data.automata.SubgraphAutomaton;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.data.automata.UnionAutomaton;
import de.uni_due.inf.ti.raven.data.automata.VertexCountingAutomaton;
import de.uni_due.inf.ti.raven.data.automata.VertexCoverAutomaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.gui.actions.AddToRepositoryAction;
import de.uni_due.inf.ti.raven.gui.actions.CreateAutomatonAction;
import de.uni_due.inf.ti.raven.gui.actions.CreateCospanAction;
import de.uni_due.inf.ti.raven.gui.actions.CreateDecompositionAction;
import de.uni_due.inf.ti.raven.gui.actions.CreateGraphAction;
import de.uni_due.inf.ti.raven.gui.actions.CreateSignatureAction;
import de.uni_due.inf.ti.raven.gui.actions.HelpAction;
import de.uni_due.inf.ti.raven.gui.actions.LoadAutomatonAction;
import de.uni_due.inf.ti.raven.gui.actions.LoadButtonAction;
import de.uni_due.inf.ti.raven.gui.actions.LoadCospanAction;
import de.uni_due.inf.ti.raven.gui.actions.LoadDecompositionAction;
import de.uni_due.inf.ti.raven.gui.actions.LoadGraphAction;
import de.uni_due.inf.ti.raven.gui.actions.LoadSignatureAction;
import de.uni_due.inf.ti.raven.gui.actions.QuitAction;
import de.uni_due.inf.ti.raven.gui.actions.RemoveFromRepositoryAction;
import de.uni_due.inf.ti.raven.gui.actions.RunScriptAction;
import de.uni_due.inf.ti.raven.gui.actions.SaveAutomatonAction;
import de.uni_due.inf.ti.raven.gui.actions.SaveButtonAction;
import de.uni_due.inf.ti.raven.gui.actions.SaveCospanAction;
import de.uni_due.inf.ti.raven.gui.actions.SaveDecompositionAction;
import de.uni_due.inf.ti.raven.gui.actions.SaveGraphAction;
import de.uni_due.inf.ti.raven.gui.actions.SaveSignatureAction;
import de.uni_due.inf.ti.raven.gui.actions.UpdateAction;
import de.uni_due.inf.ti.raven.gui.actions.VariablesAndPropertiesAction;
import de.uni_due.inf.ti.raven.gui.actions.VersionAction;
import de.uni_due.inf.ti.raven.gui.automaton.ColorabilityAutomatonInformationPanelBuilder;
import de.uni_due.inf.ti.raven.gui.automaton.DominatingSetAutomatonInformationPanelBuilder;
import de.uni_due.inf.ti.raven.gui.automaton.EdgeCountingAutomatonInformationPanelBuilder;
import de.uni_due.inf.ti.raven.gui.automaton.LinkAutomatonInformationPanelBuilder;
import de.uni_due.inf.ti.raven.gui.automaton.MaximumEdgeAutomatonInformationPanelBuilder;
import de.uni_due.inf.ti.raven.gui.automaton.MaximumVertexAutomatonInformationPanelBuilder;
import de.uni_due.inf.ti.raven.gui.automaton.MinimumEdgeAutomatonInformationPanelBuilder;
import de.uni_due.inf.ti.raven.gui.automaton.MinimumVertexAutomatonInformationPanelBuilder;
import de.uni_due.inf.ti.raven.gui.automaton.NoIsolatedNodesAutomatonInformationPanelBuilder;
import de.uni_due.inf.ti.raven.gui.automaton.PathAutomatonInformationPanelBuilder;
import de.uni_due.inf.ti.raven.gui.automaton.ProductAutomatonInformationPanelBuilder;
import de.uni_due.inf.ti.raven.gui.automaton.SubgraphAutomatonInformationPanelBuilder;
import de.uni_due.inf.ti.raven.gui.automaton.UnionAutomatonInformationPanelBuilder;
import de.uni_due.inf.ti.raven.gui.automaton.VertexCountingAutomatonInformationPanelBuilder;
import de.uni_due.inf.ti.raven.gui.automaton.VertexCoverAutomatonInformationPanelBuilder;
import de.uni_due.inf.ti.raven.io.RavenPrinter;
import de.uni_due.inf.ti.raven.io.TextAreaPrinter;
import de.uni_due.inf.ti.raven.io.TextfieldInputStream;
import de.uni_due.inf.ti.raven.main.RavenMain;
import de.uni_due.inf.ti.raven.main.Version;

/**
 * This class provides the main window of Raven. But many of the components of 
 * this window are implemented by other GUI classes.
 * 
 * @author Christoph Blume, Laura Steinert, Dennis Nolte
 * @version $Revision: 1.116 $
 * 
 * @see ControlPanel
 * @see InformationPanel
 * @see InformationPanelBuilder
 * @see TechniquePanel
 * 
 */
public class MainWindow extends JFrame implements RepositoryListener {
	
	private static final long serialVersionUID = 2283465836582992916L;

	/** The default background color for all GUI components */
	public static final Color BACKGROUND_COLOR = Color.WHITE;

	private static final String AUTOMATON_CARD_LABEL = "automatonCard";
	private static final String BDD_CARD_LABEL = "bddCard";
	private static final String COLORABILITY_AUTOMATON_CARD_LABEL = "colorabilityAutomatonCard";
	private static final String COSPAN_CARD_LABEL = "cospanCard";
	private static final String DECOMPOSITION_CARD_LABEL = "decompositionCard";
	private static final String DOMINATING_SET_AUTOMATON_CARD_LABEL = "dominatingSetAutomatonCard";
	private static final String EDGE_COUNTING_AUTOMATON_CARD_LABEL = "edgeCountingAutomatonCard";
	private static final String EMPTY_CARD_LABEL = "emptyCard";
	private static final String GRAPH_CARD_LABEL = "graphCard";
	private static final String LINK_AUTOMATON_CARD_LABEL = "linkAutomatonCard";
	private static final String MAXIMUM_EDGE_AUTOMATON_CARD_LABEL = "maximumEdgeAutomatonCard";
	private static final String MAXIMUM_VERTEX_AUTOMATON_CARD_LABEL = "maximumVertexAutomatonCard";
	private static final String MINIMUM_EDGE_AUTOMATON_CARD_LABEL = "minimumEdgeAutomatonCard";
	private static final String MINIMUM_VERTEX_AUTOMATON_CARD_LABEL = "minimumVertexAutomatonCard";
	private static final String NO_ISOLATED_NODES_AUTOMATON_CARD_LABEL = "noIsolatedNodesAutomatonCard";
	private static final String PATH_AUTOMATON_CARD_LABEL = "pathAutomatonCard";
	private static final String PATH_DECOMPOSITION_CARD_LABEL = "pathdecompositionCard";
	private static final String PRODUCT_AUTOMATON_CARD_LABEL = "productAutomatonCard";
	private static final String SIGNATURE_CARD_LABEL = "signatureCard";
	private static final String SUBGRAPH_AUTOMATON_CARD_LABEL = "subgraphAutomatonCard";
	private static final String TREE_DECOMPOSITION_CARD_LABEL = "treedecompositionCard";
	private static final String UNION_AUTOMATON_CARD_LABEL = "unionAutomatonCard";
	private static final String VERTEX_COUNTING_AUTOMATON_CARD_LABEL = "vertexCountingAutomatonCard";
	private static final String VERTEX_COVER_AUTOMATON_CARD_LABEL = "vertexCoverAutomatonCard";

	private static final String WINDOW_TITLE = "Raven - Ver. " + Version.getVersion();

	private static MainWindow window;
	private static RavenPrinter out;
	private static final GuiMessages messages = GuiMessages.getMessages();
	private static final Repository repo = Repository.getRepository();

	private JTextArea textarea;
	private JTextField consoleInput;
	private JPanel infoPanel;

	private DefaultTableModel repositoryTableModel;

	private JPanel automatonInformationPanel;
	private InformationPanel<ColorabilityAutomaton> colorabilityInformationPanel;
	private InformationPanel<DominatingSetAutomaton> dominatingSetInformationPanel;
	private InformationPanel<EdgeCountingAutomaton> edgeCountingInformationPanel;
	private InformationPanel<LinkAutomaton> linkInformationPanel;
	private InformationPanel<MaximumEdgeAutomaton> maximumEdgeInformationPanel;
	private InformationPanel<MaximumVertexAutomaton> maximumVertexInformationPanel;
	private InformationPanel<MinimumEdgeAutomaton> minimumEdgeInformationPanel;
	private InformationPanel<MinimumVertexAutomaton> minimumVertexInformationPanel;
	private InformationPanel<NoIsolatedNodesAutomaton> noIsolatedNodesInformationPanel;
	private InformationPanel<PathAutomaton> pathInformationPanel;
	private InformationPanel<ProductAutomaton> productInformationPanel;
	private InformationPanel<SubgraphAutomaton> subgraphInformationPanel;
	private InformationPanel<UnionAutomaton> unionInformationPanel;
	private InformationPanel<VertexCountingAutomaton> vertexCountingInformationPanel;
	private InformationPanel<VertexCoverAutomaton> vertexCoverInformationPanel;
	private InformationPanel<CospanDecomposition> decompositionInformationPanel;
	private InformationPanel<Cospan> cospanInformationPanel;
	private InformationPanel<TreeDecomposition> treeDecompositionInformationPanel;
	private InformationPanel<PathDecomposition> pathDecompositionInformationPanel;
	private InformationPanel<Graph> graphInformationPanel;
	private InformationPanel<Signature> signatureInformationPanel;
	private InformationPanel<RavenBDD> bddInformationPanel;

	private static HelpSet currenthelpSet = null;
	
	private final Logger logger;

	/*------------------------------------------------------------------------------------------*/
	/* 		Static Methods 																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns an instance of the main window.
	 * 
	 * @return the main window
	 */
	public static MainWindow getWindow() {
		if (window == null) {
			window = new MainWindow();

			repo.addRepositoryListener(window);
			Path dotPath = Paths.get(RavenMain.getProperty(RavenMain.PROPERTY_DOT_PATH));
			if(Files.notExists(dotPath) || !Files.isRegularFile(dotPath) || 
					!Files.isExecutable(dotPath)) {
				Utilities.showWarningMessage(window, 
						GuiMessages.DOT_PATH_NOT_FOUND_TITLE_EXCEPTION_KEY,
						GuiMessages.DOT_PATH_NOT_FOUND_TEXT_EXCEPTION_KEY);
			}
		}

		return window;
	}

	/**
	 * Returns a {@code HelpBroker} object containing help information for the
	 * usage of Raven.
	 * 
	 * @return {@code HelpBroker} object
	 */
	public static HelpBroker getHelpSystem() {
		HelpBroker helpBroker = null;

		if (currenthelpSet == null) {
			// Find the HelpSet file and create the HelpSet object:
			ClassLoader cl = MainWindow.class.getClassLoader();
			try {
				URL hsURL = HelpSet.findHelpSet(cl, "jhelpset.hs");
				currenthelpSet = new HelpSet(cl, hsURL);
				helpBroker = currenthelpSet.createHelpBroker();
			} catch (Exception e) {
				e.printStackTrace();
				currenthelpSet = null;
			}
		} else {
			helpBroker = currenthelpSet.createHelpBroker();
		}

		return helpBroker;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new Raven Main Window.
	 */
	private MainWindow() {
		super(WINDOW_TITLE);
		logger = Logger.getLogger(getClass().getName());

		BufferedImage img = null;
		try {
			img = ImageIO.read(new File(RavenMain.RAVEN_ICON_FILENAME));
			super.setIconImage(img);
		} catch (IOException e) {
			// Do nothing
		}

		addWindowListener(new WindowAdapter() {

			@Override
			public void windowClosing(WindowEvent e) {
				dispose();
			}

			@Override
			public void windowClosed(WindowEvent e) {
				RavenMain.exit(RavenMain.EXIT_CODE_NORMAL_TERMINATION);
			}
		});

		File imageFile = new File(RavenMain.BIG_RAVEN_ICON_FILENAME); 
		if(!imageFile.exists()) {
			textarea = new JTextArea();
		}
		else {
			final ImageIcon icon = new ImageIcon(imageFile.getAbsolutePath());
			// Create output text area and install default printer
			textarea = new JTextArea() {
	
				private static final long serialVersionUID = 7066386582585853861L;
				
				private final Image backgroundImage;
	
				{
					backgroundImage = new BufferedImage(icon.getIconWidth(), 
							icon.getIconHeight(), BufferedImage.TYPE_INT_RGB);
					Graphics2D g2d = (Graphics2D) backgroundImage.getGraphics();
					g2d.setBackground(Color.WHITE);
					g2d.clearRect(0, 0, icon.getIconWidth(), 
							icon.getIconHeight());
					AlphaComposite alpha = AlphaComposite.getInstance(
							AlphaComposite.SRC_OVER, 0.3f);
					g2d.setComposite(alpha);
					g2d.drawImage(icon.getImage(), 0, 0, null);
				}
	
				@Override
				public void paintComponent(Graphics g) {
					Rectangle rect = getVisibleRect();
					
					int x = rect.x + (rect.width-backgroundImage.getWidth(this))/2;
					int y = rect.y + (rect.height-backgroundImage.getHeight(this))/2;
					g.drawImage(backgroundImage, x, y, this);
					setOpaque(false);
	
					super.paintComponent(g);
				}
			};
		}
		textarea.setEditable(false);
		 
		out = TextAreaPrinter.getPrinter(textarea);
		consoleInput = new JTextField(20);
		try (TextfieldInputStream tis = new TextfieldInputStream(consoleInput)) {
			Console console = Console.getConsole(tis, out, false);
			new Thread(console).start();
		}
		catch (IOException e) {
			logger.throwing(getClass().getName(), "MainWindow", e);
		}

		// Create content pane elements
		JSplitPane splitpane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);
		splitpane.setBackground(BACKGROUND_COLOR);
		splitpane.setLeftComponent(createMainPanel());
		splitpane.setRightComponent(createRepositoryPanel());
		splitpane.setBorder(BorderFactory.createRaisedBevelBorder());
		splitpane.setResizeWeight(0.9);
		setContentPane(splitpane);
		// Install the text area as default output stream

		// Create main menu
		setJMenuBar(createMainMenuBar());
		setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
		
		// Set up window properties
		setMinimumSize(new Dimension(800, 800));
		setExtendedState(JFrame.MAXIMIZED_BOTH);
		setLocationRelativeTo(null);
		pack();
		setVisible(true);
	}

	/*------------------------------------------------------------------------------------------*/
	/* Public Methods																			*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public void dataAdded(String name, Data data) {
		String dataType = messages.getString(
				GuiMessages.DATA_TYPE_PREFIX_KEY + data.getType().toString());
		String text = null;

		if (data instanceof Automaton) {
			Automaton automaton = (Automaton) data;
			String maxInterface = Integer.toString(automaton.getMaximumInterface());
			text = messages.getString(GuiMessages.AUTOMATON_ADDED_MESSAGE_KEY, name, maxInterface);
		} 
		else if (data instanceof Cospan) {
			text = messages.getString(GuiMessages.COSPAN_ADDED_MESSAGE_KEY, name);
		}
		else if (data instanceof Decomposition<?>) {
			text = messages.getString(GuiMessages.DECOMPOSITION_ADDED_MESSAGE_KEY, name);
		}
		else if (data instanceof Graph) {
			text = messages.getString(GuiMessages.GRAPH_ADDED_MESSAGE_KEY, name);
		}
		else if (data instanceof Signature) {
			text = messages.getString(GuiMessages.SIGNATURE_ADDED_MESSAGE_KEY, name);
		}
		else {
			text = messages.getString(GuiMessages.UNKNOWN_DATA_ADDED_MESSAGE_KEY, name);
		}

		repositoryTableModel.addRow(new Object[] { name, dataType });

		out.println(text);
	}

	@Override
	public void dataRemoved(String name, Data data) {
		for (int i = 0; i < repositoryTableModel.getRowCount(); i++) {
			if (repositoryTableModel.getValueAt(i, 0).equals(name)) {
				repositoryTableModel.removeRow(i);
			}
		}
	}
	
	@Override
	public void dataRenamed(String name, String replacement) {
		repositoryTableModel.fireTableDataChanged();
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates the main menu bar.
	 * 
	 * @return the main menu bar
	 */
	private JMenuBar createMainMenuBar() {
		JMenuBar mainMenuBar = new JMenuBar();

		// Create File menu
		JMenu fileMenu = new JMenu(messages.getString(GuiMessages.FILE_MENU_KEY));
		JMenu loadMenu = new JMenu(messages.getString(GuiMessages.LOAD_KEY));
		loadMenu.add(new JMenuItem(new LoadAutomatonAction(this)));
		loadMenu.add(new JMenuItem(new LoadCospanAction(this)));
		loadMenu.add(new JMenuItem(new LoadDecompositionAction(this)));
		loadMenu.add(new JMenuItem(new LoadGraphAction(this)));
		loadMenu.add(new JMenuItem(new LoadSignatureAction(this)));
		fileMenu.add(loadMenu);
		JMenu saveMenu = new JMenu(messages.getString(GuiMessages.SAVE_KEY));
		saveMenu.add(new JMenuItem(new SaveAutomatonAction(this)));
		saveMenu.add(new JMenuItem(new SaveCospanAction(this)));		
		saveMenu.add(new JMenuItem(new SaveDecompositionAction(this)));
		saveMenu.add(new JMenuItem(new SaveGraphAction(this)));
		saveMenu.add(new JMenuItem(new SaveSignatureAction(this)));
		fileMenu.add(saveMenu);
		fileMenu.addSeparator();
		fileMenu.add(new RunScriptAction(this, out));
		fileMenu.addSeparator();
		fileMenu.add(new QuitAction());
		mainMenuBar.add(fileMenu);

		// Create Repository menu
		JMenu repositoryMenu = new JMenu(messages.getString(GuiMessages.REPOSITORY_MENU_KEY));
		mainMenuBar.add(repositoryMenu);
		repositoryMenu.add(new CreateAutomatonAction(this));
		repositoryMenu.add(new CreateCospanAction(this));
		repositoryMenu.add(new CreateDecompositionAction(this));
		repositoryMenu.add(new CreateGraphAction(this));
		repositoryMenu.add(new CreateSignatureAction(this));
		
		// Create Info menu
		JMenu infoMenu = new JMenu(messages.getString(GuiMessages.INFO_MENU_KEY));
		infoMenu.add(new VariablesAndPropertiesAction(this));
		infoMenu.add(new VersionAction(this));
		infoMenu.add(new UpdateAction());
		infoMenu.addSeparator();
		infoMenu.add(new HelpAction(this));
		mainMenuBar.add(infoMenu);

		return mainMenuBar;
	}

	/**
	 * Creates the main panel.
	 * @return the main panel
	 */
	private JPanel createMainPanel() {
		JPanel panel = new JPanel(new BorderLayout());

		panel.setOpaque(false);
		panel.add(new TechniquePanel(), BorderLayout.NORTH);
		panel.add(createTextPanel(), BorderLayout.CENTER);

		return panel;
	}

	/**
	 * Creates the panel on which repository information will be displayed.
	 * @return the repository panel
	 */
	private JSplitPane createRepositoryPanel() {
		repositoryTableModel = new RepositoryTableModel();
		repositoryTableModel.addColumn(messages.getString(GuiMessages.NAME_LABEL_KEY));
		repositoryTableModel.addColumn(messages.getString(GuiMessages.TYPE_LABEL_KEY));

		final JTable repositoryTable = new JTable(repositoryTableModel);
		repositoryTable.setOpaque(false);
		repositoryTable.setBackground(MainWindow.BACKGROUND_COLOR);

		repositoryTable.setAutoCreateColumnsFromModel(true);
		repositoryTable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		repositoryTable.getSelectionModel().addListSelectionListener(
				new ListSelectionListener() {

					@Override
					public void valueChanged(ListSelectionEvent e) {
						if (!e.getValueIsAdjusting()) {
							CardLayout layout = (CardLayout) (infoPanel.getLayout());
							int row = repositoryTable.getSelectedRow();

							if (row > -1) {
								Data data = repo.get(repositoryTable.getValueAt(
										repositoryTable.convertRowIndexToModel(row),
										repositoryTable.convertColumnIndexToModel(0)).toString());

								if (data instanceof Automaton) {
									CardLayout automatonPanelLayout = (CardLayout) (automatonInformationPanel.getLayout());
									String cardLabel = EMPTY_CARD_LABEL;

									if(data instanceof ColorabilityAutomaton) {
										colorabilityInformationPanel.setData((ColorabilityAutomaton) data);
										cardLabel = COLORABILITY_AUTOMATON_CARD_LABEL;
									}
									else if(data instanceof DominatingSetAutomaton) {
										dominatingSetInformationPanel.setData((DominatingSetAutomaton) data);
										cardLabel = DOMINATING_SET_AUTOMATON_CARD_LABEL;
									}
									else if(data instanceof EdgeCountingAutomaton) {
										edgeCountingInformationPanel.setData((EdgeCountingAutomaton) data);
										cardLabel = EDGE_COUNTING_AUTOMATON_CARD_LABEL;
									}
									else if(data instanceof LinkAutomaton) {
										linkInformationPanel.setData((LinkAutomaton) data);
										cardLabel = LINK_AUTOMATON_CARD_LABEL;
									}
									else if(data instanceof MaximumEdgeAutomaton) {
										maximumEdgeInformationPanel.setData((MaximumEdgeAutomaton) data);
										cardLabel = MAXIMUM_EDGE_AUTOMATON_CARD_LABEL;
									}
									else if(data instanceof MaximumVertexAutomaton) {
										maximumVertexInformationPanel.setData((MaximumVertexAutomaton) data);
										cardLabel = MAXIMUM_VERTEX_AUTOMATON_CARD_LABEL;
									}
									else if(data instanceof MinimumEdgeAutomaton) {
										minimumEdgeInformationPanel.setData((MinimumEdgeAutomaton) data);
										cardLabel = MINIMUM_EDGE_AUTOMATON_CARD_LABEL;
									}
									else if(data instanceof MinimumVertexAutomaton) {
										minimumVertexInformationPanel.setData((MinimumVertexAutomaton) data);
										cardLabel = MINIMUM_VERTEX_AUTOMATON_CARD_LABEL;
									}
									else if(data instanceof NoIsolatedNodesAutomaton) {
										noIsolatedNodesInformationPanel.setData((NoIsolatedNodesAutomaton) data);
										cardLabel = NO_ISOLATED_NODES_AUTOMATON_CARD_LABEL;
									}
									else if(data instanceof PathAutomaton) {
										pathInformationPanel.setData((PathAutomaton) data);
										cardLabel = PATH_AUTOMATON_CARD_LABEL;
									}
									else if(data instanceof ProductAutomaton) {
										productInformationPanel.setData((ProductAutomaton) data);
										cardLabel = PRODUCT_AUTOMATON_CARD_LABEL;
									}
									else if(data instanceof SubgraphAutomaton) {
										subgraphInformationPanel.setData((SubgraphAutomaton) data);
										cardLabel = SUBGRAPH_AUTOMATON_CARD_LABEL;
									}
									else if(data instanceof UnionAutomaton) {
										unionInformationPanel.setData((UnionAutomaton) data);
										cardLabel = UNION_AUTOMATON_CARD_LABEL;
									}
									else if(data instanceof VertexCountingAutomaton) {
										vertexCountingInformationPanel.setData((VertexCountingAutomaton) data);
										cardLabel = VERTEX_COUNTING_AUTOMATON_CARD_LABEL;
									}
									else if(data instanceof VertexCoverAutomaton) {
										vertexCoverInformationPanel.setData((VertexCoverAutomaton) data);
										cardLabel = VERTEX_COVER_AUTOMATON_CARD_LABEL;
									}
									
									automatonPanelLayout.show(automatonInformationPanel, cardLabel);
									layout.show(infoPanel, AUTOMATON_CARD_LABEL);
								} else if (data instanceof Graph) {
									graphInformationPanel.setData((Graph) data);
									layout.show(infoPanel, GRAPH_CARD_LABEL);
								} else if (data instanceof Signature) {
									signatureInformationPanel.setData((Signature) data);
									layout.show(infoPanel, SIGNATURE_CARD_LABEL);
								} else if (data instanceof RavenBDD) {
									bddInformationPanel.setData((RavenBDD) data);
									layout.show(infoPanel, BDD_CARD_LABEL);
								} else if (data instanceof CospanDecomposition) {
									decompositionInformationPanel.setData((CospanDecomposition) data);
									layout.show(infoPanel, DECOMPOSITION_CARD_LABEL);
								} else if (data instanceof Cospan) {
									cospanInformationPanel.setData((Cospan) data);
									layout.show(infoPanel, COSPAN_CARD_LABEL);
								} else if (data instanceof TreeDecomposition) {
									treeDecompositionInformationPanel.setData((TreeDecomposition) data);
									layout.show(infoPanel, TREE_DECOMPOSITION_CARD_LABEL);
								} else if (data instanceof PathDecomposition) {
									pathDecompositionInformationPanel.setData((PathDecomposition) data);
									layout.show(infoPanel, PATH_DECOMPOSITION_CARD_LABEL);
								}
							}
							else {
								layout.show(infoPanel, EMPTY_CARD_LABEL);
							}
						}

					}
				});
		

		JToolBar tableToolBar = new JToolBar();
		tableToolBar.setBorderPainted(false);
		tableToolBar.setOpaque(false);
		tableToolBar.setFloatable(false);
		tableToolBar.add(new AddToRepositoryAction(this));
		tableToolBar.add(new RemoveFromRepositoryAction(repositoryTable, out));
		tableToolBar.add(new LoadButtonAction(this));
		tableToolBar.add(new SaveButtonAction(this));

		JScrollPane scrollpane = new JScrollPane(repositoryTable);
		scrollpane.setOpaque(false);
		scrollpane.getViewport().setOpaque(false);
		
		JPanel tablePanel = new JPanel(new BorderLayout());
		tablePanel.setOpaque(false);
		tablePanel.add(scrollpane, BorderLayout.CENTER);
		tablePanel.add(tableToolBar, BorderLayout.SOUTH);

		Utilities.setCompoundBorder(tablePanel, messages.getString(GuiMessages.REPOSITORY_PANEL_LABEL));
		colorabilityInformationPanel = ColorabilityAutomatonInformationPanelBuilder.createPanel();
		dominatingSetInformationPanel = DominatingSetAutomatonInformationPanelBuilder.createPanel();
		edgeCountingInformationPanel = EdgeCountingAutomatonInformationPanelBuilder.createPanel();
		linkInformationPanel = LinkAutomatonInformationPanelBuilder.createPanel();
		maximumEdgeInformationPanel = MaximumEdgeAutomatonInformationPanelBuilder.createPanel();
		maximumVertexInformationPanel = MaximumVertexAutomatonInformationPanelBuilder.createPanel();
		minimumEdgeInformationPanel = MinimumEdgeAutomatonInformationPanelBuilder.createPanel();
		minimumVertexInformationPanel = MinimumVertexAutomatonInformationPanelBuilder.createPanel();
		noIsolatedNodesInformationPanel = NoIsolatedNodesAutomatonInformationPanelBuilder.createPanel();
		pathInformationPanel = PathAutomatonInformationPanelBuilder.createPanel();
		productInformationPanel = ProductAutomatonInformationPanelBuilder.createPanel();
		subgraphInformationPanel = SubgraphAutomatonInformationPanelBuilder.createPanel();
		unionInformationPanel = UnionAutomatonInformationPanelBuilder.createPanel();
		vertexCountingInformationPanel = VertexCountingAutomatonInformationPanelBuilder.createPanel();
		vertexCoverInformationPanel = VertexCoverAutomatonInformationPanelBuilder.createPanel();
		
		JPanel emptyPanel = new JPanel();
		emptyPanel.setOpaque(false);
		
		automatonInformationPanel = new JPanel(new CardLayout());
		automatonInformationPanel.setOpaque(false);
		automatonInformationPanel.add(emptyPanel, EMPTY_CARD_LABEL);
		automatonInformationPanel.add(colorabilityInformationPanel, COLORABILITY_AUTOMATON_CARD_LABEL);
		automatonInformationPanel.add(dominatingSetInformationPanel, DOMINATING_SET_AUTOMATON_CARD_LABEL);
		automatonInformationPanel.add(edgeCountingInformationPanel, EDGE_COUNTING_AUTOMATON_CARD_LABEL);
		
		automatonInformationPanel.add(linkInformationPanel, LINK_AUTOMATON_CARD_LABEL);
		automatonInformationPanel.add(maximumEdgeInformationPanel, MAXIMUM_EDGE_AUTOMATON_CARD_LABEL);
		automatonInformationPanel.add(maximumVertexInformationPanel, MAXIMUM_VERTEX_AUTOMATON_CARD_LABEL);
		automatonInformationPanel.add(minimumEdgeInformationPanel, MINIMUM_EDGE_AUTOMATON_CARD_LABEL);
		automatonInformationPanel.add(minimumVertexInformationPanel, MINIMUM_VERTEX_AUTOMATON_CARD_LABEL);
		automatonInformationPanel.add(noIsolatedNodesInformationPanel, NO_ISOLATED_NODES_AUTOMATON_CARD_LABEL);
		automatonInformationPanel.add(pathInformationPanel, PATH_AUTOMATON_CARD_LABEL);
		automatonInformationPanel.add(productInformationPanel, PRODUCT_AUTOMATON_CARD_LABEL);
		automatonInformationPanel.add(subgraphInformationPanel, SUBGRAPH_AUTOMATON_CARD_LABEL);
		automatonInformationPanel.add(unionInformationPanel, UNION_AUTOMATON_CARD_LABEL);
		automatonInformationPanel.add(vertexCountingInformationPanel, VERTEX_COUNTING_AUTOMATON_CARD_LABEL);
		automatonInformationPanel.add(vertexCoverInformationPanel, VERTEX_COVER_AUTOMATON_CARD_LABEL);

		decompositionInformationPanel = CospanDecompositionInformationPanelBuilder.createPanel();
		cospanInformationPanel = CospanInformationPanelBuilder.createPanel();
		treeDecompositionInformationPanel = TreeDecompositionInformationPanelBuilder.createPanel();
		pathDecompositionInformationPanel = PathDecompositionInformationPanelBuilder.createPanel();
		graphInformationPanel = GraphInformationPanelBuilder.createPanel();
		signatureInformationPanel = SignatureInformationPanelBuilder.createPanel();
		bddInformationPanel = BDDInformationPanelBuilder.createPanel();

		emptyPanel = new JPanel();
		emptyPanel.setOpaque(false);

		infoPanel = new JPanel(new CardLayout());
		infoPanel.setOpaque(false);
		infoPanel.add(emptyPanel, EMPTY_CARD_LABEL);
		infoPanel.add(automatonInformationPanel, AUTOMATON_CARD_LABEL);
		infoPanel.add(decompositionInformationPanel, DECOMPOSITION_CARD_LABEL);
		infoPanel.add(cospanInformationPanel, COSPAN_CARD_LABEL);
		infoPanel.add(treeDecompositionInformationPanel, TREE_DECOMPOSITION_CARD_LABEL);
		infoPanel.add(pathDecompositionInformationPanel, PATH_DECOMPOSITION_CARD_LABEL);
		infoPanel.add(graphInformationPanel, GRAPH_CARD_LABEL);
		infoPanel.add(signatureInformationPanel, SIGNATURE_CARD_LABEL);
		infoPanel.add(bddInformationPanel, BDD_CARD_LABEL);

		JScrollPane infoScrollpane = new JScrollPane(infoPanel);
		Utilities.setCompoundBorder(infoScrollpane, messages.getString(GuiMessages.DATA_INFORMATION_PANEL_KEY));
		infoScrollpane.setOpaque(false);
		infoScrollpane.getViewport().setOpaque(false);
		
		JSplitPane splitPane = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
		splitPane.setOpaque(false);
		splitPane.setDividerLocation(310);
		splitPane.add(tablePanel);
		splitPane.add(infoScrollpane);

		return splitPane;
	}

	/**
	 * Creates the panel on which output text area will be displayed.
	 * Additionally a input field is added to execute console commands.
	 * @return the output text area panel
	 */
	private JPanel createTextPanel() {
		JPanel panel = new JPanel(new BorderLayout());
		panel.setBackground(MainWindow.BACKGROUND_COLOR);
		
		JScrollPane scrollpane = new JScrollPane(textarea,
				JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
				JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
		scrollpane.getViewport().setBackground(MainWindow.BACKGROUND_COLOR);
		scrollpane.setBackground(MainWindow.BACKGROUND_COLOR);
		
		panel.add(scrollpane, BorderLayout.CENTER);

		// create the input panel
		JPanel consoleInputPanel = new JPanel(new BorderLayout());
		consoleInputPanel.add(consoleInput, BorderLayout.CENTER);
		consoleInputPanel.setBackground(BACKGROUND_COLOR);
		panel.add(consoleInputPanel, BorderLayout.SOUTH);

		// Create borders
		Utilities.setCompoundBorder(consoleInputPanel, 
				messages.getString(GuiMessages.COMMAND_LINE_LABEL_KEY));

		// Create borders
		Utilities.setCompoundBorder(panel, messages.getString(GuiMessages.OUTPUT_LABEL_KEY));
		return panel;
	}

}
