/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui;

import javax.swing.SwingUtilities;

import com.jgoodies.forms.builder.DefaultFormBuilder;

import de.uni_due.inf.ti.raven.actions.ComputationEvent;
import de.uni_due.inf.ti.raven.actions.ComputationEvent.EventType;
import de.uni_due.inf.ti.raven.actions.ComputationListener;
import de.uni_due.inf.ti.raven.actions.MembershipAction;
import de.uni_due.inf.ti.raven.algorithms.DefaultCospanDecompositionAlgorithm;
import de.uni_due.inf.ti.raven.algorithms.LibTWGreedyFillInAlgorithm;
import de.uni_due.inf.ti.raven.data.Cospan;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.data.automata.Automaton;

/**
 * This class implements a {@code ControlPanel} for the invocation of the different membership 
 * checking algorithms implemented in Raven.
 * @author Christoph Blume
 * @version $Revision: 1.12 $
 * 
 */
public final class MembershipControlPanel extends ControlPanel {
	
	/**
	 * This class is used as a worker thread to perform the membership check. 
	 */
	private class MembershipWorkerThread extends WorkerThread {

		private String automatonName;
		private String graphDecompositionName;
		private CospanDecomposition decomposition;
		private MembershipAction action;

		@Override
		protected Void doInBackground() throws Exception {
			// Get the parameter for the invariant checking algorithm
			Automaton automaton = (Automaton) (automatonBox.getSelectedData());

			// Store the names of the parameter (which will be used later on for the output)
			automatonName = automatonBox.getSelectedName();
			graphDecompositionName = inputBox.getSelectedName();

			if(automaton==null) {
				Utilities.showWarningMessage(MembershipControlPanel.this,
						GuiMessages.NO_AUTOMATON_SELECTED_TITLE_EXCEPTION_KEY,
						GuiMessages.NO_AUTOMATON_SELECTED_TEXT_EXCEPTION_KEY,
						messages.getString(GuiMessages.AUTOMATON_LABEL_KEY));
				cancel(true);
				fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_CANCELLED, "", 0, 0));
				return null;
			}
			else if(graphDecompositionName==null) {
				Utilities.showWarningMessage(MembershipControlPanel.this,
						GuiMessages.NO_DECOMPOSITION_SELECTED_TITLE_EXCEPTION_KEY,
						GuiMessages.NO_DECOMPOSITION_SELECTED_TEXT_EXCEPTION_KEY,
						messages.getString(GuiMessages.INPUT_DECOMPOSITION_LABEL_KEY));
				cancel(true);
				fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_CANCELLED, "", 0, 0));
				return null;
			}
			
			// Create membership action
			action = null;
			
			if(inputBox.getSelectedData() instanceof Graph) {
				Graph graph = (Graph) (inputBox.getSelectedData());
				LibTWGreedyFillInAlgorithm algorithm = new LibTWGreedyFillInAlgorithm();
				decomposition = CospanDecomposition.getCospanDecomposition(algorithm.decompose(graph));
			} 
			if(inputBox.getSelectedData() instanceof Cospan) {
				Cospan cospan = (Cospan) (inputBox.getSelectedData());
				DefaultCospanDecompositionAlgorithm algorithm = new DefaultCospanDecompositionAlgorithm();
				decomposition = algorithm.decompose(cospan);
			} 
			else {
				decomposition = (CospanDecomposition) (inputBox.getSelectedData());
			}

			action = new MembershipAction(automaton, decomposition);
			// Register listeners
			for(ComputationListener listener : listeners) {
				action.addComputationListener(listener);
			}

			// Start computation
			action.execute();
			return null;
		}

		@Override
		protected void done() {
			//If the worker thread has been cancelled
			if(isCancelled()) {
				if(action!=null) {
					//also cancel the computation of the action
					action.cancelComputation();
				}
				
				automatonName = null;
				graphDecompositionName = null;
			}
			//else bring up message dialog to display result
			else {
				boolean isMember = action.getResult();
				String titleKey;
				String textKey;
				
				if (isMember) {
					titleKey = GuiMessages.MEMBERSHIP_DOES_HOLD_DIALOG_TITLE_KEY;
					textKey = GuiMessages.MEMBERSHIP_DOES_HOLD_DIALOG_TEXT_KEY;
				} 
				else {
					titleKey = GuiMessages.MEMBERSHIP_DOES_NOT_HOLD_DIALOG_TITLE_KEY;
					textKey = GuiMessages.MEMBERSHIP_DOES_NOT_HOLD_DIALOG_TEXT_KEY;
				}

				int value = Utilities.showConfirmDialog(null, titleKey, textKey, 
						automatonName, graphDecompositionName);
				
				//If the user wants to see the automaton's run
				if(value==RavenDialog.YES_OPTION) {				
					//Get the automaton and the decomposition
					Repository repo = Repository.getRepository();
					final Automaton automaton = (Automaton) repo.get(automatonName);
					decomposition = (CospanDecomposition) Repository.getRepository().get(graphDecompositionName);

					//Show the resulting run to the user
					SwingUtilities.invokeLater(new Runnable() {
						
						@Override
						public void run() {
							GraphWindow.getWindow().showRun(automaton, decomposition);
						}
					});
				}
			}
		}

	}
	
	private static final long serialVersionUID = 4882608907510085358L;

	/** Automaton Box Label (used for Testing) */
	protected static final String AUTOMATON_BOX_NAME = "automatonBox";
	
	/** Input Box Label (used for Testing) */
	protected static final String INPUT_BOX_LABEL = "inputBox";
	
	private DataComboBox automatonBox;
	private DataComboBox inputBox;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public void setEnabled(boolean isEnabled) {
		automatonBox.setEnabled(isEnabled);
		inputBox.setEnabled(isEnabled);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	protected void initialize(DefaultFormBuilder builder) {
		Repository repo = Repository.getRepository();
		
		automatonBox = new DataComboBox(repo);
		automatonBox.setName(AUTOMATON_BOX_NAME);
		automatonBox.addDataClass(Automaton.class);
		append(builder, GuiMessages.AUTOMATON_LABEL_KEY, automatonBox);
		
		inputBox = new DataComboBox(repo);
		inputBox.setName(INPUT_BOX_LABEL);
		inputBox.addDataClass(Cospan.class);
		inputBox.addDataClass(CospanDecomposition.class);
		inputBox.addDataClass(Graph.class);
		append(builder, GuiMessages.INPUT_GRAPH_DECOMPOSITION_LABEL_KEY, inputBox);
		
		// set Color
		automatonBox.setBackground(getBackground());
		inputBox.setBackground(getBackground());
	}

	@Override
	protected WorkerThread getWorkerThread() {
		return new MembershipWorkerThread();
	}

}
