/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;
import javax.swing.SpringLayout;

import de.uni_due.inf.ti.raven.data.Edge;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.Node;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class provides a dialog to create a new edge for a given graph.
 * @author Christoph Blume, Dennis Nolte
 * @version $Revision: 1.9 $
 * 
 * @see Graph
 * @see Edge
 * @see Node
 */
public class NewEdgeDialog extends RavenDialog {

	/**
	 * This class provides information about an edge which is to be added.
	 * @author Christoph Blume
	 * @version $Revision: 1.9 $
	 * 
	 * @see Edge
	 */
	public static final class EdgeData {
		
		private String name;
		private String label;
		private List<Node> incidentNodes;

		/**
		 * Creates a new {@code EdgeData}-object containing the name, the label
		 * and the sequence of incident nodes of a new edge.
		 * @param name the name of the new edge
		 * @param label the label of the new edge
		 * @param incidentNodes the sequence of incident nodes
		 */
		public EdgeData(String name, String label, List<Node> incidentNodes) {
			this.name = name;
			this.label = label;
			this.incidentNodes = incidentNodes;
		}
		
		/**
		 * Returns the name of the edge.
		 * @return the name
		 */
		public String getName() {
			return name;
		}

		/**
		 * Returns the label of the edge.
		 * @return the label
		 */
		public String getLabel() {
			return label;
		}

		/**
		 * Returns the incident nodes of the edge.
		 * @return the incident nodes
		 */
		public List<Node> getIncidentNodes() {
			return incidentNodes;
		}
		
		@Override
		public String toString() {
			return name+"="+label+"("+incidentNodes+")";
		}
	}
	
	private static final long serialVersionUID = 3601676806719487497L;

	private static final String NAME_FIELD_LABEL = "nameField";
	private static final String LABEL_FIELD_LABEL = "labelField";
	private static final String INCIDENT_NODES_LIST_LABEL = "incidentNodesList";

	private static NewEdgeDialog dialog;
	private static EdgeData edgeData;

	private Graph graph;
	private JTextField nameField;
	private JTextField labelField;
	private JList<Node> availableNodesList;
	private JList<Node> incidentNodesList;

	/*------------------------------------------------------------------------------------------*/
	/* 		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Brings up a new dialog on which the user can enter the name and the label 
	 * and select the incident nodes of the new edge. 
	 * @param owner the parent window of the dialog
	 * @param graph the graph to which the new edge is to be added
	 * @param presetNodes the set of preset nodes which are used for the sequence of incident nodes
	 * @return an {@code EdgeData}-object which contains the name, the label and the sequence of incident nodes select by the user
	 * @throws NullPointerException thrown if either the graph or the preset nodes are {@code null}.
	 * @throws IllegalArgumentException thrown if at least one of the preset nodes is not contained in the graph.
	 */
	public static EdgeData showDialog(Window owner, Graph graph, 
			List<Node> presetNodes) {
		if(graph==null) {
			throw new NullPointerException(RavenExceptionMessages.GRAPH_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(presetNodes==null) {
			throw new NullPointerException(RavenExceptionMessages.PRESET_NODE_SET_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(!graph.getNodes().containsAll(presetNodes)) {
			throw new IllegalArgumentException(RavenExceptionMessages.NODE_NOT_EXISTING_EXCEPTION);
		}

		dialog = new NewEdgeDialog(owner);
		dialog.initialize(graph, presetNodes);
		dialog.setLocationRelativeTo(owner);
		dialog.setVisible(true);
		
		return edgeData;
	}

	/**
	 * Brings up a new dialog on which the user can enter the name and the label 
	 * and select the incident nodes of the new edge. 
	 * @param owner the parent window of the dialog
	 * @param graph the graph to which the new edge is to be added
	 * @return an {@code EdgeData}-object which contains the name, the label and the sequence of incident nodes select by the user
	 * @throws NullPointerException thrown if either the graph is {@code null}.
	 */
	public static EdgeData showDialog(Window owner, Graph graph) {
		return showDialog(owner, graph, Collections.<Node>emptyList());
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * 
	 * @param owner the parent window of this dialog
	 */
	private NewEdgeDialog(Window owner) {
		super(owner, messages.getString(GuiMessages.NEW_EDGE_DIALOG_TITLE_KEY), 
				ACCEPT_CANCEL_LAYOUT);
		getContentPane().setBackground(MainWindow.BACKGROUND_COLOR);
		AbstractAction acceptAction = new AbstractAction(
				messages.getString(GuiMessages.ACCEPT_LABEL_KEY)) {

			private static final long serialVersionUID = 2418165190397918534L;

			@Override
			public void actionPerformed(ActionEvent e) {
				String edgeName = ((JTextField) Utilities.getComponentByName(
						NAME_FIELD_LABEL, NewEdgeDialog.this)).getText();
				String edgeLabel = ((JTextField) Utilities.getComponentByName(
						LABEL_FIELD_LABEL, NewEdgeDialog.this)).getText();

				@SuppressWarnings("unchecked")
				JList<Node> list = ((JList<Node>) Utilities.getComponentByName(
						INCIDENT_NODES_LIST_LABEL, NewEdgeDialog.this));
				List<Node> incidentNodes = Collections.list(
						((DefaultListModel<Node>) list.getModel()).elements());

				if (!Operation.isLabelValid(edgeLabel)) {
					showErrorMessage(GuiMessages.INVALID_LABEL_TITLE_EXCEPTION_KEY,
							GuiMessages.INVALID_LABEL_TEXT_EXCEPTION_KEY, edgeLabel);
					edgeData = null;
					return;
				}
				if (edgeName.isEmpty()) {
					showErrorMessage(GuiMessages.EMPTY_INPUTFIELD_TITLE_EXCEPTION_KEY,
							GuiMessages.EMPTY_INPUTFIELD_TEXT_EXCEPTION_KEY,
							messages.getString(GuiMessages.NAME_LABEL_KEY));
					edgeData = null;
					return;
				}
				if (edgeLabel.isEmpty()) {
					showErrorMessage(GuiMessages.EMPTY_INPUTFIELD_TITLE_EXCEPTION_KEY,
							GuiMessages.EMPTY_INPUTFIELD_TEXT_EXCEPTION_KEY,
							messages.getString(GuiMessages.LABEL_LABEL_KEY));
					edgeData = null;
					return;
				}
				if(graph.getLabels().keySet().contains(edgeLabel) && 
						graph.getLabels().get(edgeLabel)!=incidentNodes.size()) {
					showErrorMessage(GuiMessages.INVALID_LABEL_ARITY_TITLE_EXCEPTION_KEY,
							GuiMessages.INVALID_LABEL_ARITY_TEXT_EXCEPTION_KEY, edgeLabel);
					edgeData = null;
					return;
				}
				if(graph.hasEdge(edgeName)) {
					showErrorMessage(GuiMessages.EXISTING_EDGE_NAME_TITLE_EXCEPTION_KEY,
							GuiMessages.EXISTING_EDGE_NAME_TEXT_EXCEPTION_KEY,
							edgeName);
					edgeData = null;
					return;
				}

				edgeData = new EdgeData(edgeName, edgeLabel, incidentNodes);
				NewEdgeDialog.this.dispose();
			}

		};

		setMainPane(createMainPanel());
		setAction(ACCEPT_BUTTON, acceptAction);
		pack();
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates the main panel for the New-Edge-Dialog.
	 * @return the main panel
	 */
	private JPanel createMainPanel() {
		String labelSuffix = ": ";

		// Create panel for name and the label fields
		JLabel nameLabel = new JLabel(
				messages.getString(GuiMessages.NAME_LABEL_KEY)
				+ labelSuffix);
		nameField = new JTextField(10);
		nameField.setName(NAME_FIELD_LABEL);

		JLabel labelLabel = new JLabel(
				messages.getString(GuiMessages.LABEL_LABEL_KEY)
				+ labelSuffix);
		labelField = new JTextField(10);
		labelField.setName(LABEL_FIELD_LABEL);

		SpringLayout layout = new SpringLayout();
		JPanel nameLabelPanel = new JPanel(layout);
		nameLabelPanel.add(nameLabel);
		nameLabelPanel.add(nameField);
		nameLabelPanel.add(labelLabel);
		nameLabelPanel.add(labelField);

		Utilities.makeGridLayout(layout, nameLabelPanel, 4, 10, 10);

		// Create panel for the selection of the incident nodes
		JPanel incidentNodePanel = new JPanel(new GridBagLayout());
		GridBagConstraints c;
		JScrollPane scrollpane;

		availableNodesList = new JList<>(new DefaultListModel<Node>());
		availableNodesList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		c = new GridBagConstraints();
		c.weightx = 2;
		c.gridx = 0;
		c.gridy = 0;
		c.gridwidth = 2;
		c.gridheight = 5;
		scrollpane = new JScrollPane(availableNodesList);
		scrollpane.setPreferredSize(new Dimension(150, 300));
		incidentNodePanel.add(scrollpane, c);

		incidentNodesList = new JList<>(new DefaultListModel<Node>());
		incidentNodesList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		incidentNodesList.setName(INCIDENT_NODES_LIST_LABEL);

		c = new GridBagConstraints();
		c.weightx = 2;
		c.gridx = 3;
		c.gridy = 0;
		c.gridwidth = 2;
		c.gridheight = 5;
		scrollpane = new JScrollPane(incidentNodesList);
		scrollpane.setPreferredSize(new Dimension(150, 300));
		incidentNodePanel.add(scrollpane, c);

		JPanel buttonPanel = new JPanel(new GridLayout(4, 1));

		JButton shiftRightButton = new JButton(new AbstractAction(messages.getString(GuiMessages.SHIFT_RIGHT_LABEL_KEY)) {

			private static final long serialVersionUID = -2632710493259765026L;

			@Override
			public void actionPerformed(ActionEvent e) {
				Node node = availableNodesList.getSelectedValue();
				if (node == null) {
					Utilities.showInformationMessage(NewEdgeDialog.this,
							GuiMessages.NEW_EDGE_DIALOG_NO_SELECTION_TITLE_KEY,
							GuiMessages.NEW_EDGE_DIALOG_NO_SELECTION_TEXT_KEY);
				} else {
					((DefaultListModel<Node>) incidentNodesList
							.getModel()).addElement(node);
				}
			}
		});

		JButton shiftLeftButton = new JButton(new AbstractAction(messages.getString(GuiMessages.SHIFT_LEFT_LABEL_KEY)) {

			private static final long serialVersionUID = -5067848443035530041L;

			@Override
			public void actionPerformed(ActionEvent e) {
				Node node = incidentNodesList.getSelectedValue();
				if (node == null) {
					Utilities.showInformationMessage(NewEdgeDialog.this,
							GuiMessages.NEW_EDGE_DIALOG_NO_SELECTION_TITLE_KEY,
							GuiMessages.NEW_EDGE_DIALOG_NO_SELECTION_TEXT_KEY);
				} else {
					((DefaultListModel<Node>) incidentNodesList
							.getModel()).removeElement(node);
				}
			}
		});

		JButton shiftUpButton = new JButton(new AbstractAction(messages.getString(GuiMessages.SHIFT_UP_LABEL_KEY)) {

			private static final long serialVersionUID = -9121620289728882224L;

			@Override
			public void actionPerformed(ActionEvent e) {
				int selectedIndex = incidentNodesList.getSelectedIndex();
				if(selectedIndex > 0) {
					DefaultListModel<Node> model = (DefaultListModel<Node>) incidentNodesList.getModel();
					Node shiftHelp = model.set(selectedIndex-1, model.get(selectedIndex));
					model.set(selectedIndex, shiftHelp);	
					incidentNodesList.setSelectedIndex(selectedIndex-1);
				}
			}
		});

		JButton shiftDownButton = new JButton(new AbstractAction(messages.getString(GuiMessages.SHIFT_DOWN_LABEL_KEY)) {

			private static final long serialVersionUID = 6986887350538094829L;

			@Override
			public void actionPerformed(ActionEvent e) {
				int selectedIndex = incidentNodesList.getSelectedIndex();
				if(selectedIndex < incidentNodesList.getLastVisibleIndex()) {
					DefaultListModel<Node> model = (DefaultListModel<Node>) incidentNodesList.getModel();
					Node shiftHelp = model.set(selectedIndex+1, model.get(selectedIndex));
					model.set(selectedIndex, shiftHelp);	
					incidentNodesList.setSelectedIndex(selectedIndex+1);
				}

			}
		});

		buttonPanel.add(shiftRightButton);
		buttonPanel.add(shiftLeftButton);
		buttonPanel.add(shiftUpButton);
		buttonPanel.add(shiftDownButton);

		c = new GridBagConstraints();
		c.weightx = 2;
		c.gridx = 2;
		c.gridy = 0;
		c.gridwidth = 1;
		c.gridheight = 3;
		incidentNodePanel.add(buttonPanel, c);


		JPanel mainPanel = new JPanel(new BorderLayout());
		nameLabelPanel.setOpaque(false);
		incidentNodePanel.setOpaque(false);
		mainPanel.add(nameLabelPanel, BorderLayout.NORTH);
		mainPanel.add(incidentNodePanel, BorderLayout.CENTER);
		mainPanel.setOpaque(false);
		return mainPanel;
	}

	/**
	 * Initializes the New-Edge-Dialog.
	 */
	private void initialize(Graph graph, List<Node> presetNodes) {
		this.graph = graph;
		edgeData = null;
		nameField.setText("e"+graph.getEdgeCount());
		labelField.setText("");

		DefaultListModel<Node> model = (DefaultListModel<Node>) availableNodesList.getModel();
		model.removeAllElements();

		List<Node> nodeList = new LinkedList<>(graph.getNodes());
		Collections.sort(nodeList);

		for (Node node : nodeList) {
			model.addElement(node);
		}

		model = (DefaultListModel<Node>) incidentNodesList.getModel();
		model.removeAllElements();

		for(Node node: presetNodes) {
				model.addElement(node);					
		}

	}

}
