/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.gui;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.IOException;

import javax.imageio.ImageIO;
import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JPanel;
import javax.swing.JSeparator;
import javax.swing.KeyStroke;

import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * The main class for creating a Raven dialog window.
 * @author Christoph Blume
 * @version $Revision: 1.33 $
 *
 */
public abstract class RavenDialog extends JDialog {

	private static final long serialVersionUID = 3314435372872483891L;

	/** Indicating that an ACCEPT and a CANCEL BUTTON will be shown. */
	public static final int ACCEPT_CANCEL_LAYOUT = 100;

	/** Indicating that an CANCEL BUTTON will be shown. */
	public static final int CANCEL_LAYOUT = 101;

	/** Indicating that an CLOSE BUTTON will be shown. */
	public static final int CLOSE_LAYOUT = 102;

	/** Indicating that an SAVE and a CANCEL BUTTON will be shown. */
	public static final int SAVE_CANCEL_LAYOUT = 103;

	/** Indicating that an LOAD and a CANCEL BUTTON will be shown. */
	public static final int LOAD_CANCEL_LAYOUT = 104;

	/** Indicating that a YES and a NO BUTTON will be shown. */
	public static final int YES_NO_LAYOUT = 105;
	
	/** Indicating that a YES, a NO and a CANCEL BUTTON will be shown. */
	public static final int YES_NO_CANCEL_LAYOUT = 106;
	
	/** Indicating the ACCEPT button. */
	public static final int ACCEPT_BUTTON = 200;

	/** Indicating the ACCEPT button. */
	public static final int CANCEL_BUTTON = 201;

	/** Indicating the ACCEPT button. */
	public static final int CLOSE_BUTTON = 202;
	
	/** Indicating the ACCEPT button. */
	public static final int SAVE_BUTTON = 203;

	/** Indicating the ACCEPT button. */
	public static final int LOAD_BUTTON = 204;

	/** Indicating the ACCEPT button. */
	public static final int YES_BUTTON = 205;

	/** Indicating the ACCEPT button. */
	public static final int NO_BUTTON = 206;
	
	/** Return value form class method if ACCEPT is chosen. */
	public static final int ACCEPT_OPTION = 300;
	
	/** Return value form class method if CANCEL is chosen. */
	public static final int CANCEL_OPTION = 301;
	
	/** Return value form class method if CLOSE is chosen. */
	public static final int CLOSE_OPTION = 302;
	
	/** Return value form class method if SAVE is chosen. */
	public static final int SAVE_OPTION = 303;
	
	/** Return value form class method if LOAD is chosen. */
	public static final int LOAD_OPTION = 304;
	
	/** Return value form class method if YES is chosen. */
	public static final int YES_OPTION = 305;
	
	/** Return value form class method if NO is chosen. */
	public static final int NO_OPTION = 306;

	/** Accept Button Label (used for Testing) */
	public static final String ACCEPT_LABEL = "RavenDialogAcceptButton";
	
	/** Cancel Button Label (used for Testing) */
	public static final String CANCEL_LABEL = "RavenDialogCancelButton";
	
	/** Close Button Label (used for Testing) */
	public static final String CLOSE_LABEL = "RavenDialogCloseButton";
	
	/** Save Button Label (used for Testing) */
	public static final String SAVE_LABEL = "RavenDialogSaveButton";
	
	/** Load Button Label (used for Testing) */
	public static final String LOAD_LABEL = "RavenDialogLoadButton";

	/** Load Button Label (used for Testing) */
	public static final String YES_LABEL = "RavenDialogYesButton";
	
	/** Load Button Label (used for Testing) */
	public static final String NO_LABEL = "RavenDialogNoButton";
	
	/** */
	protected static GuiMessages messages = GuiMessages.getMessages();

	private JPanel buttonPanel = null;

	private JButton acceptButton = null;	
	private JButton cancelButton = null;
	private JButton closeButton = null;
	private JButton loadButton = null;
	private JButton noButton = null;
	private JButton saveButton = null;
	private JButton yesButton = null;

	private JPanel contentPane;	
	
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new button for this dialog.
	 * @param action the action which is assigned to the button
	 * @param name the name for the button (used only for internal purposes)
	 * @param key the key which indicates the label of the button
	 * @return the created button
	 */	
	protected static JButton createButton(AbstractAction action, String name, String key) {
		JButton button = new JButton(action);
		button.setName(name);
		button.setText(messages.getString(key));

		return button;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new dialog.
	 * @param parent the parent of the dialog
	 * @param title the dialog's title
	 * @param buttonLayout the button layout of the dialog
	 * 
	 * @see #ACCEPT_CANCEL_LAYOUT
	 * @see #CANCEL_LAYOUT
	 * @see #CLOSE_LAYOUT
	 */
	protected RavenDialog(Component parent, String title, int buttonLayout) {
//		super(SwingUtilities.getWindowAncestor(parent));
		setModalityType(ModalityType.APPLICATION_MODAL);
		setTitle(title);
		
		try {
			setIconImage(ImageIO.read(new File(RavenMain.RAVEN_ICON_FILENAME)));
		} catch (IOException e) {
			//Do nothing
		}
		
		if(!(buttonLayout==ACCEPT_CANCEL_LAYOUT || buttonLayout==CANCEL_LAYOUT || 
				buttonLayout==CLOSE_LAYOUT || buttonLayout == SAVE_CANCEL_LAYOUT ||
				buttonLayout == LOAD_CANCEL_LAYOUT || buttonLayout == YES_NO_LAYOUT ||
				buttonLayout == YES_NO_CANCEL_LAYOUT)) {
			//Replace string literal by constant
			throw new IllegalArgumentException("This value is not a legal button layout!");
		}		
		
		JPanel panel = createButtonPanel(buttonLayout);
		contentPane = new JPanel(new BorderLayout());
		contentPane.add(new JPanel(), BorderLayout.CENTER);
		contentPane.add(panel, BorderLayout.SOUTH);
		setContentPane(contentPane);
		
		switch (buttonLayout) {
		case ACCEPT_CANCEL_LAYOUT:
			getRootPane().setDefaultButton(acceptButton);
			break;
			
		case SAVE_CANCEL_LAYOUT:
			getRootPane().setDefaultButton(saveButton);			
			break;

		case LOAD_CANCEL_LAYOUT:
			getRootPane().setDefaultButton(loadButton);			
			break;

		case CANCEL_LAYOUT:
			getRootPane().setDefaultButton(cancelButton);
			
		case CLOSE_LAYOUT:
			getRootPane().setDefaultButton(closeButton);
			break;

		case YES_NO_LAYOUT:
		case YES_NO_CANCEL_LAYOUT:
			getRootPane().setDefaultButton(yesButton);
			break;

		default:
			getRootPane().setDefaultButton(closeButton);
			break;
		}
		
		//Dispose dialog on "escape"
		rootPane.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(KeyStroke.getKeyStroke(
				KeyEvent.VK_ESCAPE, 0), "dispose");
		rootPane.getActionMap().put("dispose", new AbstractAction() {

			private static final long serialVersionUID = -8222251245662454026L;

			@Override
			public void actionPerformed(ActionEvent e) {
				dispose();
			}
		});
		
		addWindowListener(new WindowAdapter() {
			
			@Override
			public void windowClosing(WindowEvent e) {
				dispose();
			}

		});
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Sets the dialog's main pane.
	 * @param comp the main pane component for this dialog
	 */
	public final void setMainPane(JComponent comp) {
		getContentPane().add(comp, BorderLayout.CENTER);
		repaint();
	}

	/**
	 * Sets the {@code action} for the specified button.
	 * @param buttonType the type of the button
	 * @param action the {@code Action} which is executed when the button is pressed
	 */
	public void setAction(int buttonType, Action action) {
		if(buttonType==ACCEPT_BUTTON) {
			acceptButton.setAction(action);
		}
		if(buttonType==SAVE_BUTTON) {
			saveButton.setAction(action);
		}
		if(buttonType==LOAD_BUTTON) {
			loadButton.setAction(action);
		}
		else if(buttonType==CANCEL_BUTTON) {
			cancelButton.setAction(action);
		}
		else if(buttonType==CLOSE_BUTTON) {
			closeButton.setAction(action);
		}
		else if(buttonType==YES_BUTTON) {
			yesButton.setAction(action);
		}
		else if(buttonType==NO_BUTTON) {
			noButton.setAction(action);
		}
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Adds the certain button to the dialog's button panel.
	 * @param button the button which is to be added
	 */
	protected final void addButtonToPanel(JButton button) {
		buttonPanel.add(button, -1);
		repaint();
	}
	
	/**
	 * Brings up a message dialog showing a warning message.
	 * @param titleKey the key indicating the title string
	 * @param messageKey the key indicating the message string
	 * @param parameter additional parameter for the message string
	 */
	protected final void showWarningMessage(String titleKey, String messageKey, String... parameter) {
		Utilities.showWarningMessage(this, titleKey, messageKey, parameter);
	}
	
	/**
	 * Brings up a message dialog showing a error message.
	 * @param titleKey the key indicating the title string
	 * @param messageKey the key indicating the message string
	 * @param parameter additional parameter for the message string
	 */
	protected final void showErrorMessage(String titleKey, String messageKey, String... parameter) {
		Utilities.showErrorMessage(this, titleKey, messageKey, parameter);
	}
	
	/**
	 * Brings up a message dialog showing a information message.
	 * @param titleKey the key indicating the title string
	 * @param messageKey the key indicating the message string
	 * @param parameter additional parameter for the message string
	 */
	protected final void showInformationMessage(String titleKey, String messageKey, String... parameter) {
		Utilities.showInformationMessage(this, titleKey, messageKey, parameter);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the panel which displays the default buttons.
	 * @param layout value indicating the layout of the button panel
	 * @return the button panel
	 */
	private JPanel createButtonPanel(int layout) {
		JPanel panel = null;

		AbstractAction defaultAction = new AbstractAction() {

			private static final long serialVersionUID = 7257271464106642570L;

			@Override
			public void actionPerformed(ActionEvent e) {
				RavenDialog.this.dispose();
			}
		};

		buttonPanel = new JPanel(new FlowLayout(FlowLayout.RIGHT));
		if(layout==ACCEPT_CANCEL_LAYOUT) {
			acceptButton = createButton(defaultAction, ACCEPT_LABEL, GuiMessages.ACCEPT_LABEL_KEY);
			cancelButton = createButton(defaultAction, CANCEL_LABEL, GuiMessages.CANCEL_LABEL_KEY);
			
			buttonPanel.add(acceptButton);
			buttonPanel.add(cancelButton);
		}
		else if(layout==SAVE_CANCEL_LAYOUT) {
			saveButton = createButton(defaultAction, SAVE_LABEL, GuiMessages.SAVE_LABEL_KEY);
			cancelButton = createButton(defaultAction, CANCEL_LABEL, GuiMessages.CANCEL_LABEL_KEY);
			
			buttonPanel.add(saveButton);
			buttonPanel.add(cancelButton);
		}
		else if(layout==LOAD_CANCEL_LAYOUT) {
			loadButton = createButton(defaultAction, LOAD_LABEL, GuiMessages.LOAD_LABEL_KEY);
			cancelButton = createButton(defaultAction, CANCEL_LABEL, GuiMessages.CANCEL_LABEL_KEY);
			
			buttonPanel.add(loadButton);
			buttonPanel.add(cancelButton);
		}
		else if(layout==CANCEL_LAYOUT) {
			cancelButton = createButton(defaultAction, CANCEL_LABEL, GuiMessages.CANCEL_LABEL_KEY);
			
			buttonPanel.add(cancelButton);
		}
		else if(layout==CLOSE_LAYOUT) {
			closeButton = createButton(defaultAction, CLOSE_LABEL, GuiMessages.CLOSE_LABEL_KEY);
			
			buttonPanel.add(closeButton);
		}
		else if(layout==YES_NO_LAYOUT) {
			yesButton = createButton(defaultAction, YES_LABEL, GuiMessages.YES_LABEL_KEY);
			noButton = createButton(defaultAction, NO_LABEL, GuiMessages.NO_LABEL_KEY);
			
			buttonPanel.add(yesButton);
			buttonPanel.add(noButton);
		}
		else if(layout==YES_NO_CANCEL_LAYOUT) {
			yesButton = createButton(defaultAction, YES_LABEL, GuiMessages.YES_LABEL_KEY);
			noButton = createButton(defaultAction, NO_LABEL, GuiMessages.NO_LABEL_KEY);
			cancelButton = createButton(defaultAction, CANCEL_LABEL, GuiMessages.CANCEL_LABEL_KEY);
			
			buttonPanel.add(yesButton);
			buttonPanel.add(noButton);
			buttonPanel.add(cancelButton);
		}

		panel = new JPanel(new BorderLayout());
		panel.add(new JSeparator(), BorderLayout.NORTH);
		panel.add(buttonPanel, BorderLayout.CENTER);

		return panel;
	}
	
}
