/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui;

import java.io.File;

import javax.swing.JFileChooser;
import javax.swing.filechooser.FileFilter;

import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * This class provides a {@code FileFilter} that filters the set of files shown 
 * to the user based on the file name extension. The filter accepts only those 
 * files which are relevant for Raven.
 * @author Christoph Blume
 * @version $Revision: 1.5 $
 *
 * @see FileFilter
 * @see JFileChooser
 * @see RavenMain
 */
public final class RavenFileFilter extends FileFilter {

	private static final GuiMessages messages = GuiMessages.getMessages();

	/** Automaton File Filter */
	public static final FileFilter DEFAULT_AUTOMATON_FILE_FILTER = new RavenFileFilter(
			GuiMessages.AUTOMATON_FILE_EXTENSION_KEY, RavenMain.AUTOMATON_FILE_EXTENSION);
	
	/** Cospan File Filter */
	public static final FileFilter DEFAULT_COSPAN_FILE_FILTER = new RavenFileFilter(
			GuiMessages.COSPAN_GXL_FILE_EXTENSION_KEY, RavenMain.COSPAN_GXL_FILE_EXTENSION);
	
	/** Decomposition File Filter */
	public static final FileFilter DEFAULT_DECOMPOSITION_FILE_FILTER = new RavenFileFilter(
			GuiMessages.DECOMPOSITION_FILE_EXTENSION_KEY, RavenMain.DECOMPOSITION_FILE_EXTENSION);
	
	/** Graph GXL File Filter */
	public static final FileFilter GRAPH_GXL_FILE_FILTER = new RavenFileFilter(
			GuiMessages.GRAPH_GXL_FILE_EXTENSION_KEY, RavenMain.GRAPH_GXL_FILE_EXTENSION);
	
	/** Graph SVG File Filter */
	public static final FileFilter GRAPH_SVG_FILE_FILTER = new RavenFileFilter(
			GuiMessages.SVG_FILE_EXTENSION_LABEL_KEY, RavenMain.SVG_FILE_EXTENSION);
	
	/** Signature File Filter */
	public static final FileFilter DEFAULT_SIGNATURE_FILE_FILTER = new RavenFileFilter(
			GuiMessages.SIGNATURE_FILE_EXTENSION_KEY, RavenMain.SIGNATURE_FILE_EXTENSION);
	
	/** Raven Script File Filter */
	public static final FileFilter RAVEN_SCRIPT_FILE_FILTER = new RavenFileFilter(
			GuiMessages.RAVEN_SCRIPT_FILE_EXTENSION_KEY, RavenMain.RAVEN_SCRIPT_FILE_EXTENSION);

	private String descriptionKey;
	private String fileExtension;
	
	/*------------------------------------------------------------------------------------------*/
	/* 		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a file filter with a description given by the {@code descriptionKey} based on the 
	 * file extension {@code fileExtension}.
	 * @param descriptionKey the GuiMessage key used for the description
	 * @param fileExtension the file extension used for the filer
	 */
	private RavenFileFilter(String descriptionKey, String fileExtension) {
		this.descriptionKey = descriptionKey;
		this.fileExtension = fileExtension;
	}
	
	@Override
	public boolean accept(File f) {
		return f.isDirectory() || f.getName().endsWith("."+fileExtension);
	}

	@Override
	public String getDescription() {
		return messages.getString(descriptionKey) + " (*." + fileExtension + ")";
	}

}
