/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.BorderFactory;
import javax.swing.DefaultListCellRenderer;
import javax.swing.DefaultListModel;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.ListCellRenderer;
import javax.swing.ListSelectionModel;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import com.jgoodies.forms.builder.DefaultFormBuilder;
import com.jgoodies.forms.factories.CC;
import com.jgoodies.forms.layout.FormLayout;

import de.uni_due.inf.ti.raven.data.Edge;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.Node;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class provides a dialog to select and remove an existing edge (and all incident nodes) 
 * from a given graph.
 * @author Christoph Blume
 * @version $Revision: 1.3 $
 *
 * @see Graph
 * @see Edge
 * @see Node
 */
public final class RemoveEdgeDialog extends RavenDialog {
	
	private static final long serialVersionUID = 3601676806719487497L;

	private static final String EDGES_LIST_LABEL = "edgesList";
	private static final String LABEL_FIELD_LABEL = "labelField";
	private static final String INCIDENT_NODES_LIST_LABEL = "incidentNodesList";

	private static RemoveEdgeDialog dialog;
	private static Edge selectedEdge;

	private final Graph graph;
	private JList<Edge> edgeList;
	private JList<Node> incidentNodesList;
	private JTextField labelField;

	/*------------------------------------------------------------------------------------------*/
	/* 		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Brings up a new dialog on which the user can enter the name and the label 
	 * and select the incident nodes of the new edge. 
	 * @param owner the parent window of the dialog
	 * @param graph the graph to which the new edge is to be added
	 * @return an {@code EdgeData}-object which contains the name, the label and the sequence of incident nodes select by the user
	 * @throws NullPointerException thrown if either the graph is {@code null}.
	 */
	public static Edge showDialog(Window owner, Graph graph) {
		if(graph==null) {
			throw new NullPointerException(RavenExceptionMessages.GRAPH_MUST_NOT_BE_NULL_EXCEPTION);
		}

		dialog = new RemoveEdgeDialog(owner, graph);
		dialog.setLocationRelativeTo(owner);
		dialog.setVisible(true);
		
		return selectedEdge;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * 
	 * @param owner the parent window of this dialog
	 */
	private RemoveEdgeDialog(Window owner, Graph graph) {
		super(owner, messages.getString(GuiMessages.NEW_EDGE_DIALOG_TITLE_KEY), 
				ACCEPT_CANCEL_LAYOUT);
		this.graph = graph;
		
		getContentPane().setBackground(MainWindow.BACKGROUND_COLOR);
		AbstractAction acceptAction = new AbstractAction(
				messages.getString(GuiMessages.REMOVE_EDGE_LABEL_KEY)) {

			private static final long serialVersionUID = 2418165190397918534L;

			@Override
			public void actionPerformed(ActionEvent e) {
				selectedEdge = edgeList.getSelectedValue();
				
				if(selectedEdge!=null) {
					RemoveEdgeDialog.this.dispose();
				}
				else {
					showWarningMessage(GuiMessages.NO_EDGE_SELECTED_TITLE_EXCEPTION_KEY, 
							GuiMessages.NO_EDGE_SELECTED_TEXT_EXCEPTION_KEY);
				}
			}

		};

		setMainPane(createMainPanel());
		setAction(ACCEPT_BUTTON, acceptAction);
		pack();
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates the main panel for the New-Edge-Dialog.
	 * @return the main panel
	 */
	private JPanel createMainPanel() {
		String labelSuffix = ": ";
		DefaultFormBuilder builder = null;

		// Create panel for the label fields
		JLabel edgesLabel = new JLabel(
				messages.getString(GuiMessages.EDGES_LABEL_KEY) + labelSuffix);

		DefaultListModel<Edge> model =  new DefaultListModel<>();
		List<Edge> edges = new LinkedList<>(graph.getEdges());
		Collections.sort(edges);
		for (Edge edge : edges) {
			model.addElement(edge);
		}
		
		edgeList = new JList<>(model);
		edgeList.setName(EDGES_LIST_LABEL);
		edgeList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		edgeList.addListSelectionListener(new ListSelectionListener() {
			
			@Override
			public void valueChanged(ListSelectionEvent e) {
				Edge edge = edgeList.getSelectedValue();
				
				labelField.setText(edge.getLabel());
				
				DefaultListModel<Node> model = (DefaultListModel<Node>) incidentNodesList.getModel();
				model.removeAllElements();
				
				for(Node node: edge.getIncidentNodes()) {
					model.addElement(node);
				}
				
			}
		});
		edgeList.setCellRenderer(new DefaultListCellRenderer() {

			private static final long serialVersionUID = -3505795828519566784L;
			
			@Override
			public Component getListCellRendererComponent(JList<?> list, Object object, 
					int index, boolean isSelected, boolean cellHasFocus) {			
				return super.getListCellRendererComponent(list, ((Edge) object).getName(), index, 
						isSelected, cellHasFocus);
			}


		});
		
		JScrollPane scrollpane = new JScrollPane(edgeList);
		scrollpane.setPreferredSize(new Dimension(150, 300));
		
		JPanel edgeListPanel = new JPanel();
		edgeListPanel.setOpaque(false);
		edgeListPanel.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 2));

		builder = new DefaultFormBuilder(
				new FormLayout("fill:max(20dlu;pref):grow", "pref, 3dlu, fill:pref:grow"), edgeListPanel);
		builder.add(edgesLabel, CC.xy(1, 1));
		builder.add(scrollpane, CC.xy(1, 3));
		
		// Create panel for the selection of the incident nodes
		JLabel labelLabel = new JLabel(
				messages.getString(GuiMessages.LABEL_LABEL_KEY) + labelSuffix);

		labelField = new JTextField(10);
		labelField.setEditable(false);
		labelField.setName(LABEL_FIELD_LABEL);
		
		JLabel incidentNodesLabel = new JLabel(
				messages.getString(GuiMessages.INCIDENT_NODES_LABEL_KEY) + labelSuffix);

		incidentNodesList = new JList<>(new DefaultListModel<Node>());
		incidentNodesList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		incidentNodesList.setName(INCIDENT_NODES_LIST_LABEL);
		incidentNodesList.setCellRenderer(new ListCellRenderer<Node>() {
			
			private final int formatLength = String.valueOf(
					RemoveEdgeDialog.this.graph.getEdgeCount()).length();

			private final String formatIndex(int index) {
				StringBuilder format = new StringBuilder(String.valueOf(index));
				
				//Insert leading white space
				for(int i=0;i<formatLength-format.length();i++) {
					format.insert(0, " ");
				}
				
				return format.toString();
			}
			
			@Override
			public Component getListCellRendererComponent(JList<? extends Node> list, Node node, 
					int index, boolean isSelected, boolean cellHasFocus) {
				return new JLabel(formatIndex(index)+": "+node.getName());
			}

		});
		
		scrollpane = new JScrollPane(incidentNodesList);
		scrollpane.setPreferredSize(new Dimension(150, 300));
		
		JPanel edgeInformationPanel = new JPanel();
		edgeInformationPanel.setOpaque(false);
		edgeInformationPanel.setBorder(BorderFactory.createEmptyBorder(5, 2, 5, 5));
		
		builder = new DefaultFormBuilder(
				new FormLayout("fill:max(20dlu;pref):grow", "pref, 3dlu, fill:pref, 3dlu, pref, 3dlu, fill:pref:grow"), 
				edgeInformationPanel);
		builder.add(labelLabel, CC.xy(1, 1));
		builder.add(labelField, CC.xy(1, 3));
		builder.add(incidentNodesLabel, CC.xy(1, 5));
		builder.add(scrollpane, CC.xy(1, 7));
		
		JPanel mainPanel = new JPanel();
		mainPanel.setOpaque(false);

		builder = new DefaultFormBuilder(
				new FormLayout("fill:max(20dlu;pref):grow, 3dlu, fill:max(20dlu;pref):grow", "fill:max(20dlu;pref):grow"), 
				mainPanel);
		builder.add(edgeListPanel, CC.xy(1, 1));
		builder.add(edgeInformationPanel, CC.xy(3, 1));

		return mainPanel;
	}

}
