/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.gui;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;

import javax.swing.Action;
import javax.swing.BoxLayout;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.border.EmptyBorder;
import javax.swing.filechooser.FileFilter;
import javax.swing.filechooser.FileSystemView;
import javax.swing.plaf.basic.BasicFileChooserUI;

import de.uni_due.inf.ti.raven.actions.SaveAction;
import de.uni_due.inf.ti.raven.data.Data;
import de.uni_due.inf.ti.raven.data.NameDataPair;
import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * This class is the abstract base class for all {@code Data} save dialogs. It provides a simple
 * mechanism for the user to choose to {@code Data} file to be saved.
 * @author Christoph Blume, Laura Steinert, Dennis Nolte
 * @version $Revision: 1.23 $
 *
 * @see Data
 * @see SaveAction
 */
public abstract class SaveDataDialog extends RavenDialog {

	private static final long serialVersionUID = -2316808700155215140L;
	
	/** GUI Messages Resource Bundle */
	protected static final GuiMessages messages = GuiMessages.getMessages();

	private final boolean isDataChooserVisible;

	private JFileChooser chooser;
	private JLabel dataNameLabel;
	private DataComboBox dataComboBox;
	
	/*------------------------------------------------------------------------------------------*/
	/* 		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Check whether saving a file denoted by filename will overwrite any existing data.
	 * @param filename the name of the file to be checked
	 * @return if this will overwrite anything
	 */
	protected static boolean isFileExisting(String filename){
		return new File(filename).exists();
	}

	/**
	 * Checks whether the filename {@code filename} has the file extension {@code extension}.
	 * If not the filename is concatenated with the file extension 
	 * @param filename the filename to be checked
	 * @param extension the file extension to be used
	 * @return the checked filename
	 */
	protected static String checkFileExtension(String filename, String extension) {
		if (!filename.endsWith("."+extension)){
			filename += "." + extension;
		}

		return filename;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new dialog to save the specified kind of data from the repository.
	 * @param owner the parent window of this dialog
	 * @param title the title of the dialog
	 * @param defaultDirectory the default directory which is used for the file system view
	 */
	protected SaveDataDialog(Window owner, String title, File defaultDirectory) {
		super(owner, title, RavenDialog.SAVE_CANCEL_LAYOUT);

		isDataChooserVisible = false;
		getContentPane().add(createMainPanel(defaultDirectory));
		setLocationByPlatform(true);
		pack();
		setLocationRelativeTo(owner);
		
		setAction(RavenDialog.SAVE_BUTTON, getSaveAction());
	}

	/**
	 * Creates a new dialog to save the specified kind of data from the repository.
	 * @param owner the parent window of this dialog
	 * @param title the title of the dialog
	 * @param defaultDirectory the default directory which is used for the file system view
	 * @param dataClass the class indicating the type of data
	 */
	protected SaveDataDialog(Window owner, String title, File defaultDirectory, Class<? extends Data> dataClass) {
		super(owner, title, RavenDialog.SAVE_CANCEL_LAYOUT);

		isDataChooserVisible = true;
		getContentPane().add(createMainPanel(defaultDirectory));
		setLocationByPlatform(true);
		pack();
		dataComboBox.addDataClass(dataClass);
		
		setAction(RavenDialog.SAVE_BUTTON, getSaveAction());
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Abstract Protected Methods															*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the {@code Action} which will be performed whenever the user clicks the save button.
	 * @return the save action
	 */
	protected abstract Action getSaveAction();
	
	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the currently selected file
	 * @return the currently selected file
	 */
	protected File getSelectedFile() {
		BasicFileChooserUI ui = (BasicFileChooserUI) chooser.getUI();
		return new File(chooser.getCurrentDirectory(), ui.getFileName());
	}
	
	/**
	 * Returns the currently selected file filter
	 * @return the currently selected file filter
	 */
	protected FileFilter getSelectedFileFilter() {
		return chooser.getFileFilter();
	}
	
	/**
	 * Returns the name for the data object entered by the user
	 * @return the name for the data object
	 */
	protected NameDataPair getNameDataPair() {
			return ((NameDataPair) dataComboBox.getSelectedItem());
	}
	
	/**
	 * Sets the label for the data name textfield
	 * @param text the label of the textfield
	 */	
	protected void setDataNameLabel(String text) {
		dataNameLabel.setText(text+":");
		Dimension d = dataNameLabel.getPreferredSize();
		dataNameLabel.setPreferredSize(new Dimension(d.width + 5, d.height));
	}
	
	/**
	 * Sets the file filter which can be used by this dialog.
	 * @param fileFilters a list of available file filters
	 */
	protected void setFileFilters(FileFilter... fileFilters) {
		chooser.resetChoosableFileFilters();
		for(FileFilter filter: fileFilters) {
			chooser.addChoosableFileFilter(filter);
		}
	}
	
	/*------------------------------------------------------------------------------------------*/
	/* 		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the panel which displays the selection of an automaton from the repository.
	 * @param defaultDirectory the default directory which is used for the file system view
	 * @return the main panel
	 */
	private JPanel createMainPanel(File defaultDirectory) {
		chooser = new JFileChooser();
		FileSystemView view = new RavenDirectoryFileSystemView(new File(RavenMain.SAVE_DIR),
				defaultDirectory);
		chooser.setFileSystemView(view);
		chooser.setCurrentDirectory(view.getDefaultDirectory());
		
		chooser.setControlButtonsAreShown(false);
		chooser.setAcceptAllFileFilterUsed(false);
		
		JPanel mainPanel = new JPanel(new BorderLayout());
		mainPanel.add(chooser, BorderLayout.CENTER);

		if(isDataChooserVisible) {
			dataNameLabel = new JLabel();
			dataComboBox = new DataComboBox(Repository.getRepository());
			
			// sets the name of the item to be saved as default filename
			dataComboBox.addActionListener(new ActionListener() {			
				@Override
				public void actionPerformed(ActionEvent e) {
					chooser.setSelectedFile(new File(((NameDataPair) dataComboBox.getSelectedItem()).getName()));
				}
			});
			dataNameLabel.setLabelFor(dataComboBox);
			
			JPanel dataNamePanel = new JPanel();
			dataNamePanel.setLayout(new BoxLayout(dataNamePanel, BoxLayout.LINE_AXIS));
			dataNamePanel.setBorder(new EmptyBorder(12, 12, 0, 11));
			dataNamePanel.add(dataNameLabel);
			dataNamePanel.add(dataComboBox);

			mainPanel.add(dataNamePanel, BorderLayout.NORTH);
		}

		return mainPanel;
	}
	
}
