/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.gui;

import java.awt.Window;
import java.awt.event.ActionEvent;
import java.io.File;

import javax.swing.AbstractAction;
import javax.swing.Action;

import de.uni_due.inf.ti.raven.actions.SaveAction;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Decomposition;
import de.uni_due.inf.ti.raven.data.NameDataPair;
import de.uni_due.inf.ti.raven.io.CospanDecompositionFileWriter;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * This class provides a simple mechanism for the user to save a decomposition to a file.
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.19 $
 *
 * @see CospanDecomposition
 * @see CospanDecompositionFileWriter
 *
 */
public class SaveDecompositionDialog extends SaveDataDialog {
	
	private static final long serialVersionUID = 3670659781434810975L;
	
	private static Window dialog;

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Brings up a dialog to save decompositions from the repository.
	 * @param owner the parent window of this dialog
	 */
	public static void showDialog(Window owner) {
		dialog = new SaveDecompositionDialog(owner);
		dialog.setLocationRelativeTo(owner);
		dialog.setVisible(true);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new dialog to save decompositions from the repository.
	 * @param owner the parent window of this dialog
	 */
	private SaveDecompositionDialog(Window owner) {
		super(owner, messages.getString(GuiMessages.DECOMPOSITION_SAVE_DIALOG_TITLE_KEY), 
				new File(RavenMain.DECOMPOSITION_SAVE_DIR), Decomposition.class);

		setDataNameLabel(messages.getString(GuiMessages.DECOMPOSITION_NAME_LABEL_KEY));
		setFileFilters(RavenFileFilter.DEFAULT_DECOMPOSITION_FILE_FILTER);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	protected Action getSaveAction() {
		return new AbstractAction(messages.getString(GuiMessages.SAVE_DECOMPOSITION_ACTION_KEY)) {

			private static final long serialVersionUID = 197458776008048957L;

			@Override
			public void actionPerformed(ActionEvent e) {
				NameDataPair nameDataPair = getNameDataPair();
				if(nameDataPair==null) {
					showErrorMessage(GuiMessages.NO_DECOMPOSITION_SELECTED_TITLE_EXCEPTION_KEY,
							GuiMessages.NO_DECOMPOSITION_SELECTED_TEXT_EXCEPTION_KEY);
					return;
				}

				String filename = getSelectedFile().getAbsolutePath();
				filename = filename.trim();
				if(filename==null || filename.isEmpty()) {
					showErrorMessage(GuiMessages.EMPTY_FILENAME_TITLE_EXCEPTION_KEY,
							GuiMessages.EMPTY_FILENAME_TEXT_EXCEPTION_KEY);
					return;
				}
				
				filename = checkFileExtension(filename, RavenMain.DECOMPOSITION_FILE_EXTENSION);				
				if(isFileExisting(filename)){
					int value = Utilities.showConfirmDialog(
							SaveDecompositionDialog.this, 
							GuiMessages.FILE_ALREADY_EXISTING_TITLE_KEY,
							GuiMessages.FILE_ALREADY_EXISTING_TEXT_KEY, filename);
					if(value==RavenDialog.NO_OPTION) {
						return;
					}
				}

				CospanDecomposition decomp = (CospanDecomposition) nameDataPair.getData();
				new SaveAction(decomp, filename).execute();
				showInformationMessage(GuiMessages.DECOMPOSITION_SAVED_TITLE_KEY,
						GuiMessages.DECOMPOSITION_SAVED_TEXT_KEY, filename);
	  			dialog.dispose();
			}
		};
	}

}
