/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui;

import java.awt.Window;
import java.awt.event.ActionEvent;
import java.io.File;
import java.io.FileNotFoundException;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.filechooser.FileFilter;

import de.uni_due.inf.ti.raven.actions.SaveAction;
import de.uni_due.inf.ti.raven.actions.SaveAction.GraphPersistenceType;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.NameDataPair;
import de.uni_due.inf.ti.raven.io.GXLWriter;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * This class provides a simple mechanism for the user to save a graph to a file.
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.28 $
 *
 * @see Graph
 * @see GXLWriter
 *
 */
public class SaveGraphDialog extends SaveDataDialog {

	private static final long serialVersionUID = 3670659781434810975L;
	
	private static Window dialog;

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Brings up a dialog to save graphs from the repository.
	 * @param owner the parent window of this dialog
	 */
	public static void showDialog(Window owner) {
		showDialog(owner, RavenFileFilter.GRAPH_GXL_FILE_FILTER,
				RavenFileFilter.GRAPH_SVG_FILE_FILTER);
	}

	/**
	 * Brings up a dialog to save graphs from the repository to file types depending on the given
	 * file filters.
	 * @param owner the parent window of this dialog
	 * @param fileFilters the list of available file filters 
	 */
	public static void showDialog(Window owner, FileFilter... fileFilters) {
		dialog = new SaveGraphDialog(owner, fileFilters);
		dialog.setLocationRelativeTo(owner);
		dialog.setVisible(true);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new dialog to save graphs from the repository.
	 * @param owner the parent window of this dialog
	 */
	private SaveGraphDialog(Window owner, FileFilter... fileFilters) {
		super(owner, messages.getString(GuiMessages.GRAPH_SAVE_DIALOG_TITLE_KEY), 
				new File(RavenMain.GRAPH_SAVE_DIR), Graph.class);

		setDataNameLabel(messages.getString(GuiMessages.GRAPH_NAME_LABEL_KEY));
		setFileFilters(fileFilters);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	protected Action getSaveAction() {
		return new AbstractAction(messages.getString(GuiMessages.SAVE_GRAPH_ACTION_KEY)) {

			private static final long serialVersionUID = 197458776008048957L;

			@Override
			public void actionPerformed(ActionEvent e) {
				NameDataPair nameDataPair = getNameDataPair();
				if(nameDataPair==null) {
					showErrorMessage(GuiMessages.NO_GRAPH_SELECTED_TITLE_EXCEPTION_KEY,
							GuiMessages.NO_GRAPH_SELECTED_TEXT_EXCEPTION_KEY);
					return;
				}

				String filename = getSelectedFile().getAbsolutePath();
				filename = filename.trim();
				if(filename==null || filename.isEmpty()) {
					showErrorMessage(GuiMessages.EMPTY_FILENAME_TITLE_EXCEPTION_KEY,
							GuiMessages.EMPTY_FILENAME_TEXT_EXCEPTION_KEY);
					return;
				}

				Graph graph = (Graph) nameDataPair.getData();
				GraphPersistenceType type = null;
				//If the user wants to save the graph to a GXL file
				if(getSelectedFileFilter()==RavenFileFilter.GRAPH_GXL_FILE_FILTER) {
					type = GraphPersistenceType.GXL_FORMAT;
					filename = checkFileExtension(filename, RavenMain.GRAPH_GXL_FILE_EXTENSION);
				}
				//If the user wants to save the graph to a SVG file
				else if(getSelectedFileFilter()==RavenFileFilter.GRAPH_SVG_FILE_FILTER) {
					type = GraphPersistenceType.SVG_FORMAT;
					filename = checkFileExtension(filename, RavenMain.SVG_FILE_EXTENSION);
				}
				
				//Try to save the selected graph to the specified file
				if (isFileExisting(filename)){
					int value = Utilities.showConfirmDialog(SaveGraphDialog.this, 
							GuiMessages.FILE_ALREADY_EXISTING_TITLE_KEY,
							GuiMessages.FILE_ALREADY_EXISTING_TEXT_KEY, filename);
					if(value==RavenDialog.NO_OPTION) {
						return;
					}
				}
				
				try {
					new SaveAction(type, graph, filename).execute();
					showInformationMessage(GuiMessages.GRAPH_SAVED_TITLE_KEY,
							GuiMessages.GRAPH_SAVED_TEXT_KEY, filename);
				} catch (FileNotFoundException ex) {
					showErrorMessage(GuiMessages.FILE_NOT_FOUND_TITLE_EXCEPTION_KEY,
							GuiMessages.FILE_NOT_FOUND_TEXT_EXCEPTION_KEY, filename);
				}
				finally {
					dialog.dispose();
				}
			}
		};
	}

}
