/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui;

import java.awt.FlowLayout;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.text.DecimalFormat;
import java.util.HashSet;

import javax.swing.AbstractAction;
import javax.swing.DefaultListModel;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JFormattedTextField;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.JToolBar;
import javax.swing.ListSelectionModel;
import javax.swing.SpringLayout;

import com.jgoodies.forms.builder.DefaultFormBuilder;
import com.jgoodies.forms.factories.Borders;
import com.jgoodies.forms.layout.FormLayout;

import de.uni_due.inf.ti.raven.actions.SignatureAction;
import de.uni_due.inf.ti.raven.cli.Console;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.Operation.Connect;
import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * This class provides a dialog to create a new signature.
 * @author Christoph Blume
 * @version $Revision: 1.32 $
 *
 */
public class SignatureCreationDialog extends RavenDialog {

	private class AddOperationAction extends AbstractAction {

		private static final long serialVersionUID = 8494547536935118434L;

		private Operation operation;

		public AddOperationAction(Operation operation) {
			super(messages.getString(GuiMessages.ADD_OPERATION_KEY, 
					operation.toString()));

			this.operation = operation;
		}

		@Override
		public void actionPerformed(ActionEvent e) {
			operationListModel.addElement(operation);
			popup.remove((JComponent) e.getSource());
		}

	}

	/**
	 * This class provides a dialog to create a new label which can be used for 
	 * a new CONNECT-operation.
	 * @author Christoph Blume
	 * @version $Revision: 1.32 $
	 *
	 * @see Operation
	 */
	private static class NewLabelDialog extends RavenDialog {

		private static final long serialVersionUID = -1735339928124098876L;

		private static NewLabelDialog dialog;

		private final JTextField LABEL_FIELD;
		private final JTextField ARITY_FIELD;

		/*------------------------------------------------------------------------------------------*/
		/*		Static Methods																		*/
		/*------------------------------------------------------------------------------------------*/

		/**
		 * Shows the dialog for the creation of new labels.
		 * @param owner the parent window of the dialog
		 */
		public static void showDialog(Window owner) {
			dialog = new NewLabelDialog(owner);			
			dialog.setLocationRelativeTo(owner);
			dialog.setVisible(true);
		}

		/*------------------------------------------------------------------------------------------*/
		/*		Constructors																		*/
		/*------------------------------------------------------------------------------------------*/

		/**
		 * Creates a new dialog for the creation of new labels.
		 * @param owner the parent window of this dialog
		 */
		private NewLabelDialog(Window owner) {
			super(owner, messages.getString(GuiMessages.NEW_LABEL_DIALOG_TITLE_KEY), RavenDialog.ACCEPT_CANCEL_LAYOUT);

			String labelSuffix = ": ";

			JLabel labelLabel = new JLabel(
					messages.getString(GuiMessages.LABEL_LABEL_KEY)+labelSuffix);
			LABEL_FIELD = new JTextField(10);

			JLabel arityLabel = new JLabel(
					messages.getString(GuiMessages.ARITY_LABEL_KEY)+labelSuffix);
			ARITY_FIELD = new JFormattedTextField(new DecimalFormat("0"));
			ARITY_FIELD.setColumns(10);

			SpringLayout layout = new SpringLayout();
			JPanel panel = new JPanel();
			panel.setBackground(MainWindow.BACKGROUND_COLOR);
			panel.add(labelLabel);
			panel.add(LABEL_FIELD);
			panel.add(arityLabel);
			panel.add(ARITY_FIELD);

			Utilities.makeGridLayout(layout, panel, 4, 10, 10);

			AbstractAction acceptAction = new AbstractAction(
					messages.getString(GuiMessages.ACCEPT_LABEL_KEY)) {

				private static final long serialVersionUID = 321250982627352495L;

				@Override
				public void actionPerformed(ActionEvent e) {
					String label = LABEL_FIELD.getText();
					String arityString = ARITY_FIELD.getText();

					if (!Operation.isLabelValid(label)) {
						showErrorMessage(GuiMessages.INVALID_LABEL_TITLE_EXCEPTION_KEY,
								GuiMessages.INVALID_LABEL_TEXT_EXCEPTION_KEY, label);
						return;
					}
					if(label.isEmpty()) {
						showErrorMessage(GuiMessages.EMPTY_INPUTFIELD_TITLE_EXCEPTION_KEY,
								GuiMessages.EMPTY_INPUTFIELD_TEXT_EXCEPTION_KEY,
								messages.getString(GuiMessages.LABEL_LABEL_KEY));
						return;
					}
					if(arityString.isEmpty()) {
						showErrorMessage(GuiMessages.EMPTY_INPUTFIELD_TITLE_EXCEPTION_KEY,
								GuiMessages.EMPTY_INPUTFIELD_TEXT_EXCEPTION_KEY,
								messages.getString(GuiMessages.ARITY_LABEL_KEY));
						return;
					}

					int arity = Integer.parseInt(arityString);
					if(arity<0) {
						showErrorMessage(GuiMessages.VALUE_MUST_BE_NON_NEGATIVE_TITLE_EXCEPTION_KEY,
								GuiMessages.VALUE_MUST_BE_NON_NEGATIVE_TEXT_EXCEPTION_KEY,
								messages.getString(GuiMessages.ARITY_LABEL_KEY));
						return;
					}

					operationListModel.addElement(Operation.getConnect(label, arity));
					NewLabelDialog.this.dispose();
				}
			};

			setMainPane(panel);
			setAction(ACCEPT_BUTTON, acceptAction);
			pack();
		}

	}

	/**
	 * 
	 * @author Christoph Blume
	 *
	 */
	private static final long serialVersionUID = -7550248945827525277L;

	private static SignatureCreationDialog dialog;
	private static GuiMessages messages;
	private static DefaultListModel<Operation> operationListModel;

	private JTextField nameField;
	private JPopupMenu popup;

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Shows the dialog for the creation of new signatures.
	 * @param owner the parent window of the dialog
	 */
	public static void showDialog(Window owner) {
		messages = GuiMessages.getMessages();
		dialog = new SignatureCreationDialog(owner);

		dialog.setLocationRelativeTo(owner);
		dialog.setVisible(true);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new dialog for the creation of new signatures.
	 * @param owner the parent window of this dialog
	 */
	private SignatureCreationDialog(Window owner) {
		super(owner, messages.getString(GuiMessages.SIGNATURE_CREATION_DIALOG_TITLE_KEY), 
				RavenDialog.ACCEPT_CANCEL_LAYOUT);
		getContentPane().setBackground(MainWindow.BACKGROUND_COLOR);
		AbstractAction acceptAction = new AbstractAction(
				messages.getString(GuiMessages.ACCEPT_LABEL_KEY)) {

			private static final long serialVersionUID = 2418165190397918534L;

			@Override
			public void actionPerformed(ActionEvent e) {
				String name = nameField.getText();
				if(name==null || name.isEmpty()) {
					showErrorMessage(GuiMessages.EMPTY_INPUTFIELD_TITLE_EXCEPTION_KEY,
							GuiMessages.EMPTY_INPUTFIELD_TEXT_EXCEPTION_KEY,
							messages.getString(GuiMessages.NAME_LABEL_KEY));
					return;
				}
				if(!Repository.isNameValid(name)){
					showErrorMessage(GuiMessages.INVALID_NAME_TITLE_EXCEPTION_KEY,
							GuiMessages.INVALID_NAME_TEXT_EXCEPTION_KEY, name);
					return;
				}

				if (Repository.getRepository().contains(name)) {
					showInformationMessage(GuiMessages.NAME_ALREADY_IN_USE_TITLE_EXCEPTION_KEY,
							GuiMessages.NAME_ALREADY_IN_USE_TEXT_EXCEPTION_KEY);
					return;
				} 

				HashSet<Operation> operations = new HashSet<>();
				for(int i=0;i<operationListModel.size();i++) {
					operations.add(operationListModel.get(i));
				}
				new SignatureAction(name, operations).execute();
				SignatureCreationDialog.this.dispose();
			}

		};

		popup = new JPopupMenu();
		setMainPane(createMainPanel());
		setAction(ACCEPT_BUTTON, acceptAction);
		setResizable(false);
		pack();
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the panel which displays components for the creation of the signature.
	 * @return the main panel
	 */
	private JPanel createMainPanel() {
		DefaultFormBuilder builder = new DefaultFormBuilder(new FormLayout("pref, 3dlu, 80dlu:grow", ""));
		builder.border(Borders.DIALOG);
		
		String labelSuffix = ": ";
		JLabel nameLabel = new JLabel(
				messages.getString(GuiMessages.NAME_LABEL_KEY)+labelSuffix);
		nameLabel.setOpaque(false);
		nameField = new JTextField(15);
		nameField.setOpaque(false);

		builder.append(nameLabel, nameField);
		builder.appendSeparator(messages.getString(GuiMessages.OPERATIONS_LABEL_KEY));
		
		operationListModel = new DefaultListModel<>();
		final JList<Operation> operationList = new JList<>(operationListModel);
		operationList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

		popup.add(new AbstractAction(messages.getString(
				GuiMessages.ADD_OPERATION_KEY, Console.CONNECT_TOKEN)) {

			private static final long serialVersionUID = -899663186839016035L;

			@Override
			public void actionPerformed(ActionEvent e) {
				NewLabelDialog.showDialog(SignatureCreationDialog.this);
			}

		});
		popup.add(new AddOperationAction(Operation.getFuse()));

		operationListModel.addElement(Operation.getPerm());
		operationListModel.addElement(Operation.getRes());
		operationListModel.addElement(Operation.getTrans());
		operationListModel.addElement(Operation.getVertex());
		
		//Create Add/Remove Toolbar
		JToolBar tableToolBar = new JToolBar();
		tableToolBar.setOpaque(false);
		tableToolBar.setBackground(MainWindow.BACKGROUND_COLOR);
		tableToolBar.setFloatable(false);
		tableToolBar.setBorderPainted(false);
		tableToolBar.add(new AbstractAction("", 
				new ImageIcon(RavenMain.PLUS_ICON_FILENAME)) {

			private static final long serialVersionUID = 6480710296758476298L;

			@Override
			public void actionPerformed(ActionEvent e) {
				JButton button = (JButton) e.getSource();
				popup.show(button, button.getWidth()/2, button.getHeight()/2);
			}
		});
		tableToolBar.add(new AbstractAction("", 
				new ImageIcon(RavenMain.MINUS_ICON_FILENAME)) {

			private static final long serialVersionUID = 6691022816927260856L;

			@Override
			public void actionPerformed(ActionEvent e) {
				int index = operationList.getSelectedIndex();
				Operation operation = operationList.getSelectedValue();

				if(index!=-1) {
					operationListModel.remove(index);

					if(!(operation instanceof Connect)) {
						popup.insert(new AddOperationAction(operation), getListIndex(operation));
					}
				}
				else {
					showInformationMessage(GuiMessages.NO_OPERATION_SELECTED_TITLE_KEY,
							GuiMessages.NO_OPERATION_SELECTED_TEXT_KEY);
				}
			}

			private int getListIndex(Operation operation) {
				int offset = operationListModel.getSize();

				if(operation instanceof Operation.Fuse) {
					return offset-(offset-1);
				}
				if(operation instanceof Operation.Perm) {
					return offset-(offset-2);
				}
				if(operation instanceof Operation.Res) {
					return offset-(offset-3);
				}
				if(operation instanceof Operation.Trans) {
					return offset-(offset-4);
				}
				if(operation instanceof Operation.Vertex) {
					return offset-(offset-5);
				}				

				return 0;
			}

		});
		
		operationList.setBackground(MainWindow.BACKGROUND_COLOR);
		builder.append(new JScrollPane(operationList),3);
		builder.nextLine();
		
		JPanel toolBarPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
		toolBarPanel.setOpaque(false);
		toolBarPanel.add(tableToolBar);
		builder.append(toolBarPanel,3);
		
		return builder.getPanel();
	}
}
