/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui;

import javax.swing.JComboBox;
import javax.swing.JTextField;

import com.jgoodies.forms.builder.DefaultFormBuilder;

import de.uni_due.inf.ti.raven.actions.ComputationEvent;
import de.uni_due.inf.ti.raven.actions.ComputationListener;
import de.uni_due.inf.ti.raven.actions.SimulationComputationAction;
import de.uni_due.inf.ti.raven.actions.ComputationEvent.EventType;
import de.uni_due.inf.ti.raven.algorithms.Algorithm.SearchType;
import de.uni_due.inf.ti.raven.algorithms.SimulationAlgorithm;
import de.uni_due.inf.ti.raven.algorithms.SimulationAlgorithm.Implementations;
import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.data.automata.Automaton;

/**
 * This class implements a {@code ControlPanel} for the invocation of the different simulation 
 * computation algorithms implemented in Raven.
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.7 $
 * 
 */
public final class SimulationControlPanel extends ControlPanel {

	private static final long serialVersionUID = -4084627539102346116L;

	/**
	 * This class is used as a worker thread to perform the computation of the simulation relation. 
	 */
	private class SimulationWorkerThread extends WorkerThread {

		private String name;
		private SimulationComputationAction action;

		@Override
		protected Void doInBackground() throws Exception {
			// Get the parameter for the invariant checking algorithm
			Implementations algorithm = (Implementations) simulationAlgorithmBox.getSelectedItem();
			Automaton automaton = (Automaton) automatonBox.getSelectedData();

			// Store the names of the parameter (which will be used later on for the output)
			name = nameField.getText();
			if (Repository.getRepository().contains(name)) {
				Utilities.showWarningMessage(SimulationControlPanel.this, 
						GuiMessages.NAME_ALREADY_IN_USE_TITLE_EXCEPTION_KEY,
						GuiMessages.NAME_ALREADY_IN_USE_TEXT_EXCEPTION_KEY,
						name);
				cancel(true);
				fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_CANCELLED, "", 0, 0));
				return null;
			}
			
			//Check whether the name is valid
			if (name.isEmpty()) {
				Utilities.showWarningMessage(SimulationControlPanel.this,
						GuiMessages.EMPTY_INPUTFIELD_TITLE_EXCEPTION_KEY,
						GuiMessages.EMPTY_INPUTFIELD_TEXT_EXCEPTION_KEY,
						messages.getString(GuiMessages.NAME_LABEL_KEY));
				cancel(true);
				fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_CANCELLED, "", 0, 0));
				return null;
			}
			if(!Repository.isNameValid(name)) {
				Utilities.showWarningMessage(SimulationControlPanel.this,
						GuiMessages.INVALID_NAME_TITLE_EXCEPTION_KEY,
						GuiMessages.INVALID_NAME_TEXT_EXCEPTION_KEY,
						name);
				cancel(true);
				fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_CANCELLED, "", 0, 0));
				return null;
			}
			if(Repository.getRepository().contains(name)) {
				Utilities.showWarningMessage(SimulationControlPanel.this,
						GuiMessages.NAME_ALREADY_IN_USE_TITLE_EXCEPTION_KEY,
						GuiMessages.NAME_ALREADY_IN_USE_TEXT_EXCEPTION_KEY,
						name);
				cancel(true);
				fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_CANCELLED, "", 0, 0));
				return null;
			}
			if(automaton==null) {
				Utilities.showWarningMessage(SimulationControlPanel.this,
						GuiMessages.NO_AUTOMATON_SELECTED_TITLE_EXCEPTION_KEY,
						GuiMessages.NO_AUTOMATON_SELECTED_TEXT_EXCEPTION_KEY,
						messages.getString(GuiMessages.AUTOMATON_LABEL_KEY));
				cancel(true);
				fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_CANCELLED, "", 0, 0));
				return null;
			}

			// Create simulation computation action
			action = null;
			if (algorithm==Implementations.ForwardSimulationAlgorithm) {
				action = new SimulationComputationAction(SearchType.forward, automaton);
			}
			else if (algorithm==Implementations.BackwardSimulationAlgorithm) {
				action = new SimulationComputationAction(SearchType.backward, automaton);
			}

			// Register listeners
			for(ComputationListener listener : listeners) {
				action.addComputationListener(listener);
			}

			// Start computation
			action.execute();
			return null;
		}

		@Override
		protected void done() {
			//If the worker thread has been cancelled
			if(isCancelled()) {
				if(action!=null) {
					//also cancel the computation of the action
					action.cancelComputation();
				}
				
				action = null;
				name = null;
			}
			//else bring up message dialog to display result
			else {
				//TODO Alter this Panel (compute the simulation by SimulationComputationAction, check the simulation by SimulationCheckingAction)
				//boolean okay = (boolean) action.getResult();
				//if (okay) {
				//	JOptionPane.showMessageDialog(MainWindow.getWindow(), messages
				//			.getString(GuiMessages.SIMULATION_CREATED_TEXT_KEY), messages
				//			.getString(GuiMessages.SIMULATION_CREATED_TEXT_KEY),
				//			JOptionPane.INFORMATION_MESSAGE);
				//}
				//else {
				//	JOptionPane.showMessageDialog(MainWindow.getWindow(), messages
				//			.getString(GuiMessages.SIMULATION_NOT_CREATED_TEXT_KEY), messages
				//			.getString(GuiMessages.SIMULATION_NOT_CREATED_TEXT_KEY),
				//			JOptionPane.INFORMATION_MESSAGE);
				//}
			}
		}

	}

	/** Name Text Field Label (used for Testing) */
	protected static final String NAME_FIELD_NAME = "name";
	
	/** Decomposition Algorithm Box Label (used for Testing) */
	protected static final String ALGORITHM_BOX_NAME = "decompositionAlgorithm";
	
	/** Automaton Box Label (used for Testing) */
	protected static final String AUTOMATON_BOX_NAME = "automatonBox";
	
	private JTextField nameField;
	private JComboBox<SimulationAlgorithm.Implementations> simulationAlgorithmBox;
	private DataComboBox automatonBox;

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public void setEnabled(boolean isEnabled) {
		nameField.setEnabled(isEnabled);
		simulationAlgorithmBox.setEnabled(isEnabled);
		automatonBox.setEnabled(isEnabled);
		
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	protected void initialize(DefaultFormBuilder builder) {
		Repository repo = Repository.getRepository();
		
		nameField = new JTextField(10);
		nameField.setName(NAME_FIELD_NAME);
		append(builder, GuiMessages.NAME_LABEL_KEY, nameField);

		simulationAlgorithmBox = new JComboBox<>(Implementations.values());
		simulationAlgorithmBox.setName(ALGORITHM_BOX_NAME);
		append(builder, GuiMessages.ALGORITHM_LABEL_KEY, simulationAlgorithmBox);

		automatonBox = new DataComboBox(repo);
		automatonBox.setName(AUTOMATON_BOX_NAME);
		automatonBox.addDataClass(Automaton.class);
		append(builder, GuiMessages.AUTOMATON_LABEL_KEY, automatonBox);

		// set Color
		nameField.setBackground(getBackground());
		simulationAlgorithmBox.setBackground(getBackground());
		automatonBox.setBackground(getBackground());

	}

	@Override
	protected WorkerThread getWorkerThread() {
		return new SimulationWorkerThread();
	}

}
