/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui;

import java.awt.BorderLayout;
import java.awt.CardLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;

import javax.swing.AbstractAction;
import javax.swing.DefaultListCellRenderer;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JToolBar;

import com.jgoodies.forms.builder.DefaultFormBuilder;
import com.jgoodies.forms.factories.Borders;
import com.jgoodies.forms.layout.FormLayout;

import de.uni_due.inf.ti.raven.actions.ComputationEvent;
import de.uni_due.inf.ti.raven.actions.ComputationEvent.EventType;
import de.uni_due.inf.ti.raven.actions.ComputationListener;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * This class provides a panel on which the user can choose and start the different 
 * techniques, algorithms and goals provided by Raven.
 * @author Christoph Blume
 * @version $Revision: 1.13 $
 *
 * @see ControlPanelFactory
 * @see ControlPanel
 */
public final class TechniquePanel extends JPanel {

	private static final long serialVersionUID = 5512343439393102987L;
	
	private static final String LAYOUT_SPECIFICATION = "left:max(40dlu;pref), 3dlu, 40dlu:grow";

	private static final GuiMessages messages = GuiMessages.getMessages();
	
	private static final String DECOMPOSITION_LABEL = "decomposition";
	private static final String EMPTINESS_LABEL = "emptiness";
	private static final String INVARIANT_CHECKING_TECHNIQUE_LABEL = "invariantChecking";
	private static final String LANGUAGE_INCLUSION_TECHNIQUE_LABEL = "languageInclusion";
	private static final String MEMBERSHIP_TECHNIQUE_LABEL = "membership";
	private static final String SIMULATION_TECHNIQUE_LABEL = "simulation";
	private static final String UNIVERSALITY_TECHNIQUE_LABEL = "universality";
	
	private static final String[] TECHNIQUES = new String[] {
		DECOMPOSITION_LABEL,
		EMPTINESS_LABEL, 
		INVARIANT_CHECKING_TECHNIQUE_LABEL,
		LANGUAGE_INCLUSION_TECHNIQUE_LABEL, 
		MEMBERSHIP_TECHNIQUE_LABEL,
		SIMULATION_TECHNIQUE_LABEL, 
		UNIVERSALITY_TECHNIQUE_LABEL
	};
	
	private JComboBox<String> techniqueBox;
	private ControlPanel decompositionPanel;
	private ControlPanel emptinessPanel;
	private ControlPanel invariantCheckingPanel;
	private ControlPanel languageInclusionPanel;
	private ControlPanel membershipPanel;
	private ControlPanel simulationPanel;
	private ControlPanel universalityPanel;

	private JButton startButton;

	private JButton stopButton;
	
	/**
	 * Creates the panel on which controls for the usage of Raven will be displayed.
	 */
	public TechniquePanel() {
		super(new BorderLayout());
		setOpaque(false);
		
		decompositionPanel = ControlPanelFactory.getDecompositionControlPanel();
		emptinessPanel = ControlPanelFactory.getEmptinessControlPanel();
		invariantCheckingPanel = ControlPanelFactory.getInvariantCheckingControlPanel();
		languageInclusionPanel = ControlPanelFactory.getLanguageInclusionControlPanel();
		membershipPanel = ControlPanelFactory.getMembershipControlPanel();
		simulationPanel = ControlPanelFactory.getSimulationControlPanel();
		universalityPanel = ControlPanelFactory.getUniversalityControlPanel();

		final JPanel centerPanel = new JPanel(new CardLayout());
		centerPanel.setOpaque(false);
		centerPanel.add(invariantCheckingPanel, INVARIANT_CHECKING_TECHNIQUE_LABEL);
		centerPanel.add(languageInclusionPanel, LANGUAGE_INCLUSION_TECHNIQUE_LABEL);
		centerPanel.add(membershipPanel, MEMBERSHIP_TECHNIQUE_LABEL);
		centerPanel.add(simulationPanel, SIMULATION_TECHNIQUE_LABEL);
		centerPanel.add(universalityPanel, UNIVERSALITY_TECHNIQUE_LABEL);
		centerPanel.add(emptinessPanel, EMPTINESS_LABEL);
		centerPanel.add(decompositionPanel, DECOMPOSITION_LABEL);
		
		FormLayout layout = new FormLayout(LAYOUT_SPECIFICATION, "");
		DefaultFormBuilder builder = new DefaultFormBuilder(layout);
		builder.border(Borders.DLU4);
		
		techniqueBox = new JComboBox<>(TECHNIQUES);
		techniqueBox.setBackground(MainWindow.BACKGROUND_COLOR);
		techniqueBox.setRenderer(new DefaultListCellRenderer() {

			private static final long serialVersionUID = 2290896130644751404L;

			@Override
			public Component getListCellRendererComponent(JList<?> list, Object value, int index, 
					boolean isSelected, boolean cellHasFocus) {
				return super.getListCellRendererComponent(list,
						messages.getString(GuiMessages.LABEL_PREFIX + value.toString()), index, 
						isSelected, cellHasFocus);
			}

		});
		techniqueBox.addItemListener(new ItemListener() {

			@Override
			public void itemStateChanged(ItemEvent e) {
				if (e.getStateChange() == ItemEvent.SELECTED) {
					String technique = e.getItem().toString();
					CardLayout layout = (CardLayout) (centerPanel.getLayout());
					layout.show(centerPanel, technique);
				}
			}
		});
		techniqueBox.setSelectedIndex(techniqueBox.getItemCount()-1);
		builder.append(messages.getString(GuiMessages.PROBLEM_LABEL_KEY), techniqueBox).setOpaque(false);
		builder.appendSeparator();

		JPanel algorithmPanel = builder.getPanel();
		algorithmPanel.setOpaque(false);
		
		JToolBar toolBar = new JToolBar();
		toolBar.setBorderPainted(false);
		toolBar.setOpaque(false);
		toolBar.setFloatable(false);
		toolBar.addSeparator(new Dimension(170,0));
		
		final RavenProgressBarPanel bar = new RavenProgressBarPanel();
		
		startButton = new JButton(
				new AbstractAction("", new ImageIcon(RavenMain.RUN_ICON_FILENAME)) {

					private static final long serialVersionUID = -4458661949935231754L;
					
					private ControlPanel panel;

					@Override
					public void actionPerformed(ActionEvent e) {

						((JButton)e.getSource()).setEnabled(false);
						// Start the chosen algorithm
						switch (techniqueBox.getSelectedItem().toString()) {
						case INVARIANT_CHECKING_TECHNIQUE_LABEL:
							panel = invariantCheckingPanel;
							break;

						case LANGUAGE_INCLUSION_TECHNIQUE_LABEL:
							panel = languageInclusionPanel;
							break;

						case MEMBERSHIP_TECHNIQUE_LABEL:
							panel = membershipPanel;
							break;

						case SIMULATION_TECHNIQUE_LABEL:
							panel = simulationPanel;
							break;

						case UNIVERSALITY_TECHNIQUE_LABEL:
							panel = universalityPanel;
							break;

						case EMPTINESS_LABEL:
							panel = emptinessPanel;
							break;

						case DECOMPOSITION_LABEL:
							panel = decompositionPanel;
							break;

						default:
							break;
						}
						
						setPanelEnabled(false);
						panel.addComputationListener(bar);
						panel.addComputationListener(new ComputationListener() {
							
							@Override
							public void computationAdvanded(ComputationEvent e) {
								if(e.getType()==EventType.COMPUTATION_FINISHED || 
										e.getType()==EventType.COMPUTATION_CANCELLED) {
									setPanelEnabled(true);
								}
							}
						});
						
						new Thread(new Runnable() {
							
							@Override
							public void run() {
								panel.startWorkerThread();
							}
						}).start();
					}


				});
		startButton.setOpaque(false);
		
		stopButton = new JButton(
				new AbstractAction("", new ImageIcon(RavenMain.STOP_ICON_FILENAME)) {

					private static final long serialVersionUID = 6025284437011235156L;

					@Override
					public void actionPerformed(ActionEvent e) {
						startButton.setEnabled(true);
						// Stop the chosen algorithm
						switch (techniqueBox.getSelectedItem().toString()) {
						case INVARIANT_CHECKING_TECHNIQUE_LABEL:
							invariantCheckingPanel.stopWorkerThread();
							break;

						case LANGUAGE_INCLUSION_TECHNIQUE_LABEL:
							languageInclusionPanel.stopWorkerThread();
							break;

						case MEMBERSHIP_TECHNIQUE_LABEL:
							membershipPanel.stopWorkerThread();
							break;

						case SIMULATION_TECHNIQUE_LABEL:
							simulationPanel.stopWorkerThread();
							break;

						case UNIVERSALITY_TECHNIQUE_LABEL:
							universalityPanel.stopWorkerThread();
							break;

						case EMPTINESS_LABEL:
							emptinessPanel.stopWorkerThread();
							break;

						case DECOMPOSITION_LABEL:
							decompositionPanel.stopWorkerThread();
							break;

						default:
							break;
						}
					}
					
				});
		stopButton.setOpaque(false);
		
		toolBar.add(startButton);
		toolBar.addSeparator(new Dimension(5,0));
		//TODO Re-activate stop button
//		toolBar.add(stopButton);
		toolBar.addSeparator(new Dimension(5,0));
		toolBar.add(bar);
		
		add(algorithmPanel, BorderLayout.NORTH);
		add(centerPanel, BorderLayout.CENTER);
		add(toolBar, BorderLayout.SOUTH);
		Utilities.setCompoundBorder(this, messages.getString(GuiMessages.TECHNIQUE_PANEL_LABEL));
	}

	private void setPanelEnabled(boolean isEnabled) {		
		techniqueBox.setEnabled(isEnabled);
		decompositionPanel.setEnabled(isEnabled);
		emptinessPanel.setEnabled(isEnabled);
		invariantCheckingPanel.setEnabled(isEnabled);
		languageInclusionPanel.setEnabled(isEnabled);
		membershipPanel.setEnabled(isEnabled);
		simulationPanel.setEnabled(isEnabled);
		universalityPanel.setEnabled(isEnabled);
		startButton.setEnabled(isEnabled);
	}
	
}
