/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui;

import java.awt.Component;
import java.awt.FlowLayout;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;

import javax.swing.ButtonGroup;
import javax.swing.DefaultListCellRenderer;
import javax.swing.JComboBox;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JRadioButton;

import com.jgoodies.forms.builder.DefaultFormBuilder;

import de.uni_due.inf.ti.raven.actions.AntichainUniversalityAction;
import de.uni_due.inf.ti.raven.actions.ComputationEvent;
import de.uni_due.inf.ti.raven.actions.ComputationListener;
import de.uni_due.inf.ti.raven.actions.SimulationAntichainUniversalityAction;
import de.uni_due.inf.ti.raven.actions.UniversalityAction;
import de.uni_due.inf.ti.raven.actions.ComputationEvent.EventType;
import de.uni_due.inf.ti.raven.algorithms.Algorithm.SearchType;
import de.uni_due.inf.ti.raven.algorithms.UniversalityAlgorithm.Implementations;
import de.uni_due.inf.ti.raven.data.CospanCounterExample;
import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.data.automata.Automaton;

/**
 * This class implements a {@code ControlPanel} for the invocation of the different universality 
 * checking algorithms implemented in Raven.
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.10 $
 * 
 */
public final class UniversalityControlPanel extends ControlPanel {

	private static final long serialVersionUID = -5130697822106444828L;

	/**
	 * This class is used as a worker thread to perform the universality check. 
	 */
	private class UniversalityWorkerThread extends WorkerThread {

		private String automatonName;
		private UniversalityAction action;

		@Override
		protected Void doInBackground() throws Exception {
			// Get the parameter for the invariant checking algorithm
			Automaton automaton = (Automaton) automatonBox.getSelectedData();

			// Store the names of the parameter (which will be used later on for the output)
			automatonName = automatonBox.getSelectedName();

			if(automaton==null) {
				Utilities.showWarningMessage(UniversalityControlPanel.this,
						GuiMessages.NO_AUTOMATON_SELECTED_TITLE_EXCEPTION_KEY,
						GuiMessages.NO_AUTOMATON_SELECTED_TEXT_EXCEPTION_KEY,
						messages.getString(GuiMessages.AUTOMATON_LABEL_KEY));
				cancel(true);
				fireComputationEvent(new ComputationEvent(this, EventType.COMPUTATION_CANCELLED, "", 0, 0));
				return null;
			}
			
			// Determine the search type to be used (either forward or backward)
			SearchType type;
			if (forwardButton.isSelected()) {
				type = SearchType.forward;
			}
			else {
				type = SearchType.backward;
			}

			// Determine whether to compute a counter-example (if the result is negative)
			boolean computeCounterEx;
			if (computeCE.isSelected()) {
				computeCounterEx = true;
			}
			else {
				computeCounterEx = false;
			}

			// Create universality action
			action = null;
			Implementations algorithm = (Implementations) universalityAlgorithmBox.getSelectedItem();
			if (algorithm == Implementations.AntichainUniversality) {
				action = new AntichainUniversalityAction(type, computeCounterEx, automaton);
			} 
			else if (algorithm == Implementations.SimulationAntichainUniversality) {
				action = new SimulationAntichainUniversalityAction(type, computeCounterEx, automaton);
			}

			// Register listeners
			for(ComputationListener listener : listeners) {
				action.addComputationListener(listener);
			}

			// Start computation
			action.execute();
			return null;
		}

		@Override
		protected void done() {
			//If the worker thread has been cancelled
			if(isCancelled()) {
				if(action!=null) {
					//also cancel the computation of the action
					action.cancelComputation();
				}
				
				action = null;
				automatonName = null;
			}
			//else bring up message dialog to display result
			else {
				boolean isIncluded = action.getResult();
				
				if (isIncluded) {
					Utilities.showInformationMessage(UniversalityControlPanel.this,
							GuiMessages.UNIVERSALITY_DOES_HOLD_DIALOG_TITLE_KEY, 
							GuiMessages.UNIVERSALITY_DOES_HOLD_DIALOG_TEXT_KEY,
							automatonName);
				} else {
					CospanCounterExample counterExample = action.getCounterExample();
					
					Utilities.showInformationMessage(UniversalityControlPanel.this,
							GuiMessages.UNIVERSALITY_DOES_NOT_HOLD_DIALOG_TITLE_KEY,
							GuiMessages.UNIVERSALITY_DOES_NOT_HOLD_DIALOG_TEXT_KEY,
							automatonName);
					
					if (counterExample != null) {
						CounterExampleWindow.getWindow().showCounterExampleChooser(
								counterExample.getCospan(), 
								messages.getString(GuiMessages.COUNTEREXAMPLE_LABEL_KEY));
					}
				}
			}
		}

	}

	/** Algorithm Box Label (used for Testing) */
	protected static final String ALGORITHM_BOX_NAME = "algorithmBox";
	
	/** Automaton Box Label (used for Testing) */
	protected static final String AUTOMATON_BOX_NAME = "secondAutomatonBox";

	/** Forward Search Radio Button Label (used for Testing) */
	protected static final String FORWARD_SEARCH_RADIO_BUTTON_NAME = "forwardSearch";

	/** Backward Search Radio Button Label (used for Testing) */
	protected static final String BACKWARD_SEARCH_RADIO_BUTTON_NAME = "backwardSearch";

	/** Compute Counter-Example Radio Button Label (used for Testing) */
	protected static final String COMPUTE_COUNTER_EXAMPLE_RADIO_BUTTON_NAME = "computeCounterExampleBox";

	/** Do Not Compute Counter-Example Radio Button Label (used for Testing) */
	protected static final String DO_NOT_COMPUTE_COUNTER_EXAMPLE_RADIO_BUTTON_NAME = "doNotComputeCounterExampleBox";
	
	private JComboBox<Implementations> universalityAlgorithmBox;
	private DataComboBox automatonBox;
	private JRadioButton forwardButton;
	private JRadioButton backwardButton;
	private JRadioButton computeCE;
	private JRadioButton doNotComputeCE;

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public void setEnabled(boolean isEnabled) {
		universalityAlgorithmBox.setEnabled(isEnabled);
		automatonBox.setEnabled(isEnabled);
		forwardButton.setEnabled(isEnabled);
		backwardButton.setEnabled(isEnabled);
		computeCE.setEnabled(isEnabled);
		doNotComputeCE.setEnabled(isEnabled);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	protected void initialize(DefaultFormBuilder builder) {
		Repository repo = Repository.getRepository();

		universalityAlgorithmBox = new JComboBox<>(Implementations.values());
		universalityAlgorithmBox.setName(ALGORITHM_BOX_NAME);
		universalityAlgorithmBox.setRenderer(new DefaultListCellRenderer() {

			private static final long serialVersionUID = 48576265719193199L;

			@Override
			public Component getListCellRendererComponent(JList<?> list,
					Object value, int index, boolean isSelected,
					boolean cellHasFocus) {
				return super.getListCellRendererComponent(
						list, messages.getString(GuiMessages.LABEL_PREFIX + value.toString()), 
						index, isSelected, cellHasFocus);
			}
		});
		universalityAlgorithmBox.addItemListener(new ItemListener() {
			
			@Override
			public void itemStateChanged(ItemEvent e) {
				boolean setEnabled = e.getItem().equals(
						Implementations.AntichainUniversality);
				
				forwardButton.setEnabled(setEnabled);
				backwardButton.setEnabled(setEnabled);
			}
		});
		
		append(builder, GuiMessages.ALGORITHM_LABEL_KEY, universalityAlgorithmBox);

		automatonBox = new DataComboBox(repo);
		automatonBox.setName(AUTOMATON_BOX_NAME);
		automatonBox.addDataClass(Automaton.class);
		append(builder, GuiMessages.AUTOMATON_LABEL_KEY, automatonBox);


		forwardButton = new JRadioButton(messages.getString(GuiMessages.FORWARD_LABEL_KEY));
		forwardButton.setName(FORWARD_SEARCH_RADIO_BUTTON_NAME);
		forwardButton.setOpaque(false);
		forwardButton.setSelected(true);
		
		backwardButton = new JRadioButton(messages.getString(GuiMessages.BACKWARD_LABEL_KEY));
		backwardButton.setName(BACKWARD_SEARCH_RADIO_BUTTON_NAME);
		backwardButton.setOpaque(false);
		
		ButtonGroup searchTypeGroup = new ButtonGroup();
		searchTypeGroup.add(forwardButton);
		searchTypeGroup.add(backwardButton);

		JPanel searchTypePanel = new JPanel(new FlowLayout(FlowLayout.LEFT, 0, 0));
		searchTypePanel.setOpaque(false);
		searchTypePanel.add(forwardButton);
		searchTypePanel.add(backwardButton);
		append(builder, GuiMessages.SEARCHTYPE_LABEL_KEY, searchTypePanel);

		computeCE = new JRadioButton(messages.getString(GuiMessages.YES_LABEL_KEY));
		computeCE.setName(COMPUTE_COUNTER_EXAMPLE_RADIO_BUTTON_NAME);
		computeCE.setOpaque(false);
		computeCE.setSelected(true);
		
		doNotComputeCE = new JRadioButton(messages.getString(GuiMessages.NO_LABEL_KEY));
		doNotComputeCE.setName(DO_NOT_COMPUTE_COUNTER_EXAMPLE_RADIO_BUTTON_NAME);
		doNotComputeCE.setOpaque(false);
		
		ButtonGroup computeCounterExample = new ButtonGroup();
		computeCounterExample.add(computeCE);
		computeCounterExample.add(doNotComputeCE);

		JPanel counterExamplePanel = new JPanel(new FlowLayout(FlowLayout.LEFT, 0, 0));
		counterExamplePanel.setOpaque(false);
		counterExamplePanel.add(computeCE);
		counterExamplePanel.add(doNotComputeCE);
		append(builder, GuiMessages.COMPUTE_COUNTER_EXAMPLE_LABEL_KEY, counterExamplePanel);

		// set Color
		universalityAlgorithmBox.setBackground(getBackground());
		automatonBox.setBackground(getBackground());
		forwardButton.setBackground(getBackground());
		backwardButton.setBackground(getBackground());
		computeCE.setBackground(getBackground());
		doNotComputeCE.setBackground(getBackground());
	}

	@Override
	protected WorkerThread getWorkerThread() {
		return new UniversalityWorkerThread();
	}

}
