/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.gui;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Container;
import java.awt.event.ActionEvent;
import java.util.LinkedList;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.BorderFactory;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.Spring;
import javax.swing.SpringLayout;
import javax.swing.border.Border;

import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * A collection of utility methods for the graphical user interface of Raven
 * @author Christoph Blume
 * @version $Revision: 1.15 $
 *
 */
public final class Utilities {
	
	/**
	 * This class implements a standard dialog box that prompts users for a value 
	 * or informs them of something.
	 */
	private static class MessageDialog extends RavenDialog {

		private static final long serialVersionUID = 2557342959488307795L;

		/**
		 * Creates a new {@code MessageDialog}-object to display messages to the user.
		 * @param parent the parent component of this dialog
		 * @param icon the icon to be used
		 * @param titleKey the key indicating the title string
		 * @param messageKey the key indicating the message string
		 * @param parameter parameters which are used for the message string
		 */
		protected MessageDialog(Component parent, Icon icon, String titleKey, String messageKey, 
				String... parameter) {
			this(parent, icon, titleKey, RavenDialog.CLOSE_LAYOUT, messageKey, parameter);
		}
		
		/**
		 * Creates a new {@code MessageDialog}-object to display messages to the user.
		 * @param parent the parent component of this dialog
		 * @param icon the icon to be used
		 * @param titleKey the key indicating the title string
		 * @param buttonLayout the button layout of the dialog
		 * @param messageKey the key indicating the message string
		 * @param parameter parameters which are used for the message string
		 */
		protected MessageDialog(Component parent, Icon icon, String titleKey, 
				int buttonLayout, String messageKey, String... parameter) {
			super(parent, messages.getString(titleKey), buttonLayout);

			JLabel messageLabel = new JLabel(
					convertToHTMLString(messages.getString(messageKey, parameter)), 
					icon, JLabel.LEFT);
			messageLabel.setIconTextGap(10);
			
			JPanel mainPane = new JPanel(new BorderLayout());
			mainPane.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));
			mainPane.add(messageLabel, BorderLayout.CENTER);
			setMainPane(mainPane);
			setResizable(false);
			pack();
			setLocationRelativeTo(parent);
		}
		
	}
	
	private static final GuiMessages messages = GuiMessages.getMessages();
	private static int value = RavenDialog.CANCEL_OPTION;
	
	/**
	 * This constructor will do nothing since this class should not be implement any object.
	 */
	private Utilities() {
		//Do nothing
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Static Methods																*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Converts the string {@code text} to an HTML string, i.e. the string is surrounded with 
	 * an &lt;html&gt;-tag and all line breaks are replaced by &lt;br&gt;-tags.
	 * @param text the string which is to be converted
	 * @return the converted string
	 */
	public static String convertToHTMLString(String text) {
		return "<html>"+text.replaceAll("\n", "<br>")+"</html>";
	}
	
	/**
	 * Brings up a message dialog showing a warning message.
	 * @param parent the parent component on whose frame the message dialog will be displayed
	 * @param titleKey the key indicating the title string
	 * @param messageKey the key indicating the message string
	 * @param parameter additional parameter for the message string
	 */
	public static void showWarningMessage(Component parent, String titleKey, String messageKey, 
			String... parameter) {
		Icon icon = new ImageIcon(RavenMain.DIALOG_WARNING_ICON_FILENAME);
		MessageDialog dialog = new MessageDialog(parent, icon, titleKey, messageKey, parameter);
		dialog.setVisible(true);
	}
	
	/**
	 * Brings up a message dialog showing a error message.
	 * @param parent the parent component on whose frame the message dialog will be displayed 
	 * @param titleKey the key indicating the title string
	 * @param messageKey the key indicating the message string
	 * @param parameter additional parameter for the message string
	 */
	public static void showErrorMessage(Component parent, String titleKey, String messageKey, 
			String... parameter) {
		Icon icon = new ImageIcon(RavenMain.DIALOG_ERROR_ICON_FILENAME);
		MessageDialog dialog = new MessageDialog(parent, icon, titleKey, messageKey, parameter);
		dialog.setVisible(true);
	}
	
	/**
	 * Brings up a message dialog showing a information message.
	 * @param parent the parent component on whose frame the message dialog will be displayed
	 * @param titleKey the key indicating the title string
	 * @param messageKey the key indicating the message string
	 * @param parameter additional parameter for the message string
	 */
	public static void showInformationMessage(Component parent, String titleKey, String messageKey, 
			String... parameter) {
		Icon icon = new ImageIcon(RavenMain.DIALOG_INFORMATION_ICON_FILENAME);
		MessageDialog dialog = new MessageDialog(parent, icon, titleKey, messageKey, parameter);
		dialog.setVisible(true);
	}
	
	/**
	 * Brings up a message dialog showing a information message.
	 * @param parent the parent component on whose frame the message dialog will be displayed
	 * @param titleKey the key indicating the title string
	 * @param messageKey the key indicating the message string
	 * @param parameter additional parameter for the message string
	 * @return indicating the option chosen by the user
	 */
	public static int showConfirmDialog(Component parent, String titleKey, String messageKey, 
			String... parameter) {
		Icon icon = new ImageIcon(RavenMain.DIALOG_QUESTION_ICON_FILENAME);
		final MessageDialog dialog = new MessageDialog(parent, icon, titleKey, 
				RavenDialog.YES_NO_LAYOUT, messageKey, parameter);
		dialog.setAction(RavenDialog.YES_BUTTON, new AbstractAction(
				messages.getString(GuiMessages.YES_LABEL_KEY)) {
			
			private static final long serialVersionUID = -1064230459341987373L;

			@Override
			public void actionPerformed(ActionEvent e) {
				value = RavenDialog.YES_OPTION;
				dialog.dispose();
			}
		});
		dialog.setAction(RavenDialog.NO_BUTTON, new AbstractAction(
				messages.getString(GuiMessages.NO_LABEL_KEY)) {
			
			private static final long serialVersionUID = -1064230459341987373L;

			@Override
			public void actionPerformed(ActionEvent e) {
				value = RavenDialog.NO_OPTION;
				dialog.dispose();
			}
		});
		
		dialog.setVisible(true);
		
		return value;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Static Methods															*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns the component with the name {@code name} which belongs to the container {@code root}
	 * (or one of its sub-containers).
	 * @param name the component's name
	 * @param root the root container in which the search for the component starts
	 * @return the wanted component or null if no component with the given name can be found
	 */
	protected static Component getComponentByName(String name, Container root) {
		List<Container> list = null;

		list = new LinkedList<>();
		list.add(root);
		while(!list.isEmpty()) {
			for (Component comp: list.remove(0).getComponents()) {
				if(name.equals(comp.getName())) {
					return comp;
				}
				else if(comp instanceof Container) {
					list.add((Container) comp);
				}
			}
		}

		return null;
	}
	
	/**
	 * Creates a new grid-based layout {@code layout} placing all components of the panel {@code panel}
	 * into {@code cols} different columns.
	 * @param layout the layout which is used for the laying out the panel's components
	 * @param panel the panel whose components will be laid out
	 * @param cols the number of columns of the grid layout
	 * @param xPadding the padding in x-direction
	 * @param yPadding the padding in y-direction
	 */
	protected static void makeGridLayout(SpringLayout layout, JPanel panel, 
			int cols, int xPadding, int yPadding) {
		Spring[] maxWidthSprings = null;
		Component comp = null;
		Component adjustmentComp = null;
		Component eastComp = null;
		Component southComp = null;
		SpringLayout.Constraints cons = null;
	
		if(cols<1) {
			throw new IllegalArgumentException();
		}
		else if(cols==1) {
			makeListLayout(layout, panel, xPadding, yPadding);
		}
		else {
			if(panel.getComponentCount()==0) {
				return;
			}
			
			//Initialize springs for width of the several columns
			maxWidthSprings = new Spring[cols];
			for(int i=0;i<cols;i++) {
				maxWidthSprings[i] = Spring.constant(0);
			}
			
			//Get the maximum width for each column (i.e. the maximum width of each component in a column) 
			for(int i=0;i<panel.getComponentCount();i++) {
				comp = panel.getComponent(i);
				comp = panel.getComponent(i);
				
				cons = layout.getConstraints(comp);
				//Ignore the last component if it is the only one in that row
				if(i%cols!=0 || i!=panel.getComponentCount()-1) {
					maxWidthSprings[i%cols] = Spring.max(maxWidthSprings[i%cols], 
							cons.getWidth());
				}
			}
			
			//Initialize most-east and most-south component (to compute the width and the height of the panel)
			eastComp = panel.getComponent(0);
			southComp = panel.getComponent(0);
			//Lay out the components on the panel
			for(int i=0;i<panel.getComponentCount();i++) {
				comp = panel.getComponent(i);
				//The first component is attached to the left and the top edge of the panel
				if(i==0) {
					layout.putConstraint(SpringLayout.NORTH, comp, yPadding, 
							SpringLayout.NORTH, panel);
					layout.putConstraint(SpringLayout.WEST, comp, xPadding, 
							SpringLayout.WEST, panel);
				}
				//every component which belongs to the first column is adjusted by the most-south component
				//in the row above
				else if(i%cols==0) {
					for(int j=i-cols;j<i;j++) {
						if(layout.getConstraint(SpringLayout.SOUTH, adjustmentComp).getValue() <
								layout.getConstraint(SpringLayout.SOUTH, panel.getComponent(j)).getValue()) {
							adjustmentComp = panel.getComponent(j);
						}
					}
					
					layout.putConstraint(SpringLayout.NORTH, comp, yPadding, 
							SpringLayout.SOUTH, adjustmentComp);
					layout.putConstraint(SpringLayout.WEST, comp, xPadding, 
							SpringLayout.WEST, panel);
				}
				//every other component is adjusted by its left neighbor
				else {
					adjustmentComp = panel.getComponent(i-1);
					layout.putConstraint(SpringLayout.NORTH, comp, 0, 
							SpringLayout.NORTH, adjustmentComp);
					layout.putConstraint(SpringLayout.WEST, comp, xPadding, 
							SpringLayout.EAST, adjustmentComp);
				}
				
				//Set the width of each component of the column to the columns' width
				//except for the last component if it is the only one in that row
				if(i%cols!=0 || i!=panel.getComponentCount()-1) {
					layout.getConstraints(comp).setWidth(maxWidthSprings[i%cols]);
				}
				
				//Check whether the current component lays more to right than the previous
				if(layout.getConstraint(SpringLayout.EAST, eastComp).getValue() < 
						layout.getConstraint(SpringLayout.EAST, comp).getValue()) {
					eastComp = comp;
				}
				
				//Check whether the current component lays more to bottom than the previous
				if(layout.getConstraint(SpringLayout.SOUTH, southComp).getValue() < 
						layout.getConstraint(SpringLayout.SOUTH, comp).getValue()) {
					southComp = comp;
				}
			}
			
			//Set the panel's width and height such that 
			//every component on this panel is visible
			layout.putConstraint(SpringLayout.EAST, panel, xPadding, 
					SpringLayout.EAST, eastComp);
			layout.putConstraint(SpringLayout.SOUTH, panel, yPadding, 
					SpringLayout.SOUTH, southComp);
		}
	}
	
	/**
	 * Sets the border of the component {@code component} to a lowered bevel border with 
	 * title {@code title}.
	 * @param component the component whose border is set
	 * @param title the title of the border
	 */
	protected static void setCompoundBorder(JComponent component, String title) {
		Border loweredbevel = BorderFactory.createLoweredBevelBorder();
		Border titledBorder = BorderFactory.createTitledBorder(loweredbevel, " "+title+" ");
		
		component.setBorder(titledBorder);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Private Static Methods																*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new list-based layout {@code layout} placing all components of the panel {@code panel}.
	 * @param layout the layout which is used for the laying out the panel's components
	 * @param panel the panel whose components will be laid out
	 * @param xPadding the padding in x-direction
	 * @param yPadding the padding in y-direction
	 */
	private static void makeListLayout(SpringLayout layout, JPanel panel, int xPadding, int yPadding) {
		Spring maxWidthSpring = null;
		Component comp = null;
		Component adjustmentComp = null;
		SpringLayout.Constraints cons = null;		
		
		if(panel.getComponentCount()==0) {
			return;
		}
		
		//Initialize spring for the maximum width of the column
		maxWidthSpring = Spring.constant(0);
		
		//Get the maximum width for each column (i.e. the maximum width of each component in a column) 
		for(int i=0;i<panel.getComponentCount();i++) {
			comp = panel.getComponent(i);
			comp = panel.getComponent(i);
			
			cons = layout.getConstraints(comp);
			maxWidthSpring = Spring.max(maxWidthSpring, cons.getWidth());
		}
		
		//Lay out the components on the panel
		for(int i=0;i<panel.getComponentCount();i++) {
			comp = panel.getComponent(i);
			//The first component is attached to the left and the top edge of the panel
			if(i==0) {
				layout.putConstraint(SpringLayout.NORTH, comp, yPadding, 
						SpringLayout.NORTH, panel);
				layout.putConstraint(SpringLayout.WEST, comp, xPadding, 
						SpringLayout.WEST, panel);
			}
			//every other component is adjusted by its left neighbor
			else {
				adjustmentComp = panel.getComponent(i-1);
				layout.putConstraint(SpringLayout.NORTH, comp, yPadding, 
						SpringLayout.SOUTH, adjustmentComp);
				layout.putConstraint(SpringLayout.WEST, comp, 0, 
						SpringLayout.WEST, adjustmentComp);
			}
			
			//Set the width of each component of the column to the columns' maximum width
			layout.getConstraints(comp).setWidth(maxWidthSpring);
		}
		
		//Set the panel's width and height such that 
		//every component on this panel is visible
		layout.putConstraint(SpringLayout.EAST, panel, xPadding, 
				SpringLayout.EAST, panel.getComponent(panel.getComponentCount()-1));
		layout.putConstraint(SpringLayout.SOUTH, panel, yPadding, 
				SpringLayout.SOUTH, panel.getComponent(panel.getComponentCount()-1));
	}
	
}
