/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.DefaultListModel;
import javax.swing.ImageIcon;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.JToolBar;
import javax.swing.ListSelectionModel;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import com.jgoodies.forms.builder.DefaultFormBuilder;
import com.jgoodies.forms.factories.Borders;
import com.jgoodies.forms.layout.FormLayout;

import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * Displays current variables and properties in a window and enables the user to
 * edit and create variables/properties.
 * 
 * @author Laura Steinert, Dennis Nolte
 * @version $Revision: 1.13 $
 */
public class VariablesAndPropertiesDialog extends RavenDialog {

	private static class PropertyComparator implements Comparator<String> {

		@Override
		public int compare(String s, String t) {
			if(RavenMain.isSystemProperty(s)==RavenMain.isSystemProperty(t)) {
				return s.compareTo(t);
			}
			else if(RavenMain.isSystemProperty(s)) {
				return -1;
			}
			
			return 1;
		}
		
	}

	/**
	 * 
	 * @author Christoph Blume
	 *
	 */
	private static class EditPropertyDialog extends RavenDialog {

		private static final long serialVersionUID = -1735339928124098876L;

		private static EditPropertyDialog dialog;

		private final JTextField KEY_FIELD;
		private final JTextField VALUE_FIELD;

		/*------------------------------------------------------------------------------------------*/
		/*		Static Methods																		*/
		/*------------------------------------------------------------------------------------------*/

		/**
		 * Shows the dialog for the creation/update of properties.
		 * @param owner the parent window of the dialog
		 * @param key the key of the property which is to be edited
		 */
		public static void showDialog(Window owner, String key) {
			dialog = new EditPropertyDialog(owner, key);
			dialog.setLocationRelativeTo(owner);
			dialog.setVisible(true);
		}

		/*------------------------------------------------------------------------------------------*/
		/*		Constructors																		*/
		/*------------------------------------------------------------------------------------------*/

		/**
		 * Creates a new dialog for the creation/update of properties.
		 * @param owner the parent window of this dialog
		 */
		private EditPropertyDialog(Window owner, final String key) {
			super(owner, messages.getString(
					GuiMessages.UPDATE_PROPERTY_DIALOG_TITLE_KEY),
					RavenDialog.ACCEPT_CANCEL_LAYOUT);

			DefaultFormBuilder builder = new DefaultFormBuilder(new FormLayout("pref, 3dlu, 80dlu:grow", ""));
			builder.border(Borders.DIALOG);

			JLabel keyLabel = new JLabel(messages.getString(GuiMessages.KEY_LABEL_KEY));
			KEY_FIELD = new JTextField(20);

			JLabel valueLabel = new JLabel(messages.getString(GuiMessages.VALUE_LABEL_KEY));
			VALUE_FIELD = new JTextField(20);

			if(edit) {
				KEY_FIELD.setText(key);
				KEY_FIELD.setEditable(!RavenMain.isSystemProperty(key));
				VALUE_FIELD.setText(RavenMain.getProperty(key));
			}
			else {
				KEY_FIELD.setEditable(true);
			}

			builder.append(keyLabel, KEY_FIELD);
			builder.nextLine();
			builder.append(valueLabel, VALUE_FIELD);

			AbstractAction acceptAction = new AbstractAction(
					messages.getString(GuiMessages.ACCEPT_LABEL_KEY)) {

				private static final long serialVersionUID = 321250982627352495L;

				@Override
				public void actionPerformed(ActionEvent e) {
					String newKey = KEY_FIELD.getText();
					String newValue = VALUE_FIELD.getText();

					if(newKey.isEmpty()) {
						showErrorMessage(GuiMessages.EMPTY_INPUTFIELD_TITLE_EXCEPTION_KEY,
								GuiMessages.EMPTY_INPUTFIELD_TEXT_EXCEPTION_KEY,
								messages.getString(GuiMessages.KEY_LABEL_KEY));
						return;
					}
					if(newValue.isEmpty()) {
						showErrorMessage(GuiMessages.EMPTY_INPUTFIELD_TITLE_EXCEPTION_KEY,
								GuiMessages.EMPTY_INPUTFIELD_TEXT_EXCEPTION_KEY,
								messages.getString(GuiMessages.VALUE_LABEL_KEY));
						return;
					}

					if(edit) {
						if(!key.equals(newKey)) {
							RavenMain.removeProperty(key);
							keyModel.removeElementAt(keyList.getSelectedIndex());
							addKey(newKey);
						}
						RavenMain.setProperty(newKey, newValue);
					}
					else {
						RavenMain.setProperty(newKey, newValue);
						addKey(newKey);
					}
					RavenMain.saveProperties();
					keyList.setSelectedValue(newKey, true);
					valueField.setText(RavenMain.getProperty(newKey));
					EditPropertyDialog.this.dispose();
				}

				/**
				 * Adds the key {@code key} to the list model.
				 * @param key the key to be added
				 */
				private void addKey(String key) {
					int index = 0;
					while(index<keyModel.size() && 
							COMPARATOR.compare(keyModel.get(index), key)<1) {
						index++;
					}
					keyModel.add(index, key);
				}
			};

			setMainPane(builder.getPanel());
			getContentPane().setBackground(MainWindow.BACKGROUND_COLOR);
			setAction(ACCEPT_BUTTON, acceptAction);
			pack();
			setSize(new Dimension(getWidth()+200, getHeight()));
		}
		
		/*------------------------------------------------------------------------------------------*/
		/*		Private Methods																		*/
		/*------------------------------------------------------------------------------------------*/

	}

	private static final long serialVersionUID = -6070816682340188463L;
	
	private static final PropertyComparator COMPARATOR = new PropertyComparator();
	
	private static VariablesAndPropertiesDialog dialog;
	private static JList<String> keyList;
	private static boolean edit;
	private static DefaultListModel<String> keyModel;
	private static JToolBar tableToolBar;
	private static JTextField valueField;

	private String key;
	private String value;

	/*------------------------------------------------------------------------------------------*/
	/* 		Static Methods 																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Brings up the dialog to display, add, edit and remove the variables and properties. 
	 * @param owner the parent window
	 */
	public static void showDialog(Window owner) {
		if(dialog==null) {
			dialog = new VariablesAndPropertiesDialog(owner);
		}

		dialog.setVisible(true);
	}

	/*------------------------------------------------------------------------------------------*/
	/* 		Constructors 																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new {@code VariablesAndPropertiesDialog}-object to manipulate variables and 
	 * properties.
	 * @param owner the parent window
	 */
	private VariablesAndPropertiesDialog(Window owner) {
		super(owner, messages.getString(GuiMessages.PROPERTIES_DIALOG_TITLE_KEY), RavenDialog.CLOSE_LAYOUT);
		setMainPane(createMainPanel());
		getContentPane().setBackground(MainWindow.BACKGROUND_COLOR);
		pack();
		setSize(new Dimension(getWidth()+200, getHeight()));
		setLocationRelativeTo(null); // centralize window
	}

	/*------------------------------------------------------------------------------------------*/
	/* 		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/*------------------------------------------------------------------------------------------*/
	/* 		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	private JPanel createMainPanel() {
		DefaultFormBuilder builder = new DefaultFormBuilder(new FormLayout("pref, 3dlu, 80dlu:grow", ""));
		builder.border(Borders.DIALOG);

		List<String> propertyNames = new LinkedList<>(RavenMain.getPropertyNames());
		Collections.sort(propertyNames, COMPARATOR);
		keyModel = new DefaultListModel<>();
		for(String property: propertyNames) {
			keyModel.addElement(property);
		}
		
		keyList = new JList<>(keyModel);
		keyList.setOpaque(false);
		keyList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		keyList.addListSelectionListener(new ListSelectionListener() {

			@Override
			public void valueChanged(ListSelectionEvent e) {
				int index = keyList.getSelectedIndex();

				if(index!=-1) {
					key = keyList.getSelectedValue();
					value = RavenMain.getProperty(key);

					valueField.setText(value);

					if(RavenMain.isSystemProperty(key)) {
						setAllButtonEnabled(true);
						tableToolBar.getComponent(1).setEnabled(false);
					}
					else {
						setAllButtonEnabled(true);
					}
				}
				else {
					valueField.setText("");
					setAllButtonEnabled(false);
					tableToolBar.getComponent(0).setEnabled(true);
				}
			}
		});

		builder.appendSeparator(messages.getString(GuiMessages.KEY_LABEL_KEY));
		builder.append(new JScrollPane(keyList),3);
		builder.nextLine();

		tableToolBar = new JToolBar();
		tableToolBar.setOpaque(false);
		tableToolBar.setBackground(MainWindow.BACKGROUND_COLOR);
		tableToolBar.setFloatable(false);
		tableToolBar.setBorderPainted(false);

		tableToolBar.add(new AbstractAction("", 
				new ImageIcon(RavenMain.PLUS_ICON_FILENAME)) {

			private static final long serialVersionUID = 6480710296758476298L;

			@Override
			public void actionPerformed(ActionEvent e) {
				edit = false;
				EditPropertyDialog.showDialog(
						VariablesAndPropertiesDialog.this, key);
			}
		});

		tableToolBar.add(new AbstractAction("", 
				new ImageIcon(RavenMain.MINUS_ICON_FILENAME)) {

			private static final long serialVersionUID = 6691022816927260856L;

			@Override
			public void actionPerformed(ActionEvent e) {
				int index = keyList.getSelectedIndex();

				if(index!=-1) {
					RavenMain.removeProperty(key);
					keyModel.removeElementAt(keyList.getSelectedIndex());
					RavenMain.saveProperties();
				}
				else {
					showInformationMessage(GuiMessages.NO_KEY_SELECTED_TITLE_KEY,
							GuiMessages.NO_KEY_SELECTED_TEXT_KEY);
				}
			}
		});

		tableToolBar.add(new AbstractAction("", 
				new ImageIcon(RavenMain.EDIT_ICON_FILENAME)) {

			private static final long serialVersionUID = 6480710296758476298L;

			@Override
			public void actionPerformed(ActionEvent e) {
				int index = keyList.getSelectedIndex();

				if(index!=-1) {
					edit = true;
					EditPropertyDialog.showDialog(
							VariablesAndPropertiesDialog.this, key);
					
				}
				else {
					showInformationMessage(GuiMessages.NO_KEY_SELECTED_TITLE_KEY,
							GuiMessages.NO_KEY_SELECTED_TEXT_KEY);
				}
			}
		});

		JPanel toolBarPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
		toolBarPanel.setOpaque(false);
		toolBarPanel.add(tableToolBar);
		builder.append(toolBarPanel,3);

		setAllButtonEnabled(false);
		tableToolBar.getComponent(0).setEnabled(true);

		valueField = new JTextField(40);
		valueField.setOpaque(false);
		valueField.setEditable(false);

		builder.appendSeparator(messages.getString(GuiMessages.VALUE_LABEL_KEY));
		builder.append(valueField,3);

		JPanel mainPanel = builder.getPanel();
		mainPanel.setOpaque(false);

		return mainPanel;
	}

	private static void setAllButtonEnabled(boolean bool) {

		for(Component comp: tableToolBar.getComponents()) {
			comp.setEnabled(bool);
		}
	}

}
