/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui.actions;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.event.ActionEvent;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.BorderFactory;
import javax.swing.ImageIcon;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextArea;

import de.uni_due.inf.ti.raven.gui.GuiMessages;
import de.uni_due.inf.ti.raven.gui.MainWindow;
import de.uni_due.inf.ti.raven.gui.RavenDialog;
import de.uni_due.inf.ti.raven.gui.Utilities;
import de.uni_due.inf.ti.raven.main.RavenMain;
import de.uni_due.inf.ti.raven.main.RavenUpdater;

/**
 * Checks whether a new version exists and if so asks the user if the update shall be executed.
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.8 $
 *
 */
public class UpdateAction extends AbstractAction{

	private static class UpdateWindow extends RavenDialog {
		
		private static final long serialVersionUID = 4609603726035904084L;
		
		private static int value;
		
		/**
		 * Shows an update window which asks the user to execute the update.
		 * @return {@link RavenDialog#YES_OPTION} if the user has agreed to perform the update or 
		 * {@link RavenDialog#NO_OPTION} otherwise
		 */
		public static int showUpdateWindow() {
			final UpdateWindow dialog = new UpdateWindow(MainWindow.getWindow());
			dialog.setAction(RavenDialog.YES_BUTTON, new AbstractAction(
					messages.getString(GuiMessages.YES_LABEL_KEY)) {
				
				private static final long serialVersionUID = -1064230459341987373L;

				@Override
				public void actionPerformed(ActionEvent e) {
					value = RavenDialog.YES_OPTION;
					dialog.dispose();
				}
			});
			dialog.setAction(RavenDialog.NO_BUTTON, new AbstractAction(
					messages.getString(GuiMessages.NO_LABEL_KEY)) {
				
				private static final long serialVersionUID = -1064230459341987373L;

				@Override
				public void actionPerformed(ActionEvent e) {
					value = RavenDialog.NO_OPTION;
					dialog.dispose();
				}
			});
			
			dialog.setVisible(true);
			
			return value;
		}
		
		/**
		 * Creates a new update window which asks the user whether to perform the update or not.
		 */
		private UpdateWindow(Component parent) {
			super(parent, messages.getString(GuiMessages.UPDATE_AVAILABLE_TITLE_KEY), 
					RavenDialog.YES_NO_LAYOUT);

			JLabel messageLabel = new JLabel(Utilities.convertToHTMLString(
					messages.getString(GuiMessages.UPDATE_AVAILABLE_TEXT_KEY)), 
					new ImageIcon(RavenMain.DIALOG_QUESTION_ICON_FILENAME), 
					JLabel.LEFT);
			messageLabel.setIconTextGap(10);
			
			String versionString = RavenUpdater.getMajorVersion() + "." + RavenUpdater.getMinorVersion();
			String updateInformation = RavenUpdater.getUpdateInformation();
			if(updateInformation==null || updateInformation.isEmpty()) {
				updateInformation = messages.getString(
						GuiMessages.NO_INFORMATION_AVAILABLE_LABEL_KEY);
			}
			
			JLabel informationLabel = new JLabel(messages.getString(
					GuiMessages.UPDATE_INFORMATION_LABEL_KEY, versionString)+":");
			informationLabel.setBorder(BorderFactory.createEmptyBorder(20, 0, 5, 0));
			
			JTextArea informationArea = new JTextArea(updateInformation);
			informationArea.setEditable(false);
			informationArea.setCursor(null);
			informationArea.setFocusable(false);
			informationArea.setLineWrap(true);
			informationArea.setOpaque(false);
			informationArea.setWrapStyleWord(true);

			JPanel informationPanel = new JPanel(new BorderLayout());
			informationPanel.add(informationLabel, BorderLayout.NORTH);
			informationPanel.add(informationArea, BorderLayout.CENTER);
			
			JPanel mainPane = new JPanel(new BorderLayout());
			mainPane.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));
			mainPane.add(messageLabel, BorderLayout.NORTH);
			mainPane.add(informationPanel, BorderLayout.CENTER);
			setMainPane(mainPane);
			setResizable(false);
			pack();
			setLocationRelativeTo(parent);
		}
		
	}
	
	private static final long serialVersionUID = 8822491429279494910L;
	
	private static GuiMessages messages = GuiMessages.getMessages();
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates an {@code Action} for updating Raven.
	 */
	public UpdateAction() {
		putValue( Action.NAME, messages.getString(GuiMessages.UPDATE_MENU_KEY));
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public void actionPerformed(ActionEvent e) {
		//Look for updates (if updates are available the update-object is not null
		int updateStatus = RavenUpdater.updateAvailable();
		
		if(updateStatus==RavenUpdater.STATUS_CODE_UPDATE_AVAILABLE) {
			int result = UpdateWindow.showUpdateWindow();		
			if (result == RavenDialog.YES_OPTION){
				int status = RavenUpdater.downloadUpdate();
				if(status==RavenUpdater.STATUS_CODE_SUCCESS) {
					//If the update has been successfully installed, try to restart Raven
					RavenMain.restart();
				}
				else {
					Utilities.showErrorMessage(MainWindow.getWindow(), 
							GuiMessages.UPDATE_ERROR_TITLE_EXCEPTION_KEY, 
							GuiMessages.UPDATE_ERROR_TEXT_EXCEPTION_KEY);
				}
			}
		}
		else {
			Utilities.showInformationMessage(MainWindow.getWindow(), 
					GuiMessages.NO_UPDATE_AVAILABLE_TITLE_KEY, 
					GuiMessages.NO_UPDATE_AVAILABLE_TEXT_KEY);
		}
	}

}
