/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui.automaton;

import java.awt.BorderLayout;
import java.awt.CardLayout;
import java.awt.Component;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.HashMap;
import java.util.Map;

import javax.swing.AbstractAction;
import javax.swing.JComboBox;
import javax.swing.JPanel;
import javax.swing.JTextField;

import com.jgoodies.forms.builder.DefaultFormBuilder;
import com.jgoodies.forms.factories.Borders;
import com.jgoodies.forms.layout.FormLayout;

import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.gui.GuiMessages;
import de.uni_due.inf.ti.raven.gui.MainWindow;
import de.uni_due.inf.ti.raven.gui.RavenDialog;
import de.uni_due.inf.ti.raven.gui.RavenProgressBarPanel;

/**
 * This class provides a dialog to create a new graph automaton. Please note
 * that the dialog is modal, i.e. the {@code showDialog} method blocks the
 * caller until the user's interaction is complete.
 * 
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.19 $
 * 
 */
public class AutomatonCreationDialog extends RavenDialog {

	private static final long serialVersionUID = 5280749376137803746L;

	/** Name Text Field Label (used for Testing) */
	protected static final String AUTOMATON_NAME_LABEL = "automatonName";
	
	/** Automaton Type Box Label (used for Testing) */
	protected static final String AUTOMATON_TYPE_LABEL = "automatonType";

	private static final Automaton.Type COLORABILITY_TYPE = Automaton.Type.colorability;
	private static final Automaton.Type DOMINATING_SET_TYPE = Automaton.Type.dominatingSet;
	private static final Automaton.Type EDGE_COUNTING_TYPE = Automaton.Type.edgeCounting;
	private static final Automaton.Type LINK_TYPE = Automaton.Type.link;
	private static final Automaton.Type MAXIMUM_EDGE_TYPE = Automaton.Type.maximumEdge;
	private static final Automaton.Type MAXIMUM_VERTEX_TYPE = Automaton.Type.maximumVertex;
	private static final Automaton.Type MINIMUM_EDGE_TYPE = Automaton.Type.minimumEdge;
	private static final Automaton.Type MINIMUM_VERTEX_TYPE = Automaton.Type.minimumVertex;
	private static final Automaton.Type NO_ISOLATED_NODES_TYPE = Automaton.Type.noIsolatedNodes;
	private static final Automaton.Type PATH_TYPE = Automaton.Type.path;
	private static final Automaton.Type PRODUCT_TYPE = Automaton.Type.product;
	private static final Automaton.Type SUBGRAPH_TYPE = Automaton.Type.subgraph;
	private static final Automaton.Type UNION_TYPE = Automaton.Type.union;
	private static final Automaton.Type VERTEX_COUNTING_TYPE = Automaton.Type.vertexCounting;
	private static final Automaton.Type VERTEX_COVER_TYPE = Automaton.Type.vertexCover;

	private static final GuiMessages messages = GuiMessages.getMessages();
	
	private final Map<Automaton.Type, AutomatonCreationPanel> panelMap;

	private JTextField automatonNameField;
	private JComboBox<Automaton.Type> automatonTypeBox;
	private JPanel propertiesPanel;
	
	/** Progress panel to visualize the computation progress */
	private RavenProgressBarPanel progressPanel;


	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Shows the dialog for creation of new automata.
	 * 
	 * @param owner the parent window of the dialog
	 */
	public static void showDialog(Window owner) {
		// has to be created every time in case the repository has changed
		AutomatonCreationDialog dialog = new AutomatonCreationDialog(owner);
		dialog.setVisible(true);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new dialog for creation of a new automaton.
	 * 
	 * @param owner the parent window of this dialog
	 */
	protected AutomatonCreationDialog(Window owner) {
		super(owner, messages.getString(GuiMessages.AUTOMATON_CREATION_DIALOG_TITLE_KEY),
				RavenDialog.ACCEPT_CANCEL_LAYOUT);
		getContentPane().setBackground(MainWindow.BACKGROUND_COLOR);
		AbstractAction acceptAction = new AbstractAction(
				messages.getString(GuiMessages.ACCEPT_LABEL_KEY)) {

			private static final long serialVersionUID = -4391546903154013085L;

			@Override
			public void actionPerformed(ActionEvent e) {
				final String automatonName = automatonNameField.getText();
				
				if(automatonName.isEmpty()) {
					showWarningMessage(GuiMessages.EMPTY_INPUTFIELD_TITLE_EXCEPTION_KEY,
							GuiMessages.EMPTY_INPUTFIELD_TEXT_EXCEPTION_KEY,
							messages.getString(GuiMessages.NAME_LABEL_KEY));
					return;
				}
				if (!Repository.isNameValid(automatonName)) {
					showWarningMessage(GuiMessages.INVALID_NAME_TITLE_EXCEPTION_KEY,
							GuiMessages.INVALID_NAME_TEXT_EXCEPTION_KEY, automatonName);
					return;
				}

				if (Repository.getRepository().contains(automatonName)) {
					showInformationMessage(GuiMessages.NAME_ALREADY_IN_USE_TITLE_EXCEPTION_KEY,
							GuiMessages.NAME_ALREADY_IN_USE_TEXT_EXCEPTION_KEY, automatonName);
					return;
				}

				new Thread(new Runnable() {
					
					@Override
					public void run() {
						Automaton.Type type = automatonTypeBox.getItemAt(automatonTypeBox.getSelectedIndex());
						Automaton automaton = panelMap.get(type).createAutomaton(progressPanel);
						if(automaton!=null) {
							Repository.getRepository().put(automatonName, automaton);
							dispose();
						}
						
					}
				}).start();
			}

		};

		panelMap = new HashMap<>();
		progressPanel = new RavenProgressBarPanel();
		progressPanel.setOpaque(false);
		
		JPanel mainPanel = new JPanel(new BorderLayout());
		mainPanel.add(createChooserPanel(), BorderLayout.NORTH);
		mainPanel.add(createPropertiesPanel(), BorderLayout.CENTER);
		mainPanel.add(progressPanel, BorderLayout.SOUTH);
		mainPanel.setOpaque(false);
		
		setMainPane(mainPanel);
		setAction(ACCEPT_BUTTON, acceptAction);
		pack();
		setLocationRelativeTo(owner);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	private Component createPropertiesPanel() {
		panelMap.put(COLORABILITY_TYPE, new ColorabilityAutomatonCreationPanel());
		panelMap.put(DOMINATING_SET_TYPE, new DominatingSetAutomatonCreationPanel());
		panelMap.put(EDGE_COUNTING_TYPE, new EdgeCountingAutomatonCreationPanel());
		panelMap.put(LINK_TYPE, new LinkAutomatonCreationPanel());
		panelMap.put(MAXIMUM_EDGE_TYPE, new MaximumEdgeAutomatonCreationPanel());
		panelMap.put(MAXIMUM_VERTEX_TYPE, new MaximumVertexAutomatonCreationPanel());
		panelMap.put(MINIMUM_EDGE_TYPE, new MinimumEdgeAutomatonCreationPanel());
		panelMap.put(MINIMUM_VERTEX_TYPE, new MinimumVertexAutomatonCreationPanel());
		panelMap.put(NO_ISOLATED_NODES_TYPE, new NoIsolatedNodesAutomatonCreationPanel());
		panelMap.put(PATH_TYPE, new PathAutomatonCreationPanel());
		panelMap.put(PRODUCT_TYPE, new ProductAutomatonCreationPanel());
		panelMap.put(SUBGRAPH_TYPE, new SubgraphAutomatonCreationPanel());
		panelMap.put(UNION_TYPE, new UnionAutomatonCreationPanel());
		panelMap.put(VERTEX_COUNTING_TYPE, new VertexCountingAutomatonCreationPanel());
		panelMap.put(VERTEX_COVER_TYPE, new VertexCoverAutomatonCreationPanel());

		propertiesPanel = new JPanel(new CardLayout());
		propertiesPanel.add(panelMap.get(COLORABILITY_TYPE), COLORABILITY_TYPE.toString());
		propertiesPanel.add(panelMap.get(DOMINATING_SET_TYPE), DOMINATING_SET_TYPE.toString());
		propertiesPanel.add(panelMap.get(EDGE_COUNTING_TYPE), EDGE_COUNTING_TYPE.toString());
		propertiesPanel.add(panelMap.get(LINK_TYPE), LINK_TYPE.toString());
		propertiesPanel.add(panelMap.get(MAXIMUM_EDGE_TYPE), MAXIMUM_EDGE_TYPE.toString());
		propertiesPanel.add(panelMap.get(MAXIMUM_VERTEX_TYPE), MAXIMUM_VERTEX_TYPE.toString());
		propertiesPanel.add(panelMap.get(MINIMUM_EDGE_TYPE), MINIMUM_EDGE_TYPE.toString());
		propertiesPanel.add(panelMap.get(MINIMUM_VERTEX_TYPE), MINIMUM_VERTEX_TYPE.toString());
		propertiesPanel.add(panelMap.get(NO_ISOLATED_NODES_TYPE), NO_ISOLATED_NODES_TYPE.toString());
		propertiesPanel.add(panelMap.get(PATH_TYPE), PATH_TYPE.toString());
		propertiesPanel.add(panelMap.get(PRODUCT_TYPE), PRODUCT_TYPE.toString());
		propertiesPanel.add(panelMap.get(SUBGRAPH_TYPE), SUBGRAPH_TYPE.toString());
		propertiesPanel.add(panelMap.get(UNION_TYPE), UNION_TYPE.toString());
		propertiesPanel.add(panelMap.get(VERTEX_COUNTING_TYPE), VERTEX_COUNTING_TYPE.toString());
		propertiesPanel.add(panelMap.get(VERTEX_COVER_TYPE), VERTEX_COVER_TYPE.toString());

		propertiesPanel.setOpaque(false);
		return propertiesPanel;
	}

	/**
	 * Creates the main panel to set the new automaton's properties.
	 * 
	 * @return the main panel
	 */
	private JPanel createChooserPanel() {
		String labelSuffix = ": ";

		FormLayout layout = new FormLayout("left:max(40dlu;pref), 3dlu, 40dlu:grow", "");
		DefaultFormBuilder builder = new DefaultFormBuilder(layout);
		builder.border(Borders.DIALOG);
		builder.appendSeparator(messages.getString(GuiMessages.AUTOMATON_CREATION_PROPERTIES_TEXT_KEY));
		builder.nextLine();

		automatonNameField = new JTextField();
		automatonNameField.setName(AUTOMATON_NAME_LABEL);
		builder.append(messages.getString(GuiMessages.NAME_LABEL_KEY) + labelSuffix, automatonNameField);
		builder.nextLine();

		automatonTypeBox = new JComboBox<>(Automaton.Type.getBasicTypes());
		automatonTypeBox.setName(AUTOMATON_TYPE_LABEL);
		automatonTypeBox.setOpaque(false);
		automatonTypeBox.setBackground(MainWindow.BACKGROUND_COLOR);
		automatonTypeBox.addItemListener(new ItemListener() {

			@Override
			public void itemStateChanged(ItemEvent e) {
				if (e.getStateChange() == ItemEvent.SELECTED) {
					CardLayout layout = (CardLayout) (propertiesPanel.getLayout());
					layout.show(propertiesPanel, e.getItem().toString());
				}
			}
		});
		builder.append(messages.getString(GuiMessages.AUTOMATON_TYPE_LABEL_KEY) + labelSuffix, 
				automatonTypeBox);

		return builder.getPanel();
	}

}
