/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.gui.automaton;

import java.awt.Dimension;
import java.text.DecimalFormat;

import javax.swing.JComponent;
import javax.swing.JFormattedTextField;
import javax.swing.JPanel;

import com.jgoodies.forms.builder.DefaultFormBuilder;
import com.jgoodies.forms.factories.Borders;
import com.jgoodies.forms.layout.FormLayout;

import de.uni_due.inf.ti.raven.actions.ComputationListener;
import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.data.Signature;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.gui.DataComboBox;
import de.uni_due.inf.ti.raven.gui.GuiMessages;
import de.uni_due.inf.ti.raven.gui.MainWindow;
import de.uni_due.inf.ti.raven.gui.Utilities;

/**
 * The {@code AutomatonCreationPanel} is the abstract base class for the different creation panel. 
 * This super class implements the general input components needed for creation of the different 
 * graph automata. The concrete subclasses add additional input components.
 * @author Christoph Blume
 * @version $Revision: 1.16 $
 *
 */
public abstract class AutomatonCreationPanel extends JPanel {

	private static final String LAYOUT_SPECIFICATION = "left:max(40dlu;pref), 3dlu, 40dlu:grow";

	private static final long serialVersionUID = 7567010988034528901L;

	/** */
	protected static final GuiMessages messages = GuiMessages.getMessages();

	private static final String LABEL_SUFFIX = ": ";

	/** Alphabet List Label (used for Testing) */
	protected static final String ALPHABET_LABEL = "Alphabet";
	
	/** Maximum Interface Text Field Label (used for Testing) */
	protected static final String MAXIMUM_INTERFACE_LABEL = "maximumInterface";
	
	/** Inner Interface Text Field Label (used for Testing) */
	protected static final String INNER_INTERFACE_LABEL = "innerInterface";
	
	/** Outer Interface Text Field Label (used for Testing) */
	protected static final String OUTER_INTERFACE_LABEL = "outerInterface";
	
	/** Alphabet List */
	protected DataComboBox alphabetBox;

	/** Maximum Interface Text Field */
	protected JFormattedTextField maximumInterfaceField;
	
	/** Inner Interface Text Field */
	protected JFormattedTextField innerInterfaceField;
	
	/** Outer Interface Text Field */
	protected JFormattedTextField outerInterfaceField;
	
	private final DefaultFormBuilder builder;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the integer which has been typed in the specified input field.
	 * @param textfield the input field whose value will be read
	 * @return the integer which has been typed in or {@code null} if no integer has been typed in
	 */
	protected static Integer getInteger(JFormattedTextField textfield) {
		Object value = textfield.getValue();
		
		if(value!=null && value instanceof Number) {
			return ((Number) value).intValue();
		}
		
		return null;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a panel (with default input fields) for the creation of automata.
	 * @param addDefaultFields indicates whether the default input fields will be added
	 */
	protected AutomatonCreationPanel(boolean addDefaultFields) {
		builder = new DefaultFormBuilder(new FormLayout(LAYOUT_SPECIFICATION, ""), this);
		builder.border(Borders.DIALOG);
		builder.appendSeparator(messages.getString(GuiMessages.AUTOMATON_CREATION_AUTOMATON_PROPERTIES_TEXT_KEY));
		builder.nextLine();

		if(addDefaultFields) {
			alphabetBox = new DataComboBox(Repository.getRepository());
			alphabetBox.setName(ALPHABET_LABEL);
			alphabetBox.addDataClass(Signature.class);
			alphabetBox.setMinimumSize(new Dimension(40, 25));
			alphabetBox.setOpaque(false);
			alphabetBox.setBackground(MainWindow.BACKGROUND_COLOR);
			append(GuiMessages.SIGNATURE_LABEL_KEY, alphabetBox);
			
			maximumInterfaceField = new JFormattedTextField(new DecimalFormat("0"));
			maximumInterfaceField.setName(MAXIMUM_INTERFACE_LABEL);
			maximumInterfaceField.setColumns(10);
			append(GuiMessages.MAXIMUM_INTERFACE_LABEL_KEY, maximumInterfaceField);
			
			innerInterfaceField = new JFormattedTextField(new DecimalFormat("0"));
			innerInterfaceField.setName(INNER_INTERFACE_LABEL);
			append(GuiMessages.INNER_INTERFACE_SIZE_LABEL_KEY, innerInterfaceField);
			
			outerInterfaceField = new JFormattedTextField(new DecimalFormat("0"));
			outerInterfaceField.setName(OUTER_INTERFACE_LABEL);
			append(GuiMessages.OUTER_INTERFACE_SIZE_LABEL_KEY, outerInterfaceField);
		}
	}
	
	/**
	 * Creates a panel (with default input fields) for the creation of automata.
	 */
	protected AutomatonCreationPanel() {
		this(true);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Protected Abstract Methods															*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates an automaton depending on the values of input fields of the panel.
	 * @param listeners the listener to receive computation events
	 * @return the created automaton or {@code null} if an invalid value has been entered
	 */
	protected abstract Automaton createAutomaton(ComputationListener... listeners);
	
	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Adds a separator with the given text specified by the label key.
     * @param key the separator title text
	 */
	protected void appendSeparator(String key) {
		builder.appendSeparator(messages.getString(key) + LABEL_SUFFIX);
		builder.nextLine();
	}
	
	/**
	 * Adds a text specified by the label key to the panel
	 * @param key the label text
	 */
	protected void append(String key) {
		builder.append(messages.getString(key) + LABEL_SUFFIX);
		builder.nextLine();
	}
	
	/**
	 * Adds a text label specified by the label key to the panel that spans a specified number of 
	 * columns
	 * @param key the label text
	 * @param columnSpan the size of the column span
	 */
	protected void append(String key, int columnSpan) {
		builder.setColumnSpan(columnSpan);
		builder.append(messages.getString(key) + LABEL_SUFFIX);
		builder.nextLine();
	}
	
	/**
	 * Adds a text label specified by the label key and a component to the panel.
     * @param key the label text
     * @param component the component to be added
	 */
	protected void append(String key, JComponent component) {
		builder.append(messages.getString(key) + LABEL_SUFFIX, component);
		builder.nextLine();
	}
	
	/**
	 * Adds a component to the panel that spans a specified number of columns
	 * @param component the component to be added
	 * @param columnSpan the size of the column span
	 */
	protected void append(JComponent component, int columnSpan) {
		builder.append(component, columnSpan);
		builder.nextLine();
	}
	
	/**
	 * Brings up a message dialog showing a warning message.
	 * @param titleKey the key indicating the title string
	 * @param messageKey the key indicating the message string
	 * @param parameter additional parameter for the message string
	 */
	protected void showWarningMessage(String titleKey, String messageKey, String... parameter) {
		for(int i=0;i<parameter.length;i++) {
			parameter[i] = messages.getString(parameter[i]);
		}
		
		Utilities.showWarningMessage(this, titleKey, messageKey, parameter);
	}
	
}
