/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui.automaton;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.text.DecimalFormat;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JButton;
import javax.swing.JList;
import javax.swing.JTextField;

import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.gui.BDDDialog;
import de.uni_due.inf.ti.raven.gui.EncodingDialog;
import de.uni_due.inf.ti.raven.gui.GuiMessages;
import de.uni_due.inf.ti.raven.gui.InformationPanel;
import de.uni_due.inf.ti.raven.gui.MainWindow;
import de.uni_due.inf.ti.raven.gui.Utilities;

/**
 * This class provides graphical components to display the properties of graph automata objects.
 * @author Christoph Blume
 * @version $Revision: 1.8 $
 * @param <T> the type of automaton 
 *
 */
public abstract class AutomatonInformationPanel<T extends Automaton> extends InformationPanel<T> {

	private static final long serialVersionUID = 6104628483325920310L;

	/** */
	protected Repository repo = Repository.getRepository();
	
	/** */
	protected GuiMessages messages = GuiMessages.getMessages();
	
	// Automaton Information Panel Component Labels
	/** Automaton Name Text Field Label (used for Testing) */
	protected static final String AUTOMATON_NAME_LABEL = "automatonName";
	
	/** Automaton Type Text Field Label (used for Testing) */
	protected static final String TYPE_LABEL = "automatonType";
	
	/** Number of States Text Field Label (used for Testing) */
	protected static final String NUMBER_OF_STATES_LABEL = "numberOfStates";
	
	/** Number of Initial States Text Field Label (used for Testing) */
	protected static final String NUMBER_OF_INITIAL_STATES_LABEL = "numberOfInitialStates";
	
	/** Number of Final States Text Field Label (used for Testing) */
	protected static final String NUMBER_OF_FINAL_STATES_LABEL = "numberOfFinalStates";
	
	/** Maximum Interface Text Field Label (used for Testing) */
	protected static final String MAXIMUM_INTERFACE_LABEL = "maximumInterface";
	
	/** Inner Interface Text Field Label (used for Testing) */
	protected static final String INNER_INTERFACE_LABEL = "innerInterface";
	
	/** Outer Interface Text Field Label (used for Testing) */
	protected static final String OUTER_INTERFACE_LABEL = "outerInterface";
	
	/** Encoding Name Text Field Label (used for Testing) */
	protected static final String ENCODING_NAME_LABEL = "encodingName";
	
	/** Number of BDD Nodes Text Field Label (used for Testing) */
	protected static final String NUMBER_OF_BDD_NODES_LABEL = "numberOfBDDNodes";
	
	/** Show Encoding Button Label (used for Testing) */
	protected static final String SHOW_ENCODING_LABEL = "showEncoding";
	
	/** Alphabet List Label (used for Testing) */
	protected static final String ALPHABET_LABEL = "Alphabet";
	
	/** Show BDD Button Label (used for Testing) */
	protected static final String SHOW_BDD_LABEL = "showBDD";

	
	// Automaton Information Panel Components
	/** Automaton Name Text Field */
	protected JTextField automatonNameField;
	
	/** Automaton Type Text Field */
	protected JTextField typeField;
	
	/** Number of States Text Field */
	protected JTextField numberOfStatesField;
	
	/** Number of Initial States Text Field */
	protected JTextField numberOfInitialStatesField;
	
	/** Number of Final States Text Field */
	protected JTextField numberOfFinalStatesField;
	
	/** Maximum Interface Text Field */
	protected JTextField maximumInterfaceField;
	
	/** Inner Interface Text Field */
	protected JTextField innerInterfaceField;
	
	/** Outer Interface Text Field */
	protected JTextField outerInterfaceField;
	
	/** Encoding Name Text Field */
	protected JTextField encodingField;
	
	/** Number of BDD Nodes Text Field */
	protected JTextField numberOfBDDNodesField;
	
	/** Show Encoding Button */
	protected JButton showEncodingButton;
	
	/** Alphabet List */
	protected JList<Operation> alphabetList;
	
	/** Show BDD Button */
	protected JButton showBDDButton;
	
	/*------------------------------------------------------------------------------------------*/
	/* 		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Updates the components displaying the special information about the specific type of automaton.
	 * @param automaton the automaton whose information will be displayed
	 */
	public abstract void setAdditionalData(final T automaton);
	
	/*------------------------------------------------------------------------------------------*/
	/* 		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public void setData(final T automaton) {
		automatonNameField.setText(repo.getName(automaton));
		typeField.setText(messages.getString(
				GuiMessages.DATA_TYPE_PREFIX_KEY+automaton.getType().toString()));
		numberOfStatesField.setText(DecimalFormat.getInstance().format(
				automaton.getNumberOfStates()));
		numberOfInitialStatesField.setText(DecimalFormat.getInstance().format(
				automaton.getNumberOfInitialStates()));
		numberOfFinalStatesField.setText(DecimalFormat.getInstance().format(
				automaton.getNumberOfFinalStates()));
		maximumInterfaceField.setText(String.valueOf(automaton.getMaximumInterface()));
		innerInterfaceField.setText(String.valueOf(automaton.getInnerInterface()));
		outerInterfaceField.setText(String.valueOf(automaton.getOuterInterface()));
		encodingField.setText(automaton.getEncoding().toString());
		numberOfBDDNodesField.setText(DecimalFormat.getInstance().format(
				automaton.getSize()));

		List<Operation> list = new LinkedList<>(automaton.getAlphabet().getOperations());
		Collections.sort(list, new Comparator<Operation>() {

			@Override
			public int compare(Operation o1, Operation o2) {
				return o1.toString().compareTo(o2.toString());
			}

		});
		alphabetList.setListData(list.toArray(new Operation[0]));

		
		for(ActionListener l: showEncodingButton.getActionListeners()) {
			showEncodingButton.removeActionListener(l);
		}
		showEncodingButton.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				new Thread(new Runnable() {

					@Override
					public void run() {
						EncodingDialog.showDialog(null, automaton.getEncoding());
					}

				}).start();
			}
		});

		for(ActionListener l: showBDDButton.getActionListeners()) {
			showBDDButton.removeActionListener(l);
		}
		showBDDButton.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				if(alphabetList.isSelectionEmpty()) {
					Utilities.showWarningMessage(MainWindow.getWindow(), 
							GuiMessages.NO_OPERATION_SELECTED_TITLE_EXCEPTION_KEY, 
							GuiMessages.NO_OPERATION_SELECTED_TEXT_EXCEPTION_KEY);
				}
				else {
					new Thread(new Runnable() {

						@Override
						public void run() {
							BDDDialog.showDialog(MainWindow.getWindow(), automaton, 
									alphabetList.getSelectedValue());
						}
					}).start();
				}

			}
		});
		
		setAdditionalData(automaton);
	}
	
}
