/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.gui.automaton;

import javax.swing.JButton;
import javax.swing.JList;
import javax.swing.JScrollPane;
import javax.swing.JTextField;

import com.jgoodies.forms.builder.DefaultFormBuilder;
import com.jgoodies.forms.layout.FormLayout;

import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.gui.GuiMessages;
import de.uni_due.inf.ti.raven.gui.InformationPanelBuilder;

/**
 * This class provides an {@code InformationPanel} for displaying information about graph automata.
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.11 $
 *
 * @see Automaton
 */
public abstract class AutomatonInformationPanelBuilder extends InformationPanelBuilder<Automaton> {
	
	private AutomatonInformationPanel<? extends Automaton> panel;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates the panel on which automaton information will be displayed.
	 * @param panel the form builder used to build the panel
	 */
	protected AutomatonInformationPanelBuilder(AutomatonInformationPanel<? extends Automaton> panel) {
		super(new DefaultFormBuilder(new FormLayout(LAYOUT_SPECIFICATION), panel));

		this.panel = panel;
		addDefaultComponents();
		addAdditionalComponents();
		addEncodingAndAlphabetComponents();
	}
	
	/*------------------------------------------------------------------------------------------*/
	/* 		Abstract Methods																	*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Adds additional components to display further information about the automaton to the 
	 * information panel.
	 */
	protected abstract void addAdditionalComponents();
	
	/*------------------------------------------------------------------------------------------*/
	/* 		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Adds the default fields to display the automaton name, type, maximum interface, inner 
	 * interface and outer interface to the information panel.
	 */
	private void addDefaultComponents() {
		panel.automatonNameField = new JTextField(10);
		panel.automatonNameField.setName(AutomatonInformationPanel.AUTOMATON_NAME_LABEL);
		panel.automatonNameField.setEditable(false);
		panel.automatonNameField.setOpaque(false);
		append(GuiMessages.NAME_LABEL_KEY, panel.automatonNameField);

		panel.typeField = new JTextField(10);
		panel.typeField.setName(AutomatonInformationPanel.TYPE_LABEL);
		panel.typeField.setEditable(false);
		panel.typeField.setOpaque(false);
		append(GuiMessages.TYPE_LABEL_KEY, panel.typeField);

		appendSeparator();
		
		panel.numberOfStatesField = new JTextField(10);
		panel.numberOfStatesField.setName(AutomatonInformationPanel.NUMBER_OF_STATES_LABEL);
		panel.numberOfStatesField.setEditable(false);
		panel.numberOfStatesField.setOpaque(false);
		append(GuiMessages.NUMBER_OF_STATES_LABEL_KEY, panel.numberOfStatesField);

		panel.numberOfInitialStatesField = new JTextField(10);
		panel.numberOfInitialStatesField.setName(AutomatonInformationPanel.NUMBER_OF_INITIAL_STATES_LABEL);
		panel.numberOfInitialStatesField.setEditable(false);
		panel.numberOfInitialStatesField.setOpaque(false);
		append(GuiMessages.NUMBER_OF_INITIAL_STATES_LABEL_KEY, panel.numberOfInitialStatesField);

		panel.numberOfFinalStatesField = new JTextField(10);
		panel.numberOfFinalStatesField.setName(AutomatonInformationPanel.NUMBER_OF_FINAL_STATES_LABEL);
		panel.numberOfFinalStatesField.setEditable(false);
		panel.numberOfFinalStatesField.setOpaque(false);
		append(GuiMessages.NUMBER_OF_FINAL_STATES_LABEL_KEY, panel.numberOfFinalStatesField);
		
		appendSeparator();
		
		panel.maximumInterfaceField = new JTextField(10);
		panel.maximumInterfaceField.setName(AutomatonInformationPanel.MAXIMUM_INTERFACE_LABEL);
		panel.maximumInterfaceField.setEditable(false);
		panel.maximumInterfaceField.setOpaque(false);
		append(GuiMessages.MAXIMUM_INTERFACE_LABEL_KEY, panel.maximumInterfaceField);

		panel.innerInterfaceField = new JTextField(10);
		panel.innerInterfaceField.setName(AutomatonInformationPanel.INNER_INTERFACE_LABEL);
		panel.innerInterfaceField.setEditable(false);
		panel.innerInterfaceField.setOpaque(false);
		append(GuiMessages.INNER_INTERFACE_SIZE_LABEL_KEY, panel.innerInterfaceField);

		panel.outerInterfaceField = new JTextField(10);
		panel.outerInterfaceField.setName(AutomatonInformationPanel.OUTER_INTERFACE_LABEL);
		panel.outerInterfaceField.setEditable(false);
		panel.outerInterfaceField.setOpaque(false);
		append(GuiMessages.OUTER_INTERFACE_SIZE_LABEL_KEY, panel.outerInterfaceField);
		
		appendSeparator();
	}
	
	/**
	 * Adds the default components to display the encoding and alphabet used by the automaton to the 
	 * information panel.
	 */
	private void addEncodingAndAlphabetComponents() {
		panel.encodingField = new JTextField(10);
		panel.encodingField.setName(AutomatonInformationPanel.ENCODING_NAME_LABEL);
		panel.encodingField.setEditable(false);
		panel.encodingField.setOpaque(false);
		append(GuiMessages.ENCODING_LABEL_KEY, panel.encodingField);
		
		panel.numberOfBDDNodesField = new JTextField(10);
		panel.numberOfBDDNodesField.setName(AutomatonInformationPanel.NUMBER_OF_BDD_NODES_LABEL);
		panel.numberOfBDDNodesField.setEditable(false);
		panel.numberOfBDDNodesField.setOpaque(false);
		append(GuiMessages.NUMBER_OF_BDD_NODES_LABEL_KEY, panel.numberOfBDDNodesField);
		
		panel.showEncodingButton = new JButton();
		panel.showEncodingButton.setName(AutomatonInformationPanel.SHOW_ENCODING_LABEL);
		appendButton(GuiMessages.SHOW_ENCODING_LABEL_KEY, panel.showEncodingButton);
		
		appendSeparator();
		
		JScrollPane scrollPane = new JScrollPane();		
		panel.alphabetList = new JList<>();
		panel.alphabetList.setName(AutomatonInformationPanel.ALPHABET_LABEL);
		panel.alphabetList.setVisibleRowCount(6);
		panel.alphabetList.setOpaque(false);
		scrollPane.setViewportView(panel.alphabetList);
		scrollPane.getViewport().setOpaque(false);
		scrollPane.setOpaque(false);
		append(GuiMessages.OPERATIONS_LABEL_KEY, scrollPane);
			
		panel.showBDDButton = new JButton();
		panel.showBDDButton.setName(AutomatonInformationPanel.SHOW_BDD_LABEL);
		appendButton(GuiMessages.SHOW_BDD_LABEL_KEY, panel.showBDDButton);
	}
	
}
