/*
 * Copyright (c) 2009-2013, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * Neither the name of the "Theoretical Computer Science, University of 
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui.automaton;

import java.awt.Dimension;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Logger;

import com.mxgraph.layout.mxParallelEdgeLayout;
import com.mxgraph.swing.mxGraphComponent;
import com.mxgraph.util.mxEvent;
import com.mxgraph.util.mxEventObject;
import com.mxgraph.util.mxEventSource.mxIEventListener;
import com.mxgraph.util.mxRectangle;
import com.mxgraph.view.mxGraph;

import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.gui.GuiMessages;
import de.uni_due.inf.ti.raven.gui.Utilities;
import de.uni_due.inf.ti.raven.gui.graphs.VisualGraph;

/**
 * This class provides a panel which is intended to display the run of an 
 * {@code Automaton} on a {@code CospanDecomposition}. Since the automata are 
 * highly non-deterministic the runs are depicted as trees whose nodes are the 
 * reachable states of the underlying automata and whose edges are labeled by 
 * atomic cospans. 
 * @author Christoph Blume
 * @version $Revision: 1.4 $
 * 
 * @see Automaton
 * @see CospanDecomposition
 * @see Operation
 * @see VisualGraph
 *
 */
public final class AutomatonRunView extends mxGraphComponent {

	private static final long serialVersionUID = -1490070705232131487L;

	private final Logger logger;

	private final List<mxIEventListener> listeners;
	private VisualGraph visualGraph;

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * 
	 */
	public AutomatonRunView() {
		super(new mxGraph());

		listeners = new LinkedList<>();
		logger = Logger.getLogger(getClass().getName());
	}
	
	/**
	 * Adds the selection listener {@code listener} to the list of listeners.
	 * @param listener the listener to be registered
	 */
	public void addSelectionListener(mxIEventListener listener) {
		listeners.add(listener);
		if(visualGraph!=null) {			
			visualGraph.getSelectionModel().addListener(mxEvent.CHANGE, listener);
		}
	}
	
	/**
	 * Removes the selection listener {@code listener} from the list of listeners.
	 * @param listener the listener to be removed
	 */
	public void removeSelectionListener(mxIEventListener listener) {
		listeners.remove(listener);
		if(visualGraph!=null) {
			visualGraph.getSelectionModel().removeListener(listener, mxEvent.CHANGE);
		}
	}

	/**
	 * Prints the run of the automaton {@code automaton} on the cospan decomposition {@code decomposition}.
	 * @param automaton the automaton used for the run
	 * @param decomposition the cospan decomposition which is the input for this run
	 */
	public synchronized void print(final Automaton automaton, 
			final CospanDecomposition decomposition) {
		try {
			visualGraph = VisualGraph.getVisualGraph(automaton, decomposition);
		} catch (FileNotFoundException e) {
			logger.throwing(getClass().getName(), "print", e);
			Utilities.showWarningMessage(AutomatonRunView.this, 
					GuiMessages.DOT_PATH_NOT_FOUND_TITLE_EXCEPTION_KEY, 
					GuiMessages.DOT_PATH_NOT_FOUND_TEXT_EXCEPTION_KEY);
			return;
		} catch (IOException | InterruptedException e) {
			logger.throwing(getClass().getName(), "print", e);
			Utilities.showWarningMessage(AutomatonRunView.this, 
					GuiMessages.DOT_ERROR_OCCURRED_TITLE_EXCEPTION_KEY, 
					GuiMessages.DOT_ERROR_OCCURRED_TEXT_EXCEPTION_KEY);
			return;
		}

		setConnectable(false);
		setPanning(true);
		setAutoExtend(true);
		setGraph(visualGraph);

		final mxParallelEdgeLayout layout = new mxParallelEdgeLayout(visualGraph);
		layout.execute(visualGraph.getDefaultParent());
		visualGraph.addListener("cellsMoved", new mxIEventListener() {

			@Override
			public void invoke(Object sender, mxEventObject evt) {
				layout.execute(visualGraph.getDefaultParent());
			}
		});
		
		for(mxIEventListener listener: listeners) {
			visualGraph.getSelectionModel().addListener(mxEvent.CHANGE, listener);
		}

		getGraphControl().addMouseWheelListener(new MouseWheelListener() {

			@Override
			public void mouseWheelMoved(MouseWheelEvent e) {
				int rotation = e.getWheelRotation();
				if(rotation<0) {
					rotation *= -1;
					for(int i=0;i<rotation;i++) {
						AutomatonRunView.this.zoomIn();
					}
				}
				else if(rotation>0) {
					for(int i=0;i<rotation;i++) {
						AutomatonRunView.this.zoomOut();
					}
				}
			}
		});
		
		visualGraph.getSelectionModel().setSingleSelection(true);

		// Move graph center to center point of the panel
		//		int y = Math.max(SAVE_IMAGE_RECTANGLE.y + SAVE_IMAGE_RECTANGLE.height, 
		//				MAXIMIZE_VIEW_IMAGE_RECTANGLE.y + MAXIMIZE_VIEW_IMAGE_RECTANGLE.height);
		//		mxPoint translationPoint = new mxPoint(0, y+IMAGE_GAP);
		//		visualGraph.getView().setTranslate(translationPoint);

		mxRectangle bounds = visualGraph.getGraphBounds();
		setPreferredSize(new Dimension((int) bounds.getWidth(), (int) bounds.getHeight()));
		
		repaint();
	}

}
