/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui.automaton;

import java.text.DecimalFormat;

import javax.swing.JFormattedTextField;

import de.uni_due.inf.ti.raven.actions.ComputationListener;
import de.uni_due.inf.ti.raven.actions.DominatingSetAutomatonAction;
import de.uni_due.inf.ti.raven.data.Signature;
import de.uni_due.inf.ti.raven.data.automata.DominatingSetAutomaton;
import de.uni_due.inf.ti.raven.gui.GuiMessages;

/**
 * This class provides an {@code AutomatonCreationPanel} for creating a new {@code 
 * DominatingSetAutomaton}. 
 * @author Christoph Blume
 * @version $Revision: 1.14 $
 *
 * @see DominatingSetAutomaton
 * @see DominatingSetAutomatonAction
 */
public final class DominatingSetAutomatonCreationPanel extends AutomatonCreationPanel {

	private static final long serialVersionUID = -1114836716923047435L;

	/** Maximum Dominating Set Size Text Field Label (used for Testing) */
	protected static final String MAXIMUM_DOMINATING_SET_SIZE_LABEL = "maximumDominatingSetSize";
	
	private JFormattedTextField maximumDominatingSetSizeField;

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * 
	 */
	public DominatingSetAutomatonCreationPanel() {
		appendSeparator(GuiMessages.AUTOMATON_CREATION_ADVANCED_PROPERTIES_TEXT_KEY);
		
		maximumDominatingSetSizeField = new JFormattedTextField(new DecimalFormat("0"));
		maximumDominatingSetSizeField.setName(MAXIMUM_DOMINATING_SET_SIZE_LABEL);
		append(GuiMessages.MAXIMUM_DOMINATING_SET_SIZE_LABEL_KEY, maximumDominatingSetSizeField);
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	protected DominatingSetAutomaton createAutomaton(ComputationListener... listeners) {
		Integer maximumInterface = getInteger(maximumInterfaceField);
		Integer innerInterface = getInteger(innerInterfaceField);
		Integer outerInterface = getInteger(outerInterfaceField);
		Integer sizeOfDominatingSet = getInteger(maximumDominatingSetSizeField);
		Signature signature = (Signature) alphabetBox.getSelectedData();

		if(signature == null){
			showWarningMessage(GuiMessages.NO_SIGNATURE_SELECTED_TITLE_EXCEPTION_KEY, 
					GuiMessages.NO_SIGNATURE_SELECTED_TEXT_EXCEPTION_KEY);
			return null;
		}
		if(maximumInterface==null) {
			showWarningMessage(GuiMessages.EMPTY_INPUTFIELD_TITLE_EXCEPTION_KEY, 
					GuiMessages.EMPTY_INPUTFIELD_TEXT_EXCEPTION_KEY,
					GuiMessages.MAXIMUM_INTERFACE_LABEL_KEY); 
			return null;
		}
		if(maximumInterface<0) {
			showWarningMessage(GuiMessages.VALUE_MUST_BE_NON_NEGATIVE_TITLE_EXCEPTION_KEY, 
					GuiMessages.VALUE_MUST_BE_NON_NEGATIVE_TEXT_EXCEPTION_KEY,
					GuiMessages.MAXIMUM_INTERFACE_LABEL_KEY); 
			return null;
		}
		if(innerInterface==null) {
			showWarningMessage(GuiMessages.EMPTY_INPUTFIELD_TITLE_EXCEPTION_KEY, 
					GuiMessages.EMPTY_INPUTFIELD_TEXT_EXCEPTION_KEY,
					GuiMessages.INNER_INTERFACE_SIZE_LABEL_KEY); 
			return null;
		}
		if(innerInterface<0) {
			showWarningMessage(GuiMessages.VALUE_MUST_BE_NON_NEGATIVE_TITLE_EXCEPTION_KEY, 
					GuiMessages.VALUE_MUST_BE_NON_NEGATIVE_TEXT_EXCEPTION_KEY,
					GuiMessages.INNER_INTERFACE_SIZE_LABEL_KEY);
			return null;
		}
		if(innerInterface>maximumInterface) {
			showWarningMessage(GuiMessages.VALUE_GREATER_THAN_TITLE_EXCEPTION_KEY, 
					GuiMessages.VALUE_GREATER_THAN_TEXT_EXCEPTION_KEY,
					GuiMessages.INNER_INTERFACE_SIZE_LABEL_KEY, 
					GuiMessages.MAXIMUM_INTERFACE_LABEL_KEY);
			return null;
		}
		if(outerInterface==null) {
			showWarningMessage(GuiMessages.EMPTY_INPUTFIELD_TITLE_EXCEPTION_KEY, 
					GuiMessages.EMPTY_INPUTFIELD_TEXT_EXCEPTION_KEY,
					GuiMessages.OUTER_INTERFACE_SIZE_LABEL_KEY); 
			return null;
		}
		if(outerInterface<0) {
			showWarningMessage(GuiMessages.VALUE_MUST_BE_NON_NEGATIVE_TITLE_EXCEPTION_KEY, 
					GuiMessages.VALUE_MUST_BE_NON_NEGATIVE_TEXT_EXCEPTION_KEY,
					GuiMessages.OUTER_INTERFACE_SIZE_LABEL_KEY); 
			return null;
		}
		if(outerInterface>maximumInterface) {
			showWarningMessage(GuiMessages.VALUE_GREATER_THAN_TITLE_EXCEPTION_KEY, 
					GuiMessages.VALUE_GREATER_THAN_TEXT_EXCEPTION_KEY,
					GuiMessages.OUTER_INTERFACE_SIZE_LABEL_KEY, 
					GuiMessages.MAXIMUM_INTERFACE_LABEL_KEY);
			return null;
		}
		if(sizeOfDominatingSet==null) {
			showWarningMessage(GuiMessages.EMPTY_INPUTFIELD_TITLE_EXCEPTION_KEY, 
					GuiMessages.EMPTY_INPUTFIELD_TEXT_EXCEPTION_KEY,
					GuiMessages.MAXIMUM_DOMINATING_SET_SIZE_LABEL_KEY); 
			return null;
		}
		if (sizeOfDominatingSet < 1) {
			showWarningMessage(GuiMessages.VALUE_MUST_BE_POSITIVE_TITLE_EXCEPTION_KEY, 
					GuiMessages.VALUE_MUST_BE_POSITIVE_TEXT_EXCEPTION_KEY,
					GuiMessages.MAXIMUM_DOMINATING_SET_SIZE_LABEL_KEY);
			return null;
		}

		DominatingSetAutomatonAction action = new DominatingSetAutomatonAction(maximumInterface, 
				innerInterface, outerInterface, signature, sizeOfDominatingSet);
		for(ComputationListener listener: listeners) {
			action.addComputationListener(listener);
		}
		
		action.execute();
		return action.getResult();
	}

}
