/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui.automaton;

import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.text.DecimalFormat;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import javax.swing.AbstractAction;
import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JFormattedTextField;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.ListModel;
import javax.swing.ListSelectionModel;

import com.jgoodies.forms.builder.DefaultFormBuilder;
import com.jgoodies.forms.layout.FormLayout;

import de.uni_due.inf.ti.raven.actions.ComputationListener;
import de.uni_due.inf.ti.raven.actions.EdgeCountingAutomatonAction;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.Operation.Connect;
import de.uni_due.inf.ti.raven.data.Signature;
import de.uni_due.inf.ti.raven.data.automata.EdgeCountingAutomaton;
import de.uni_due.inf.ti.raven.gui.GuiMessages;

/**
 * This class provides an {@code AutomatonCreationPanel} for creating a new {@code 
 * EdgeCountingAutomaton}.
 * @author Christoph Blume
 * @version $Revision: 1.17 $
 *
 * @see EdgeCountingAutomaton
 * @see EdgeCountingAutomatonAction
 *
 */
public final class EdgeCountingAutomatonCreationPanel extends AutomatonCreationPanel {

	private static final long serialVersionUID = 5998135498211072174L;

	/** Divisor Text Field Label (used for Testing) */
	protected static final String DIVISOR_LABEL = "divisor";
	
	/** Remainder Text Field Label (used for Testing) */
	protected static final String REMAINDER_LABEL = "remainder";
	
	/** Available Edges List Label (used for Testing) */
	protected static final String AVAILABLE_EDGES_LABEL = "availableEdges";
	
	/** Selected Edges List Label (used for Testing) */
	protected static final String SELECTED_EDGES_LABEL = "selectedEdges";

	/** Shift-Right Button Label (used for Testing) */
	protected static final String SHIFT_RIGHT_LABEL = "shiftRight";

	/** Shift-Left Button Label (used for Testing) */
	protected static final String SHIFT_LEFT_LABEL = "shiftLeft";
	
	private JFormattedTextField remainderEdgeField;
	private JFormattedTextField divisorEdgeField;

	private JList<Connect> availableOperationsList;
	private JList<Connect> edgeLabelSet;

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns a {@code ListModel} object containing all {@code Connect}-operations contained in 
	 * the signature {@code signature.}
	 * @param signature the signature whose {@code Connect}-operations will be used 
	 * @return the {@code ListModel} containing all {@code Connect}-operations of the signature
	 * 
	 * @see ListModel
	 * @see Connect
	 */
	private static ListModel<Connect> getListModel(Signature signature) {
		DefaultListModel<Connect> model = new DefaultListModel<>();
		if(signature != null)
		{
			for(Operation letter: signature) {
				if(letter instanceof Connect) {
					model.addElement((Connect) letter);
				}
			}
		}
		return model;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * 
	 */
	public EdgeCountingAutomatonCreationPanel() {
		appendSeparator(GuiMessages.AUTOMATON_CREATION_ADVANCED_PROPERTIES_TEXT_KEY);
		
		divisorEdgeField = new JFormattedTextField(new DecimalFormat("0"));
		divisorEdgeField.setName(DIVISOR_LABEL);
		append(GuiMessages.DIVISOR_LABEL_KEY, divisorEdgeField);		

		remainderEdgeField = new JFormattedTextField(new DecimalFormat("0"));
		remainderEdgeField.setName(REMAINDER_LABEL);
		append(GuiMessages.REMAINDER_EDGES_LABEL_KEY, remainderEdgeField);
		
		append(createEdgeSelectionPanel(), 3);

		alphabetBox.addItemListener(new ItemListener() {

		@Override
		public void itemStateChanged(ItemEvent e) {
				if (e.getStateChange() == ItemEvent.SELECTED) {
					availableOperationsList.setModel(getListModel(
							(Signature) alphabetBox.getSelectedData()));
					edgeLabelSet.setModel(new DefaultListModel<Connect>());
				}
			}
		});
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	protected EdgeCountingAutomaton createAutomaton(ComputationListener... listeners) {
		Integer maximumInterface = getInteger(maximumInterfaceField);
		Integer innerInterface = getInteger(innerInterfaceField);
		Integer outerInterface = getInteger(outerInterfaceField);
		Integer divisor = getInteger(divisorEdgeField);
		Integer remainder = getInteger(remainderEdgeField);
		Signature signature = (Signature) alphabetBox.getSelectedData();
		DefaultListModel<Connect> model = (DefaultListModel<Connect>) edgeLabelSet.getModel();
		Set<Connect> edgeSet = new HashSet<>(Collections.list(model.elements()));

		if(signature == null){
			showWarningMessage(GuiMessages.NO_SIGNATURE_SELECTED_TITLE_EXCEPTION_KEY, 
					GuiMessages.NO_SIGNATURE_SELECTED_TEXT_EXCEPTION_KEY,
					GuiMessages.SIGNATURE_LABEL_KEY);
			return null;
		}
		if(maximumInterface==null) {
			showWarningMessage(GuiMessages.EMPTY_INPUTFIELD_TITLE_EXCEPTION_KEY, 
					GuiMessages.EMPTY_INPUTFIELD_TEXT_EXCEPTION_KEY,
					GuiMessages.MAXIMUM_INTERFACE_LABEL_KEY); 
			return null;
		}
		if(maximumInterface<0) {
			showWarningMessage(GuiMessages.VALUE_MUST_BE_NON_NEGATIVE_TITLE_EXCEPTION_KEY, 
					GuiMessages.VALUE_MUST_BE_NON_NEGATIVE_TEXT_EXCEPTION_KEY,
					GuiMessages.MAXIMUM_INTERFACE_LABEL_KEY); 
			return null;
		}
		if(innerInterface==null) {
			showWarningMessage(GuiMessages.EMPTY_INPUTFIELD_TITLE_EXCEPTION_KEY, 
					GuiMessages.EMPTY_INPUTFIELD_TEXT_EXCEPTION_KEY,
					GuiMessages.INNER_INTERFACE_SIZE_LABEL_KEY); 
			return null;
		}
		if(innerInterface<0) {
			showWarningMessage(GuiMessages.VALUE_MUST_BE_NON_NEGATIVE_TITLE_EXCEPTION_KEY, 
					GuiMessages.VALUE_MUST_BE_NON_NEGATIVE_TEXT_EXCEPTION_KEY,
					GuiMessages.INNER_INTERFACE_SIZE_LABEL_KEY);
			return null;
		}
		if(innerInterface>maximumInterface) {
			showWarningMessage(GuiMessages.VALUE_GREATER_THAN_TITLE_EXCEPTION_KEY, 
					GuiMessages.VALUE_GREATER_THAN_TEXT_EXCEPTION_KEY,
					GuiMessages.INNER_INTERFACE_SIZE_LABEL_KEY, 
					GuiMessages.MAXIMUM_INTERFACE_LABEL_KEY);
			return null;
		}
		if(outerInterface==null) {
			showWarningMessage(GuiMessages.EMPTY_INPUTFIELD_TITLE_EXCEPTION_KEY, 
					GuiMessages.EMPTY_INPUTFIELD_TEXT_EXCEPTION_KEY,
					GuiMessages.OUTER_INTERFACE_SIZE_LABEL_KEY); 
			return null;
		}
		if(outerInterface<0) {
			showWarningMessage(GuiMessages.VALUE_MUST_BE_NON_NEGATIVE_TITLE_EXCEPTION_KEY, 
					GuiMessages.VALUE_MUST_BE_NON_NEGATIVE_TEXT_EXCEPTION_KEY,
					GuiMessages.OUTER_INTERFACE_SIZE_LABEL_KEY); 
			return null;
		}
		if(outerInterface>maximumInterface) {
			showWarningMessage(GuiMessages.VALUE_GREATER_THAN_TITLE_EXCEPTION_KEY, 
					GuiMessages.VALUE_GREATER_THAN_TEXT_EXCEPTION_KEY,
					GuiMessages.OUTER_INTERFACE_SIZE_LABEL_KEY, 
					GuiMessages.MAXIMUM_INTERFACE_LABEL_KEY);
			return null;
		}
		if(divisor==null) {
			showWarningMessage(GuiMessages.EMPTY_INPUTFIELD_TITLE_EXCEPTION_KEY, 
					GuiMessages.EMPTY_INPUTFIELD_TEXT_EXCEPTION_KEY,
					GuiMessages.DIVISOR_LABEL_KEY); 
			return null;
		}
		if(divisor<1) {
			showWarningMessage(GuiMessages.VALUE_MUST_BE_POSITIVE_TITLE_EXCEPTION_KEY, 
					GuiMessages.VALUE_MUST_BE_POSITIVE_TEXT_EXCEPTION_KEY,
					GuiMessages.DIVISOR_LABEL_KEY); 
			return null;
		}
		if(remainder==null) {
			showWarningMessage(GuiMessages.EMPTY_INPUTFIELD_TITLE_EXCEPTION_KEY, 
					GuiMessages.EMPTY_INPUTFIELD_TEXT_EXCEPTION_KEY,
					GuiMessages.REMAINDER_EDGES_LABEL_KEY); 
			return null;
		}
		if(remainder<0) {
			showWarningMessage(GuiMessages.VALUE_MUST_BE_NON_NEGATIVE_TITLE_EXCEPTION_KEY, 
					GuiMessages.VALUE_MUST_BE_NON_NEGATIVE_TEXT_EXCEPTION_KEY,
					GuiMessages.REMAINDER_EDGES_LABEL_KEY); 
			return null;
		}
		if(remainder>=divisor) {
			showWarningMessage(GuiMessages.VALUE_GREATER_THAN_TITLE_EXCEPTION_KEY, 
					GuiMessages.VALUE_GREATER_THAN_TEXT_EXCEPTION_KEY,
					GuiMessages.REMAINDER_EDGES_LABEL_KEY, 
					GuiMessages.DIVISOR_LABEL_KEY);
			return null;
		}
		if(edgeSet.isEmpty()) {
			showWarningMessage(GuiMessages.EMPTY_LIST_TITLE_EXCEPTION_KEY, 
					GuiMessages.EMPTY_LIST_TEXT_EXCEPTION_KEY, 
					GuiMessages.EDGE_SELECTION_LABEL_KEY);
			return null;
		}

		EdgeCountingAutomatonAction action = new EdgeCountingAutomatonAction(maximumInterface, 
				innerInterface, outerInterface, signature, divisor, remainder, edgeSet);
		for(ComputationListener listener: listeners) {
			action.addComputationListener(listener);
		}
		
		action.execute();
		return action.getResult();
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a panel for (de)selecting the edges (or {@code Connect}-operations to be precise) 
	 * which will be counted by the edge-counting automaton.
	 * @return the edge selection panel
	 */
	private JPanel createEdgeSelectionPanel() {
		JScrollPane scrollpane;

		FormLayout layout = new FormLayout("40dlu:grow, 3dlu, left:pref, 3dlu, 40dlu:grow", "");
		DefaultFormBuilder builder = new DefaultFormBuilder(layout);

		builder.append(new JLabel(messages.getString(GuiMessages.EDGE_SELECTION_LABEL_KEY)), 5);
		builder.nextLine();

		availableOperationsList = new JList<>(getListModel(
				(Signature) alphabetBox.getSelectedData()));
		availableOperationsList.setName(AVAILABLE_EDGES_LABEL);
		availableOperationsList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		availableOperationsList.setModel(getListModel(
				(Signature) alphabetBox.getSelectedData()));
		scrollpane = new JScrollPane(availableOperationsList);
		scrollpane.setPreferredSize(new Dimension(150, 100));
		builder.append(scrollpane);

		JButton shiftRightButton = new JButton(new AbstractAction(">>") {

			private static final long serialVersionUID = -2632710493259765026L;

			@Override
			public void actionPerformed(ActionEvent e) {
				DefaultListModel<Connect> leftModel = (DefaultListModel<Connect>) 
						availableOperationsList.getModel();
				DefaultListModel<Connect> rightModel = (DefaultListModel<Connect>) 
						edgeLabelSet.getModel();

				int index = availableOperationsList.getSelectedIndex();
				if(index>-1) {
					rightModel.addElement(leftModel.remove(index));
				}
			}
		});
		shiftRightButton.setName(SHIFT_RIGHT_LABEL);

		JButton shiftLeftButton = new JButton(new AbstractAction("<<") {

			private static final long serialVersionUID = -5067848443035530041L;

			@Override
			public void actionPerformed(ActionEvent e) {
				DefaultListModel<Connect> leftModel = (DefaultListModel<Connect>) 
						availableOperationsList.getModel();
				DefaultListModel<Connect> rightModel = (DefaultListModel<Connect>) 
						edgeLabelSet.getModel();

				int index = edgeLabelSet.getSelectedIndex();
				if(index>-1) {
					leftModel.addElement(rightModel.remove(index));
				}
			}
		});
		shiftLeftButton.setName(SHIFT_LEFT_LABEL);

		JPanel buttonPanel = new JPanel(new GridLayout(2, 1, 0, 5));
		buttonPanel.add(shiftRightButton);
		buttonPanel.add(shiftLeftButton);
		builder.append(buttonPanel);

		edgeLabelSet = new JList<>();
		edgeLabelSet.setName(SELECTED_EDGES_LABEL);
		edgeLabelSet.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		edgeLabelSet.setModel(new DefaultListModel<Connect>());
		scrollpane = new JScrollPane(edgeLabelSet);
		scrollpane.setPreferredSize(new Dimension(150, 100));
		builder.append(scrollpane);

		return builder.getPanel();
	}

}
