/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui.automaton;

import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import javax.swing.AbstractAction;
import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.ListModel;
import javax.swing.ListSelectionModel;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;

import com.jgoodies.forms.builder.DefaultFormBuilder;
import com.jgoodies.forms.layout.FormLayout;

import de.uni_due.inf.ti.raven.actions.ComputationListener;
import de.uni_due.inf.ti.raven.actions.PathAutomatonAction;
import de.uni_due.inf.ti.raven.data.Signature;
import de.uni_due.inf.ti.raven.data.automata.PathAutomaton;
import de.uni_due.inf.ti.raven.gui.GuiMessages;

/**
 * This class provides an {@code AutomatonCreationPanel} for creating a new {@code 
 * PathAutomaton}.
 * @author Christoph Blume, Dennis Nolte
 * @version $Revision: 1.14 $
 *
 * @see PathAutomaton
 * @see PathAutomatonAction
 */
public final class PathAutomatonCreationPanel extends AutomatonCreationPanel {

	private static final String LAYOUT_SPECIFICATION = "40dlu:grow, 3dlu, left:pref, 3dlu, 40dlu:grow";

	private static final long serialVersionUID = 8295074236284917134L;

	/** Available Source Nodes List Label (used for Testing) */
	protected static final String AVAILABLE_SOURCE_NODES_LIST_LABEL = "availableSourceNodes";
	
	/** Selected Source Nodes List Label (used for Testing) */
	protected static final String SELECTED_SOURCE_NODES_LIST_LABEL = "selectedSourceNodes";

	/** Shift-Right Source Nodes Button Label (used for Testing) */
	protected static final String SHIFT_SOURCE_NODES_RIGHT_LABEL = "shiftRightSource";

	/** Shift-Left Source Nodes Button Label (used for Testing) */
	protected static final String SHIFT_SOURCE_NODES_LEFT_LABEL = "shiftLeftSource";
	
	/** Available Target Nodes List Label (used for Testing) */
	protected static final String AVAILABLE_TARGET_NODES_LIST_LABEL = "availableTargetNodes";
	
	/** Selected Target Nodes List Label (used for Testing) */
	protected static final String SELECTED_TARGET_NODES_LIST_LABEL = "selectedTargetNodes";
	
	/** Shift-Right Target Nodes Button Label (used for Testing) */
	protected static final String SHIFT_TARGET_NODES_RIGHT_LABEL = "shiftRightTarget";

	/** Shift-Left Target Nodes Button Label (used for Testing) */
	protected static final String SHIFT_TARGET_NODES_LEFT_LABEL = "shiftLeftTarget";
	
	private JList<Integer> availableSourceNodeList;
	private JList<Integer> selectedSourceNodeList;
	private JList<Integer> availableTargetNodeList;
	private JList<Integer> selectedTargetNodeList;

	/*------------------------------------------------------------------------------------------*/
	/*		Static	Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a list model containing the number from 0 to {@code size-1} as elements.
	 * @param size the number of elements
	 * @return a list model for the numbers 0 to {@code size-1}
	 */
	private static ListModel<Integer> getListModel(Integer size) {
		DefaultListModel<Integer> model = new DefaultListModel<>();
		if(size != null && size > 0) {
			for(int i=0;i<size;i++) {
				model.addElement(i);
			}
		}

		return model;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new creation panel for the instantiation of a new path automaton.
	 */
	protected PathAutomatonCreationPanel() {
		appendSeparator(GuiMessages.AUTOMATON_CREATION_ADVANCED_PROPERTIES_TEXT_KEY);
	
		append(createNodeSelectionPanel(), 3);
		
		innerInterfaceField.getDocument().addDocumentListener(new DocumentListener() {
			
			@Override
			public void removeUpdate(DocumentEvent e) {
				availableSourceNodeList.setModel(getListModel(getInteger(innerInterfaceField)));
				selectedSourceNodeList.setModel(new DefaultListModel<Integer>());
			}
			
			@Override
			public void insertUpdate(DocumentEvent e) {
				availableSourceNodeList.setModel(getListModel(getInteger(innerInterfaceField)));
				selectedSourceNodeList.setModel(new DefaultListModel<Integer>());				
			}
			
			@Override
			public void changedUpdate(DocumentEvent e) {
				availableSourceNodeList.setModel(getListModel(getInteger(innerInterfaceField)));
				selectedSourceNodeList.setModel(new DefaultListModel<Integer>());
			}
		});
		
		outerInterfaceField.getDocument().addDocumentListener(new DocumentListener() {
			
			@Override
			public void removeUpdate(DocumentEvent e) {
				availableTargetNodeList.setModel(getListModel(getInteger(outerInterfaceField)));
				selectedTargetNodeList.setModel(new DefaultListModel<Integer>());
			}
			
			@Override
			public void insertUpdate(DocumentEvent e) {
				availableTargetNodeList.setModel(getListModel(getInteger(outerInterfaceField)));
				selectedTargetNodeList.setModel(new DefaultListModel<Integer>());
			}
			
			@Override
			public void changedUpdate(DocumentEvent e) {
				availableTargetNodeList.setModel(getListModel(getInteger(outerInterfaceField)));
				selectedTargetNodeList.setModel(new DefaultListModel<Integer>());
			}
		});
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods																	*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	protected PathAutomaton createAutomaton(ComputationListener... listeners) {
		Integer maximumInterface = getInteger(maximumInterfaceField);
		Integer innerInterface = getInteger(innerInterfaceField);
		Integer outerInterface= getInteger(outerInterfaceField);
		DefaultListModel<Integer> model = (DefaultListModel<Integer>) selectedSourceNodeList.getModel();
		Set<Integer> sourceSet = new HashSet<>(Collections.list(model.elements()));
		model = (DefaultListModel<Integer>) selectedTargetNodeList.getModel();
		Set<Integer> targetSet = new HashSet<>(Collections.list(model.elements()));
		
		Signature signature = (Signature) alphabetBox.getSelectedData();

		if(signature == null){
			showWarningMessage(GuiMessages.NO_SIGNATURE_SELECTED_TITLE_EXCEPTION_KEY, 
					GuiMessages.NO_SIGNATURE_SELECTED_TEXT_EXCEPTION_KEY,
					GuiMessages.SIGNATURE_LABEL_KEY);
			return null;
		}
		if(maximumInterface == null){
			showWarningMessage(GuiMessages.EMPTY_INPUTFIELD_TITLE_EXCEPTION_KEY, 
					GuiMessages.EMPTY_INPUTFIELD_TEXT_EXCEPTION_KEY,
					GuiMessages.MAXIMUM_INTERFACE_LABEL_KEY);
			return null;
		}
		if(maximumInterface<0) {
			showWarningMessage(GuiMessages.VALUE_MUST_BE_NON_NEGATIVE_TITLE_EXCEPTION_KEY, 
					GuiMessages.VALUE_MUST_BE_NON_NEGATIVE_TEXT_EXCEPTION_KEY,
					GuiMessages.MAXIMUM_INTERFACE_LABEL_KEY); 
			return null;
		}
		if(innerInterface == null){
			showWarningMessage(GuiMessages.EMPTY_INPUTFIELD_TITLE_EXCEPTION_KEY, 
					GuiMessages.EMPTY_INPUTFIELD_TEXT_EXCEPTION_KEY,
					GuiMessages.INNER_INTERFACE_SIZE_LABEL_KEY);
			return null;
		}
		if(innerInterface<0) {
			showWarningMessage(GuiMessages.VALUE_MUST_BE_NON_NEGATIVE_TITLE_EXCEPTION_KEY, 
					GuiMessages.VALUE_MUST_BE_NON_NEGATIVE_TEXT_EXCEPTION_KEY,
					GuiMessages.INNER_INTERFACE_SIZE_LABEL_KEY);
			return null;
		}
		if(innerInterface>maximumInterface) {
			showWarningMessage(GuiMessages.VALUE_GREATER_THAN_TITLE_EXCEPTION_KEY, 
					GuiMessages.VALUE_GREATER_THAN_TEXT_EXCEPTION_KEY,
					GuiMessages.INNER_INTERFACE_SIZE_LABEL_KEY, 
					GuiMessages.MAXIMUM_INTERFACE_LABEL_KEY);
			return null;
		}
		if(outerInterface == null){
			showWarningMessage(GuiMessages.EMPTY_INPUTFIELD_TITLE_EXCEPTION_KEY, 
					GuiMessages.EMPTY_INPUTFIELD_TEXT_EXCEPTION_KEY,
					GuiMessages.OUTER_INTERFACE_SIZE_LABEL_KEY);
			return null;
		}
		if(outerInterface<0) {
			showWarningMessage(GuiMessages.VALUE_MUST_BE_NON_NEGATIVE_TITLE_EXCEPTION_KEY, 
					GuiMessages.VALUE_MUST_BE_NON_NEGATIVE_TEXT_EXCEPTION_KEY,
					GuiMessages.OUTER_INTERFACE_SIZE_LABEL_KEY); 
			return null;
		}
		if(outerInterface>maximumInterface) {
			showWarningMessage(GuiMessages.VALUE_GREATER_THAN_TITLE_EXCEPTION_KEY, 
					GuiMessages.VALUE_GREATER_THAN_TEXT_EXCEPTION_KEY,
					GuiMessages.OUTER_INTERFACE_SIZE_LABEL_KEY, 
					GuiMessages.MAXIMUM_INTERFACE_LABEL_KEY);
			return null;
		}
						
		PathAutomatonAction action = new PathAutomatonAction(maximumInterface, 
				innerInterface, outerInterface, signature, sourceSet, targetSet);
		for(ComputationListener listener: listeners) {
			action.addComputationListener(listener);
		}
		
		action.execute();
		return action.getResult();
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a panel which holds the controls for the selection of the node and target nodes.
	 * @return the node selection panel
	 */
	private JPanel createNodeSelectionPanel() {
		JScrollPane scrollpane;

		FormLayout layout = new FormLayout(LAYOUT_SPECIFICATION, "");
		DefaultFormBuilder builder = new DefaultFormBuilder(layout);

		builder.append(new JLabel(messages.getString(GuiMessages.SOURCE_NODES_LABEL_KEY)), 5);
		builder.nextLine();

		availableSourceNodeList = new JList<>(getListModel(getInteger(innerInterfaceField)));
		availableSourceNodeList.setName(AVAILABLE_SOURCE_NODES_LIST_LABEL);
		availableSourceNodeList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		availableSourceNodeList.setModel(getListModel(getInteger(innerInterfaceField)));
		scrollpane = new JScrollPane(availableSourceNodeList);
		builder.append(scrollpane);

		JButton shiftSourceRightButton = new JButton(new AbstractAction(">>") {

			private static final long serialVersionUID = -2632710493259765026L;

			@Override
			public void actionPerformed(ActionEvent e) {
				DefaultListModel<Integer> leftModel = (DefaultListModel<Integer>) 
						availableSourceNodeList.getModel();
				DefaultListModel<Integer> rightModel = (DefaultListModel<Integer>) 
						selectedSourceNodeList.getModel();

				int index = availableSourceNodeList.getSelectedIndex();
				if(index>-1) {
					rightModel.addElement(leftModel.remove(index));
				}
			}
		});
		shiftSourceRightButton.setName(SHIFT_SOURCE_NODES_RIGHT_LABEL);

		JButton shiftSourceLeftButton = new JButton(new AbstractAction("<<") {

			private static final long serialVersionUID = -5067848443035530041L;

			@Override
			public void actionPerformed(ActionEvent e) {
				DefaultListModel<Integer> leftModel = (DefaultListModel<Integer>) 
						availableSourceNodeList.getModel();
				DefaultListModel<Integer> rightModel = (DefaultListModel<Integer>) 
						selectedSourceNodeList.getModel();

				int index = selectedSourceNodeList.getSelectedIndex();
				if(index>-1) {
					leftModel.addElement(rightModel.remove(index));
				}
			}
		});
		shiftSourceLeftButton.setName(SHIFT_SOURCE_NODES_LEFT_LABEL);

		JPanel buttonPanel = new JPanel(new GridLayout(2, 1, 0, 5));
		buttonPanel.add(shiftSourceRightButton);
		buttonPanel.add(shiftSourceLeftButton);
		builder.append(buttonPanel);

		selectedSourceNodeList = new JList<>();
		selectedSourceNodeList.setName(SELECTED_SOURCE_NODES_LIST_LABEL);
		selectedSourceNodeList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		selectedSourceNodeList.setModel(new DefaultListModel<Integer>());
		scrollpane = new JScrollPane(selectedSourceNodeList);
		builder.append(scrollpane);
		
		builder.nextLine();
		builder.append(new JLabel(messages.getString(GuiMessages.TARGET_NODES_LABEL_KEY)), 5);
		builder.nextLine();

		availableTargetNodeList = new JList<>(getListModel(getInteger(outerInterfaceField)));
		availableTargetNodeList.setName(AVAILABLE_TARGET_NODES_LIST_LABEL);
		availableTargetNodeList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		availableTargetNodeList.setModel(getListModel(
				getInteger(outerInterfaceField)));
		scrollpane = new JScrollPane(availableTargetNodeList);
		builder.append(scrollpane);

		JButton shiftTargetRightButton = new JButton(new AbstractAction(">>") {

			private static final long serialVersionUID = -2632710493259765026L;

			@Override
			public void actionPerformed(ActionEvent e) {
				DefaultListModel<Integer> leftModel = (DefaultListModel<Integer>) 
						availableTargetNodeList.getModel();
				DefaultListModel<Integer> rightModel = (DefaultListModel<Integer>) 
						selectedTargetNodeList.getModel();

				int index = availableTargetNodeList.getSelectedIndex();
				if(index>-1) {
					rightModel.addElement(leftModel.remove(index));
				}
			}
		});
		shiftTargetRightButton.setName(SHIFT_TARGET_NODES_RIGHT_LABEL);

		JButton shiftTargetLeftButton = new JButton(new AbstractAction("<<") {

			private static final long serialVersionUID = -5067848443035530041L;

			@Override
			public void actionPerformed(ActionEvent e) {
				DefaultListModel<Integer> leftModel = (DefaultListModel<Integer>) 
						availableTargetNodeList.getModel();
				DefaultListModel<Integer> rightModel = (DefaultListModel<Integer>) 
						selectedTargetNodeList.getModel();

				int index = selectedTargetNodeList.getSelectedIndex();
				if(index>-1) {
					leftModel.addElement(rightModel.remove(index));
				}
			}
		});
		shiftTargetLeftButton.setName(SHIFT_TARGET_NODES_LEFT_LABEL);

		buttonPanel = new JPanel(new GridLayout(2, 1, 0, 5));
		buttonPanel.add(shiftTargetRightButton);
		buttonPanel.add(shiftTargetLeftButton);
		builder.append(buttonPanel);

		selectedTargetNodeList = new JList<>();
		selectedTargetNodeList.setName(SELECTED_TARGET_NODES_LIST_LABEL);
		selectedTargetNodeList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		selectedTargetNodeList.setModel(new DefaultListModel<Integer>());
		scrollpane = new JScrollPane(selectedTargetNodeList);
		builder.append(scrollpane);
	
		return builder.getPanel();
	}
	
}
