/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui.automaton;

import de.uni_due.inf.ti.raven.actions.ComputationListener;
import de.uni_due.inf.ti.raven.actions.ProductAutomatonAction;
import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.data.automata.ProductAutomaton;
import de.uni_due.inf.ti.raven.gui.DataComboBox;
import de.uni_due.inf.ti.raven.gui.GuiMessages;
import de.uni_due.inf.ti.raven.gui.MainWindow;

/**
 * This class provides an {@code AutomatonCreationPanel} for creating a new {@code 
 * ProductAutomaton}.
 * @author Christoph Blume
 * @version $Revision: 1.10 $
 *
 * @see ProductAutomaton
 * @see ProductAutomatonAction
 *
 */
public final class ProductAutomatonCreationPanel extends AutomatonCreationPanel {

	private static final long serialVersionUID = 5998135498211072174L;

	/** First Automaton Combo Box Label (Used for Testing) */
	protected static final String FIRST_AUTOMATON_BOX_LABEL = "firstAutomaton";
	
	/** Second Automaton Combo Box Label (Used for Testing) */
	protected static final String SECOND_AUTOMATON_BOX_LABEL = "secondAutomaton";
	
	private DataComboBox firstAutomatonBox;
	private DataComboBox secondAutomatonBox;

	/**
	 * 
	 */
	public ProductAutomatonCreationPanel() {
		super(false);
		
		Repository repo = Repository.getRepository();
		firstAutomatonBox = new DataComboBox(repo);
		firstAutomatonBox.setName(FIRST_AUTOMATON_BOX_LABEL);
		firstAutomatonBox.addDataClass(Automaton.class);
		firstAutomatonBox.setOpaque(false);
		firstAutomatonBox.setBackground(MainWindow.BACKGROUND_COLOR);
		append(GuiMessages.FIRST_AUTOMATON_LABEL_KEY, firstAutomatonBox);

		secondAutomatonBox = new DataComboBox(repo);
		secondAutomatonBox.setName(SECOND_AUTOMATON_BOX_LABEL);
		secondAutomatonBox.addDataClass(Automaton.class);
		secondAutomatonBox.setOpaque(false);
		secondAutomatonBox.setBackground(MainWindow.BACKGROUND_COLOR);
		append(GuiMessages.SECOND_AUTOMATON_LABEL_KEY, secondAutomatonBox);		
	}

	@Override
	protected ProductAutomaton createAutomaton(ComputationListener... listeners) {
		Automaton first = (Automaton) firstAutomatonBox.getSelectedData();
		Automaton second = (Automaton) secondAutomatonBox.getSelectedData();

		if(first==null) {
			showWarningMessage(GuiMessages.NO_AUTOMATON_SELECTED_TITLE_EXCEPTION_KEY, 
					GuiMessages.NO_AUTOMATON_SELECTED_TEXT_EXCEPTION_KEY,
					GuiMessages.FIRST_AUTOMATON_LABEL_KEY); 
			return null;
		}
		if(second==null) {
			showWarningMessage(GuiMessages.NO_AUTOMATON_SELECTED_TITLE_EXCEPTION_KEY, 
					GuiMessages.NO_AUTOMATON_SELECTED_TEXT_EXCEPTION_KEY,
					GuiMessages.SECOND_AUTOMATON_LABEL_KEY); 
			return null;
		}
		if(first.equals(second)) {
			showWarningMessage(GuiMessages.AUTOMATA_MUST_NOT_BE_EQUAL_TITLE_EXCEPTION_KEY, 
					GuiMessages.AUTOMATA_MUST_NOT_BE_EQUAL_TEXT_EXCEPTION_KEY, 
					GuiMessages.FIRST_AUTOMATON_LABEL_KEY,
					GuiMessages.SECOND_AUTOMATON_LABEL_KEY);
			return null;
		}
		if(!first.getAlphabet().equals(second.getAlphabet())) {
			showWarningMessage(GuiMessages.ALPHABETS_MUST_BE_EQUAL_TITLE_EXCEPTION_KEY, 
					GuiMessages.ALPHABETS_MUST_BE_EQUAL_TEXT_EXCEPTION_KEY);
			return null;
		}
		if(first.getMaximumInterface()!=second.getMaximumInterface()) {
			showWarningMessage(GuiMessages.MAXIMUM_INTERFACE_SIZES_MUST_BE_EQUAL_TITLE_EXCEPTION_KEY, 
					GuiMessages.MAXIMUM_INTERFACE_SIZES_MUST_BE_EQUAL_TEXT_EXCEPTION_KEY);
			return null;
		}
		if(first.getInnerInterface()!=second.getInnerInterface()) {
			showWarningMessage(GuiMessages.INNER_INTERFACE_SIZES_MUST_BE_EQUAL_TITLE_EXCEPTION_KEY, 
					GuiMessages.INNER_INTERFACE_SIZES_MUST_BE_EQUAL_TEXT_EXCEPTION_KEY);
			return null;
		}
		if(first.getOuterInterface()!=second.getOuterInterface()) {
			showWarningMessage(GuiMessages.OUTER_INTERFACE_SIZES_MUST_BE_EQUAL_TITLE_EXCEPTION_KEY, 
					GuiMessages.OUTER_INTERFACE_SIZES_MUST_BE_EQUAL_TEXT_EXCEPTION_KEY);
			return null;
		}
		
		ProductAutomatonAction action = new ProductAutomatonAction(first, second);
		for(ComputationListener listener: listeners) {
			action.addComputationListener(listener);
		}
		
		action.execute();
		return action.getResult();
	}

}
