/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
*/

package de.uni_due.inf.ti.raven.gui.automaton;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.List;

import javax.swing.JButton;
import javax.swing.JTextField;

import de.uni_due.inf.ti.raven.data.automata.SubgraphAutomaton;
import de.uni_due.inf.ti.raven.gui.GraphWindow;
import de.uni_due.inf.ti.raven.gui.GuiMessages;
import de.uni_due.inf.ti.raven.gui.InformationPanel;

/**
 * This class provides an {@code InformationPanel} for displaying information about subgraph graph 
 * automata.
 * @author Christoph Blume
 * @version $Revision: 1.7 $
 *
 * @see AutomatonInformationPanelBuilder
 * @see SubgraphAutomaton
 */
public final class SubgraphAutomatonInformationPanelBuilder extends AutomatonInformationPanelBuilder {
	
	/**
	 * 
	 * @author Christoph Blume
	 *
	 */
	private static class SubgraphAutomatonInformationPanel extends AutomatonInformationPanel<SubgraphAutomaton> {

		private static final long serialVersionUID = -3267332555302461823L;

		private static String orderToString(List<?> order) {
			StringBuilder builder = new StringBuilder();
			for(Object o: order) {
				if(builder.length()>0) {
					builder.append(" < ");
				}

				builder.append(o.toString());
			}
			
			return builder.toString();
		}
		
		/*------------------------------------------------------------------------------------------*/
		/* 		Public Methods																		*/
		/*------------------------------------------------------------------------------------------*/
		
		@Override
		public void setAdditionalData(final SubgraphAutomaton automaton) {
			subgraphNameField.setText(repo.getName(automaton.getSubgraph()));
			
			showSubgraphButton.addActionListener(new ActionListener() {

				
				@Override
				public void actionPerformed(ActionEvent e) {
					GraphWindow window = GraphWindow.getWindow();
					window.print(automaton.getSubgraph());
				}
			});
			
			nodeOrderField.setText(orderToString(automaton.getNodeOrder()));
			
			edgeOrderField.setText(orderToString(automaton.getEdgeOrder()));
		}
		
	}

	/** Subgraph Name Text Field Label (used for Testing) */
	protected static final String SUBGRAPH_NAME_LABEL = "subgraphName";
	
	/** Show Subgraph Button Label (used for Testing) */
	protected static final String SHOW_SUBGRAPH_LABEL = "showSubgraph";

	/** Node Order Text Field Label (used for Testing) */
	protected static final String NODE_ORDER_LABEL = "nodeOrder";

	/** Edge Order Text Field Label (used for Testing) */
	protected static final String EDGE_ORDER_LABEL = "edgeOrder";
	
	private static SubgraphAutomatonInformationPanelBuilder panelBuilder;
	private static AutomatonInformationPanel<SubgraphAutomaton> panel;
	private static JTextField subgraphNameField;
	private static JTextField nodeOrderField;
	private static JTextField edgeOrderField;
	private static JButton showSubgraphButton;

	/*------------------------------------------------------------------------------------------*/
	/* 		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates an information panel displaying properties of a given subgraph automaton.
	 * @return the edge counting automaton information panel
	 */
	public static InformationPanel<SubgraphAutomaton> createPanel() {
		if(panelBuilder==null) {
			panel = new SubgraphAutomatonInformationPanel();
			panelBuilder = new SubgraphAutomatonInformationPanelBuilder(panel);
		}
	
		return panel;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates the panel on which automaton information will be displayed.
	 * @param panel the form builder used to build the panel
	 */
	private SubgraphAutomatonInformationPanelBuilder(
			AutomatonInformationPanel<SubgraphAutomaton> panel) {
		super(panel);
	}

	@Override
	protected void addAdditionalComponents() {
		subgraphNameField = new JTextField(10);
		subgraphNameField.setName(SUBGRAPH_NAME_LABEL);
		subgraphNameField.setEditable(false);
		subgraphNameField.setOpaque(false);
		append(GuiMessages.SUBGRAPH_LABEL_KEY, subgraphNameField);
		
		showSubgraphButton = new JButton();
		showSubgraphButton.setName(SHOW_SUBGRAPH_LABEL);
		appendButton(GuiMessages.SHOW_SUBGRAPH_LABEL_KEY, showSubgraphButton);
		
		nodeOrderField = new JTextField(10);
		nodeOrderField.setName(NODE_ORDER_LABEL);
		nodeOrderField.setEditable(false);
		nodeOrderField.setOpaque(false);
		append(GuiMessages.NODE_ORDER_LABEL_KEY, nodeOrderField);
		
		edgeOrderField = new JTextField(10);
		edgeOrderField.setName(EDGE_ORDER_LABEL);
		edgeOrderField.setEditable(false);
		edgeOrderField.setOpaque(false);
		append(GuiMessages.EDGE_ORDER_LABEL_KEY, edgeOrderField);

	}
	
}
