/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui.graphs;

import java.awt.Component;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.lang.ProcessBuilder.Redirect;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import org.antlr.v4.runtime.ANTLRInputStream;
import org.antlr.v4.runtime.CommonTokenStream;
import org.antlr.v4.runtime.tree.ParseTree;

import com.mxgraph.swing.mxGraphComponent;
import com.mxgraph.util.mxPoint;
import com.mxgraph.util.mxRectangle;
import com.mxgraph.view.mxGraph;

import de.uni_due.inf.ti.raven.antlr.DOTBaseVisitor;
import de.uni_due.inf.ti.raven.antlr.DOTLexer;
import de.uni_due.inf.ti.raven.antlr.DOTParser;
import de.uni_due.inf.ti.raven.antlr.DOTParser.A_listContext;
import de.uni_due.inf.ti.raven.antlr.DOTParser.Edge_stmtContext;
import de.uni_due.inf.ti.raven.antlr.DOTParser.Node_idContext;
import de.uni_due.inf.ti.raven.antlr.DOTParser.Node_stmtContext;
import de.uni_due.inf.ti.raven.antlr.DOTVisitor;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.encodings.BDDEncoding;
import de.uni_due.inf.ti.raven.gui.GuiMessages;
import de.uni_due.inf.ti.raven.gui.Utilities;
import de.uni_due.inf.ti.raven.io.BDDPrinter;
import de.uni_due.inf.ti.raven.io.DotPrinter;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * This class provides a {@code VisualGraph} to visualize <i>binary decision diagrams</i>.
 * @author Christoph Blume
 * @version $Revision: 1.29 $
 *
 * @see VisualGraph
 * @see RavenBDD
 */
public class BDDView extends mxGraphComponent implements BDDPrinter {

	private static final long serialVersionUID = 7012970645664007628L;

	private static final String DOT_ATTRIBUTE_SEPARATOR_TOKEN = ",";
	private static final String DOT_BOX_TOKEN = "box";
//	private static final String DOT_ELLIPSE_TOKEN = "ellipse";
	private static final String DOT_LABEL_TOKEN = "label";
	private static final String DOT_POSITION_TOKEN = "pos";
	private static final String DOT_SHAPE_TOKEN = "shape";
	private static final String DOT_STYLE_TOKEN = "style";

	private static final String DOTTED_EDGE_TYPE = "dotted";
	private static final String FILLED_EDGE_TYPE = "filled";

	private final Component parent;
	private final Logger logger;

	private VisualGraph visualGraph;

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new {@code GraphView} component.
	 * @param parent the parent component of this graph view 
	 * @param encoding the encoding used to encode the BDD
	 */
	public BDDView(Component parent, BDDEncoding encoding) {
		super(new mxGraph());
		this.parent = parent;

		logger = Logger.getLogger(getClass().getName());
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods 																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public void print(RavenBDD bdd) {
		File tempSourceFile = null;

		//Print the graph in DOT-format to a temporary file
		try {
			tempSourceFile = File.createTempFile("raven_source_", "."+RavenMain.DOT_FILE_EXTENSION, 
					new File(RavenMain.TEMP_DIR));
			DotPrinter printer = new DotPrinter(tempSourceFile.getAbsolutePath());
			printer.print(bdd);
		} catch (IOException e) {
			logger.throwing(getClass().getName(), "print", e);
			Utilities.showWarningMessage(parent, GuiMessages.IO_EXCEPTION_TITLE_EXCEPTION_KEY, 
					GuiMessages.IO_EXCEPTION_TEXT_EXCEPTION_KEY);
			return;
		}

		//Initialize the visual graph
		visualGraph = new VisualGraph();		
		visualGraph.setCellsEditable(false);
		visualGraph.setAllowDanglingEdges(false);
		visualGraph.setAllowLoops(false);
		visualGraph.setCellsDeletable(false);
		visualGraph.setCellsCloneable(false);
		visualGraph.setCellsDisconnectable(false);
		visualGraph.setDropEnabled(false);
		visualGraph.setSplitEnabled(false);
		visualGraph.setCellsBendable(false);
		visualGraph.setConnectableEdges(false);
		visualGraph.setEdgeLabelsMovable(false);
		visualGraph.setVertexLabelsMovable(false);		
		visualGraph.setCellsResizable(false);
		final Object parentChild = this.visualGraph.getDefaultParent();
		final Map<String, Object> nodeMap = new HashMap<>();

		Path dotPath = Paths.get(RavenMain.getProperty(RavenMain.PROPERTY_DOT_PATH));
		if(Files.exists(dotPath) && Files.isRegularFile(dotPath) &&
				Files.isExecutable(dotPath)) {
			final ProcessBuilder pb = new ProcessBuilder(
					dotPath.toFile().getAbsolutePath(), "-Tdot", 
					"-K"+RavenMain.DOT_BDD_LAYOUTER, 
					tempSourceFile.getAbsolutePath());

//			pb.redirectError(Redirect.PIPE);
			pb.redirectOutput(Redirect.PIPE);
			pb.redirectError(Redirect.INHERIT);
//			pb.redirectOutput(Redirect.INHERIT);

			try {
				final Process dotProcess = pb.start();
				
				//Connect error stream of the process to the logger
				try (BufferedReader reader = new BufferedReader(
						new InputStreamReader(dotProcess.getErrorStream()))) {
					String line;
					while((line = reader.readLine())!=null) {
						logger.warning(line);
					}
				}
				
//				BufferedReader reader = new BufferedReader(
//						new InputStreamReader(dotProcess.getInputStream()));
//				try {
//					String line;
//					while((line = reader.readLine())!=null) {
//						logger.warning(line);
//					}
//				}
//				finally {
//					
//				}
				
				//Initialize ANTLR Parser and create parsing tree
				ANTLRInputStream input = new ANTLRInputStream(dotProcess.getInputStream());
				DOTLexer lexer = new DOTLexer(input);
				CommonTokenStream tokens = new CommonTokenStream(lexer);
				DOTParser parser = new DOTParser(tokens);
				ParseTree tree = parser.graph();
				
				//Evaluate parsing tree
				DOTVisitor<Object> eval = new DOTBaseVisitor<Object>() {
					
					@Override
					public Object visitNode_stmt(Node_stmtContext ctx) {
						try {
							Map<String, String> map = visitAttr_list(ctx.attr_list());
							String[] position = map.get(DOT_POSITION_TOKEN).split(
									DOT_ATTRIBUTE_SEPARATOR_TOKEN);
							int x = (int) Float.parseFloat(position[0]);
							int y = (int) Float.parseFloat(position[1]);
							String label = map.get(DOT_LABEL_TOKEN);

							String style = null;
							if(map.containsKey(DOT_SHAPE_TOKEN) && 
									map.get(DOT_SHAPE_TOKEN).equals(DOT_BOX_TOKEN)) {
								if(label.equals("0")) {
									style = VisualGraph.BDD_TERMINAL_ZERO_STYLE;
								}
								else if(label.equals("1")) {
									style = VisualGraph.BDD_TERMINAL_ONE_STYLE;
								}
							}
							else {
								style = VisualGraph.BDD_VERTEX_STYLE;
							}

							Object node = visualGraph.insertVertex(parentChild, null, 
									label, x, -y, 30, 30, style);
							nodeMap.put(ctx.node_id().getText(), node);
//							if(style.equals(VisualGraph.BDD_VERTEX_STYLE)) {
//								node.setString(ENCODING_NODE_TYPE_KEY, encoding.getBitType(label/2);
//							}

							return node;
						}
						catch (Exception e) {
							e.printStackTrace();
							throw e;
						}
					}

					@Override
					public Object visitEdge_stmt(Edge_stmtContext ctx) {
						try {
							Map<String, String> map = visitAttr_list(ctx.attr_list());
							String sourceName = visitId(ctx.node_id().id());
							String targetName;
							String label = map.get(DOT_LABEL_TOKEN);
							String style = map.get(DOT_STYLE_TOKEN);

							if(style.equals(DOTTED_EDGE_TYPE)) {
								style = VisualGraph.BDD_LOW_EDGE_STYLE;
							}
							else if(style.equals(FILLED_EDGE_TYPE)) {
								style = VisualGraph.BDD_HIGH_EDGE_STYLE;						
							}

							for(Node_idContext nodeID: ctx.edgeRHS().node_id()) {
								targetName = visitId(nodeID.id());
								Object source = nodeMap.get(sourceName);
								Object target = nodeMap.get(targetName);
								visualGraph.insertEdge(parentChild, null, label, 
										source, target, style);
								sourceName = targetName;
							}

							return null;
						}
						catch (Exception e) {
							e.printStackTrace();
							throw e;
						}
					}

					@Override
					public List<String> visitEdgeRHS(DOTParser.EdgeRHSContext ctx) {
						try {
							List<String> list = new LinkedList<>();
							for(Node_idContext nodeID: ctx.node_id()) {
								list.add(visitId(nodeID.id()));
							}
							return list;
						}
						catch (Exception e) {
							e.printStackTrace();
							throw e;
						}
					}

					@Override
					public Map<String, String> visitAttr_list(DOTParser.Attr_listContext ctx) {
						try {
							Map<String, String> map = new HashMap<>();

							for(A_listContext child: ctx.a_list()) {
								map.putAll(visitA_list(child));
							}
							return map;
						}
						catch (Exception e) {
							e.printStackTrace();
							throw e;
						}
					};

					@Override
					public Map<String, String> visitA_list(A_listContext ctx) {
						try {
							Map<String, String> map = new HashMap<>();

							visitChildren(ctx);
							for(int i=0;i<ctx.getChildCount();i++) {
								String s = ctx.getChild(i).getText();
								if(s.equals("=")) {
									String key = ctx.getChild(i-1).getText().replaceAll("\"", "").trim();
									String value = ctx.getChild(i+1).getText().replaceAll("\"", "").trim();
									map.put(key, value);
									i++;
								}
							}

							return map;
						}
						catch (Exception e) {
							e.printStackTrace();
							throw e;
						}
					}

					@Override
					public String visitId(DOTParser.IdContext ctx) {
						try {
							return ctx.getText();
						}
						catch (Exception e) {
							e.printStackTrace();
							throw e;
						}
					}

				};

				//Run the process and check for errors
				int returnCode = dotProcess.waitFor();
				if(returnCode!=0) {
					Utilities.showWarningMessage(parent, 
							GuiMessages.IO_EXCEPTION_TITLE_EXCEPTION_KEY, 
							GuiMessages.IO_EXCEPTION_TEXT_EXCEPTION_KEY);
					return;
				}

				visualGraph.getModel().beginUpdate();
				eval.visit(tree);
				visualGraph.getModel().endUpdate();

				//Delete temporary layout file and "close" the visual graph
				if(!tempSourceFile.delete()) {
					tempSourceFile.deleteOnExit();
				}
			} catch (InterruptedException | IOException e) {
				logger.throwing(getClass().getName(), "print", e);
				Utilities.showWarningMessage(parent, 
						GuiMessages.DOT_ERROR_OCCURRED_TITLE_EXCEPTION_KEY, 
						GuiMessages.DOT_ERROR_OCCURRED_TEXT_EXCEPTION_KEY);
				e.printStackTrace();
				return;
			}
		}
		else {
			Utilities.showWarningMessage(parent, GuiMessages.DOT_PATH_NOT_FOUND_TITLE_EXCEPTION_KEY, 
					GuiMessages.DOT_PATH_NOT_FOUND_TEXT_EXCEPTION_KEY);
			return;
		}

		setConnectable(false);
		setPanning(true);
		setAutoExtend(true);
		setGraph(visualGraph);

		//Move graph center to center point of the panel
		mxRectangle bounds = visualGraph.getGraphBounds();
		mxPoint translationPoint = new mxPoint(0, bounds.getHeight());
		visualGraph.getView().setTranslate(translationPoint);		
		
		repaint();
	}

}