/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui.graphs;


import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.logging.Logger;

import javax.imageio.ImageIO;
import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.JFrame;

import com.mxgraph.swing.mxGraphComponent;
import com.mxgraph.util.mxPoint;
import com.mxgraph.util.mxRectangle;
import com.mxgraph.view.mxGraph;

import de.uni_due.inf.ti.raven.data.Cospan;
import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.gui.GuiMessages;
import de.uni_due.inf.ti.raven.gui.MainWindow;
import de.uni_due.inf.ti.raven.gui.RavenDialog;
import de.uni_due.inf.ti.raven.gui.RavenFileFilter;
import de.uni_due.inf.ti.raven.gui.SaveDataDialog;
import de.uni_due.inf.ti.raven.gui.Utilities;
import de.uni_due.inf.ti.raven.io.CospanPrinter;
import de.uni_due.inf.ti.raven.io.SVGWriter;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * This class provides a {@code VisualGraph} to visualize <i>cospans</i>.
 * @author Christoph Blume, Dennis Nolte
 * @version $Revision: 1.13 $
 * 
 * @see VisualGraph
 * @see Cospan
 */
public class CospanView extends mxGraphComponent implements CospanPrinter {

	private class SVGSaveDialog extends SaveDataDialog {

		private static final long serialVersionUID = 2755260582295749164L;

		/**
		 * 
		 */
		protected SVGSaveDialog() {
			super(null, messages.getString(GuiMessages.COSPAN_SAVE_DIALOG_TITLE_KEY), 
					new File(RavenMain.COSPAN_SAVE_DIR));
			setFileFilters(RavenFileFilter.GRAPH_SVG_FILE_FILTER);
		}

		@Override
		protected Action getSaveAction() {
			return new AbstractAction(messages.getString(GuiMessages.SAVE_COSPAN_ACTION_KEY)) {

				private static final long serialVersionUID = 197458776008048957L;

				@Override
				public void actionPerformed(ActionEvent e) {
					String filename = getSelectedFile().getAbsolutePath();
					filename = filename.trim();
					if(filename==null || filename.isEmpty()) {
						showErrorMessage(GuiMessages.EMPTY_FILENAME_TITLE_EXCEPTION_KEY,
								GuiMessages.EMPTY_FILENAME_TEXT_EXCEPTION_KEY);
						return;
					}
					
					//Try to save the displayed graph to the specified file
					filename = checkFileExtension(filename, RavenMain.SVG_FILE_EXTENSION);
					if (isFileExisting(filename)){
						int value = Utilities.showConfirmDialog(SVGSaveDialog.this, 
								GuiMessages.FILE_ALREADY_EXISTING_TITLE_KEY,
								GuiMessages.FILE_ALREADY_EXISTING_TEXT_KEY, filename);
						if(value==RavenDialog.NO_OPTION) {
							return;
						}
					}
					
					try {
						new SVGWriter().write(visualGraph, filename);
						showInformationMessage(GuiMessages.COSPAN_SAVED_TITLE_KEY,
								GuiMessages.COSPAN_SAVED_TEXT_KEY, filename);
					} catch (FileNotFoundException ex) {
						logger.throwing(getClass().getName(), "getSaveAction", ex);
						showErrorMessage(GuiMessages.FILE_NOT_FOUND_TITLE_EXCEPTION_KEY,
								GuiMessages.FILE_NOT_FOUND_TEXT_EXCEPTION_KEY, filename);
					} catch (IOException ex) {
						logger.throwing(getClass().getName(), "getSaveAction", ex);
						showErrorMessage(GuiMessages.FILE_NOT_FOUND_TITLE_EXCEPTION_KEY,
								GuiMessages.FILE_NOT_FOUND_TEXT_EXCEPTION_KEY, filename);
					}
					finally {
						SVGSaveDialog.this.dispose();
					}
				}
			};
		}
		
	}
	
	private static final long serialVersionUID = -8057196269526181353L;

	private static final int IMAGE_X = 10;
	private static final int IMAGE_Y = 10;
	private static final int IMAGE_GAP = 10;
	
	private final BufferedImage OPAQUE_SAVE_IMAGE;
	private final BufferedImage TRANSPARENT_SAVE_IMAGE;
	private final Rectangle SAVE_IMAGE_RECTANGLE;
	
	private final BufferedImage OPAQUE_MAXIMIZE_VIEW_IMAGE;
	private final BufferedImage TRANSPARENT_MAXIMIZE_VIEW_IMAGE;
	private final Rectangle MAXIMIZE_VIEW_IMAGE_RECTANGLE;
	
	private VisualGraph visualGraph;
	private Cospan cospan;
	
	private boolean isSaveButtonVisible;
	private boolean isMaximizeViewButtonVisible;
	private BufferedImage visibleSaveImage;
	private BufferedImage visibleMaximizeViewImage;

	private SVGSaveDialog dialog;
	
	private final Logger logger;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods 																		*/
	/*------------------------------------------------------------------------------------------*/
	
	private static BufferedImage createTransparentImage(BufferedImage image) {
		BufferedImage transparentImage = new BufferedImage(image.getWidth(), 
				image.getHeight(), image.getType());
		Graphics2D g2d = (Graphics2D) transparentImage.getGraphics();
		g2d.setBackground(Color.WHITE);
		AlphaComposite alpha = AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.3f);
		g2d.setComposite(alpha);
		g2d.drawImage(image, 0, 0, null);		
		return transparentImage;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * 
	 */
	public CospanView() {
		super(new mxGraph());
		
		logger = Logger.getLogger(getClass().getName());
		
		BufferedImage saveImage = null;
		try {
			//Try to load the save icon file
			saveImage = ImageIO.read(new File(RavenMain.SAVE_ICON_FILENAME));
		} catch (IOException e) {
			logger.throwing(getClass().getName(), "GraphView", e);
			saveImage = new BufferedImage(16, 16, BufferedImage.TYPE_INT_RGB);
		}
		finally {
			//Create an opaque and a transparent image from the previously loaded save icon
			OPAQUE_SAVE_IMAGE = saveImage;
			TRANSPARENT_SAVE_IMAGE = createTransparentImage(OPAQUE_SAVE_IMAGE);
			SAVE_IMAGE_RECTANGLE = new Rectangle(IMAGE_X, IMAGE_Y, OPAQUE_SAVE_IMAGE.getWidth(), 
					OPAQUE_SAVE_IMAGE.getHeight());
			visibleSaveImage = TRANSPARENT_SAVE_IMAGE;
			
			//Install mouse listener (for the hover effect)
			getGraphControl().addMouseMotionListener(new MouseAdapter() {
				
				@Override
				public void mouseMoved(MouseEvent e) {
					if(SAVE_IMAGE_RECTANGLE.contains(e.getPoint())) {
						visibleSaveImage = OPAQUE_SAVE_IMAGE;
					}
					else {
						visibleSaveImage = TRANSPARENT_SAVE_IMAGE;
					}
					repaint();
				}
				
			});
			
			//Install mouse listener (for the mouse click)
			getGraphControl().addMouseListener(new MouseAdapter() {

				@Override
				public void mouseClicked(MouseEvent e) {
					if(SAVE_IMAGE_RECTANGLE.contains(e.getPoint())) {
						dialog = new SVGSaveDialog();
						dialog.setVisible(true);
					}
				}
			});
		}
		
		BufferedImage maximizeViewImage = null;
		try {
			//Try to load the save icon file
			maximizeViewImage = ImageIO.read(new File(RavenMain.MAXIMIZE_VIEW_ICON_FILENAME));
		} catch (IOException e) {
			logger.throwing(getClass().getName(), "GraphView", e);
			maximizeViewImage = new BufferedImage(16 ,16, BufferedImage.TYPE_INT_RGB);
		}
		finally {
			//Create an opaque and a transparent image from the previously loaded save icon
			OPAQUE_MAXIMIZE_VIEW_IMAGE = maximizeViewImage;
			TRANSPARENT_MAXIMIZE_VIEW_IMAGE = createTransparentImage(OPAQUE_MAXIMIZE_VIEW_IMAGE);
			MAXIMIZE_VIEW_IMAGE_RECTANGLE = new Rectangle(
					SAVE_IMAGE_RECTANGLE.x + SAVE_IMAGE_RECTANGLE.width + IMAGE_GAP, 
					SAVE_IMAGE_RECTANGLE.y, OPAQUE_MAXIMIZE_VIEW_IMAGE.getWidth(), 
					OPAQUE_MAXIMIZE_VIEW_IMAGE.getHeight());
			visibleMaximizeViewImage = TRANSPARENT_MAXIMIZE_VIEW_IMAGE;
			
			//Install mouse listener (for the hover effect)
			getGraphControl().addMouseMotionListener(new MouseAdapter() {
				
				@Override
				public void mouseMoved(MouseEvent e) {
					if(MAXIMIZE_VIEW_IMAGE_RECTANGLE.contains(e.getPoint())) {
						visibleMaximizeViewImage = OPAQUE_MAXIMIZE_VIEW_IMAGE;
					}
					else {
						visibleMaximizeViewImage = TRANSPARENT_MAXIMIZE_VIEW_IMAGE;
					}
					repaint();
				}
				
			});
			
			//Install mouse listener (for the mouse click)
			getGraphControl().addMouseListener(new MouseAdapter() {

				private JFrame window;

				@Override
				public void mouseClicked(MouseEvent e) {
					if(MAXIMIZE_VIEW_IMAGE_RECTANGLE.contains(e.getPoint())) {
						if(window==null) {
							CospanView view = new CospanView();
							view.print(cospan);
							view.setOpaque(false);
							view.getViewport().setOpaque(false);					
							window = new JFrame(Repository.getRepository().getName(cospan));
							window.getContentPane().add(view);
							window.getContentPane().setBackground(MainWindow.BACKGROUND_COLOR);
							window.setSize(500, 500);
							window.setLocationRelativeTo(null);
							
							window.addWindowListener(new WindowAdapter() {
								
								@Override
								public void windowClosed(WindowEvent arg0) {
									window = null;							
								}
								@Override
								public void windowClosing(WindowEvent arg0) {
									window.dispose();							
								}

							});
							window.setVisible(true);
						}
					}
				}
			});
		}
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods 																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Sets the save button, which is used to save the displayed graph, visible/invisible.
	 * @param isVisible {@code true} to enable the save button, {@code false} to disable it
	 */
	public void setSaveButtonVisible(boolean isVisible) {
		isSaveButtonVisible = isVisible;
	}
	
	/**
	 * Sets the maximize view button, which is used to maximize the displayed graph, visible/invisible.
	 * @param isVisible {@code true} to enable the maximize view button, {@code false} to disable it
	 */
	public void setMaximizeViewButtonVisible(boolean isVisible) {
		isMaximizeViewButtonVisible = isVisible;
	}
	
	@Override
	public void print(final Cospan cospan) {
		this.cospan = cospan;
		try {
			visualGraph = VisualGraph.getVisualGraph(cospan);
		} catch (FileNotFoundException e) {
			logger.throwing(getClass().getName(), "print", e);
			Utilities.showWarningMessage(null, GuiMessages.DOT_PATH_NOT_FOUND_TITLE_EXCEPTION_KEY, 
					GuiMessages.DOT_PATH_NOT_FOUND_TEXT_EXCEPTION_KEY);
			return;
		} catch (IOException | InterruptedException e) {
			logger.throwing(getClass().getName(), "print", e);
			Utilities.showWarningMessage(null, 
					GuiMessages.DOT_ERROR_OCCURRED_TITLE_EXCEPTION_KEY, 
					GuiMessages.DOT_ERROR_OCCURRED_TEXT_EXCEPTION_KEY);
			return;
		}

		setConnectable(false);
		setPanning(true);
		setAutoExtend(true);
		setGraph(visualGraph);
		
		//Move graph center to center point of the panel
		int y = Math.max(SAVE_IMAGE_RECTANGLE.y + SAVE_IMAGE_RECTANGLE.height, 
				MAXIMIZE_VIEW_IMAGE_RECTANGLE.y + MAXIMIZE_VIEW_IMAGE_RECTANGLE.height);
		mxRectangle bounds = visualGraph.getGraphBounds();
		mxPoint translationPoint = new mxPoint(-bounds.getX(), y+IMAGE_GAP);
		visualGraph.getView().setTranslate(translationPoint);
		
		repaint();
	}
	
	@Override
	public void paint(Graphics g) {
		super.paint(g);

		if(isSaveButtonVisible) {
			g.drawImage(visibleSaveImage, SAVE_IMAGE_RECTANGLE.x, SAVE_IMAGE_RECTANGLE.y, this);
		}
		
		if(isMaximizeViewButtonVisible) {
			g.drawImage(visibleMaximizeViewImage, MAXIMIZE_VIEW_IMAGE_RECTANGLE.x, 
					MAXIMIZE_VIEW_IMAGE_RECTANGLE.y, this);
		}
	}
	
	
}