/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui.graphs;

import com.mxgraph.model.mxCell;
import com.mxgraph.model.mxGeometry;
import com.mxgraph.model.mxGraphModel;
import com.mxgraph.swing.mxGraphComponent;
import com.mxgraph.view.mxGraph;

import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.GraphDecomposition;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.Operation.Connect;
import de.uni_due.inf.ti.raven.data.Operation.Fuse;
import de.uni_due.inf.ti.raven.data.Operation.Perm;
import de.uni_due.inf.ti.raven.data.Operation.Res;
import de.uni_due.inf.ti.raven.data.Operation.Trans;
import de.uni_due.inf.ti.raven.data.Operation.Vertex;
import de.uni_due.inf.ti.raven.io.DecompositionPrinter;

/**
 * This class provides a {@code Display} to visualize <i>graph decompositions</i>.
 * @author Christoph Blume
 * @version $Revision: 1.22 $
 *
 * @see GraphDecomposition
 */
public class DecompositionView extends mxGraphComponent implements DecompositionPrinter {

	private static final int UNICODE_SUPERSCRIPT_0 = 8304;
	private static final int UNICODE_SUPERSCRIPT_1 = 185;
	private static final int UNICODE_SUPERSCRIPT_2 = 178;
	private static final int UNICODE_SUPERSCRIPT_3 = 179;
	private static final int UNICODE_SUPERSCRIPT_4 = 8308;
	private static final int UNICODE_SUPERSCRIPT_5 = 8309;
	private static final int UNICODE_SUPERSCRIPT_6 = 8310;
	private static final int UNICODE_SUPERSCRIPT_7 = 8311;
	private static final int UNICODE_SUPERSCRIPT_8 = 8312;
	private static final int UNICODE_SUPERSCRIPT_9 = 8313;

	private static final long serialVersionUID = -1458478463143158777L;
	
	private static final int X_OFFSET = 50;
	private static final int Y_OFFSET = 100;
	
	private static final int X_MARGIN = 200;
	private static final int Y_MARGIN = 100;

	private static final int INTERFACE_NODE_SIZE = 25;
	private static final int HYPERGRAPH_NODE_SIZE = 25;
	private static final int HYPERGRAPH_EDGE_SIZE = 25;
	

	private VisualGraph visualGraph;
	private int counter;
	private int lastLetter;
	private mxCell group;
	private Object[] previousNodes;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	private static String getSuperscriptNumber(int width) {
		char[] unicodes = {
					UNICODE_SUPERSCRIPT_0, 
					UNICODE_SUPERSCRIPT_1, 
					UNICODE_SUPERSCRIPT_2, 
					UNICODE_SUPERSCRIPT_3,
					UNICODE_SUPERSCRIPT_4, 
					UNICODE_SUPERSCRIPT_5, 
					UNICODE_SUPERSCRIPT_6,
					UNICODE_SUPERSCRIPT_7,
					UNICODE_SUPERSCRIPT_8,
					UNICODE_SUPERSCRIPT_9
				};
		 
		 
		  String numberString = Integer.toString(width);
		  String erg = "";
		  
		  for (int i=0;i<numberString.length();i++)
		  {
		   erg += unicodes[Integer.parseInt(Character.toString(numberString.charAt(i)))];
		  }
		  
		  return erg;
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new {@code DecompositionView}.
	 */
	public DecompositionView() {
		super(new mxGraph());
		counter = 0;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods 																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public void print(final CospanDecomposition decomposition) {
	
		counter = 0;
		visualGraph = new VisualGraph();
		visualGraph.setModel(new mxGraphModel());
		visualGraph.setCellsResizable(false);
		visualGraph.setCellsDisconnectable(false);
		visualGraph.setCellsEditable(false);
		
		visualGraph.getModel().beginUpdate();
		try {
			int currentWidth = decomposition.getInnerInterfaceSize();
			
			lastLetter = decomposition.getLength();
			
			for(Operation letter: decomposition) {
				if(counter == 0 && currentWidth==0)
				{
					addEmptyInnerInterface(X_OFFSET,Y_OFFSET);
					previousNodes = new Object[currentWidth];	
				}
				else if(counter == 0 && currentWidth!=0) {
					previousNodes = new Object[currentWidth];
					for(int i=0;i<currentWidth;i++) {
						previousNodes[i] = addInterfaceNode(i, String.valueOf(i));
						((mxCell)previousNodes[i]).setStyle(VisualGraph.INTERFACE_NO_BACKGROUND_STYLE);
					}
					group = (mxCell) visualGraph.groupCells(null, 5, previousNodes);
					group.setStyle(VisualGraph.INNER_INTERFACE_BACKGROUND_STYLE);
				}
				
				counter++;
				if(letter instanceof Connect) {
					Connect connect = (Connect) letter;
					createLetterLabel(letter.toString() + getSuperscriptNumber(currentWidth));
					addConnectCospan(currentWidth, connect.getLabel(), connect.getArity());
				}
				else if(letter instanceof Fuse) {
					createLetterLabel(letter.toString() + getSuperscriptNumber(currentWidth));
					currentWidth--;
					addFuseCospan(currentWidth);
				}
				else if(letter instanceof Perm) {
					createLetterLabel(letter.toString() + getSuperscriptNumber(currentWidth));
					addPermCospan(currentWidth);
				}
				else if(letter instanceof Res) {
					createLetterLabel(letter.toString() + getSuperscriptNumber(currentWidth));
					currentWidth--;
					addResCospan(currentWidth);
				}
				else if(letter instanceof Trans) {
					createLetterLabel(letter.toString() + getSuperscriptNumber(currentWidth));
					addTransCospan(currentWidth);
				}
				else if(letter instanceof Vertex) {
					createLetterLabel(letter.toString() + getSuperscriptNumber(currentWidth));
					currentWidth++;
					addVertexCospan(currentWidth);
				}
			}
			
		}
		finally {
			
			visualGraph.getModel().endUpdate();
		}
		
		visualGraph.setBorder(X_OFFSET);
		setEnabled(false);
		setConnectable(false);
		setGraph(visualGraph);
		repaint();
	}


	@Override
	public void print(GraphDecomposition decomposition) {
		// TODO Auto-generated method stub
		
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	private void addConnectCospan(int width, String label, int arity) {
		Object[] childNodes = new Object[width];
		Object parent = visualGraph.getDefaultParent();
		Object[] graphNodes = new Object[width];
		
		//Add vertex for hyperedge label
		int x = X_OFFSET + counter*X_MARGIN-X_MARGIN/2+2*HYPERGRAPH_NODE_SIZE;
		mxCell edge = (mxCell) visualGraph.insertVertex(parent, null, label, x, 0, 
				HYPERGRAPH_EDGE_SIZE, HYPERGRAPH_EDGE_SIZE, VisualGraph.HYPERGRAPH_EDGE_STYLE);
		
		//
		int minY = Integer.MAX_VALUE;
		int maxY = Integer.MIN_VALUE;
		
		for(int i=0;i<width;i++) {
			graphNodes[i] = addGraphNode(i);
			childNodes[i] = addInterfaceNode(i,String.valueOf(i));
			
			if(i>=width-arity) {
				minY = Math.min(minY, (int) ((mxCell) graphNodes[i]).getGeometry().getY());
				maxY = Math.max(maxY, (int) ((mxCell) graphNodes[i]).getGeometry().getY());

				visualGraph.insertEdge(parent, null, String.valueOf(i-width+arity), edge, graphNodes[i], 
						VisualGraph.HYPERGRAPH_TENTACLE_STYLE);
			}
			
			visualGraph.insertEdge(parent, null, null, previousNodes[i], graphNodes[i], VisualGraph.COSPAN_EDGE_STYLE_RED);
		}
		
		//Update edge (vertex) geometry
		mxGeometry geometry = edge.getGeometry();
		geometry.setY(minY+(maxY-minY)/2);
		edge.setGeometry(geometry);
		
		checkLastOperation(width, childNodes);

		for(int i=0;i<width;i++) {
			visualGraph.insertEdge(parent, null, null, childNodes[i], graphNodes[i], VisualGraph.COSPAN_EDGE_STYLE_RED);				
		}
		
		previousNodes = new Object[width];
		for(int i=0;i<width;i++) {
			previousNodes[i] = childNodes[i];
		}
	}

	private void addFuseCospan(int width) {
		Object parent = visualGraph.getDefaultParent();
		Object[] childNodes = new Object[width];
		Object[] graphNodes = new Object[width+1];

		for(int i=0;i<width;i++) {
			graphNodes[i] = addGraphNode(i);
			visualGraph.insertEdge(parent, null, null, previousNodes[i], graphNodes[i], VisualGraph.COSPAN_EDGE_STYLE_RED);
		}
		visualGraph.insertEdge(parent, null, null, previousNodes[width], graphNodes[width-1], VisualGraph.COSPAN_EDGE_STYLE_RED);
		
		for(int i=0;i<width-1;i++) {
			childNodes[i] = addInterfaceNode(i,String.valueOf(i));
		}
		String label = String.valueOf(width-1)+","+String.valueOf(width);
		childNodes[width-1] = addInterfaceNode(width-1,label);
		
		checkLastOperation(width, childNodes);
		
		for(int i=0;i<width;i++) {
			visualGraph.insertEdge(parent, null, null, childNodes[i], graphNodes[i], VisualGraph.COSPAN_EDGE_STYLE_RED);
		}
		
		previousNodes = new Object[width];
		for(int i=0;i<width;i++) {
			previousNodes[i] = childNodes[i];
		}
	}
	
	private void addPermCospan(int width) {
		Object parent = visualGraph.getDefaultParent();
		Object[] childNodes = new Object[width];
		Object[] graphNodes = new Object[width];
		
		for(int i=0;i<width;i++) {
			graphNodes[i] = addGraphNode(i);
			if((i-1) % width <0) {
				childNodes[i] = addInterfaceNode(i,String.valueOf((i-1+width)%width));
			}
			else {
				childNodes[i] = addInterfaceNode(i,String.valueOf((i-1)%width));				
			}
			visualGraph.insertEdge(parent, null, null, previousNodes[i], graphNodes[i], VisualGraph.COSPAN_EDGE_STYLE_RED);
		}
		checkLastOperation(width, childNodes);
		for(int i=0;i<width;i++) {
			if((i-1) % width <0) {
				visualGraph.insertEdge(parent, null, null, childNodes[i], graphNodes[(i-1+width)%width], VisualGraph.COSPAN_EDGE_STYLE_RED);			
			}
			else {
				visualGraph.insertEdge(parent, null, null, childNodes[i], graphNodes[(i-1)%width], VisualGraph.COSPAN_EDGE_STYLE_RED);			
			}
		}
		
		previousNodes = new Object[width];
		for(int i=0;i<width;i++) {
			previousNodes[i] = childNodes[i];
		}
		
	}
	
	private void addResCospan(int width) {
		Object parent = visualGraph.getDefaultParent();
		Object[] childNodes = new Object[width];
		Object[] graphNodes = new Object[width+1];

		for(int i=0;i<width+1;i++) {
			graphNodes[i] = addGraphNode(i);
			visualGraph.insertEdge(parent, null, null, previousNodes[i], graphNodes[i], VisualGraph.COSPAN_EDGE_STYLE_RED);
		}
		for(int i=0;i<width;i++) {
			childNodes[i] = addInterfaceNode(i,String.valueOf(i));
		}
		checkLastOperation(width, childNodes);
		
		if(width==0 && counter == lastLetter) {
			addEmptyOuterInterface(X_OFFSET + counter*X_MARGIN, Y_OFFSET);
			previousNodes = new Object[0];
		}
		else if(width==0 && counter != lastLetter) {
			addEmptyInterface(X_OFFSET + counter*X_MARGIN, Y_OFFSET);
			previousNodes = new Object[0];
		}
		else {
			for(int i=0;i<width;i++) {
				visualGraph.insertEdge(parent, null, null, childNodes[i], graphNodes[i], VisualGraph.COSPAN_EDGE_STYLE_RED);				
			}
			previousNodes = new Object[width];
			for(int i=0;i<width;i++) {
				previousNodes[i] = childNodes[i];
			}
		}
	}
	
	private void addTransCospan(int width) {
		Object parent = visualGraph.getDefaultParent();
		Object[] childNodes = new Object[width];
		Object[] graphNodes = new Object[width];
		
		for(int i=0;i<2;i++) {
			graphNodes[i] = addGraphNode(i);
			childNodes[i] = addInterfaceNode(i,String.valueOf((i+1)%2));
			visualGraph.insertEdge(parent, null, null, previousNodes[i], graphNodes[i], VisualGraph.COSPAN_EDGE_STYLE_RED);
			
		}
		for(int i=2;i<width;i++) {
			graphNodes[i] = addGraphNode(i);
			childNodes[i] = addInterfaceNode(i,String.valueOf(i));
			visualGraph.insertEdge(parent, null, null, previousNodes[i], graphNodes[i], VisualGraph.COSPAN_EDGE_STYLE_RED);
		}
		checkLastOperation(width, childNodes);
		for(int i=0;i<2;i++) {
			visualGraph.insertEdge(parent, null, null, childNodes[i], graphNodes[(i+1)%2], VisualGraph.COSPAN_EDGE_STYLE_RED);
		}
		for(int i=2;i<width;i++) {
			visualGraph.insertEdge(parent, null, null, childNodes[i], graphNodes[i], VisualGraph.COSPAN_EDGE_STYLE_RED);			
		}
		
		previousNodes = new Object[width];
		for(int i=0;i<width;i++) {
			previousNodes[i] = childNodes[i];
		}
			
	}

	private void addVertexCospan(int width) {
		Object parent = visualGraph.getDefaultParent();
		Object[] childNodes = new Object[width];
		Object[] graphNodes = new Object[width];	
		
		for(int i=0;i<width;i++) {
			graphNodes[i] = addGraphNode(i);
			childNodes[i] = addInterfaceNode(i,String.valueOf(i));
		}
		for(int i=0;i<width-1;i++) {
			visualGraph.insertEdge(parent, null, null, previousNodes[i], graphNodes[i], VisualGraph.COSPAN_EDGE_STYLE_RED);
		}
		
		checkLastOperation(width, childNodes);
		for(int i=0;i<width;i++) {
			visualGraph.insertEdge(parent, null, null, childNodes[i], graphNodes[i], VisualGraph.COSPAN_EDGE_STYLE_RED);			
		}
		
		previousNodes = new Object[width];
		for(int i=0;i<width;i++) {
			previousNodes[i] = childNodes[i];
		}
	}
	
	private mxCell addInterfaceNode(int number,String label) {
		Object parent = visualGraph.getDefaultParent();
		int x = X_OFFSET + counter*X_MARGIN;
		int y = Y_OFFSET + number*Y_MARGIN;
		
		return (mxCell) visualGraph.insertVertex(parent, null, label, x, y, 
				INTERFACE_NODE_SIZE, INTERFACE_NODE_SIZE, VisualGraph.INTERFACE_VERTEX_STYLE);
	}
	
	private mxCell addGraphNode(int number) {
		Object parent = visualGraph.getDefaultParent();
		int x = X_OFFSET + counter*X_MARGIN-X_MARGIN/2;
		int y = Y_OFFSET + number*Y_MARGIN;
		
		return (mxCell) visualGraph.insertVertex(parent, null, "", x, y, 
				HYPERGRAPH_NODE_SIZE, HYPERGRAPH_NODE_SIZE, VisualGraph.HYPERGRAPH_VERTEX_STYLE);
	}
	
	private void addEmptyInterface(int x, int y) {
		Object parent = visualGraph.getDefaultParent();
		Object[] emptyInterface = new Object[1];
		emptyInterface[0] = visualGraph.insertVertex(parent, null, "\u00D8", x, y, 
				INTERFACE_NODE_SIZE, INTERFACE_NODE_SIZE, VisualGraph.NO_VERTEX_STYLE);
		visualGraph.groupCells(null, 5, emptyInterface); 
	}
	
	private void addEmptyInnerInterface(int x, int y) {
		Object parent = visualGraph.getDefaultParent();
		Object[] emptyInterface = new Object[1];
		emptyInterface[0] = visualGraph.insertVertex(parent, null, "\u00D8", x, y, 
				INTERFACE_NODE_SIZE, INTERFACE_NODE_SIZE, VisualGraph.NO_VERTEX_STYLE);
		group = (mxCell) visualGraph.groupCells(null, 5, emptyInterface); 
		group.setStyle(VisualGraph.INNER_INTERFACE_BACKGROUND_STYLE);
	}
	
	private void addEmptyOuterInterface(int x, int y) {
		Object parent = visualGraph.getDefaultParent();
		Object[] emptyInterface = new Object[1];
		emptyInterface[0] = visualGraph.insertVertex(parent, null, "\u00D8", x, y, 
				INTERFACE_NODE_SIZE, INTERFACE_NODE_SIZE, VisualGraph.NO_VERTEX_STYLE);
		group = (mxCell) visualGraph.groupCells(null, 5, emptyInterface); 
		group.setStyle(VisualGraph.OUTER_INTERFACE_BACKGROUND_STYLE);
	}
	
	private mxCell createLetterLabel(String letterLabel) {
		Object parent = visualGraph.getDefaultParent();
		int x = X_OFFSET + counter*X_MARGIN-X_MARGIN/2;
		int y = Y_OFFSET - Y_MARGIN/2;
		
		return (mxCell) visualGraph.insertVertex(parent, null, letterLabel, x, y, 
				HYPERGRAPH_NODE_SIZE, HYPERGRAPH_NODE_SIZE, VisualGraph.LABEL_STYLE);
	}

	/**
	 * @param width the current width
	 * @param childNodes the outer interface nodes
	 */
	private void checkLastOperation(int width, Object[] childNodes) {
		if(counter == lastLetter) {
			for(int i=0;i<width;i++) {
				((mxCell)childNodes[i]).setStyle(VisualGraph.INTERFACE_NO_BACKGROUND_STYLE);
			}
		}
		
		group = (mxCell) visualGraph.groupCells(null, 5, childNodes);
		if(counter == lastLetter) {
			group.setStyle(VisualGraph.OUTER_INTERFACE_BACKGROUND_STYLE);
		}
	}
	
	

}