/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui.graphs;

import com.mxgraph.model.mxCell;
import com.mxgraph.model.mxGraphModel;
import com.mxgraph.swing.mxGraphComponent;
import com.mxgraph.view.mxGraph;

import de.uni_due.inf.ti.raven.encodings.BDDEncoding;
import de.uni_due.inf.ti.raven.encodings.BDDEncoding.StateType;
import de.uni_due.inf.ti.raven.encodings.BoundEncoding;
import de.uni_due.inf.ti.raven.encodings.ColorabilityEncoding;
import de.uni_due.inf.ti.raven.encodings.CountingEncoding;
import de.uni_due.inf.ti.raven.encodings.DominatingSetEncoding;
import de.uni_due.inf.ti.raven.encodings.LinkEncoding;
import de.uni_due.inf.ti.raven.encodings.NoIsolatedNodesEncoding;
import de.uni_due.inf.ti.raven.encodings.PathEncoding;
import de.uni_due.inf.ti.raven.encodings.SubgraphEncoding;
import de.uni_due.inf.ti.raven.encodings.VertexCoverEncoding;
import de.uni_due.inf.ti.raven.gui.GuiMessages;
import de.uni_due.inf.ti.raven.io.EncodingPrinter;
import de.uni_due.inf.ti.raven.messages.Messages;


/**
 * This class provides a {@code VisualGraph} to visualize <i>encodings</i>.
 * @author Christoph Blume, Dennis Nolte
 * @version $Revision: 1.23 $
 *
 * @see VisualGraph
 * @see BDDEncoding
 */
public class EncodingView extends mxGraphComponent implements EncodingPrinter {

	/**
	 * This class is the abstract base class for all {@code EncodingVisualizer}
	 * used to visualize the different types of {@code BDDEncodings}
	 * @param <E> the type of {@code BDDEncoding} which can be visualized with 
	 * the certain visualizer component
	 */
	protected abstract static class EncodingVisualizer<E extends BDDEncoding> {

		/** */
		protected final Messages messages;
		
		/** */
		protected int leftShift;
		
		/** The offset in x-direction */
		protected static final int X_OFFSET = 180;
		
		/** The offset in y-direction */
		protected static final int Y_OFFSET = 100;
		
		/** The margin in x-direction */
		protected static final int X_MARGIN = 100;
		
		/** The margin in y-direction */
		protected static final int Y_MARGIN = 200;
		
		/** */
		protected static final int GAP_CONSTANT = 5;
		
		/** The size of the interface nodes */
		protected static final int INTERFACE_VERTEX_SIZE = 25;
		
		/** The size of the state nodes */
		protected static final int STATE_VERTEX_SIZE = 25;
		
		/** */
		protected final VisualGraph visualGraph;
		
		/**
		 * 
		 */
		public EncodingVisualizer() {
			visualGraph = new VisualGraph();
			visualGraph.setModel(new mxGraphModel());
			visualGraph.setCellsResizable(false);
			visualGraph.setCellsDisconnectable(false);
			visualGraph.setCellsEditable(false);
			
			messages = GuiMessages.getMessages();
		}
		
		/*------------------------------------------------------------------------------------------*/
		/*		Abstract Methods 																	*/
		/*------------------------------------------------------------------------------------------*/
		
		/**
		 * Visualizes the state bits of the respective encoding based on the 
		 * state type. 
		 * @param type the state type indicating which bits are to be visualized
		 * @param encoding the encoding whose bits are to be visualized
		 */
		protected abstract void visualizeStateBits(StateType type, E encoding);
		
		/*------------------------------------------------------------------------------------------*/
		/*		Protected Methods 																	*/
		/*------------------------------------------------------------------------------------------*/
		
		/**
		 * Visualizes the given {@code BDDEncoding}.
		 * @param encoding the encoding to be visualized
		 * @return a visual graph which visualizes the given encoding
		 */
		protected VisualGraph visualize(E encoding) {
			visualizeInterfaceBits(StateType.current, encoding);
			visualizeInterfaceBits(StateType.successor, encoding);
			visualizeStateBits(StateType.current, encoding);
			visualizeStateBits(StateType.successor, encoding);
			
			return visualGraph;
		}
		
		/**
		 * Visualizes the interface bits of the respective encoding based on the 
		 * state type. 
		 * @param type the state type indicating which bits are to be visualized
		 * @param encoding the encoding whose bits are to be visualized
		 */
		protected void visualizeInterfaceBits(StateType type, E encoding) {
			if(type==StateType.current) {
				addStateTypeLabel(type, messages.getString(GuiMessages.CURRENT_INTERFACE_LABEL_KEY));	
			}
			else {
				addStateTypeLabel(type, messages.getString(GuiMessages.SUCCESSOR_INTERFACE_LABEL_KEY));
			}
			
			visualGraph.getModel().beginUpdate();
			try {
				int offset = getOffset(encoding, type);
				int interfaceLength = encoding.getInterfaceLength();
				leftShift = interfaceLength%2 == 0 ? 1 : 0;
				int groupPos = interfaceLength/2;
				
				Object[] interfaceNodes = new Object[interfaceLength];
				for(int i=0; i<interfaceLength;i++) {
					interfaceNodes[i] = addInterfaceVertex(type, i, String.valueOf(offset+i*2));
					addBitLabel(type, i, String.valueOf(i));
					if(i == groupPos) {
						addGroupLabel(type, i, GuiMessages.getMessages().getString(
								GuiMessages.INTERFACE_BITS_LABEL_KEY));
					}
				}
				mxCell group = (mxCell) visualGraph.groupCells(null, 5, interfaceNodes);
				group.setStyle(VisualGraph.INTERFACE_WHITE_BACKGROUND_STYLE);
			}
			finally{
				visualGraph.getModel().endUpdate();
			}
		}
		
		/**
		 * Returns the offset of the first bit of the encoding relative to the 
		 * BDD universe.
		 * @param encoding the encoding for which the offset is to be computed
		 * @param type the state type indicating which bits are to be used
		 * @return the offset
		 */
		protected int getOffset(E encoding, StateType type) {
			int offset = encoding.getBit(StateType.current, 0).var();
			
			if(type==StateType.successor) {
				offset += 1;
			}
			return offset;
		}
		
		/**
		 * Adds a group label with the text {@code label} at the specified 
		 * x-position {@code xPos}. The y-position is determined by the state 
		 * type {@code type}.
		 * @param type the state type used to indicate the y-position
		 * @param xPos the x-position of the group label
		 * @param label the text of the group label
		 */
		protected void addGroupLabel(StateType type, int xPos, String label) {
			int row = type==StateType.current ? 0 : 1;
			Object parent = visualGraph.getDefaultParent();
			int x = X_OFFSET + xPos*X_MARGIN - leftShift * X_MARGIN/2;
			int y = Y_OFFSET + row*Y_MARGIN-Y_MARGIN/4 - GAP_CONSTANT;
			
			visualGraph.insertVertex(parent, null, label, x, y, 
					INTERFACE_VERTEX_SIZE, INTERFACE_VERTEX_SIZE, VisualGraph.LABEL_STYLE);
		}
		
		/**
		 * Adds a label to each interface and state vertex indicating the position
		 * of the certain node in the encoding.
		 * @param type the state type used to indicate the position
		 * @param index the index of the (visual) vertex
		 * @param label the text of the label
		 */
		protected void addBitLabel(StateType type, int index, String label) {
			int row = type==StateType.current ? 0 : 1;
			Object parent = visualGraph.getDefaultParent();
			int x = X_OFFSET + index*X_MARGIN;
			int y = Y_OFFSET + row*Y_MARGIN-Y_MARGIN/8 - GAP_CONSTANT;
			
			visualGraph.insertVertex(parent, null, 
					messages.getString(GuiMessages.BIT_LABEL_KEY, label), x, y, 
					INTERFACE_VERTEX_SIZE, INTERFACE_VERTEX_SIZE, VisualGraph.LABEL_STYLE);
		}
		
		/**
		 * 
		 * Adds a state label with the text {@code label} in front of the 
		 * encoding. The y-position is determined by the state type {@code type}.
		 * @param type the state type used to indicate the y-position
		 * @param label the text of the state label
		 */
		protected void addStateTypeLabel(StateType type, String label) {
			int row = type==StateType.current ? 0 : 1;
			Object parent = visualGraph.getDefaultParent();
			int x = X_OFFSET - X_MARGIN;
			int y = Y_OFFSET + row*Y_MARGIN;
			
			visualGraph.insertVertex(parent, null, label, x, y, 
					INTERFACE_VERTEX_SIZE, INTERFACE_VERTEX_SIZE, VisualGraph.LABEL_STYLE);
		}
		
		/**
		 * Adds a (visual) interface vertex with the text {@code label}. The 
		 * position is determined by {@code type} (for the y-position) and 
		 * {@code index} for the x-position.
		 * @param type the state type used to indicate the y-position
		 * @param index the index of the vertex in the encoding used to indicate 
		 * the x-position
		 * @param label the text of the group label
		 * @return the new (visual) interface vertex
		 */
		protected mxCell addInterfaceVertex(StateType type, int index, String label) {
			int row = type==StateType.current ? 0 : 1;
			Object parent = visualGraph.getDefaultParent();
			int x = X_OFFSET + index*X_MARGIN;
			int y = Y_OFFSET + row*Y_MARGIN;
			
			return (mxCell) visualGraph.insertVertex(parent, null, label, x, y, 
					INTERFACE_VERTEX_SIZE, INTERFACE_VERTEX_SIZE, VisualGraph.INTERFACE_VERTEX_STYLE);
		}
		
		/**
		 * Adds a (visual) state vertex with the text {@code label}. The 
		 * position is determined by {@code type} (for the y-position) and 
		 * {@code index} for the x-position.
		 * @param type the state type used to indicate the y-position
		 * @param index the index of the vertex in the encoding used to indicate 
		 * the x-position
		 * @param label the text of the group label
		 * @return the new (visual) state vertex
		 */
		protected mxCell addStateVertex(StateType type, int index, String label) {
			int row = type==StateType.current ? 0 : 1;
			Object parent = visualGraph.getDefaultParent();
			int x = X_OFFSET + index*X_MARGIN;
			int y = Y_OFFSET + row*Y_MARGIN;
			
			return (mxCell) visualGraph.insertVertex(parent, null, label, x, y, 
					STATE_VERTEX_SIZE, STATE_VERTEX_SIZE, VisualGraph.HYPERGRAPH_VERTEX_STYLE);
		}
	}
	
	private static final long serialVersionUID = -8191175059947002723L;

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new {@code EncodingView}.
	 */
	public EncodingView() {
		super(new mxGraph());
		setOpaque(false);
		getViewport().setOpaque(false);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods 																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public void print(BDDEncoding encoding) {	
		VisualGraph visualGraph;
		
		if(encoding instanceof BoundEncoding) {
			BoundEncodingVisualizer visualizer = new BoundEncodingVisualizer();
			visualGraph = visualizer.visualize((BoundEncoding) encoding);
		}
		else if(encoding instanceof ColorabilityEncoding) {
			ColorabilityEncodingVisualizer visualizer = new ColorabilityEncodingVisualizer();
			visualGraph = visualizer.visualize((ColorabilityEncoding) encoding);
		}
		else if(encoding instanceof CountingEncoding) {
			CountingEncodingVisualizer visualizer = new CountingEncodingVisualizer();
			visualGraph = visualizer.visualize((CountingEncoding) encoding);
		}
		else if(encoding instanceof DominatingSetEncoding) {
			DominatingSetEncodingVisualizer visualizer = new DominatingSetEncodingVisualizer();
			visualGraph = visualizer.visualize((DominatingSetEncoding) encoding);
		}
		else if(encoding instanceof LinkEncoding) {
			LinkEncodingVisualizer visualizer = new LinkEncodingVisualizer();
			visualGraph = visualizer.visualize((LinkEncoding) encoding);
		}
		else if(encoding instanceof NoIsolatedNodesEncoding) {
			NoIsolatedNodesEncodingVisualizer visualizer = new NoIsolatedNodesEncodingVisualizer();
			visualGraph = visualizer.visualize((NoIsolatedNodesEncoding) encoding);
		}
		else if(encoding instanceof PathEncoding) {
			PathEncodingVisualizer visualizer = new PathEncodingVisualizer();
			visualGraph = visualizer.visualize((PathEncoding) encoding);
		}
		else if(encoding instanceof SubgraphEncoding) {
			SubgraphEncodingVisualizer visualizer = new SubgraphEncodingVisualizer();
			visualGraph = visualizer.visualize((SubgraphEncoding) encoding);
		}
		else if(encoding instanceof VertexCoverEncoding) {
			VertexCoverEncodingVisualizer visualizer = new VertexCoverEncodingVisualizer();
			visualGraph = visualizer.visualize((VertexCoverEncoding) encoding);
		}
		else {
			DefaultEncodingVisualizer visualizer = new DefaultEncodingVisualizer();
			visualGraph = visualizer.visualize(encoding);
		}
			
		setEnabled(false);
		setConnectable(false);
		setGraph(visualGraph);
	}

}