/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui.graphs;

import com.mxgraph.model.mxCell;

import de.uni_due.inf.ti.raven.encodings.BDDEncoding;
import de.uni_due.inf.ti.raven.encodings.BDDEncoding.StateType;
import de.uni_due.inf.ti.raven.encodings.PathEncoding;
import de.uni_due.inf.ti.raven.gui.GuiMessages;
import de.uni_due.inf.ti.raven.gui.graphs.EncodingView.EncodingVisualizer;

/**
 * This class provides a {@code Display} to visualize <i>Path encodings</i>.
 * @author Christoph Blume, Dennis Nolte
 * @version $Revision: 1.5 $
 *
 * @see BDDEncoding
 */
public class PathEncodingVisualizer extends EncodingVisualizer<PathEncoding> {
	
	/*------------------------------------------------------------------------------------------*/
	/*		Protected Methods 																	*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	protected void visualizeStateBits(StateType type, PathEncoding encoding) {
		visualGraph.getModel().beginUpdate();
		try {
			int offset = getOffset(encoding, type);
			int interfaceLength = encoding.getInterfaceLength();
			leftShift = interfaceLength % 2 == 0 ? 1 : 0;
			int groupPos = interfaceLength/2;
			
			Object[] interfaceNodes = new Object[interfaceLength];
			for(int i=0; i<interfaceLength;i++) {
				interfaceNodes[i] = addInterfaceVertex(type, i, String.valueOf(offset+i*2));
				addBitLabel(type, i, String.valueOf(i));
				if(i == groupPos) {
					addGroupLabel(type, i, GuiMessages.getMessages().getString(
							GuiMessages.INTERFACE_BITS_LABEL_KEY));
				}
			}
			mxCell group = (mxCell) visualGraph.groupCells(null, 5, interfaceNodes);
			group.setStyle(VisualGraph.INTERFACE_WHITE_BACKGROUND_STYLE);
			
			int maxInterfaceSize = encoding.getMaximumInterface();
			int PathBitCounter = 1;
			Object[] encodingPathNodes;
			int currentPos = interfaceLength;
			int currentRest;
			
			for(int vertexCounter=0;vertexCounter<maxInterfaceSize;vertexCounter++) {
				
				currentRest = maxInterfaceSize - vertexCounter;
				leftShift = currentRest % 2 == 0 ? 1 : 0;
				groupPos = currentRest/2;				
				encodingPathNodes = new Object[currentRest];
				
				for(int i=currentPos;i<currentPos+currentRest;i++) {
					
					encodingPathNodes[i-currentPos] = addStateVertex(type, i, String.valueOf(offset+i*2));
					addBitLabel(type, i,String.valueOf(i));
					addPathBitLabel(type, i,String.valueOf(i-currentPos));
					if(i-currentPos == groupPos) {
						addGroupLabel(type, i, messages.getString(GuiMessages.VERTEX_LABEL_KEY, String.valueOf(vertexCounter)));
					}
					if(PathBitCounter == currentRest) {
						int index = (vertexCounter%(VisualGraph.COLOR_STYLE.length-2))+2; 
						// -2/+2 to skip the black and white color
						
						group = (mxCell) visualGraph.groupCells(null, 5, encodingPathNodes);
						group.setStyle(VisualGraph.COLOR_STYLE[index]);					
						PathBitCounter = 1;
					}
					else {
						PathBitCounter++;					
					}
				}
				
				currentPos = currentPos + currentRest;		
			}
			
			Object[] SourcePathNodes = new Object[maxInterfaceSize];
			leftShift = maxInterfaceSize % 2 == 0 ? 1 : 0;
			groupPos = maxInterfaceSize/2;	
			currentPos = encoding.getLength()-maxInterfaceSize;
			
			for(int i=currentPos;i<encoding.getLength();i++) {
				
				SourcePathNodes[i-currentPos] = addStateVertex(type, i,String.valueOf(offset+i*2));
				addBitLabel(type, i, String.valueOf(i));
				addSourcePathBitLabel(type, i, String.valueOf(i-currentPos));
				if(i-currentPos == groupPos) {
					addGroupLabel(type, i, messages.getString(GuiMessages.SOURCE_PATH_BITS_LABEL_KEY));
				}
			}
			
			group = (mxCell) visualGraph.groupCells(null, 5, SourcePathNodes);
			group.setStyle(VisualGraph.INTERFACE_WHITE_BACKGROUND_STYLE);
			
		}
		finally{
			visualGraph.getModel().endUpdate();
		}
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods 																	*/
	/*------------------------------------------------------------------------------------------*/
	
	private mxCell addPathBitLabel(StateType type, int number, String label) {
		int row = type==StateType.current ? 0 : 1;
		Object parent = visualGraph.getDefaultParent();
		int x = X_OFFSET + number*X_MARGIN;
		int y = Y_OFFSET + row*Y_MARGIN+Y_MARGIN/5 + GAP_CONSTANT;
		
		return (mxCell) visualGraph.insertVertex(parent, null, 
				messages.getString(GuiMessages.PATH_BIT_LABEL_KEY, label), x, y, 
				INTERFACE_VERTEX_SIZE, INTERFACE_VERTEX_SIZE, VisualGraph.LABEL_STYLE);
	}
	
	private mxCell addSourcePathBitLabel(StateType type, int number, String label) {
		int row = type==StateType.current ? 0 : 1;
		Object parent = visualGraph.getDefaultParent();
		int x = X_OFFSET + number*X_MARGIN;
		int y = Y_OFFSET + row*Y_MARGIN+Y_MARGIN/5 + GAP_CONSTANT;
		
		return (mxCell) visualGraph.insertVertex(parent, null,
				messages.getString(GuiMessages.SOURCE_PATH_BIT_LABEL_KEY, label), x, y, 
				INTERFACE_VERTEX_SIZE, INTERFACE_VERTEX_SIZE, VisualGraph.LABEL_STYLE);
	}

}