/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.gui.graphs;

import java.awt.Color;
import java.awt.FontMetrics;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.lang.ProcessBuilder.Redirect;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;

import org.antlr.v4.runtime.ANTLRInputStream;
import org.antlr.v4.runtime.CommonTokenStream;
import org.antlr.v4.runtime.tree.ParseTree;

import com.mxgraph.layout.mxParallelEdgeLayout;
import com.mxgraph.model.mxCell;
import com.mxgraph.model.mxGeometry;
import com.mxgraph.model.mxIGraphModel;
import com.mxgraph.util.mxConstants;
import com.mxgraph.util.mxEventObject;
import com.mxgraph.util.mxRectangle;
import com.mxgraph.util.mxUtils;
import com.mxgraph.view.mxGraph;
import com.mxgraph.view.mxStylesheet;

import de.uni_due.inf.ti.raven.antlr.DOTBaseVisitor;
import de.uni_due.inf.ti.raven.antlr.DOTLexer;
import de.uni_due.inf.ti.raven.antlr.DOTParser;
import de.uni_due.inf.ti.raven.antlr.DOTParser.A_listContext;
import de.uni_due.inf.ti.raven.antlr.DOTParser.Edge_stmtContext;
import de.uni_due.inf.ti.raven.antlr.DOTParser.Node_idContext;
import de.uni_due.inf.ti.raven.antlr.DOTParser.Node_stmtContext;
import de.uni_due.inf.ti.raven.antlr.DOTVisitor;
import de.uni_due.inf.ti.raven.data.Cospan;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.Node;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.gui.GuiMessages;
import de.uni_due.inf.ti.raven.io.DotPrinter;
import de.uni_due.inf.ti.raven.main.RavenMain;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class is a utility class to layout the different "graphical" objects used 
 * in Raven such as <i>graphs</i>, <i>cospans</i>, <i>automata</i>, ...
 * @author Christoph Blume
 * @version $Revision: 1.34 $
 *
 * @see Automaton
 * @see Cospan
 * @see CospanDecomposition
 * @see Graph
 */
public class VisualGraph extends mxGraph {

	/** 
	 * Enumeration of the different interface types.
	 */
	private static enum InterfaceType {
		/** Inner Interface Type */
		inner,

		/** Outer Interface Type */
		outer
	}

	private static final String DOT_POINT_SEPARATOR_TOKEN = ",";

	//	private static final String DOT_BOX_TOKEN = "box";
	private static final String DOT_ELLIPSE_TOKEN = "ellipse";
	private static final String DOT_LABEL_TOKEN = "label";
	private static final String DOT_POSITION_TOKEN = "pos";
	private static final String DOT_SHAPE_TOKEN = "shape";
	private static final String DOT_TYPE_TOKEN = "type";
	
	/** The DOT Token To Indicate Final States */
	public static final String DOT_FINAL_TOKEN = "final";
	
	/** The DOT Token To Indicate Initial States */
	public static final String DOT_INITIAL_TOKEN = "initial";
	
	/** The DOT Token To Indicate States Which Are Both Initial And Final */
	public static final String DOT_INITIAL_FINAL_TOKEN = "initialFinal";

	private static final String EMPTY_INTERFACE_TOKEN = GuiMessages.EMPTY_INTERFACE_LABEL_KEY;

	
	// The Graph Styles
	/** The Default Label Style */
	public static final String LABEL_STYLE = "ravenLabelStyle";

	/** The "Empty Vertex" Style  */
	public static final String NO_VERTEX_STYLE = "ravenNoVertexStyle";

	/** The Default Hypergraph Vertex Style */
	public static final String HYPERGRAPH_VERTEX_STYLE = "ravenHypergraphVertexStyle";

	/** The Hypergraph Vertex Style Used To Indicate Nodes Contained In The Inner Interface */
	public static final String HYPERGRAPH_INNER_VERTEX_STYLE = "ravenHypergraphInnerVertexStyle";

	/** The Hypergraph Vertex Style Used To Indicate Nodes Contained In The Outer Interface */
	public static final String HYPERGRAPH_OUTER_VERTEX_STYLE = "ravenHypergraphOuterVertexStyle";

	/** The Hypergraph Vertex Style Used To Indicate Nodes Contained In Both The Inner And The Outer Interface */
	public static final String HYPERGRAPH_INNER_OUTER_VERTEX_STYLE = "ravenHypergraphInnerOuterVertexStyle";

	/** The Default Hypergraph Edge Style */
	public static final String HYPERGRAPH_EDGE_STYLE = "ravenHypergraphEdgeStyle";

	/** The Default Hypergraph Tentacle Style */
	public static final String HYPERGRAPH_TENTACLE_STYLE = "ravenHypergraphTentacleStyle";

	/** The Default BDD Vertex Style To Indicate The Terminal "Zero" */
	public static final String BDD_TERMINAL_ZERO_STYLE = "ravenBDDTerminalZeroStyle";

	/** The Default BDD Vertex Style To Indicate The Terminal "One" */
	public static final String BDD_TERMINAL_ONE_STYLE = "ravenBDDTerminalOneStyle";

	/** The Default BDD Vertex Style To Indicate Inner Nodes */
	public static final String BDD_VERTEX_STYLE = "ravenBDDVertexStyle";

	/** The Default BDD Vertex Style To Indicate High Edges */
	public static final String BDD_HIGH_EDGE_STYLE = "ravenBDDHighEdgeStyle";

	/** The Default BDD Vertex Style To Indicate Low Edges */
	public static final String BDD_LOW_EDGE_STYLE = "ravenBDDLowEdgeStyle";
	
	/** The Cospan Mapping Style Which Used Red Colored Arrows */
	public static final String COSPAN_EDGE_STYLE_RED = "ravenCospanEdgeStyleRed";

	/** The Default Background Style For The Inner Interface */
	public static final String INNER_INTERFACE_BACKGROUND_STYLE = "ravenInnerInterfaceStyle";

	/** The Default Background Style For The Outer Interface */
	public static final String OUTER_INTERFACE_BACKGROUND_STYLE = "ravenOuterInterfaceStyle";

	/** The Default Style For Interface Nodes */
	public static final String INTERFACE_VERTEX_STYLE = "ravenInterfaceVertexStyle";

	/** The Background Style For Interfaces Without Background */
	public static final String INTERFACE_NO_BACKGROUND_STYLE = "ravenInterfaceNoBackgroundStyle";

	/** The Background Style For Interfaces With A White Background */
	public static final String INTERFACE_WHITE_BACKGROUND_STYLE = "ravenInterfaceWhiteBackgroundStyle";
	
	private static final String COSPAN_EDGE_STYLE_BLACK = "ravenCospanEdgeStyleBlack";
	private static final String INTERFACE_BLACK_BACKGROUND_STYLE = "ravenInterfaceBlackBackgroundStyle";
	private static final String LEGEND_BACKGROUND_STYLE = "ravenLegendBackgroundStyle";
	private static final String LEGEND_VERTEX_STYLE = "ravenLegendVertexStyle";
	private static final String LEGEND_INNER_VERTEX_STYLE = "ravenLegendInnerVertexStyle";
	private static final String LEGEND_OUTER_VERTEX_STYLE = "ravenLegendOuterVertexStyle";
	private static final String LEGEND_INNER_OUTER_VERTEX_STYLE = "ravenLegendInnerOuterVertexStyle";
	private static final String AUTOMATON_STATE_STYLE = "ravenAutomatonStateStyle";
	private static final String AUTOMATON_FINAL_STATE_STYLE = "ravenAutomatonFinalStateStyle";
	private static final String AUTOMATON_FINAL_INITIAL_STATE_STYLE = "ravenAutomatonFinalInitialStateStyle";
	private static final String AUTOMATON_INITIAL_STATE_STYLE = "ravenAutomatonInitialStateStyle";
	private static final String AUTOMATON_TRANSITION_STYLE = "ravenAutomatonTransitionStyle";

	
	// The color constants

	/** White Color Index */
	public static final int WHITE = 0;

	/** Black Color Index */
	public static final int BLACK = 1;

	/** Red Color Index */
	public static final int RED = 2;

	/** Green Color Index */
	public static final int GREEN = 3;

	/** Blue Color Index */
	public static final int BLUE = 4;

	/** Yellow Color Index */
	public static final int YELLOW = 5;

	/** Cyan Color Index */
	public static final int CYAN = 6;

	/** Orange Color Index */
	public static final int ORANGE = 7;

	/** Magenta Color Index */
	public static final int MAGENTA = 8;

	/** Gray Color Index */
	public static final int GRAY = 9;

	/** The list of available colors */
	public static final String[] COLOR_STYLE = {
		mxUtils.hexString(Color.WHITE),
		mxUtils.hexString(Color.BLACK),
		mxUtils.hexString(Color.RED),
		mxUtils.hexString(Color.GREEN),
		mxUtils.hexString(Color.BLUE),
		mxUtils.hexString(Color.YELLOW),
		mxUtils.hexString(Color.CYAN),
		mxUtils.hexString(Color.ORANGE),
		mxUtils.hexString(Color.MAGENTA),
		mxUtils.hexString(Color.GRAY)
	};

	private static final GuiMessages messages = GuiMessages.getMessages();

	// Must be between -1 and 1
	private static final double EDGE_LABEL_POSITION = 0;
	// Distance in pixels
	private static final double EDGE_LABEL_DISTANCE = 10;

	private static final int X_MARGIN = 50;
	private static final int Y_MARGIN = 50;

	private static final int INTERFACE_NODE_SIZE = 25;
	private static final int HYPERGRAPH_NODE_SIZE = 30;

	private static Graph graph;
	private static List<Node> innerInterface;
	private static List<Node> outerInterface;

	private Map<String, mxCell> nameToVertexMap;

	/*------------------------------------------------------------------------------------------*/
	/*		Public Static Methods 																*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Converts a Raven graph to its graph visualization.
	 * @param graph the graph which is to be visualized
	 * @return the visual graph visualizing the given graph
	 * @throws FileNotFoundException thrown if the DOT executable can not be found
	 * @throws IOException thrown if an I/O exception is thrown during the execution of DOT
	 * @throws InterruptedException thrown if DOT is interrupted during execution
	 */
	public static VisualGraph getVisualGraph(Graph graph) 
			throws FileNotFoundException, IOException, InterruptedException {
		File tempSourceFile = null;

		//Print the graph in DOT-format to a temporary file
		tempSourceFile = File.createTempFile("raven_source_", "."+RavenMain.DOT_FILE_EXTENSION, 
				new File(RavenMain.TEMP_DIR));
		DotPrinter printer = new DotPrinter(tempSourceFile.getAbsolutePath());
		printer.print(graph);

		//Initialize the visual graph
		final VisualGraph visualGraph = new VisualGraph();
		visualGraph.setCellsEditable(false);
		visualGraph.setAllowDanglingEdges(false);
		visualGraph.setAllowLoops(false);
		visualGraph.setCellsDeletable(false);
		visualGraph.setCellsCloneable(false);
		visualGraph.setCellsDisconnectable(false);
		visualGraph.setDropEnabled(false);
		visualGraph.setSplitEnabled(false);
		visualGraph.setCellsBendable(false);
		visualGraph.setConnectableEdges(false);
		visualGraph.setEdgeLabelsMovable(false);
		visualGraph.setVertexLabelsMovable(false);		
		visualGraph.setCellsResizable(false);
		final Object parentChild = visualGraph.getDefaultParent();

		final Map<String, Object> nodeMap = new HashMap<>();
		final Map<String, Object> edgeMap = new HashMap<>();

		Path dotPath = Paths.get(RavenMain.getProperty(RavenMain.PROPERTY_DOT_PATH));
		if(!Files.exists(dotPath) || !Files.isRegularFile(dotPath) ||
				!Files.isExecutable(dotPath)) {
			throw new FileNotFoundException(RavenExceptionMessages.DOT_FILE_NOT_FOUND_EXCEPTION);
		}

		final ProcessBuilder pb = new ProcessBuilder(
				dotPath.toFile().getAbsolutePath(), "-Tdot", 
				"-K"+RavenMain.DOT_GRAPH_LAYOUTER, 
				tempSourceFile.getAbsolutePath());

		pb.redirectError(Redirect.PIPE);
		pb.redirectOutput(Redirect.PIPE);

		final Process dotProcess = pb.start();
		
		new Thread(new Runnable() {

			@Override
			public void run() {
				//Connect error stream of the process to the logger
				try (BufferedReader reader = new BufferedReader(
						new InputStreamReader(dotProcess.getErrorStream()))) {
					String line;
					while((line = reader.readLine())!=null) {
						Logger.getLogger(VisualGraph.class.getName()).warning(line);
					}
				} catch (IOException e) {
					Logger.getLogger(VisualGraph.class.getName()).throwing(getClass().getName(), "run", e);
				}
			}
			
		}).start();

		//Initialize ANTLR Parser and create parsing tree
		ANTLRInputStream input = new ANTLRInputStream(dotProcess.getInputStream());
		DOTLexer lexer = new DOTLexer(input);
		CommonTokenStream tokens = new CommonTokenStream(lexer);
		DOTParser parser = new DOTParser(tokens);
		ParseTree tree = parser.graph();

		//Evaluate parsing tree
		DOTVisitor<Object> eval = new DOTBaseVisitor<Object>() {

			@Override
			public Object visitNode_stmt(Node_stmtContext ctx) {
				Map<String, String> map = visitAttr_list(ctx.attr_list());
				String[] position = map.get(DOT_POSITION_TOKEN).split(
						DOT_POINT_SEPARATOR_TOKEN);
				int x = (int) Float.parseFloat(position[0]);
				int y = (int) -Float.parseFloat(position[1]);

				String style = null;
				Map<String, Object> vertexMap = null;
				if(map.get(DOT_SHAPE_TOKEN).equals(DOT_ELLIPSE_TOKEN)) {
					style = HYPERGRAPH_VERTEX_STYLE;
					vertexMap = nodeMap;
				}
				else {
					style = HYPERGRAPH_EDGE_STYLE;
					vertexMap = edgeMap;
				}

				String id = ctx.node_id().getText();
				Object node = visualGraph.insertVertex(parentChild, id, 
						map.get(DOT_LABEL_TOKEN), x, -y, 30, 30, style);
				vertexMap.put(id, node);
				return node;
			}

			@Override
			public Object visitEdge_stmt(Edge_stmtContext ctx) {
				Map<String, String> map = visitAttr_list(ctx.attr_list());
				String sourceName = visitId(ctx.node_id().id());
				String targetName;
				String label = map.get(DOT_LABEL_TOKEN);

				for(Node_idContext nodeID: ctx.edgeRHS().node_id()) {
					targetName = visitId(nodeID.id());

					Object source = nodeMap.containsKey(sourceName) ? 
							nodeMap.get(sourceName) : edgeMap.get(sourceName);
							Object target = nodeMap.containsKey(targetName) ?
									nodeMap.get(targetName) : edgeMap.get(targetName);
									mxCell edge = (mxCell) visualGraph.insertEdge(parentChild, null, 
											label, source, target, HYPERGRAPH_TENTACLE_STYLE);
									sourceName = targetName;

									mxGeometry geometry = edge.getGeometry();
									geometry.setX(EDGE_LABEL_POSITION);
									geometry.setY(EDGE_LABEL_DISTANCE);
				}

				return null;
			}

			@Override
			public List<String> visitEdgeRHS(DOTParser.EdgeRHSContext ctx) {
				List<String> list = new LinkedList<>();
				for(Node_idContext nodeID: ctx.node_id()) {
					list.add(visitId(nodeID.id()));
				}
				return list;
			}

			@Override
			public Map<String, String> visitAttr_list(DOTParser.Attr_listContext ctx) {
				Map<String, String> map = new HashMap<>();

				for(A_listContext child: ctx.a_list()) {
					map.putAll(visitA_list(child));
				}
				return map;
			};

			@Override
			public Map<String, String> visitA_list(A_listContext ctx) {
				Map<String, String> map = new HashMap<>();

				visitChildren(ctx);
				for(int i=0;i<ctx.getChildCount();i++) {
					String s = ctx.getChild(i).getText();
					if(s.equals("=")) {
						String key = ctx.getChild(i-1).getText().replaceAll("\"", "").trim();
						String value = ctx.getChild(i+1).getText().replaceAll("\"", "").trim();
						map.put(key, value);
						i++;
					}
				}

				return map;
			}

			@Override
			public String visitId(DOTParser.IdContext ctx) {
				return ctx.getText();
			}
		};

		//Run the process and check for errors
		int returnCode = dotProcess.waitFor();
		if(returnCode!=0) {
			return null;
		}

		visualGraph.getModel().beginUpdate();
		eval.visit(tree);
		visualGraph.getModel().endUpdate();

		//Delete temporary layout file and "close" the visual graph
		if(!tempSourceFile.delete()) {
			tempSourceFile.deleteOnExit();
		}

		return visualGraph;
	}

	/**
	 * Converts a Raven cospan to its graph visualization.
	 * @param cospan the cospan which is to be visualized
	 * @return the visual graph visualizing the given cospan
	 * @throws FileNotFoundException thrown if the DOT executable can not be found
	 * @throws IOException thrown if an I/O exception is thrown during the execution of DOT
	 * @throws InterruptedException thrown if DOT is interrupted during execution
	 */
	public static VisualGraph getVisualGraph(Cospan cospan)
			throws FileNotFoundException, IOException, InterruptedException {
		graph = cospan.getGraph();
		innerInterface = cospan.getInnerInterface();
		outerInterface = cospan.getOuterInterface();
		VisualGraph visualGraph = getVisualGraph(cospan.getGraph());
		mxRectangle boundingBox = visualGraph.getGraphBounds();

		//Create the interfaces and labels
		visualGraph.getModel().beginUpdate();
		try {
			//Create the labels for the interfaces and the graph
			double y = boundingBox.getY() - Y_MARGIN/2;
			mxCell labelCell = null;

			if(graph.getNodeCount()>0) {
				labelCell = createLabel(visualGraph, GuiMessages.GRAPH_LABEL_KEY, 
						boundingBox.getX()+(boundingBox.getWidth()/2), y, LABEL_STYLE); 
			} 
			else {
				labelCell = createLabel(visualGraph, EMPTY_INTERFACE_TOKEN, 
						boundingBox.getX()+(boundingBox.getWidth()/2), y, LABEL_STYLE);
			}
			mxGeometry geometry = labelCell.getGeometry();
			geometry.setX(boundingBox.getCenterX() - geometry.getWidth()/2);

			labelCell = createLabel(visualGraph, GuiMessages.INNER_INTERFACE_LABEL_KEY, 
					0, y, LABEL_STYLE);
			geometry = labelCell.getGeometry();
			geometry.setX(boundingBox.getX() - geometry.getWidth() - X_MARGIN/2);
			mxCell interfaceCell = createInterface(visualGraph, InterfaceType.inner, innerInterface, 
					geometry.getCenterX(), boundingBox.getY());
			geometry = interfaceCell.getGeometry();
			geometry.setX(geometry.getX()-geometry.getWidth()/2);

			labelCell = createLabel(visualGraph, GuiMessages.OUTER_INTERFACE_LABEL_KEY, 
					0, y, LABEL_STYLE);
			geometry = labelCell.getGeometry();
			geometry.setX(boundingBox.getX() + boundingBox.getWidth() + X_MARGIN/2);
			double legendX = geometry.getX() + geometry.getWidth() + X_MARGIN/2;

			interfaceCell = createInterface(visualGraph, InterfaceType.outer, outerInterface, 
					geometry.getCenterX(), boundingBox.getY());
			geometry = interfaceCell.getGeometry();
			geometry.setX(geometry.getX()-geometry.getWidth()/2);

			createLegend(visualGraph, legendX, y);
		}
		finally {
			visualGraph.getModel().endUpdate();
		}

		return visualGraph;
	}


	/**
	 * Creates a graph visualization of the run of the automaton {@code automaton} on 
	 * the decomposition {@code decomposition}.
	 * @param automaton the automaton used for the run
	 * @param decomposition the cospan decomposition which is the input for this run
	 * @return the visual graph visualizing the run of the automaton on the decomposition
	 * @throws FileNotFoundException thrown if the DOT executable can not be found
	 * @throws IOException thrown if an I/O exception is thrown during the execution of DOT
	 * @throws InterruptedException thrown if DOT is interrupted during execution
	 */
	public static VisualGraph getVisualGraph(Automaton automaton, 
			CospanDecomposition decomposition) 
					throws FileNotFoundException, IOException, InterruptedException {
		File tempSourceFile = null;

		//Print the graph in DOT-format to a temporary file
		tempSourceFile = File.createTempFile("raven_source_", "."+RavenMain.DOT_FILE_EXTENSION, 
				new File(RavenMain.TEMP_DIR));
		DotPrinter printer = new DotPrinter(tempSourceFile.getAbsolutePath());
		printer.print(automaton, decomposition);

		//Initialize the visual graph
		final VisualGraph visualGraph = new VisualGraph();
		visualGraph.setCellsEditable(false);
		visualGraph.setAllowDanglingEdges(false);
		visualGraph.setAllowLoops(false);
		visualGraph.setCellsDeletable(false);
		visualGraph.setCellsCloneable(false);
		visualGraph.setCellsDisconnectable(false);
		visualGraph.setDropEnabled(false);
		visualGraph.setSplitEnabled(false);
		visualGraph.setCellsBendable(false);
		visualGraph.setConnectableEdges(false);
		visualGraph.setEdgeLabelsMovable(false);
		visualGraph.setVertexLabelsMovable(false);		
		visualGraph.setCellsResizable(false);
		final Object parentChild = visualGraph.getDefaultParent();

		final Map<String, Object> nodeMap = new HashMap<>();

		Path dotPath = Paths.get(RavenMain.getProperty(RavenMain.PROPERTY_DOT_PATH));
		if(!Files.exists(dotPath) || !Files.isRegularFile(dotPath) ||
				!Files.isExecutable(dotPath)) {
			throw new FileNotFoundException(RavenExceptionMessages.DOT_FILE_NOT_FOUND_EXCEPTION);
		}

		final ProcessBuilder pb = new ProcessBuilder(
				dotPath.toFile().getAbsolutePath(), "-Tdot", 
				"-K"+RavenMain.DOT_RUN_LAYOUTER, 
				tempSourceFile.getAbsolutePath());

		pb.redirectError(Redirect.PIPE);
		pb.redirectOutput(Redirect.PIPE);

		final Process dotProcess = pb.start();
		//Connect error stream of the process to the logger
		new Thread(new Runnable() {
			
			@Override
			public void run() {
				try (BufferedReader reader = new BufferedReader(
						new InputStreamReader(dotProcess.getErrorStream()))) {
					String line;
					while((line = reader.readLine())!=null) {
						Logger.getLogger(VisualGraph.class.getName()).warning(line);
					}
				} catch (IOException e) {
					Logger.getLogger(VisualGraph.class.getName()).throwing(getClass().getName(), "run", e);
				}
				
			}
		}).start();
		
		ANTLRInputStream input = new ANTLRInputStream(dotProcess.getInputStream());
		DOTLexer lexer = new DOTLexer(input);
		CommonTokenStream tokens = new CommonTokenStream(lexer);
		DOTParser parser = new DOTParser(tokens);
		ParseTree tree = parser.graph();

		//Evaluate parsing tree
		DOTVisitor<Object> eval = new DOTBaseVisitor<Object>() {

			@Override
			public Object visitNode_stmt(Node_stmtContext ctx) {
				Map<String, String> map = visitAttr_list(ctx.attr_list());
				String[] position = map.get(DOT_POSITION_TOKEN).split(
						DOT_POINT_SEPARATOR_TOKEN);
				int x = (int) Float.parseFloat(position[0]);
				int y = (int) -Float.parseFloat(position[1]);

				String style = null;
				if(map.containsKey(DOT_TYPE_TOKEN) && map.get(DOT_TYPE_TOKEN).equals(DOT_FINAL_TOKEN)) {
					style = AUTOMATON_FINAL_STATE_STYLE;
				}
				else if(map.containsKey(DOT_TYPE_TOKEN) && map.get(DOT_TYPE_TOKEN).equals(DOT_INITIAL_TOKEN)) {
					style = AUTOMATON_INITIAL_STATE_STYLE;
				}
				else if(map.containsKey(DOT_TYPE_TOKEN) && map.get(DOT_TYPE_TOKEN).equals(DOT_INITIAL_FINAL_TOKEN)) {
					style = AUTOMATON_FINAL_INITIAL_STATE_STYLE;
				}
				else {
					style = AUTOMATON_STATE_STYLE;
				}
				
				String id = ctx.node_id().getText();
				Object node = visualGraph.insertVertex(parentChild, id, 
						map.get(DOT_LABEL_TOKEN), x, -y, 30, 30, style);
				nodeMap.put(id, node);
				return node;
			}

			@Override
			public Object visitEdge_stmt(Edge_stmtContext ctx) {
				Map<String, String> map = visitAttr_list(ctx.attr_list());
				String sourceName = visitId(ctx.node_id().id());
				String targetName;
				String label = map.get(DOT_LABEL_TOKEN);

				for(Node_idContext nodeID: ctx.edgeRHS().node_id()) {
					targetName = visitId(nodeID.id());
					Object source = nodeMap.get(sourceName);
					Object target = nodeMap.get(targetName);
										
					mxCell edge = (mxCell) visualGraph.insertEdge(parentChild, null, label, source, 
							target, AUTOMATON_TRANSITION_STYLE);

					mxGeometry geometry = edge.getGeometry();
					geometry.setX(EDGE_LABEL_POSITION);
					geometry.setY(EDGE_LABEL_DISTANCE);
				}

				return null;
			}

			@Override
			public List<String> visitEdgeRHS(DOTParser.EdgeRHSContext ctx) {
				List<String> list = new LinkedList<>();
				for(Node_idContext nodeID: ctx.node_id()) {
					list.add(visitId(nodeID.id()));
				}
				return list;
			}

			@Override
			public Map<String, String> visitAttr_list(DOTParser.Attr_listContext ctx) {
				Map<String, String> map = new HashMap<>();

				for(A_listContext child: ctx.a_list()) {
					map.putAll(visitA_list(child));
				}
				return map;
			};

			@Override
			public Map<String, String> visitA_list(A_listContext ctx) {
				Map<String, String> map = new HashMap<>();

				visitChildren(ctx);
				for(int i=0;i<ctx.getChildCount();i++) {
					String s = ctx.getChild(i).getText();
					if(s.equals("=")) {
						String key = ctx.getChild(i-1).getText().replaceAll("\"", "").trim();
						String value = ctx.getChild(i+1).getText().replaceAll("\"", "").trim();
						map.put(key, value);
						i++;
					}
				}

				return map;
			}

			@Override
			public String visitId(DOTParser.IdContext ctx) {
				return ctx.getText();
			}
		};

		//Run the process and check for errors
		int returnCode = dotProcess.waitFor();
		if(returnCode!=0) {
			return null;
		}

		visualGraph.getModel().beginUpdate();
		eval.visit(tree);
		visualGraph.getModel().endUpdate();

		//Delete temporary layout file and "close" the visual graph
		if(!tempSourceFile.delete()) {
			tempSourceFile.deleteOnExit();
		}
		
		return visualGraph;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Private Static Methods 																*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Sets the default stylesheets for the {@code mxGraph} graph.
	 * @param graph the graph for which the default stylesheets should be set
	 */
	private static void setDefaultStyleSheet(mxGraph graph) {
		Hashtable<String, Object> style;
		mxStylesheet stylesheet = graph.getStylesheet();

		// base style
		Hashtable<String, Object> baseStyle = new Hashtable<>();
		baseStyle.put(mxConstants.STYLE_STROKECOLOR, COLOR_STYLE[RED]);

		//Decomposition Style
		//Decomposition Edge Style Red
		style = new Hashtable<>(baseStyle);
		style.put(mxConstants.STYLE_STROKECOLOR, COLOR_STYLE[RED]);
		stylesheet.putCellStyle(COSPAN_EDGE_STYLE_RED, style);

		//Decomposition Edge Style Black
		style = new Hashtable<>(baseStyle);
		style.put(mxConstants.STYLE_STROKECOLOR, COLOR_STYLE[BLACK]);
		stylesheet.putCellStyle(COSPAN_EDGE_STYLE_BLACK, style);

		//No Vertex Style
		style = new Hashtable<>(baseStyle);
		style.put(mxConstants.STYLE_ALIGN, mxConstants.ALIGN_CENTER);
		style.put(mxConstants.STYLE_FONTSIZE, 14);
		style.put(mxConstants.STYLE_MOVABLE, 0);
		style.put(mxConstants.STYLE_OPACITY, 0);
		style.put(mxConstants.STYLE_SHAPE, mxConstants.NONE);
		style.put(mxConstants.STYLE_VERTICAL_ALIGN, mxConstants.ALIGN_MIDDLE);
		stylesheet.putCellStyle(NO_VERTEX_STYLE, style);

		//Label style
		style = new Hashtable<>(baseStyle);
		style.put(mxConstants.STYLE_FONTCOLOR, COLOR_STYLE[BLACK]);
		style.put(mxConstants.STYLE_FONTSIZE, 16);
		style.put(mxConstants.STYLE_OPACITY, 0);
		style.put(mxConstants.STYLE_MOVABLE, 0);
		style.put(mxConstants.STYLE_SHAPE, mxConstants.NONE);
		stylesheet.putCellStyle(LABEL_STYLE, style);

		//Hypergraph Style
		//Hypergraph Vertex style (Default)
		style = new Hashtable<>(baseStyle);
		style.put(mxConstants.STYLE_ALIGN, mxConstants.ALIGN_CENTER);
		style.put(mxConstants.STYLE_SHAPE, mxConstants.SHAPE_ELLIPSE);
		style.put(mxConstants.STYLE_VERTICAL_ALIGN, mxConstants.ALIGN_MIDDLE);
		stylesheet.putCellStyle(HYPERGRAPH_VERTEX_STYLE, style);

		//Hypergraph Inner Vertex style
		style = new Hashtable<>(baseStyle);
		style.put(mxConstants.STYLE_ALIGN, mxConstants.ALIGN_CENTER);
		style.put(mxConstants.STYLE_FILLCOLOR, COLOR_STYLE[ORANGE]);
		style.put(mxConstants.STYLE_SHAPE, mxConstants.SHAPE_ELLIPSE);
		style.put(mxConstants.STYLE_VERTICAL_ALIGN, mxConstants.ALIGN_MIDDLE);
		stylesheet.putCellStyle(HYPERGRAPH_INNER_VERTEX_STYLE, style);

		//Hypergraph Outer Vertex style
		style = new Hashtable<>(baseStyle);
		style.put(mxConstants.STYLE_ALIGN, mxConstants.ALIGN_CENTER);
		style.put(mxConstants.STYLE_FILLCOLOR, COLOR_STYLE[GREEN]);
		style.put(mxConstants.STYLE_SHAPE, mxConstants.SHAPE_ELLIPSE);
		style.put(mxConstants.STYLE_VERTICAL_ALIGN, mxConstants.ALIGN_MIDDLE);
		stylesheet.putCellStyle(HYPERGRAPH_OUTER_VERTEX_STYLE, style);

		//Hypergraph Inner And Outer Vertex style
		style = new Hashtable<>(baseStyle);
		style.put(mxConstants.STYLE_ALIGN, mxConstants.ALIGN_CENTER);
		style.put(mxConstants.STYLE_FILLCOLOR, COLOR_STYLE[WHITE]);
		style.put(mxConstants.STYLE_SHAPE, mxConstants.SHAPE_ELLIPSE);
		style.put(mxConstants.STYLE_VERTICAL_ALIGN, mxConstants.ALIGN_MIDDLE);
		stylesheet.putCellStyle(HYPERGRAPH_INNER_OUTER_VERTEX_STYLE, style);

		//Hypergraph Edge style
		style = new Hashtable<>(baseStyle);
		style.put(mxConstants.STYLE_SHAPE, mxConstants.SHAPE_RECTANGLE);
		stylesheet.putCellStyle(HYPERGRAPH_EDGE_STYLE, style);

		//Hypergraph Tentacle style
		style = new Hashtable<>(baseStyle);
		style.put(mxConstants.STYLE_ENDARROW, mxConstants.NONE);
		style.put(mxConstants.STYLE_STROKECOLOR, COLOR_STYLE[BLUE]);
		style.put(mxConstants.STYLE_TARGET_PERIMETER_SPACING, 0);
		stylesheet.putCellStyle(HYPERGRAPH_TENTACLE_STYLE, style);

		//BDD Style
		//BDD Terminal Zero Style
		style = new Hashtable<>(baseStyle);
		style.put(mxConstants.STYLE_ALIGN, mxConstants.ALIGN_CENTER);
		style.put(mxConstants.STYLE_SHAPE, mxConstants.SHAPE_RECTANGLE);
		style.put(mxConstants.STYLE_VERTICAL_ALIGN, mxConstants.ALIGN_MIDDLE);
		stylesheet.putCellStyle(BDD_TERMINAL_ZERO_STYLE, style);

		//BDD Terminal Zero Style
		style = new Hashtable<>(baseStyle);
		style.put(mxConstants.STYLE_ALIGN, mxConstants.ALIGN_CENTER);
		style.put(mxConstants.STYLE_SHAPE, mxConstants.SHAPE_RECTANGLE);
		style.put(mxConstants.STYLE_VERTICAL_ALIGN, mxConstants.ALIGN_MIDDLE);
		stylesheet.putCellStyle(BDD_TERMINAL_ONE_STYLE, style);

		//BDD Vertex Style
		style = new Hashtable<>(baseStyle);
		style.put(mxConstants.STYLE_ALIGN, mxConstants.ALIGN_CENTER);
		style.put(mxConstants.STYLE_FILLCOLOR, "#BFBFBF");
		style.put(mxConstants.STYLE_VERTICAL_ALIGN, mxConstants.ALIGN_MIDDLE);
		style.put(mxConstants.STYLE_SHAPE, mxConstants.SHAPE_ELLIPSE);
		style.put(mxConstants.STYLE_STROKECOLOR, COLOR_STYLE[GRAY]);
		stylesheet.putCellStyle(BDD_VERTEX_STYLE, style);

		//BDD Low-edge Style
		style = new Hashtable<>(baseStyle);
		style.put(mxConstants.STYLE_DASHED, true);
		style.put(mxConstants.STYLE_ELBOW, mxConstants.ELBOW_VERTICAL);
		style.put(mxConstants.STYLE_FILLCOLOR, COLOR_STYLE[RED]);
		style.put(mxConstants.STYLE_SHAPE, mxConstants.SHAPE_CURVE);
		style.put(mxConstants.STYLE_STROKECOLOR, COLOR_STYLE[RED]);
		stylesheet.putCellStyle(BDD_LOW_EDGE_STYLE, style);

		//BDD High-edge Style
		style = new Hashtable<>(baseStyle);
		style.put(mxConstants.STYLE_DASHED, false);
		style.put(mxConstants.STYLE_ELBOW, mxConstants.ELBOW_HORIZONTAL);
		style.put(mxConstants.STYLE_FILLCOLOR, COLOR_STYLE[BLUE]);
		style.put(mxConstants.STYLE_SHAPE, mxConstants.SHAPE_CURVE);
		style.put(mxConstants.STYLE_STROKECOLOR, COLOR_STYLE[BLUE]);
		stylesheet.putCellStyle(BDD_HIGH_EDGE_STYLE, style);

		//Interface Style

		//Inner Interface style
		style = new Hashtable<>(baseStyle);
		style.put(mxConstants.STYLE_FILLCOLOR, COLOR_STYLE[ORANGE]);
		style.put(mxConstants.STYLE_MOVABLE, 0);
		stylesheet.putCellStyle(INNER_INTERFACE_BACKGROUND_STYLE, style);

		//Outer Interface style
		style = new Hashtable<>(baseStyle);
		style.put(mxConstants.STYLE_FILLCOLOR, COLOR_STYLE[GREEN]);
		style.put(mxConstants.STYLE_MOVABLE, 0);
		stylesheet.putCellStyle(OUTER_INTERFACE_BACKGROUND_STYLE, style);

		//Interface Vertex style
		style = new Hashtable<>(baseStyle);
		style.put(mxConstants.STYLE_SHAPE, mxConstants.SHAPE_ELLIPSE);
		style.put(mxConstants.STYLE_VERTICAL_ALIGN, mxConstants.ALIGN_MIDDLE);
		style.put(mxConstants.STYLE_ALIGN, mxConstants.ALIGN_CENTER);
		stylesheet.putCellStyle(INTERFACE_VERTEX_STYLE, style);

		//Interface No Background style
		style = new Hashtable<>(baseStyle);
		style.put(mxConstants.STYLE_ALIGN, mxConstants.ALIGN_CENTER);
		style.put(mxConstants.STYLE_MOVABLE, 0);
		style.put(mxConstants.STYLE_FILLCOLOR, mxConstants.NONE);
		style.put(mxConstants.STYLE_SHAPE, mxConstants.SHAPE_ELLIPSE);
		style.put(mxConstants.STYLE_VERTICAL_ALIGN, mxConstants.ALIGN_MIDDLE);
		stylesheet.putCellStyle(INTERFACE_NO_BACKGROUND_STYLE, style);

		//Background style
		for(int i=0; i<COLOR_STYLE.length;i++) {
			style = new Hashtable<>(baseStyle);
			style.put(mxConstants.STYLE_FILLCOLOR, COLOR_STYLE[i]);
			stylesheet.putCellStyle(COLOR_STYLE[i], style);		
		}

		style = new Hashtable<>(baseStyle);
		style.put(mxConstants.STYLE_FILLCOLOR, COLOR_STYLE[BLACK]);
		stylesheet.putCellStyle(INTERFACE_BLACK_BACKGROUND_STYLE, style);

		style = new Hashtable<>(baseStyle);
		style.put(mxConstants.STYLE_FILLCOLOR, COLOR_STYLE[WHITE]);
		stylesheet.putCellStyle(INTERFACE_WHITE_BACKGROUND_STYLE, style);

		//Legend style
		style = new Hashtable<>(baseStyle);
		style.put(mxConstants.STYLE_FILLCOLOR, COLOR_STYLE[WHITE]);
		style.put(mxConstants.STYLE_STROKECOLOR, COLOR_STYLE[BLACK]);
		stylesheet.putCellStyle(LEGEND_BACKGROUND_STYLE, style);

		//Legend Vertex style (Default)
		style = new Hashtable<>(stylesheet.getStyles().get(HYPERGRAPH_VERTEX_STYLE));
		style.put(mxConstants.STYLE_MOVABLE, 0);
		stylesheet.putCellStyle(LEGEND_VERTEX_STYLE, style);

		//Legend Inner Vertex style
		style = new Hashtable<>(stylesheet.getStyles().get(HYPERGRAPH_INNER_VERTEX_STYLE));
		style.put(mxConstants.STYLE_MOVABLE, 0);
		stylesheet.putCellStyle(LEGEND_INNER_VERTEX_STYLE, style);

		//Legend Outer Vertex style
		style = new Hashtable<>(stylesheet.getStyles().get(HYPERGRAPH_OUTER_VERTEX_STYLE));
		style.put(mxConstants.STYLE_MOVABLE, 0);
		stylesheet.putCellStyle(LEGEND_OUTER_VERTEX_STYLE, style);

		//Legend Inner and Outer Vertex style
		style = new Hashtable<>(stylesheet.getStyles().get(HYPERGRAPH_INNER_OUTER_VERTEX_STYLE));
		style.put(mxConstants.STYLE_MOVABLE, 0);
		stylesheet.putCellStyle(LEGEND_INNER_OUTER_VERTEX_STYLE, style);

		//Automaton Run View Styles
		//Automaton State Style
		style = new Hashtable<>(baseStyle);
		style.put(mxConstants.STYLE_ALIGN, mxConstants.ALIGN_CENTER);
		style.put(mxConstants.STYLE_SHAPE, mxConstants.SHAPE_ELLIPSE);
		style.put(mxConstants.STYLE_VERTICAL_ALIGN, mxConstants.ALIGN_MIDDLE);
		stylesheet.putCellStyle(AUTOMATON_STATE_STYLE, style);

		//Automaton Final State Style
		style = new Hashtable<>(baseStyle);
		style.put(mxConstants.STYLE_ALIGN, mxConstants.ALIGN_CENTER);
		style.put(mxConstants.STYLE_STROKECOLOR, COLOR_STYLE[BLACK]);
		style.put(mxConstants.STYLE_FILLCOLOR, COLOR_STYLE[ORANGE]);
		style.put(mxConstants.STYLE_SHAPE, mxConstants.SHAPE_RECTANGLE);
		style.put(mxConstants.STYLE_VERTICAL_ALIGN, mxConstants.ALIGN_MIDDLE);
		stylesheet.putCellStyle(AUTOMATON_FINAL_STATE_STYLE, style);
		
		//Automaton Final and Initial State Style
		style = new Hashtable<>(baseStyle);
		style.put(mxConstants.STYLE_ALIGN, mxConstants.ALIGN_CENTER);
		style.put(mxConstants.STYLE_STROKECOLOR, COLOR_STYLE[BLACK]);
		style.put(mxConstants.STYLE_FILLCOLOR, COLOR_STYLE[WHITE]);
		style.put(mxConstants.STYLE_SHAPE, mxConstants.SHAPE_HEXAGON);
		style.put(mxConstants.STYLE_VERTICAL_ALIGN, mxConstants.ALIGN_MIDDLE);
		stylesheet.putCellStyle(AUTOMATON_FINAL_INITIAL_STATE_STYLE, style);
		
		//Automaton Initial State Style
		style = new Hashtable<>(baseStyle);
		style.put(mxConstants.STYLE_ALIGN, mxConstants.ALIGN_CENTER);
		style.put(mxConstants.STYLE_STROKECOLOR, COLOR_STYLE[BLACK]);
		style.put(mxConstants.STYLE_FILLCOLOR, COLOR_STYLE[GREEN]);
		style.put(mxConstants.STYLE_SHAPE, mxConstants.SHAPE_ELLIPSE);
		style.put(mxConstants.STYLE_VERTICAL_ALIGN, mxConstants.ALIGN_MIDDLE);
		stylesheet.putCellStyle(AUTOMATON_INITIAL_STATE_STYLE, style);

		//Automaton Transition Style
		style = new Hashtable<>(baseStyle);
		style.put(mxConstants.STYLE_DASHED, false);
		style.put(mxConstants.STYLE_ELBOW, mxConstants.ELBOW_VERTICAL);
		style.put(mxConstants.STYLE_STROKECOLOR, COLOR_STYLE[BLACK]);
		style.put(mxConstants.STYLE_FILLCOLOR, COLOR_STYLE[BLACK]);
		style.put(mxConstants.STYLE_SHAPE, mxConstants.SHAPE_CURVE);
		stylesheet.putCellStyle(AUTOMATON_TRANSITION_STYLE, style);
	}

	private static mxCell createLabel(VisualGraph visualGraph, String key, 
			double x, double y, String styleName) {
		Object parent = visualGraph.getDefaultParent();
		// get font metrics
		Map<String, Object> style = visualGraph.getStylesheet().getStyles().get(styleName);
		FontMetrics metrics = mxUtils.getFontMetrics(mxUtils.getFont(style));

		String label = messages.getString(key);
		return (mxCell) visualGraph.insertVertex(parent, null, label, 
				x, y, metrics.stringWidth(label), HYPERGRAPH_NODE_SIZE, styleName);
	}

	private static mxCell createInterface(VisualGraph visualGraph, InterfaceType type, 
			List<Node> interfaceList, double x, double y) {
		Object parent = visualGraph.getDefaultParent();
		Object[] visualNodes = new Object[interfaceList.size()];
		mxCell group = null;

		//If there are any nodes in the inner interface create the necessary vertices...
		if(interfaceList.size()>0) {
			for(int i=0;i<interfaceList.size();i++) {
				//Create a new node for the interface
				String id = type==InterfaceType.inner ? "in" : "out";
				visualNodes[i] = visualGraph.insertVertex(parent, id+i, i, X_MARGIN, i*Y_MARGIN, 
						INTERFACE_NODE_SIZE, INTERFACE_NODE_SIZE, 
						INTERFACE_NO_BACKGROUND_STYLE);
				//Get the corresponding node of the middle graph
				mxCell graphVertex = visualGraph.getVertexByName(interfaceList.get(i).getName());

				//Set the new vertex style
				String style = null;
				switch (graphVertex.getStyle()) {
				case HYPERGRAPH_VERTEX_STYLE:
					switch (type) {
					case inner:
						style = HYPERGRAPH_INNER_VERTEX_STYLE;
						break;

					case outer:
						style = HYPERGRAPH_OUTER_VERTEX_STYLE;
						break;
					}
					break;

				case HYPERGRAPH_INNER_VERTEX_STYLE:
					switch (type) {
					case inner:
						style = HYPERGRAPH_INNER_VERTEX_STYLE;
						break;

					case outer:
						style = HYPERGRAPH_INNER_OUTER_VERTEX_STYLE;
						break;
					}
					break;

				case HYPERGRAPH_OUTER_VERTEX_STYLE:
					switch (type) {
					case inner:
						style = HYPERGRAPH_INNER_OUTER_VERTEX_STYLE;
						break;

					case outer:
						style = HYPERGRAPH_OUTER_VERTEX_STYLE;
						break;
					}
					break;

				case HYPERGRAPH_INNER_OUTER_VERTEX_STYLE:
					style = HYPERGRAPH_INNER_OUTER_VERTEX_STYLE;
					break;
				}

				visualGraph.setCellStyle(style, new Object[]{ graphVertex });
				//Add the edge between the interface and the middle graph's node
				visualGraph.insertEdge(parent, null, null, visualNodes[i], 
						graphVertex, COSPAN_EDGE_STYLE_BLACK);
			}

			//Group the interface nodes
			group = (mxCell) visualGraph.groupCells(null, 5, visualNodes);
			mxGeometry geometry = group.getGeometry();
			geometry.setX(x);
			geometry.setY(y);
			group.setGeometry(geometry);
			switch (type) {
			case inner:
				group.setStyle(INNER_INTERFACE_BACKGROUND_STYLE);
				break;

			case outer:
				group.setStyle(OUTER_INTERFACE_BACKGROUND_STYLE);
				break;
			}
		}
		//...else create an empty inner interface
		else {
			int color = type==InterfaceType.inner ? ORANGE : GREEN;
			Object[] emptyInterface = new Object[] {
					visualGraph.insertVertex(visualGraph.getDefaultParent(), null,
							messages.getString(EMPTY_INTERFACE_TOKEN), X_MARGIN, Y_MARGIN, 
							INTERFACE_NODE_SIZE, INTERFACE_NODE_SIZE, 
							NO_VERTEX_STYLE)
			};

			group = (mxCell) visualGraph.groupCells(null, 5, emptyInterface);
			mxGeometry geometry = group.getGeometry();
			geometry.setX(x);
			geometry.setY(y);
			group.setGeometry(geometry);
			group.setStyle(COLOR_STYLE[color]);
		}

		return group;
	}


	private static void createLegend(VisualGraph visualGraph, double x, double y) {
		Object parent = visualGraph.getDefaultParent();

		Object[] legendNodes = new Object[10];
		legendNodes[0] = createLabel(visualGraph, GuiMessages.LEGEND_LABEL_KEY, x, 0, LABEL_STYLE);

		legendNodes[1] = visualGraph.insertVertex(parent, null, "", x, Y_MARGIN, 
				HYPERGRAPH_NODE_SIZE, HYPERGRAPH_NODE_SIZE, LEGEND_VERTEX_STYLE);
		legendNodes[2] = createLabel(visualGraph, GuiMessages.LEGEND_NOT_IN_INTERFACE_LABEL_KEY, 
				x + HYPERGRAPH_NODE_SIZE + 10, Y_MARGIN, LABEL_STYLE);

		legendNodes[3] = visualGraph.insertVertex(parent, null, "", x, Y_MARGIN*2, 
				HYPERGRAPH_NODE_SIZE, HYPERGRAPH_NODE_SIZE, LEGEND_INNER_VERTEX_STYLE);
		legendNodes[4] = createLabel(visualGraph, GuiMessages.LEGEND_ONLY_IN_INNER_INTERFACE_LABEL_KEY,
				x + HYPERGRAPH_NODE_SIZE + 10, Y_MARGIN*2, LABEL_STYLE);

		legendNodes[5] = visualGraph.insertVertex(parent, null, "", x, Y_MARGIN*3, 
				HYPERGRAPH_NODE_SIZE, HYPERGRAPH_NODE_SIZE, LEGEND_OUTER_VERTEX_STYLE);
		legendNodes[6] = createLabel(visualGraph, GuiMessages.LEGEND_ONLY_IN_OUTER_INTERFACE_LABEL_KEY,
				x + HYPERGRAPH_NODE_SIZE + 10, Y_MARGIN*3, LABEL_STYLE);

		legendNodes[7] = visualGraph.insertVertex(parent, null, "", x, Y_MARGIN*4, 
				HYPERGRAPH_NODE_SIZE, HYPERGRAPH_NODE_SIZE, LEGEND_INNER_OUTER_VERTEX_STYLE);
		legendNodes[8] = createLabel(visualGraph, GuiMessages.LEGEND_IN_BOTH_INTERFACES_LABEL_KEY,
				x + HYPERGRAPH_NODE_SIZE + 10, Y_MARGIN*4, LABEL_STYLE);

		mxCell group = (mxCell) visualGraph.groupCells(null, 5, legendNodes);
		mxGeometry geometry = group.getGeometry();
		geometry.setX(x);
		geometry.setY(y);
		group.setGeometry(geometry);
		group.setStyle(LEGEND_BACKGROUND_STYLE);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new {@code VisualGraph}-object for the visualization of graphs.
	 */
	public VisualGraph() {
		this(null, null);
	}

	/**
	 * Creates a new {@code VisualGraph}-object for the visualization of graphs
	 * depending on the {@code mxIGraphModel} model.
	 * @param model the underlying graph model
	 */
	public VisualGraph(mxIGraphModel model) {
		this(model, null);
	}

	/**
	 * Creates a new {@code VisualGraph}-object for the visualization of graphs
	 * initialized with the {@code myStylesheet} stylesheet.
	 * @param stylesheet the initial stylesheet
	 */
	public VisualGraph(mxStylesheet stylesheet) {
		this(null, stylesheet);
	}

	/**
	 * Creates a new {@code VisualGraph}-object for the visualization of graphs
	 * depending on the {@code mxIGraphModel} model and initialized with the 
	 * {@code myStylesheet} stylesheet.
	 * @param model the underlying graph model
	 * @param stylesheet the initial stylesheet
	 */
	public VisualGraph(mxIGraphModel model, mxStylesheet stylesheet) {
		super(model, stylesheet);

		setDefaultStyleSheet(this);
		nameToVertexMap = new HashMap<>();

		final mxParallelEdgeLayout layout = new mxParallelEdgeLayout(this);
		layout.execute(getDefaultParent());
		addListener("cellsMoved", new mxIEventListener() {

			@Override
			public void invoke(Object sender, mxEventObject evt) {
				layout.execute(getDefaultParent());
			}
		});
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public Object insertVertex(Object parent, String id, Object value,
			double x, double y, double width, double height, String style) {
		mxCell vertex = (mxCell) super.insertVertex(parent, id, value, x, y, width, height, style);

		if(id!=null && !id.isEmpty()) {
			nameToVertexMap.put(id, vertex);
		}

		return vertex;
	}

	@Override
	public Object insertVertex(Object parent, String id, Object value, double x, double y, 
			double width, double height) {
		return insertVertex(parent, id, value, x, y, width, height, null);
	}

	@Override
	public boolean isCellSelectable(Object o) {
		if (o != null) {
			if (o instanceof mxCell) {
				mxCell cell = (mxCell) o;             
				return cell.getStyle()!=null && (
						cell.getStyle().equals(HYPERGRAPH_EDGE_STYLE)  ||
						cell.getStyle().equals(HYPERGRAPH_INNER_VERTEX_STYLE)  ||
						cell.getStyle().equals(HYPERGRAPH_OUTER_VERTEX_STYLE)  ||
						cell.getStyle().equals(HYPERGRAPH_VERTEX_STYLE)  ||
						cell.getStyle().equals(HYPERGRAPH_INNER_OUTER_VERTEX_STYLE) ||
						cell.getStyle().equals(AUTOMATON_STATE_STYLE) ||
						cell.getStyle().equals(AUTOMATON_FINAL_STATE_STYLE) ||
						cell.getStyle().equals(AUTOMATON_FINAL_INITIAL_STATE_STYLE) ||
						cell.getStyle().equals(AUTOMATON_INITIAL_STATE_STYLE));
			}
		}
		return super.isCellSelectable(o);
	}

	/**
	 * Checks whether the given cell is a vertex (of the hypergraph displayed by 
	 * this {@code ViusalGraph}).
	 * @param cell the cell to be checked
	 * @return {@code true} if the cell represents a vertex, {@code false} otherwise
	 */
	public boolean isVertex(mxCell cell) {
		return nameToVertexMap.containsValue(cell);
	}

	/**
	 * Returns the {@code mxCell}-object representing the vertex with the id {@code name}.
	 * @param name the id of the vertex
	 * @return the vertex with the given id
	 * @throws NullPointerException thrown if the id is {@code null}
	 * @throws IllegalArgumentException thrown if the id is empty
	 */
	public mxCell getVertexByName(String name) {
		if(name==null) {
			throw new NullPointerException(RavenExceptionMessages.NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(name.isEmpty()) {
			throw new IllegalArgumentException(RavenExceptionMessages.NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}

		return nameToVertexMap.get(name);
	}

	/**
	 * Sets the style of the {@code mxCell}-object representing the vertex with the id 
	 * {@code name} to the style given by {@code style}.
	 * @param name the id of the vertex
	 * @param style the name of the style
	 * @return {@code true} if the vertex has been found, {@code false} if no vertex could be found 
	 * @throws NullPointerException thrown if the id is {@code null}
	 * @throws IllegalArgumentException thrown if the id is empty
	 */
	public boolean setVertexStyle(String name, String style) {
		if(name==null) {
			throw new NullPointerException(RavenExceptionMessages.NAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(name.isEmpty()) {
			throw new IllegalArgumentException(RavenExceptionMessages.NAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}

		if(style==null || style.isEmpty()) {
			style = HYPERGRAPH_VERTEX_STYLE;
		}

		mxCell vertex = getVertexByName(name);
		if(vertex!=null) {
			setCellStyle(style, new Object[]{ vertex });
			return true;
		}

		return false;
	}

}
