/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.io;

import java.io.PrintStream;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import org.fusesource.jansi.AnsiConsole;

import de.uni_due.inf.ti.raven.cli.ConsoleMessages;
import de.uni_due.inf.ti.raven.data.Cospan;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Edge;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.GraphDecomposition;
import de.uni_due.inf.ti.raven.data.GraphDecomposition.Bag;
import de.uni_due.inf.ti.raven.data.Node;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.PathDecomposition;
import de.uni_due.inf.ti.raven.data.Repository;
import de.uni_due.inf.ti.raven.data.Signature;
import de.uni_due.inf.ti.raven.data.TreeDecomposition;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class offers methods to print data to the console.
 * @author Christoph Blume
 * @version $Revision: 1.42 $
 *
 */
public class ConsolePrinter implements RavenPrinter, BDDPrinter {

	private static ConsolePrinter printer;

	private final PrintStream out;
	private final ConsoleMessages messages; 

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns a printer for printing to the console.
	 * @return a console printer
	 */
	public static ConsolePrinter getPrinter() {
		if(printer==null) {
			printer = new ConsolePrinter();
		}

		return printer;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new console printer which prints to the default system output.
	 */
	private ConsolePrinter() {
		out = new PrintStream(AnsiConsole.wrapOutputStream(System.out));
		messages = ConsoleMessages.getMessages();
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public void print(String string) {
		out.print(string);
		out.flush();
	}
	
	@Override
	public void print(Object object) {
		out.print(object);
		out.flush();
	}

	@Override
	public void println(String string) {
		out.println(string);
		out.flush();
	}

	@Override
	public void println(Object object) {
		out.println(object);
		out.flush();
	}

	@Override
	public void println() {
		out.println();
		out.flush();
	}

	@Override
	public void printAutomaton(Automaton automaton) {
		if(automaton==null) {
			throw new NullPointerException(
					RavenExceptionMessages.AUTOMATON_MUST_NOT_BE_NULL_EXCEPTION);
		}

		try {
			String automatonName = Repository.getRepository().getName(automaton);
			
			TextTable table = new TextTable();
			table.addHeader(0, messages.getString(ConsoleMessages.AUTOMATON_KEY) + " " + 
					messages.getString(ConsoleMessages.INFORMATION_KEY)+":");
			table.addContent(0, messages.getString(ConsoleMessages.ENCODING_KEY)+":", 
					automaton.getEncoding().toString());
			table.addContent(0, "");
			table.addContent(0, "");
			table.addContent(0, "");
			table.addContent(1, messages.getString(ConsoleMessages.NAME_KEY)+":", 
					(automatonName==null ? "---" : automatonName));
			table.addContent(1, messages.getString(ConsoleMessages.TYPE_KEY)+":", automaton.getType());
			table.addContent(1, messages.getString(ConsoleMessages.MAXIMUM_INTERFACE_KEY)+":", 
					automaton.getMaximumInterface());
			table.addContent(1, messages.getString(ConsoleMessages.ENCODING_KEY)+" " + 
					messages.getString(ConsoleMessages.LENGTH_KEY)+":", automaton.getEncoding().getLength());
			table.addContent(2, 0, messages.getString(ConsoleMessages.OPERATION_KEY));
			table.addContent(2, 1, messages.getString(ConsoleMessages.BDD_SIZE_KEY));
			
			int totalNodes = 0;
			for(Operation operation: automaton.getAlphabet()) {
				RavenBDD temp = automaton.getTransitionRelation(operation);
				table.addContent(3, 0, operation.toString());
				table.addContent(3, 1, temp.nodeCount());
				totalNodes += temp.nodeCount();
			}
			table.addContent(3, 0, messages.getString(ConsoleMessages.TOTAL_KEY));
			table.addContent(3, 1, totalNodes);
			table.addContent(4, 0, "# "+messages.getString(ConsoleMessages.STATES_KEY));
			table.addContent(4, 1, automaton.getNumberOfStates());
			
			table.setHeaderSpan(0, 3);
			table.setColumnSpan(0, 0, 3);
			out.println(table.toString());
		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}

	@Override
	public void print(Graph graph) {
		if(graph==null) {
			throw new NullPointerException(
					RavenExceptionMessages.GRAPH_MUST_NOT_BE_NULL_EXCEPTION);
		}

		String graphName = Repository.getRepository().getName(graph);
		
		
		out.println("Graph '"+(graphName==null ? "---" : graphName)+"'");
		
		List<Node> nodeList = new LinkedList<>(graph.getNodes());
		Collections.sort(nodeList);
		
		for(Node node: nodeList) {
			out.println("\t"+messages.getString(ConsoleMessages.NODE_KEY)+": "+node.getName());
		}
		out.println();
		
		List<Edge> edgeList = new LinkedList<>(graph.getEdges());
		Collections.sort(edgeList);
		
		for(Edge edge: edgeList) {
			int arity = edge.getArity();
			out.print("\t"+messages.getString(ConsoleMessages.EDGE_KEY)+": " + 
						edge.getName() + "=" + edge.getLabel() +"(");
			int i=0;
			for(Node node: edge.getIncidentNodes()) {
				out.print(node.getName());
				if(i<arity-1) {
					out.print(",");
					i++;
				}
			}
			out.println(")");
		}
		out.println();
	}

	@Override
	public void print(Signature signature) {
		if(signature==null) {
			throw new NullPointerException(
					RavenExceptionMessages.SIGNATURE_MUST_NOT_BE_NULL_EXCEPTION);
		}

		TextTable table = new TextTable();
		table.addHeader(0, messages.getString(ConsoleMessages.SIGNATURE_KEY)+" " + 
				messages.getString(ConsoleMessages.INFORMATION_KEY)+":");
		for(Operation operation: signature) {
			table.addContent(0, 0, operation.toString());
		}

		out.println(table.toString());
	}

	@Override
	public void print(RavenBDD bdd) {
		if(bdd==null) {
			throw new NullPointerException(
					RavenExceptionMessages.BDD_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		bdd.printDot(out);
	}

	@Override
	public void print(Cospan cospan) {
		if(cospan==null) {
			throw new NullPointerException(
					RavenExceptionMessages.COSPAN_MUST_NOT_BE_NULL_EXCEPTION);
		}

		String cospanName = Repository.getRepository().getName(cospan);

		out.println(messages.getString(ConsoleMessages.COSPAN_KEY)+" '"+ 
						(cospanName==null ? "---" : cospanName)+"'");
		out.println("- "+messages.getString(ConsoleMessages.INNER_INTERFACE_KEY)+":");
		
		int counter = 0;
		for(Node node: cospan.getInnerInterface()) {
			out.println("\t* "+(counter++)+": "+node.getName());
		}
		
		out.println("- "+messages.getString(ConsoleMessages.OUTER_INTERFACE_KEY)+":");
		
		counter = 0;
		for(Node node: cospan.getOuterInterface()) {
			out.println("\t* "+(counter++)+": "+node.getName());
		}
		
		out.println("- "+messages.getString(ConsoleMessages.MIDDLE_GRAPH_KEY)+":");
		for(Node node: cospan.getNodes()) {
			out.println("\t* "+messages.getString(ConsoleMessages.NODE_KEY)+": "+node.getName());
		}
		out.println();
		if(!cospan.getEdges().isEmpty()) {
			for(Edge edge: cospan.getEdges()) {
				int arity = edge.getArity();
				out.print("\t* "+messages.getString(ConsoleMessages.EDGE_KEY)+": "+
							edge.getName() + "=" + edge.getLabel() +"(");
				int i=0;
				for(Node node: edge.getIncidentNodes()) {
					out.print(node.getName());
					if(i<arity-1) {
						out.print(",");
						i++;
					}
				}
				out.println(")");
			}
		}
	}
	
	@Override
	public void print(CospanDecomposition decomposition) {
		boolean isFirst = true;
		
		println(messages.getString(ConsoleMessages.TREE_DECOMPOSITION_KEY)+" '" + 
					Repository.getRepository().getName(decomposition)+"': ");
		TextTable table = new TextTable();
		table.addHeader(0, messages.getString(ConsoleMessages.INTERFACE_SIZES_KEY)+":");
		table.addContent(0, 
				messages.getString(ConsoleMessages.INNER_INTERFACE_KEY)+":",
				messages.getString(ConsoleMessages.OUTER_INTERFACE_KEY)+":", 
				messages.getString(ConsoleMessages.MAXIMUM_INTERFACE_KEY)+":", 
				messages.getString(ConsoleMessages.LENGTH_KEY)+":");
		table.addContent(0, decomposition.getInnerInterfaceSize(), 
				decomposition.getOuterInterfaceSize(), decomposition.getWidth(), 
				decomposition.getLength());
//		table.addContent(2, 0, messages.getString(ConsoleMessages.TOTAL_KEY));
//		table.addContent(2, 1, automaton.getSize());
//		table.addContent(3, 0, "# "+messages.getString(ConsoleMessages.STATES_KEY));
//		table.addContent(3, 1, automaton.getNumberOfStates());

		table.setHeaderSpan(0, 1);
		println(table.toString());
		
		print("\t");
		for(Operation operation: decomposition) {
			if(isFirst) {
				isFirst = false;
			}
			else {
				print("; ");
			}

			print(operation.toString());
		}
		
		println();
	}

	@Override
	public void print(GraphDecomposition decomposition) {
		if(decomposition.getClass().equals(TreeDecomposition.class)) {
			println(messages.getString(ConsoleMessages.TREE_DECOMPOSITION_KEY)+":");
		}
		else if(decomposition.getClass().equals(PathDecomposition.class)) {
			println(messages.getString(ConsoleMessages.PATH_DECOMPOSITION_KEY)+":");
		}
		else {
			println(messages.getString(ConsoleMessages.GRAPH_DECOMPOSITION_KEY)+":");
		}
		
		StringBuilder builder = new StringBuilder();
		for(Bag bag: decomposition) {
			print(bag.getIdentifier()+": ");
			boolean isFirst = true;
			
			for(Bag neighbor: bag.getNeighbors()) {
				if(!isFirst) {
					print(", ");
				}
				
				print(neighbor.getIdentifier());
				isFirst = false;
			}
			println();
			builder.append(bag.getIdentifier()+" = "+bag.getNodes()+System.lineSeparator());
		}
		
		println();
		println(builder.toString());
	}

}
