/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.io;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.List;

import de.uni_due.inf.ti.raven.cli.Console;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * Class for writing operation files containing a list of operations from which a 
 * cospan decomposition can be obtained.
 * @author Laura Steinert
 * @version $Revision: 1.4 $
 *
 */
public class CospanDecompositionFileWriter {
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Static Methods																*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Writes a cospan decomposition to a file.
	 * @param filename the name of the file to which the cospan decomposition is written
	 * @param decomposition the cospan decomposition to be written
	 * @throws IOException thrown if an I/O error occurred
	 */
	public static void writeOperations(String filename, CospanDecomposition decomposition) 
			throws IOException {
		File file = new File(filename);
		if(file.exists() && !file.isFile()) {
			throw new IllegalArgumentException(RavenExceptionMessages.FILE_NOT_FOUND_EXCEPTION);
		}
		
		// creates file if it doesn't already exist
		file.getParentFile().mkdirs();
		file.createNewFile();
		
		try(BufferedWriter br = new BufferedWriter(new FileWriter(file))) {
			br.write(Console.INTERFACE_SIZE_TOKEN+" = "+decomposition.getInnerInterfaceSize());
			br.newLine();
			
			List<Operation> operations = decomposition.getOperations();
			// write all but last operation
			for(int i=0; i<operations.size()-1; i++){
				br.write(operations.get(i).parse());
				br.newLine();
			}
			// write last without \n
			br.write(operations.get(operations.size()-1).parse());
		};
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	private CospanDecompositionFileWriter() {}
	
}
