/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.io;

import java.io.BufferedOutputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintStream;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Edge;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.GraphDecomposition;
import de.uni_due.inf.ti.raven.data.GraphDecomposition.Bag;
import de.uni_due.inf.ti.raven.data.Node;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDDomain;
import de.uni_due.inf.ti.raven.gui.graphs.VisualGraph;

/**
 * This class provides methods to print BDDs, cospan decompositions or graphs to files in 
 * DOT format.
 * @author Christoph Blume
 * @version $Revision: 1.24 $
 * 
 * @see RavenBDD
 * @see CospanDecomposition
 * @see Graph
 */
public class DotPrinter implements BDDPrinter, DecompositionPrinter, GraphPrinter, RunPrinter {

	private static class Tree {

		private static int counter = 0;
		
		private final int uid;
		private final RavenBDD content;
		private final List<Tree> children;
		private int height;
		
		/**
		 * 
		 * @param content the content of (the root of) this tree
		 */
		public Tree(RavenBDD content) {
			this.content = content;
			this.children = new LinkedList<>();
			uid = counter++;
		}

		/**
		 * 
		 * @param child the child to be added to the root of this tree
		 */
		public void addChild(Tree child) {
			children.add(child);
			child.height = height+1;
		}
		
		public String getNodeName() {
			return String.valueOf(RavenBDDDomain.getInteger(content));
		}
		
		@Override
		public String toString() {
			return "node"+uid;
		}
		
	}
	
	private BufferedOutputStream ostream;
	private final Logger logger;

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new printer which prints its output in DOT syntax to the file named by {@code filename}.
	 * @param filename the name of the output file
	 * @throws FileNotFoundException thrown if the file named by {@code filename} can not be found
	 */
	public DotPrinter(String filename) throws FileNotFoundException {
		ostream = new BufferedOutputStream(new FileOutputStream(filename));

		// Register logger
		logger = Logger.getLogger(getClass().getName());
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public void print(CospanDecomposition decomposition) {
		for(Operation operation: decomposition) {
			operation.toString();
		}
	}

	@Override
	public void print(Graph graph) {
		// Log entering
		logger.entering(getClass().getName(), "print", graph);

		try (OutputStreamWriter writer = new OutputStreamWriter(ostream)) {
			writer.write("graph hypergraph {\n");

			for(Node node: graph.getNodes()) {
				writer.write("\t"+node.getName()+" [label=\""+node.getName()+"\", shape=\"ellipse\"];\n");
			}

			for(Edge edge: graph.getEdges()) {
				writer.write("\t"+edge.getName()+" [label=\""+edge.getLabel()+"\", shape=\"box\"];\n");
				int i=0;
				for(Node node: edge.getIncidentNodes()) {
					writer.write("\t"+node.getName()+" -- "+edge.getName()+" [label=\""+i+"\"];\n");
					i++;
				}
			}

			writer.write("}");
		} catch (IOException e) {
			logger.warning(e.getMessage());
			logger.throwing(getClass().getName(), "print", e);
		}

		// Log exiting
		logger.exiting(getClass().getName(), "print");
	}

	@Override
	public void print(GraphDecomposition decomposition) {
		// Log entering
		logger.entering(getClass().getName(), "print", decomposition);

		try (OutputStreamWriter writer = new OutputStreamWriter(ostream)) {
			writer.write("strict graph decomp {\n");


			for(Bag bag: decomposition) {
				writer.write("\t"+bag.getIdentifier()+" [label=\""+bag+"\"];\n");
				for(Bag neighbor: bag.getNeighbors()) {
					writer.write("\t"+bag.getIdentifier()+" -- "+neighbor.getIdentifier()+";\n");
				}
			}

			writer.write("}");
		} catch (IOException e) {
			logger.warning(e.getMessage());
			logger.throwing(getClass().getName(), "print", e);
		}

		// Log exiting
		logger.exiting(getClass().getName(), "print");
	}

	@Override
	public void print(RavenBDD bdd) {
		// Log entering
		logger.entering(getClass().getName(), "print", bdd);

		try (PrintStream printStream = new PrintStream(ostream)) {
			bdd.printDot(printStream);			
		}

		// Log exiting
		logger.exiting(getClass().getName(), "print");
	}

	@Override
	public void print(Automaton automaton, CospanDecomposition decomposition) {
		// Log entering
		logger.entering(getClass().getName(), "print", decomposition);
		try (OutputStreamWriter writer = new OutputStreamWriter(ostream)) {
			writer.write("digraph run {\n");
			writer.write("\trankdir = LR;\n");

			RavenBDDDomain stateDomain = automaton.getEncoding().getCurrentStateDomain();
			List<Tree> treeList = new LinkedList<>();
			
			for(RavenBDD state: automaton.getInitialStates().getSatAssignments(stateDomain.getDomainBDD())) {
				Tree tree = new Tree(state);
				treeList.add(tree);

				writer.write("\t"+tree+" [label=\""+tree.getNodeName()+"\" ,type=\""+VisualGraph.DOT_INITIAL_TOKEN+"\"];\n");
			}

			List<Tree> currentElements = new LinkedList<>(treeList);
			List<Tree> successorElements;
			
			for(Operation letter: decomposition) {
				successorElements = new LinkedList<>();

				for(Tree currentElement: currentElements) {
					RavenBDD successorStates = automaton.getSuccessorStates(currentElement.content, letter);

					for(RavenBDD successorState: successorStates.getSatAssignments(stateDomain.getDomainBDD())) {
						Tree tree = new Tree(successorState);
						currentElement.addChild(tree);
						successorElements.add(tree);
						
						String type = automaton.isAcceptingStateSet(successorState) ? ", type=\""+VisualGraph.DOT_FINAL_TOKEN+"\"" : "";
						writer.write("\t"+tree+" [label=\""+tree.getNodeName()+"\""+type+"];\n");
					}
				}

				currentElements = successorElements;
			}
			
			while(!treeList.isEmpty()) {
				Tree tree = treeList.remove(0);
				
				for(Tree child: tree.children) {
					treeList.add(child);
					
					writer.write("\t"+tree+" -> "+child+" [label=\""+decomposition.get(tree.height)+"\"];\n");
				}
				
			}

			writer.write("}");
		} catch (IOException e) {
			logger.warning(e.getMessage());
			logger.throwing(getClass().getName(), "print", e);
		}

		// Log exiting
		logger.exiting(getClass().getName(), "print");
	}

}
