/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.io;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Logger;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import de.uni_due.inf.ti.raven.data.Cospan;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.Node;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class provides operations for reading graphs and cospans from files.
 * @author Christoph Blume, Dennis Nolte, Laura Steinert
 * @version $Revision: 1.30 $
 * 
 */
public class GXLReader implements RavenReader {

	// GXL Tags
	private static final String GXL_TAG_ATTR = "attr";
	private static final String GXL_TAG_EDGE = "edge";
	private static final String GXL_TAG_GRAPH = "graph";
	private static final String GXL_TAG_NODE = "node";
	private static final String GXL_TAG_REL = "rel";
	private static final String GXL_TAG_RELEND = "relend";
	private static final String GXL_TAG_STRING = "string";

	// GXL Attributes
	private static final String GXL_ATTRIBUTE_FROM = "from";
	private static final String GXL_ATTRIBUTE_EDGEIDS = "edgeids";
	private static final String GXL_ATTRIBUTE_ID = "id";
	private static final String GXL_ATTRIBUTE_INT = "int";
	private static final String GXL_ATTRIBUTE_POS = "pos";
	private static final String GXL_ATTRIBUTE_TARGET = "target";
	private static final String GXL_ATTRIBUTE_TO = "to";

	// GXL Values
	private static final String GXL_VALUE_INNER_INTERFACE = "innerInterface";
	private static final String GXL_VALUE_LABEL = "label";
	private static final String GXL_VALUE_OUTER_INTERFACE = "outerInterface";
	private static final String GXL_VALUE_TRUE = "true";

	private static final Logger logger = Logger.getLogger(GXLReader.class.getName());

	/*------------------------------------------------------------------------------------------*/
	/* 		Public Static Methods																*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Reads a graph from a GXL file and returns it.
	 * 
	 * @param filename the name of the file from which the graph is read
	 * @return the graph read from the file
	 * @throws FileNotFoundException thrown if the file to be read could not be found or if the 
	 * corresponding DTD file could not be found
	 * @throws IOException thrown if an I/O error occurred
	 */
	public static Graph readGraph(String filename) throws FileNotFoundException, IOException {
		String graphName;
		Graph graph = null;// = new Graph(name, nodes, edges);
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		// Initialisiere XML-Parser
		factory.setValidating(true);
		factory.setIgnoringElementContentWhitespace(true);
		DocumentBuilder builder;
		try {
			builder = factory.newDocumentBuilder();
			GXLErrorHandler errorHandler = new GXLErrorHandler();
			builder.setErrorHandler(errorHandler);
			Document document = builder.parse(new File(filename));

			// check whether parsing went okay
			// otherwise return null
			if (errorHandler.getErrorOccured()) {
				return null;
			}
			// open file and get the graph node
			Element root = document.getDocumentElement();
			Element graphElement = (Element) root.getElementsByTagName(GXL_TAG_GRAPH).item(0);
			graphName = graphElement.getAttribute(GXL_ATTRIBUTE_ID);

			graph = new Graph(graphName);

			// get nodes
			extractNodes(graphElement, graph);

			// get edges
			extractEdges(graphElement, graph);
		} catch (ParserConfigurationException e) {
			logger.throwing(GXLReader.class.getName(), "readGraph", e);
			return null;
		} catch (SAXException e) {
			logger.throwing(GXLReader.class.getName(), "readGraph", e);
			return null;
		}
		return graph;
	}

	/**
	 * Reads a cospan from a GXL file and returns it.
	 * 
	 * @param filename the name of the file from which the cospan is read
	 * @return the cospan read from the file
	 * @throws FileNotFoundException thrown if the file to be read could not be found or if the 
	 * corresponding DTD file could not be found
	 * @throws IOException thrown if an I/O error occurred
	 */
	public static Cospan readCospan(String filename) throws FileNotFoundException, IOException {
		String graphName;
		Cospan cospan = null; // = new Cospan(graph,innerInterface,outerInterface)
		Graph graph = null;// = new Graph(name, nodes, edges);
		List<String> innerInterfaceNames = new LinkedList<>();
		List<String> outerInterfaceNames = new LinkedList<>();
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();

		// Initialisiere XML-Parser
		factory.setValidating(true);
		factory.setIgnoringElementContentWhitespace(true);
		DocumentBuilder builder;
		try {
			builder = factory.newDocumentBuilder();
			GXLErrorHandler errorHandler = new GXLErrorHandler();
			builder.setErrorHandler(errorHandler);
			Document document = builder.parse(new File(filename));

			// check whether parsing went okay
			// otherwise return null
			if (errorHandler.getErrorOccured()) {
				return null;
			}

			Element innerInterfaceElement = null;
			Element graphElement = null;
			Element outerInterfaceElement = null;


			// open file and get the inner interface node
			Element root = document.getDocumentElement();
			NodeList allElements = root.getElementsByTagName(GXL_TAG_GRAPH);

			if(allElements.getLength() == 3) {
				for(int i=0; i<allElements.getLength();i++) {
					Element element = (Element) allElements.item(i);

					if(element.getAttribute(GXL_ATTRIBUTE_ID).equals(GXL_VALUE_INNER_INTERFACE)) {
						innerInterfaceElement = element;
					}
					else if(element.getAttribute(GXL_ATTRIBUTE_ID).equals(GXL_VALUE_OUTER_INTERFACE)) {
						outerInterfaceElement = element;
					}
					else {
						graphElement = element;
					}
				}				
			}
			else {
				return null;
			}

			innerInterfaceNames = extractInterfaceNodes(innerInterfaceElement);

			// get the graph node
			graphName = graphElement.getAttribute(GXL_ATTRIBUTE_ID);
			graph = new Graph(graphName);
			extractNodes(graphElement, graph);
			// get edges
			extractEdges(graphElement, graph);

			//get the outer interface node
			outerInterfaceNames = extractInterfaceNodes(outerInterfaceElement);
		} catch (ParserConfigurationException e) {
			logger.throwing(GXLReader.class.getName(), "readCospan", e);
			return null;
		} catch (SAXException e) {
			logger.throwing(GXLReader.class.getName(), "readCospan", e);
			return null;
		} catch (Exception e) {
			logger.throwing(GXLReader.class.getName(), "readCospan", e);
			return null;
		}

		List<Node> innerInterface = buildInterface(graph, innerInterfaceNames);
		List<Node> outerInterface = buildInterface(graph, outerInterfaceNames);

		cospan = new Cospan(graph, innerInterface, outerInterface);
		return cospan;
	}

	/*------------------------------------------------------------------------------------------*/
	/* 		Private Static Methods																*/
	/*------------------------------------------------------------------------------------------*/

	private static List<String> extractInterfaceNodes(Element innerInterfaceElement) {
		List<String> interfaceList = new LinkedList<>();
		String target = null;
		int pos = 0;
		//get the nodes
		NodeList innerElements = innerInterfaceElement.getElementsByTagName(GXL_TAG_NODE);
		String[] interfaceOrder = new String[innerElements.getLength()];
		for (int i = 0; i < innerElements.getLength(); i++) {
			org.w3c.dom.Node n = innerElements.item(i);
			target = null;
			pos = 0;

			//get all childs in a node
			NodeList attributesList = n.getChildNodes();
			for(int j=0; j< attributesList.getLength();j++) {

				//get childs of attr j
				if(attributesList.item(j).getNodeName().equals(GXL_TAG_ATTR)) {
					//check the attr name to decide which parameter is stored within
					NamedNodeMap attr = attributesList.item(j).getAttributes();
					for(int k=0; k<attr.getLength();k++) {

						NodeList childs = attributesList.item(j).getChildNodes();
						//if the attr node j is a node named target...
						if(attr.item(k).getNodeValue().equals(GXL_ATTRIBUTE_TARGET)) {

							//...extract the string saved within
							for (int m = 0; m < childs.getLength(); m++) {
								if (childs.item(m).getNodeName().equals(GXL_TAG_STRING)) {
									target = childs.item(m).getTextContent();
								}
							}	
						}
						//else if the attr node j is a node named pos...
						else if(attr.item(k).getNodeValue().equals(GXL_ATTRIBUTE_POS)) {

							//...extract the int saved within
							for (int m = 0; m < childs.getLength(); m++) {
								if (childs.item(m).getNodeName().equals(GXL_ATTRIBUTE_INT)) {
									pos = Integer.parseInt(childs.item(m).getTextContent());
								}
							}	
						}
					}
				}	
			}
			if((pos>=0) && (pos<interfaceOrder.length)) {
				interfaceOrder[pos] = target;				
			}
		}
		for(String nodeName: interfaceOrder) {
			if(!(nodeName == null)) {
				interfaceList.add(nodeName);
			}
		}

		return interfaceList;
	}


	/**
	 * Creates an interface list for the graph {@code graph} containing all nodes 
	 * given by the name list {@code nodeNames} 
	 * @param graph the graph for which the interface is build
	 * @param nodeNames the list of the node names
	 * @return a list of nodes representing the interface
	 */
	private static List<Node> buildInterface(Graph graph,
			List<String> nodeNames) {
		List<Node> interfaceList = new LinkedList<>();

		for(String name: nodeNames) {
			interfaceList.add(graph.getNode(name));
		}
		return interfaceList;
	}

	/**
	 * Extracts all edges from the graph indicated by the GXL-Graph-Tag 
	 * {@code graphElement} and stores them to the graph data structure {@code graph}.
	 * @param graphElement the graph tag which indicates the graph
	 * @param graph the graph in which the nodes are stored
	 */
	private static void extractNodes(Element graphElement, Graph graph) {
		if(graphElement==null) {
			throw new NullPointerException(
					RavenExceptionMessages.GXL_TAG_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(!graphElement.getNodeName().equals(GXL_TAG_GRAPH)) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_GXL_TAG_EXCEPTION);
		}
		if(graph==null) {
			throw new NullPointerException(
					RavenExceptionMessages.GRAPH_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		NodeList nodeElements = graphElement.getElementsByTagName(GXL_TAG_NODE);
		for (int i = 0; i < nodeElements.getLength(); i++) {
			org.w3c.dom.Node n = nodeElements.item(i);
			String name = n.getAttributes().getNamedItem(GXL_ATTRIBUTE_ID)
					.getNodeValue();
			graph.addNode(name);
		}
	}
	
	/**
	 * Extracts all edges from the graph indicated by the GXL-Graph-Tag 
	 * {@code graphElement} and stores them to the graph data structure {@code graph}.
	 * @param graphElement the graph tag which indicates the graph  
	 * @param graph the graph in which the nodes are stored
	 */
	private static void extractEdges(Element graphElement, Graph graph) {
		if(graphElement==null) {
			throw new NullPointerException(
					RavenExceptionMessages.GXL_TAG_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if(!graphElement.getNodeName().equals(GXL_TAG_GRAPH)) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_GXL_TAG_EXCEPTION);
		}
		if(graph==null) {
			throw new NullPointerException(
					RavenExceptionMessages.GRAPH_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		String edgeIds = graphElement.getAttribute(GXL_ATTRIBUTE_EDGEIDS);
		boolean edgeIdsUsed = false;
		if (edgeIds.equals(GXL_VALUE_TRUE)) {
			edgeIdsUsed = true;
		}

		NodeList edgeElements = graphElement.getElementsByTagName(GXL_TAG_EDGE);
		for (int i = 0; i < edgeElements.getLength(); i++) {
			org.w3c.dom.Node e = edgeElements.item(i);
			String name = "";
			if (edgeIdsUsed) {
				name = e.getAttributes().getNamedItem(GXL_ATTRIBUTE_ID).getNodeValue();
			}

			String from = e.getAttributes().getNamedItem(GXL_ATTRIBUTE_FROM).getNodeValue();
			String to = e.getAttributes().getNamedItem(GXL_ATTRIBUTE_TO).getNodeValue();
			Node source = graph.getNode(from);
			Node target = graph.getNode(to);
			LinkedList<Node> nodesForEdge = new LinkedList<>();
			nodesForEdge.add(source);
			nodesForEdge.add(target);
			graph.addEdge(name, name, nodesForEdge);
		}

		// get edges named rel
		NodeList relElements = graphElement.getElementsByTagName(GXL_TAG_REL);
		for (int i = 0; i < relElements.getLength(); i++) {
			org.w3c.dom.Node rel = relElements.item(i);
			String id = "";
			if (edgeIdsUsed) {
				id = rel.getAttributes().getNamedItem(GXL_ATTRIBUTE_ID).getNodeValue();
			}
			String label = id;

			List<Node> incidentNodes = new LinkedList<>();
			NodeList labelNodes = rel.getChildNodes();
			for (int j = 0; j < labelNodes.getLength(); j++) {
				if (labelNodes.item(j).getNodeName().equals(GXL_TAG_RELEND)) {
					String target = labelNodes.item(j).getAttributes().getNamedItem(GXL_ATTRIBUTE_TARGET).getNodeValue();
					incidentNodes.add(graph.getNode(target));
				} else if (labelNodes.item(j).getNodeName().equals(GXL_TAG_ATTR)) {
					NamedNodeMap attributes = labelNodes.item(j).getAttributes();
					for (int k = 0; k < attributes.getLength(); k++) {
						if (attributes.item(k).getNodeValue().equals(GXL_VALUE_LABEL)) {
							NodeList children = labelNodes.item(j).getChildNodes();
							for (int m = 0; m < children.getLength(); m++) {
								if (children.item(m).getNodeName().equals(GXL_TAG_STRING)) {
									label = children.item(m).getTextContent();
								}
							}
						}
					}
				}
			}
			graph.addEdge(id, label, incidentNodes);
		}
	}

	/*------------------------------------------------------------------------------------------*/
	/* 		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * 
	 */
	private GXLReader() {
		// Do nothing
	}


}
