/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.io;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.nio.charset.StandardCharsets;
import java.util.List;
import java.util.Set;

import de.uni_due.inf.ti.raven.data.Cospan;
import de.uni_due.inf.ti.raven.data.Edge;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.Node;
import de.uni_due.inf.ti.raven.main.RavenMain;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * Writes a cospan or graph to a GXL file.
 * @author Christoph Blume, Dennis Nolte, Laura Steinert
 * @version $Revision: 1.24 $
 */
public class GXLWriter implements GraphWriter, CospanWriter {

	private enum InterfaceType {
		/** Inner Interface Type */
		inner(GXL_VALUE_INNER_INTERFACE),
		
		/** Outer Interface Type */
		outer(GXL_VALUE_OUTER_INTERFACE);
		
		private String name;

		private InterfaceType(String name) {
			this.name = name;
		}
		
		@Override
		public String toString() {
			return name;
		}
	}
	
	private static final String XML_DECLARATION = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>";
	private static final String DOCUMENT_TYPE_DECLARATION = "<!DOCTYPE gxl SYSTEM \"" + RavenMain.GXL_DTD_URI + "\">";
	
	// GXL Tags
	private static final String GXL_TAG_ATTR = "attr";
	private static final String GXL_TAG_GRAPH = "graph";
	private static final String GXL_TAG_GXL = "gxl";
	private static final String GXL_TAG_INT = "int";
	private static final String GXL_TAG_NODE = "node";
	private static final String GXL_TAG_REL = "rel";
	private static final String GXL_TAG_RELEND = "relend";
	private static final String GXL_TAG_STRING = "string";

	// GXL Attributes
	private static final String GXL_ATTRIBUTE_EDGEIDS = "edgeids";
	private static final String GXL_ATTRIBUTE_EDGEMODE = "edgemode";
	private static final String GXL_ATTRIBUTE_HYPERGRAPH = "hypergraph";
	private static final String GXL_ATTRIBUTE_ID = "id";
	private static final String GXL_ATTRIBUTE_NAME = "name";
	private static final String GXL_ATTRIBUTE_ROLE = "role";
	private static final String GXL_ATTRIBUTE_STARTORDER = "startorder";
	private static final String GXL_ATTRIBUTE_TARGET = "target";

	// GXL Values
	private static final String GXL_VALUE_INNER_INTERFACE = "innerInterface";
	private static final String GXL_VALUE_LABEL = "label";
	private static final String GXL_VALUE_OUTER_INTERFACE = "outerInterface";
	private static final String GXL_VALUE_POS = "pos";
	private static final String GXL_VALUE_TARGET = "target";
	private static final String GXL_VALUE_TRUE = "true";
	private static final String GXL_VALUE_UNDIRECTED = "undirected";
	private static final String GXL_VALUE_VERTEX = "vertex";
	
	private static GXLWriter writer;

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns an instance of the GXL Writer
	 * @return an GXL Writer instance
	 */
	public static GXLWriter getWriter() {
		if(writer==null) {
			writer = new GXLWriter();
		}
		
		return writer;
	}
	
	/**
	 * Writes the graph {@code graph} using the printer {@code out}.
	 * @param graph the graph to be written
	 * @param out the printer to be used
	 */
	private static void writeGraph(Graph graph, PrintWriter out) {
		out.println("<"+GXL_TAG_GRAPH+" " + 
						GXL_ATTRIBUTE_ID + "=\"" + graph.getName() + "\" " + 
						GXL_ATTRIBUTE_EDGEIDS + "=\"" + GXL_VALUE_TRUE + "\" " +
						GXL_ATTRIBUTE_EDGEMODE + "= \"" + GXL_VALUE_UNDIRECTED + "\" " + 
						GXL_ATTRIBUTE_HYPERGRAPH + "=\"" + GXL_VALUE_TRUE + "\">");

		Set<Node> nodes = graph.getNodes();
		Set<Edge> edges = graph.getEdges();

		// write nodes to the file
		for (Node n : nodes){
			out.println("<"+GXL_TAG_NODE+" " + 
							GXL_ATTRIBUTE_ID + "=\"" + n.getName() + "\">");
			out.println("</"+GXL_TAG_NODE+">");
		}

		// write edges to the file as relations
		for (Edge e : edges){
			out.println("<"+GXL_TAG_REL+" " + 
							GXL_ATTRIBUTE_ID + "=\"" + e.getName() +"\">");
			out.println("<"+GXL_TAG_ATTR+" " + 
							GXL_ATTRIBUTE_NAME + "=\"" + GXL_VALUE_LABEL + "\">");
			out.println("<"+GXL_TAG_STRING+">" + e.getLabel() + "</"+GXL_TAG_STRING+">");
			out.println("</"+GXL_TAG_ATTR+">");
			List<Node> incidentNodes = e.getIncidentNodes();
			for (int i=0; i<incidentNodes.size(); i++){
				out.println("<"+GXL_TAG_RELEND+" "+ 
								GXL_ATTRIBUTE_TARGET+"=\"" + e.getIncidentNodes().get(i) + "\" " +
								GXL_ATTRIBUTE_ROLE + "=\"" + GXL_VALUE_VERTEX + "\" " +
								GXL_ATTRIBUTE_STARTORDER + "=\"" + i + "\"/>");					
			}
			out.println("</"+GXL_TAG_REL+">");
		}

		out.println("</"+GXL_TAG_GRAPH+">");
	}

	/**
	 * Writes the inner/outer interface (determined by {@code type}) of the cospan using 
	 * the printer {@code out}.
	 * @param type the type of the interface
	 * @param cospan the cospan whose interface is to be written
	 * @param out the printer to be used
	 */
	private static void writeInterface(InterfaceType type, Cospan cospan, PrintWriter out) {
		String id = null;
		List<Node> list = null;
		
		switch (type) {
		case inner:
			id = "in";
			list = cospan.getInnerInterface();
			break;
			
		case outer:
			id = "out";
			list = cospan.getOuterInterface();
			break;
		}
		
		out.println("<"+GXL_TAG_GRAPH+" " + GXL_ATTRIBUTE_ID + "=\""+ type.toString() + "\">");
		int pos = 0;
		for (Node n : list){
			out.println("<"+GXL_TAG_NODE+" " + GXL_ATTRIBUTE_ID + "=\""+ id + pos + "\">");

			out.println("<"+GXL_TAG_ATTR+" " + 
							GXL_ATTRIBUTE_NAME + "=\"" + GXL_VALUE_POS + "\">");
			out.println("<"+GXL_TAG_INT+">" + pos + "</"+GXL_TAG_INT+">");
			out.println("</"+GXL_TAG_ATTR+">");

			out.println("<"+GXL_TAG_ATTR+" " + 
							GXL_ATTRIBUTE_NAME + "=\"" + GXL_VALUE_TARGET + "\">");
			out.println("<"+GXL_TAG_STRING+">" + n.getName() + "</"+GXL_TAG_STRING+">");
			out.println("</"+GXL_TAG_ATTR+">");

			out.println("</"+GXL_TAG_NODE+">");

			pos++;
		}
		out.println("</"+GXL_TAG_GRAPH+">");
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	private GXLWriter() {
		//Do nothing
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public void write(Graph graph, String filename) 
			throws FileNotFoundException, IOException, UnsupportedEncodingException {		
		if (graph == null){
			throw new NullPointerException(
					RavenExceptionMessages.GRAPH_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if (filename == null) {
			throw new NullPointerException(
					RavenExceptionMessages.FILENAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if (filename.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.FILENAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}

		File file = new File(filename);
		if(!file.exists()) {
			file.getParentFile().mkdirs();
			file.createNewFile();
		}
		else if(!file.isFile()) {
			throw new IllegalArgumentException(RavenExceptionMessages.FILE_NOT_FOUND_EXCEPTION);
		}
			
		try (PrintWriter out = new PrintWriter(filename, StandardCharsets.UTF_8.name())) {
			// print gxl header
			out.println(XML_DECLARATION);
			out.println(DOCUMENT_TYPE_DECLARATION);
			out.println("<"+GXL_TAG_GXL+">");
			writeGraph(graph, out);
			out.println("</"+GXL_TAG_GXL+">");
			out.flush();
		}
	}

	@Override
	public void write(Cospan cospan, String filename) 
			throws FileNotFoundException, IOException, UnsupportedEncodingException {
		if (cospan == null){
			throw new NullPointerException(
					RavenExceptionMessages.COSPAN_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if (filename == null) {
			throw new NullPointerException(
					RavenExceptionMessages.FILENAME_MUST_NOT_BE_NULL_EXCEPTION);
		}
		if (filename.isEmpty()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.FILENAME_MUST_NOT_BE_EMPTY_EXCEPTION);
		}
		
		File file = new File(filename);
		if(!file.exists()) {
			file.getParentFile().mkdirs();
			file.createNewFile();
		}
		else if(!file.isFile()) {
			throw new IllegalArgumentException(RavenExceptionMessages.FILE_NOT_FOUND_EXCEPTION);
		}

		Graph graph = cospan.getGraph();
		try (PrintWriter out = new PrintWriter(filename, StandardCharsets.UTF_8.name())) {
			//Print gxl header
			out.println(XML_DECLARATION);
			out.println(DOCUMENT_TYPE_DECLARATION);
			out.println("<"+GXL_TAG_GXL+">");

			//Write the cospan
			writeInterface(InterfaceType.inner, cospan, out);
			writeGraph(graph, out);
			writeInterface(InterfaceType.outer, cospan, out);
			out.println("</"+GXL_TAG_GXL+">");
			out.flush();
		}
	}

}
