/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.io;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.Scanner;

import de.uni_due.inf.ti.raven.cli.Console;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.main.RavenMain;

/**
 * Class for reading operation files containing a list of operations from which a 
 * cospan decomposition can be obtained.
 * @author Christoph Blume
 * @version $Revision: 1.16 $
 *
 */
public class OperationFileReader implements RavenReader {
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Static Methods																*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Reads an operation file {@code file} and returns the cospan decomposition obtained from 
	 * the operation list.
	 * @param file the file containing the operation list
	 * @return the cospan decomposition obtained from the operation list
	 * @throws FileNotFoundException thrown if the specified file can not be found
	 * @throws IOException thrown if the specified file can not be read
	 * 
	 * @see #readOperations(String)
	 * @see Operation
	 * @see CospanDecomposition
	 */
	public static CospanDecomposition readOperations(File file) 
			throws FileNotFoundException, IOException {
		CospanDecomposition decomp = new CospanDecomposition(); 
		try (BufferedReader reader = new BufferedReader(new FileReader(file))) {
			String line;
			boolean isFirstLine = true;
			while((line = reader.readLine())!=null) {
				int index = line.indexOf(RavenMain.FILE_COMMENT_SIGN);
				if(index>-1) {
					line = line.substring(0, index);
				}
				line = line.trim();
				
				if(!line.isEmpty()) {
					if(isFirstLine && line.matches(Console.INTERFACE_PATTERN)) {
						isFirstLine = false;
						try (Scanner scanner = new Scanner(line)) {
							scanner.findInLine(Console.INTERFACE_PATTERN);
							int size = Integer.valueOf(scanner.match().group(1)); 
							decomp = new CospanDecomposition(size);
						};
					}
					else {
						decomp.add(Operation.parse(line));
					}

				}
			}
		}

		return decomp;
	}
	
	/**
	 * Reads an operation file specified by the filename {@code filename} and returns 
	 * the cospan decomposition obtained from the operation list.
	 * @param filename the file name of the operation file
	 * @return the cospan decomposition obtained from the operation list
	 * @throws FileNotFoundException thrown if the specified file can not be found
	 * @throws IOException thrown if the specified file can not be read
	 * 
	 * @see #readOperations(File)
	 * @see Operation
	 * @see CospanDecomposition
	 */
	public static CospanDecomposition readOperations(String filename) 
			throws FileNotFoundException, IOException {
		return readOperations(new File(filename));
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	private OperationFileReader() {}
	
}
