/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.io;

import java.io.BufferedWriter;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;

import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDFactory.BDDOperator;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDDPairing;
import de.uni_due.inf.ti.raven.encodings.BDDEncoding;

/**
 * Writes an automaton to a plain text file.
 * @author Christoph Blume
 * @version $Revision: 1.9 $
 *
 */
public class PlainTextWriter implements AutomatonWriter {

	private static PlainTextWriter writer;

	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns a {@code PlainTextWriter} to save {@code Data} objects to files.
	 * @return the file writer
	 */
	public static PlainTextWriter getWriter() {
		if(writer==null) {
			writer = new PlainTextWriter();
		}

		return writer;
	}
	

	/**
	 * Creates and returns the text header with general informations about the printed automaton.
	 * @param automaton the automaton which is printed out
	 * @return the text header string
	 */
	private static String getHeader(Automaton automaton) {
		StringBuilder header = new StringBuilder();

		header.append("//\n");
		header.append("// Automaton:\n");
		header.append("//     Type: "+automaton.getType()+"\n");
		header.append("//     Maximum Interface Size: "+automaton.getMaximumInterface()+"\n");
		header.append("//     Number of States: "+automaton.getNumberOfStates()+"\n");
		header.append("//\n");
		header.append("// Alphabet:\n");
		for(Operation operation: automaton.getAlphabet()) {
			header.append("//     "+operation.toString()+"\n");
		}
		header.append("//\n");

		return header.toString();
	}

	/**
	 * Returns a string representing all transitions of the automaton {@code automaton} for the letter 
	 * {@code letter}. 
	 * @param automaton the automaton which is printed out
	 * @param letter the letter whose transitions are printed out
	 * @return the transition string
	 */
	private static String getTransition(Automaton automaton, Operation letter) {
		StringBuilder transitions = new StringBuilder();
		BDDEncoding encoding = automaton.getEncoding();
		RavenBDD currentStateDomainBDD = encoding.getCurrentStateDomain().getDomainBDD();
		RavenBDD successorStateDomainBDD = encoding.getSuccessorStateDomain().getDomainBDD();

		RavenBDD transitionBDD = automaton.getTransitionRelation(letter);
		RavenBDD currentStatesBDD = transitionBDD.exists(successorStateDomainBDD);

		for(RavenBDD currentState: currentStatesBDD.getSatAssignments(currentStateDomainBDD)) {
			transitions.append("s"+automaton.getStateName(currentState)+" -"+letter+"->");

			RavenBDD successorStatesBDD = transitionBDD.applyEx(currentState, BDDOperator.and, 
					currentStateDomainBDD);
			for(RavenBDD successorState: successorStatesBDD.getSatAssignments(successorStateDomainBDD)) {
				RavenBDDPairing pairing = RavenBDDPairing.makePair();
				pairing.set(encoding.getSuccessorStateDomain(), encoding.getCurrentStateDomain());

				transitions.append(" s"+ automaton.getStateName(successorState.replace(pairing)));
			}
			transitions.append("\n");
		}
		transitions.append("\n");

		return transitions.toString();
	}

	/**
	 * Returns a string representing all initial states of the automaton {@code automaton}.
	 * @param automaton the automaton which is printed out
	 * @return the initial states string
	 */
	private static String getInitialStates(Automaton automaton) {
		StringBuilder initialStates = new StringBuilder();
		BDDEncoding encoding = automaton.getEncoding();
		RavenBDD currentStateDomainBDD = encoding.getCurrentStateDomain().getDomainBDD();

		initialStates.append("starting:");
		RavenBDD initialStatesBDD = automaton.getFinalStates();
		for(RavenBDD currentStateBDD: initialStatesBDD.getSatAssignments(currentStateDomainBDD)) {
			initialStates.append(" s"+automaton.getStateName(currentStateBDD));
		}
		initialStates.append("\n");

		return initialStates.toString();
	}

	/**
	 * Returns a string representing all final states of the automaton {@code automaton}.
	 * @param automaton the automaton which is printed out
	 * @return the final states string
	 */
	private static String getFinalStates(Automaton automaton) {
		StringBuilder finalStates = new StringBuilder();
		BDDEncoding encoding = automaton.getEncoding();
		RavenBDD currentStateDomainBDD = encoding.getCurrentStateDomain().getDomainBDD();

		finalStates.append("accept:");
		RavenBDD finalStatesBDD = automaton.getFinalStates();
		for(RavenBDD currentStateBDD: finalStatesBDD.getSatAssignments(currentStateDomainBDD)) {
			finalStates.append(" s"+automaton.getStateName(currentStateBDD));
		}
		finalStates.append("\n");

		return finalStates.toString();
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new writer which writes its output in a simple text format.
	 */
	private PlainTextWriter() {}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	@Override
	public void write(Automaton automaton, String filename) 
			throws FileNotFoundException, IOException {
		try (BufferedWriter writer = new BufferedWriter(new FileWriter(filename))) {
			writer.write(getHeader(automaton));
			writer.write("\n");

			writer.write("// Adding Transitions: \n");
			for(Operation operation: automaton.getAlphabet()) {
				writer.write("//    "+operation+"-Transitions: \n");
				writer.write(getTransition(automaton, operation));
				writer.write("\n");
			}

			writer.write("// Adding Starting States: \n");
			writer.write(getInitialStates(automaton));
			writer.write("\n");

			writer.write("// Adding Accepting States: \n");
			writer.write(getFinalStates(automaton));

			writer.flush();
		}
	}

}
