/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.io;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.logging.Logger;

import com.mxgraph.canvas.mxICanvas;
import com.mxgraph.canvas.mxSvgCanvas;
import com.mxgraph.util.mxCellRenderer;
import com.mxgraph.util.mxDomUtils;
import com.mxgraph.util.mxUtils;
import com.mxgraph.util.mxXmlUtils;
import com.mxgraph.util.mxCellRenderer.CanvasFactory;

import de.uni_due.inf.ti.raven.data.Cospan;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.gui.graphs.VisualGraph;

/**
 * Writes a cospan or graph to a SVG file.
 * @author Christoph Blume
 * @version $Revision: 1.8 $
 *
 */
public class SVGWriter implements GraphWriter, CospanWriter {

	private final Logger logger;

	private CanvasFactory factory;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * 
	 */
	public SVGWriter() {
		factory = new CanvasFactory() {

			@Override
			public mxICanvas createCanvas(int width, int height) {
				mxSvgCanvas canvas = new mxSvgCanvas(
						mxDomUtils.createSvgDocument(width, height));
				canvas.setEmbedded(true);
				return canvas;
			}

		};
		
		logger = Logger.getLogger(getClass().getName());
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public void write(Graph graph, String filename)
			throws FileNotFoundException, IOException {
		try {
			write(VisualGraph.getVisualGraph(graph), filename);
		} catch (InterruptedException e) {
			logger.throwing(getClass().getName(), "writeGraph", e);
		}
	}

	@Override
	public void write(Cospan cospan, String filename)
			throws FileNotFoundException, IOException {
		try {
			write(VisualGraph.getVisualGraph(cospan), filename);
		} catch (Exception e) {
			logger.throwing(getClass().getName(), "writeCospan", e);
		}
	}
	
	/**
	 * Writes a graph to the file specified by {@code filename}.
	 * @param graph the visual graph to be written
	 * @param filename the filename of the file to which the visual graph is written
	 * @throws FileNotFoundException thrown if the file exists but is a directory rather than a 
	 * regular file, does not exist but cannot be created, or cannot be opened for any other reason
	 * @throws IOException thrown if an I/O error occurs while writing the visual graph
	 */
	public void write(VisualGraph graph, String filename) 
			throws FileNotFoundException, IOException {
		mxSvgCanvas canvas = (mxSvgCanvas) mxCellRenderer.drawCells(
				graph, null, 1, null, factory);
		mxUtils.writeFile(mxXmlUtils.getXml(canvas.getDocument()), filename);
	}
	
}
