/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.io;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.data.Operation;
import de.uni_due.inf.ti.raven.data.Signature;
import de.uni_due.inf.ti.raven.messages.Messages;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class provides operations for writing signatures to files.
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.8 $
 *
 */
public class SignatureFilePrinter implements SignaturePrinter {

	private static Logger logger = Logger.getLogger(Messages.class.getName());
	
	private File file;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Static Methods																*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Reads a signature from a file and returns it.
	 * @param filename the name of the file from which the signature is read
	 * @param signature the signature to be written
	 * @throws IOException thrown if an I/O error occurred
	 */
	public static void writeSignature(String filename, Signature signature) 
			throws IOException {
		// creates file if it doesn't already exist
		File file = new File(filename);
		if(!file.exists()) {
			file.getParentFile().mkdirs();
			file.createNewFile();
		}
		else if(!file.isFile()) {
			throw new IllegalArgumentException(RavenExceptionMessages.FILE_NOT_FOUND_EXCEPTION);
		}

		try (BufferedWriter br = new BufferedWriter(new FileWriter(filename))) {
			// write all operations into the file, one operation per line
			for(Operation op : signature.getOperations()){
				br.write(op.parse());
				br.newLine();
			}
		};
	}
	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	private SignatureFilePrinter(String filename) throws IOException {
		// creates file if it doesn't already exist
		file = new File(filename);
		file.createNewFile();
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public void print(Signature signature) {
		try (BufferedWriter writer = new BufferedWriter(new FileWriter(file))) {
			// write all operations into the file, one operation per line
			for(Operation op : signature.getOperations()){
				writer.write(op.parse());
				writer.newLine();
			}
			writer.flush();
		} catch (IOException e) {
			logger.throwing(getClass().getName(), "print", e);
		}
	}
	
}
