/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.io;

import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;

import javax.swing.JTextArea;

import de.uni_due.inf.ti.raven.data.Cospan;
import de.uni_due.inf.ti.raven.data.CospanDecomposition;
import de.uni_due.inf.ti.raven.data.Edge;
import de.uni_due.inf.ti.raven.data.Graph;
import de.uni_due.inf.ti.raven.data.GraphDecomposition;
import de.uni_due.inf.ti.raven.data.Node;
import de.uni_due.inf.ti.raven.data.Signature;
import de.uni_due.inf.ti.raven.data.automata.Automaton;
import de.uni_due.inf.ti.raven.data.bdd.RavenBDD;
import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class offers methods to print data to a specified {@code JTextArea}.
 * @author Christoph Blume
 * $Revision: 1.13 $
 *
 */
public class TextAreaPrinter implements RavenPrinter {

	private static final String LINE_BREAK = System.getProperty("line.separator");

	private static TextAreaPrinter printer;
	
	private final PrintStream out;
	private final JTextArea textarea;
	
	/*------------------------------------------------------------------------------------------*/
	/*		Static Methods																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Returns a printer for printing to the {@code JTextArea} {@code textarea}.
	 * @param textarea the {@code JTextArea} which is used for printing the output
	 * @return a text area printer
	 */
	public static TextAreaPrinter getPrinter(JTextArea textarea) {
		if(printer==null || printer.textarea.equals(textarea)) {
			printer = new TextAreaPrinter(textarea);
		}

		return printer;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Creates a new text area printer which prints to the {@code JTextArea} {@code textarea} and 
	 * to the default system output.
	 * @param textarea the {@code JTextArea} which is used for printing the output
	 */
	private TextAreaPrinter(final JTextArea textarea) {
		if(textarea==null) {
			throw new NullPointerException(
					RavenExceptionMessages.TEXT_AREA_MUST_NOT_BE_NULL_EXCEPTION);
		}
		
		this.textarea = textarea;
		final PrintStream systemOut = System.out;
		out = new PrintStream(new OutputStream() {

			@Override
			public void write(int n) throws IOException {
				textarea.append(Character.toString((char) n));
				systemOut.write(n);
			}

		});

		System.setOut(out);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public void print(String string) {
		out.print(string);
		textarea.setCaretPosition(textarea.getDocument().getLength());
	}
	
	@Override
	public void print(Object object) {
		print(object.toString());
	}

	@Override
	public void println(String string) {
		print(string+System.lineSeparator());		
	}
	
	@Override
	public void println(Object object) {
		println(object.toString());
	}

	@Override
	public void println() {
		print(System.lineSeparator());
	}
	
	@Override
	public void printAutomaton(Automaton automaton) {
		// TODO Auto-generated method stub

	}

	@Override
	public void print(RavenBDD bdd) {
		// TODO Auto-generated method stub

	}

	@Override
	public void print(Cospan cospan) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void print(CospanDecomposition decomposition) {
		// TODO Auto-generated method stub

	}

	@Override
	public void print(GraphDecomposition decomposition) {
		// TODO Auto-generated method stub

	}

	@Override
	public void print(Graph graph) {
		textarea.append("Nodes:"+LINE_BREAK);
		for(Node node: graph.getNodes()) {
			textarea.append("* "+node.getName());
		}
		
		textarea.append(LINE_BREAK);
		textarea.append("Edges:"+LINE_BREAK);
		for(Edge edge: graph.getEdges()) {
			textarea.append("* "+edge.getName()+":"+edge.getLabel()+":");
			
			boolean isFirst = true;
			for(Node node: edge.getIncidentNodes()) {
				textarea.append((isFirst ? "" : "-")+node.getName());
				isFirst = false;
			}
			textarea.append(LINE_BREAK);
		}
	}

	@Override
	public void print(Signature signature) {
		// TODO Auto-generated method stub

	}

}
