/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.io;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import de.uni_due.inf.ti.raven.messages.RavenExceptionMessages;

/**
 * This class provides a simple text table which can be used to structure  
 * contents in a table-like manner.
 * @author Christoph Blume
 * @version $Revision: 1.17 $
 *
 */
public final class TextTable {

	/**
	 * 
	 * @author Christoph Blume
	 * @version $Revision: 1.17 $
	 */
	private enum Alignment {
		left, 
		center, 
		right;
	}

	private static final String VERT = "|";
	private static final String HORZ = "=";
	private static final String CONN = "#";

	/**
	 * 
	 * @author Christoph Blume
	 * @version $Revision: 1.17 $
	 *
	 */
	private class Row {

		private List<Cell> cells;
		private List<Set<Integer>> columns;
		private int height;

		/*------------------------------------------------------------------------------------------*/
		/*		Constructors																		*/
		/*------------------------------------------------------------------------------------------*/
		
		/**
		 * Creates a new row.
		 */
		public Row() {
			cells = new LinkedList<>();
			columns = new ArrayList<>();
			height = 0;
		}
		
		/*------------------------------------------------------------------------------------------*/
		/*		Public Methods																		*/
		/*------------------------------------------------------------------------------------------*/

		/**
		 * Returns the {@code i}-th cell.
		 * @param i the index of the cell
		 * @return the requested cell
		 */
		public Cell getCell(int i) {
			return cells.get(i);
		}

		/**
		 * Returns the number of cells in this row.
		 * @return the number of cells
		 */
		public int getCellCount() {
			return columns.size();
		}

		/**
		 * Adds content {@code content} to the {@code column}-th column of this row.
		 * @param content the content to be added
		 * @param column the index of the column
		 */
		public void addContent(String content, int column) {
			for(int i=getCellCount();i<=column;i++) {
				cells.add(new Cell());
				Set<Integer> set = new HashSet<>();
				set.add(i);
				columns.add(set);
			}

			Cell cell = getCell(column);
			cell.add(content);
			height = Math.max(height, cell.lines());
		}

		/**
		 * Spans a column span from {@code from} to {@code to} (inclusive).
		 * @param from the index of the column in which the span starts
		 * @param to the index of the column in which the span end
		 */
		public void addColumnSpan(int from, int to) {
			if(from<0 || from>=to || to>=getCellCount()) {
				throw new IllegalArgumentException(
						RavenExceptionMessages.ILLEGAL_RANGE_EXCEPTION);
			}

			for(int i=from+1;i<=to;i++) {
				Set<Integer> set = columns.remove(from+1);
				columns.get(from).addAll(set);
				for(int j: set) {
					getCell(j).align = getCell(from).align;
				}
			}
		}

		@Override
		public String toString() {
			StringBuilder builder = new StringBuilder();

			for(int line=0;line<height;line++) {
				builder.append(vert);
				for(int column=0;column<getCellCount();column++) {
					builder.append(" ");

					int width = getCellWidth(column);
					String content = getCellContent(column, line);

					switch (getAlignment(column)) {
					case left:
						builder.append(content);
						builder.append(getSpacing(width-content.length()));
						break;

					case center:
						int leftSpace = width-content.length();
						int rightSpace = leftSpace/2;
						leftSpace = leftSpace-rightSpace;

						builder.append(getSpacing(leftSpace));
						builder.append(content);
						builder.append(getSpacing(rightSpace));
						break;

					case right:
						builder.append(getSpacing(width-content.length()));
						builder.append(content);
						break;

					default:
						break;
					}

					builder.append(vert);
				}
				builder.append("\n");
			}

			return builder.toString();
		}

		/*------------------------------------------------------------------------------------------*/
		/*		Private Methods																		*/
		/*------------------------------------------------------------------------------------------*/
		
		/**
		 * Returns the alignment of the {@code column}-th column.
		 * @param column the index of the column
		 * @return the alignment
		 */
		private Alignment getAlignment(int column) {
			for(int i: columns.get(column)) {
				return getCell(i).align;
			}
			return Alignment.left;
		}

		/**
		 * Returns the content of the cell in the {@code column}-th column and 
		 * {@code line}-th line within the cell.
		 * @param column the index of the column
		 * @param line the index of the line within the cell
		 * @return the content string
		 */
		private String getCellContent(int column, int line) {
			StringBuilder builder = new StringBuilder();

			for(int i: columns.get(column)) {
				Cell cell = getCell(i);

				if(line<cell.lines()) {
					builder.append(cell.getLine(line));
				}
				builder.append(" ");
			}

			return builder.toString();
		}

		/**
		 * Returns the width of the {@code i}-th cell.
		 * @param i the index of the cell
		 * @return the width of the cell
		 */
		private int getCellWidth(int i) {
			int width = (columns.get(i).size()-1)*3 + 1;

			for(int j: columns.get(i)) {
				width += columnWidth.get(j);
			}

			return width;
		}

		/**
		 * Returns the minimum width overall lines of the {@code i}-th cell.
		 * @param i the index of the cell
		 * @return the minimum width of the cell
		 */
		private int getMinimumCellWidth(int i) {
			int width = (columns.get(i).size()-1);

			for(int j: columns.get(i)) {
				width += cells.get(j).width;
			}

			return width;
		}

		/**
		 * Returns a string of length {@code width} containing
		 * whitespace signs only.
		 * @param width the length of the spacing
		 * @return the whitespace string
		 */
		private String getSpacing(int width) {
			char[] temp = new char[width];
			Arrays.fill(temp, ' ');
			return String.valueOf(temp);
		}

	}

	/**
	 * 
	 * @author Christoph Blume
	 * @version $Revision: 1.17 $
	 *
	 */
	private class Cell {

		private List<String> content;
		private Alignment align;
		private int width;

		/*------------------------------------------------------------------------------------------*/
		/*		Constructors																		*/
		/*------------------------------------------------------------------------------------------*/
		
		/**
		 * Creates a new cell.
		 */
		public Cell() {
			content = new LinkedList<>();
			align = Alignment.left;
		}
		
		/*------------------------------------------------------------------------------------------*/
		/*		Public Methods																		*/
		/*------------------------------------------------------------------------------------------*/

		/**
		 * Adds a new line containing the string {@code content} to this cell.
		 * @param content the content of the new line
		 */
		public void add(String content) {
			if(this.content.size()==0 || 
					!getLine(this.content.size()-1).isEmpty()) {
				this.content.add(content);
			}
			else {
				this.content.set(this.content.size()-1, content);
			}

			width = Math.max(width, content.length());
		}

		/**
		 * Returns the number of lines of this cell.
		 * @return the number of lines
		 */
		public int lines() {
			return content.size();
		}

		/**
		 * Returns the {@code i}-th line of this cell.
		 * @param i the index of the line
		 * @return the request line
		 * 
		 * @throws IllegalArgumentException thrown if <em>i&lt;0</em> or 
		 * <em>i&gt;=#lines</em>
		 */
		public String getLine(int i) {
			if(i<0 || i>=content.size()) {
				throw new IllegalArgumentException(
						RavenExceptionMessages.INDEX_OUT_OF_BOUNDS_EXCEPTION);
			}

			return content.get(i);
		}

	}

	private Row header;
	private List<Row> rows;
	private int columns;
	private List<Integer> columnWidth;
	
	private String vert;
	private String horz;
	private String conn;
	
	/**
	 * Creates a new text-based table with default table delimiters.
	 * 
	 * @see #TextTable(char, char, char)
	 */
	public TextTable() {
		rows = new ArrayList<>();
		columnWidth = new ArrayList<>();
		header = new Row();
		vert = VERT;
		horz = HORZ;
		conn = CONN;
	}

	/**
	 * Creates a new text-based table using the given delimiters.
	 * @param vert the vertical delimiter
	 * @param horz the horizontal delimiter
	 * @param conn the connecting delimiter (used at connection points)
	 */
	public TextTable(char vert, char horz, char conn) {
		this();
		this.vert = String.valueOf(vert);
		this.horz = String.valueOf(horz);
		this.conn = String.valueOf(conn);
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Adds a new header with content {@code object} at the end of the header row.
	 * @param object the content object
	 */
	public void addHeader(Object object) {
		addHeader(header.getCellCount(), object);
	}

	/**
	 * Adds a new header at position {@code column} with content {@code object}.
	 * @param column the index of the column to which the header will be added
	 * @param object the content object
	 */
	public void addHeader(int column, Object object) {
		String content = object.toString();
		header.addContent(content, column);
		columns = Math.max(columns, column+1);

		for(int i=columnWidth.size();i<=column;i++) {
			columnWidth.add(0);
		}
		columnWidth.set(column, Math.max(columnWidth.get(column), 
				content.length()));
	}

	/**
	 * Adds to the {@code row}-th row a new column with content {@code objects}.
	 * If the table has less than {@code row+1} rows, additional rows will be added.
	 * @param row the index of the row
	 * @param objects the content objects
	 * 
	 * <emph>Note:</emph> Each content object will be placed in a single line.
	 */
	public void addContent(int row, Object... objects) {
		for(int i=rows.size();i<=row;i++) {
			rows.add(new Row());
		}

		int column = rows.get(row).getCellCount();
		for(Object object: objects) {
			addContent(row, column, object);
		}
	}

	/**
	 * Adds to the {@code row}-th row and the {@code column}-th column the content 
	 * {@code object}. If the table has less than {@code row+1} rows and/or less than 
	 * {@code column+1} columns, additional rows and/or columns will be added.
	 * @param row the index of the row
	 * @param column the index of the column
	 * @param object the content object
	 */
	public void addContent(int row, int column, Object object) {
		for(int i=rows.size();i<=row;i++) {
			rows.add(new Row());
		}

		String content = object.toString();
		rows.get(row).addContent(content, column);
		columns = Math.max(columns, column+1);

		for(int i=columnWidth.size();i<=column;i++) {
			columnWidth.add(0);
		}
		columnWidth.set(column, Math.max(columnWidth.get(column), 
				content.length()));
	}

	/**
	 * Spans a column span in the header row ranging from {@code from} to 
	 * {@code to} (inclusive).
	 * @param from the index of the column in which the span starts
	 * @param to the index of the column in which the span end
	 */
	public void setHeaderSpan(int from, int to) {
		if(from<0 || from>=to) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_RANGE_EXCEPTION);
		}

		for(int i=header.getCellCount();i<=to;i++) {
			addHeader("");
		}

		header.addColumnSpan(from, to);
		for(int i=from;i<=to;i++) {
			columnWidth.set(i, 0);

			for(int j=0;j<rows.size();j++) {
				if(i<rows.get(j).getCellCount()) {
					//TODO Spaltenbreite ist inkorrekt, falls der Inhalt des Columnspans laenger ist als die Summe der Breite beteiligten Spalten
					columnWidth.set(i, Math.max(columnWidth.get(i), 
							rows.get(j).getMinimumCellWidth(i)));
				}
			}
		}
	}

	/**
	 * Spans a column span in the row {@code row} ranging from {@code from} to 
	 * {@code to} (inclusive).
	 * @param row the index of the row
	 * @param from the index of the column in which the span starts
	 * @param to the index of the column in which the span end
	 */
	public void setColumnSpan(int row, int from, int to) {
		if(row<0 || row>=rows.size()) {
			throw new IllegalArgumentException(
					RavenExceptionMessages.INDEX_OUT_OF_BOUNDS_EXCEPTION);
		}
		if(from<0 || from>=to || to>=rows.get(row).getCellCount()) {
			System.out.printf("from=%d, to=%d, #rows=%d%n", from, to, rows.get(row).getCellCount());
			throw new IllegalArgumentException(
					RavenExceptionMessages.ILLEGAL_RANGE_EXCEPTION);
		}

		rows.get(row).addColumnSpan(from, to);
	}
	
	@Override
	public String toString() {
		StringBuilder table = new StringBuilder();
		//Print header and cells
		if(header.columns.size()>0) {
			table.append(conn);
			for(int i=0;i<header.columns.size();i++) {
				char[] temp = new char[header.getCellWidth(i)+1];
				Arrays.fill(temp, horz.charAt(0));
				table.append(temp);
				table.append(conn);
			}
			table.append("\n");
			table.append(header.toString());
			table.append(getHorizontalBorder(header, false));
		}
		else {
			table.append(getHorizontalBorder(rows.get(0), true));
		}
		for(Row row: rows) {
			table.append(row.toString());
			table.append(getHorizontalBorder(row, false));
		}
		
		return table.toString();
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Private Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Returns the horizontal text border between for the {@code row}-th and the {@code row+1}-th 
	 * rows. If {@code isTop} is {@code true}, the {@code row+1}-th row will not be considered. This 
	 * flag can be used to compute the border of the top-most row.
	 * @param row the index of the row
	 * @param isTop if {@code true}
	 * @return the horizontal text border
	 */
	private String getHorizontalBorder(Row row, boolean isTop) {
		int index = rows.indexOf(row);
		Row	nextRow = null;
		if(index<rows.size()-1 && !isTop) {
			nextRow = rows.get(index+1);
		}

		StringBuilder builder = new StringBuilder(conn);
		for(int i=0;i<row.columns.size();i++) {
			char[] temp = new char[row.getCellWidth(i)+1];
			Arrays.fill(temp, horz.charAt(0));
			builder.append(temp);
			builder.append(conn);
		}

		if(nextRow!=null) {
			int pos = 0;
			for(int i=0;i<nextRow.columns.size();i++) {
				pos += nextRow.getCellWidth(i) + 2;
				if(pos<builder.length()) {
					builder.replace(pos, pos+1, conn);
				}
				else {
					char[] temp = new char[pos-builder.length()];
					Arrays.fill(temp, horz.charAt(0));
					builder.append(temp);
					builder.append(conn);
				}
			}
		}
		builder.append("\n");

		return builder.toString();
	}

}