/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.io;

import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.io.IOException;
import java.io.InputStream;
import java.util.logging.Logger;

import javax.swing.JTextField;

/**
 * This class provides an {@code InputStream} from which the contents of the 
 * specified {@code JTextField} can be read. 
 * @author Christoph Blume, Laura Steinert
 * @version $Revision: 1.6 $
 *
 */
public class TextfieldInputStream extends InputStream implements KeyListener {

	private JTextField textfield;
	private byte[] contents;
	private int pointer;
	private boolean LF_Returned = false;
	private boolean CR_Returned = false;
	
	private Logger logger;

	/*------------------------------------------------------------------------------------------*/
	/*		Constructors																		*/
	/*------------------------------------------------------------------------------------------*/
	
	/**
	 * Creates a new input stream which reads from the specified JTextField {@code textfield}.
	 * @param textfield the textfield from which the content is read
	 */
	public TextfieldInputStream(JTextField textfield) {
		this.textfield = textfield;
		textfield.addKeyListener(this);
		contents = null;
		
		logger = Logger.getLogger(getClass().getName());
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Public Methods																		*/
	/*------------------------------------------------------------------------------------------*/
	
	@Override
	public int read() throws IOException {
		synchronized (this) {
			//If no content is currently available, wait until input is available
			if(contents==null) {
				try {
					wait();
					LF_Returned = false;
					CR_Returned = false;
				} catch (InterruptedException e) {
					logger.throwing(getClass().getName(), "read", e);
				}
			}
			
			//If the end of the input is reached mark the end with a "new line"
			//and end the stream
			if(pointer>=contents.length) {
				if(!LF_Returned) {
					LF_Returned = true;
					return '\n';
				}
				else if(!CR_Returned) {
					CR_Returned = true;
					return '\r';
				}
				else {
					contents = null;
					return -1;
				}
			}
			
			//If there is still some content return it
			return contents[pointer++];
		}
	}

	@Override
	public void keyPressed(KeyEvent e) {
		//Update the content only if the ENTER key was pressed
		if(e.getKeyCode()==KeyEvent.VK_ENTER) {
			synchronized (this) {
				contents = textfield.getText().getBytes();
				textfield.setText("");
				pointer = 0;
				
				notify();
			}
		}
	}

	@Override
	public void keyReleased(KeyEvent e) {
		//Do nothing
	}

	@Override
	public void keyTyped(KeyEvent e) {
		//Do nothing
	}

}
