/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.main;

import java.awt.SplashScreen;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.logging.LogManager;
import java.util.logging.Logger;

import de.uni_due.inf.ti.raven.cli.Console;
import de.uni_due.inf.ti.raven.cli.ConsoleMessages;
import de.uni_due.inf.ti.raven.gui.MainWindow;
import de.uni_due.inf.ti.raven.gui.SplashScreenWithProgressBar;

/**
 * This class is the starting point for Raven. It implements the main()-method and 
 * initializes the main services.
 * @author Christoph Blume
 * @version $Revision: 1.85 $
 *
 */
public class RavenMain {

	/**
	 * Program modes in which Raven can be started. Used to determine different program
	 * behavior.
	 * @author Christoph Blume
	 * @version $Revision: 1.85 $
	 *
	 */
	private enum Mode {
		/** Console Mode */
		CONSOLE_MODE,
		
		 /** GUI Mode */
		GUI_MODE
	}
	
	private static final String RAVEN_NAME = "Raven Tool Suite";
	private static final String RAVEN_VENDOR = "Theoretical Computer Science,\n" +
			"University of Duisburg-Essen, Germany.";
	private static final String RAVEN_PROJECT_URL = "http://www.ti.inf.uni-due.de/research/tools/raven";

	private static final String PROPERTIES_FILENAME = "raven.properties";

	// Exit Codes

	/** Indicates normal termination */
	public static final int EXIT_CODE_NORMAL_TERMINATION = 0;

	/** Indicates the occurrence of an unexpected error (yes, all other errors were expected. ;) */
	public static final int EXIT_CODE_UNEXPECTED_ERROR = 1;

	/** Indicates that the directory where saved files will be stored could not be created */
	public static final int EXIT_CODE_SAVE_DIR_CREATION_FAILED = 2;

	/** Indicates that the directory where to place scripts could not be created */
	public static final int EXIT_CODE_SCRIPT_DIR_CREATION_FAILED = 3;

	/** Indicates that the temporary file directory could not be created */
	public static final int EXIT_CODE_TEMP_DIR_CREATION_FAILED = 4;

	/** Indicates that a security exception was thrown which led to a severe error */
	public static final int EXIT_CODE_SECURITY_EXCEPTION = 5;

	/** Indicates that the properties file could not be found which led to a severe error  */
	public static final int EXIT_CODE_PROPERTIES_FILE_NOT_FOUND = 6;

	/** Indicates that a severe I/O exception occurred during loading the properties file  */
	public static final int EXIT_CODE_PROPERTIES_IO_EXCEPTION = 7;

	/** Indicates that Raven has been run out of memory */
	public static final int EXIT_CODE_OUT_OF_MEMORY = 8;


	// Directory Constants

	/** The Raven Working Directory directory */
	public static final String MAIN_DIR = System.getProperty("user.dir") + File.separator;
	
	/** The resource file directory */
	public static final String RES_DIR = MAIN_DIR + "resources" + File.separator;

	/** The save file directory */
	public static final String SAVE_DIR = MAIN_DIR + "savedata" + File.separator;

	/** The automaton save file directory */
	public static final String AUTOMATON_SAVE_DIR = SAVE_DIR + "automata" + File.separator;

	/** The automaton save file directory */
	public static final String COSPAN_SAVE_DIR = SAVE_DIR + "cospans" + File.separator;

	/** The automaton save file directory */
	public static final String DECOMPOSITION_SAVE_DIR = SAVE_DIR + "decompositions" + File.separator;
	
	/** The command documentation files folder */
	public static final String DOCUMENTATION_FOLDER = RES_DIR + File.separator + "cli" + 
												File.separator + "documentation" + File.separator;
	
	/** The DOT save file directory */
	public static final String DOT_SAVE_DIR = SAVE_DIR + "dot" + File.separator;
	
	/** The graph save file directory */
	public static final String GRAPH_SAVE_DIR = SAVE_DIR + "graphs" + File.separator;
	
	/** The script file directory */
	public static final String SCRIPT_DIR = MAIN_DIR + "script" + File.separator;

	/** The signature save file directory */
	public static final String SIGNATURE_SAVE_DIR = SAVE_DIR + "signatures" + File.separator;
	
	/** The temporary file directory */
	public static final String TEMP_DIR = MAIN_DIR + "tmp" + File.separator;

	/** The resource file directory */
	public static final String ICON_DIR = RES_DIR + "icons" + File.separator;
	

	// External Program Constants

	/** The JAVA APPLICATION LAUNCHER executable */
	public static final String JAVA_LAUNCHER_COMMAND = System.getProperty("java.home") + 
			File.separator + "bin" + File.separator + "java";
	
	/** The DOT LAYOUTER FOR BDDS */
	public static final String DOT_BDD_LAYOUTER = "dot";
	
	/** The DOT LAYOUTER FOR GRAPHS */
	public static final String DOT_GRAPH_LAYOUTER = "neato";
	
	/** The DOT LAYOUTER FOR AUTOMATON RUNS */
	public static final String DOT_RUN_LAYOUTER = "dot";

	/** The GXL DTD URI */
	public static final String GXL_DTD_URI = "http://www.gupro.de/GXL/gxl-1.0.dtd";
	
	/** */
	public static final String RAVEN_MANUAL_FILE_NAME = RES_DIR + "documentation.pdf";


	// File Extension Constants

	/** The file extension for automaton files */
	public static final String AUTOMATON_FILE_EXTENSION = "aut";

	/** The file extension for BDD files */
	public static final String BDD_FILE_EXTENSION = "bdd";
	
	/** The file extension for cospan files */
	public static final String COSPAN_GXL_FILE_EXTENSION = "cos";

	/** The file extension for decomposition files */
	public static final String DECOMPOSITION_FILE_EXTENSION = "dec";

	/** The file extension for DOT files */
	public static final String DOT_FILE_EXTENSION = "dot";
	
	/** The file extension for graph files */
	public static final String GRAPH_GXL_FILE_EXTENSION = "gxl";
	
	/** The file extension for JAVA ARCHIVE files */
	public static final String JAR_FILE_EXTENSION = "jar";

	/** The file extension for raven script files */
	public static final String RAVEN_SCRIPT_FILE_EXTENSION = "txt";
	
	/** The file extension for serialized object files */
	public static final String SERIALIZATION_FILE_EXTENSION = "obj";
	
	/** The file extension for Signature files */
	public static final String SIGNATURE_FILE_EXTENSION = "sig";
	
	/** The file extension for graph files */
	public static final String SVG_FILE_EXTENSION = "svg";

	
	// Raven Image Files
	
	/** The Dialog Error Icon Filename */
	public static final String DIALOG_ERROR_ICON_FILENAME = ICON_DIR + "dialog-error.png";
	
	/** The Dialog Information Icon Filename */
	public static final String DIALOG_INFORMATION_ICON_FILENAME = ICON_DIR + "dialog-information.png";
	
	/** The Dialog Question Icon Filename */
	public static final String DIALOG_QUESTION_ICON_FILENAME = ICON_DIR + "dialog-question.png";
	
	/** The Dialog Warning Icon Filename */
	public static final String DIALOG_WARNING_ICON_FILENAME = ICON_DIR + "dialog-warning.png";
	
	/** The Edit Icon Filename */
	public static final String EDIT_ICON_FILENAME = ICON_DIR + "list-edit-alternative.png";
	
	/** The Help Icon Filename */
	public static final String HELP_ICON_FILENAME = ICON_DIR + "icon-help-small.png";
	
	/** The Maximize View Icon Filename */
	public static final String MAXIMIZE_VIEW_ICON_FILENAME = ICON_DIR + "view-fullscreen.png";
	
	/** The Minus Icon Filename */
	public static final String MINUS_ICON_FILENAME = ICON_DIR + "list-remove-alternative.png";
	
	/** The Open Icon Filename */
	public static final String OPEN_ICON_FILENAME = ICON_DIR + "document-open.png";
	
	/** The Plus Icon Filename */
	public static final String PLUS_ICON_FILENAME = ICON_DIR + "list-add-alternative.png";
	
	/** The Raven Icon File */
	public static final String RAVEN_ICON_FILENAME = ICON_DIR + "raven_icon.png";
	
	/** The Big Raven Icon File */
	public static final String BIG_RAVEN_ICON_FILENAME = ICON_DIR + "raven_icon_big.png";
	
	/** The Raven Logo Filename */
	public static final String RAVEN_LOGO_FILENAME = ICON_DIR + "raven_logo.png";
	
	/** The Run Icon Filename */
	public static final String RUN_ICON_FILENAME = ICON_DIR + "media-playback-start-alternative.png";
	
	/** The Save Icon Filename */
	public static final String SAVE_ICON_FILENAME = ICON_DIR + "document-save.png";
	
	/** The Stop Icon Filename */
	public static final String STOP_ICON_FILENAME = ICON_DIR + "media-playback-stop-alternative.png";

	
	// Raven Update Information

	/** Raven Update Description File */
	public static final String RAVEN_UPDATE_FILE = "raven-update.xml";

	/** Raven Update Mirrors */
	public static final String[] RAVEN_UPDATE_URLS = new String[] {
		"http://www.ti.inf.uni-due.de/uploads/media/"
	};
	
	/** Raven Update XML Schema validation file */
	public static final String RAVEN_UPDATE_SCHEMA_FILENAME = RES_DIR + "raven-update.xsd";
	
	
	// Default Encodings

	/** The default dominating set encoding */
	public static final String DEFAULT_DOMINATING_SET_ENCODING = "DominatingSet.encoding";

	/** The default vertex cover encoding */
	public static final String DEFAULT_VERTEX_COVER_ENCODING = "VertexCover.encoding";

	/** The default colorability encoding */
	public static final String DEFAULT_COLORABILITY_ENCODING = "Colorability.encoding";

	
	/** */
	public static final String FILE_COMMENT_SIGN = "#";
	
	
	// Property Constants
	
	private static final String RAVEN_PROPERTY_PREFIX = "_";
	
	/** */
	public static final String PROPERTY_DEFAULT_LEVEL = ".level";
	
	/** */
	public static final String PROPERTY_DOT_PATH = "dot";
	
	/** */
	public static final String PROPERTY_DOWNLOAD_BETA_VERSION = RAVEN_PROPERTY_PREFIX + "beta";
	
	/** */
	public static final String PROPERTY_FILE_HANDLER_FORMATTER = "java.util.logging.FileHandler.formatter";
	
	/** */
	public static final String PROPERTY_FILE_HANDLER_LIMIT = "java.util.logging.FileHandler.limit";
	
	/** */
	public static final String PROPERTY_FILE_HANDLER_PATTERN = "java.util.logging.FileHandler.pattern";
	
	/** */
	public static final String PROPERTY_FILE_HANDLER_LEVEL = "java.util.logging.FileHandler.level";
	
	/** */
	public static final String PROPERTY_HANDLERS = "handlers";
	
	/** */
	public static final String PROPERTY_RAVEN_LOGGING_HANDLER_LEVEL = "de.uni_due.inf.ti.raven.main.RavenLoggingHandler.level";
	
	/** The sloppy mode property */
	public static final String PROPERTY_SLOPPY_MODE = RAVEN_PROPERTY_PREFIX + "sloppy";
	
	/** The time mode property */
	public static final String PROPERTY_TIME = RAVEN_PROPERTY_PREFIX + "time";
	
	/** The verbose mode property */
	public static final String PROPERTY_VERBOSE = RAVEN_PROPERTY_PREFIX + "verbose";
	
	private static final String HANDLER_LEVEL_INFO = "INFO";
	private static final String HANDLER_LEVEL_CONFIG = "CONFIG";
	private static final String HANDLER_LEVEL_FINE = "FINE";
	private static final String FILE_HANDLER = "java.util.logging.FileHandler";
	private static final String FILE_HANDLER_DEFAULT_FORMATTER = "java.util.logging.SimpleFormatter";
	private static final String FILE_HANDLER_DEFAULT_LIMIT = "10000000";
	private static final String FILE_HANDLER_DEFAULT_PATTERN = "./raven_log%u.log";
	private static final String RAVEN_LOGGING_HANDLER = "de.uni_due.inf.ti.raven.main.RavenLoggingHandler";

	
	//Raven Options
	
	private static final String OPTION_TOKEN = "-";
	private static final String IS_TOKEN = "=";
	private static final String TRUE_TOKEN = "true";
	private static final String CONSOLE_TOKEN = "console";
	private static final String CONSOLE_SHORTCUT_TOKEN = "c";
	private static final String CONSOLE_OPTION = CONSOLE_TOKEN+"|"+CONSOLE_SHORTCUT_TOKEN;
	private static final List<String> OPTION_LIST = new LinkedList<>();

	
	private static final Logger logger = Logger.getLogger(RavenMain.class.getName());

	private static Mode mode;
	private static Properties properties;

	/*------------------------------------------------------------------------------------------*/
	/*		Public Static Methods																*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Starts Raven.
	 * @param args the arguments passed to Raven
	 */
	public static void main(String[] args) {
		//Create directories (if not existing)
		checkAndCreateDirectory(SAVE_DIR, EXIT_CODE_SAVE_DIR_CREATION_FAILED);
		checkAndCreateDirectory(AUTOMATON_SAVE_DIR, EXIT_CODE_SAVE_DIR_CREATION_FAILED);
		checkAndCreateDirectory(COSPAN_SAVE_DIR, EXIT_CODE_SAVE_DIR_CREATION_FAILED);
		checkAndCreateDirectory(DECOMPOSITION_SAVE_DIR, EXIT_CODE_SAVE_DIR_CREATION_FAILED);
		checkAndCreateDirectory(GRAPH_SAVE_DIR, EXIT_CODE_SAVE_DIR_CREATION_FAILED);
		checkAndCreateDirectory(SIGNATURE_SAVE_DIR, EXIT_CODE_SAVE_DIR_CREATION_FAILED);
		checkAndCreateDirectory(DOT_SAVE_DIR, EXIT_CODE_SAVE_DIR_CREATION_FAILED);
		checkAndCreateDirectory(SCRIPT_DIR, EXIT_CODE_SCRIPT_DIR_CREATION_FAILED);
		checkAndCreateDirectory(TEMP_DIR, EXIT_CODE_TEMP_DIR_CREATION_FAILED);

		for(File file: new File(TEMP_DIR).listFiles()) {
			file.delete();
		}
		
		OPTION_LIST.add(CONSOLE_OPTION);

		properties = new Properties();
		try {
			properties.load(new FileReader(PROPERTIES_FILENAME));
		} catch (FileNotFoundException e) {
			properties.setProperty(PROPERTY_DEFAULT_LEVEL, HANDLER_LEVEL_FINE);
			properties.setProperty(PROPERTY_DOWNLOAD_BETA_VERSION, String.valueOf(false));
			properties.setProperty(PROPERTY_FILE_HANDLER_LEVEL, HANDLER_LEVEL_CONFIG);
			properties.setProperty(PROPERTY_FILE_HANDLER_PATTERN, FILE_HANDLER_DEFAULT_PATTERN);
			properties.setProperty(PROPERTY_FILE_HANDLER_LIMIT, FILE_HANDLER_DEFAULT_LIMIT);
			properties.setProperty(PROPERTY_FILE_HANDLER_FORMATTER, FILE_HANDLER_DEFAULT_FORMATTER);
			properties.setProperty(PROPERTY_HANDLERS, RAVEN_LOGGING_HANDLER + ", " + FILE_HANDLER);
			properties.setProperty(PROPERTY_RAVEN_LOGGING_HANDLER_LEVEL, HANDLER_LEVEL_INFO);
			properties.setProperty(PROPERTY_SLOPPY_MODE, String.valueOf(true));
			properties.setProperty(PROPERTY_TIME, String.valueOf(false));
			properties.setProperty(PROPERTY_VERBOSE, String.valueOf(false));
			
			saveProperties();
			
		} catch (IOException e) {
			logger.throwing(RavenMain.class.getName(), "main", e);
			exit(EXIT_CODE_PROPERTIES_IO_EXCEPTION);
		}
		try {
			LogManager.getLogManager().readConfiguration(
					new FileInputStream(PROPERTIES_FILENAME));
		} catch (SecurityException e) {
			logger.throwing(RavenMain.class.getName(), "main", e);
			exit(EXIT_CODE_SECURITY_EXCEPTION);
		} catch (IOException e) {
			logger.throwing(RavenMain.class.getName(), "main", e);
			exit(EXIT_CODE_PROPERTIES_IO_EXCEPTION);
		}

		Map<String, String> options = parseOptions(args);
		List<String> parameter = parseParameter(
				Arrays.copyOfRange(args, options.size(), args.length));
		
		switch (parameter.size()) {
		case 0:
			if(options.containsKey(CONSOLE_OPTION)) {
				mode = Mode.CONSOLE_MODE;
				Console console = Console.getConsole();
				new Thread(console).start();
			}
			else {
				mode = Mode.GUI_MODE;
				SplashScreen splash = SplashScreen.getSplashScreen();
		        if (splash != null) {
					SplashScreenWithProgressBar splashScreenWithProgressBar 
						= new SplashScreenWithProgressBar(splash);
					splashScreenWithProgressBar.start();
		        }
				
				new Thread(new Runnable() {

					@Override
					public void run() {
						MainWindow.getWindow();
					}
				}).start();
			}
			break;

		case 1:
			if(options.containsKey(CONSOLE_OPTION)) {
				try {
					mode = Mode.CONSOLE_MODE;
					Console.runBatchFile(SCRIPT_DIR+File.separator+parameter.get(parameter.size()-1));
				} catch (FileNotFoundException e) {
					System.err.println("Invalid filename");
				}
			}
//			else {
//				//TODO Implement possibility to load script files in GUI mode 
//			}
			break;

		default:
			System.err.println("Wrong parameter");
			break;
		}

	}

		/**
	 * Returns informations about Raven.
	 * @return an information string
	 */
	public static String getInfo() {
		return RAVEN_NAME + ", Ver. " + Version.getVersion() + "\n" +
				RAVEN_VENDOR + "\n" + "Homepage: " + RAVEN_PROJECT_URL + "\n";
	}
	
	/**
	 * Returns the Raven logo as ASCII art.
	 * @return the raven logo
	 */
	public static String getASCIIArt(){
		return
				"                                            :#W#" + "\n" + 
				"                                          :#WWWWW@##" + "\n" + 
				"                                         :#WWWWWWWW@##*-" + "\n" + 
				"                                        :WWWWWWWWWWWWWWW@-" + "\n" + 
				"                                       +WWWWWWWW=" + "\n" + 
				"                                    -@WWWWWWWW#" + "\n" + 
				"                                 -*@WWWWWWWWW#" + "\n" + 
				"                             -=@WWWWWWWWWWWWW*" + "\n" + 
				"                         :#WWWWWWWWWWWWWWWWWW+" + "\n" + 
				"                      :#WWWWWWWWWWWWWWWWWWWW@-" + "\n" + 
				"                    -=WWWWWWWWWWWWWWWWWWWWW@-" + "\n" + 
				"                 -@WWWWWWWWWWWWWWWWWWWWWWW@-" + "\n" + 
				"               -#WWWWWWWWWWWWWWWWWWWWWWWW@-" + "\n" + 
				"             *WWWWWWWWWWWWWWWWWWWWWWWWWW*" + "\n" + 
				"           -:=@WWWWWWWWWWWWWWWWWWWWWWW#-" + "\n" + 
				"       :#WWWWWWWWWWWWWWWWWWWWWWWWWWWW=-" + "\n" + 
				"    -#W@@WWW@#*:-         -:*=@WW#WW@" + "\n" + 
				" :#+=WWWWW@+-               *WW@+WW-" + "\n" + 
				"+@WWW@=-                     -WW*.#@" + "\n" + 
				"                               *WW  #@#" + "\n" + 
				"                                 #=## WW#+-" + "\n";
	}

	/**
	 * Returns a set of names of properties currently set for Raven.
	 * @return the property names
	 */
	public static Set<String> getPropertyNames() {
		return properties.stringPropertyNames();
	}
	
	/**
	 * Returns the value of a property.
	 * @param name the name of the property
	 * @return the value of the property
	 */
	public static String getProperty(String name) {
		return properties.getProperty(name);
	}
	
	/**
	 * Sets the property {@code name} to the value {@code value}. If the property 
	 * is not existing yet, it is added.
	 * @param name the name of the property
	 * @param value the value of the property
	 */
	public static void setProperty(String name, String value) {
		properties.put(name, value);
	}
	
	/**
	 * Removes the property {@code name}.
	 * @param name the name of the property
	 * <em>NOTE: Only user-defined properties can be removed.</em>
	 */
	public static void removeProperty(String name) {
		
		if(!isSystemProperty(name)) {
			properties.remove(name);			
		}
	}
	
	/**
	 * Checks if the property {@code name} is a system property.
	 * @param name the name of the property
	 * <em>NOTE: Only user-defined properties can be removed.</em>
	 * @return {@code true} if the property is a system property, {@code false}
	 * otherwise
	 */
	public static boolean isSystemProperty(String name) {
		return  name.equals(PROPERTY_DEFAULT_LEVEL) ||
				name.equals(PROPERTY_DOT_PATH) ||
				name.equals(PROPERTY_DOWNLOAD_BETA_VERSION) ||
				name.equals(PROPERTY_FILE_HANDLER_FORMATTER) ||
				name.equals(PROPERTY_FILE_HANDLER_LEVEL) ||
				name.equals(PROPERTY_FILE_HANDLER_LIMIT) ||
				name.equals(PROPERTY_FILE_HANDLER_PATTERN) ||
				name.equals(PROPERTY_HANDLERS) ||
				name.equals(PROPERTY_RAVEN_LOGGING_HANDLER_LEVEL) ||
				name.equals(PROPERTY_SLOPPY_MODE) ||
				name.equals(PROPERTY_TIME) ||
				name.equals(PROPERTY_VERBOSE);
	}
	
	/**
	 * Save the properties in a file.
	 */
	public static void saveProperties() {
		try {
			properties.store(new FileOutputStream(PROPERTIES_FILENAME), null);
		} catch (IOException ex) {
			logger.throwing(RavenMain.class.getName(), "main", ex);
			exit(EXIT_CODE_PROPERTIES_IO_EXCEPTION);
		}
	}
	
	/**
	 * Exits Raven and returns it's status.
	 * @param status a value indicating the exit status
	 */
	public static void exit(int status) {
		_exit(status);
	}

	/**
	 * Restarts Raven.
	 */
	public static void restart() {
		try {
			final File currentJar = new File(
					RavenMain.class.getProtectionDomain().getCodeSource().getLocation().toURI());
			if(!currentJar.getName().endsWith("."+JAR_FILE_EXTENSION)) {
				System.err.println("The Raven JAR file could not be found. Please restart Raven manually.");
				return;
			}

			/* Build command: java -jar application.jar */
			final ArrayList<String> command = new ArrayList<>();
			command.add(JAVA_LAUNCHER_COMMAND);
			command.add("-jar");
			command.add(currentJar.getPath());
			if(mode==Mode.CONSOLE_MODE) {
				command.add(OPTION_TOKEN+CONSOLE_SHORTCUT_TOKEN);
			}

			final ProcessBuilder builder = new ProcessBuilder(command);
			builder.start();
		} catch (URISyntaxException e) {
			logger.throwing(RavenMain.class.getName(), "restart", e);
			System.err.println("The Raven JAR file could not be found. Please restart Raven manually.");
			e.printStackTrace();
		} catch (IOException e) {
			logger.throwing(RavenMain.class.getName(), "restart", e);
			System.err.println("An error occurred. Please restart Raven manually.");
			e.printStackTrace();
		}
		finally {
			System.exit(0);
		}
	}
	
	/*------------------------------------------------------------------------------------------*/
	/*		Private Static Methods																*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Parses the options from the input arguments.
	 * @param args the arguments array
	 * @return a map of options
	 */
	private static Map<String, String> parseOptions(String[] args) {
		boolean isValid;
		
		Map<String, String> options = new HashMap<>();
		for(String argument: args) {
			isValid = false;
			
			if(argument.startsWith(OPTION_TOKEN)) {
				for(String option: OPTION_LIST) {
					String[] array = argument.substring(OPTION_TOKEN.length()).split(IS_TOKEN, 1);
					if(array[0].matches(option)) {
						String value;
						if(array.length==1) {
							value = TRUE_TOKEN;
						}
						else {
							value = array[1];
						}

						options.put(option, value);
						isValid = true;
						break;
					}
				}
				
				if(!isValid) {
					_exit(9, argument);
				}
			}
			else {
				break;
			}
		}

		return options;
	}

	/**
	 * Parses the parameter from the input arguments.
	 * @param args the arguments array
	 * @return a list of parameter
	 */
	private static List<String> parseParameter(String[] args) {
		List<String> parameter = new LinkedList<>();

		for(String argument: args) {
			parameter.add(argument);
		}
		return parameter;
	}
	
	/**
	 * Checks whether the directory indicated by the file {@code directory} is existing. If not
	 * the specified directory is created. If the creation fails, Raven is exited with an error 
	 * indicated by {@code errorType}
	 * @param directory the directory to be checked
	 * @param errorType the error type to returned if Raven is exited due to an error
	 */
	private static void checkAndCreateDirectory(String directory, int errorType) {
		if(directory==null || errorType<1) {
			_exit(EXIT_CODE_UNEXPECTED_ERROR);
		}
		
		File directoryFile = new File(directory);
		if(!directoryFile.exists()) {
			if(!directoryFile.mkdir()) {
				_exit(errorType, directory);
			}
		}
	}
	
	/**
	 * Exits Raven and returns it's status.
	 * @param status a value indicating the exit status
	 */
	private static void _exit(int status, String... variables) {
		if(status!=EXIT_CODE_NORMAL_TERMINATION) {
			System.out.println(ConsoleMessages.getMessages().getString(
					ConsoleMessages.EXIT_STATUS_KEY[status-1], variables));
		}
		System.exit(status);
	}

}
