/*
 * Copyright (c) 2009-2014, Christoph Blume
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution.
 * 3. Neither the name of the "Theoretical Computer Science, University of
 * Duisburg-Essen, Germany" nor the names of its contributors may be used to 
 * endorse or promote products derived from this software without specific 
 * prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 */

package de.uni_due.inf.ti.raven.main;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.DigestInputStream;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.Collections;
import java.util.logging.Logger;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import java.util.zip.ZipOutputStream;

import javax.xml.XMLConstants;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.validation.SchemaFactory;

import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 * This class provides an update mechanism for Raven. The updater checks whether 
 * an update is available and if so, tries to download and to install the new 
 * update.
 * @author Christoph Blume
 * @version $Revision: 1.9 $
 *
 */
public final class RavenUpdater {

	// Status Codes for the availability check and the update download

	/** Status code if a new update is available */
	public static final int STATUS_CODE_UPDATE_AVAILABLE = 0;
	
	/** Status code if no update is available */
	public static final int STATUS_CODE_NO_UPDATE_AVAILABLE = 1;
	
	/** Status code if the schema file has not been found */
	public static final int STATUS_CODE_UPDATE_SCHEMA_FILE_NOT_FOUND = 2;
	
	/** Status code if the schema file has been corrupted */
	public static final int STATUS_CODE_UPDATE_SCHEMA_FILE_CORRUPTED = 3;
	
	/** Status code if the building of the parser failed */
	public static final int STATUS_CODE_BUILDING_PARSER_FAILED = 4;
	
	/** Status code if the download location is invalid/malformed */
	public static final int STATUS_CODE_UPDATE_LOCATION_INVALID = 5;
	
	/** Status code if the download URL was not valid */
	public static final int STATUS_CODE_IOERROR_OCCURRED = 6;

	/** Status code if the download was successful */
	public static final int STATUS_CODE_SUCCESS = 7;
	
	/** Status code if the download was corrupted */
	public static final int STATUS_CODE_DOWNLOAD_CORRUPTED = 8;
	
	/** Status code if the update files could not be found */
	public static final int STATUS_CODE_UPDATE_FILES_NOT_FOUND = 9;
	
	/** Status code if the download failed for an unknown reason */
	public static final int STATUS_CODE_FAILED_FOR_UNKNOWN_REASON = 10;

	private static final String TI_PREFIX = "ti";
	private static final String MD_ALGORITHM = "SHA";
	private static final int BUFFER_SIZE = 1024;

	private static final String LOCATION = "location";
	private static final String VERSION = "version";
	private static final String MAJOR_VERSION = "majorVersion";
	private static final String MINOR_VERSION = "minorVersion";
	private static final String CHECKSUM = "checksum";
	private static final String INFORMATION = "information";

	private static DocumentBuilderFactory factory = null;
	private static Document xmlDocument = null;
	
	private static URL updateLocation = null;
	private static byte[] updateChecksum = null;
	
	private static int majorVersion = -1;
	private static int minorVersion = -1;
	private static String updateInformation = null;

	private static Logger logger = Logger.getLogger(RavenUpdater.class.getName());

	/*------------------------------------------------------------------------------------------*/
	/*		Public Static Methods																*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Checks whether updates for Raven are available.
	 * @return {@code true} if updates are available, {@code false} otherwise 
	 */
	public static synchronized int updateAvailable() {
		SchemaFactory schemaFactory = null;
		DocumentBuilder builder = null;
		Node root = null;
		Node node = null;
		NodeList list = null;
		NamedNodeMap attr = null;

		//Create XML document builder
		factory = DocumentBuilderFactory.newInstance();
		factory.setNamespaceAware(true);
		factory.setIgnoringElementContentWhitespace(true);
		schemaFactory = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);
		
		//Check whether the update schema file exists, if not return with an error code
		File updateSchemaFile = new File(RavenMain.RAVEN_UPDATE_SCHEMA_FILENAME);
		if(!updateSchemaFile.exists()) {
			return STATUS_CODE_UPDATE_SCHEMA_FILE_NOT_FOUND;
		}
		
		//else try to set the schema for the factory
		try {
			factory.setSchema(schemaFactory.newSchema(updateSchemaFile));
		} catch (SAXException e) {
			logger.throwing(RavenUpdater.class.getName(), "updateAvailable", e);
			return STATUS_CODE_UPDATE_SCHEMA_FILE_CORRUPTED;
		}

		try {
			builder = factory.newDocumentBuilder();
			builder.setErrorHandler(new ErrorHandler() {
				
				@Override
				public void warning(SAXParseException e) throws SAXException {
					logger.throwing(RavenUpdater.class.getName(), "updateAvailable", e);
				}
				
				@Override
				public void fatalError(SAXParseException e) throws SAXException {
					logger.throwing(RavenUpdater.class.getName(), "updateAvailable", e);
				}
				
				@Override
				public void error(SAXParseException e) throws SAXException {
					logger.throwing(RavenUpdater.class.getName(), "updateAvailable", e);
				}
			});
		} catch (ParserConfigurationException e) {
			logger.throwing(RavenUpdater.class.getName(), "updateAvailable", e);
			return STATUS_CODE_BUILDING_PARSER_FAILED;
		}

		//Try all update locations (until an update is available)
		for(String urlString: RavenMain.RAVEN_UPDATE_URLS) {
			//Parse document and get root node
			urlString = urlString + RavenMain.RAVEN_UPDATE_FILE;
			try {
				xmlDocument = builder.parse(new BufferedInputStream(new URL(urlString).openStream()));
			}
			catch (FileNotFoundException e) {
				//Skip file since it was not found
				continue;
			} catch (MalformedURLException e) {
				//This exception should not occur since the URLs for the update resources are hard-coded
				logger.throwing(RavenUpdater.class.getName(), "updateAvailable", e);
				continue;
			} catch (SAXException e) {
				logger.throwing(RavenUpdater.class.getName(), "updateAvailable", e);
				continue;
			} catch (IOException e) {
				logger.throwing(RavenUpdater.class.getName(), "updateAvailable", e);
				continue;
			}
			root = xmlDocument.getFirstChild();
			attr = root.getAttributes();

			list = root.getChildNodes();
			for (int i = 0; i < list.getLength(); i++) {
				node = list.item(i);
				if (node.getNodeName().equals(TI_PREFIX + ":" + VERSION)) {
					attr = node.getAttributes();

					majorVersion = Integer.parseInt(attr.getNamedItem(MAJOR_VERSION).getTextContent());
					minorVersion = Integer.parseInt(attr.getNamedItem(MINOR_VERSION).getTextContent());
				} else if (node.getNodeName().equals(TI_PREFIX + ":" + LOCATION)) {
					try {
						updateLocation = new URL(node.getTextContent());
					} catch (MalformedURLException e) {
						logger.throwing(RavenUpdater.class.getName(), "updateAvailable", e);
						return STATUS_CODE_UPDATE_LOCATION_INVALID;
					}
				} else if (node.getNodeName().equals(TI_PREFIX + ":" + CHECKSUM)) {
					updateChecksum = hexToByteArray(node.getTextContent());
				} else if (node.getNodeName().equals(TI_PREFIX + ":" + INFORMATION)) {
					updateInformation = node.getTextContent();
				}
			}

			System.out.println(urlString);
			if(compareVersion(majorVersion, minorVersion)>0) {
				//If an update is available stop further update requests
				return STATUS_CODE_UPDATE_AVAILABLE;
			}

			updateLocation = null;
			updateChecksum = null;
		}

		return STATUS_CODE_NO_UPDATE_AVAILABLE;
	}

	/**
	 * Returns the major version of the last found update. 
	 * <em>Please note:</em> First use {@code updateAvailable} to check for updates
	 * before using this method
	 * @return the major version of the found update or {@code -1} if no update was found
	 */
	public static synchronized int getMajorVersion() {
		return majorVersion;
	}
	
	/**
	 * Returns the minor version of the last found update. 
	 * <em>Please note:</em> First use {@code updateAvailable} to check for updates
	 * before using this method
	 * @return the minor version of the found update or {@code -1} if no update was found
	 */
	public static synchronized int getMinorVersion() {
		return minorVersion;
	}
	
	/**
	 * Returns the update information of the last found update. 
	 * <em>Please note:</em> First use {@code updateAvailable} to check for updates
	 * before using this method
	 * @return the update information of the found update or {@code null} if no update was found
	 */
	public static synchronized String getUpdateInformation() {
		return updateInformation;
	}
	
	/**
	 * 
	 * @return {@code true} if the download was successful, {@code false} otherwise
	 */
	public static synchronized int downloadUpdate() {
		int status;
		DigestInputStream checkedIn = null;
		OutputStream out = null;
		File tempFile = null;
		ZipFile tempZipFile = null;
		byte[] chunk = null;
		byte[] digestValue = null;
		int chunkSize;
		MessageDigest digest = null;

		try {
			tempFile = File.createTempFile("raven", ".tmp");
			digest = MessageDigest.getInstance(MD_ALGORITHM);
			checkedIn = new DigestInputStream(new BufferedInputStream(
					updateLocation.openStream()), digest);
			out = new BufferedOutputStream(new FileOutputStream(tempFile));
			chunk = new byte[BUFFER_SIZE];

			//Try to download the given file
			try {
				//Write the content in a temporary file and compute the checksum
				while((chunkSize = checkedIn.read(chunk))!=-1) {
					out.write(chunk, 0, chunkSize);
				}
			} catch(Exception e) {
				e.printStackTrace();
			} finally {
				//Close all streams
				checkedIn.close();
				out.flush();
				out.close();
			}

			//If the checksum is correct
			digestValue = digest.digest();
			if(MessageDigest.isEqual(updateChecksum, digestValue)) {
				tempZipFile = new ZipFile(tempFile);

				for (ZipEntry entry : Collections.list(tempZipFile.entries())) {
					unzipEntry(tempZipFile, entry, RavenMain.MAIN_DIR);
				}
				tempZipFile.close();

				//Show success message
				status = STATUS_CODE_SUCCESS;
			}
			else {
				status = STATUS_CODE_DOWNLOAD_CORRUPTED;
			}
		} catch (FileNotFoundException e) {
			//Write stack trace to the error-log
			logger.throwing(RavenUpdater.class.getName(), "downloadFile", e);
			status = STATUS_CODE_UPDATE_FILES_NOT_FOUND;
		} catch (IOException e) {
			//Write stack trace to the error-log
			logger.throwing(RavenUpdater.class.getName(), "downloadFile", e);
			status = STATUS_CODE_IOERROR_OCCURRED;
		} catch (NoSuchAlgorithmException e) {
			//NOTE: This exception should not be thrown as long 
			//		as 'standard' hash algorithm are used!
			
			//Write stack trace to the error-log
			logger.throwing(RavenUpdater.class.getName(), "downloadFile", e);
			status = STATUS_CODE_FAILED_FOR_UNKNOWN_REASON;
		}
		finally {
			//Delete the temporary file
			if(!tempFile.delete()) {
				tempFile.deleteOnExit();							
			}
			updateLocation = null;
			updateChecksum = null;
		}

		return status;
	}

	/*------------------------------------------------------------------------------------------*/
	/*		Private Static Methods																*/
	/*------------------------------------------------------------------------------------------*/

	/**
	 * Unzips from the file {@code zipFile} the entry {@code entry} to the destination 
	 * given by the string {@code destination}.
	 * @param zipFile the zip file from the entry is extracted
	 * @param entry the zip entry to be extracted 
	 * @param destination the destination of the extraction
	 * @throws IOException thrown if an I/O exception occurs
	 * 
	 * @see ZipFile
	 * @see ZipEntry
	 * @see ZipOutputStream
	 */
	private static void unzipEntry(ZipFile zipFile, ZipEntry entry, String destination)
			throws IOException {
		File file = new File(destination, entry.getName());
		byte[] chunk = null;
		int chunkSize;

		if(entry.isDirectory()) {
			file.mkdirs();
		}
		else {
			new File(file.getParent()).mkdirs();

			try (InputStream in = zipFile.getInputStream(entry);
					OutputStream out = new FileOutputStream(file)) {
				chunk = new byte[BUFFER_SIZE];

				while((chunkSize = in.read(chunk))!=-1) {
					out.write(chunk, 0, chunkSize);
				}
			}
		}
	}

	/**
	 * Converts a hexadecimal String into a byte array.
	 * @param hexString the hexadecimal string to be converted
	 * @return the corresponding byte array.
	 */
	private static byte[] hexToByteArray(String hexString) { 
		int length = 0;
		byte[] data = null; 

		length = hexString.length();
		data = new byte[length/2];
		for(int i=0;i<length;i+=2) { 
			data[i/2] = (byte) ((Character.digit(hexString.charAt(i), 16) << 4) 
					+ Character.digit(hexString.charAt(i+1), 16)); 
		}
		return data; 
	}

	/**
	 * Compares this version with the version given version numbers. 
	 * @param majorVersion the major version number
	 * @param minorVersion the minor version number
	 * @return Returns {@code 1} if the this version is more up-to-date, 
	 * {@code 0} if both version are equal and {@code -1} otherwise.
	 */
	private static int compareVersion(int majorVersion, int minorVersion) {
		int value = 0;

		//Check whether the major version of this version is smaller
		if(Version.getMajorVersion()<majorVersion) {
			value = 1;
		}
		//Check whether both major version are equal
		else if(Version.getMajorVersion()==majorVersion) {
			//compare both minor version
			//if minor version of the version is smaller
			if(Version.getMinorVersion()<minorVersion) {
				value = 1;
			}
			//if both minor version are equal
			else if(Version.getMinorVersion()==minorVersion) {
				value = 0;
			}
			//if this version has a greater minor version
			else {
				value = -1;
			}
		}
		//if this version has a greater major version
		else {
			value = -1;
		}

		return value;
	}

	/*------------------------------------------------------------------------------------------*/
	/* 		Constructors 																		*/
	/*------------------------------------------------------------------------------------------*/

	private RavenUpdater() {}

}
